/*-----------------------------------------------------------------------
 * File: CERT_ENC.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/*
 * This file contains the high level routines necessary to encode 
 * one or more X509v3 certificates
 */ 


#include "ber_der.h"
#include "clutil.h"
#include "internal.h"
#include "oidsx509.h" 
#include "clderll.h"
#include <cssmMemTrack.h>

/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_DerEncodeAlgorithmId
 *
 * Description:
 * This function packs the algorithm structure into a DER encoded memory blob.
 *
 * Parameters: 
 * algorithm (input) : The algorithm structure to be DER encoded
 *
 * results_ptr (output): The DER encoded algorithm structure
 * 
 * NOTE: This will DER-encode an algorithmId structure even if the OID is NULL.
 *       It is strictly syntactic in nature.  It is done this way to allow for 
 *       the DER-encoding of unsigned certs and subjectPublicKeyInfo structs where the 
 *       key algorithm is unknown.
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_DerEncodeAlgorithmId(CSSM_CLUTIL_CONTEXT *pContext,
                                                  CSSM_HANDLE AppHandle, 
                                                  const CSSM_X509_ALGORITHM_IDENTIFIER *algorithm,
                                                  CSSM_DATA_PTR *results_ptr)
{
    
    uint32                  num_types = 2;
    CL_DER_CHILD_INFO       algorithm_id_struct_array[2];
    BER_PARSED_ITEM         temp_ber_item;
    CL_DER_TAG_TYPE         temp_param_tag;
    CL_DER_NODE_HANDLE      root = 0;
    CSSM_RETURN             ret = CSSM_OK;

    if (!results_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *results_ptr = NULL;
    /* Initializations */
    port_memset(algorithm_id_struct_array, 0, num_types*sizeof(CL_DER_CHILD_INFO));

    /* Create the algorithm struct tree */
    ret =  
        cl_der_CreateLeafNode(pContext,DER_TAG_OID, algorithm->algorithm.Length, 
                              algorithm->algorithm.Data, &algorithm_id_struct_array[0].ChildHandle);
    if (!algorithm->parameters.Data || !algorithm->parameters.Length)
    {
        algorithm_id_struct_array[1].IsOptional = CSSM_TRUE;
        ret =  
            cl_der_CreateLeafNode(pContext,DER_TAG_NULL, algorithm->parameters.Length, algorithm->parameters.Data, &(algorithm_id_struct_array[1].ChildHandle));

        if(ret != CSSM_OK)
        {
                cl_der_DeleteLeafNode(pContext,algorithm_id_struct_array[0].ChildHandle);
                return ret; 
        }
    
    }   
    else 
    {
        BER_ExpandItem(algorithm->parameters.Data, algorithm->parameters.Length, &temp_ber_item);
        temp_param_tag    = temp_ber_item.Tag[0];
        if (temp_param_tag == DER_TAG_SEQUENCE + BER_CONSTRUCTED ||
            temp_param_tag == DER_TAG_SET      + BER_CONSTRUCTED )
            temp_param_tag -= BER_CONSTRUCTED;
        algorithm_id_struct_array[1].IsOptional = CSSM_TRUE;
        ret =  
            cl_der_CreateLeafNode(pContext,temp_param_tag, temp_ber_item.ContentLength, (uint8*)temp_ber_item.Content, &(algorithm_id_struct_array[1].ChildHandle));
 
        if(ret != CSSM_OK)
        {
                cl_der_DeleteLeafNode(pContext,algorithm_id_struct_array[0].ChildHandle);
                return ret; 
        }
    
    
    }
    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_types, algorithm_id_struct_array, &root);

    if(ret != CSSM_OK)
    {
        cl_der_DeleteTree (pContext,root);
        return ret;
    }

    /* Pack the algorithm struct into a DER encoded blob */
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, results_ptr);

    if(ret != CSSM_OK)
    {

        cl_der_DeleteTree (pContext,root);
        return ret;
    }

    /* Cleanup */
    cl_der_DeleteTree (pContext,root);

    /* Return */
 
    return ret;
}



/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_DerEncodeName
 *
 * Description:
 * This function packs an X509 distinguished name structure into 
 * a DER encoded memory blob.
 *
 * Parameters: 
 * name (input) : the distinguished name to be DER encoded
 *
 * return_results_ptr (output) : A CSSM_DATA structure which contains the DER encoded
 *                        distinguished name.
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_DerEncodeName(CSSM_CLUTIL_CONTEXT *pContext,
                                           CSSM_HANDLE AppHandle,
                                           const CSSM_X509_NAME *name,
                                           CSSM_DATA_PTR *return_results_ptr)
{
    CSSM_DATA_PTR           results_ptr = NULL;
    CL_DER_CHILD_INFO_PTR   rdn_array = NULL, attr_array = NULL;
    uint32                  num_attr_value_types = 2;
    CSSM_RETURN             ret = CSSM_OK;
    CL_DER_CHILD_INFO       attr_type_value_struct[2];
    CL_DER_NODE_HANDLE      root = 0;
    uint32 i=0, j=0, k=0;
    if (!return_results_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *return_results_ptr = NULL;
    /* Initializations */
    rdn_array = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, name->numberOfRDNs, sizeof(CL_DER_CHILD_INFO));
    if (!rdn_array)
    {
        return CSSMERR_CL_MEMORY_ERROR;        
    }
    port_memset(attr_type_value_struct, 0, num_attr_value_types * sizeof(CL_DER_CHILD_INFO));

    /* For each RDN, create its attribute array */
    for (i=0; i < name->numberOfRDNs; i++)
    {
        /* Allocate the array */
        attr_array = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 
            name->RelativeDistinguishedName[i].numberOfPairs, sizeof(CL_DER_CHILD_INFO));
        if (!attr_array)
        {
            for (k=0; k<i; k++)
            {
                cl_der_DeleteParentNode(pContext,rdn_array[k].ChildHandle);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, rdn_array);
            return CSSMERR_CL_MEMORY_ERROR;            
        }

        /* Create a leaf for each attribute type/value pair */
        for (j=0; j < name->RelativeDistinguishedName[i].numberOfPairs; j++)
        {
            ret =  
                cl_der_CreateLeafNode(pContext,DER_TAG_OID, 
                name->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Length,
                name->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Data, &(attr_type_value_struct[0].ChildHandle));
 
            if (ret != CSSM_OK)
            {
                for (k=0; k<j; k++)
                {
                    cl_der_DeleteParentNode(pContext,attr_array[k].ChildHandle);
                }
                for (k=0; k<i; k++)
                {
                    cl_der_DeleteParentNode(pContext,rdn_array[k].ChildHandle);
                }
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, attr_array);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, rdn_array);
                return ret;                  
                
            }

            if (name->RelativeDistinguishedName[i].AttributeTypeAndValue[j].valueType)
            {
                ret =  
                    cl_der_CreateLeafNode(pContext,name->RelativeDistinguishedName[i].AttributeTypeAndValue[j].valueType, 
                    name->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Length,
                    name->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Data, &(attr_type_value_struct[1].ChildHandle));
    
        
                if (ret != CSSM_OK)
                {
                    for (k=0; k<j; k++)
                    {
                        cl_der_DeleteParentNode(pContext,attr_array[k].ChildHandle);
                    }
                    for (k=0; k<i; k++)
                    {
                        cl_der_DeleteParentNode(pContext,rdn_array[k].ChildHandle);
                    }
                    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, attr_array);
                    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, rdn_array);
                    return ret;                
                }
            }
            else
            {
                ret =  
                    cl_der_CreateLeafNode(pContext,DER_TAG_PRINTABLE_STRING, 
                    name->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Length,
                    name->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Data, &(attr_type_value_struct[1].ChildHandle));

                if (ret != CSSM_OK)
                {
                    for (k=0; k<j; k++)
                    {
                        cl_der_DeleteParentNode(pContext,attr_array[k].ChildHandle);
                    }
                    for (k=0; k<i; k++)
                    {
                        cl_der_DeleteParentNode(pContext,rdn_array[k].ChildHandle);
                    }
                    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, attr_array);
                    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, rdn_array);
                    return ret;                
                }
            }
            ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_attr_value_types, attr_type_value_struct, &attr_array[j].ChildHandle);

            if (ret != CSSM_OK)
            {
                for (k=0; k<j; k++)
                {
                    cl_der_DeleteParentNode(pContext,attr_array[k].ChildHandle);
                }
                for (k=0; k<i; k++)
                {
                    cl_der_DeleteParentNode(pContext,rdn_array[k].ChildHandle);
                }
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, attr_array);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, rdn_array);
                return ret;                
            }

        }

        /* Create the parent node for this RDN */
        ret = cl_der_CreateParentNode(pContext,DER_TAG_SET, name->RelativeDistinguishedName[i].numberOfPairs, attr_array, &rdn_array[i].ChildHandle);
 
        if (ret != CSSM_OK)
        {
            for (k=0; k<j; k++)
            {
                cl_der_DeleteParentNode(pContext,attr_array[k].ChildHandle);
            }
            for (k=0; k<i; k++)
            {
                cl_der_DeleteParentNode(pContext,rdn_array[k].ChildHandle);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, attr_array);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, rdn_array);
            return ret;                       
        }

        /* Free the attribute array */
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, attr_array);
    }

    /* Create the parent node for this name struct */
    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, name->numberOfRDNs, rdn_array, &root);

    if (ret != CSSM_OK)
    {
        for (k=0; k<j; k++)
        {
            cl_der_DeleteParentNode(pContext,attr_array[k].ChildHandle);
        }
        for (k=0; k<i; k++)
        {
            cl_der_DeleteParentNode(pContext,rdn_array[k].ChildHandle);
        }
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, attr_array);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, rdn_array);
        return ret;                       
    }
    
    
    
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, rdn_array);

    /* Pack the key info struct into a DER encoded blob */
    
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, &results_ptr);

    
    if (ret != CSSM_OK)
    {
        for (k=0; k<j; k++)
        {
            cl_der_DeleteParentNode(pContext,attr_array[k].ChildHandle);
        }
        for (k=0; k<i; k++)
        {
            cl_der_DeleteParentNode(pContext,rdn_array[k].ChildHandle);
        }
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, attr_array);
        return ret;                       
    }

    /* Delete the parse tree */
    cl_der_DeleteTree (pContext,root);

    /* Return */
   
    *return_results_ptr = results_ptr;
    return ret;
}



/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_DerEncodeSPKI
 *
 * Description:
 * This function packs the certificate's subject key into 
 * a DER encoded memory blob.
 *
 * Parameters: 
 * key (input)     : the key info to be DER encoded
 *
 * results_ptr (output): A CSSM_DATA structure which contains the DER encoded key blob.
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_DerEncodeSPKI(CSSM_CLUTIL_CONTEXT *pContext,
                                           CSSM_HANDLE AppHandle, 
                                           const CSSM_X509_SUBJECT_PUBLIC_KEY_INFO *key,
                                           CSSM_DATA_PTR *results_ptr)
{
    CSSM_RETURN         ret = CSSM_OK;
    uint32              num_types = 2;
    CL_DER_CHILD_INFO   key_info_struct[2];
    CSSM_DATA_PTR       enc_alg_ptr = NULL;
    CL_DER_NODE_HANDLE  root = 0;

    if (!results_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *results_ptr = NULL;
    /* Initializations */
    port_memset(key_info_struct, 0, num_types*sizeof(CL_DER_CHILD_INFO));
    
    ret = CSSM_CLUTIL_X509_DerEncodeAlgorithmId(pContext,AppHandle, &key->algorithm, &enc_alg_ptr);
    if (ret != CSSM_OK)
    {
        return ret;
    }

    /* Create the key info tree */

    ret =  
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 
        cl_der_LengthOfContent(enc_alg_ptr),
        &enc_alg_ptr->Data[cl_der_LengthOfTag(enc_alg_ptr) + cl_der_LengthOfLength(enc_alg_ptr)], &(key_info_struct[0].ChildHandle));

    if (ret != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_alg_ptr->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_alg_ptr);
        return ret;        
    }

    ret =  
        cl_der_CreateLeafNode(pContext,DER_TAG_BIT_STRING, 
        8*key->subjectPublicKey.Length,
        key->subjectPublicKey.Data, &(key_info_struct[1].ChildHandle));

    if (ret != CSSM_OK)
    {
        cl_der_DeleteLeafNode(pContext,key_info_struct[0].ChildHandle);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_alg_ptr->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_alg_ptr);
        return ret;        
    }


    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_alg_ptr->Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_alg_ptr);
 
    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_types, key_info_struct, &root);
 
    if( ret != CSSM_OK)
    {
        cl_der_DeleteLeafNode(pContext,key_info_struct[0].ChildHandle);
        cl_der_DeleteLeafNode(pContext,key_info_struct[1].ChildHandle);
        cl_der_DeleteTree (pContext,root);
        return ret;  
    }
    /* Pack the key info struct into a DER encoded blob */
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, results_ptr);

    if(ret != CSSM_OK)
    {
        cl_der_DeleteTree (pContext,root);
        return ret;
    }

    /* Delete the parse tree */
    cl_der_DeleteTree (pContext,root);
    /* Return */
   
    return ret;
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_DerEncodeExtension
 *
 * Description:
 * This function packs one Intel X509V3 certificate extension into 
 * a DER encoded memory blob.
 *
 * Parameters: 
 * Extension (input)     : the extension to be DER encoded
 *
 * results_ptr (output)  :A CSSM_DATA structure which contains the DER encoded extension.
  *---------------------------------------------------------------------------*/
static  CSSM_DATA null_data_s;
CSSM_RETURN cssm_clutil_x509_DerEncodeExtension(const CSSM_CLUTIL_CONTEXT *pContext,
                                                CSSM_HANDLE AppHandle,
                                                const CSSM_X509_EXTENSION *Extension,
                                                CSSM_DATA_PTR *results_ptr)
{
    CSSM_DATA_PTR temp_extnValue = NULL;
    CSSM_X509EXT_PAIR local_extnValue;

    uint32 numBytes[2];
    CSSM_RETURN ret = CSSM_OK;
    null_data_s.Length = 0;
    null_data_s.Data = NULL;

    if (!results_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *results_ptr = NULL;
    local_extnValue;  
    numBytes;


    if (Extension->value.tagAndValue && (Extension->value.tagAndValue)->value.Data && (Extension->value.tagAndValue)->value.Length)
    {
        ret = cl_der_EncodeObject (pContext, AppHandle, (Extension->value.tagAndValue)->type, 
                                                &((Extension->value.tagAndValue)->value), &temp_extnValue);

    } 
    else 
    {
        temp_extnValue = &null_data_s;
    }
   *results_ptr = temp_extnValue;
   return ret;
}

CSSM_RETURN CSSM_CLUTIL_X509_DerEncodeExtension(
                                        CSSM_CLUTIL_CONTEXT *pContext,
                                        CSSM_HANDLE AppHandle, 
                                        const CSSM_X509_EXTENSION *Extension,
                                        CSSM_DATA_PTR *results_ptr)
{
    
    uint32              num_types = 3;
    CL_DER_CHILD_INFO   extension_array[3];
    uint8               critical_dflt;
    CSSM_DATA           temp_critical_dflt;
    CSSM_DATA_PTR       temp_extnValue;
    CL_DER_NODE_HANDLE  root;
    CSSM_DATA           null_data;
    CSSM_RETURN         ret = CSSM_OK;
   
    if (!results_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    null_data.Length = 0;
    null_data.Data = NULL;
    *results_ptr = NULL;
    /* Initializations */
    port_memset(extension_array, 0, num_types*sizeof(CL_DER_CHILD_INFO));
    critical_dflt = 0x00;
    temp_critical_dflt.Length = 1;
    temp_critical_dflt.Data   = &critical_dflt;

    /* Code */
    ret = cssm_clutil_x509_DerEncodeExtension(pContext,AppHandle, Extension, &temp_extnValue);
    if (ret != CSSM_OK)
    {
       return ret;
    }

    /* Create the extension tree */
    ret =  
        cl_der_CreateLeafNode(pContext,DER_TAG_OID, 
        Extension->extnId.Length, Extension->extnId.Data, &(extension_array[0].ChildHandle));
    extension_array[1].HasDefault = CSSM_TRUE;
    extension_array[1].DefaultValue = temp_critical_dflt;

    if(ret != CSSM_OK)
    {
         return ret;
    }
    
    ret =  
        cl_der_CreateLeafNode(pContext,DER_TAG_BOOLEAN, 
        sizeof(CSSM_BOOL), (uint8 *) &Extension->critical, &(extension_array[1].ChildHandle));
 
    
    if (ret != CSSM_OK)
    {
         cl_der_DeleteLeafNode(pContext,extension_array[0].ChildHandle);
         return ret;
    }
    ret =  
         cl_der_CreateLeafNode(pContext,DER_TAG_OCTET_STRING, 
         temp_extnValue->Length, temp_extnValue->Data, &(extension_array[2].ChildHandle));
    
    if (ret != CSSM_OK)
    {
         cl_der_DeleteLeafNode(pContext,extension_array[0].ChildHandle);
         cl_der_DeleteLeafNode(pContext,extension_array[1].ChildHandle);
         return ret;
    }


    if (temp_extnValue->Data)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, temp_extnValue->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, temp_extnValue);
    }


    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_types, extension_array, &root);

    if (ret != CSSM_OK)
    {
        cl_der_DeleteLeafNode(pContext,extension_array[0].ChildHandle);
        cl_der_DeleteLeafNode(pContext,extension_array[1].ChildHandle);
        cl_der_DeleteLeafNode(pContext,extension_array[2].ChildHandle);
        cl_der_DeleteTree (pContext,root);
        return ret;       
    }

    /* Pack the signature struct into a DER encoded blob */
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, results_ptr);
    if (ret != CSSM_OK)
    {
         cl_der_DeleteTree (pContext,root);
         return ret;
    }

    /* Delete the parse tree */
    cl_der_DeleteTree (pContext,root);
        

    /* Return  */
   
    return ret;
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_DerEncodeExtensions
 *
 * Description:
 * This function packs the Intel X509V3 certificate extensions array 
 * into a DER encoded memory blob.
 *
 * Parameters: 
 * Extensions (input)     : the array of extensions to be DER encoded
 * NumExtensions (input)  : the number of extensions in the array
 *
 * results_ptr (output): The DER encoded extensions
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_DerEncodeExtensions(CSSM_CLUTIL_CONTEXT *pContext,
                                                 CSSM_HANDLE AppHandle, 
                                                 const CSSM_X509_EXTENSION *Extensions, 
                                                 uint32 NumExtensions,
                                                 CSSM_DATA_PTR *results_ptr)
{
    CSSM_RETURN             ret = CSSM_OK;
    CL_DER_CHILD_INFO_PTR   extensions_array = NULL;
    CSSM_DATA_PTR           enc_ext_ptr = NULL;
    CL_DER_NODE_HANDLE      root = 0;
    uint32 i=0, j=0;

    if (!results_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *results_ptr = NULL;
    /* Initializations */
    extensions_array = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 
        NumExtensions, sizeof(CL_DER_CHILD_INFO) );
    if (!extensions_array)
    {
        return CSSMERR_CL_MEMORY_ERROR;        
    }

    /* For each extension */
    for (i=0; i < NumExtensions; i++)
    {
        /* Der encode the extension */
        ret = CSSM_CLUTIL_X509_DerEncodeExtension(pContext,AppHandle, &Extensions[i], &enc_ext_ptr);

        if (!enc_ext_ptr)
        {
            for (j=0; j<i; j++)
            {
                cl_der_DeleteLeafNode(pContext,extensions_array[j].ChildHandle);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, extensions_array);
            return ret;
        }

        /* Create a leaf node for it */
        ret =
            cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 
            cl_der_LengthOfContent(enc_ext_ptr),
            &enc_ext_ptr->Data[cl_der_LengthOfTag(enc_ext_ptr) + cl_der_LengthOfLength(enc_ext_ptr)], &(extensions_array[i].ChildHandle));
                                    

        /* Delete the DER encoded extension */
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_ext_ptr->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_ext_ptr);

        if (!extensions_array[i].ChildHandle)
        {
            for(j=0; j<i; j++)
            {
                cl_der_DeleteLeafNode(pContext,extensions_array[j].ChildHandle);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, extensions_array);
            return CSSMERR_CL_INTERNAL_ERROR;
        }
    }

    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, NumExtensions, extensions_array, &root);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, extensions_array);

    /* Pack the signatures into a DER encoded blob */
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, results_ptr);
    
    if(ret != CSSM_OK)
    {
         cl_der_DeleteTree (pContext,root);
         return ret;
    }
    /* Delete the parse tree */
    cl_der_DeleteTree (pContext,root);

    /* Return */

    return ret;

}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_DerEncodeTbsCertificate
 *
 * Description:
 * This function packs the ToBeSigned portion of an Intel X509V3 signed certificate 
 * into a DER encoded memory blob.
 *
 * Parameters: 
 * cert509 (input)         : The ToBeSigned portion of a certificate to be DER encoded
 *
 * results_ptr (output)    : The DER encoded ToBeSigned portion of an X509v3 certificate
 * 
 *---------------------------------------------------------------------------*/
typedef struct parameters 
{
    CSSM_DATA_PTR   enc_extensions_ptr;
    CSSM_DATA_PTR   enc_key_info_ptr;
    CSSM_DATA_PTR   enc_subjectName;
    CSSM_DATA_PTR   enc_issuerName;
    CSSM_DATA_PTR   enc_signature;
} PARAMETERS;

void CleanUpTBSCertificate( CSSM_CLUTIL_CONTEXT *pContext,
                            CSSM_HANDLE AppHandle,
                            sint32 num, 
                            CL_DER_CHILD_INFO *cert_ptr, 
                            PARAMETERS *params_ptr, 
                            CL_DER_CHILD_INFO *validity_ptr)
{

    CSSM_RETURN retVal = CSSM_OK;
    if (!params_ptr)
    {
        return;
    }
    if(params_ptr->enc_extensions_ptr)
    {
        if(params_ptr->enc_extensions_ptr->Data)
        {
           CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, params_ptr->enc_extensions_ptr->Data);
        }

       CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, params_ptr->enc_extensions_ptr);
    }
 
    if(params_ptr->enc_key_info_ptr)
    {
        if(params_ptr->enc_key_info_ptr->Data)
        {
           CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, params_ptr->enc_key_info_ptr->Data);
        }

       CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, params_ptr->enc_key_info_ptr);
    }
    
    if(params_ptr->enc_subjectName)
    {
        if(params_ptr->enc_subjectName->Data)
        {
           CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, params_ptr->enc_subjectName->Data);
        }

       CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, params_ptr->enc_subjectName);
    }
 
    if(params_ptr->enc_issuerName)
    {
        if(params_ptr->enc_issuerName->Data)
        {
           CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, params_ptr->enc_issuerName->Data);
        }

       CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, params_ptr->enc_issuerName);
    }

    if(params_ptr->enc_signature)
    {
        if(params_ptr->enc_signature->Data)
        {
           CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, params_ptr->enc_signature->Data);
        }

       CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, params_ptr->enc_signature);
    }

    if (num >= 0)
    {
        int i;
        for (i=0; i<=num; i++) 
        {
            retVal =cl_der_DeleteLeafNode(pContext,cert_ptr[i].ChildHandle);
        }
    }
    if (!validity_ptr)
    {
        return;
    }
    if (validity_ptr[0].ChildHandle != 0)
    {
        cl_der_DeleteLeafNode(pContext,validity_ptr[0].ChildHandle);
    }
    if (validity_ptr[1].ChildHandle != 0)
    {
        cl_der_DeleteLeafNode(pContext,validity_ptr[1].ChildHandle);        
    }
    
}

CSSM_RETURN CSSM_CLUTIL_X509_DerEncodeTbsCertificate (CSSM_CLUTIL_CONTEXT *pContext,
                                                      CSSM_HANDLE AppHandle, 
                                                      const CSSM_X509_TBS_CERTIFICATE *cert509,
                                                      CSSM_DATA_PTR *results_ptr)
{
    
    uint32              num_types = 10;
    CL_DER_CHILD_INFO   cert509_struct_array[10];
    uint8               version_dflt;
    CSSM_DATA           temp_version;
    uint32              num_validity_types = 2;
    CL_DER_CHILD_INFO   validity_struct_array[2];
    uint32              notBefore_length, notAfter_length;

    CL_DER_NODE_HANDLE  root;
    uint32              i;
    CSSM_RETURN         ret = CSSM_OK;
    PARAMETERS TBSCertificateParams;

    if (!results_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }    
    *results_ptr = NULL;
    port_memset(cert509_struct_array, 0, 10 * sizeof(CL_DER_CHILD_INFO));
    port_memset(validity_struct_array, 0, 2 * sizeof(CL_DER_CHILD_INFO));
    port_memset(&TBSCertificateParams, 0, sizeof(PARAMETERS)); 


   
    /* Initializations */
    port_memset(cert509_struct_array, 0, num_types*sizeof(CL_DER_CHILD_INFO));
    port_memset(validity_struct_array, 0, num_validity_types*sizeof(CL_DER_CHILD_INFO));
    version_dflt = 0x00;
    temp_version.Length = 1;
    temp_version.Data   = &version_dflt;

    if ( cert509->validity.notBefore.time.Data )
    {
         notBefore_length = strlen((char *)cert509->validity.notBefore.time.Data) + 1;
    }
    else 
    {
        notBefore_length = 0;
    }
    if ( cert509->validity.notAfter.time.Data )
    {
        notAfter_length = strlen((char *)cert509->validity.notAfter.time.Data) + 1;
    }
    else 
    {
        notAfter_length = 0;
    }

    /* Create the DER encoded structs */
    if (notBefore_length <= UTC_TIME_STRLEN +1)
    {
        ret =  
            cl_der_CreateLeafNode(pContext,DER_TAG_UTC_TIME, 
            notBefore_length, (uint8*) cert509->validity.notBefore.time.Data, &validity_struct_array[0].ChildHandle);
    }
    else
    {
        ret =  
            cl_der_CreateLeafNode(pContext,DER_TAG_GENERALIZED_TIME, 
            notBefore_length, (uint8*) cert509->validity.notBefore.time.Data, &validity_struct_array[0].ChildHandle);
    }
    if (ret != CSSM_OK)
    {
        CleanUpTBSCertificate(pContext, AppHandle, -1, cert509_struct_array, 
                                &TBSCertificateParams, validity_struct_array); 
        return ret;
    }

    
    
    if (notAfter_length <= UTC_TIME_STRLEN +1)
    {
        ret =  
            cl_der_CreateLeafNode(pContext,DER_TAG_UTC_TIME, 
            notAfter_length, (uint8*) cert509->validity.notAfter.time.Data, &validity_struct_array[1].ChildHandle);
    }
    else 
    {
        ret =  
            cl_der_CreateLeafNode(pContext,DER_TAG_GENERALIZED_TIME, 
            notAfter_length, (uint8*) cert509->validity.notAfter.time.Data, &validity_struct_array[1].ChildHandle);
    }
    if (ret != CSSM_OK)
    {       
        CleanUpTBSCertificate(pContext, AppHandle, -1, cert509_struct_array, 
                                &TBSCertificateParams, validity_struct_array); 
        return ret;
    }

    /*Encode the Algorithm back to Der notation */
    ret   = CSSM_CLUTIL_X509_DerEncodeAlgorithmId(pContext,AppHandle, &cert509->signature, &TBSCertificateParams.enc_signature);
    
    if (ret != CSSM_OK)
    {
        CleanUpTBSCertificate(pContext, AppHandle, -1, cert509_struct_array, 
                                &TBSCertificateParams, validity_struct_array); 
        return ret;
    }

    ret  = CSSM_CLUTIL_X509_DerEncodeName(pContext, AppHandle, &cert509->issuer, &TBSCertificateParams.enc_issuerName); 

    if (ret != CSSM_OK)
    {        
        CleanUpTBSCertificate(pContext, AppHandle, -1, cert509_struct_array, 
                                &TBSCertificateParams, validity_struct_array); 
        return ret;
    }

    ret = CSSM_CLUTIL_X509_DerEncodeName(pContext, AppHandle, &cert509->subject, &TBSCertificateParams.enc_subjectName); 

    if (ret != CSSM_OK)
    {
        CleanUpTBSCertificate(pContext, AppHandle, -1, cert509_struct_array, 
                                &TBSCertificateParams, validity_struct_array); 
        return ret;
    }
 
    ret    = CSSM_CLUTIL_X509_DerEncodeSPKI(pContext, AppHandle, &cert509->subjectPublicKeyInfo, &TBSCertificateParams.enc_key_info_ptr);

    if (ret != CSSM_OK )
    {
        CleanUpTBSCertificate(pContext, AppHandle, -1, cert509_struct_array, 
                                &TBSCertificateParams, validity_struct_array); 
        return ret;
    }

    ret = CSSM_CLUTIL_X509_DerEncodeExtensions(pContext, AppHandle, 
                                cert509->extensions.extensions, 
                                cert509->extensions.numberOfExtensions, 
                                &TBSCertificateParams.enc_extensions_ptr);

    if (ret != CSSM_OK)
    {
        CleanUpTBSCertificate(pContext, AppHandle, -1, cert509_struct_array, 
                                &TBSCertificateParams, validity_struct_array); 
        return ret;
    }


    /* Create the DER encoded X509 structure */
    cert509_struct_array[0].HasCtxSpecificTag = CSSM_TRUE;
    cert509_struct_array[0].IsExplicit = CSSM_TRUE;
    cert509_struct_array[0].HasDefault = CSSM_TRUE;
    cert509_struct_array[0].DefaultValue = temp_version;
    i = cert509->version.Length;
 
    ret =  
        cl_der_CreateLeafNode(pContext, DER_TAG_INTEGER, i, cert509->version.Data, &cert509_struct_array[0].ChildHandle);

    if(ret != CSSM_OK)
    {
        CleanUpTBSCertificate(pContext, AppHandle, 0, cert509_struct_array, 
                                &TBSCertificateParams, validity_struct_array); 
        return ret;
    }

    i = cert509->serialNumber.Length;

    ret =  
        cl_der_CreateLeafNode(pContext, DER_TAG_INTEGER, i, cert509->serialNumber.Data, &cert509_struct_array[1].ChildHandle);

    if(ret != CSSM_OK)
    {
        CleanUpTBSCertificate(pContext, AppHandle, 1, cert509_struct_array, 
                                &TBSCertificateParams, validity_struct_array); 
        return ret;
    }

    i = cl_der_LengthOfContent(TBSCertificateParams.enc_signature);

    if (i != 0)
    {
        ret =  
            cl_der_CreateLeafNode(pContext, DER_TAG_SEQUENCE, i,
                &TBSCertificateParams.enc_signature->Data[cl_der_LengthOfTag(TBSCertificateParams.enc_signature) + cl_der_LengthOfLength(TBSCertificateParams.enc_signature)], &cert509_struct_array[2].ChildHandle);        
    }
    else
    {
        ret = cl_der_CreateLeafNode(pContext, DER_TAG_SEQUENCE, 0, NULL, &cert509_struct_array[2].ChildHandle);
    }

    if (ret != CSSM_OK  )
    {
        CleanUpTBSCertificate(pContext, AppHandle, 2, cert509_struct_array, 
                                &TBSCertificateParams, validity_struct_array); 
        return ret;
    }

    i = cl_der_LengthOfContent(TBSCertificateParams.enc_issuerName);

    if (i != 0)
    {
        ret =  
            cl_der_CreateLeafNode(pContext, DER_TAG_SEQUENCE, i,
                &TBSCertificateParams.enc_issuerName->Data[ cl_der_LengthOfTag(TBSCertificateParams.enc_issuerName) + cl_der_LengthOfLength(TBSCertificateParams.enc_issuerName)], &cert509_struct_array[3].ChildHandle);
    }
    else
    {
        ret = cl_der_CreateLeafNode(pContext, DER_TAG_SEQUENCE, 0, NULL, &cert509_struct_array[3].ChildHandle);
    }

    if (ret != CSSM_OK )
    {
        CleanUpTBSCertificate(pContext, AppHandle, 3, cert509_struct_array, 
                                &TBSCertificateParams, validity_struct_array); 

        return ret;
    }

    i = num_validity_types;

    if (i != 0) 
    {
        ret =  
            cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, i, validity_struct_array, &cert509_struct_array[4].ChildHandle);
    }
    else
    {
        ret =  
            cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, 0, NULL, &cert509_struct_array[4].ChildHandle);
    }


    if (ret != CSSM_OK)
    {
        CleanUpTBSCertificate(pContext, AppHandle, 4, cert509_struct_array, 
                                &TBSCertificateParams, validity_struct_array); 
        return ret;
    }

    i = cl_der_LengthOfContent(TBSCertificateParams.enc_subjectName);

    if (i != 0)
    {
        ret =  
            cl_der_CreateLeafNode(pContext, DER_TAG_SEQUENCE, i,
                &TBSCertificateParams.enc_subjectName->Data[ cl_der_LengthOfTag(TBSCertificateParams.enc_subjectName) 
                    + cl_der_LengthOfLength(TBSCertificateParams.enc_subjectName)], 
                    &cert509_struct_array[5].ChildHandle);
    }
    else
    {
        ret =  cl_der_CreateLeafNode(pContext, DER_TAG_SEQUENCE, 0, NULL, &cert509_struct_array[5].ChildHandle);
    }

    if (ret != CSSM_OK )
    {
        CleanUpTBSCertificate(pContext, AppHandle, 5, cert509_struct_array, 
                                &TBSCertificateParams, validity_struct_array); 
        return ret;
    }

    i = cl_der_LengthOfContent(TBSCertificateParams.enc_key_info_ptr);
    if (i != 0)
    {
        ret =  
            cl_der_CreateLeafNode(pContext, DER_TAG_SEQUENCE, i,
                &TBSCertificateParams.enc_key_info_ptr->Data[ 
                        cl_der_LengthOfTag(TBSCertificateParams.enc_key_info_ptr) 
                        + cl_der_LengthOfLength(TBSCertificateParams.enc_key_info_ptr)], 
                &cert509_struct_array[6].ChildHandle);
    }
    else
    {
        ret =  cl_der_CreateLeafNode(pContext, DER_TAG_SEQUENCE, 0, NULL, &cert509_struct_array[6].ChildHandle);
    }

    if (ret != CSSM_OK)
    {
        CleanUpTBSCertificate(pContext, AppHandle, 6, cert509_struct_array, 
                                &TBSCertificateParams, validity_struct_array); 

        return ret;
    }
 
    cert509_struct_array[7].IsOptional = CSSM_TRUE;
    cert509_struct_array[7].HasCtxSpecificTag = CSSM_TRUE;
    i = 8*cert509->issuerUniqueIdentifier.Length;

    if (i != 0)
    {
        ret =  
            cl_der_CreateLeafNode(pContext,DER_TAG_BIT_STRING, i, cert509->issuerUniqueIdentifier.Data, &cert509_struct_array[7].ChildHandle);
    }
    else
    {
        ret =  
            cl_der_CreateLeafNode(pContext,DER_TAG_BIT_STRING, 0, NULL, &cert509_struct_array[7].ChildHandle);
    }

    if (ret != CSSM_OK)
    {
        CleanUpTBSCertificate(pContext, AppHandle, 7, cert509_struct_array, 
                                &TBSCertificateParams, validity_struct_array); 
        
        return ret;
    }

    cert509_struct_array[8].IsOptional = CSSM_TRUE;
    cert509_struct_array[8].HasCtxSpecificTag = CSSM_TRUE;
    i = 8*cert509->subjectUniqueIdentifier.Length;

    if (i != 0)
    {
        ret =  
            cl_der_CreateLeafNode(pContext,DER_TAG_BIT_STRING, i, cert509->subjectUniqueIdentifier.Data, &cert509_struct_array[8].ChildHandle);
    }
    else
    {
        ret = cl_der_CreateLeafNode(pContext,DER_TAG_BIT_STRING, 0, NULL, &cert509_struct_array[8].ChildHandle);
    }

    if (ret != CSSM_OK)
    {
        CleanUpTBSCertificate(pContext, AppHandle, 8, cert509_struct_array, 
                                &TBSCertificateParams, validity_struct_array); 
        return ret;
    }
    
    cert509_struct_array[9].IsOptional = CSSM_TRUE;
    cert509_struct_array[9].HasCtxSpecificTag = CSSM_TRUE;
    cert509_struct_array[9].IsExplicit = CSSM_TRUE;
    i = cl_der_LengthOfContent(TBSCertificateParams.enc_extensions_ptr);

    if (i != 0)
    {
        ret =  
            cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, i, 
                &TBSCertificateParams.enc_extensions_ptr->Data[ 
                    cl_der_LengthOfTag(TBSCertificateParams.enc_extensions_ptr) 
                    + cl_der_LengthOfLength(TBSCertificateParams.enc_extensions_ptr)], 
                &(cert509_struct_array[9].ChildHandle));
    }
    else
    {
        ret = cl_der_CreateLeafNode(pContext, DER_TAG_SEQUENCE, 0, NULL, &(cert509_struct_array[9].ChildHandle));
    }

    if (ret != CSSM_OK )
    {
        CleanUpTBSCertificate(pContext, AppHandle, 9, cert509_struct_array, 
                                &TBSCertificateParams, validity_struct_array); 
        return ret;
    }

    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, TBSCertificateParams.enc_extensions_ptr->Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, TBSCertificateParams.enc_extensions_ptr);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, TBSCertificateParams.enc_key_info_ptr->Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, TBSCertificateParams.enc_key_info_ptr);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, TBSCertificateParams.enc_subjectName->Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, TBSCertificateParams.enc_subjectName);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, TBSCertificateParams.enc_issuerName->Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, TBSCertificateParams.enc_issuerName);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, TBSCertificateParams.enc_signature->Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, TBSCertificateParams.enc_signature);

    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_types, cert509_struct_array, &root);

    if(ret != CSSM_OK)
    {
        cl_der_DeleteTree (pContext,root);
        return ret;
    }

    /* Pack the signature struct into a DER encoded blob */
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, results_ptr);
    
    if(ret != CSSM_OK)
    {
         cl_der_DeleteTree (pContext,root);
         return ret;
    }
    /* Delete the parse tree */
    cl_der_DeleteTree (pContext,root);

    /* Return */
   
    return ret;
}



/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_DerEncodeSignature
 *
 * Description: 
 * This function packs one Intel X509V3 certificate signature into a DER encoded memory blob.
 *
 * Parameters: 
 * Signature (input)     : the signature to be DER encoded
 *
 * result_val (output)   : A CSSM_DATA structure which contains the DER encoded signature.
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_DerEncodeSignature (CSSM_CLUTIL_CONTEXT *pContext,
                                                 CSSM_HANDLE AppHandle,
                                                 const CSSM_X509_SIGNATURE *Signature,
                                                 CSSM_DATA_PTR *result_val)
{
    CSSM_DATA_PTR       results_ptr = NULL;
    uint32              num_sigstruct_types = 2;
    CL_DER_CHILD_INFO   signature_struct_array[2];
    CSSM_DATA_PTR       enc_alg_ptr = NULL;
    CL_DER_NODE_HANDLE  root = 0;
    CSSM_RETURN         ret = CSSM_OK;

    /* Initializations */
    port_memset(signature_struct_array, 0, num_sigstruct_types*sizeof(CL_DER_CHILD_INFO));

    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }    

    *result_val = NULL;

    /* Create the signatureAlgorithm sub-structure */
    ret = CSSM_CLUTIL_X509_DerEncodeAlgorithmId(pContext,AppHandle, &Signature->algorithmIdentifier, &enc_alg_ptr);
    if (ret != CSSM_OK)
    {
        return ret;
    }
    /* Create the signature from the algorithm and signature */
    ret =  
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 
            cl_der_LengthOfContent(enc_alg_ptr),
            &enc_alg_ptr->Data[cl_der_LengthOfTag(enc_alg_ptr) + cl_der_LengthOfLength(enc_alg_ptr)], &(signature_struct_array[0].ChildHandle));
 
    if(ret != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_alg_ptr->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_alg_ptr);
        return ret;
    }
    
    ret =  
        cl_der_CreateLeafNode(pContext,DER_TAG_OCTET_STRING, 
        Signature->encrypted.Length, Signature->encrypted.Data, &(signature_struct_array[1].ChildHandle));


    
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_alg_ptr->Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_alg_ptr);

    if (!signature_struct_array[0].ChildHandle || !signature_struct_array[1].ChildHandle)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_sigstruct_types, 
        signature_struct_array, &root);

    if(ret != CSSM_OK)
    {
       cl_der_DeleteTree (pContext,root);
       return ret;
    }
    /* Pack the signature struct into a DER encoded blob */
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, &results_ptr);

    /* Delete the parse tree */
    cl_der_DeleteTree (pContext,root);

    /* Return the result */
    if (CSSM_OK != ret)
    {
        return ret;
    }

    *result_val = results_ptr;
    return  CSSM_OK;
}




/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_DerEncodeSignedCertificate
 *
 * Description:
 * This function packs a Intel X509V3 signed certificate structure 
 * into a DER encoded memory blob.
 *
 * Parameters: 
 * Cert (input)       : The certificate to be DER encoded
 *
 * results_ptr (output) : The DER encoded certificate
 * 
 * NOTE: allows an unsigned certificate to be DER-encoded 
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_DerEncodeSignedCertificate (CSSM_CLUTIL_CONTEXT *pContext,
                                                         CSSM_HANDLE AppHandle, 
                                                         const CSSM_X509_SIGNED_CERTIFICATE *Cert,
                                                         CSSM_DATA_PTR *results_ptr)
{
    
    uint32              num_certstruct_types = 3;
    CL_DER_CHILD_INFO   cert_struct_array[3];
    CSSM_DATA_PTR       enc_x509_ptr = NULL,
                        enc_sigalg_ptr = NULL;
    CL_DER_NODE_HANDLE  root = 0;
    CSSM_RETURN         ret = CSSM_OK;

    if (!results_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }    

    *results_ptr = NULL;
    /* Initializations */
    port_memset(cert_struct_array, 0, num_certstruct_types*sizeof(CL_DER_CHILD_INFO));

    /* DER encode the certificate components */
    ret = 
        CSSM_CLUTIL_X509_DerEncodeTbsCertificate(pContext, AppHandle, &Cert->certificate, &enc_x509_ptr);
    if (ret != CSSM_OK)
    {
        return ret;
    }
    ret = CSSM_CLUTIL_X509_DerEncodeAlgorithmId(pContext,AppHandle, &Cert->signature.algorithmIdentifier, &enc_sigalg_ptr);

    if (ret != CSSM_OK) 
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_x509_ptr->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_x509_ptr);
        return ret;
    }

    /* Create the DER encoded Cert */
    ret =  
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 
            cl_der_LengthOfContent(enc_x509_ptr),
            &enc_x509_ptr->Data[cl_der_LengthOfTag(enc_x509_ptr) + cl_der_LengthOfLength(enc_x509_ptr)], &cert_struct_array[0].ChildHandle);

    

    if(ret != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_x509_ptr->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_x509_ptr);
        return ret;
    }
 
    ret =  
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 
            cl_der_LengthOfContent(enc_sigalg_ptr),
            &enc_sigalg_ptr->Data[cl_der_LengthOfTag(enc_sigalg_ptr) + cl_der_LengthOfLength(enc_sigalg_ptr)], &cert_struct_array[1].ChildHandle);
 
    if(ret != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_x509_ptr->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_x509_ptr);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_sigalg_ptr->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_sigalg_ptr);
        return ret;
    }

    ret =  
        cl_der_CreateLeafNode(pContext,DER_TAG_BIT_STRING, 
            8*Cert->signature.encrypted.Length, 
            Cert->signature.encrypted.Data, &cert_struct_array[2].ChildHandle);

    if(ret != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_x509_ptr->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_x509_ptr);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_sigalg_ptr->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_sigalg_ptr);
        cl_der_DeleteLeafNode(pContext,cert_struct_array[0].ChildHandle);
        cl_der_DeleteLeafNode(pContext,cert_struct_array[1].ChildHandle);
        return ret;
    }
    
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_x509_ptr->Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_x509_ptr);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_sigalg_ptr->Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_sigalg_ptr);
    
   
    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_certstruct_types, cert_struct_array, &root);

    if(ret != CSSM_OK)
    {   
        cl_der_DeleteTree (pContext,root);
        return ret;
    }
    /* Pack the signature struct into a DER encoded blob */
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, results_ptr);
    if(ret != CSSM_OK)
    {
         cl_der_DeleteTree (pContext,root);
         return ret;
    }

    /* Delete the parse tree */
    cl_der_DeleteTree (pContext,root);
    /* Return */
   
    return ret;
}


/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_DerEncodeCertGroup
 *
 * Description:
 * This function packs a CSSM CertGroup structure 
 * into a single DER encoded memory blob.
 *
 * Parameters: 
 * CertGroupPtr (input) : The certificate group to be DER encoded
 *
 * result_val (output)  : The DER encoded certificate group
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_DerEncodeCertGroup(CSSM_CLUTIL_CONTEXT *pContext,
                                           CSSM_HANDLE AppHandle, 
                                           const CSSM_CERTGROUP *CertGroupPtr,
                                           CSSM_DATA_PTR *result_val)
{
    CSSM_DATA_PTR           results_ptr = NULL;
    CL_DER_CHILD_INFO_PTR   cert_array = NULL;
    CSSM_DATA_PTR           enc_cert_ptr = NULL;
    CL_DER_NODE_HANDLE      root = 0;
    uint32 i=0, j=0;
    CSSM_RETURN             ret = CSSM_OK;

    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }    

    *result_val = NULL;

    /* If there are no certs, return the encoding for a NULL SET */
    if ( !CertGroupPtr ||
         (!CertGroupPtr->GroupList.CertList && !CertGroupPtr->NumCerts))
    {
        
        return cssm_clutil_EncodeNullSet(pContext,AppHandle, result_val);
    }
    if (!CertGroupPtr->GroupList.CertList || !CertGroupPtr->NumCerts)
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }

    /* Initializations */
    cert_array = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, CertGroupPtr->NumCerts, sizeof(CL_DER_CHILD_INFO) );
    if (!cert_array)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* For each cert */
    for (i=0; i < CertGroupPtr->NumCerts; i++)
    {
        enc_cert_ptr = &CertGroupPtr->GroupList.CertList[i];

        /* Create a leaf node */
        ret =
            cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 
            cl_der_LengthOfContent(enc_cert_ptr),
            &enc_cert_ptr->Data[cl_der_LengthOfTag(enc_cert_ptr) + cl_der_LengthOfLength(enc_cert_ptr)], &(cert_array[i].ChildHandle));
                                    
        if (ret != CSSM_OK)
        {
            for(j=0; j<i; j++)
            {
                cl_der_DeleteLeafNode(pContext,cert_array[j].ChildHandle);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, cert_array);
            return CSSMERR_CL_INTERNAL_ERROR;
            
        }
    }

    ret = cl_der_CreateParentNode(pContext,DER_TAG_SET, CertGroupPtr->NumCerts, cert_array, &root);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, cert_array);

    /* Pack the signatures into a DER encoded blob */
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, &results_ptr);

    if(ret != CSSM_OK)
    {   
        cl_der_DeleteTree (pContext,root);
        return ret;
    }
    /* Delete the parse tree */
    cl_der_DeleteTree (pContext,root);
    /* Return the result */
    
    *result_val = results_ptr;
    return CSSM_OK;
}
