/*-----------------------------------------------------------------------
 * File: CRLDECOD.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 

/* This file defines functions for use in decoding a DER-encoded Certificate
   Revocation List (CRL) */


#include "clutil.h"
#include "guids.h"
#include "ber_der.h"
#include "crllist.h"
#include "internal.h"
#include <cssmMemTrack.h>


/*-----------------------------------------------------------------------------
* Name: cl_GetDerTimeChoice
*
* Description:
* This is a private helper function that creates and returns a der-type structure
* that describes a ASN.1 CHOICE field, for a choice between a UTCTime and
* GeneralizedTime.  The return value from this function can be used in the
* 'Choices' array of a DER_COMPLEX_TYPE structure.
*
* Parameters: 
* index:    The index in a Choices array for which this structure is being
*           created.  
*
* Return value:
* DER_EXPLICIT_CHOICES_PTR:     The der-type structure as described above, or
*                               NULL if a memory error occurs.
*---------------------------------------------------------------------------*/
DER_EXPLICIT_CHOICES_PTR cl_GetDerTimeChoice(const CSSM_CLUTIL_CONTEXT *pContext,
                                             CSSM_BOOL ToAppMem,
                                             CSSM_HANDLE AppHandle, 
                                             sint32 index)
{
    CL_DER_TAG_TYPE         tag = 0;
    DER_EXPLICIT_CHOICES_PTR choice = NULL;

    
    /* allocate choice struct */
    choice = wrapped_calloc_func(pContext,ToAppMem, AppHandle, 1, sizeof(DER_EXPLICIT_CHOICES));
    if (!choice)
    {
        return NULL;
    }
    choice->NumberOfChoices = 2;

    /* Field is not explicit -> ParentField same as choice index that this will be assigned
       to */
    choice->ParentField = index;

    /* Allocate oids array and members - these are the tags we are looking for */
    choice->Oids = wrapped_calloc_func(pContext,ToAppMem, AppHandle, 2, sizeof(BER_PARSED_ITEM_PTR));
    if (!choice->Oids)
    {
        wrapped_free_func(pContext,ToAppMem, AppHandle, choice);
        return NULL;
    }
    choice->Oids[0] = wrapped_calloc_func(pContext,ToAppMem, AppHandle, 1, sizeof(BER_PARSED_ITEM));
    choice->Oids[1] = wrapped_calloc_func(pContext,ToAppMem, AppHandle, 1, sizeof(BER_PARSED_ITEM));
    if (!choice->Oids[0] || !choice->Oids[1])
    {
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[0]);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[1]);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids);
        wrapped_free_func(pContext,ToAppMem, AppHandle, choice);
        return NULL;
    }

    /* Since field is not explicit, only need to specify the tag fields */
    choice->Oids[0]->Tag = 
        wrapped_calloc_func(pContext,ToAppMem, AppHandle, 1, sizeof(CL_DER_TAG_TYPE));
    choice->Oids[1]->Tag = 
        wrapped_calloc_func(pContext,ToAppMem, AppHandle, 1, sizeof(CL_DER_TAG_TYPE));
    if (!choice->Oids[0]->Tag || !choice->Oids[1]->Tag)
    {
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[0]->Tag);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[1]->Tag);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[0]);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[1]);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids);
        wrapped_free_func(pContext,ToAppMem, AppHandle, choice);
        return NULL;
    }
    tag = DER_TAG_UTC_TIME;
    port_memcpy((void*)choice->Oids[0]->Tag, &tag, sizeof(CL_DER_TAG_TYPE));
    tag = DER_TAG_GENERALIZED_TIME;
    port_memcpy((void*)choice->Oids[1]->Tag, &tag, sizeof(CL_DER_TAG_TYPE));

    /* Allocate Subtypes array and members- these are the corresponding tags we will store */
    choice->Subtypes = 
        wrapped_calloc_func(pContext,ToAppMem, AppHandle, 2, sizeof(DER_COMPLEX_TYPE_PTR));
    if (!choice->Subtypes)
    {
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[0]->Tag);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[1]->Tag);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[0]);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[1]);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids);
        wrapped_free_func(pContext,ToAppMem, AppHandle, choice);
        return NULL;
    }
    choice->Subtypes[0] = 
        wrapped_calloc_func(pContext,ToAppMem, AppHandle, 1, sizeof(DER_COMPLEX_TYPE));
    choice->Subtypes[1] = 
        wrapped_calloc_func(pContext,ToAppMem, AppHandle, 1, sizeof(DER_COMPLEX_TYPE));
    if (!choice->Subtypes[0] || !choice->Subtypes[1])
    {
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Subtypes[0]);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Subtypes[1]);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Subtypes);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[0]->Tag);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[1]->Tag);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[0]);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[1]);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids);
        wrapped_free_func(pContext,ToAppMem, AppHandle, choice);
        return NULL;
    }

    /* Since it's not *really* a subtype, just fill TypeTag field */
    choice->Subtypes[0]->TypeTag = wrapped_calloc_func(pContext,ToAppMem, AppHandle, 1, sizeof(CL_DER_TAG_TYPE));
    choice->Subtypes[1]->TypeTag = wrapped_calloc_func(pContext,ToAppMem, AppHandle, 1, sizeof(CL_DER_TAG_TYPE));
    if (!choice->Subtypes[0]->TypeTag || !choice->Subtypes[1]->TypeTag)
    {
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Subtypes[0]->TypeTag);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Subtypes[1]->TypeTag);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Subtypes[0]);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Subtypes[1]);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Subtypes);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[0]->Tag);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[1]->Tag);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[0]);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids[1]);
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)choice->Oids);
        wrapped_free_func(pContext,ToAppMem, AppHandle, choice);
        return NULL;
    }
    tag = DER_TAG_UTC_TIME;
    port_memcpy((void*)choice->Subtypes[0]->TypeTag, &tag, sizeof(CL_DER_TAG_TYPE));
    tag = DER_TAG_GENERALIZED_TIME;
    port_memcpy((void*)choice->Subtypes[1]->TypeTag, &tag, sizeof(CL_DER_TAG_TYPE));

    return choice;

}


/*-----------------------------------------------------------------------------
* Name: x509_cl_NumRecordsInRevokedList
*
* Description:
* This function looks through a der encoded revoked cert list and finds how many entries 
* ('records') are in the list.
*
* Parameters: 
* revokedList (input):  The encoded revoked list to query.
* numRecords (output):  This pointer will be used to return the number of records
*                       found in the revoked certs list.
*
* Return value:
* CSSM_RETURN:  CSSM_OK if the operation is successful, Error code otherwise.
*---------------------------------------------------------------------------*/
CSSM_RETURN x509_cl_NumRecordsInRevokedList(const CSSM_CLUTIL_CONTEXT *pContext,
                                            const CSSM_DATA *revokedList,
                                            sint32 *numRecords)
{
    BER_PARSED_ITEM sequenceInfo = {NULL, 0, NULL};
    const uint8           *expand_rtn = NULL;
    
    

    /* expand the der-string and count the number of items in the sequence */
    expand_rtn = BER_ExpandItem(revokedList->Data, revokedList->Length, &sequenceInfo);
    if (!expand_rtn)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }
    *numRecords = BER_CountItems(sequenceInfo.Content, sequenceInfo.ContentLength);

    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
* Name: x509_cl_NumRecordsInTbsCertList
*
* Description:
* This function looks through a der encoded tbsCertList and finds how many entries 
* ('records') are in the revoked certificates list.
*
* Parameters: 
* tbsCertList (input):  The encoded tbsCertList to query.
* numRecords (output):  This pointer will be used to return the number of records
*                       found in the revoked certs list.
*
* Return value:
* CSSM_RETURN:  CSSM_OK if the operation is successful, Error code otherwise.
*
*---------------------------------------------------------------------------*/

#define CERTLIST_NUM_TYPES 7
CSSM_RETURN x509_cl_NumRecordsInTbsCertList(const CSSM_CLUTIL_CONTEXT *pContext,
                                            const CSSM_DATA *tbsCertList, 
                                            sint32 *numRecords)
{
    sint32              items_parsed = 0;
    uint8               **originalTags = NULL;
    uint8               **contextTags = NULL;
    DER_EXPLICIT_CHOICES_PTR *choices = NULL;
    DER_COMPLEX_TYPE_PTR    type_ptr = NULL;
    BER_PARSED_ITEM     fields[7];
    CSSM_DATA           revoked_list;
    uint32              tag_length = 0;
    uint32              length_length = 0;
    CSSM_RETURN         return_val = CSSM_OK;
    sint32              i=0;

    if (!numRecords)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    /* allocate */
    type_ptr = CSSM_CALLOC_WRAPPER(internal_calloc, 1, sizeof(DER_COMPLEX_TYPE), NULL);
    originalTags = CSSM_CALLOC_WRAPPER(internal_calloc, CERTLIST_NUM_TYPES, sizeof(CL_DER_TAG_TYPE *), NULL);
    contextTags  = CSSM_CALLOC_WRAPPER(internal_calloc, CERTLIST_NUM_TYPES, sizeof(CL_DER_TAG_TYPE *), NULL);
    choices      = CSSM_CALLOC_WRAPPER(internal_calloc, CERTLIST_NUM_TYPES, sizeof(DER_EXPLICIT_CHOICES_PTR), NULL);
    if ((type_ptr == NULL) || (originalTags == NULL) || (contextTags == NULL) || (choices == NULL))
    {
        if (type_ptr != NULL)
        {
            CSSM_FREE_WRAPPER(internal_free, type_ptr, NULL);
        }
        if (originalTags != NULL)
        {
            CSSM_FREE_WRAPPER(internal_free, originalTags, NULL);
        }
        if (contextTags != NULL)
        {
            CSSM_FREE_WRAPPER(internal_free, contextTags, NULL);
        }
        if (choices != NULL)
        {
            CSSM_FREE_WRAPPER(internal_free, choices, NULL);
        }
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    type_ptr->OriginalTags = (const uint8 **)originalTags;  /* ifdef VMS apply cast */


    type_ptr->ContextTags = (const uint8 **)contextTags;    /* ifdef VMS apply cast */
    type_ptr->Choices = choices;
    type_ptr->NumberOfDefaults = CERTLIST_NUM_TYPES;

    /* Initialize */ 
    port_memset(fields, 0, CERTLIST_NUM_TYPES*sizeof(BER_PARSED_ITEM));

    /* Allocate Memory */
    for (i=0; i < CERTLIST_NUM_TYPES; i++)
    {
        if (i != 3 && i != 4)
        {
            originalTags[i] = CSSM_MALLOC_WRAPPER(internal_malloc, sizeof(CL_DER_TAG_TYPE), NULL);
            contextTags[i]  = CSSM_MALLOC_WRAPPER(internal_malloc, sizeof(CL_DER_TAG_TYPE), NULL);
            if ((originalTags[i] == NULL) || (contextTags[i] == NULL))
            {
                cl_der_InternalDeleteType(pContext,CSSM_FALSE, 0, type_ptr);
                return CSSMERR_CL_INTERNAL_ERROR;
                
            }
        }
    }

    /* Find where content portion of tbsCertList begins */
    tag_length = cl_der_LengthOfTag(tbsCertList);
    length_length = cl_der_LengthOfLength(tbsCertList);

    /* Prepare for expanding of the tbsCertList sequence */
    *originalTags[0] = BER_INTEGER; /*version*/
    *contextTags[0] = BER_INTEGER;
    *originalTags[1] = BER_CONSTRUCTED_SEQUENCE;    /*signature*/
    *contextTags[1] = BER_CONSTRUCTED_SEQUENCE;
    *originalTags[2] = BER_CONSTRUCTED_SEQUENCE;    /*issuer*/
    *contextTags[2] = BER_CONSTRUCTED_SEQUENCE;
    /** thisUpdate  **/
    /* get choices struct */
    choices[3] = cl_GetDerTimeChoice(pContext,CSSM_FALSE, 0, 3);
    if (!choices[3])
    {
        cl_der_InternalDeleteType(pContext,CSSM_FALSE, 0, type_ptr);
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    /** nextUpdate **/
    /* get choices struct */
    choices[4] = cl_GetDerTimeChoice(pContext,CSSM_FALSE, 0, 4);
    if (!choices[4])
    {
        cl_der_InternalDeleteType(pContext,CSSM_FALSE, 0, type_ptr);
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }
    
    *originalTags[5] = BER_CONSTRUCTED_SEQUENCE;    /*revokedCertificates*/
    *contextTags[5] = BER_CONSTRUCTED_SEQUENCE;
    *originalTags[6] = BER_CONSTRUCTED_SEQUENCE;    /*extensions*/
    *contextTags[6] = 0xA0;
    /* Expand the first level of the tbsCertList */
    items_parsed = BER_ExpandSequence(tbsCertList->Data + tag_length + length_length,
                                               tbsCertList->Length - tag_length - length_length,
                                               CERTLIST_NUM_TYPES,
                                               choices,
                                               (const uint8 **)contextTags,	/* ifdef VMS apply cast */
                                               (const uint8 **)originalTags,    /* ifdef VMS apply cast */
                                               NULL,
                                               fields);
    if (items_parsed < CERTLIST_NUM_TYPES)
    {
        cl_der_InternalDeleteType(pContext,CSSM_FALSE, 0, type_ptr);
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    /* Cleanup */
    cl_der_InternalDeleteType(pContext,CSSM_FALSE, 0, type_ptr);

    /* If no revoked cert list was found, return 0 */
    if (!fields[5].Tag)
    {
        *numRecords = 0;
        return CSSM_OK;
    }
    /* Otherwise, look at the revoked cert list data */
    revoked_list.Length = BER_SizeofObject(fields[5].Tag);
    revoked_list.Data = CSSM_MALLOC_WRAPPER(internal_malloc, revoked_list.Length, NULL);
    if (!revoked_list.Length || !revoked_list.Data)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }
    port_memcpy(revoked_list.Data, fields[5].Tag, revoked_list.Length);
    return_val =  x509_cl_NumRecordsInRevokedList(pContext, &revoked_list, numRecords); 

    /* Cleanup */
    CSSM_FREE_WRAPPER(internal_free, revoked_list.Data, NULL);

    return return_val;
}



/*-----------------------------------------------------------------------------
* Name: x509_cl_NumRecordsInCrl
*
* Description:
* This function looks through a der-encoded crl and finds how many entries ('records') are
* in the revoked certificates list.
*
* Parameters: 
* crl (input):  The encoded crl to query.
* numRecords (output):  This pointer will be used to return the number of records
*                       found in the revoked certs list.
*
* Return value:
* CSSM_RETURN:  CSSM_OK if the operation is successful, CSSM_FAIL otherwise.
*---------------------------------------------------------------------------*/
CSSM_RETURN x509_cl_NumRecordsInCrl(const CSSM_CLUTIL_CONTEXT *pContext,
                                    const CSSM_DATA *crl, 
                                    sint32 *numRecords)
{
    uint32          tag_length = 0;
    uint32          length_length = 0;
    uint32          tbsCertList_length = 0;
    CSSM_DATA_PTR   tbsCertList = NULL;
    CSSM_RETURN     return_val = CSSM_OK;

    if (!numRecords)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    /* Find the beginning of the tbsCertList - it should begin right after the
       tag and length fields of the crl */
    tag_length = cl_der_LengthOfTag(crl);
    length_length = cl_der_LengthOfLength(crl);
    tbsCertList_length = BER_SizeofObject(crl->Data + tag_length + length_length);

    /* Copy tbsCertList portion into its own structure, and then look into that
    structure */
    tbsCertList = CSSM_MALLOC_WRAPPER(internal_malloc, sizeof(CSSM_DATA), NULL);
    if (!tbsCertList)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    tbsCertList->Data = CSSM_MALLOC_WRAPPER(internal_malloc, tbsCertList_length, NULL);
    if (!tbsCertList->Data)
    {
        CSSM_FREE_WRAPPER(internal_free, tbsCertList, NULL);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    tbsCertList->Length = tbsCertList_length;
    port_memcpy(tbsCertList->Data, crl->Data + tag_length + length_length, tbsCertList_length);

    return_val = x509_cl_NumRecordsInTbsCertList(pContext,tbsCertList, numRecords);

    CSSM_FREE_WRAPPER(internal_free, tbsCertList->Data, NULL);
    CSSM_FREE_WRAPPER(internal_free, tbsCertList, NULL);

    return return_val;
}




/*-----------------------------------------------------------------------------
* Name: x509_cl_DetermineRevokedEntryFieldExistence
*
* Description:
* This function fills a boolean array whose elements correspond to the fields of
* a entry in the RevokedCertList.  For each field, the array will hold true if the
* field is present in the input DER parse tree, or false otherwise.
*
* Note that only OPTIONAL fields may ever be omitted (via a CSSM_FALSE entry in the
* return array).
*
* Parameters: 
* NumberOfChildren: The number of elements present in the input ChildrenArray.  An
*                   error will occur if this is not equal to 3.
* ChildrenArray:    An array of CL_DER_NODE's, the result of a parse of a der-encoded
*                   data.
*
* 
* results_val (output): The array described in the 'Description' above, or NULL if an error
*                       occurs.
*
*---------------------------------------------------------------------------*/
CSSM_RETURN x509_cl_DetermineRevokedEntryFieldExistence(const CSSM_CLUTIL_CONTEXT *pContext,
                                                        CSSM_HANDLE AppHandle, 
                                                        uint32 NumberOfChildren,
                                                        CL_DER_NODE_PTR ChildrenArray,
                                                        CSSM_BOOL ** results_val)
{
    CSSM_BOOL *return_val = NULL;
    uint32 i=0;

    if (!results_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    /* NumberOfChildren should always be 3 for a RevokedCertList entry */
    if (NumberOfChildren != 3)
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }

    /* Allocate memory for existence array */
    return_val = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, NumberOfChildren*sizeof(CSSM_BOOL));
    if (!return_val)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* all fields present by default */
    for (i=0; i< NumberOfChildren; i++)
    {
        return_val[i] = CSSM_TRUE;
    }

    /* Determine if the extensions are available */
    if (!ChildrenArray[2].DerNodeType.ParentNode->ChildrenArray)
    {
        return_val[2] = CSSM_FALSE;
    }
    *results_val = return_val;

    return CSSM_OK;
}




/*-----------------------------------------------------------------------------
* Name: cl_AllocateComplexType
*
* Description:
* This function provides error checking of inputs, and allocates and initializes
* a DER_COMPLEX_TYPE structure for use by some functions that create der types.  
* Of the sub-lists in the DER_COMPLEX_TYPE, it only allocates memory for the 
* original tags, context tags, defaults, and choices.
*
* Parameters: 
* num_types: The number of fields in the DER type to be allocated.
*
* result_val (output): The newly allocated and initialized DER_COMPLEX_TYPE structure 
*---------------------------------------------------------------------------*/
CSSM_RETURN cl_AllocateComplexType(const CSSM_CLUTIL_CONTEXT *pContext,
                                            CSSM_HANDLE AppHandle, sint32 num_types, 
                                            DER_COMPLEX_TYPE_PTR *result_val)
{
    DER_COMPLEX_TYPE_PTR    return_val = NULL;
    sint32                  i = 0;

    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    *result_val = NULL;
    /* Allocate a type structure */
    return_val = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(DER_COMPLEX_TYPE));
    if (!return_val)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* allocate and initialize sublists */
    return_val->TypeTag = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(CL_DER_TAG_TYPE));
    return_val->OriginalTags = 
        CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, num_types, sizeof(CL_DER_TAG_TYPE *));
    return_val->ContextTags = 
        CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, num_types, sizeof(CL_DER_TAG_TYPE *));
    return_val->Defaults = 
        CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, num_types, sizeof(BER_PARSED_ITEM));
    return_val->Choices =
        CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, num_types, sizeof(DER_EXPLICIT_CHOICES_PTR));
    if (!return_val->TypeTag || !return_val->OriginalTags || !return_val->ContextTags ||
        !return_val->Defaults || !return_val->Choices)
    {
        cl_der_DeleteType(pContext,AppHandle, return_val);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Allocate the original and context specific tags */
    for (i=0; i < num_types; i++)
    {
        return_val->OriginalTags[i] = 
            CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(CL_DER_TAG_TYPE));
        return_val->ContextTags[i] = 
            CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(CL_DER_TAG_TYPE));
        if (!return_val->OriginalTags[i] || !return_val->ContextTags[i])
        {
            cl_der_DeleteType(pContext,AppHandle, return_val);
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
    }
    *result_val = return_val;
    return CSSM_OK;
}



/*-----------------------------------------------------------------------------
* Name: x509_cl_CreateRevokedEntryDerType
*
* Description:
* This function creates the DER type structure required to properly decode a
* DER-encoded CRL record (i.e. a member of the RevokedCertList)
*
* Parameters: 
* 
*
* result_val (output):The DER type structure corresponding to a CRL record.
*---------------------------------------------------------------------------*/
CSSM_RETURN x509_cl_CreateRevokedEntryDerType(const CSSM_CLUTIL_CONTEXT *pContext,
                                              CSSM_HANDLE AppHandle,
                                              DER_COMPLEX_TYPE_PTR *result_val)
{
    DER_COMPLEX_TYPE_PTR    return_val = NULL;
    sint32                  num_types = 3;
    CL_DER_TAG_TYPE         tag = 0;
    CSSM_RETURN             ret = CSSM_OK;
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    *result_val = NULL;
    /* Allocate type structure and initialize (includes input error checks) */
    ret = cl_AllocateComplexType(pContext,AppHandle, num_types, &return_val);
    if (CSSM_OK != ret)
    {
        /* Let lower level error pass through */
        return ret;
    }

    /* Fill in the type tag and numberOfDefaults */
    tag = BER_CONSTRUCTED_SEQUENCE;
    port_memcpy((void*)return_val->TypeTag, &tag, sizeof(CL_DER_TAG_TYPE));
    return_val->NumberOfDefaults = num_types;

    /* Fill in the original tags for the data */
    tag = DER_TAG_INTEGER;  /* serial number */
    port_memcpy((void*)return_val->OriginalTags[0], &tag, sizeof(CL_DER_TAG_TYPE));
   
    /* revocation date - CHOICE between UTCTime and GeneralizedTime */
    /* original tag should be null */
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (void*)return_val->OriginalTags[1]);
    return_val->OriginalTags[1] = NULL;
    /* fill choice field */
    return_val->Choices[1] = cl_GetDerTimeChoice(pContext,CSSM_TRUE, AppHandle, 1);
    if (!return_val->Choices[1])
    {
        cl_der_DeleteType(pContext,AppHandle, return_val);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    tag = BER_CONSTRUCTED_SEQUENCE; /* extensions */
    port_memcpy((void*)return_val->OriginalTags[2], &tag, sizeof(CL_DER_TAG_TYPE));

   /* Fill in the context-specific tags for the data */
    tag = DER_TAG_INTEGER;  /* serial number */
    port_memcpy((void*)return_val->ContextTags[0], &tag, sizeof(CL_DER_TAG_TYPE));
    /* Should be NULL for revocation date */
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (void*)return_val->ContextTags[1]);
    return_val->ContextTags[1] = NULL;
    tag = BER_CONSTRUCTED_SEQUENCE; /* extensions */
    port_memcpy((void*)return_val->ContextTags[2], &tag, sizeof(CL_DER_TAG_TYPE));

    *result_val = return_val;
    return CSSM_OK;
}



/*-----------------------------------------------------------------------------
* Name: x509_cl_CreateRevokedCertsDerType
*
* Description:
* This function creates the DER type structure required to properly decode
* a DER-encoded RevokedCertList structure.
*
* Parameters: 
* numRecords:   The number of records (entries in the revoked cert list) to
*               create types for.
*
* result_val (output):The DER type structure corresponding to a RevokedCertList.
*---------------------------------------------------------------------------*/
CSSM_RETURN x509_cl_CreateRevokedCertsDerType(const CSSM_CLUTIL_CONTEXT *pContext,
                                              CSSM_HANDLE AppHandle, 
                                              sint32 numRecords,
                                              DER_COMPLEX_TYPE_PTR *result_val)
{
    DER_COMPLEX_TYPE_PTR    return_val = NULL;
    CL_DER_TAG_TYPE         tag = 0;
    sint32                  i=0;
    CSSM_RETURN ret = CSSM_OK;
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;

    /* Allocate a type structure for the revokedCertificates list */
    return_val = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(DER_COMPLEX_TYPE));
    if (!return_val)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Allocate TypeTag and Subtypes */
    return_val->TypeTag = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(CL_DER_TAG_TYPE));
    return_val->Subtypes = 
        CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, numRecords, sizeof(DER_COMPLEX_TYPE_PTR));
    if (!return_val->TypeTag || !return_val->Subtypes)
    {
        cl_der_DeleteType(pContext,AppHandle, return_val);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Set TypeTag and number of defaults */
    tag = BER_CONSTRUCTED_SEQUENCE;
    port_memcpy((void*)return_val->TypeTag, &tag, sizeof(CL_DER_TAG_TYPE));
    return_val->NumberOfDefaults = numRecords;

    /* Set Subtypes */
    for (i=0; i < numRecords; i++)
    {
        ret = x509_cl_CreateRevokedEntryDerType(pContext,AppHandle, &(return_val->Subtypes[i]));
        if (CSSM_OK != ret)
        {
            cl_der_DeleteType(pContext,AppHandle, return_val);
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
    }

    *result_val = return_val;
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
* Name: x509_cl_CreateTbsCertListDerType
*
* Description:
* This function creates the DER type structure required to properly decode
* a DER-encoded tbsCertList structure.
*
* Parameters: 
* numRecords:   The number of records that will be stored in the revoked cert list.
* result_val (output): The DER type structure corresponding to a tbsCertList.
*
*---------------------------------------------------------------------------*/
CSSM_RETURN x509_cl_CreateTbsCertListDerType(const CSSM_CLUTIL_CONTEXT *pContext,
                                             CSSM_HANDLE AppHandle, 
                                             sint32 numRecords,
                                             DER_COMPLEX_TYPE_PTR *result_val)
{
    DER_COMPLEX_TYPE_PTR    return_val = NULL;
    sint32                  num_types = 7;
    CL_DER_TAG_TYPE         tag = 0;
    CSSM_RETURN ret = CSSM_OK;
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;
    /* Allocate type structure and initialize (includes input error checks) */
    ret = cl_AllocateComplexType(pContext,AppHandle, num_types, &return_val);
    if (CSSM_OK != ret)
    {
        /* Let lower level error pass through */
        return ret;
    }

    /* Fill in the type tag and numberOfDefaults */
    tag = BER_CONSTRUCTED_SEQUENCE;
    port_memcpy((void*)return_val->TypeTag, &tag, sizeof(CL_DER_TAG_TYPE));
    return_val->NumberOfDefaults = num_types;

    /* Fill in the original tags for the data portion of a CRL */
    /* version */
    tag = DER_TAG_INTEGER;  
    port_memcpy((void*)return_val->OriginalTags[0], &tag, sizeof(CL_DER_TAG_TYPE));

    /*** thisUpdate - CHOICE between UTCTime and GeneralizedTime ***/
    /* original tag should be null */
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (void*)return_val->OriginalTags[3]);
    return_val->OriginalTags[3] = NULL;
    /* fill choice field */
    return_val->Choices[3] = cl_GetDerTimeChoice(pContext,CSSM_TRUE, AppHandle, 3);
    if (!return_val->Choices[3])
    {
        cl_der_DeleteType(pContext,AppHandle, return_val);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /*** nextUpdate - CHOICE between UTCTime and GeneralizedTime ***/
    /* original tag should be null */
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (void*)return_val->OriginalTags[4]);
    return_val->OriginalTags[4] = NULL;
    /* fill choice field */
    return_val->Choices[4] = cl_GetDerTimeChoice(pContext,CSSM_TRUE, AppHandle, 4);
    if (!return_val->Choices[4])
    {
        cl_der_DeleteType(pContext,AppHandle, return_val);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* signature, issuer, revokedCertificates, extensions */
    tag = BER_CONSTRUCTED_SEQUENCE; 
    port_memcpy((void*)return_val->OriginalTags[1], &tag, sizeof(CL_DER_TAG_TYPE));
    port_memcpy((void*)return_val->OriginalTags[2], &tag, sizeof(CL_DER_TAG_TYPE));
    port_memcpy((void*)return_val->OriginalTags[5], &tag, sizeof(CL_DER_TAG_TYPE));
    port_memcpy((void*)return_val->OriginalTags[6], &tag, sizeof(CL_DER_TAG_TYPE));

    /* Fill in the context-specific tags for the data portion of a CRL */
    /* version */
    tag = DER_TAG_INTEGER;  
    port_memcpy((void*)return_val->ContextTags[0], &tag, sizeof(CL_DER_TAG_TYPE));

    /* thisUpdate and nextUpdate should have null field here */
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (void*)return_val->ContextTags[3]);
    return_val->ContextTags[3] = NULL;
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (void*)return_val->ContextTags[4]);
    return_val->ContextTags[4] = NULL;

    /* signature, issuer, revokedCertificates */
    tag = BER_CONSTRUCTED_SEQUENCE; 
    port_memcpy((void*)return_val->ContextTags[1], &tag, sizeof(CL_DER_TAG_TYPE));
    port_memcpy((void*)return_val->ContextTags[2], &tag, sizeof(CL_DER_TAG_TYPE));
    port_memcpy((void*)return_val->ContextTags[5], &tag, sizeof(CL_DER_TAG_TYPE));
    /* extensions */
    tag = 0xA0;     
    port_memcpy((void*)return_val->ContextTags[6], &tag, sizeof(CL_DER_TAG_TYPE));

    /* Allocate and fill in subtype for revokedCertificates list */
    return_val->Subtypes = 
        CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, num_types, sizeof(DER_COMPLEX_TYPE_PTR));
    if (!return_val->Subtypes)
    {
        cl_der_DeleteType(pContext,AppHandle, return_val);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

     
    x509_cl_CreateRevokedCertsDerType(pContext,AppHandle, numRecords, &return_val->Subtypes[5]);
    if (!return_val->Subtypes[5])
    {
        cl_der_DeleteType(pContext,AppHandle, return_val);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    *result_val = return_val;
    return CSSM_OK;
}



/*-----------------------------------------------------------------------------
 * Name: x509_cl_CreateSignedCrlDerType
 *
 * Description:
 * This function creates the DER type structure required to properly decode
 * the DER-encoded CRL structure
 *
 * Parameters: 
 * numRecords: The number of CRL records that will be stored in the revoked
 *              certificates list.
 *
 * result_val (output): The DER type structure corresponding to a CRL
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN  x509_cl_CreateSignedCrlDerType(const CSSM_CLUTIL_CONTEXT *pContext,
                                            CSSM_HANDLE AppHandle, 
                                            sint32 numRecords,
                                            DER_COMPLEX_TYPE_PTR *results_val)
{

    DER_COMPLEX_TYPE_PTR    return_val = NULL;
    sint32                  num_types = 3;
    CL_DER_TAG_TYPE         tag = 0;

    
    if (!results_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    /* Allocate a type structure for the signed crl */
    return_val = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(DER_COMPLEX_TYPE));
    if (!return_val)
    {
        return CSSMERR_CL_MEMORY_ERROR;

    }

    /* set typeTag */
    return_val->TypeTag = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, sizeof(CL_DER_TAG_TYPE));
    if (!return_val->TypeTag)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, return_val);
        return CSSMERR_CL_MEMORY_ERROR;
    }
    tag = BER_CONSTRUCTED_SEQUENCE;
    port_memcpy((void*)return_val->TypeTag, &tag, sizeof(CL_DER_TAG_TYPE));
    return_val->NumberOfDefaults = num_types;

    /* Allocate a sub-type structure for the data portion of the crl */
    return_val->Subtypes = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, num_types, sizeof(DER_COMPLEX_TYPE_PTR));
    if (!return_val->Subtypes)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (void*)return_val->TypeTag);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, return_val);
        return CSSMERR_CL_MEMORY_ERROR;
    }

    /* make tbsCertList subtype der type */
    x509_cl_CreateTbsCertListDerType(pContext,AppHandle, numRecords, &return_val->Subtypes[0]);
    if (!return_val->Subtypes[0])
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (void*)return_val->Subtypes);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (void*)return_val->TypeTag);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, return_val);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    *results_val = return_val;
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: x509_cl_CreateSetofSignedCrlDerType
 *
 * Description:
 *
 * Parameters: 
 *
 * Return value:
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN x509_cl_CreateSetofSignedCrlDerType (const CSSM_CLUTIL_CONTEXT *pContext,
                                                 CSSM_HANDLE AppHandle, 
                                                 uint32 NumCrls,
                                                 uint8 *EncCrlGroupPtr,
                                                 DER_COMPLEX_TYPE_PTR *result_val)
{
    DER_COMPLEX_TYPE_PTR der_type_ptr = NULL;
    CL_DER_TAG_TYPE tag = 0;
    CSSM_DATA       temp_data = {0, NULL};
    uint8           **crl_obj_array = NULL;
    uint32          num_crls = 0;
    sint32          num_records = 0;
    uint32 i=0;
    CSSM_RETURN ret = CSSM_OK;
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;
    if (!NumCrls || !EncCrlGroupPtr)
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }
    temp_data.Length = BER_SizeofObject(EncCrlGroupPtr);
    temp_data.Data = EncCrlGroupPtr;
    if( CSSM_CLUTIL_DerGetFirstLevelTagPtrs(pContext,AppHandle, &temp_data, &crl_obj_array, &num_crls) != CSSM_OK)
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }
    if (num_crls != NumCrls)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, crl_obj_array);
        return CSSMERR_CL_INVALID_DATA;
        
    }

    /* Allocate a type structure for the set of signed CRLs */
    der_type_ptr = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(DER_COMPLEX_TYPE));
    if (!der_type_ptr)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, crl_obj_array);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    der_type_ptr->TypeTag = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, sizeof(CL_DER_TAG_TYPE));
    der_type_ptr->ContextTags  = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, NumCrls, sizeof(CL_DER_TAG_TYPE));
    der_type_ptr->Subtypes = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, NumCrls, sizeof(DER_COMPLEX_TYPE_PTR));
    if (!der_type_ptr->TypeTag || !der_type_ptr->ContextTags || !der_type_ptr->Subtypes)
    {
        cl_der_DeleteType(pContext,AppHandle, der_type_ptr);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, crl_obj_array);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    tag = BER_CONSTRUCTED_SET;
    port_memcpy((void*)der_type_ptr->TypeTag, &tag, sizeof(CL_DER_TAG_TYPE));
    der_type_ptr->NumberOfDefaults = 0;  /* No default crls in this set */

    tag = BER_CONSTRUCTED_SEQUENCE;
    for (i=0; i < NumCrls; i++)
    {
        temp_data.Length = BER_SizeofObject(crl_obj_array[i]);
        temp_data.Data = crl_obj_array[i];
        if (x509_cl_NumRecordsInCrl(pContext,&temp_data, &num_records) != CSSM_OK)
        {
            cl_der_DeleteType(pContext,AppHandle, der_type_ptr);
            return CSSMERR_CL_INTERNAL_ERROR;
            
        }

        ret = x509_cl_CreateSignedCrlDerType(pContext,AppHandle, num_records, &der_type_ptr->Subtypes[i]);
        der_type_ptr->ContextTags[i]  = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(CL_DER_TAG_TYPE));
        if (!der_type_ptr->Subtypes[i] || !der_type_ptr->ContextTags[i])
        {
            cl_der_DeleteType(pContext,AppHandle, der_type_ptr);
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
        port_memcpy((void*)der_type_ptr->ContextTags[i], &tag, sizeof(CL_DER_TAG_TYPE));
    }

    *result_val = der_type_ptr;
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
* Name: x509_cl_RevokedEntryParseTreeToStruct
*
* Description:
* This function translates a parse tree into a revoked certificate entry struct.
*
* Parameters: 
* parsed_tree:  The tree to be translated, obtained by parsing a der-encoded revoked
*               certificate entry.
*
* results_val (output):The revoked cert entry structure filled with data
*                      corresponding to the input parse tree.
*
*---------------------------------------------------------------------------*/
CSSM_RETURN x509_cl_RevokedEntryParseTreeToStruct(const CSSM_CLUTIL_CONTEXT *pContext,
                                                  CSSM_HANDLE AppHandle, 
                                                  CL_DER_NODE_PTR parsed_tree,
                                                  CSSM_X509_REVOKED_CERT_ENTRY_PTR *results_val)
{
    CSSM_X509_REVOKED_CERT_ENTRY_PTR return_val = NULL;
    CSSM_BOOL                   *existence = NULL;
    CL_DER_PARENT_NODE_PTR      tree_extensions = NULL;
    CSSM_RETURN ret = CSSM_OK;
    if (!results_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    /* Error checking on first level of tree */
    if (parsed_tree->DerNodeType.ParentNode->NumberOfChildren != 3 ||
        CSSM_CLUTIL_X509_IsBadReadPtr(parsed_tree->DerNodeType.ParentNode->ChildrenArray, 3*sizeof(CL_DER_NODE)))
    {
        return CSSMERR_CL_INVALID_DATA;
        
    } 
    x509_cl_DetermineRevokedEntryFieldExistence(pContext,
                                                AppHandle, 
                                                parsed_tree->DerNodeType.ParentNode->NumberOfChildren,
                                                parsed_tree->DerNodeType.ParentNode->ChildrenArray,
                                                &existence);
    if (parsed_tree->DerNodeType.ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Tag != DER_TAG_INTEGER  ||
        (parsed_tree->DerNodeType.ParentNode->ChildrenArray[1].DerNodeType.LeafNode->Tag != DER_TAG_UTC_TIME &&
         parsed_tree->DerNodeType.ParentNode->ChildrenArray[1].DerNodeType.LeafNode->Tag != DER_TAG_GENERALIZED_TIME) ||
        (existence[2] == CSSM_TRUE && parsed_tree->DerNodeType.ParentNode->ChildrenArray[2].DerNodeType.ParentNode->Tag != DER_TAG_SEQUENCE))
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }

    /* Allocate new structure */
    return_val = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(CSSM_X509_REVOKED_CERT_ENTRY));
    if (!return_val)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Fill in serial number */
    return_val->certificateSerialNumber.Length = 
        parsed_tree->DerNodeType.ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Data.Length;
    if (return_val->certificateSerialNumber.Length)
    {
        return_val->certificateSerialNumber.Data = 
            CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, return_val->certificateSerialNumber.Length);
        if (!return_val->certificateSerialNumber.Data)
        {
            CSSM_CLUTIL_X509_FreeRevokedCertEntry(pContext,AppHandle, return_val);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, existence);
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
        port_memcpy(return_val->certificateSerialNumber.Data, 
                    parsed_tree->DerNodeType.ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Data.Data, 
                    return_val->certificateSerialNumber.Length);
    }
    else
    {
        return_val->certificateSerialNumber.Data = NULL;
    }

    /* Fill in the revocation date */
        /* note parse tree includes terminating null in length */
    if (parsed_tree->DerNodeType.ParentNode->ChildrenArray[1].DerNodeType.LeafNode->Data.Length - 1)
    {
        return_val->revocationDate.time.Data = 
            CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, 
                                         parsed_tree->DerNodeType.ParentNode->ChildrenArray[1].DerNodeType.LeafNode->Data.Length);
        if (!return_val->revocationDate.time.Data)
        {
            CSSM_CLUTIL_X509_FreeRevokedCertEntry(pContext,AppHandle, return_val);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, existence);
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
        return_val->revocationDate.time.Length = parsed_tree->DerNodeType.ParentNode->ChildrenArray[1].DerNodeType.LeafNode->Data.Length;
        port_memcpy(return_val->revocationDate.time.Data, 
                    parsed_tree->DerNodeType.ParentNode->ChildrenArray[1].DerNodeType.LeafNode->Data.Data, 
                    parsed_tree->DerNodeType.ParentNode->ChildrenArray[1].DerNodeType.LeafNode->Data.Length);
    }
    else
    {
        return_val->revocationDate.time.Data = NULL;
        return_val->revocationDate.time.Length = 0;
    }

    /* Fill in extensions */
    if (existence[2] == CSSM_TRUE)
    {
        tree_extensions = 
            parsed_tree->DerNodeType.ParentNode->ChildrenArray[2].DerNodeType.ParentNode;    
        return_val->extensions.numberOfExtensions = tree_extensions->NumberOfChildren;
        if (return_val->extensions.numberOfExtensions)
        {
            ret = x509_cl_ExtensionsParseTreeToArray(pContext,AppHandle, &parsed_tree->DerNodeType.ParentNode->ChildrenArray[2], &return_val->extensions.extensions);
            if (CSSM_OK != ret)
            {
                CSSM_CLUTIL_X509_FreeRevokedCertEntry(pContext,AppHandle, return_val);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, existence);
                
                return ret;
            }
        }
    }
    else
    {
        return_val->extensions.numberOfExtensions = 0;
        return_val->extensions.extensions = NULL;
    }

    /* Cleanup */
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, existence);
    *results_val = return_val;
    return ret;
}



/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_DerDecodeRevokedCertEntry
*
* Description:
* This function unpacks a X509V2 CRL revocation record from a  
* a DER encoded memory blob into a CSSM revoked certificate list structure.
*
* Parameters: 
* data: The data to be decoded.
*
* result_val (output): A revoked cert entry structure filled with data
*                      corresponding to the der encoded input data, or
*                      NULL if an error occurs.
*
*---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_DerDecodeRevokedCertEntry(const CSSM_CLUTIL_CONTEXT *pContext,
                                                       CSSM_HANDLE AppHandle,
                                                       const CSSM_DATA *data,
                                                       CSSM_X509_REVOKED_CERT_ENTRY_PTR *result_val)
{
    CSSM_X509_REVOKED_CERT_ENTRY_PTR return_val = NULL;
    CL_DER_NODE_PTR             parsed_tree = NULL;
    DER_COMPLEX_TYPE_PTR        der_type_ptr = NULL;
    CSSM_RETURN                 ret = CSSM_OK;
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
 
    *result_val = NULL;
    /** Unpack the revoked entry data */

    /* Create the der type */
    ret = x509_cl_CreateRevokedEntryDerType(pContext,AppHandle, &der_type_ptr);
    if (CSSM_OK != ret)
    {
        return ret;
        
    }

    /* Build Parse tree */
    ret = cl_der_ParseDerObject(pContext,AppHandle, data, der_type_ptr, &parsed_tree);
    cl_der_DeleteType(pContext,AppHandle, der_type_ptr);
    if (CSSM_OK != ret)
    {
        return ret;
        
    }

    /* Build the revoked entry structure */
    ret = x509_cl_RevokedEntryParseTreeToStruct(pContext,AppHandle, parsed_tree, &return_val);
    if (CSSM_OK != ret)
    {
        return ret;
        
    }

    /* Cleanup */
    ret = cl_der_DeleteParsedDerObject(pContext,AppHandle, parsed_tree);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, parsed_tree);

    *result_val = return_val;
    return CSSM_OK;

}




/*-----------------------------------------------------------------------------
* Name: x509_cl_RevokedListParseTreeToStruct
*
* Description:
* This function translates a parse tree into a revoked certificates list structure.
*
* Parameters: 
* parsed_tree:  The tree to be translated, obtained by parsing a der-encoded revoked
*               certificate list.
*
* result_val (output): The revoked cert list structure filled with data
*                      corresponding to the input parse tree, or NULL if
*                      an error occurs.
*
*---------------------------------------------------------------------------*/
CSSM_RETURN x509_cl_RevokedListParseTreeToStruct(const CSSM_CLUTIL_CONTEXT *pContext,
                                                 CSSM_HANDLE AppHandle, 
                                                 CL_DER_NODE_PTR parsed_tree,
                                                 X509_REVOKED_CERT_LIST_PTR *result_val)
{
    X509_REVOKED_CERT_LIST_PTR  return_val = NULL;
    CSSM_X509_REVOKED_CERT_ENTRY_PTR RCEntry = NULL;
    uint32                      i = 0;
    CSSM_RETURN ret;
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    
    *result_val = NULL;
    /* Error checking on first level of tree */
    if (CSSM_CLUTIL_X509_IsBadReadPtr(parsed_tree->DerNodeType.ParentNode->ChildrenArray, parsed_tree->DerNodeType.ParentNode->NumberOfChildren * sizeof(CL_DER_NODE)))
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }

    /* Create a new empty revoked list */
    ret = CSSM_CLUTIL_X509_RevokedListCreate(pContext,AppHandle,&return_val);
    if (CSSM_OK != ret)
    {
        return ret;
        
    }

    /* For each revoked list entry in the parse tree, translate it and add to list */
    for (i = 0; i < parsed_tree->DerNodeType.ParentNode->NumberOfChildren; i++)
    {
        /* Translate parsed entry */
        
        ret = x509_cl_RevokedEntryParseTreeToStruct(pContext,AppHandle, &parsed_tree->DerNodeType.ParentNode->ChildrenArray[i], &RCEntry);
        /* Add entry struct to list */
        ret = CSSM_CLUTIL_X509_InsertRevokedCertEntry(pContext,AppHandle, return_val, RCEntry);
        if( CSSM_OK != ret)
        {
            CSSM_CLUTIL_X509_FreeRevokedCertEntry(pContext,AppHandle, RCEntry);
            CSSM_CLUTIL_X509_FreeRevokedCertList(pContext,AppHandle, return_val);
            return ret;
            
        }
        /* Free up the RCEntry pointer for use on next iteration */
        CSSM_CLUTIL_X509_FreeRevokedCertEntry(pContext,AppHandle, RCEntry);
    }

    /* return the list */
    *result_val = return_val;
    return ret;
}


/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_DerDecodeRevokedCertList
*
* Description:
* This function unpacks the revoked certificates list portion of a x509 CRL from
* a DER encoded memory blob into a CSSM revoked certificate list structure.
*
* Parameters: 
* data: The data to be decoded.
*
* result_val (output): A revoked cert list structure filled with data
*                      corresponding to the der encoded input data, or
*                      NULL if an error occurs.
*
*---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_DerDecodeRevokedCertList(const CSSM_CLUTIL_CONTEXT *pContext,
                                                      CSSM_HANDLE AppHandle,
                                                      const CSSM_DATA *data,
                                                      X509_REVOKED_CERT_LIST_PTR *result_val)
{
    X509_REVOKED_CERT_LIST_PTR  return_val = NULL;
    CL_DER_NODE_PTR             parsed_tree = NULL;
    DER_COMPLEX_TYPE_PTR        der_type_ptr = NULL;
    sint32                      numRecords = 0;
    CSSM_RETURN                 ret = CSSM_OK;
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
   
    *result_val = NULL;
    /* Need to know the number of records in the list before decoding */
    ret = x509_cl_NumRecordsInRevokedList(pContext,data, &numRecords);
    if (CSSM_OK != ret)
    {
        return ret;
        
    }

    /* Unpack the revoked list data */
    ret = x509_cl_CreateRevokedCertsDerType(pContext,AppHandle, numRecords, &der_type_ptr);
    if (!der_type_ptr)
    {
        return ret;
        
    }

    /* Build parse tree */
    ret = cl_der_ParseDerObject(pContext,AppHandle, data, der_type_ptr, &parsed_tree);
    cl_der_DeleteType(pContext,AppHandle, der_type_ptr);
    if (!parsed_tree)
    {
        return ret;
        
    }

    /* Build the revoked list structure */
    ret = x509_cl_RevokedListParseTreeToStruct(pContext,AppHandle, parsed_tree, &return_val);
    if (!return_val)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    /* Cleanup */
    ret = cl_der_DeleteParsedDerObject(pContext,AppHandle, parsed_tree);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, parsed_tree);

    *result_val = return_val;
    return CSSM_OK;
}



/*-----------------------------------------------------------------------------
* Name: x509_cl_TbsCertListParseTreeToStruct
*
* Description:
* This function translates a CL parse tree into a CRL tbsCertList struct.
*
* Parameters: 
* parsed_tree:  The parse tree to be translated.
*
* result_val (output):A new tbsCertList struct, initialized with data
*                     corresponding to the input parse tree, or NULL if
*                     an error occurs.
*
*---------------------------------------------------------------------------*/
CSSM_RETURN x509_cl_TbsCertListParseTreeToStruct(const CSSM_CLUTIL_CONTEXT *pContext,
                                                 CSSM_HANDLE AppHandle, 
                                                 CL_DER_NODE_PTR parsed_tree,
                                                 X509_TBS_CERTLIST_PTR *result_val)
{
    X509_TBS_CERTLIST_PTR           return_val = NULL;
    CL_DER_PARENT_NODE_PTR          tree_signature = NULL,
                                    tree_extensions = NULL;
    CSSM_X509_NAME_PTR                   temp_name_ptr = NULL;
    CSSM_X509_ALGORITHM_IDENTIFIER_PTR   algorithm_ptr = NULL;
    CSSM_BOOL                       existence[7];
    CSSM_RETURN                     ret = CSSM_OK;
    
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *result_val = NULL;
    if (x509_cl_IsBadTbsCertListParseTree(parsed_tree->DerNodeType.ParentNode, existence))
    {
        return CSSMERR_CL_INVALID_DATA;
    }
    tree_signature = parsed_tree->DerNodeType.ParentNode->ChildrenArray[1].DerNodeType.ParentNode;

    /* Allocate Memory for the new tbsCertList */
    return_val = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(X509_TBS_CERTLIST));
    if (!return_val)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    
    /** Fill in data values **/
    /* Fill in the version */
    if (existence[0])
    {
        return_val->version.Length = 
            parsed_tree->DerNodeType.ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Data.Length;
        return_val->version.Data = 
            CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, return_val->version.Length);
        if (!return_val->version.Data)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, return_val);
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
        port_memcpy(return_val->version.Data, 
                    parsed_tree->DerNodeType.ParentNode->ChildrenArray[0].DerNodeType.LeafNode->Data.Data,
                    return_val->version.Length);
    }

    /* Fill in the signature algorithm */
    ret = x509_cl_AlgorithmParseTreeToStruct(pContext,AppHandle, &parsed_tree->DerNodeType.ParentNode->ChildrenArray[1], &algorithm_ptr);
    if (CSSM_OK != ret)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, return_val);
        return ret;
        
    }
    return_val->signature = *algorithm_ptr;
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, algorithm_ptr);

    /* Fill in the issuer */
    ret = x509_cl_X509NameParseTreeToStruct(pContext,AppHandle, &parsed_tree->DerNodeType.ParentNode->ChildrenArray[2], &temp_name_ptr);
    if (CSSM_OK != ret)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, return_val);
        return ret;
        
    }
    return_val->issuer = *temp_name_ptr;
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, temp_name_ptr);

    /* Fill in thisUpdate field */
    /* note parse tree includes terminating null in length */
    if ((parsed_tree->DerNodeType.ParentNode->ChildrenArray[3].DerNodeType.LeafNode->Data.Length - 1) > 0)
    {
        return_val->thisUpdate.time.Data = 
            CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, 
                                         parsed_tree->DerNodeType.ParentNode->ChildrenArray[3].DerNodeType.LeafNode->Data.Length);
        if (!return_val->thisUpdate.time.Data)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, return_val);
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
        return_val->thisUpdate.time.Length = parsed_tree->DerNodeType.ParentNode->ChildrenArray[3].DerNodeType.LeafNode->Data.Length;
        port_memcpy(return_val->thisUpdate.time.Data,
                    parsed_tree->DerNodeType.ParentNode->ChildrenArray[3].DerNodeType.LeafNode->Data.Data,
                    parsed_tree->DerNodeType.ParentNode->ChildrenArray[3].DerNodeType.LeafNode->Data.Length);
    }
    else
    {
        return_val->thisUpdate.time.Data = NULL;
        return_val->thisUpdate.time.Length = 0;
    }

    /* Fill in nextUpdate field */
    /* note parse tree includes terminating null in length */
    if (parsed_tree->DerNodeType.ParentNode->ChildrenArray[4].DerNodeType.LeafNode->Data.Length - 1)
    {
        return_val->nextUpdate.time.Data = 
            CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, 
                                         parsed_tree->DerNodeType.ParentNode->ChildrenArray[4].DerNodeType.LeafNode->Data.Length);
        if (!return_val->nextUpdate.time.Data)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, return_val);
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
        port_memcpy(return_val->nextUpdate.time.Data,
                    parsed_tree->DerNodeType.ParentNode->ChildrenArray[4].DerNodeType.LeafNode->Data.Data,
                    parsed_tree->DerNodeType.ParentNode->ChildrenArray[4].DerNodeType.LeafNode->Data.Length);
    }
    else
    {
        return_val->nextUpdate.time.Data = NULL;
        return_val->nextUpdate.time.Length = 0;
    }

    /* Fill in revokedCertificates field */
    if (existence[5])
    {
        ret = x509_cl_RevokedListParseTreeToStruct(pContext,AppHandle, &parsed_tree->DerNodeType.ParentNode->ChildrenArray[5], &return_val->revokedCertificates);
        if (CSSM_OK != ret)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, return_val);
            return ret;
        }
    }
    else
    {
        /* no list - assign NULL*/
        return_val->revokedCertificates = NULL;
    }

    /* Fill in extensions */
    if (existence[6])
    {
        tree_extensions = 
            parsed_tree->DerNodeType.ParentNode->ChildrenArray[6].DerNodeType.ParentNode->ChildrenArray[0].DerNodeType.ParentNode;
        return_val->extensions.numberOfExtensions = tree_extensions->NumberOfChildren;
        if (return_val->extensions.numberOfExtensions)
        {
            ret = x509_cl_ExtensionsParseTreeToArray(pContext,AppHandle, &parsed_tree->DerNodeType.ParentNode->ChildrenArray[6].DerNodeType.ParentNode->ChildrenArray[0], &return_val->extensions.extensions);
            if (CSSM_OK != ret)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, return_val);
                return ret;
            }

        }
        else
        {
            return_val->extensions.extensions = NULL;
        }
    }
    else
    {
        return_val->extensions.numberOfExtensions = 0;
        return_val->extensions.extensions = NULL;
    }

    /*Return */
    *result_val = return_val;
    return ret;
}
         


/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_DerDecodeTbsCertList
*
* Description:
* This function unpacks the tbsCertList portion of a x509V2 CRL from a DER encoded
* memory blob into a CSSM CRL tbsCertList structure.
*
* Parameters: 
* data: The DER-encoded tbsCertList data to be decoded.
*
* result_val (output):A new tbsCertList structure, initialized with data
*                     corresponding to the input der-encoded data, or NULL
*                     if an error occurs.
*
*---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_DerDecodeTbsCertList(const CSSM_CLUTIL_CONTEXT *pContext,
                                                  CSSM_HANDLE AppHandle,
                                                  const CSSM_DATA *data,
                                                  X509_TBS_CERTLIST_PTR *result_val)
{
    X509_TBS_CERTLIST_PTR   return_val = NULL;
    CL_DER_NODE_PTR         parsed_tree = NULL;
    DER_COMPLEX_TYPE_PTR    der_type_ptr = NULL;
    sint32                  numRecords = 0;
    CSSM_RETURN             ret = CSSM_OK;
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
   
    *result_val = NULL;
    /* need to know the number of records before decoding */
    ret = x509_cl_NumRecordsInTbsCertList(pContext,data, &numRecords);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    

    /** Unpack the tbsCertList data **/
    /* create der type */
    ret = x509_cl_CreateTbsCertListDerType(pContext,AppHandle, numRecords, &der_type_ptr);
    if (!der_type_ptr)
    {        
        return ret;
    }
    
    /* build parse tree */
    ret = cl_der_ParseDerObject(pContext,AppHandle, data, der_type_ptr, &parsed_tree);
    cl_der_DeleteType(pContext,AppHandle, der_type_ptr);
    if (!parsed_tree)
    {        
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    /* Get the tbsCertList structure */
    ret = x509_cl_TbsCertListParseTreeToStruct(pContext,AppHandle, parsed_tree, &return_val);
    ret = cl_der_DeleteParsedDerObject(pContext,AppHandle, parsed_tree);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, parsed_tree);
    if (!return_val)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    *result_val = return_val;
    return CSSM_OK;
}



/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_DerDecodeSignedCrl
*
* Description:
* This function unpacks an X509V2 CRL from a DER encoded memory blob into a 
* CSSM CRL structure.
*
* Parameters: 
* crl_data:     The DER-encoded CRL data to be decoded. 
*
* result_val(output):   A new CRL structure, initialized with data corresponding
*                       to the der-encoded input crl_data, or NULL if an error
*                       occurs.
*
*---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_DerDecodeSignedCrl(const CSSM_CLUTIL_CONTEXT *pContext,
                                                CSSM_HANDLE AppHandle,
                                                const CSSM_DATA *crl_data,
                                                X509_SIGNED_CRL_PTR *result_val)
{
    X509_SIGNED_CRL_PTR             return_val = NULL;
    CL_DER_NODE_PTR                 parsed_tree = NULL;
    X509_TBS_CERTLIST_PTR           tbsCertList_ptr = NULL;
    CSSM_X509_ALGORITHM_IDENTIFIER_PTR   temp_sigalg_ptr = NULL;
    DER_COMPLEX_TYPE_PTR            der_type_ptr = NULL;
    sint32                          numRecords = 0;
    CSSM_RETURN                     ret = CSSM_OK;
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    
    *result_val = NULL;

    /* need to know number of records before decoding */
    if (x509_cl_NumRecordsInCrl(pContext,crl_data, &numRecords) != CSSM_OK)
    {
        return CSSMERR_CL_INVALID_DATA;
    }

    
    
    /* Unpack the CRL data */
    ret = x509_cl_CreateSignedCrlDerType(pContext,AppHandle, numRecords, &der_type_ptr);
    if (CSSM_OK != ret)
    {
        return ret;
        
    }
    
    /* build parse tree */
    ret = cl_der_ParseDerObject(pContext,AppHandle, crl_data, der_type_ptr, &parsed_tree);
    cl_der_DeleteType(pContext,AppHandle, der_type_ptr);
    if (!parsed_tree)
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }

    /* Error checking on the tree */
    if (x509_cl_IsBadCrlParseTree(parsed_tree->DerNodeType.ParentNode))
    {
        ret = cl_der_DeleteParsedDerObject(pContext,AppHandle, parsed_tree);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, parsed_tree);
        return CSSMERR_CL_INVALID_DATA;
        
    }

    /* Allocate a crl structure */
    return_val = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(X509_SIGNED_CRL));
    if (!return_val)
    {
        ret = cl_der_DeleteParsedDerObject(pContext,AppHandle, parsed_tree);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, parsed_tree);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Get the tbsCertList structure */
    ret = x509_cl_TbsCertListParseTreeToStruct(pContext,AppHandle, &parsed_tree->DerNodeType.ParentNode->ChildrenArray[0], &tbsCertList_ptr);
    if (CSSM_OK != ret)
    {
        cl_der_DeleteParsedDerObject(pContext,AppHandle, parsed_tree);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, parsed_tree);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, return_val);
        return ret;
    }
    return_val->tbsCertList = *tbsCertList_ptr;
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, tbsCertList_ptr);

    /* Get the signature algorithm */
    ret = 
        x509_cl_AlgorithmParseTreeToStruct(pContext,AppHandle, &parsed_tree->DerNodeType.ParentNode->ChildrenArray[1], &temp_sigalg_ptr);
    if (CSSM_OK != ret)
    {
        cl_der_DeleteParsedDerObject(pContext,AppHandle, parsed_tree);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, parsed_tree);
        CSSM_CLUTIL_X509_FreeTbsCertListData(pContext,AppHandle, &return_val->tbsCertList);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, return_val);
        return ret;
    }
    return_val->signature.algorithmIdentifier = *temp_sigalg_ptr;
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, temp_sigalg_ptr);

    /* get signature date */
    return_val->signature.encrypted.Length =
        parsed_tree->DerNodeType.ParentNode->ChildrenArray[2].DerNodeType.LeafNode->Data.Length;
    if (return_val->signature.encrypted.Length)
    {
        if (return_val->signature.encrypted.Length %8)
        {
            return_val->signature.encrypted.Length = return_val->signature.encrypted.Length/8 +1;
        }
        else
        {
            return_val->signature.encrypted.Length = return_val->signature.encrypted.Length/8;
        }

        return_val->signature.encrypted.Data = 
            CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, return_val->signature.encrypted.Length);
        if (!return_val->signature.encrypted.Data)
        {
            cl_der_DeleteParsedDerObject(pContext,AppHandle, parsed_tree);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, parsed_tree);
            CSSM_CLUTIL_X509_FreeX509V2SignedCrl(pContext,AppHandle, return_val);
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
        port_memcpy(return_val->signature.encrypted.Data,
                    parsed_tree->DerNodeType.ParentNode->ChildrenArray[2].DerNodeType.LeafNode->Data.Data,
                    return_val->signature.encrypted.Length);
    }
    else
    {
        return_val->signature.encrypted.Data = NULL;
    }

    /* cleanup */
    cl_der_DeleteParsedDerObject(pContext,AppHandle, parsed_tree);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, parsed_tree);
    
    *result_val = return_val;
    return ret;
}
