/*-----------------------------------------------------------------------
 * File: MISC_NAM.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */



#include "cssmport.h"
#include "clutil.h"
#include "oidsattr.h"
#include "internal.h"
#include <cssmMemTrack.h>


static CSSM_RETURN
CreateOidLdapTable(const CSSM_CLUTIL_CONTEXT *pContext,
                   CSSM_HANDLE AppHandle,
                   CSSM_OID_LDAP_PAIR_PTR *OidLdapTablePtr,
                   uint32 *NumOidLdapEntries);

static CSSM_RETURN
DeleteOidLdapTable(const CSSM_CLUTIL_CONTEXT *pContext,
                   CSSM_HANDLE AppHandle,
                   uint32 NumOidLdapEntries,
                   CSSM_OID_LDAP_PAIR_PTR *OidLdapTablePtr);


static CSSM_RETURN
AddValueToTable(const CSSM_CLUTIL_CONTEXT *pContext,
                CSSM_HANDLE AppHandle,
                CSSM_DATA Value,
                CSSM_BOOL IsValueLDAPFormatted,
                CSSM_OID_LDAP_PAIR_PTR OidLdapPair);

static CSSM_RETURN
GetOneLDAPValueFromTable(const CSSM_CLUTIL_CONTEXT *pContext,
                         CSSM_HANDLE AppHandle,
                         CSSM_OID_LDAP_PAIR OidLdapPair,
                         uint32 ValueNumberToGet,
                         CSSM_DATA_PTR LDAPValue);

static CSSM_RETURN
AddValueToLDAPString(const CSSM_CLUTIL_CONTEXT *pContext,
                     CSSM_HANDLE AppHandle,
                     CSSM_DATA Value,
                     CSSM_DATA TypeLDAP,
                     CSSM_DATA_PTR LDAPString);

static CSSM_RETURN
LDAPFormatAString(const CSSM_CLUTIL_CONTEXT *pContext,
                  CSSM_HANDLE AppHandle,
                  CSSM_DATA Value,
                  CSSM_DATA_PTR LDAPValue);

static CSSM_RETURN
UnformatAnLDAPString(const CSSM_CLUTIL_CONTEXT *pContext,
                     CSSM_HANDLE AppHandle,
                     CSSM_DATA LDAPValue,
                     CSSM_DATA_PTR Value);

static CSSM_RETURN
CountSpecialLDAPChars(const CSSM_CLUTIL_CONTEXT *pContext,
                      CSSM_DATA Value,
                      uint32 *NumSpecialChars);

static CSSM_RETURN
DuplicateOID(const CSSM_CLUTIL_CONTEXT *pContext,
             CSSM_HANDLE AppHandle,
             CSSM_OID SrcOID,
             CSSM_OID_PTR DestOID);

static CSSM_RETURN
GetNumLdapValues(const CSSM_CLUTIL_CONTEXT *pContext,
                 const CSSM_DATA *LDAPName,
                 uint32 *NumValues);

static CSSM_RETURN
GetLDAPTypeValueInfo(const CSSM_CLUTIL_CONTEXT *pContext,
                     CSSM_HANDLE AppHandle,
                     const CSSM_DATA *LDAPName,
                     uint32 TypeValueNumber,
                     CSSM_DATA_PTR LDAPType,
                     CSSM_DATA_PTR LDAPValue);



/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_DuplicateCNameStruct
 *
 * Description:
 * This function duplicates an CSSM_X509_NAME structure
 *
 * Parameters: 
 * AppHandle (input) : handle to be used in memory allocation and freeing
 * SrcName   (input) : CSSM_X509_NAME structure to duplicate
 *
 * result_val(output):
 * Success: CSSM_X509_NAME_PTR to the duplicated tree
 * Failure: NULL
 * 
 *---------------------------------------------------------------------------*/


CSSM_RETURN CSSM_CLUTIL_X509_DuplicateCNameStruct(const CSSM_CLUTIL_CONTEXT *pContext,
                                      CSSM_HANDLE AppHandle, 
                                      const CSSM_X509_NAME *SrcName,
                                      CSSM_X509_NAME_PTR *result_val)
{
    CSSM_X509_NAME_PTR DestName = NULL;
    uint32 i=0, j=0;
    CSSM_BOOL Error;
    *result_val = NULL;
        /* validate parameters */
        if((NULL == SrcName) ||
           (port_IsBadReadPtr(SrcName,
                              sizeof(CSSM_X509_NAME_PTR))) ||
           (port_IsBadReadPtr(SrcName->RelativeDistinguishedName,
                              SrcName->numberOfRDNs * sizeof(CSSM_X509_RDN))))
        {
            return CSSMERR_CL_INVALID_INPUT_POINTER;
        }
        else
        {
            /* allocate a CSSM_X509_NAME for a return value */
            DestName = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                   1,
                                                   sizeof(CSSM_X509_NAME));
            if(!DestName)
            {
                return CSSMERR_CL_MEMORY_ERROR;
            }
            else
            {
                /* set the number of RDNs and allocate their array */
                DestName->numberOfRDNs = SrcName->numberOfRDNs;
                DestName->RelativeDistinguishedName = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                                  1,
                                                                                  (sizeof(CSSM_X509_RDN) * SrcName->numberOfRDNs));
                /* check for memory errors */
                if(!DestName->RelativeDistinguishedName)
                {
                    CSSM_CLUTIL_X509_FreeName(pContext,AppHandle,
                                              DestName);
                    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,
                                              (void *)DestName);
                    DestName = NULL;
                    return CSSMERR_CL_MEMORY_ERROR;
                }
                else
                {
                    i = 0;
                    Error = CSSM_FALSE;
                    /* For each RDN fill in the info */
                    while((CSSM_FALSE == Error) &&
                          (i < SrcName->numberOfRDNs))
                    {
                        DestName->RelativeDistinguishedName[i].numberOfPairs = SrcName->RelativeDistinguishedName[i].numberOfPairs;
                        DestName->RelativeDistinguishedName[i].AttributeTypeAndValue = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                                                                   1,
                                                                                                                   (sizeof(CSSM_X509_TYPE_VALUE_PAIR) * SrcName->RelativeDistinguishedName[i].numberOfPairs));
                        if(!DestName->RelativeDistinguishedName[i].AttributeTypeAndValue)
                        {
                            CSSM_CLUTIL_X509_FreeName(pContext,AppHandle,
                                                      DestName);
                            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,
                                                      (void *)DestName);
                            DestName = NULL;
                            Error = CSSM_TRUE;
                            return CSSMERR_CL_MEMORY_ERROR;
                        }
                        else
                        {
                            j = 0;
                            /* For each attribute/value pair */
                            while((CSSM_FALSE == Error) &&
                                  (j < SrcName->RelativeDistinguishedName[i].numberOfPairs))
                            {
                                /* copy the type */
                                DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Length = SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Length;
                                DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                                                                                        1,
                                                                                                                                        SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Length);
    
                                /* if there was an error, stop */
                                if(!DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Data)
                                {
                                    CSSM_CLUTIL_X509_FreeName(pContext,AppHandle,
                                                              DestName);
                                    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,
                                                              (void *)DestName);
                                    DestName = NULL;
                                    Error = CSSM_TRUE;
                                    return CSSMERR_CL_MEMORY_ERROR;
                                }
                                else
                                {
                                    port_memcpy(DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Data,
                                           SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Data,
                                           SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Length);


                                    /* copy the valueType */
                                    DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].valueType = 
                                        SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].valueType;


                                    /* copy the value */
                                    DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Length = SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Length;
                                    DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                                                                                             1,
                                                                                                                                             SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Length);
                                    if(!DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Data)
                                    {
                                        CSSM_CLUTIL_X509_FreeName(pContext,AppHandle,
                                                                  DestName);
                                        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,
                                                                  (void *)DestName);
                                        DestName = NULL;
                                        return CSSMERR_CL_MEMORY_ERROR;
                                    }
                                    else
                                    {
                                        port_memcpy(DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Data,
                                               SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Data,
                                               SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Length);
                                    }/* end else */
                                }/* end else */
                                j++;
                            }/* end while */
                        }/* end else */
                        i++;
                    }/* end while */
                }/* end else */
            }/* end else */
        }/* end else */
   
   *result_val =  DestName;
   return CSSM_OK;
}



/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_CNameStructToLDAPString
 *
 * Description:
 * This function takes an CSSM_X509_NAME structure and constructs its LDAP string
 * equivalent.
 *
 * Parameters: 
 * AppHandle (input) : handle to be used in memory allocation and freeing
 * x509Name  (input) : CSSM_X509_NAME structure to convert
 * LDAPName  (output): CSSM_X509_NAME converted to an LDAP string
 *
 * Return value:
 * Success: CSSM_OK
 * Failure: Error Code
 * 
 *---------------------------------------------------------------------------*/

CSSM_RETURN
CSSM_CLUTIL_X509_CNameStructToLDAPString(const CSSM_CLUTIL_CONTEXT *pContext,
                                         CSSM_HANDLE AppHandle,
                                         const CSSM_X509_NAME *x509Name,
                                         CSSM_DATA_PTR *LDAPName)
{
    CSSM_OID_LDAP_PAIR_PTR OidLdapTablePtr = NULL;
    uint32 NumOidLdapEntries = 0;
    CSSM_RETURN Ret = CSSM_OK;
    CSSM_RETURN tempRet;
    sint32 i;
    uint32 j, k;
    CSSM_DATA_PTR tempLDAPName = NULL;
    CSSM_BOOL ValueAdded;
    CSSM_DATA tempLDAPValue;

   
    /* we'll change this if all goes well */
    *LDAPName = NULL;

    /* create the OID/LDAP table with initialized values */
    Ret = CreateOidLdapTable(pContext,AppHandle,
                             &OidLdapTablePtr,
                             &NumOidLdapEntries);

    /* fill in the table with values from the CSSM_X509_NAME structure */
    if(CSSM_OK == Ret)
    {
        /* loop through the CSSM_X509_NAME structure */
        i = 0;
        while((CSSM_OK == Ret) &&
              (i < (sint32)x509Name->numberOfRDNs))
        {
            j = 0;
            while((CSSM_OK == Ret) &&
                  (j < x509Name->RelativeDistinguishedName[i].numberOfPairs))
            {
                /* add the value in the correct table location */
                ValueAdded = CSSM_FALSE;
                k = 0;
                while((CSSM_FALSE == ValueAdded) &&
                      (k < NumOidLdapEntries) &&
                      (CSSM_OK == Ret))
                {
                    if(cssm_clutil_IsEquivalentCssmData((CSSM_DATA_PTR)&(x509Name->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type),
                                                        &(OidLdapTablePtr[k].TypeOID)))
                    {
                        Ret = AddValueToTable(pContext,
                                              AppHandle,
                                              x509Name->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value,
                                              CSSM_FALSE,
                                              &(OidLdapTablePtr[k]));
                        if(CSSM_OK == Ret)
                            ValueAdded = CSSM_TRUE;
                    }
                    else
                    {
                        k++;
                    }
                }

                /* did we fail because we don't support this OID */
                if(k == NumOidLdapEntries)
                {
                    Ret = CSSMERR_CL_UNKNOWN_TAG;
                }

                j++;
            }   /* end while */
            i++;
        }   /* end while */
      

        /* if there's no error, let's construct the return LDAP string now */
        if(CSSM_OK == Ret)
        {
            tempLDAPName = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                       1,
                                                       sizeof(CSSM_DATA));
            if(NULL == tempLDAPName)
            {
                Ret = CSSMERR_CL_MEMORY_ERROR;
            }
            else
            {
                i = 0;
                while((CSSM_OK == Ret) &&
                      (i < (sint32)NumOidLdapEntries))
                {
                    j = 0;
                    tempLDAPValue.Length = 0;
                    tempLDAPValue.Data = NULL;
                    while((CSSM_OK == Ret) &&
                          (j < OidLdapTablePtr[i].NumLDAPValues))
                    {
                        /* get the value we're about to add */
                        Ret = GetOneLDAPValueFromTable(pContext,AppHandle,
                                                       OidLdapTablePtr[i],
                                                       j,
                                                       &tempLDAPValue);

                        /* add the value to the output string */
                        if(CSSM_OK == Ret)
                        {
                            Ret = AddValueToLDAPString(pContext,AppHandle,
                                                       tempLDAPValue,
                                                       OidLdapTablePtr[i].TypeLDAP,
                                                       tempLDAPName);
                        }

                        if (tempLDAPValue.Data != NULL)
                        {
                            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,
                                                tempLDAPValue.Data);
                        }
                        tempLDAPValue.Length = 0;
                        tempLDAPValue.Data = NULL;

                        j++;
                    }   /* end while */
                    i++;
                }   /* end while */

                /* if we've had no errors, we need to change the last position
                   of the string from a comma to a NULL character
                */
                if(CSSM_OK == Ret)
                    tempLDAPName->Data[tempLDAPName->Length-1] = 0;

            }   /* end else */
        }   /* end if */

        /* if we allocate memory for the table, we need to free it */
        if(NULL != OidLdapTablePtr)
        {
            tempRet = DeleteOidLdapTable(pContext,AppHandle,
                                         NumOidLdapEntries,
                                         &OidLdapTablePtr);
            if(CSSM_OK == Ret)
                Ret = tempRet;
        }
    }   /* end if */
    
    /* prepare for return */
    if(CSSM_OK == Ret)
    {
        *LDAPName = tempLDAPName;
    }
    else
    {
        if(tempLDAPName)
        {
            if(tempLDAPName->Data)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,
                                          tempLDAPName->Data);
            }
        }
    }
    return Ret;
}



/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_LDAPStringToCNameStruct
 *
 * Description:
 * This function takes an LDAP string DN and constructs its CSSM_X509_NAME structure
 * equivalent.
 *
 * Parameters: 
 * AppHandle (input) : handle to be used in memory allocation and freeing
 * LDAPName  (input) : LDAP string name to convert
 * x509Name  (output): LDAP string converted to an CSSM_X509_NAME
 *
 * Return value:
 * Success: CSSM_OK
 * Failure: Error Code
 * 
 *---------------------------------------------------------------------------*/

CSSM_RETURN
CSSM_CLUTIL_X509_LDAPStringToCNameStruct(const CSSM_CLUTIL_CONTEXT *pContext,
                                         CSSM_HANDLE AppHandle,
                                         const CSSM_DATA *LDAPName,
                                         CSSM_X509_NAME_PTR *CName)
{
    CSSM_RETURN Ret = CSSM_OK;
    CSSM_OID_LDAP_PAIR_PTR OidLdapTablePtr = NULL;
    uint32 NumOidLdapEntries;
    uint32 NumValues;
    sint32 i;
    uint32 j, k;
    CSSM_DATA LDAPType;
    CSSM_DATA LDAPValue;

  
        /* this will be changed if there are no errors */
        *CName = NULL;

        Ret = CreateOidLdapTable(pContext,AppHandle,
                                 &OidLdapTablePtr,
                                 &NumOidLdapEntries);
        if(CSSM_OK == Ret)
        {
            Ret = GetNumLdapValues(pContext,LDAPName,
                                   &NumValues);
            if(CSSM_OK == Ret)
            {
                i = 0;
                while((CSSM_OK == Ret) &&
                      (i < (sint32)NumValues))
                {
                    Ret = GetLDAPTypeValueInfo(pContext,AppHandle,
                                               LDAPName,
                                               i,
                                               &LDAPType,
                                               &LDAPValue);

                    j = 0;
                    while((j < NumOidLdapEntries) &&
                          (cssm_clutil_IsEquivalentCssmData(&LDAPType,
                                                            &(OidLdapTablePtr[j].TypeLDAP)) == CSSM_FALSE))
                    {
                        j++;
                    }
                    if (LDAPType.Data)
                    {
                        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,
                                                  LDAPType.Data);
                        LDAPType.Length = 0;
                        LDAPType.Data = NULL;
                    }

                    if(j == NumOidLdapEntries)
                    {
                        Ret = CSSMERR_CL_INVALID_DATA;
                    }
                    else
                    {
                        Ret = AddValueToTable(pContext,AppHandle,
                                              LDAPValue,
                                              CSSM_TRUE,
                                              &OidLdapTablePtr[j]);

                        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,
                                                  LDAPValue.Data);
                    }
                    i++;
                }
            }
     
        if(CSSM_OK == Ret)
        {
            *CName = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                 1,
                                                 sizeof(CSSM_X509_NAME));
            if(*CName)
            {
                (*CName)->numberOfRDNs = NumValues;
                if (NumValues)
                {
                    (*CName)->RelativeDistinguishedName = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                                      1,
                                                                                      sizeof(CSSM_X509_RDN)*NumValues);
                    if((*CName)->RelativeDistinguishedName)
                    {
                        i = 0;
                        while((i < (sint32)NumValues) &&
                              (CSSM_OK == Ret))
                        {
                            (*CName)->RelativeDistinguishedName[i].AttributeTypeAndValue = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                                                                       1,
                                                                                                                       sizeof(CSSM_X509_TYPE_VALUE_PAIR));
                            if((*CName)->RelativeDistinguishedName[i].AttributeTypeAndValue)
                                (*CName)->RelativeDistinguishedName[i].numberOfPairs = 1;
                            else
                            {
                                Ret = CSSMERR_CL_MEMORY_ERROR;
                            
                            }
                            i++;
                        }
                    }
                    else
                    {
                        Ret = CSSMERR_CL_MEMORY_ERROR;
                    
                    }
                }
            }
            else
            {
                Ret = CSSMERR_CL_MEMORY_ERROR;
                
            }
        }

        if(CSSM_OK == Ret)
        {
            i = NumOidLdapEntries - 1;
            k = 0;
            while((CSSM_OK == Ret) &&
                  (0 <= i))
            {
                j = 0;
                while((CSSM_OK == Ret) &&
                      (j < OidLdapTablePtr[i].NumLDAPValues))
                {
                    Ret = DuplicateOID(pContext,AppHandle,
                                       OidLdapTablePtr[i].TypeOID,
                                       &((*CName)->RelativeDistinguishedName[k].AttributeTypeAndValue[0].type));
                    if(CSSM_OK == Ret)
                    {
                        Ret = GetOneLDAPValueFromTable(pContext,AppHandle,
                                                       OidLdapTablePtr[i],
                                                       j,
                                                       &LDAPValue);
                        if(CSSM_OK == Ret)
                        {
                            (*CName)->RelativeDistinguishedName[k].AttributeTypeAndValue[0].valueType = DER_TAG_PRINTABLE_STRING;
                            Ret = UnformatAnLDAPString(pContext,AppHandle,
                                                       LDAPValue,
                                                       &((*CName)->RelativeDistinguishedName[k].AttributeTypeAndValue[0].value));

                            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,
                                                      LDAPValue.Data);
                            k++;
                        }
                    }
                    j++;
                }
                i--;
            }
        }

        if(NULL != OidLdapTablePtr)
        {
            Ret = DeleteOidLdapTable(pContext,AppHandle,
                                     NumOidLdapEntries,
                                     &OidLdapTablePtr);
        }
    }
    return Ret;
}



/*-----------------------------------------------------------------------------
 * Name: CreateOidLdapTable
 *
 * Description:
 * This function constructs a table that matches OIDs to an LDAP type string.
 * It is intended that this table then be used for the parsing of CSSM_X509_NAME and
 * LDAP string names.  Memory allocated by this function is freed most easily
 * by calling DeleteOidLdapTable.
 *
 * Parameters: 
 * AppHandle          (input) : handle to be used in memory allocation and freeing
 * OidLdapTablePtr   (output) : newly allocated and filled in table
 * NumOidLdpatEntries (output): number of elements in the OidLdapTable
 *
 * Return value:
 * Success: CSSM_OK
 * Failure: Error Code
 * 
 *---------------------------------------------------------------------------*/

static CSSM_RETURN
CreateOidLdapTable(const CSSM_CLUTIL_CONTEXT *pContext,
                   CSSM_HANDLE AppHandle,
                   CSSM_OID_LDAP_PAIR_PTR *OidLdapTablePtr,
                   uint32 *NumOidLdapEntries)
{
    CSSM_RETURN Ret = CSSM_OK;
    uint32 i = 0;
    uint32 j = 0;        
   
        /* set the number of attributes and allocate memory for the table */
        *NumOidLdapEntries = NUMBER_OF_NAME_ATTRIBUTE_OIDS;
        *OidLdapTablePtr = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                       1,
                                                       sizeof(CSSM_OID_LDAP_PAIR) * *NumOidLdapEntries);

        if (!*OidLdapTablePtr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
        }
        /*** now we set the OID and LDAP information for each entry ***/

        i = 0;
        /* Common Name */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_CommonName.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_CommonName.Length);

        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_CommonName.Data,
               CSSMOID_CommonName.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("CN");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("CN"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "CN",
               strlen("CN"));

        i++;

        /* Initials */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_Initials.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_Initials.Length);

        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_Initials.Data,
               CSSMOID_Initials.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("initials");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("initials"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "initials",
               strlen("initials"));
        i++;
        /* Given Name */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_GivenName.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_GivenName.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_GivenName.Data,
               CSSMOID_GivenName.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("givenName");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("givenName"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "givenName",
               strlen("givenName"));

        i++;

        /* Surname */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_Surname.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_Surname.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_Surname.Data,
               CSSMOID_Surname.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("surname");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("surname"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "surname",
               strlen("surname"));

        i++;
        /* Name */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_Name.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_Name.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_Name.Data,
               CSSMOID_Name.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("name");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("name"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "name",
               strlen("name"));
        i++;
        /* Title */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_Title.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_Title.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_Title.Data,
               CSSMOID_Title.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("title");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("title"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "title",
               strlen("title"));
        i++;
        /* Description */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_Description.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_Description.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_Description.Data,
               CSSMOID_Description.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("description");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("description"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "description",
               strlen("description"));

        i++;
        /* Organizational Unit Name */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_OrganizationalUnitName.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_OrganizationalUnitName.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_OrganizationalUnitName.Data,
               CSSMOID_OrganizationalUnitName.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("OU");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("OU"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "OU",
               strlen("OU"));

        i++;
        /* Collective Organizational Unit Name */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_CollectiveOrganizationalUnitName.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_CollectiveOrganizationalUnitName.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_CollectiveOrganizationalUnitName.Data,
               CSSMOID_CollectiveOrganizationalUnitName.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("collectiveOrganizationalUnitName");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("collectiveOrganizationalUnitName"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "collectiveOrganizationalUnitName",
               strlen("collectiveOrganizationalUnitName"));
        i++;
        /* Organization Name */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_OrganizationName.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_OrganizationName.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_OrganizationName.Data,
               CSSMOID_OrganizationName.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("O");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("O"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "O",
               strlen("O"));
        i++;
        /* Collective Organization Name */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_CollectiveOrganizationName.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_CollectiveOrganizationName.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_CollectiveOrganizationName.Data,
               CSSMOID_CollectiveOrganizationName.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("collectiveOrganizationName");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("collectiveOrganizationName"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "collectiveOrganizationName",
               strlen("collectiveOrganizationName"));
        i++;
        /* Post Office Box */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_PostOfficeBox.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_PostOfficeBox.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_PostOfficeBox.Data,
               CSSMOID_PostOfficeBox.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("postOfficeBox");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("postOfficeBox"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "postOfficeBox",
               strlen("postOfficeBox"));
        i++;
        /* Collective Post Office Box */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_CollectivePostOfficeBox.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_CollectivePostOfficeBox.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_CollectivePostOfficeBox.Data,
               CSSMOID_CollectivePostOfficeBox.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("collectivePostOfficeBox");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("collectivePostOfficeBox"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "collectivePostOfficeBox",
               strlen("collectivePostOfficeBox"));
        i++;
        /* Physical Delivery Office Name */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_PhysicalDeliveryOfficeName.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_PhysicalDeliveryOfficeName.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_PhysicalDeliveryOfficeName.Data,
               CSSMOID_PhysicalDeliveryOfficeName.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("physicalDeliveryOfficeName");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("physicalDeliveryOfficeName"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "physicalDeliveryOfficeName",
               strlen("physicalDeliveryOfficeName"));
        i++;

        /* Collective Physical Delivery Office Name */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_CollectivePhysicalDeliveryOfficeName.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_CollectivePhysicalDeliveryOfficeName.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_CollectivePhysicalDeliveryOfficeName.Data,
               CSSMOID_CollectivePhysicalDeliveryOfficeName.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("collectivePhysicalDeliveryOfficeName");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("collectivePhysicalDeliveryOfficeName"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "collectivePhysicalDeliveryOfficeName",
               strlen("collectivePhysicalDeliveryOfficeName"));
        i++;
        /* Telephone Number */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_TelephoneNumber.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_TelephoneNumber.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_TelephoneNumber.Data,
               CSSMOID_TelephoneNumber.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("telephoneNumber");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("telephoneNumber"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "telephoneNumber",
               strlen("telephoneNumber"));
        i++;
        /* Collective Telephone Number */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_CollectiveTelephoneNumber.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_CollectiveTelephoneNumber.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_CollectiveTelephoneNumber.Data,
               CSSMOID_CollectiveTelephoneNumber.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("collectiveTelephoneNumber");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("collectiveTelephoneNumber"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "collectiveTelephoneNumber",
               strlen("collectiveTelephoneNumber"));
        i++;
        /* Telex Number */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_TelexNumber.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_TelexNumber.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_TelexNumber.Data,
               CSSMOID_TelexNumber.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("telexNumber");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("telexNumber"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "telexNumber",
               strlen("telexNumber"));
        i++;
        /* Collective Telex Number */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_CollectiveTelexNumber.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_CollectiveTelexNumber.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_CollectiveTelexNumber.Data,
               CSSMOID_CollectiveTelexNumber.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("collectiveTelexNumber");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("collectiveTelexNumber"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "collectiveTelexNumber",
               strlen("collectiveTelexNumber"));
        i++;
        /* Telex Terminal Identifier */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_TelexTerminalIdentifier.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_TelexTerminalIdentifier.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_TelexTerminalIdentifier.Data,
               CSSMOID_TelexTerminalIdentifier.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("telexTerminalIdentifier");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("telexTerminalIdentifier"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "telexTerminalIdentifier",
               strlen("telexTerminalIdentifier"));
        i++;
        /* Collective Telex Terminal Identifier */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_CollectiveTelexTerminalIdentifier.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_CollectiveTelexTerminalIdentifier.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_CollectiveTelexTerminalIdentifier.Data,
               CSSMOID_CollectiveTelexTerminalIdentifier.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("collectiveTelexTerminalIdentifier");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("collectiveTelexTerminalIdentifier"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "collectiveTelexTerminalIdentifier",
               strlen("collectiveTelexTerminalIdentifier"));
        i++;
        /* Facsimile Telephone Number */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_FacsimileTelephoneNumber.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_FacsimileTelephoneNumber.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_FacsimileTelephoneNumber.Data,
               CSSMOID_FacsimileTelephoneNumber.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("facsimileTelephoneNumber");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("facsimileTelephoneNumber"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "facsimileTelephoneNumber",
               strlen("facsimileTelephoneNumber"));
        i++;
        /* Collective Facsimile Telephone Number */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_CollectiveFacsimileTelephoneNumber.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_CollectiveFacsimileTelephoneNumber.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_CollectiveFacsimileTelephoneNumber.Data,
               CSSMOID_CollectiveFacsimileTelephoneNumber.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("collectiveFacsimileTelephoneNumber");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("collectiveFacsimileTelephoneNumber"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "collectiveFacsimileTelephoneNumber",
               strlen("collectiveFacsimileTelephoneNumber"));
        i++;
        /* Object Class */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_ObjectClass.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_ObjectClass.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_ObjectClass.Data,
               CSSMOID_ObjectClass.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("objectClass");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("objectClass"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "objectClass",
               strlen("objectClass"));
        i++;
        /* Aliased Entry Name */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_AliasedEntryName.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_AliasedEntryName.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_AliasedEntryName.Data,
               CSSMOID_AliasedEntryName.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("aliasedEntryName");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("aliasedEntryName"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "aliasedEntryName",
               strlen("aliasedEntryName"));
        i++;
        /* Knowledge Information */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_KnowledgeInformation.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_KnowledgeInformation.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_KnowledgeInformation.Data,
               CSSMOID_KnowledgeInformation.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("knowledgeInformation");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("knowledgeInformation"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "knowledgeInformation",
               strlen("knowledgeInformation"));
        i++;
        /* Serial Number */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_SerialNumber.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_SerialNumber.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_SerialNumber.Data,
               CSSMOID_SerialNumber.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("serialNumber");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("serialNumber"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "serialNumber",
               strlen("serialNumber"));
        i++;
        /* Street Address */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_StreetAddress.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_StreetAddress.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_StreetAddress.Data,
               CSSMOID_StreetAddress.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("STREET");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("STREET"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "STREET",
               strlen("STREET"));
        i++;
        /* Collective Street Address */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_CollectiveStreetAddress.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_CollectiveStreetAddress.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_CollectiveStreetAddress.Data,
               CSSMOID_CollectiveStreetAddress.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("collectiveStreetAddress");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("collectiveStreetAddress"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "collectiveStreetAddress",
               strlen("collectiveStreetAddress"));
        i++;
        /* Locality Name */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_LocalityName.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_LocalityName.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_LocalityName.Data,
               CSSMOID_LocalityName.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("L");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("L"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "L",
               strlen("L"));
        i++;
        /* State Or Province Name */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_StateProvinceName.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_StateProvinceName.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_StateProvinceName.Data,
               CSSMOID_StateProvinceName.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("ST");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("ST"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "ST",
               strlen("ST"));
        i++;
        /* Country Name */
        (*OidLdapTablePtr)[i].TypeOID.Length = CSSMOID_CountryName.Length;
        (*OidLdapTablePtr)[i].TypeOID.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          CSSMOID_CountryName.Length);
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }

            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy(((*OidLdapTablePtr)[i].TypeOID.Data),
               CSSMOID_CountryName.Data,
               CSSMOID_CountryName.Length);
        (*OidLdapTablePtr)[i].TypeLDAP.Length = strlen("C");
        (*OidLdapTablePtr)[i].TypeLDAP.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                                          1,
                                                                          strlen("C"));
        if (!(*OidLdapTablePtr)[i].TypeOID.Data)
        {
            for(j=0;j<i;j++)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeOID.Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[j].TypeLDAP.Data);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*OidLdapTablePtr)[i].TypeOID.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *OidLdapTablePtr);
            *NumOidLdapEntries = 0;
            return CSSMERR_CL_MEMORY_ERROR;
        }

        port_memcpy((*OidLdapTablePtr)[i].TypeLDAP.Data,
               "C",
               strlen("C"));
   
    return Ret;
}



/*-----------------------------------------------------------------------------
 * Name: AddValueToTable
 *
 * Description:
 * Function adds the LDAP formatted string of the value specified in 
 * Value.Data to OidLdapPair.  IsValueLDAPFormatted is used to deterine if the
 * value is already LDAP formatted or if it needs to be formatted.  If there
 * is at least one value already present in OidLdapPair this new value will be
 * separated from the others by a comma.
 *
 * Parameters: 
 * AppHandle           (input) : handle to be used in memory allocation and 
 *                                  freeing
 * Value               (input) : value to be added to the CSSM_OID_LDAP_PAIR
 * IsValueLDAPFormatted (input): is the value already formatted, or does
 *                                  formatting need to be done
 * OidLdapPair        (oputput): CSSM_OID_LDAP_PAIR to which the value will be
 *                                  added
 *
 * Return value:
 * Success: CSSM_OK
 * Failure: Error Code
 * 
 *---------------------------------------------------------------------------*/

static CSSM_RETURN
AddValueToTable(const CSSM_CLUTIL_CONTEXT *pContext,
                CSSM_HANDLE AppHandle,
                CSSM_DATA Value,
                CSSM_BOOL IsValueLDAPFormatted,
                CSSM_OID_LDAP_PAIR_PTR OidLdapPair)
{
    CSSM_RETURN Ret = CSSM_OK;
    CSSM_DATA tempValue = {0, NULL};
    uint32 size;

   
    /* if needed, we put the value in LDAP formatting */
    if((CSSM_FALSE == IsValueLDAPFormatted))
    {
        Ret = LDAPFormatAString(pContext,AppHandle,
                                Value,
                                &tempValue);
    }
    else
    {
        tempValue.Length = Value.Length;
        tempValue.Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                     1,
                                                     Value.Length);
        if(tempValue.Data)
        {
            port_memcpy(tempValue.Data,
                   Value.Data,
                   Value.Length);
        }
        else
        {
            Ret = CSSMERR_CL_MEMORY_ERROR;
            
        }
    }

    /* add the value to the table */
    if(CSSM_OK == Ret)
    {
        /* compute the plain size */
        size = OidLdapPair->LDAPValues.Length + tempValue.Length;

        /* see if we need to add a comma */
        if(0 < OidLdapPair->NumLDAPValues)
            size++;

        OidLdapPair->LDAPValues.Data = CSSM_REALLOC_WRAPPER(pContext->MemFuncs.realloc_func, AppHandle,
                                                                    OidLdapPair->LDAPValues.Data,
                                                                    size);

        if(OidLdapPair->LDAPValues.Data)
        {
            /* copy the LDAP formatted string - check for adding commas */
            if(size == tempValue.Length)
            {
                port_memcpy(OidLdapPair->LDAPValues.Data,
                       tempValue.Data,
                       tempValue.Length);
                OidLdapPair->LDAPValues.Length = size;
            }
            else
            {
                port_memcpy(&(OidLdapPair->LDAPValues.Data[OidLdapPair->LDAPValues.Length++]),
                       ",",
                       sizeof(","));
                port_memcpy(&(OidLdapPair->LDAPValues.Data[OidLdapPair->LDAPValues.Length]),
                       tempValue.Data,
                       tempValue.Length);
                OidLdapPair->LDAPValues.Length = size;
            }

            /* increment the number of values we have */
            OidLdapPair->NumLDAPValues++;
        }
        else
        {
            if(NULL != tempValue.Data)
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,
                                          tempValue.Data);

            
            Ret = CSSMERR_CL_MEMORY_ERROR;
        }
    }

    if(NULL != tempValue.Data)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, tempValue.Data);
    }

    return Ret;
}



/*-----------------------------------------------------------------------------
 * Name: GetOneLDAPValueFromTable
 *
 * Description:
 * Function gets a single LDAP formatted string out of a CSSM_OID_LDAP_PAIR.
 * This value does not have the "type" attached - in other words, it's not in
 * name format.  ValueNumberToGet specifies which value to get from the
 * CSSM_OID_LDAP_PAIR - values 0 through N are acceptable, where N+1 is the
 * number of values in the CSSM_OID_LDAP_PAIR.
 *
 * Parameters: 
 * AppHandle         (input) : handle to be used in memory allocation and 
 *                                 freeing
 * OidLdapPair       (input) : source containing a list of values - of which
 *                                 we're about to get one
 * ValueNumberToGet   (input): number of the value to get (range 0-(n-1))
 * LDAPValue        (output) : the LDAP formatted value, memory to be freed by
 *                                 the caller
 *
 * Return value:
 * Success: CSSM_OK
 * Failure: Error Code
 * 
 *---------------------------------------------------------------------------*/

static CSSM_RETURN
GetOneLDAPValueFromTable(const CSSM_CLUTIL_CONTEXT *pContext,
                         CSSM_HANDLE AppHandle,
                         CSSM_OID_LDAP_PAIR OidLdapPair,
                         uint32 ValueNumberToGet,
                         CSSM_DATA_PTR LDAPValue)
{
    CSSM_RETURN Ret = CSSM_OK;
    uint32 CurrentValueNumber = 0;
    uint32 ValueLength = 0;
    uint32 CurrentValuesIndex = 0;
    uint32 i = 0;


       
        CurrentValueNumber = 0;

        /* move through values to correct number */
        while((CurrentValueNumber != ValueNumberToGet) &&
              (CurrentValuesIndex < OidLdapPair.LDAPValues.Length))
        {
            /* is there a comma and, if so, does the comma seperate values */
            if((OidLdapPair.LDAPValues.Data[CurrentValuesIndex] == ',') &&
               (OidLdapPair.LDAPValues.Data[CurrentValuesIndex-1] != '"'))
            {
               CurrentValueNumber++;
               /* our index should be at the beginning of the new value */
               CurrentValuesIndex++;
            }
            else
                CurrentValuesIndex++;   /* we'll check the next position */
        }

        /* were the correct number of values given */
        if(CurrentValuesIndex == OidLdapPair.LDAPValues.Length)
        {
            Ret = CSSMERR_CL_INVALID_DATA;
            
        }
        else
        {
            i = CurrentValuesIndex;
            ValueLength = 0;
            
            /* count the number of characters in this value */
            while((i < OidLdapPair.LDAPValues.Length) &&
                  ((OidLdapPair.LDAPValues.Data[i] != ',') ||
                   ((OidLdapPair.LDAPValues.Data[i-1] == '"') &&
                    (OidLdapPair.LDAPValues.Data[i] == ',') &&
                    (OidLdapPair.LDAPValues.Data[i+1] == '"'))))
            {
                i++;
                ValueLength++;
            }

            if(0!= ValueLength)
            {
                /* now allocate that length */
                LDAPValue->Length = ValueLength;
                LDAPValue->Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                              1,
                                                              LDAPValue->Length);

                if(LDAPValue->Data)
                {
                    port_memcpy(LDAPValue->Data,
                           &(OidLdapPair.LDAPValues.Data[CurrentValuesIndex]),
                           LDAPValue->Length);
                }
                else
                {
                    Ret = CSSMERR_CL_MEMORY_ERROR;
                    
                }
            }
        }
    
    return Ret;
}



/*-----------------------------------------------------------------------------
 * Name: AddValueToLDAPString
 *
 * Description:
 * Function takes a type and value and appends them to a string in LDAP format.
 * This string will have a comma ending character.  If this was the end of the
 * string, the comma should be replaced with a null character.
 * Note - this function does not confirm that TypeLDAP is a valid name type or
 *        that LDAPString is currently in a valid state for an LDAP string.
 *
 * Parameters: 
 * AppHandle  (input) : handle to be used in memory allocation and freeing
 * Value      (input) : LDAP formatted value to be added to LDAPString
 * TypeLDAP   (input) : LDAP "value" to be used with Value in the LDAPString
 * LDAPString (output): an LDAP formatted name string - the type/value has been
 *                          added with a comma at the end
 *
 * Return value:
 * Success: CSSM_OK
 * Failure: Error Code
 * 
 *---------------------------------------------------------------------------*/

static CSSM_RETURN
AddValueToLDAPString(const CSSM_CLUTIL_CONTEXT *pContext,
                     CSSM_HANDLE AppHandle,
                     CSSM_DATA Value,
                     CSSM_DATA TypeLDAP,
                     CSSM_DATA_PTR LDAPString)
{
    CSSM_RETURN Ret = CSSM_OK;
    uint32 NewStringLength;

   
        /* new length = value length + type length + previous length +
           3 for " = " + 1 for comma
        */
        NewStringLength = Value.Length + TypeLDAP.Length +
                          LDAPString->Length + 3 + 1;

        LDAPString->Data = CSSM_REALLOC_WRAPPER(pContext->MemFuncs.realloc_func, AppHandle,
                                                        LDAPString->Data,
                                                        NewStringLength);
        if(NULL != LDAPString->Data)
        {
            /* copy the type - i.e. CN */
            port_memcpy(&(LDAPString->Data[LDAPString->Length]),
                   TypeLDAP.Data,
                   TypeLDAP.Length);
            LDAPString->Length += TypeLDAP.Length;

            /* copy the " = " */
            port_memcpy(&(LDAPString->Data[LDAPString->Length]),
                   " = ",
                   strlen(" = "));
            LDAPString->Length = LDAPString->Length + (uint32)strlen(" = ");

            /* copy the value */
            port_memcpy(&(LDAPString->Data[LDAPString->Length]),
                   Value.Data,
                   Value.Length);
            LDAPString->Length += Value.Length;

            /* set the comma */
            LDAPString->Data[LDAPString->Length++] = ',';
        }
        else
        {
            Ret = CSSMERR_CL_MEMORY_ERROR;
            
        }
 
    return Ret;
}
            


/*-----------------------------------------------------------------------------
 * Name: LDAPFormatAString
 *
 * Description:
 * Function takes the string in Value.Data and places the LDAP-formatted
 * equivalent in LDAPValue->Data.  Caller is expected to free LDAPValue->Data.
 *
 * Parameters: 
 * AppHandle  (input) : handle to be used in memory allocation and freeing
 * Value      (input) : a string, possibly null terminated
 * LDAPValue  (output): the LDAP formatted equivalent of Value.Data
 *
 * Return value:
 * Success: CSSM_OK
 * Failure: Error Code
 * 
 *---------------------------------------------------------------------------*/

static CSSM_RETURN
LDAPFormatAString(const CSSM_CLUTIL_CONTEXT *pContext,
                  CSSM_HANDLE AppHandle,
                  CSSM_DATA Value,
                  CSSM_DATA_PTR LDAPValue)
{
    CSSM_RETURN Ret = CSSM_OK;
    uint32 NumSpecialChars;
    uint32 SizeOfNewString;
    uint32 SizeOfOldString;
    uint32 StringIndex = 0;
    uint32 i;
    CSSM_BOOL NullTerminated = CSSM_FALSE;

  
        if((Ret = CountSpecialLDAPChars(pContext,Value,
                                        &NumSpecialChars)) == CSSM_OK)
        {
            /* check size of string - checking if the string is NULL terminated */
            i = 0;
            while((i < Value.Length) &&
                  (CSSM_FALSE == NullTerminated))
            {
                if(0 == Value.Data[i])
                    NullTerminated = CSSM_TRUE;
                else
                    i++;
            }

            SizeOfOldString = i;

            /*  Compute the size of the buffer we need:
                length of type name + 2 for each special character to have 
                quotes 
            */
            SizeOfNewString = SizeOfOldString + 2*NumSpecialChars;

            /* allocate memory for the LDAP formatted string */
            LDAPValue->Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                          1,
                                                          SizeOfNewString);

            LDAPValue->Length = SizeOfNewString;
            StringIndex = 0;

            if(LDAPValue->Data)
            {
                /* set length and initialize string index */
                LDAPValue->Length = SizeOfNewString;
                StringIndex = 0;

                /* copy the value - need loop to check for special 
                   characters
                */
                for(i = 0; i < SizeOfOldString; i++)
                {
                    /* check if we're on a special character */
                    if((Value.Data[i] == ',') ||
                       (Value.Data[i] == '+') ||
                       (Value.Data[i] == '=') ||
                       (Value.Data[i] == '"') ||
                       (Value.Data[i] == '<') ||
                       (Value.Data[i] == '>') ||
                       (Value.Data[i] == '#') ||
                       (Value.Data[i] == ';') ||
                       (Value.Data[i] == '\n'))
                    {
                        LDAPValue->Data[StringIndex++] = '"';
                        LDAPValue->Data[StringIndex++] = Value.Data[i];
                        LDAPValue->Data[StringIndex++] = '"';
                    }
                    else
                        LDAPValue->Data[StringIndex++] = Value.Data[i];
                }
            }
            else
            {
                Ret = CSSMERR_CL_MEMORY_ERROR;
                
            }
        }
  
    return Ret;
}



/*-----------------------------------------------------------------------------
 * Name: UnformantAnLDAPString
 *
 * Description:
 * Function takes the LDAP formatted string in LDAPValue.Data and creates its
 * non-LDAP formatted equivalent in Value->Data.  Caller is responsible for
 * freeing memory allocated in Value->Data.
 *
 * Parameters: 
 * AppHandle (input) : handle to be used in memory allocation and freeing
 * LDAPValue (input) : an LDAP-formatted string
 * Value     (output): the non-LDAP-formatted equivalent of LDAPValue
 *
 * Return value:
 * Success: CSSM_OK
 * Failure: Error Code
 * 
 *---------------------------------------------------------------------------*/

static CSSM_RETURN
UnformatAnLDAPString(const CSSM_CLUTIL_CONTEXT *pContext,
                     CSSM_HANDLE AppHandle,
                     CSSM_DATA LDAPValue,
                     CSSM_DATA_PTR Value)
{
    CSSM_RETURN Ret = CSSM_OK;
    uint32 NumSpecialChars;
    uint32 SizeOfNewString;
    uint32 SizeOfOldString;
    uint32 StringIndex = 0;
    uint32 i;
    CSSM_BOOL NullTerminated = CSSM_FALSE;

        if((Ret = CountSpecialLDAPChars(pContext,LDAPValue,
                                        &NumSpecialChars)) == CSSM_OK)
        {
            /* need to compute the correct number of special chars -
               correct number is 1/3 of what we just got because the string
               was LDAP formatted
            */
            NumSpecialChars = NumSpecialChars / 3;

            /* check size of string - checking if the string is NULL terminated */
            i = 0;
            while((i < LDAPValue.Length) &&
                  (CSSM_FALSE == NullTerminated))
            {
                if(0 == LDAPValue.Data[i])
                    NullTerminated = CSSM_TRUE;
                else
                    i++;
            }

            SizeOfOldString = i;

            /*  Compute the size of the buffer we need:
                length of type name - 2 for each special character to have 
                quotes 
            */
            SizeOfNewString = SizeOfOldString - 2*NumSpecialChars;

            /* allocate memory for the unformatted string */
            Value->Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                      1,
                                                      SizeOfNewString);

            StringIndex = 0;

            if(Value->Data)
            {
                /* set length and initialize string index */
                Value->Length = SizeOfNewString;
                StringIndex = 0;

                /* copy the value - need loop to check for special 
                   characters
                */
                for(i = 0; i < SizeOfOldString; i++)
                {
                    if(LDAPValue.Data[i] == '"')
                    {
                        StringIndex++;
                        Value->Data[StringIndex++] = LDAPValue.Data[i];
                        StringIndex++;
                    }
                    else
                        Value->Data[StringIndex++] = LDAPValue.Data[i];
                }
            }
            else
            {
                Ret = CSSMERR_CL_MEMORY_ERROR;
                
            }
        }
  
    return Ret;
}



/*-----------------------------------------------------------------------------
 * Name: CountSpecialLDAPChars
 *
 * Description:
 * Function counts the number of characters in a string which LDAP considers to
 * be "special" characters.  The string may or may not be null terminated, and
 * may or may not be LDAP formatted.
 *
 * Parameters: 
 * Value           (input) : string to search for special characters
 * NumSpecialChars (output): number of special characters found in Value.Data
 *
 * Return value:
 * Success: CSSM_OK
 * Failure: Error Code
 * 
 *---------------------------------------------------------------------------*/

static CSSM_RETURN
CountSpecialLDAPChars(const CSSM_CLUTIL_CONTEXT *pContext,
                      CSSM_DATA Value,
                      uint32 *NumSpecialChars)
{
    CSSM_RETURN Ret = CSSM_OK;
    uint32 i;
    uint32 SizeOfString;
    CSSM_BOOL NullTerminated = CSSM_FALSE;

  
        /* check size of string - checking if the string is NULL terminated */
        i = 0;
        while((i < Value.Length) &&
              (CSSM_FALSE == NullTerminated))
        {
            if(0 == Value.Data[i])
                NullTerminated = CSSM_TRUE;
            else
                i++;
        }
        SizeOfString = i;

        /* initialize number of special characters counter */
        *NumSpecialChars = 0;
        for(i = 0; i < SizeOfString; i++)
        {
            if((Value.Data[i] == ',') ||
               (Value.Data[i] == '+') ||
               (Value.Data[i] == '=') ||
               (Value.Data[i] == '"') ||
               (Value.Data[i] == '<') ||
               (Value.Data[i] == '>') ||
               (Value.Data[i] == '#') ||
               (Value.Data[i] == ';') ||
               (Value.Data[i] == '\n'))
               (*NumSpecialChars)++;
        }
    
    return Ret;
}



/*-----------------------------------------------------------------------------
 * Name: DeleteOidLdapTable
 *
 * Description:
 * Function frees all memory associated with OidLdapTablePtr.
 *
 * Parameters: 
 * AppHandle              (input) : handle to be used in memory allocation and
 *                                      freeing
 * NumOidLdapEntries      (input) : number of entries in OidLdapTablePtr
 * OidLdapTablePtr (input/output) : CSSM_OID_LDAP_PAIR_PTR to free
 *
 * Return value:
 * Success: CSSM_OK
 * Failure: Error Code
 * 
 *---------------------------------------------------------------------------*/

static CSSM_RETURN
DeleteOidLdapTable(const CSSM_CLUTIL_CONTEXT *pContext,
                   CSSM_HANDLE AppHandle,
                   uint32 NumOidLdapEntries,
                   CSSM_OID_LDAP_PAIR_PTR *OidLdapTablePtr)
{
    CSSM_RETURN Ret = CSSM_OK;
    uint32 i;

   
        for(i = 0; i < NumOidLdapEntries; i++)
        {

            /* if there's something in the values, free it */
            if((*OidLdapTablePtr)[i].LDAPValues.Data != NULL)
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,
                                          (*OidLdapTablePtr)[i].LDAPValues.Data);

            /* we know there's something in the types, so we free them */
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,
                                      (*OidLdapTablePtr)[i].TypeLDAP.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,
                                      (*OidLdapTablePtr)[i].TypeOID.Data);

        }
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,
                                  *OidLdapTablePtr);
        *OidLdapTablePtr = NULL;
   
    return Ret;
}



/*-----------------------------------------------------------------------------
 * Name: DuplicateOID
 *
 * Description:
 * Function duplicates the OID specified by SrcOID and places it in DestOID.
 * The calling function is expected to free all memory associated with DestOID.
 *
 * Parameters: 
 * AppHandle  (input) : handle to be used in memory allocation and freeing
 * SrcOID     (input) : source OID to copy into DestOID
 * DestOID    (output): destination OID - caller frees memory allocated
 *
 * Return value:
 * Success: CSSM_OK
 * Failure: Error Code
 * 
 *---------------------------------------------------------------------------*/

static CSSM_RETURN
DuplicateOID(const CSSM_CLUTIL_CONTEXT *pContext,
             CSSM_HANDLE AppHandle,
             CSSM_OID SrcOID,
             CSSM_OID_PTR DestOID)
{
    CSSM_RETURN Ret = CSSM_OK;

   
        DestOID->Length = SrcOID.Length;
        DestOID->Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                    1,
                                                    DestOID->Length);
        if(DestOID->Data)
        {
            port_memcpy(DestOID->Data,
                   SrcOID.Data,
                   SrcOID.Length);
        }
        else
        {
            Ret = CSSMERR_CL_MEMORY_ERROR;
            
        }
   
    return Ret;
}



/*-----------------------------------------------------------------------------
 * Name: GetNumLdapValues
 *
 * Description:
 * Function counts the number of values in an LDAP-formatted string.
 *  Note - this function does not check for proper LDAP formatting
 *
 * Parameters: 
 * LDAPName   (input) : LDAP-formatted string to check for # of values
 * NumValues  (output): number of values found in LDAPName->Data
 *
 * Return value:
 * Success: CSSM_OK
 * Failure: Error Code
 * 
 *---------------------------------------------------------------------------*/

static CSSM_RETURN
GetNumLdapValues(const CSSM_CLUTIL_CONTEXT *pContext,
                 const CSSM_DATA *LDAPName,
                 uint32 *NumValues)
{
    CSSM_RETURN Ret = CSSM_OK;
    uint32 i;
    uint32 NumQuotes;

   
        /* initialize count */
        *NumValues = 0;
        NumQuotes = 0;

        /* loop to check */
        for(i = 0; i < LDAPName->Length; i++)
        {
            /* only a distinct value if there's a comma without quotes around it 
               or this was the NULL termination of the string
            */
            if(((LDAPName->Data[i] == ',') && (0 == NumQuotes)) ||
               (LDAPName->Data[i] == 0))
            {
               (*NumValues)++;
            }
            else
            {
                /* see what sort of situation quotes may put us in */
                if((LDAPName->Data[i] == '"') &&
                   (0 == NumQuotes))
                {
                    NumQuotes++;
                }
                else
                {
                    if(LDAPName->Data[i] == '"')
                    {
                        if(1 == NumQuotes)
                        {
                            /* make sure we don't walk off the end */
                            if((i+1 < LDAPName->Length) &&
                               (LDAPName->Data[i+1] == '"') &&
                               (LDAPName->Data[i-1] == '"'))
                            {
                                NumQuotes++;
                            }
                            else
                            {
                                NumQuotes--;
                            }
                        }
                        else
                        {
                            if(2 == NumQuotes)  /* we just got to the end of """ */
                                NumQuotes = 0;
                        }
                    }
                }
            }
        }   /* end for */
   
    return Ret;
}



/*-----------------------------------------------------------------------------
 * Name: GetLDAPTypeValueInfo
 *
 * Description:
 * Function parses through an LDAP name string and gets the "type" and "value"
 * specified by TypeValueNumber.  Possible values range from 0-(N-1), where N
 * is the total number of values in LDAP name string.
 *
 * Parameters: 
 * AppHandle       (input) : handle to be used in memory allocation and freeing
 * LDAPName        (input) : LDAPName->Data is the LDAP name string from which
 *                               the type/value info will be gathered
 * TypeValueNumber  (input): number specifying which type/value to return
 * LDAPType       (output) : LDAPType->Data is the "type" for this value
 * LDAPValue      (output) : LDAPValue->Data is the LDAP formatted value of
 *                               this entry in the table
 *
 * Return value:
 * Success: CSSM_OK
 * Failure: Error Code
 * 
 *---------------------------------------------------------------------------*/

static CSSM_RETURN
GetLDAPTypeValueInfo(const CSSM_CLUTIL_CONTEXT *pContext,
                     CSSM_HANDLE AppHandle,
                     const CSSM_DATA *LDAPName,
                     uint32 TypeValueNumber,
                     CSSM_DATA_PTR LDAPType,
                     CSSM_DATA_PTR LDAPValue)
{
    CSSM_RETURN Ret = CSSM_OK;
    uint32 i;
    uint32 NumQuotes;
    uint32 CurrentTypeValueOffset;
    uint32 CurrentTypeValueNumber;
    uint32 CurrentOffset;

    
        /* first step is to find the offset to the type/value info we want */
        
        /* initialize count */
        CurrentOffset = 0;
        CurrentTypeValueOffset = 0;
        CurrentTypeValueNumber = 0;
        NumQuotes = 0;

        /* loop to check */
        while((CurrentTypeValueNumber < TypeValueNumber) &&
              (CurrentOffset < LDAPName->Length) &&
              (CSSM_OK == Ret))
        {
            /* only a distinct value if there's a comma without quotes around it */
            if((LDAPName->Data[CurrentOffset] == ',') &&
               (0 == NumQuotes))
            {
               CurrentTypeValueNumber++;
               CurrentTypeValueOffset = CurrentOffset+1; /* since comma ends current one */
            }
            else
            {
                /* see what sort of situation quotes may put us in */
                if((LDAPName->Data[CurrentOffset] == '"') &&
                   (0 == NumQuotes))
                {
                    NumQuotes++;
                }
                else
                {
                    if(LDAPName->Data[CurrentOffset] == '"')
                    {
                        if(1 == NumQuotes)
                        {
                            /* make sure we don't walk off the end */
                            if((CurrentOffset+1 < LDAPName->Length) &&
                               (LDAPName->Data[CurrentOffset+1] == '"') &&
                               (LDAPName->Data[CurrentOffset-1] == '"'))
                            {
                                NumQuotes++;
                            }
                            else
                            {
                                NumQuotes--;
                            }
                        }
                        else
                        {
                            if(2 == NumQuotes)  /* we just got to the end of """ */
                                NumQuotes = 0;
                        }
                    }
                }
            }
            CurrentOffset++;
        }   /* end while */

        /* see if we found the correct type/value */
        if(CurrentOffset == LDAPName->Length)
        {
            Ret = CSSMERR_CL_INVALID_DATA;
            
        }
        else
        {
            /* find out how long our type is */
            i = 0;
            while((LDAPName->Data[CurrentOffset++] != ' ') &&
                  (CurrentOffset < LDAPName->Length))
            {
                i++;
            }
            
            /* copy the type into LDAPType */
            LDAPType->Length = i;
            LDAPType->Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                         1,
                                                         LDAPType->Length);
            if(NULL != LDAPType->Data)
            {
                port_memcpy(LDAPType->Data,
                       &(LDAPName->Data[CurrentTypeValueOffset]),
                       LDAPType->Length);
            }
            else
            {
                Ret = CSSMERR_CL_MEMORY_ERROR;
                
            }


            /* advance to the beginning of the value */
            i = 0;
            while((LDAPName->Data[CurrentOffset++] != ' ') &&
                  (CurrentOffset < LDAPName->Length))
            {
                i++;
            }
            if(1 != i)
            {
                Ret = CSSMERR_CL_INVALID_DATA;
                
            }

            if(CSSM_OK == Ret)
            {
                /* now we copy the value */
                CurrentTypeValueOffset = CurrentOffset;
                i = 0;
                NumQuotes = 0;

                /* loop to check */
                while((CurrentTypeValueNumber == TypeValueNumber) &&
                      (CurrentOffset < LDAPName->Length) &&
                      (CSSM_OK == Ret))
                {

                    if((LDAPName->Data[CurrentOffset] == ',') &&
                       (0 == NumQuotes))
                    {
                       CurrentTypeValueNumber++; /* since comma ends current one */
                    }
                    else
                    {
                        /* see what sort of situation quotes may put us in */
                        if((LDAPName->Data[CurrentOffset] == '"') &&
                           (0 == NumQuotes))
                        {
                            NumQuotes++;
                        }
                        else
                        {
                            if(LDAPName->Data[CurrentOffset] == '"')
                            {
                                if(1 == NumQuotes)
                                {
                                    /* make sure we don't walk off the end */
                                    if((CurrentOffset+1 < LDAPName->Length) &&
                                       (LDAPName->Data[CurrentOffset+1] == '"') &&
                                       (LDAPName->Data[CurrentOffset-1] == '"'))
                                    {
                                        NumQuotes++;
                                    }
                                    else
                                    {
                                        NumQuotes--;
                                    }
                                }
                                else
                                {
                                    if(2 == NumQuotes)  /* we just got to the end of """ */
                                        NumQuotes = 0;
                                }
                            }
                        }
                    }
                    i++;
                    CurrentOffset++;
                }   /* end while */

                /* now we can copy the value - 
                   i-1 has our length and CurrentTypeValueOffset has the offset
                */
                LDAPValue->Length = i - 1;
                LDAPValue->Data = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle,
                                                              1,
                                                              LDAPValue->Length);
                if(NULL != LDAPValue->Data)
                {
                    port_memcpy(LDAPValue->Data,
                           &(LDAPName->Data[CurrentTypeValueOffset]),
                           LDAPValue->Length);
                }
                else
                {
                    if(NULL != LDAPType->Data)
                    {
                        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,
                                                  LDAPType->Data);
                    }
                    Ret = CSSMERR_CL_MEMORY_ERROR;
                    
                }
            }
        }
   
    return Ret;
}
