/*-----------------------------------------------------------------------
 * File: CRTSIGN.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "maf_include.h"
/* CSSM includes */
#include "cssm.h"
#include "cssmspi.h"
#include "guids.h"
#include "cssmport.h"
#include "cssmutil.h"
#include "api_shim.h"

#include "spkiutil.h"
#include <cssmMemTrack.h>
#include "certutil.h"
#include "mempool.h"


extern MEMORY_POOL_FUNCS memory_pool_funcs;
extern MEMORY_POOL_PTR poolptr;


/***************************************************************************/
/* Sign a canonical S-expression cert body.                                */
/***************************************************************************/
/***************************************************************************
 CertTemplate must be a packed canonical Sexp of the form:
 (sequence 
  (cert ...)
  (public-key ...)
 )
*/
CSSM_RETURN CSSMCLI CL_CertSign (CSSM_CL_HANDLE CLHandle,
                        CSSM_CC_HANDLE CCHandle,
                        const CSSM_DATA *CertTemplate,
                        const CSSM_FIELD *SignScope,
                        uint32 ScopeSize,
                        CSSM_DATA_PTR SignedCert)
{
    CSSM_RETURN ret = CSSM_OK;
    CSSM_CSP_HANDLE CSPHandle = 0;

    CSSM_DATA data_to_sign;
    CSSM_DATA signed_data;
    CSSM_DATA hashed_data;

    CSSM_ALGORITHMS HashAlg;
    CSSM_ALGORITHMS SigAlg;
    CSSM_CONTEXT_PTR contextp;

    CSSM_LIST CertTempList;
    CSSM_LIST PubKeyList;
    CSSM_LIST TBSCertList;
    CSSM_LIST_ELEMENT *w;

    CSSM_LIST HashList;
    CSSM_LIST SigValList;
    CSSM_LIST SigList;
    CSSM_LIST CertList;

    if (port_IsBadWritePtr(SignedCert, sizeof(CSSM_DATA)))
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    SignedCert->Length = 0;
    SignedCert->Data = NULL;

    
    /* Check inputs */
    /* CLHandle and CCHandle are checked by CSSM */
    if (!CertTemplate)
    {
        return CSSMERR_CL_INVALID_CERT_POINTER;
    }
    if (port_IsBadReadPtr(CertTemplate, sizeof(CSSM_DATA)))
    {
        return CSSMERR_CL_INVALID_DATA;
    }

    if (!CertTemplate->Data || !CertTemplate->Length ||
        port_IsBadReadPtr(CertTemplate->Data, CertTemplate->Length))
    {
        return CSSMERR_CL_INVALID_DATA;
    }

    /* SignScope is not supported in this release. */
    /* Fail if SignScope is passed to this routine, 
    because that behavior is not supported */
    if (SignScope || ScopeSize)
    {
        return CSSMERR_CL_SCOPE_NOT_SUPPORTED;
    }

    
    ret = addin_GetContext(CCHandle, &contextp);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    CSPHandle = contextp->CSPHandle;
    SigAlg = contextp->AlgorithmType;
    switch(contextp->AlgorithmType)
    {
        case CSSM_ALGID_SHA1WithRSA:
        case CSSM_ALGID_SHA1WithDSA:
            HashAlg = CSSM_ALGID_SHA1;
            break;
        case CSSM_ALGID_MD5WithRSA :
            HashAlg = CSSM_ALGID_MD5;
            break;
        default:
            addin_FreeContext(contextp);
            return CSSMERR_CL_INVALID_CONTEXT_HANDLE;
    }

    addin_FreeContext(contextp);
        

    /* Convert input to a parsed CSSM_LIST structure */
    ret = ToCssmList(CertTemplate, &CertTempList);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    /* Walk to the second element in the list and ensure thats a sub-list for
       a cert. Assign that to TBSCertList */
    w= CertTempList.Head;
    if (!w || 
        w->WordID != CSSM_WORDID_SEQUENCE ||
        !w->NextElement)
    {
        /* Free local memory and return */
        drop_list(&CertTempList);
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    w = w->NextElement; /* update w to point to second elem: (cert ...) */
    if (w->WordID != CSSM_WORDID_CERT ||
        w->ElementType != CSSM_LIST_ELEMENT_SUBLIST)
    {
        /* Free local memory and return */
        drop_list(&CertTempList);
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    TBSCertList = w->Element.Sublist;
    port_memset(&(w->Element.Sublist), 0, sizeof(CSSM_LIST));

    /* Walk to the third elem and ensure thats a sub-list for a pub-key*/
    w = w->NextElement; /* update w to point to third elem: (pub_key ...) */
    if (w->WordID != CSSM_WORDID_PUBLIC_KEY  ||
        w->ElementType != CSSM_LIST_ELEMENT_SUBLIST)
    {
        /* Free local memory and return */
        drop_list(&CertTempList);
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    PubKeyList = w->Element.Sublist;
    port_memset(&(w->Element.Sublist), 0, sizeof(CSSM_LIST));


    /* Lets free the CertTemp list. We dont need it anymore */
    drop_list(&CertTempList);

    /* Pack the (cert...) into a CSSM_DATA */
    ret = CreateDataToSign(&TBSCertList, &data_to_sign);
    if (CSSM_OK != ret)
    {   
        drop_list(&PubKeyList);
        drop_list(&TBSCertList);
        return ret;
    }

    /* Sign the packed cert */
    ret = CreateSignature(CCHandle, &data_to_sign, &signed_data);
    if (CSSM_OK != ret)
    {   
        drop_list(&PubKeyList);
        drop_list(&TBSCertList);        
        CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func, data_to_sign.Data, poolptr);
        return ret;
    }


    /* hash/digest the packed cert */
    ret = CreateHash(CSPHandle, HashAlg, &data_to_sign, &hashed_data);
    if (CSSM_OK != ret)
    {   
        drop_list(&PubKeyList);
        drop_list(&TBSCertList);        
        CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func, data_to_sign.Data, poolptr);
        CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func, signed_data.Data, poolptr);
        return ret;
    }

    CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func, data_to_sign.Data, poolptr); /* We dont need this anymore since we
                              have the hash and the signature */

    ret = CreateHashList(HashAlg, &hashed_data, &HashList);
    if (CSSM_OK != ret)
    {   
        drop_list(&PubKeyList);
        drop_list(&TBSCertList);
        CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func, signed_data.Data, poolptr);
        CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func, hashed_data.Data, poolptr);
        return ret;
    }

    ret = CreateSigValList(SigAlg, &signed_data, &SigValList);
    if (CSSM_OK != ret)
    {   
        drop_list(&PubKeyList);
        drop_list(&TBSCertList);
        drop_list(&HashList);
        CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func, signed_data.Data, poolptr);
        return ret;
    }

    
    ret = CreateSigList(&HashList, &PubKeyList, &SigValList, &SigList);
    if (CSSM_OK != ret)
    {   
        drop_list(&PubKeyList);
        drop_list(&TBSCertList);
        drop_list(&HashList);
        drop_list(&SigValList);
        return ret;
    }

    ret = CreateCertList(&TBSCertList, &SigList, &PubKeyList, &CertList);
    if (CSSM_OK != ret)
    {   
        drop_list(&TBSCertList);
        drop_list(&SigList);
        return ret;
    }


    /* Pack that list into output */
    SignedCert->Length = size_list(&CertList);
    if (!SignedCert->Length )
    {
        /* Free used local mem */
        drop_list(&CertList);
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    SignedCert->Data = App_Calloc(CLHandle, 1, SignedCert->Length);
    if (!SignedCert->Data)
    {
        /* Free used local mem */
        drop_list(&CertList);
        return CSSMERR_CL_MEMORY_ERROR;
    }

    pack_list(&(SignedCert->Data), &CertList);

    /*  Free all used local memory */
    FreeCertList(&CertList);

    return CSSM_OK;

    
}
