/*-----------------------------------------------------------------------
 * File: CLBERDER.C
 *
 * Copyright (c) 1995-1999 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

/* This file contains routines that make exhaustive use of the Memory Pool 
*  and associated functions to parse and pack BER/DER structures. This 
*  significantly speeds up processing over the parsing/packing being done 
*  by the CLUTIL library. 
*   
*   There are 3 major sections to this file:
*   - Encoding routines that produce DER_NODE_PTR parse trees for packing
*     and encoded(packed) DER structures.
*   - Decoding routines that do the reverse of the above.
*   - Utility routines (involving mempools).
*
*  TODO: Add routines to perform CRL and CSSM_X509_NAME/LDAP  related 
*  BER/DER parsing/packing to this file
*/

#include "ber_der.h"
#include "mempool.h"
#include "x509defs.h"
#include "cssm.h"
#include "clberder.h"
#include "algoid.h"
#include "CSM/csm_keyparse.h"
#ifndef VMS
#pragma warning(disable: 4090) /* This disables the warnings that are generated
                                when CSSM_DATA::Data <- BER_PARSED_ITEM::Content
                                assignment takes place */
#endif

#ifdef VMS
#define DER_constructed_context_specific_0 DER_construct_context_spec_0
#define DER_constructed_context_specific_1 DER_construct_context_spec_1
#define DER_constructed_context_specific_2 DER_construct_context_spec_2
#define DER_constructed_context_specific_3 DER_construct_context_spec_3
#endif

extern PORT_TLS_INDEX TLS_MemoryPool;
extern BER_MEMORY_FUNCS berder_funcs_wrapper;
extern MEMORY_POOL_FUNCS memory_pool_funcs;


/* BER DER Tags */
/* defined in x509type.c and pkcs7.c */
#define X509MAXEXTENSIONS 16
#define X500NAMEDEPTH 16
extern unsigned char DER_constructedSequence[];
extern unsigned char DER_constructedSet[];
extern unsigned char DER_objectIDTag[];
extern unsigned char DER_signedDataObjectID[];
extern unsigned char DER_integerTag[];
extern unsigned char DER_bitStringTag[];
extern unsigned char DER_nullTag[];
extern unsigned char DER_printableStringTag[];
extern unsigned char DER_booleanTag[];
extern unsigned char DER_octetStringTag[];
extern unsigned char DER_utcTag[];
extern unsigned char DER_genTimeTag[];
extern unsigned char DER_constructed_context_specific_0[];
extern unsigned char DER_constructed_context_specific_1;
extern unsigned char DER_constructed_context_specific_2;
/* Types for BER DER parsing */
extern DER_COMPLEX_STRUCT DER_AlgorithmIdentifierType;
extern DER_COMPLEX_STRUCT DER_NameType;
extern DER_COMPLEX_STRUCT DER_AttributeSetType;
extern DER_COMPLEX_STRUCT DER_AttributeType;
extern DER_COMPLEX_STRUCT DER_CertExtensionsType;
extern DER_COMPLEX_STRUCT DER_CertExtensionType;
extern DER_COMPLEX_STRUCT DER_CertExtensionListType;
extern DER_COMPLEX_STRUCT DER_ValidityType;
extern DER_COMPLEX_STRUCT DER_PublicKeyType;
extern DER_COMPLEX_STRUCT DER_x509SignedPartType;
extern DER_COMPLEX_STRUCT DER_x509certType;

/* Defined in algoid.c */
CSSM_BOOL CSSMAPI
X509_IsEqualOid(const CSSM_OID *Oid1, const CSSM_OID *Oid2);
/*-----------------------------------------------------------------------------
 * Name: IsBadBerTag
 *
 * Description:
 * This functions checks whether the input tag is one of the supported universal tags
 *
 * Parameters: 
 * Tag (input) : The tag to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether the input tag is one of the supported universal tags
 *---------------------------------------------------------------------------*/
CSSM_BOOL IsBadBerTag(CSSM_BER_TAG Tag)
{
    
    if ((Tag >= BER_BOOLEAN && Tag <= BER_OBJECT_IDENTIFIER) ||
        Tag == BER_SEQUENCE || Tag == BER_SET ||
        Tag == BER_PRINTABLE_STRING || Tag == BER_T61STRING ||
        Tag == BER_IA5STRING || Tag == BER_UTCTIME ||
        Tag == BER_GENTIME)
        return CSSM_FALSE;
    
    return CSSM_TRUE;
}

/*-----------------------------------------------------------------------------
 * Name: MP_CssmDataToBerItemPtr
 *
 * Description:
 * This functions copies the input data into a BER_PARSED_ITEM format
 *
 * Parameters: 
 * Tag  (input) : The universal tag for the data
 * Data (input) : The structure to be copied
 *
 * ber_item_ptr (output):
 * A BER_PARSED_ITEM containing the equivalent of the input tag and data
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN  MP_CssmDataToBerItemPtr   (MEMORY_POOL_PTR poolptr,
                                        CSSM_BER_TAG Tag, 
                                        const CSSM_DATA Data,
                                        BER_PARSED_ITEM_PTR ber_item_ptr)
{

    uint8 *tmp_tag;
    if (!poolptr || !ber_item_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    
    (ber_item_ptr)->Tag = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func,
        sizeof(uint8), poolptr);
    
    if (!ber_item_ptr->Tag)
    {
        return CSSMERR_CL_MEMORY_ERROR;
       
    }

    /* to work our way around the 'const' decln in 
    BER_PARSED_ITEM we jump thru this hoop */
    tmp_tag = (uint8 *)ber_item_ptr->Tag;
    *tmp_tag = (uint8) (Tag & 0xFF);


    ber_item_ptr->ContentLength = Data.Length;
    if (ber_item_ptr->ContentLength)
    {
        ber_item_ptr->Content = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func,
            Data.Length, poolptr);
        
        if (!ber_item_ptr->Content)
        {
            return CSSMERR_CL_MEMORY_ERROR;
        }
        port_memcpy((void*)ber_item_ptr->Content, Data.Data, Data.Length);
    } 
    else
    {
        ber_item_ptr->Content = NULL;
    }
    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: MP_EncodeObjectToBerItem
 *
 * Description:
 * Encodes the Data as a Tag object
 *
 * Parameters: 
 * Tag (input) : Tag for the New Object
 * Data (input): Content for the new object
 * ber_item_ptr (output): DER Encoded object
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_EncodeObjectToBerItem(MEMORY_POOL_PTR poolptr, 
                                     CSSM_BER_TAG Tag, 
                                     CSSM_DATA Data,
                                     BER_PARSED_ITEM_PTR ber_item_ptr)
{
    CSSM_RETURN         ret = CSSM_OK;
    CSSM_BER_TAG        newtag;

    CSSM_BOOL   * tmpboolptr;
    CSSM_DATA            temp_data = {0, NULL};
    CSSM_DATA            null_data = {0, NULL};

    Tag &= 0x1F;

    if (!poolptr || !ber_item_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }


    if (IsBadBerTag(Tag))
    {
        return CSSMERR_CL_INVALID_DATA;
    }

    if ((Data.Data != NULL) && port_IsBadReadPtr(Data.Data, Data.Length))
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
        
    }
    
    switch ( Tag ) {
    case BER_SEQUENCE :
    case BER_SET :
        newtag = (CSSM_BER_TAG) (Tag|BER_CONSTRUCTED);
        ret = MP_CssmDataToBerItemPtr(poolptr, newtag, Data, ber_item_ptr);
        if (CSSM_OK != ret)
            return ret;
        break;

    case BER_BOOLEAN :
        temp_data.Length = 1;
        temp_data.Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func,
            temp_data.Length, poolptr);
        if (!temp_data.Data)
            return CSSMERR_CL_MEMORY_ERROR;
        tmpboolptr = (CSSM_BOOL *)Data.Data;
        if (*tmpboolptr)
            port_memset(temp_data.Data, 0xFF, 1);
        else
            port_memset(temp_data.Data, 0, 1);
        ret = MP_CssmDataToBerItemPtr(poolptr, Tag, temp_data, ber_item_ptr);
        if (CSSM_OK != ret)
            return ret;
        break;

    case BER_PRINTABLE_STRING :
    case BER_IA5STRING :
    case BER_UTCTIME :
    case BER_GENTIME :
        if (Data.Length)
        {
            temp_data = Data;
            /* Remove trailing NULLs */
            while (temp_data.Length != 0 && !temp_data.Data[temp_data.Length - 1])  
                temp_data.Length--;
            if (temp_data.Length == 0)
                temp_data = null_data;
            ret = MP_CssmDataToBerItemPtr(poolptr, Tag, temp_data, ber_item_ptr);
        }
        else
        {
            temp_data = null_data;
            ret = MP_CssmDataToBerItemPtr(poolptr, Tag, temp_data, ber_item_ptr);
        }
        if (CSSM_OK != ret)
        {
            return ret;
        }
        break;

    case BER_BIT_STRING :
        if (Data.Length)
        {
            if (Data.Length % 8) 
            {
                temp_data.Length = Data.Length/8 +2;
            }
            else
            {
                temp_data.Length = Data.Length/8 +1;
            }
            temp_data.Length = 1 + BER_LengthOfLength(temp_data.Length) + temp_data.Length;
            temp_data.Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func,
                temp_data.Length, poolptr);
            if (!temp_data.Data)
            {
                return CSSMERR_CL_MEMORY_ERROR;
               
            }

            DER_EncodeBitString(Data.Data, Data.Length, temp_data.Data, temp_data.Length);
            BER_ExpandItem(temp_data.Data, temp_data.Length, ber_item_ptr);
        } 
        else 
        {
            temp_data = null_data;
            ret = MP_CssmDataToBerItemPtr(poolptr, Tag, temp_data, ber_item_ptr);
        }
        if (CSSM_OK != ret)
            return ret;
        break;

    default:
        ret = MP_CssmDataToBerItemPtr(poolptr,Tag, Data, ber_item_ptr);
        if (CSSM_OK != ret)
            return ret;
    }

    return ret;


}
/*-----------------------------------------------------------------------------
 * Name: MP_EncodeObject
 *
 * Description:
 * Encodes the given data into a DER object in a CSSM_DATA structure
 *
 * Parameters: 
 * Tag (input) : Tag for the new object
 * Data (input): Content for the new object
 * return_val(output): Encoded DER object as a CSSM_DATA
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_EncodeObject(MEMORY_POOL_PTR poolptr, 
                            CSSM_BER_TAG Tag, 
                            CSSM_DATA Data,
                            CSSM_DATA_PTR return_val)
{
    CSSM_DATA_PTR result;
    CSSM_RETURN ret = CSSM_OK;
    BER_PARSED_ITEM ber_item;
    
    if (!poolptr || !return_val)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    
    return_val->Data = NULL;
    return_val->Length = 0;

    Tag &= 0x1F;
    if (IsBadBerTag(Tag))
    {
        return CSSMERR_CL_INVALID_DATA;
    }

    ret = MP_EncodeObjectToBerItem(poolptr, Tag, Data, &ber_item);
    if (CSSM_OK != ret )
    {
        return ret;
    }

    result = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_DATA),
                    poolptr);
    if (!result)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    result->Length = BER_LengthOfTag(ber_item.Tag)
                    + BER_LengthOfLength(ber_item.ContentLength)
                    + ber_item.ContentLength;
    result->Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, result->Length,
                    poolptr);
    if (!result->Data)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    BER_OutputTag(ber_item.Tag, &result->Data[0]);
    BER_OutputLength(ber_item.ContentLength, 
        &result->Data[BER_LengthOfTag(ber_item.Tag)]);
    port_memcpy(&result->Data[BER_LengthOfTag(ber_item.Tag) + BER_LengthOfLength(ber_item.ContentLength)],
        ber_item.Content, ber_item.ContentLength);

    *return_val = *result;
    
    return CSSM_OK;


}
/*-----------------------------------------------------------------------------
 * Name: MP_CreateAlgorithmIdNode
 *
 * Description:
 * Creates a DER_NODE_PTR containing the given alg identifier
 *
 * Parameters: 
 * algorithm (input): CSSM_X509_ALGORITHM_IDENTIFIER that contains the OID
 *                    and parameters
 *
 * nodeptr (output) : DER_NODE_PTR containing the parse tree 
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_CreateAlgorithmIdNode(MEMORY_POOL_PTR poolptr, 
                                     const CSSM_X509_ALGORITHM_IDENTIFIER algorithm,
                                     DER_NODE_PTR *nodeptr)
{
    DER_NODE_PTR tmpnode_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;
    
    if (!poolptr || !nodeptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    
    *nodeptr = NULL;

    
    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, poolptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    tmpnode_ptr = DER_AllocateNode(TLS_MemoryPool, 
        &berder_funcs_wrapper,
        2);/* 2 children:Oid and Params */
    
    if (!tmpnode_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    Init_DER_NODE_Header(tmpnode_ptr, DER_constructedSequence,
                        DER_constructedSequence,
                        2, DER_AlgorithmIdentifierType);
    tmpnode_ptr->Child[0].IsLeaf = 1;
    tmpnode_ptr->Child[0].X.Node = 0;
    tmpnode_ptr->Child[0].X.Input.Tag = DER_objectIDTag;
    tmpnode_ptr->Child[0].X.Input.ContentLength = algorithm.algorithm.Length;
    tmpnode_ptr->Child[0].X.Input.Content = algorithm.algorithm.Data;
    tmpnode_ptr->Child[1].IsLeaf = 1;
    tmpnode_ptr->Child[1].X.Node = 0;
    if (!algorithm.parameters.Data || !algorithm.parameters.Length)
    {
        if (X509_IsEqualOid(&algorithm.algorithm, &CSSMOID_X9CM_DSA))
        {
            /* Do not encode the params and leave it blank
               That means we shd reduce the number of children here by 1 */
            tmpnode_ptr->Count = 1;
        }
        else
        {
            tmpnode_ptr->Child[1].X.Input.Tag = DER_nullTag;
            tmpnode_ptr->Child[1].X.Input.ContentLength = 0;
            tmpnode_ptr->Child[1].X.Input.Content = NULL;
        }
    }
    else
    {
        BER_ExpandItem(algorithm.parameters.Data, 
            algorithm.parameters.Length, &tmpnode_ptr->Child[1].X.Input);
    }
    *nodeptr = tmpnode_ptr;
    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: MP_CreateAlgorithmNameNode
 *
 * Description:
 * Creates a DER_NODE_PTR containing the given X509 Name
 *
 * Parameters: 
 * name (input): CSSM_X509_NAME that shd be translated to a parse tree for packing
 *
 * nodeptr (output) : DER_NODE_PTR containing the parse tree 
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_CreateNameNode(MEMORY_POOL_PTR poolptr, 
                             const CSSM_X509_NAME name,
                             DER_NODE_PTR *nodeptr)
{
    CSSM_RETURN ret = CSSM_OK;
    DER_NODE_PTR name_node_ptr = NULL;
    DER_NODE_PTR current_av_set_node_ptr = NULL;
    DER_NODE_PTR current_av_node_ptr = NULL;
    uint32 i=0, j=0;
    CSSM_X509_RDN current_rdn;
    CSSM_X509_TYPE_VALUE_PAIR current_av_pair;

    if (!poolptr || !nodeptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    
    *nodeptr = NULL;
    if (name.numberOfRDNs > X500NAMEDEPTH)
    {
        /* Too many RDNs */
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, poolptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    name_node_ptr = DER_AllocateNode(TLS_MemoryPool, 
                                    &berder_funcs_wrapper,
                                    name.numberOfRDNs);
    if (!name_node_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    
    Init_DER_NODE_Header(name_node_ptr, DER_constructedSequence,
        DER_constructedSequence, name.numberOfRDNs, DER_NameType);
    for(i=0;i<name.numberOfRDNs;i++)
    {
        current_rdn = name.RelativeDistinguishedName[i];

        name_node_ptr->Child[i].IsLeaf = 0;
        name_node_ptr->Child[i].X.Node = DER_AllocateNode(TLS_MemoryPool, 
                                    &berder_funcs_wrapper,
                                    current_rdn.numberOfPairs);
        if (!(name_node_ptr->Child[i].X.Node ))
        {
            return CSSMERR_CL_MEMORY_ERROR;
        }
        current_av_set_node_ptr = name_node_ptr->Child[i].X.Node;
        Init_DER_NODE_Header(current_av_set_node_ptr, DER_constructedSet,
            DER_constructedSet, current_rdn.numberOfPairs, DER_AttributeSetType);
        for(j=0;j<current_rdn.numberOfPairs;j++)
        {
            current_av_pair = current_rdn.AttributeTypeAndValue[j];
            
            current_av_set_node_ptr->Child[j].IsLeaf = 0;
            current_av_set_node_ptr->Child[j].X.Node = DER_AllocateNode(TLS_MemoryPool, 
                                                        &berder_funcs_wrapper,
                                                        2); /* Attr Oid and Value */
            if (!current_av_set_node_ptr->Child[j].X.Node)
            {
                return CSSMERR_CL_MEMORY_ERROR;
            }
            current_av_node_ptr = current_av_set_node_ptr->Child[j].X.Node;
            Init_DER_NODE_Header(current_av_node_ptr, DER_constructedSequence,
                DER_constructedSequence, 2, DER_AttributeType);
            /* Child 0 is the Type OID*/
            current_av_node_ptr->Child[0].IsLeaf = 1;
            current_av_node_ptr->Child[0].X.Node = 0;
            current_av_node_ptr->Child[0].X.Input.Tag = DER_objectIDTag;
            current_av_node_ptr->Child[0].X.Input.ContentLength = current_av_pair.type.Length;
            current_av_node_ptr->Child[0].X.Input.Content = current_av_pair.type.Data;
            /* Child 1 is the valuetype */
            current_av_node_ptr->Child[1].IsLeaf = 1;
            current_av_node_ptr->Child[1].X.Node = 0;
            if (current_av_pair.valueType)
            {
                if (current_av_pair.valueType == DER_TAG_BIT_STRING)
                {
                    current_av_pair.value.Length *= 8;
                }

                ret = MP_EncodeObjectToBerItem(poolptr, current_av_pair.valueType , current_av_pair.value,
                            &(current_av_node_ptr->Child[1].X.Input));
                if (current_av_pair.valueType == DER_TAG_BIT_STRING)
                {
                    current_av_pair.value.Length /= 8;
                }

            }
            else
            {
                ret = MP_EncodeObjectToBerItem(poolptr, BER_PRINTABLE_STRING , current_av_pair.value,
                            &(current_av_node_ptr->Child[1].X.Input));
            }         
            if (CSSM_OK != ret)
            {
                return ret;
            }
            
        }
    }
    *nodeptr = name_node_ptr;
    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: MP_EncodeName
 *
 * Description:
 * DER encodes the input X509 Name
 *
 * Parameters: 
 * name (input): CSSM_X509_NAME that shd be encoded
 *
 * result (output) : DER encoded name
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_DerEncodeName(MEMORY_POOL_PTR poolptr,
                              const CSSM_X509_NAME *name,
                              CSSM_DATA_PTR *result)
{

    DER_NODE_PTR name_node_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;

    if (!poolptr || !result)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    *result = NULL;
    if (!name)
    {
        return CSSM_OK;
    }
    
    *result = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 1, sizeof(CSSM_DATA),
                        poolptr);
    if (!*result)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    
    ret = MP_CreateNameNode(poolptr, *name, &name_node_ptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    
    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, poolptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    (*result)->Data = DER_TypedGenerate(TLS_MemoryPool, &berder_funcs_wrapper, name_node_ptr);
    if (!(*result)->Data)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    (*result)->Length = BER_SizeofObject((*result)->Data);

    return CSSM_OK;

}
/*-----------------------------------------------------------------------------
 * Name: MP_DerEncodeExtnValue
 *
 * Description:
 * DER encodes the input Extn Value
 *
 * Parameters: 
 * Extension (input): CSSM_X509_EXTENSION that needs to be encoded
 *
 * result (output) : DER Encoded extension
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_DerEncodeExtnValue(MEMORY_POOL_PTR poolptr,
                                   const CSSM_X509_EXTENSION Extension,
                                   CSSM_DATA_PTR result)
{
    
    if (!poolptr || !result)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    
    if (Extension.value.tagAndValue && (Extension.value.tagAndValue)->value.Data 
        && (Extension.value.tagAndValue)->value.Length)
    {
        return MP_EncodeObject(poolptr, (Extension.value.tagAndValue)->type, (Extension.value.tagAndValue)->value,
                            result);
        
    }
    else
    {
        result->Length = 0;
        result->Data = NULL;
    }
    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: MP_CreateExtensionsNode
 *
 * Description:
 * Creates a DER_NODE_PTR containing the given X509 Extensions array
 *
 * Parameters: 
 * Extensions (input): CSSM_X509_EXTENSION array that
 *                     shd be translated to a parse tree for packing
 * NumExtensions (input): Number of elements in the array
 * nodeptr (output) : DER_NODE_PTR containing the parse tree 
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_CreateExtensionsNode(MEMORY_POOL_PTR poolptr, 
                             const CSSM_X509_EXTENSION *Extensions, 
                             uint32 NumExtensions,
                             DER_NODE_PTR *nodeptr)
{
    DER_NODE_PTR extns_node_ptr = NULL;
    DER_NODE_PTR extnlist_node_ptr = NULL;
    DER_NODE_PTR current_extn_node_ptr = NULL;
    uint32 i=0;
    CSSM_X509_EXTENSION current_extn;
    CSSM_DATA tmp_cssm_data = {0, NULL};
    CSSM_RETURN ret = CSSM_OK;

    if (!poolptr || !nodeptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    *nodeptr = NULL;
    if (!Extensions)
    {
        return CSSM_OK;
    }
    if (NumExtensions > X509MAXEXTENSIONS)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    
    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, poolptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    extns_node_ptr = DER_AllocateNode(TLS_MemoryPool, 
                                    &berder_funcs_wrapper,
                                    1);
    if (!extns_node_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }

    Init_DER_NODE_Header(extns_node_ptr, DER_constructedSequence,
        DER_constructedSequence, 1, DER_CertExtensionsType);

    extnlist_node_ptr = DER_AllocateNode(TLS_MemoryPool, 
                                    &berder_funcs_wrapper,
                                    NumExtensions);
    if (!extnlist_node_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }

    Init_DER_NODE_Header(extnlist_node_ptr, DER_constructedSequence,
        DER_constructedSequence, NumExtensions, DER_CertExtensionListType);
    extns_node_ptr->Child[0].IsLeaf = 0;
    extns_node_ptr->Child[0].X.Node = extnlist_node_ptr;

    for(i=0;i<NumExtensions;i++)
    {
        current_extn = Extensions[i];
        extnlist_node_ptr->Child[i].IsLeaf = 0;
        extnlist_node_ptr->Child[i].X.Node = DER_AllocateNode(TLS_MemoryPool, 
                                    &berder_funcs_wrapper,
                                    3); /* OID Critical Value */
        if (!extnlist_node_ptr->Child[i].X.Node)
        {
            return CSSMERR_CL_MEMORY_ERROR;
        }
        current_extn_node_ptr = extnlist_node_ptr->Child[i].X.Node;
        Init_DER_NODE_Header(current_extn_node_ptr, DER_constructedSequence,
            DER_constructedSequence, 3, DER_CertExtensionListType);
        /* OID*/
        current_extn_node_ptr->Child[0].IsLeaf = 1;
        current_extn_node_ptr->Child[0].X.Node = 0;
        current_extn_node_ptr->Child[0].X.Input.Tag = DER_objectIDTag;
        current_extn_node_ptr->Child[0].X.Input.ContentLength = current_extn.extnId.Length;
        current_extn_node_ptr->Child[0].X.Input.Content = current_extn.extnId.Data;
        /* Critical */
        if (current_extn.critical)
        {
            current_extn_node_ptr->Child[1].IsLeaf = 1;
            current_extn_node_ptr->Child[1].X.Node = 0;
            tmp_cssm_data.Length = 1;
            tmp_cssm_data.Data = (uint8 *) &(current_extn.critical);
            ret = MP_EncodeObjectToBerItem(poolptr, BER_BOOLEAN, tmp_cssm_data, &current_extn_node_ptr->Child[1].X.Input);
            if (CSSM_OK != ret)
            {
                return ret;
            }
            /* Value */
            current_extn_node_ptr->Child[2].IsLeaf = 1;
            current_extn_node_ptr->Child[2].X.Node = 0;
            current_extn_node_ptr->Child[2].X.Input.Tag = DER_octetStringTag;
            ret = MP_DerEncodeExtnValue(poolptr,current_extn,&tmp_cssm_data);
            if (CSSM_OK != ret)
            {
                return ret;
            }
            current_extn_node_ptr->Child[2].X.Input.ContentLength = tmp_cssm_data.Length;
            current_extn_node_ptr->Child[2].X.Input.Content = tmp_cssm_data.Data;
        }
        else
        {
            current_extn_node_ptr->Count = 2;
            /* Value */
            current_extn_node_ptr->Child[1].IsLeaf = 1;
            current_extn_node_ptr->Child[1].X.Node = 0;
            current_extn_node_ptr->Child[1].X.Input.Tag = DER_octetStringTag;
            ret = MP_DerEncodeExtnValue(poolptr,current_extn,&tmp_cssm_data);
            if (CSSM_OK != ret)
            {
                return ret;
            }
            current_extn_node_ptr->Child[1].X.Input.ContentLength = tmp_cssm_data.Length;
            current_extn_node_ptr->Child[1].X.Input.Content = tmp_cssm_data.Data;
        }
    }
    *nodeptr = extns_node_ptr;
    return CSSM_OK;


}

/*-----------------------------------------------------------------------------
 * Name: MP_CreateTBSCertNode
 *
 * Description:
 * Creates a DER_NODE_PTR containing the given TBS Certificate portion
 *
 * Parameters: 
 * cert_ptr (input): TBS portion of a cert that shd be translated 
 *                   to a parse tree for packing
 *
 * nodeptr (output) : DER_NODE_PTR containing the parse tree 
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_CreateTbsCertNode(MEMORY_POOL_PTR poolptr, 
                                 CSSM_X509_TBS_CERTIFICATE_PTR tbs_cert_ptr,
                                 DER_NODE_PTR *result_val)
{
    CSSM_RETURN ret = CSSM_OK;
    DER_NODE_PTR tbs_cert_node_ptr = NULL;
    DER_NODE_PTR spki_node_ptr = NULL;
    DER_NODE_PTR validity_node_ptr = NULL;
    
    uint32 notBefore_length, notAfter_length;
    CSSM_DATA tmp_time;
    CSSM_DATA temp;
    CSSM_DATA temp_data;
    

    if (!poolptr || !result_val)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    *result_val = NULL;

    if (!tbs_cert_ptr)
    {
        return CSSM_OK;
    }

    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, poolptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    tbs_cert_node_ptr = DER_AllocateNode(TLS_MemoryPool, &berder_funcs_wrapper, 10);
    if (!tbs_cert_node_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    
    spki_node_ptr = DER_AllocateNode(TLS_MemoryPool, &berder_funcs_wrapper, 2);
    if (!spki_node_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }

    validity_node_ptr = DER_AllocateNode(TLS_MemoryPool, &berder_funcs_wrapper, 2);
    if (!validity_node_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }

    
    /* Validity Node */
    Init_DER_NODE_Header(validity_node_ptr, DER_constructedSequence,
        DER_constructedSequence, 2, DER_ValidityType);
    

    
    if (tbs_cert_ptr->validity.notBefore.time.Data)
    {
        notBefore_length = strlen((char *)tbs_cert_ptr->validity.notBefore.time.Data) + 1;
    }
    else 
    {
        notBefore_length = 0;
    }
    if (tbs_cert_ptr->validity.notAfter.time.Data)
    {
        notAfter_length = strlen((char *)tbs_cert_ptr->validity.notAfter.time.Data) + 1;
    }
    else 
    {
        notAfter_length = 0;
    }
    /* not before */
    validity_node_ptr->Child[0].IsLeaf = 1;
    validity_node_ptr->Child[0].X.Node = 0;
    tmp_time.Length = notBefore_length;
    tmp_time.Data = tbs_cert_ptr->validity.notBefore.time.Data;

    if (notBefore_length <= UTC_TIME_STRLEN +1)
    {
        ret = MP_EncodeObjectToBerItem(poolptr, BER_UTCTIME, tmp_time, &validity_node_ptr->Child[0].X.Input);
    }
    else
    {
        ret = MP_EncodeObjectToBerItem(poolptr, BER_GENTIME, tmp_time, &validity_node_ptr->Child[0].X.Input);
    }
    if (CSSM_OK != ret)
    {
        return ret;
    }

    /* not after */
    validity_node_ptr->Child[1].IsLeaf = 1;
    validity_node_ptr->Child[1].X.Node = 0;
    tmp_time.Length = notAfter_length;
    tmp_time.Data = tbs_cert_ptr->validity.notAfter.time.Data;
    if (notAfter_length <= UTC_TIME_STRLEN +1)
    {
        ret = MP_EncodeObjectToBerItem(poolptr, BER_UTCTIME, tmp_time, &validity_node_ptr->Child[1].X.Input);
    }
    else
    {
        ret = MP_EncodeObjectToBerItem(poolptr, BER_GENTIME, tmp_time, &validity_node_ptr->Child[1].X.Input);
    }
    if (CSSM_OK != ret)
    {
        return ret;
    }

    /* SPKI node */
    Init_DER_NODE_Header(spki_node_ptr, DER_constructedSequence,
        DER_constructedSequence, 2, DER_PublicKeyType);
    /* Alg Id */
    spki_node_ptr->Child[0].IsLeaf = 0;
    ret = MP_CreateAlgorithmIdNode(poolptr, tbs_cert_ptr->subjectPublicKeyInfo.algorithm, &(spki_node_ptr->Child[0].X.Node));
    if (CSSM_OK != ret)
    {
        return ret;
    }
    /* Key */
    spki_node_ptr->Child[1].IsLeaf = 1;
    spki_node_ptr->Child[1].X.Node = 0;
    if (tbs_cert_ptr->subjectPublicKeyInfo.subjectPublicKey.Length)
    {
        temp.Length = 8*tbs_cert_ptr->subjectPublicKeyInfo.subjectPublicKey.Length;
        temp.Data = tbs_cert_ptr->subjectPublicKeyInfo.subjectPublicKey.Data;
        ret = MP_EncodeObjectToBerItem(poolptr, BER_BIT_STRING, temp,
            &spki_node_ptr->Child[1].X.Input);
        if (CSSM_OK != ret)
        {
            return ret;
        }
       
    } 
    else 
    {
        spki_node_ptr->Child[1].X.Input.Tag = DER_bitStringTag;
        spki_node_ptr->Child[1].X.Input.ContentLength = 0;
        spki_node_ptr->Child[1].X.Input.Content = 0;        
    }



    /**********************  TBS CERT NODE ***********************/
    Init_DER_NODE_Header(tbs_cert_node_ptr, DER_constructedSequence,
        DER_constructedSequence, 10, DER_x509SignedPartType);
    
    /* Version */
    ret = MP_EncodeObject(poolptr, BER_INTEGER, tbs_cert_ptr->version, &temp_data);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    tbs_cert_node_ptr->Child[0].IsLeaf = 1;
    tbs_cert_node_ptr->Child[0].X.Node = 0;
    tbs_cert_node_ptr->Child[0].X.Input.Tag = DER_constructedSequence;
    tbs_cert_node_ptr->Child[0].X.Input.ContentLength = temp_data.Length;
    tbs_cert_node_ptr->Child[0].X.Input.Content = temp_data.Data;
    /* Serial Number */
    tbs_cert_node_ptr->Child[1].IsLeaf = 1;
    tbs_cert_node_ptr->Child[1].X.Node = 0;
    tbs_cert_node_ptr->Child[1].X.Input.Tag = DER_integerTag;
    tbs_cert_node_ptr->Child[1].X.Input.ContentLength = tbs_cert_ptr->serialNumber.Length;
    tbs_cert_node_ptr->Child[1].X.Input.Content = tbs_cert_ptr->serialNumber.Data;

    /*Signature Alg OID */
    tbs_cert_node_ptr->Child[2].IsLeaf = 0;
    ret = MP_CreateAlgorithmIdNode(poolptr, tbs_cert_ptr->signature, &(tbs_cert_node_ptr->Child[2].X.Node));
    if (CSSM_OK != ret )
    {
        return ret;
    }

    /* Issuer Name */
    tbs_cert_node_ptr->Child[3].IsLeaf = 0;
    ret = MP_CreateNameNode(poolptr, tbs_cert_ptr->issuer, &(tbs_cert_node_ptr->Child[3].X.Node));
    if (CSSM_OK != ret )
    {
        return ret;
    }

    /* Validity */
    tbs_cert_node_ptr->Child[4].IsLeaf = 0;
    tbs_cert_node_ptr->Child[4].X.Node = validity_node_ptr;

    /* Subject Name */
    tbs_cert_node_ptr->Child[5].IsLeaf = 0;
    ret = MP_CreateNameNode(poolptr, tbs_cert_ptr->subject, &(tbs_cert_node_ptr->Child[5].X.Node));
    if (CSSM_OK != ret )
    {
        return ret;
    }
    
    /* SPKI */
    tbs_cert_node_ptr->Child[6].IsLeaf = 0;
    tbs_cert_node_ptr->Child[6].X.Node = spki_node_ptr;

    /* Issuer UID */
    tbs_cert_node_ptr->Child[7].IsLeaf = 1;
    tbs_cert_node_ptr->Child[7].X.Node = 0;
    if (tbs_cert_ptr->issuerUniqueIdentifier.Length)
    {
     
        temp.Length = 8*tbs_cert_ptr->issuerUniqueIdentifier.Length;
        temp.Data = tbs_cert_ptr->issuerUniqueIdentifier.Data;
        ret = MP_EncodeObjectToBerItem(poolptr, BER_BIT_STRING, temp,
            &tbs_cert_node_ptr->Child[7].X.Input);
        if (CSSM_OK != ret)
        {
            return ret;
        }
   } 
   else 
   {
        tbs_cert_node_ptr->Child[7].X.Input.Tag = 0;
        tbs_cert_node_ptr->Child[7].X.Input.ContentLength = 0;
        tbs_cert_node_ptr->Child[7].X.Input.Content = 0;        
    }


    

    /* Subject UID */
    tbs_cert_node_ptr->Child[8].IsLeaf = 1;
    tbs_cert_node_ptr->Child[8].X.Node = 0;
    if (tbs_cert_ptr->subjectUniqueIdentifier.Length)
    {
        temp.Length = 8*tbs_cert_ptr->subjectUniqueIdentifier.Length;
        temp.Data = tbs_cert_ptr->subjectUniqueIdentifier.Data;
        ret = MP_EncodeObjectToBerItem(poolptr, BER_BIT_STRING, temp,
            &tbs_cert_node_ptr->Child[8].X.Input);
        if (CSSM_OK != ret)
        {
            return ret;
        }
        
    } 
    else 
    {
        tbs_cert_node_ptr->Child[8].X.Input.Tag = 0;
        tbs_cert_node_ptr->Child[8].X.Input.ContentLength = 0;
        tbs_cert_node_ptr->Child[8].X.Input.Content = 0;        
    }

    /* Extensions */
    tbs_cert_node_ptr->Child[9].IsLeaf = 0;
    ret = MP_CreateExtensionsNode(poolptr, tbs_cert_ptr->extensions.extensions, tbs_cert_ptr->extensions.numberOfExtensions,
        &(tbs_cert_node_ptr->Child[9].X.Node));
    if (CSSM_OK != ret )
    {
        return ret;
    }
    tbs_cert_node_ptr->Child[9].X.Input.Tag = 0;
    tbs_cert_node_ptr->Child[9].X.Input.ContentLength = 0;
    tbs_cert_node_ptr->Child[9].X.Input.Content = 0;

    /* tbs cert node parse tree fully polulated */
    *result_val = tbs_cert_node_ptr;
    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: MP_CreatesignedCertNode
 *
 * Description:
 * Creates a DER_NODE_PTR containing the given Certificate 
 *
 * Parameters: 
 * signed_cert_ptr (input): cert that shd be translated 
 *                          to a parse tree for packing
 *
 * result_val (output) : DER_NODE_PTR containing the parse tree 
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_CreateSignedCertNode(MEMORY_POOL_PTR poolptr, 
                                 CSSM_X509_SIGNED_CERTIFICATE_PTR signed_cert_ptr,
                                 DER_NODE_PTR *result_val)
{
    CSSM_DATA temp_data;
    CSSM_RETURN ret = CSSM_OK;
    DER_NODE_PTR signed_cert_node_ptr = NULL;
    if (!poolptr || !result_val)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    *result_val = NULL;
    if (!signed_cert_ptr)
    {
        return CSSM_OK;
    }
    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, poolptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    signed_cert_node_ptr = DER_AllocateNode(TLS_MemoryPool, &berder_funcs_wrapper, 3);
    if (!signed_cert_node_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    Init_DER_NODE_Header(signed_cert_node_ptr, DER_constructedSequence,
        DER_constructedSequence, 3, DER_x509certType);

    /* TBS cert is child[0] */
    signed_cert_node_ptr->Child[0].IsLeaf = 0;
    ret = MP_CreateTbsCertNode(poolptr, &signed_cert_ptr->certificate, &(signed_cert_node_ptr->Child[0].X.Node));
    if (CSSM_OK != ret)
    {
        return ret;
    }

    /* Signature Alg OID is child[1]*/
    signed_cert_node_ptr->Child[1].IsLeaf = 0;
    ret = MP_CreateAlgorithmIdNode(poolptr, signed_cert_ptr->signature.algorithmIdentifier,
        &(signed_cert_node_ptr->Child[1].X.Node));
    if (CSSM_OK != ret)
    {
        return ret;
    }
    
    /* Signature is the third child (child[2])*/
    signed_cert_node_ptr->Child[2].IsLeaf = 1;
    signed_cert_node_ptr->Child[2].X.Node = 0;

    if (signed_cert_ptr->signature.encrypted.Length)
    {
        temp_data.Length = signed_cert_ptr->signature.encrypted.Length +1;
        temp_data.Length = 1 + BER_LengthOfLength(temp_data.Length) + temp_data.Length;
        temp_data.Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
            temp_data.Length, poolptr);
        if (!temp_data.Data)
        {
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        DER_EncodeBitString(signed_cert_ptr->signature.encrypted.Data, 8*(signed_cert_ptr->signature.encrypted.Length), temp_data.Data, temp_data.Length);
        BER_ExpandItem(temp_data.Data, temp_data.Length, &signed_cert_node_ptr->Child[2].X.Input);
        
    }
    else 
    {
        signed_cert_node_ptr->Child[2].X.Input.Tag = DER_bitStringTag;
        signed_cert_node_ptr->Child[2].X.Input.ContentLength = 0;
        signed_cert_node_ptr->Child[2].X.Input.Content = 0;        
    }


    *result_val = signed_cert_node_ptr;
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: MP_DerEncodeTbsCertificate
 *
 * Description:
 * DER encodes the given TBS Certificate portion
 *
 * Parameters: 
 * tbs_cert_ptr (input): TBS portion of a cert that shd be encoded
 *
 * enc_tbs_cert_ptr (output) : DER encoded TBS certificate
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_DerEncodeTbsCertificate(MEMORY_POOL_PTR poolptr, 
                                       CSSM_X509_TBS_CERTIFICATE_PTR tbs_cert_ptr,
                                       CSSM_DATA_PTR *enc_tbs_cert_ptr)
{
    
    DER_NODE_PTR tbs_cert_node_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;

    if (!poolptr || !enc_tbs_cert_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    *enc_tbs_cert_ptr = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 1, sizeof(CSSM_DATA),
                        poolptr);
    if (!*enc_tbs_cert_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    ret = MP_CreateTbsCertNode(poolptr,tbs_cert_ptr,&tbs_cert_node_ptr);
    
    if (CSSM_OK != ret)
    {
        return ret;
    }

    if (!tbs_cert_node_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    
    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, poolptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    (*enc_tbs_cert_ptr)->Data = DER_Generate(TLS_MemoryPool, &berder_funcs_wrapper, tbs_cert_node_ptr);
    if (!(*enc_tbs_cert_ptr)->Data)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    (*enc_tbs_cert_ptr)->Length = BER_SizeofObject((*enc_tbs_cert_ptr)->Data);
    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: MP_DerEncodeSignedCertificate
 *
 * Description:
 * DER encodes the given Certificate
 *
 * Parameters: 
 * signed_cert_ptr (input): cert that shd be encoded
 *
 * enc_signed_cert_ptr (output) : DER encoded X509 certificate
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_DerEncodeSignedCertificate(MEMORY_POOL_PTR poolptr, 
                                       CSSM_X509_SIGNED_CERTIFICATE_PTR signed_cert_ptr,
                                       CSSM_DATA_PTR *enc_signed_cert_ptr)
{
    
    DER_NODE_PTR signed_cert_node_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;

    if (!poolptr || !enc_signed_cert_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    *enc_signed_cert_ptr = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 1, sizeof(CSSM_DATA),
                        poolptr);
    if (!*enc_signed_cert_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    ret = MP_CreateSignedCertNode(poolptr,signed_cert_ptr,&signed_cert_node_ptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, poolptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    (*enc_signed_cert_ptr)->Data = DER_TypedGenerate(TLS_MemoryPool, &berder_funcs_wrapper, signed_cert_node_ptr);
    if (!(*enc_signed_cert_ptr)->Data)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    (*enc_signed_cert_ptr)->Length = BER_SizeofObject((*enc_signed_cert_ptr)->Data);
    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: MP_DerEncodeSignature
 *
 * Description:
 * DER encodes the given CSSM_X509_SIGNATURE
 *
 * Parameters: 
 * signature (input): signature that shd be encoded
 *
 * result (output): DER Encoded signature
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_DerEncodeSignature(MEMORY_POOL_PTR poolptr, 
                                  const CSSM_X509_SIGNATURE *signature,
                                  CSSM_DATA_PTR *result)
{
    DER_NODE_PTR sign_node_ptr = NULL;
    DER_NODE_PTR algid_node_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;
    CSSM_DATA temp_data = {0,NULL};

    if (!poolptr || !result)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    *result = NULL;
    
    if (!signature)
    {
        return CSSM_OK;
    }
    
    ret = MP_CreateAlgorithmIdNode(poolptr, signature->algorithmIdentifier, &algid_node_ptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, poolptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    sign_node_ptr = DER_AllocateNode(TLS_MemoryPool, &berder_funcs_wrapper, 2);
    if (!sign_node_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    

    sign_node_ptr->Tag = DER_constructedSequence;
    sign_node_ptr->OriginalTag = DER_constructedSequence;
    sign_node_ptr->Count = 2;
    sign_node_ptr->Type = NULL;

    sign_node_ptr->Child[0].IsLeaf = 0;
    sign_node_ptr->Child[0].X.Node = algid_node_ptr;

    
    sign_node_ptr->Child[1].IsLeaf = 1;
    sign_node_ptr->Child[1].X.Node = 0;

    if (signature->encrypted.Length)
    {
        temp_data.Length = signature->encrypted.Length +1;
        temp_data.Length = 1 + BER_LengthOfLength(temp_data.Length) + temp_data.Length;
        temp_data.Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
            temp_data.Length, poolptr);
        if (!temp_data.Data)
        {
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        DER_EncodeBitString(signature->encrypted.Data, 8*(signature->encrypted.Length), temp_data.Data, temp_data.Length);
        BER_ExpandItem(temp_data.Data, temp_data.Length, &sign_node_ptr->Child[1].X.Input);
        
    } else 
    {
        sign_node_ptr->Child[1].X.Input.Tag = DER_bitStringTag;
        sign_node_ptr->Child[1].X.Input.ContentLength = 0;
        sign_node_ptr->Child[1].X.Input.Content = 0;        
    }

    *result = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 1, sizeof(CSSM_DATA),
                        poolptr);
    if (!*result)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }

    (*result)->Data = DER_Generate(TLS_MemoryPool, &berder_funcs_wrapper, sign_node_ptr);
    
    if (!(*result)->Data)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    (*result)->Length = BER_SizeofObject((*result)->Data);

    return CSSM_OK;


}

/*-----------------------------------------------------------------------------
 * Name: MP_DerDecodeItem
 *
 * Description:
 * DER decodes the given BER_PARSED_ITEM (Tag/ContenLength/Content) 
 * into a CSSM_DATA
 *
 * Parameters: 
 * BerItemPtr (input): DER encoded object that shd be decoded
 *
 * result (output) : Decoded object (one level decoded)
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_DerDecodeItem(MEMORY_POOL_PTR poolptr,
                             BER_PARSED_ITEM_PTR BerItemPtr,
                             CSSM_DATA_PTR result)
{

    static CSSM_DATA       null_data = {0, NULL};
    uint32          bit_str_length =0;
    uint32          numBytesInBitString = 0;
    
    
    if (!poolptr || !result)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
      
    }
    if (!BerItemPtr || !BerItemPtr->Tag || !BerItemPtr->ContentLength || !BerItemPtr->Content)
    {
        *result = null_data;
        return CSSM_OK;
    }
    
    switch (BerItemPtr->Tag[0]) {
    case 0:
        *result = null_data;
        break;

    case BER_BOOLEAN :
        result->Length = sizeof(CSSM_BOOL);
        result->Data = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 1, sizeof(CSSM_BOOL),
                        poolptr);
        if (!result->Data)
        {
            return CSSMERR_CL_MEMORY_ERROR;
        }
        if (!BerItemPtr->Content || !BerItemPtr->Content[0])
        {
            result->Data[0] = CSSM_FALSE;
        }
        else
        {
            result->Data[0] = CSSM_TRUE;
        }   
        break;

    case BER_BIT_STRING :
        if (BerItemPtr->ContentLength)
        {
            /* Decode the bit string, removing the leading byte */
            bit_str_length = 8*(BerItemPtr->ContentLength -1) - BerItemPtr->Content[0];
            numBytesInBitString = BerItemPtr->ContentLength -1;

            result->Length = numBytesInBitString ;
            if (numBytesInBitString)
            {
                result->Data = 
                    CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, numBytesInBitString,
                        poolptr);
                if (!result->Data)
                {
                    
                    return CSSMERR_CL_MEMORY_ERROR;
                  
                }
                port_memcpy(result->Data, &BerItemPtr->Content[1], numBytesInBitString);
            } 
            else
            {
                result->Data = NULL; 
            }
        } 
        else
        {
            *result = null_data; 
        }

        break;

    case BER_PRINTABLE_STRING :
    case BER_IA5STRING :
    case BER_UTCTIME :
    case BER_GENTIME :
        result->Length = BerItemPtr->ContentLength +1;
        result->Data = 
            CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 1, result->Length,
                        poolptr);
        if (!result->Data)
        {
            return CSSMERR_CL_MEMORY_ERROR;
           
        }
        port_memcpy(result->Data, BerItemPtr->Content, BerItemPtr->ContentLength);
        break;

    default:
        result->Length = BerItemPtr->ContentLength;
        if (BerItemPtr->ContentLength) 
        {
            result->Data = 
                CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, result->Length,
                        poolptr);
            if (!result->Data)
            {
                
                return CSSMERR_CL_MEMORY_ERROR;
              
            }
            port_memcpy(result->Data, BerItemPtr->Content, result->Length);
        }
        else
        {
            result->Data = NULL;
        }
    }

    /* Return */
    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: MP_DerDecodeObject
 *
 * Description:
 * DER decodes the given Encoded object (CSSM_DATA)
 *
 * Parameters: 
 * Src (input): DER encoded object that shd be decoded
 *
 * result (output) : Decoded object (one level decoded)
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_DerDecodeObject(MEMORY_POOL_PTR poolptr, CSSM_DATA Src, CSSM_DATA_PTR result)
{
    BER_PARSED_ITEM item;
    BER_ExpandItem(Src.Data, Src.Length, &item);
    return MP_DerDecodeItem(poolptr, &item, result);
}
/*-----------------------------------------------------------------------------
 * Name: MP_DerDecodeExtensions
 *
 * Description:
 * DER decodes the encoded X509 extensions structure
 *
 * Parameters: 
 * Value (input): DER encoded object that shd be decoded
 *
 * extensions_ptr_ptr (output) : Decoded CSSM_X509_EXTENSIONS object
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_DerDecodeExtensions(MEMORY_POOL_PTR poolptr, 
                                   const CSSM_DATA *Value, 
                                   CSSM_X509_EXTENSIONS_PTR *extensions_ptr_ptr)
{
    BER_PARSED_ITEM item;
    DER_NODE_PTR extns_node_ptr;
    CSSM_RETURN ret = CSSM_OK;   
    
    if (!poolptr || ! extensions_ptr_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    if (!BER_ExpandItem((uint8*) Value->Data, Value->Length, &item))
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, poolptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    extns_node_ptr = BER_ParseTypedObject(TLS_MemoryPool,
            &berder_funcs_wrapper, 
            &DER_CertExtensionsType, &item);

    return MP_DerDecodeExtensionsNode(poolptr, extns_node_ptr, extensions_ptr_ptr);
}

/*-----------------------------------------------------------------------------
 * Name: MP_DerDecodeExtensionsNode
 *
 * Description:
 * Converts the Parse tree into a CSSM_X509_EXTENSIONS data structure
 *
 * Parameters: 
 * extns_node_ptr (input): parse tree for the extensions
 *
 * extensions_ptr_ptr (output) : Decoded CSSM_X509_EXTENSIONS object
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_DerDecodeExtensionsNode(MEMORY_POOL_PTR poolptr, 
                                       DER_NODE_PTR extns_node_ptr, 
                                       CSSM_X509_EXTENSIONS_PTR *extensions_ptr_ptr)
{

    DER_NODE_PTR extnlist_node_ptr;
    DER_NODE_PTR current_extn_node_ptr;
    CSSM_DATA tmp_cssm_data;
    uint32 NumExtns = 0;
    uint32 i;
    CSSM_X509_EXTENSIONS_PTR extns_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;

    BER_PARSED_ITEM value_item;
    

    if (!poolptr || !extensions_ptr_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    *extensions_ptr_ptr = NULL;

    if (!extns_node_ptr || extns_node_ptr->Type != &DER_CertExtensionsType ||
        !extns_node_ptr->Child)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    extnlist_node_ptr = extns_node_ptr->Child[0].X.Node;
    if (!extnlist_node_ptr || extnlist_node_ptr->Type != &DER_CertExtensionListType ||
        !extnlist_node_ptr->Child)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    
    /* Count is the max possible and we have to ignore the NULL nodes by inspection */
    i = 0;
    while (i < (uint32)extnlist_node_ptr->Count && extnlist_node_ptr->Child[i].X.Node)
    {
        i++;
    }
    NumExtns = i;

    extns_ptr = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 1, sizeof(CSSM_X509_EXTENSIONS),
                        poolptr);
    if (!extns_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    extns_ptr->numberOfExtensions = NumExtns;
    extns_ptr->extensions = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, NumExtns, sizeof(CSSM_X509_EXTENSION),
                        poolptr);
    if (!extns_ptr->extensions)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    for(i=0;i<NumExtns;i++)
    {
        current_extn_node_ptr = extnlist_node_ptr->Child[i].X.Node;
        if (!current_extn_node_ptr || 
            current_extn_node_ptr->Type != &DER_CertExtensionType ||
            !current_extn_node_ptr->Child)
        {
            return CSSMERR_CL_UNKNOWN_FORMAT;
        }
        /* OID */
        extns_ptr->extensions[i].extnId.Length = current_extn_node_ptr->Child[0].X.Input.ContentLength;
        extns_ptr->extensions[i].extnId.Data = current_extn_node_ptr->Child[0].X.Input.Content;
        if (current_extn_node_ptr->Count == 3)
        {
            /* Critical */
            if (!current_extn_node_ptr->Child[1].X.Input.Content || !current_extn_node_ptr->Child[1].X.Input.Content[0])
            {
                extns_ptr->extensions[i].critical = CSSM_FALSE;
            }
            else
            {
                extns_ptr->extensions[i].critical = CSSM_TRUE;
            }
            /* Value */
            extns_ptr->extensions[i].value.tagAndValue = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 
                1, sizeof(CSSM_X509EXT_TAGandVALUE), poolptr);
            if (!extns_ptr->extensions[i].value.tagAndValue)
            {
                return CSSMERR_CL_MEMORY_ERROR;
            }
        
            tmp_cssm_data.Length = current_extn_node_ptr->Child[2].X.Input.ContentLength; 
            tmp_cssm_data.Data = current_extn_node_ptr->Child[2].X.Input.Content;
            if (tmp_cssm_data.Length && tmp_cssm_data.Data)
            {   
                if (!BER_ExpandItem((uint8*) tmp_cssm_data.Data, tmp_cssm_data.Length, &value_item))
                {
                    return CSSMERR_CL_UNKNOWN_FORMAT;
                }        
                ret = MP_DerDecodeItem(poolptr, &value_item, &extns_ptr->extensions[i].value.tagAndValue->value);
                if (CSSM_OK != ret)
                {
                    return ret;
                }
                extns_ptr->extensions[i].value.tagAndValue->type = value_item.Tag[0];
            
            }
        }
        else
        {
            extns_ptr->extensions[i].critical = CSSM_FALSE; /* Default value of CSSM_FALSE*/

            /* Value */
            extns_ptr->extensions[i].value.tagAndValue = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 
                1, sizeof(CSSM_X509EXT_TAGandVALUE), poolptr);
            if (!extns_ptr->extensions[i].value.tagAndValue)
            {
                return CSSMERR_CL_MEMORY_ERROR;
            }
        
            tmp_cssm_data.Length = current_extn_node_ptr->Child[1].X.Input.ContentLength; 
            tmp_cssm_data.Data = current_extn_node_ptr->Child[1].X.Input.Content;
            if (tmp_cssm_data.Length && tmp_cssm_data.Data)
            {   
                if (!BER_ExpandItem((uint8*) tmp_cssm_data.Data, tmp_cssm_data.Length, &value_item))
                {
                    return CSSMERR_CL_UNKNOWN_FORMAT;
                }        
                ret = MP_DerDecodeItem(poolptr, &value_item, &extns_ptr->extensions[i].value.tagAndValue->value);
                if (CSSM_OK != ret)
                {
                    return ret;
                }
                extns_ptr->extensions[i].value.tagAndValue->type = value_item.Tag[0];
            
            }
        }
    }
    *extensions_ptr_ptr = extns_ptr;
    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: MP_DerDecodeExtension
 *
 * Description:
 * DER decodes a single X509 extension
 *
 * Parameters: 
 * Value (input): DER encoded object that shd be decoded
 *
 * extension_ptr_ptr (output) : Single Decoded CSSM_X509_EXTENSION object
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_DerDecodeExtension(MEMORY_POOL_PTR poolptr, 
                                  const CSSM_DATA *Value, 
                                  CSSM_X509_EXTENSION_PTR *extension_ptr_ptr)
{

    BER_PARSED_ITEM item;
    BER_PARSED_ITEM value_item;
    DER_NODE_PTR current_extn_node_ptr;
    CSSM_DATA tmp_cssm_data;
    CSSM_X509_EXTENSION_PTR extn_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;

    if (!poolptr || !extension_ptr_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    *extension_ptr_ptr = NULL;

    if (!BER_ExpandItem((uint8*) Value->Data, Value->Length, &item))
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, poolptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    current_extn_node_ptr = BER_ParseTypedObject(TLS_MemoryPool,
            &berder_funcs_wrapper, 
            &DER_CertExtensionType, &item);
    if (!current_extn_node_ptr || current_extn_node_ptr->Type != &DER_CertExtensionType ||
        !current_extn_node_ptr->Child)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    
    extn_ptr = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 1, sizeof(CSSM_X509_EXTENSION),
                        poolptr);
    if (!extn_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    
    
    if (!current_extn_node_ptr || 
        current_extn_node_ptr->Type != &DER_CertExtensionType ||
        !current_extn_node_ptr->Child)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    /* OID */
    extn_ptr->extnId.Length = current_extn_node_ptr->Child[0].X.Input.ContentLength;
    extn_ptr->extnId.Data = current_extn_node_ptr->Child[0].X.Input.Content;
    /* Critical */
    if (!current_extn_node_ptr->Child[1].X.Input.Content || !current_extn_node_ptr->Child[1].X.Input.Content[0])
    {
        extn_ptr->critical = CSSM_FALSE;
    }
    else
    {
        extn_ptr->critical = CSSM_TRUE;
    }
    /* Value */
    extn_ptr->value.tagAndValue = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 
        1, sizeof(CSSM_X509EXT_TAGandVALUE), poolptr);
    if (!extn_ptr->value.tagAndValue)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    
    tmp_cssm_data.Length = current_extn_node_ptr->Child[2].X.Input.ContentLength; 
    tmp_cssm_data.Data = current_extn_node_ptr->Child[2].X.Input.Content;
    if (tmp_cssm_data.Length && tmp_cssm_data.Data)
    {   
        if (!BER_ExpandItem((uint8*) tmp_cssm_data.Data, tmp_cssm_data.Length, &value_item))
        {
            return CSSMERR_CL_UNKNOWN_FORMAT;
        }        
        extn_ptr->value.tagAndValue->type = value_item.Tag[0];
        extn_ptr->value.tagAndValue->value.Length = value_item.ContentLength;
        extn_ptr->value.tagAndValue->value.Data = value_item.Content;
    }
    
    *extension_ptr_ptr = extn_ptr;
    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: MP_DerDecodeName
 *
 * Description:
 * DER decodes the encoded X509 Name structure
 *
 * Parameters: 
 * Value (input): DER encoded object that shd be decoded
 *
 * name_ptr_ptr (output) : Decoded CSSM_X509_NAME object
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_DerDecodeName(MEMORY_POOL_PTR poolptr, 
                             const CSSM_DATA *Value, 
                             CSSM_X509_NAME_PTR *name_ptr_ptr)
{
    
    BER_PARSED_ITEM item;
    DER_NODE_PTR name_node_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;

    if (!poolptr || ! name_ptr_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    if (!BER_ExpandItem((uint8*) Value->Data, Value->Length, &item))
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
 
    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, poolptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    name_node_ptr = BER_ParseTypedObject(TLS_MemoryPool,
            &berder_funcs_wrapper, 
            &DER_NameType, &item);

    if (!name_node_ptr)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    return MP_DerDecodeNameNode(poolptr, name_node_ptr, name_ptr_ptr);
    
}
/*-----------------------------------------------------------------------------
 * Name: MP_DerDecodeNameNode
 *
 * Description:
 * Converts the X509 Name parse tree into a CDSA CSSM_X509_NAME data structure
 *
 * Parameters: 
 * name_node_ptr (input): parse tree for a X509 Name
 *
 * name_ptr_ptr (output) : Decoded CSSM_X509_NAME object
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_DerDecodeNameNode(MEMORY_POOL_PTR poolptr, 
                                 DER_NODE_PTR name_node_ptr, 
                                 CSSM_X509_NAME_PTR *name_ptr_ptr)
{

    DER_NODE_PTR rdn_node_ptr = NULL;
    DER_NODE_PTR av_node_ptr = NULL;
    uint32 i,j;
    CSSM_X509_NAME_PTR name_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;

    if (!poolptr || !name_ptr_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    if (!name_node_ptr || name_node_ptr->Type != &DER_NameType ||
        !name_node_ptr->Child)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
 
    name_ptr = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 
        1, sizeof(CSSM_X509_NAME), poolptr);
    
    if (!name_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }

    
    /* But the count returned is the max possible (0x10).
       We need to reduce it by the number of NULL nodes in the end */
    i = 0;
    while (i < (uint32)name_node_ptr->Count && name_node_ptr->Child[i].X.Node)
    {
        i++;
    }
    if (i == 0)
    {
        *name_ptr_ptr = name_ptr;
        return CSSM_OK;
    }

    name_ptr->numberOfRDNs = i;
    name_ptr->RelativeDistinguishedName = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 
        name_ptr->numberOfRDNs, sizeof(CSSM_X509_RDN), poolptr);
    if (!name_ptr->RelativeDistinguishedName )
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    for(i=0;i<name_ptr->numberOfRDNs;i++)
    {
        rdn_node_ptr = name_node_ptr->Child[i].X.Node;
        if (!rdn_node_ptr || rdn_node_ptr->Type != &DER_AttributeSetType ||
        !rdn_node_ptr->Child)
        {
            return CSSMERR_CL_UNKNOWN_FORMAT;
        }
        name_ptr->RelativeDistinguishedName[i].numberOfPairs = rdn_node_ptr->Count;
        name_ptr->RelativeDistinguishedName[i].AttributeTypeAndValue = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 
            name_ptr->RelativeDistinguishedName[i].numberOfPairs, sizeof(CSSM_X509_TYPE_VALUE_PAIR), poolptr);
        if (!name_ptr->RelativeDistinguishedName[i].AttributeTypeAndValue )
        {
            return CSSMERR_CL_MEMORY_ERROR;
        }
        for(j=0;j<name_ptr->RelativeDistinguishedName[i].numberOfPairs;j++)
        {
            av_node_ptr = rdn_node_ptr->Child[j].X.Node;
            name_ptr->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Length = av_node_ptr->Child[0].X.Input.ContentLength;
            name_ptr->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Data = av_node_ptr->Child[0].X.Input.Content;
            name_ptr->RelativeDistinguishedName[i].AttributeTypeAndValue[j].valueType = av_node_ptr->Child[1].X.Input.Tag[0];
            ret = MP_DerDecodeItem(poolptr,&av_node_ptr->Child[1].X.Input, 
                &(name_ptr->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value));
            if (CSSM_OK != ret)
            {
                return ret;
            }
        }
    }
    *name_ptr_ptr = name_ptr;
    return CSSM_OK;
}
/*-----------------------------------------------------------------------------
 * Name: MP_DerDecodeTbsCertificate
 *
 * Description:
 * DER decodes the encoded TBS portion of a X509 certificate
 *
 * Parameters: 
 * Value (input): DER encoded object that shd be decoded
 *
 * cert_ptr_ptr (output) : Decoded CSSM_X509_TBS_CERTIFICATE object
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_DerDecodeTbsCertificate(MEMORY_POOL_PTR poolptr, 
                                       const CSSM_DATA *Value, 
                                       CSSM_X509_TBS_CERTIFICATE_PTR *cert_ptr_ptr)
{
    BER_PARSED_ITEM item;
    DER_NODE_PTR cert_node_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;
    
    if (!BER_ExpandItem((uint8*) Value->Data, Value->Length, &item))
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }

    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, poolptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    cert_node_ptr = BER_ParseTypedObject(TLS_MemoryPool,
            &berder_funcs_wrapper, 
            &DER_x509SignedPartType, &item);

    if (!cert_node_ptr || cert_node_ptr->Type != &DER_x509SignedPartType ||
        !cert_node_ptr->Child)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    
    return MP_DerDecodeTbsCertificateNode(poolptr, cert_node_ptr, cert_ptr_ptr);
}

/*-----------------------------------------------------------------------------
 * Name: MP_DerDecodeTbsCertificateNode
 *
 * Description:
 * Converts the parse tree for a TBS Certificate to a CDSA CSSM_X509_TBS_CERTIFICATE
 * data structure
 *
 * Parameters: 
 * cert_node_ptr (input): Parse tree for conversion
 *
 * cert_ptr_ptr (output) : Decoded CSSM_X509_TBS_CERTIFICATE object
 *---------------------------------------------------------------------------*/ 
CSSM_RETURN MP_DerDecodeTbsCertificateNode(MEMORY_POOL_PTR poolptr, 
                                           DER_NODE_PTR cert_node_ptr, 
                                           CSSM_X509_TBS_CERTIFICATE_PTR *cert_ptr_ptr)
{

    CSSM_X509_TBS_CERTIFICATE_PTR cert_ptr;
    CSSM_RETURN ret = CSSM_OK;
    CSSM_DATA temp_data;
    DER_NODE_PTR sign_node_ptr;
    DER_NODE_PTR validity_node_ptr;
    DER_NODE_PTR spki_node_ptr;
    DER_NODE_PTR spki_sign_node_ptr;
    CSSM_X509_NAME_PTR issuer=NULL, subject=NULL;
    CSSM_X509_EXTENSIONS_PTR extensions_ptr;

    if (!poolptr || !cert_ptr_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    cert_ptr = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 
        1, sizeof(CSSM_X509_TBS_CERTIFICATE), poolptr);
    if (!cert_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    
    if (!cert_node_ptr || cert_node_ptr->Type != &DER_x509SignedPartType ||
        !cert_node_ptr->Child)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }

    /* Version */
    temp_data.Length = cert_node_ptr->Child[0].X.Input.ContentLength;
    temp_data.Data = cert_node_ptr->Child[0].X.Input.Content;
    ret = MP_DerDecodeObject(poolptr, temp_data, &cert_ptr->version);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    /* Serial Number */
    ret = MP_DerDecodeItem(poolptr, &cert_node_ptr->Child[1].X.Input, &cert_ptr->serialNumber);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    /* Signature Algorithm: OID and Parameters */
    sign_node_ptr = cert_node_ptr->Child[2].X.Node;
    if (!sign_node_ptr || sign_node_ptr->Type != &DER_AlgorithmIdentifierType
        || !sign_node_ptr->Child )
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    ret = MP_DerDecodeItem(poolptr, &sign_node_ptr->Child[0].X.Input, &cert_ptr->signature.algorithm);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    
    if (sign_node_ptr->Child[1].X.Input.Tag && sign_node_ptr->Child[1].X.Input.ContentLength)
    {
        cert_ptr->signature.parameters.Data = sign_node_ptr->Child[1].X.Input.Tag;
        cert_ptr->signature.parameters.Length = BER_LengthOfTag(sign_node_ptr->Child[1].X.Input.Tag) 
                                            + BER_LengthOfLength(sign_node_ptr->Child[1].X.Input.ContentLength)
                                            + sign_node_ptr->Child[1].X.Input.ContentLength ;
    }

    /* Issuer Name */
    ret = MP_DerDecodeNameNode(poolptr, cert_node_ptr->Child[3].X.Node, &issuer);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    cert_ptr->issuer = *issuer;

    /* Validity */
    validity_node_ptr = cert_node_ptr->Child[4].X.Node;
    ret = MP_DerDecodeItem(poolptr, &validity_node_ptr->Child[0].X.Input, &cert_ptr->validity.notBefore.time);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    cert_ptr->validity.notBefore.timeType = validity_node_ptr->Child[0].X.Input.Tag[0];

    ret = MP_DerDecodeItem(poolptr, &validity_node_ptr->Child[1].X.Input, &cert_ptr->validity.notAfter.time);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    cert_ptr->validity.notAfter.timeType = validity_node_ptr->Child[1].X.Input.Tag[0];

    /* Subject Name */
    ret = MP_DerDecodeNameNode(poolptr, cert_node_ptr->Child[5].X.Node, &subject);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    cert_ptr->subject = *subject;

    /* SPKI */
    spki_node_ptr = cert_node_ptr->Child[6].X.Node;
    /* Alg ID */
    spki_sign_node_ptr = spki_node_ptr->Child[0].X.Node;
    ret = MP_DerDecodeItem(poolptr, &spki_sign_node_ptr->Child[0].X.Input, &cert_ptr->subjectPublicKeyInfo.algorithm.algorithm);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    /* Parameter must be left DER encoded */
    
    if (spki_sign_node_ptr->Child[1].X.Input.ContentLength)
    {
        cert_ptr->subjectPublicKeyInfo.algorithm.parameters.Length = BER_LengthOfTag(spki_sign_node_ptr->Child[1].X.Input.Tag)  
            + BER_LengthOfLength(spki_sign_node_ptr->Child[1].X.Input.ContentLength) 
            + spki_sign_node_ptr->Child[1].X.Input.ContentLength;
        cert_ptr->subjectPublicKeyInfo.algorithm.parameters.Data = spki_sign_node_ptr->Child[1].X.Input.Tag;
    }
    
    /* Key*/
    ret = MP_DerDecodeItem(poolptr, &spki_node_ptr->Child[1].X.Input, &cert_ptr->subjectPublicKeyInfo.subjectPublicKey);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    
    /* Issuer UID */
    ret = MP_DerDecodeItem(poolptr, &cert_node_ptr->Child[7].X.Input, &cert_ptr->issuerUniqueIdentifier);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    
    /* Subject UID */
    ret = MP_DerDecodeItem(poolptr, &cert_node_ptr->Child[8].X.Input, &cert_ptr->subjectUniqueIdentifier);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    

    /* Extensions */
    if (cert_node_ptr->Child[9].X.Node)
    {
        ret = MP_DerDecodeExtensionsNode(poolptr, cert_node_ptr->Child[9].X.Node, &extensions_ptr);
        if (CSSM_OK != ret)
        {
            return ret;
        }
        cert_ptr->extensions = *extensions_ptr;
    }
    
    *cert_ptr_ptr = cert_ptr;
    return CSSM_OK;
}
   
/*-----------------------------------------------------------------------------
 * Name: MP_DerDecodeSignedCertificate
 *
 * Description:
 * DER decodes the encoded X509 certificate
 *
 * Parameters: 
 * Value (input): DER encoded object that shd be decoded
 *
 * signed_cert_ptr_ptr (output) : Decoded CSSM_X509_SIGNED_CERTIFICATE object
 *---------------------------------------------------------------------------*/    
CSSM_RETURN MP_DerDecodeSignedCertificate(MEMORY_POOL_PTR poolptr, 
                                    const CSSM_DATA *Value, 
                                    CSSM_X509_SIGNED_CERTIFICATE_PTR *signed_cert_ptr_ptr)
{
    BER_PARSED_ITEM item;
    DER_NODE_PTR signed_cert_node_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;
    
    if (!poolptr || !signed_cert_ptr_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    if (!BER_ExpandItem((uint8*) Value->Data, Value->Length, &item))
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }

    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, poolptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    signed_cert_node_ptr = BER_ParseTypedObject(TLS_MemoryPool,
            &berder_funcs_wrapper, 
            &DER_x509certType, &item);

    
    return MP_DerDecodeCertificateNode(poolptr, signed_cert_node_ptr, signed_cert_ptr_ptr);
}

/*-----------------------------------------------------------------------------
 * Name: MP_DerDecodeCertificateNode
 *
 * Description:
 * Converts the parse tree for a Certificate to a CDSA CSSM_X509_SIGNED_CERTIFICATE
 * data structure
 *
 * Parameters: 
 * signed_cert_node_ptr (input): Parse tree for conversion
 *
 * signed_cert_ptr_ptr (output) : Decoded CSSM_X509_SIGNED_CERTIFICATE object
 *---------------------------------------------------------------------------*/ 
CSSM_RETURN MP_DerDecodeCertificateNode(MEMORY_POOL_PTR poolptr, 
                                        DER_NODE_PTR signed_cert_node_ptr, 
                                        CSSM_X509_SIGNED_CERTIFICATE_PTR *signed_cert_ptr_ptr)
{
    DER_NODE_PTR cert_node_ptr;
    DER_NODE_PTR algoid_node_ptr;
    CSSM_X509_SIGNED_CERTIFICATE_PTR signed_cert_ptr;
    CSSM_X509_TBS_CERTIFICATE_PTR cert_ptr;
    CSSM_RETURN ret = CSSM_OK;
    
    if (!poolptr || !signed_cert_ptr_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    if (!signed_cert_node_ptr || signed_cert_node_ptr->Type != &DER_x509certType ||
        !signed_cert_node_ptr->Child)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }

    
    signed_cert_ptr = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 
        1, sizeof(CSSM_X509_SIGNED_CERTIFICATE), poolptr);
    if (!signed_cert_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }

    /* TBS Certificate part */
    cert_node_ptr = signed_cert_node_ptr->Child[0].X.Node;
    ret = MP_DerDecodeTbsCertificateNode(poolptr, cert_node_ptr, &cert_ptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    signed_cert_ptr->certificate = *cert_ptr;


    /* algoid */
    algoid_node_ptr = signed_cert_node_ptr->Child[1].X.Node;
    ret = MP_DerDecodeItem(poolptr, &algoid_node_ptr->Child[0].X.Input, &signed_cert_ptr->signature.algorithmIdentifier.algorithm);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    if (algoid_node_ptr->Child[1].X.Input.Tag && algoid_node_ptr->Child[1].X.Input.ContentLength)
    {
        signed_cert_ptr->signature.algorithmIdentifier.parameters.Data = algoid_node_ptr->Child[1].X.Input.Tag;
        signed_cert_ptr->signature.algorithmIdentifier.parameters.Length = BER_LengthOfTag(algoid_node_ptr->Child[1].X.Input.Tag) 
                                            + BER_LengthOfLength(algoid_node_ptr->Child[1].X.Input.ContentLength)
                                            + algoid_node_ptr->Child[1].X.Input.ContentLength ;
    }
    
    /* signature */
    ret = MP_DerDecodeItem(poolptr, &signed_cert_node_ptr->Child[2].X.Input, &signed_cert_ptr->signature.encrypted);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    *signed_cert_ptr_ptr = signed_cert_ptr;
    return CSSM_OK;
}




/*-----------------------------------------------------------------------------
 *  General Utility functions that make use of the mempool follow this line
 *---------------------------------------------------------------------------*/ 
/*-----------------------------------------------------------------------------
 * Name: MP_DuplicateCNameStruct
 *
 * Description:
 * Produces a copy of the input CSSM_X509_NAME struct using the memory pool
 *
 * Parameters: 
 * SrcName (input): Input CSSM_X509_NAME object
 *
 * result_val (output) : Copy of the CSSM_X509_NAME
 * NOTE: This function also exists in misc_nam.c (clutil) without using the mempool
 *---------------------------------------------------------------------------*/ 
CSSM_RETURN MP_DuplicateCNameStruct(MEMORY_POOL_PTR poolptr,
                                    const CSSM_X509_NAME *SrcName,
                                    CSSM_X509_NAME_PTR *result_val)
{
    CSSM_X509_NAME_PTR DestName = NULL;
    uint32 i=0, j=0;
    CSSM_BOOL Error;

    if (!poolptr || !result_val)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    *result_val = NULL;
    /* validate parameters */
    if((NULL == SrcName) ||
       (port_IsBadReadPtr(SrcName,
                          sizeof(CSSM_X509_NAME_PTR))) ||
       (port_IsBadReadPtr(SrcName->RelativeDistinguishedName,
                          SrcName->numberOfRDNs * sizeof(CSSM_X509_RDN))))
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
    }
    else
    {
        /* allocate a CSSM_X509_NAME for a return value */
        DestName = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 
            1, sizeof(CSSM_X509_NAME), poolptr); 
        
        if(!DestName)
        {
            return CSSMERR_CL_MEMORY_ERROR;
        }
        else
        {
            /* set the number of RDNs and allocate their array */
            DestName->numberOfRDNs = SrcName->numberOfRDNs;
            DestName->RelativeDistinguishedName = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 
            1, (sizeof(CSSM_X509_RDN) * SrcName->numberOfRDNs), poolptr); 
                
            /* check for memory errors */
            if(!DestName->RelativeDistinguishedName)
            {
                DestName = NULL;
                return CSSMERR_CL_MEMORY_ERROR;
            }
            else
            {
                i = 0;
                Error = CSSM_FALSE;
                /* For each RDN fill in the info */
                while((CSSM_FALSE == Error) &&
                      (i < SrcName->numberOfRDNs))
                {
                    DestName->RelativeDistinguishedName[i].numberOfPairs = SrcName->RelativeDistinguishedName[i].numberOfPairs;
                    DestName->RelativeDistinguishedName[i].AttributeTypeAndValue = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 
                        1, 
                        (sizeof(CSSM_X509_TYPE_VALUE_PAIR) * SrcName->RelativeDistinguishedName[i].numberOfPairs), 
                        poolptr); 
                                                                                                               
                    if(!DestName->RelativeDistinguishedName[i].AttributeTypeAndValue)
                    {
                        DestName = NULL;
                        Error = CSSM_TRUE;
                        return CSSMERR_CL_MEMORY_ERROR;
                    }
                    else
                    {
                        j = 0;
                        /* For each attribute/value pair */
                        while((CSSM_FALSE == Error) &&
                              (j < SrcName->RelativeDistinguishedName[i].numberOfPairs))
                        {
                            /* copy the type */
                            DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Length = SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Length;
                            DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Data = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 
                                1,
                                SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Length,
                                poolptr);

                            /* if there was an error, stop */
                            if(!DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Data)
                            {
                                
                                DestName = NULL;
                                Error = CSSM_TRUE;
                                return CSSMERR_CL_MEMORY_ERROR;
                            }
                            else
                            {
                                port_memcpy(DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Data,
                                       SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Data,
                                       SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].type.Length);


                                /* copy the valueType */
                                DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].valueType = 
                                    SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].valueType;


                                /* copy the value */
                                DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Length = SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Length;
                                DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Data = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 
                                1,
                                SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Length,
                                poolptr);
                                if(!DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Data)
                                {
                                    
                                    DestName = NULL;
                                    return CSSMERR_CL_MEMORY_ERROR;
                                }
                                else
                                {
                                    port_memcpy(DestName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Data,
                                           SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Data,
                                           SrcName->RelativeDistinguishedName[i].AttributeTypeAndValue[j].value.Length);
                                }/* end else */
                            }/* end else */
                            j++;
                        }/* end while */
                    }/* end else */
                    i++;
                }/* end while */
            }/* end else */
        }/* end else */
    }/* end else */

   *result_val =  DestName;
   return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: MP_DuplicateSPKI
 *
 * Description: 
 * This function duplicates one Intel X509V3 certificate Subject Public Key Info(SPKI)
 *
 * Parameters: 
 * AppHandle (input)  : Application Handle
 * SrcKey (input)     : the SPKI to be duplicated
 *
 * result_val (output):
 * A CSSM_X509_SUBJECT_PUBLIC_KEY_INFO_PTR structure which contains the duplicated SPKI.
 *
 * NOTE: This function also exists in misc_nam.c (clutil) without using the mempool
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_DuplicateSPKI (MEMORY_POOL_PTR poolptr, 
                                      const CSSM_X509_SUBJECT_PUBLIC_KEY_INFO *SrcKey,
                                      CSSM_X509_SUBJECT_PUBLIC_KEY_INFO_PTR *result_val)
{
    CSSM_X509_SUBJECT_PUBLIC_KEY_INFO_PTR DestKey = NULL;
    void    *temp_ptr = NULL;
  
    if (!poolptr || !result_val)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    *result_val = NULL;
    /* validate parameters */
    if (port_IsBadReadPtr(SrcKey, sizeof(CSSM_X509_SUBJECT_PUBLIC_KEY_INFO)))
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
        
    }
   
    if (!poolptr)
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
    }
    /* allocate a SPKI buffer for destination key */
    DestKey = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 1,
                                           sizeof(CSSM_X509_SUBJECT_PUBLIC_KEY_INFO), poolptr);
    if(!DestKey)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }


    if ((NULL != SrcKey->algorithm.algorithm.Data) && (!port_IsBadReadPtr(
        SrcKey->algorithm.algorithm.Data, SrcKey->algorithm.algorithm.Length))) {
    
         /* Allocate memory for the new data (algorithm) */
        temp_ptr = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 1, 
                                            SrcKey->algorithm.algorithm.Length, poolptr);
        if (!temp_ptr)
        {
            
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
        
        /* Copy the new data into the newly allocated memory */
        port_memcpy (temp_ptr, SrcKey->algorithm.algorithm.Data,
                                        SrcKey->algorithm.algorithm.Length);

         /* Copy input fields into the cert */
        DestKey->algorithm.algorithm.Data = temp_ptr;
        DestKey->algorithm.algorithm.Length = SrcKey->algorithm.algorithm.Length;
    }

    if ((NULL != SrcKey->algorithm.parameters.Data) && (!port_IsBadReadPtr(
        SrcKey->algorithm.parameters.Data, SrcKey->algorithm.parameters.Length))) {

         /* Allocate memory for the new data (parameters) */
        temp_ptr = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func,  1, 
                                                SrcKey->algorithm.parameters.Length, poolptr);
        if (!temp_ptr)
        {
            
            
            return CSSMERR_CL_MEMORY_ERROR;
            
        }

        /* Copy the new data into the newly allocated memory */
        port_memcpy (temp_ptr, SrcKey->algorithm.parameters.Data, SrcKey->algorithm.parameters.Length);

         /* Copy input fields into the cert */
        DestKey->algorithm.parameters.Data = temp_ptr;
        DestKey->algorithm.parameters.Length = SrcKey->algorithm.parameters.Length;
    }

    if ((NULL != SrcKey->subjectPublicKey.Data) && (!port_IsBadReadPtr(
        SrcKey->subjectPublicKey.Data, SrcKey->subjectPublicKey.Length))) {

        /* Allocate memory for the new data (SubjectPublicKey) */
        temp_ptr = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func,  1, 
                                                SrcKey->subjectPublicKey.Length, poolptr);
        if (!temp_ptr)
        {
            
            return CSSMERR_CL_MEMORY_ERROR;
            
        }

        /* Copy the new data into the newly allocated memory */
        port_memcpy (temp_ptr, SrcKey->subjectPublicKey.Data, SrcKey->subjectPublicKey.Length);

         /* Copy input fields into the cert */
        DestKey->subjectPublicKey.Data = temp_ptr;
        DestKey->subjectPublicKey.Length = SrcKey->subjectPublicKey.Length;
    }

    /*
     * Check if any input fields were valid.
     */
    if (!DestKey->algorithm.algorithm.Data && !DestKey->algorithm.parameters.Data  
                && !DestKey->subjectPublicKey.Data)
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
        
    } else {
        *result_val = DestKey;
        return CSSM_OK;
    }
}

/*-----------------------------------------------------------------------------
 * Name: MP_X509_DuplicateSignature
 *
 * Description: 
 * This function duplicates one Intel X509V3 certificate signature.
 *
 * Parameters: 
 * AppHandle (input)  : Application Handle
 * Signature (input)     : the signature to be duplicated
 *
 * result_val (output):
 * A CSSM_X509_SIGNATURE_PTR structure which contains the duplicated signature.
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_DuplicateSignature (MEMORY_POOL_PTR poolptr, 
                                      const CSSM_X509_SIGNATURE *SrcSig,
                                      CSSM_X509_SIGNATURE_PTR *result_val)
{
    CSSM_X509_SIGNATURE_PTR DestSig = NULL;

    if (!poolptr || !result_val)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
   
    *result_val = NULL;
    /* validate parameters */
    if (port_IsBadReadPtr(SrcSig, sizeof(CSSM_X509_SIGNATURE_PTR)))
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
        
    }
   
    /* allocate a CSSM_X509_SIGNATURE buffer for destination Signature */
    DestSig = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 1,
                                           sizeof(CSSM_X509_SIGNATURE), poolptr);
    if(!DestSig)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }


    if ((NULL != SrcSig->algorithmIdentifier.algorithm.Data) 
        && (!port_IsBadReadPtr(SrcSig->algorithmIdentifier.algorithm.Data,
                            SrcSig->algorithmIdentifier.algorithm.Length))) 
    {
    
        DestSig->algorithmIdentifier.algorithm.Length = SrcSig->algorithmIdentifier.algorithm.Length;
         /* Allocate memory for the new data (algorithm) */
        DestSig->algorithmIdentifier.algorithm.Data = 
            CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 
                                        1, 
                                        SrcSig->algorithmIdentifier.algorithm.Length,
                                        poolptr);
        if (!DestSig->algorithmIdentifier.algorithm.Data)
        {
            return CSSMERR_CL_MEMORY_ERROR;
        }
        
        /* Copy the new data into the newly allocated memory */
        port_memcpy (DestSig->algorithmIdentifier.algorithm.Data, 
                     SrcSig->algorithmIdentifier.algorithm.Data,
                     SrcSig->algorithmIdentifier.algorithm.Length);

         
        
    }

    if ((NULL != SrcSig->algorithmIdentifier.parameters.Data) 
        && (!port_IsBadReadPtr(SrcSig->algorithmIdentifier.parameters.Data, 
                               SrcSig->algorithmIdentifier.parameters.Length))) 
    {

        DestSig->algorithmIdentifier.parameters.Length =
                                        SrcSig->algorithmIdentifier.parameters.Length;
         /* Allocate memory for the new data (parameters) */
        DestSig->algorithmIdentifier.parameters.Data = 
            CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 1, 
                                    SrcSig->algorithmIdentifier.parameters.Length, poolptr);
        
        if (!DestSig->algorithmIdentifier.parameters.Data)
        {
            return CSSMERR_CL_MEMORY_ERROR; 
        }

        /* Copy the new data into the newly allocated memory */
        port_memcpy (DestSig->algorithmIdentifier.parameters.Data, 
                     SrcSig->algorithmIdentifier.parameters.Data, 
                     SrcSig->algorithmIdentifier.parameters.Length);

         
    }

    if ((NULL != SrcSig->encrypted.Data) 
        && (!port_IsBadReadPtr(SrcSig->encrypted.Data, 
                               SrcSig->encrypted.Length))) 
    {

        DestSig->encrypted.Length = SrcSig->encrypted.Length;
        /* Allocate memory for the new data (SubjectPublicKey) */
        DestSig->encrypted.Data = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 1, 
                                                SrcSig->encrypted.Length, poolptr);
        if (!DestSig->encrypted.Data)
        {
            return CSSMERR_CL_MEMORY_ERROR;
        }

        /* Copy the new data into the newly allocated memory */
        port_memcpy (DestSig->encrypted.Data, SrcSig->encrypted.Data, SrcSig->encrypted.Length);

    }

    /*
     * Check if any input fields were valid.
     */
    if (!DestSig->algorithmIdentifier.algorithm.Data &&
        !DestSig->algorithmIdentifier.parameters.Data  && !DestSig->encrypted.Data)
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
        
    } 
    else 
    {
        *result_val = DestSig;
        return CSSM_OK;
    }

}

/*-----------------------------------------------------------------------------
 * Name: MP_CSSMKeyToSpki
 *
 * Description:
 * This function translates the CSSM_KEY struct into an SPKI struct
 *
 * Parameters: 
 * CSSMKey  (input)   : A CSSM_KEY structure to be translated
 *
 * spki_ptr_ptr (output):
 * An SPKI representation of the CSSM_KEY
 *---------------------------------------------------------------------------*/

CSSM_RETURN MP_CSSMKeyToSpki(MEMORY_POOL_PTR poolptr, 
                             CSSM_KEY_PTR CSSMKey, 
                             CSSM_X509_SUBJECT_PUBLIC_KEY_INFO_PTR *spki_ptr_ptr)
{
    CSSM_X509_SUBJECT_PUBLIC_KEY_INFO_PTR spki_ptr = NULL;
    CSSM_OID_PTR algoid_ptr = NULL;
    BER_PARSED_ITEM key_item;
    DER_NODE_PTR key_node_ptr;
    uint32 i;
    CSSM_RETURN ret = CSSM_OK;
    CSSM_DATA tmp_data;

    if (!poolptr || !spki_ptr_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    *spki_ptr_ptr = NULL;
    if ( !CSSMKey->KeyData.Data ||
         port_IsBadReadPtr(CSSMKey->KeyData.Data, CSSMKey->KeyData.Length) ) 
    {
        return CSSM_OK;
    }

    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, poolptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    /* Allocate memory for the SPKI */
    spki_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_X509_SUBJECT_PUBLIC_KEY_INFO),
        poolptr);
    if (!spki_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }

    /* Obtain the algorithm OID */
    algoid_ptr = X509_AlgIdToAlgorithmOid(CSSMKey->KeyHeader.AlgorithmId);

    if (algoid_ptr == NULL)
    {
    
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }

    spki_ptr->algorithm.algorithm.Data = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 
        1, algoid_ptr->Length, poolptr);

    if (spki_ptr->algorithm.algorithm.Data == NULL)
    {
    
        return CSSMERR_CL_MEMORY_ERROR; 
    }
    
    spki_ptr->algorithm.algorithm.Length = algoid_ptr->Length;
    port_memcpy(spki_ptr->algorithm.algorithm.Data, algoid_ptr->Data, algoid_ptr->Length);

    if (CSSM_ALGID_DSA == CSSMKey->KeyHeader.AlgorithmId)
    {
    
        BER_ExpandItem(CSSMKey->KeyData.Data, CSSMKey->KeyData.Length, &key_item);
        key_node_ptr = BER_ParseComplexObject(TLS_MemoryPool, &berder_funcs_wrapper, NULL, &key_item);
        /* Check that the parse tree we got was the correct one */
        if (!key_node_ptr || !key_node_ptr->Child || key_node_ptr->Count != 4)
        {
            return CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT;
        }
        for(i=0;i<(uint32)key_node_ptr->Count;i++)
        {
            if (!key_node_ptr->Child[i].IsLeaf || 
                key_node_ptr->Child[i].X.Input.Tag[0] != BER_INTEGER)
            {
                return CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT;
            }
        }
        
        /* Encode the first 3 children into the params */
        key_node_ptr->Count--;
        spki_ptr->algorithm.parameters.Data = DER_Generate(TLS_MemoryPool, &berder_funcs_wrapper, key_node_ptr);
        if (!spki_ptr->algorithm.parameters.Data)
        {
            return CSSMERR_CL_UNKNOWN_FORMAT;
        }
        spki_ptr->algorithm.parameters.Length = BER_SizeofObject(spki_ptr->algorithm.parameters.Data);

        key_node_ptr->Count++;

        /* The BER_PARSED_ITEM as a CSSM_DATA */
        tmp_data.Data = key_node_ptr->Child[3].X.Input.Content;
        tmp_data.Length = key_node_ptr->Child[3].X.Input.ContentLength;
        ret = MP_EncodeObject(poolptr, BER_INTEGER, tmp_data, &spki_ptr->subjectPublicKey);
        if (CSSM_OK != ret)
        {
            return ret;
        }

    }
    else
    {
        spki_ptr->subjectPublicKey.Length = CSSMKey->KeyData.Length;
        spki_ptr->subjectPublicKey.Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
                                    spki_ptr->subjectPublicKey.Length, poolptr);
        if (!spki_ptr->subjectPublicKey.Data)
        {
            return CSSMERR_CL_MEMORY_ERROR;
        }
        port_memcpy(spki_ptr->subjectPublicKey.Data, CSSMKey->KeyData.Data, spki_ptr->subjectPublicKey.Length);
        spki_ptr->algorithm.parameters.Length = 0;
        spki_ptr->algorithm.parameters.Data = NULL;
    }

    *spki_ptr_ptr = spki_ptr;
    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: MP_CopyAlgId
 *
 * Description:
 * Makes a copy of a CSSM_X509_ALGORITHM_IDENTIFIER
 *
 * Parameters: 
 * AlgIdIn (input)   : original alg id struct
 *
 * AlgIdOut (output): Copy of AlgIdIn (using mempool)
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_CopyAlgId(MEMORY_POOL_PTR poolptr, 
                         CSSM_X509_ALGORITHM_IDENTIFIER_PTR AlgIdIn,
                         CSSM_X509_ALGORITHM_IDENTIFIER_PTR AlgIdOut)
{

    if (!poolptr || !AlgIdOut)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    
    /* Copy algorithm */
    AlgIdOut->algorithm.Length = AlgIdIn->algorithm.Length;
    if (AlgIdOut->algorithm.Length)
    {
        AlgIdOut->algorithm.Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
            AlgIdOut->algorithm.Length,
            poolptr);
        
        if (!AlgIdOut->algorithm.Data)
        {
            return CSSMERR_CL_MEMORY_ERROR;
        }
        port_memcpy(AlgIdOut->algorithm.Data, AlgIdIn->algorithm.Data, AlgIdOut->algorithm.Length);
    } 
    else 
    {
        AlgIdOut->algorithm.Data = NULL;
    }

    /* Copy parameters */
    AlgIdOut->parameters.Length = AlgIdIn->parameters.Length;
    if (AlgIdOut->parameters.Length)
    {
        AlgIdOut->parameters.Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
            AlgIdOut->parameters.Length,
            poolptr);
        if (!AlgIdOut->parameters.Data)
        {
            return CSSMERR_CL_MEMORY_ERROR;
        }
        port_memcpy(AlgIdOut->parameters.Data, AlgIdIn->parameters.Data, AlgIdOut->parameters.Length);
    } 
    else 
    {
        AlgIdOut->parameters.Data = NULL;
    }

    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: MP_SpkiToCSSMKey
 *
 * Description:
 * This function translates the CSSM_KEY struct into an SPKI struct
 *
 * Parameters: 
 * SPKI  (input)   : A SPKI structure to be encoded
 *
 * key_ptr (output): An CSSM_KEY representation of the SPKI structure
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_SpkiToCSSMKey(MEMORY_POOL_PTR poolptr,
                             CSSM_X509_SUBJECT_PUBLIC_KEY_INFO_PTR SPKI,
                             CSSM_KEY_PTR *key_ptr)
{
    
    /* Initializations */
    CSSM_RETURN         ret = CSSM_OK;
    CSSM_DATA_PTR       keydata_ptr = NULL;


    if (!poolptr || !key_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    *key_ptr = NULL;
    

    
    /* Allocate room for the new key structure */
    *key_ptr = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func, 
        1, sizeof(CSSM_KEY), poolptr);
    
    if (!*key_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }

    /* Fill in the KeyHeader fields */
    (*key_ptr)->KeyHeader.HeaderVersion = CSSM_KEYHEADER_VERSION;
    /* (*key_ptr)->KeyHeader.CspId is unknown (remains 0) */
    (*key_ptr)->KeyHeader.BlobType = CSSM_KEYBLOB_RAW;
    (*key_ptr)->KeyHeader.AlgorithmId = X509_AlgorithmOidToAlgId(&SPKI->algorithm.algorithm);
    
    switch ( (*key_ptr)->KeyHeader.AlgorithmId ) 
    {
        case CSSM_ALGID_DSA :
        case CSSM_ALGID_SHA1WithDSA :
            (*key_ptr)->KeyHeader.Format = CSSM_KEYBLOB_RAW_FORMAT_FIPS186;
            break;
        case CSSM_ALGID_RSA :
        case CSSM_ALGID_MD5WithRSA :
        case CSSM_ALGID_MD2WithRSA :
        case CSSM_ALGID_SHA1WithRSA :
            (*key_ptr)->KeyHeader.Format = CSSM_KEYBLOB_RAW_FORMAT_PKCS1;
            break;
        case CSSM_ALGID_DH :
            (*key_ptr)->KeyHeader.Format = CSSM_KEYBLOB_RAW_FORMAT_PKCS3;
            break;
        default:
            (*key_ptr)->KeyHeader.Format = CSSM_KEYBLOB_RAW_FORMAT_OTHER;
    }

    (*key_ptr)->KeyHeader.KeyClass = CSSM_KEYCLASS_PUBLIC_KEY;
    /* (*key_ptr)->KeyHeader.KeyAttr remains 0 */
    /* (*key_ptr)->KeyHeader.KeyUsage is unknown  (set to any usage) */
    (*key_ptr)->KeyHeader.KeyUsage = CSSM_KEYUSE_ANY;
    /* (*key_ptr)->KeyHeader.StartDate is unknown  (remains 0) */
    /* (*key_ptr)->KeyHeader.EndDate is unknown  (remains 0) */
    (*key_ptr)->KeyHeader.WrapAlgorithmId = CSSM_ALGID_NONE;
    (*key_ptr)->KeyHeader.WrapMode = CSSM_ALGMODE_NONE;

    /* Fill in the key based on the key data and params */
    ret = MP_GetKeyFromSpki(poolptr, (*key_ptr)->KeyHeader.AlgorithmId, SPKI, &keydata_ptr);
    if (CSSM_OK != ret || !keydata_ptr)
    {
        return ret;
    }
    (*key_ptr)->KeyData = *keydata_ptr;
    

    ret = MP_ObtainEffectiveKeySize(*key_ptr, &(*key_ptr)->KeyHeader.LogicalKeySizeInBits);
    if (CSSM_OK != ret)
    {
        return ret;
    }
    
    if (!(*key_ptr)->KeyHeader.LogicalKeySizeInBits)
        (*key_ptr)->KeyHeader.LogicalKeySizeInBits = 8*SPKI->subjectPublicKey.Length;

    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: MP_ObtainEffectiveKeySize
 *
 * Description:
 * Obtains the effective keysize (in bits) of the input key
 * 
 * Parameters: 
 * Key (input): CSSM_KEY whose effective keysize has to be determined
 * EffKeySize (output): Effective Key size in bits
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_ObtainEffectiveKeySize (CSSM_KEY_PTR Key,
                                       uint32 *EffKeySize)
{


    CSSM_RETURN ret = CSSM_OK;

    CSSM_DATA KeyParts[CSM_MAX_PUBLIC_KEY_PARTS];
    uint32 NumKeyParts = sizeof(KeyParts) / sizeof(CSSM_DATA);


    if (!EffKeySize)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    *EffKeySize = 0;

    /* Clear the key part array */
    port_memset( KeyParts, 0, sizeof(KeyParts) );

    /* Parse the key parts */
    ret = CSM_ParseKeyblob( Key->KeyHeader.AlgorithmId,
                                 Key->KeyHeader.KeyClass,
                                 Key->KeyHeader.Format,
                                 KeyParts,
                                 &NumKeyParts,
                                 &Key->KeyData );
    if ( ret == CSSM_OK )
    {
        switch ( Key->KeyHeader.AlgorithmId )
        {
        case CSSM_ALGID_RSA:
            {
                *EffKeySize = KeyParts[CSM_RSA_MODULUS].Length * 8;
                break;
            }
        case CSSM_ALGID_DSA:
            {
                *EffKeySize = KeyParts[CSM_DSA_PRIME].Length * 8;
                break;
            }
        default:
            {
                /* Assume the length of the first parameter */
                *EffKeySize = KeyParts[0].Length * 8;
                break;
            }
        }
    }
    else
    {
        *EffKeySize = 0;
    }

    
    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: MP_GetKeyFromSpki
 *
 * Description:
 * This function parses the CSSM_KEY into its key and parameter components
 * based on the key generation algorithm.
 * NOTE:  Currently, it only checks for the RSA and DSA algorithms.
 *        The RSA algorithm is equivalent to the default behavior.
 *        All other algorithms will default to the parameters = NULL and the 
 *        key data equal to whatever is in the CSSM_KEY structure for the key
 *
 * Parameters: 
 * AlgId (input) : Algorithm identifier
 * SpkiPtr (input): SPKI structure that contains the key 
 * key_ptr(output): The output key
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_GetKeyFromSpki(MEMORY_POOL_PTR poolptr,
                              CSSM_ALGORITHMS AlgId, 
                              CSSM_X509_SUBJECT_PUBLIC_KEY_INFO_PTR SpkiPtr,
                              CSSM_DATA_PTR *key_ptr)
{

    
    CSSM_RETURN ret = CSSM_OK;
    if (!poolptr || !key_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    *key_ptr = NULL;

    switch (AlgId)
    {
       case CSSM_ALGID_DSA :
           return MP_DSAKeyRejoin(poolptr, SpkiPtr, key_ptr);
           
       default :
           if (!SpkiPtr) 
           {
               return CSSMERR_CL_INVALID_INPUT_POINTER;
           }
           *key_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_DATA), poolptr);
       
           if (!*key_ptr)
           {
              return CSSMERR_CL_MEMORY_ERROR;
           }
           (*key_ptr)->Length = SpkiPtr->subjectPublicKey.Length;
           (*key_ptr)->Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, (*key_ptr)->Length, poolptr);
       
           if (!(*key_ptr)->Data)
           {
               return CSSMERR_CL_MEMORY_ERROR;
          
           }
           port_memcpy((*key_ptr)->Data, SpkiPtr->subjectPublicKey.Data, (*key_ptr)->Length);
           return ret;
    }
}


/*-----------------------------------------------------------------------------
 * Name: MP_DSAKeyRejoin
 *
 * Description:
 *
 * Parameters: 
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN MP_DSAKeyRejoin(MEMORY_POOL_PTR poolptr,                            
                            CSSM_X509_SUBJECT_PUBLIC_KEY_INFO_PTR SpkiPtr,
                            CSSM_DATA_PTR *key_ptr)
{
    /* Initializations */
    BER_PARSED_ITEM spki_params_item;
    BER_PARSED_ITEM spki_key_item;
    DER_NODE_PTR spki_params_node_ptr = NULL;
    DER_NODE_PTR key_node_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;
    uint32 i=0;

    if (!poolptr || !key_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    *key_ptr = NULL;

    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, poolptr);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    /* Parse the parameters */
    if (!BER_ExpandItem(SpkiPtr->algorithm.parameters.Data, SpkiPtr->algorithm.parameters.Length, &spki_params_item))
    {
        return CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT;
    }
    spki_params_node_ptr = BER_ParseComplexObject(TLS_MemoryPool, &berder_funcs_wrapper, NULL, &spki_params_item);
    
    /* Check that it is a sequence of 3 integers */
    /* Check that the parse tree we got was the correct one */
    if (!spki_params_node_ptr || !spki_params_node_ptr->Child || spki_params_node_ptr->Count != 3)
    {
        return CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT;
    }
    for(i=0;i<(uint32)spki_params_node_ptr->Count;i++)
    {
        if (!spki_params_node_ptr->Child[i].IsLeaf || 
            spki_params_node_ptr->Child[i].X.Input.Tag[0] != BER_INTEGER)
        {
            return CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT;
        }
    }

    if (!BER_ExpandItem(SpkiPtr->subjectPublicKey.Data, SpkiPtr->subjectPublicKey.Length, &spki_key_item))
    {
        return CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT;
    }


    
    /* Check that it is an integer */
    if (spki_key_item.Tag[0] != BER_INTEGER)
    {
        return CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT;
    }

    key_node_ptr = DER_AllocateNode(TLS_MemoryPool, &berder_funcs_wrapper, 4);
    if (!key_node_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    
    key_node_ptr->Tag = spki_params_node_ptr->Tag;
    key_node_ptr->OriginalTag = spki_params_node_ptr->OriginalTag;
    key_node_ptr->Count = 4;
    key_node_ptr->Type = spki_params_node_ptr->Type;

    /* Assign the children correctly: first 3 from spki_params_node_ptr and last from spki_key_item*/
    for(i=0;i<(uint32)spki_params_node_ptr->Count;i++)
    {
        key_node_ptr->Child[i] = spki_params_node_ptr->Child[i];
    }
    key_node_ptr->Child[3].IsLeaf = CSSM_TRUE;
    key_node_ptr->Child[3].X.Node = 0;
    key_node_ptr->Child[3].X.Input = spki_key_item;


    *key_ptr = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func,
        1, sizeof(CSSM_DATA), poolptr);
    if (!*key_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }

    (*key_ptr)->Data = DER_Generate(TLS_MemoryPool, &berder_funcs_wrapper, key_node_ptr);

    if (!(*key_ptr)->Data)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    (*key_ptr)->Length = BER_SizeofObject((*key_ptr)->Data);

    return CSSM_OK;
}
