/*-----------------------------------------------------------------------
 * File: CRTGETFI.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 

#include "maf_include.h"
    /* Windows & CSSM includes */
#ifndef VMS
#pragma warning (disable:4201 4214 4115)
#include <stdio.h>
#pragma warning (default:4201 4214 4115)
#else
#include <stdio.h>
#endif

#include "cssm.h"
#include "cssmspi.h"
#include "guids.h"
#include "cssmport.h"
#include "cssmutil.h"

    /* Certificate-specific includes */
#include "x509defs.h"
#include "x_fndefs.h"

#include "oidsx509.h"
#include "oidsalg.h"
#include "x509hlpr.h"
#include "clutil.h"
#include "x_qtable.h"
#include "internal.h"

#include "api_shim.h"
#include "maf_include.h"

#include "clberder.h"

#include <cssmMemTrack.h>

#define X509_FORMAT_VERSION 2

extern MEMORY_POOL_FUNCS memory_pool_funcs;
extern const CSSM_API_MEMORY_FUNCS Addin_APIMemFuncs;



/*-----------------------------------------------------------------------------
 * Name: CL_CertGetFirstFieldValue
 *
 * Description:
 * This function returns the value of the designated certificate field.
 * If there is more than one match for the requested field, a results handle
 * will be assigned which the application can use to obtain the other matches.
 *
 * Parameters: 
 * Cert (input)           : The certificate from which to obtain the field value
 * CertField (input)      : An object identifier which identifies 
 *                          the field value to be extracted from the Cert.
 * ResultsHandle          : An output handle which the application can use to obtain additional 
 *                          matched elements
 * NumberOfMatchedCerts   : An output which indicates the number of matching elements in the Cert
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSMCLI CL_CertGetFirstFieldValue (CSSM_CL_HANDLE CLHandle,
                                     const CSSM_DATA *Cert, 
                                     const CSSM_OID *CertField,
                                     CSSM_HANDLE_PTR ResultsHandle,
                                     uint32 *NumberOfMatchedFields,
                                     CSSM_DATA_PTR *result_val)
{
    /* Initializations */
    CSSM_RETURN                 ret = CSSM_OK;
    CSSM_X509_SIGNED_CERTIFICATE_PTR cert_ptr = NULL;
    CSSM_DATA_PTR               results_ptr = NULL, 
                                datap = NULL;

    MEMORY_POOL pool;

    CSSM_CLUTIL_CONTEXT *contextPtr;
    
    ADDIN_ATTACH_DATA *pAttachData = NULL;

    /* Check to see that we know our execution environment */
    pAttachData = (ADDIN_ATTACH_DATA *)Addin_GetAttachData();
    if ( pAttachData == NULL )
    {
      return CSSMERR_CL_INTERNAL_ERROR;
  
    } 

    contextPtr = (CSSM_CLUTIL_CONTEXT *)pAttachData;

    if (port_IsBadWritePtr (result_val, sizeof(CSSM_DATA_PTR)))
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    *result_val = NULL;
    /* Check inputs */
    /* CLHandle is cheked by CSSM */
    /* Check that Cert is a valid data pointer */
    if ( (Cert == NULL) || 
         port_IsBadReadPtr (Cert, sizeof(CSSM_DATA)) ) 
    {
        ret = CSSMERR_CL_INVALID_CERT_POINTER;
        return ret;
    }
    if ( port_IsBadReadPtr (Cert->Data, Cert->Length) ||
        !Cert->Data || 
        !Cert->Length) 
    {
        ret = CSSMERR_CL_INVALID_CERT_POINTER;
        return ret;
    }
    if ( port_IsBadReadPtr (CertField, sizeof(CSSM_OID)) || !CertField )  
    {
        ret = CSSMERR_CL_INVALID_INPUT_POINTER;
        return ret;
    }
    if ( port_IsBadReadPtr (CertField->Data, CertField->Length) || 
        !CertField->Data || 
        !CertField->Length)  
    {
        ret = CSSMERR_CL_INVALID_INPUT_POINTER;
        return ret;
    }
    if ( port_IsBadWritePtr (ResultsHandle, sizeof(CSSM_HANDLE)) ||
         port_IsBadWritePtr (NumberOfMatchedFields, sizeof(uint32)))
    {
        ret = CSSMERR_CL_INVALID_OUTPUT_POINTER;
        return ret;
    }

    /* Setup the Memory pool */
    ret = (*memory_pool_funcs.init)((CSSM_MEMORY_FUNCS_PTR)&Addin_APIMemFuncs, NULL, &pool);
    if (CSSM_OK != ret)
    {
        return ret;
    }
  
    /* Unpack the Cert */
    ret = MP_DerDecodeSignedCertificate(&pool, Cert, &cert_ptr);

    if(ret != CSSM_OK)
    {
        if (CSSMERR_CL_MEMORY_ERROR != ret)
        {
            ret = CSSMERR_CL_UNKNOWN_FORMAT;
        }
        (*memory_pool_funcs.recycle)(&pool);
        return ret;
    }
    

    /* Check that Cert is a CSSM certificate */
    if (CSSM_CLUTIL_X509_IsBadCertPtr(cert_ptr))
    {
        (*memory_pool_funcs.recycle)(&pool);
        return  CSSMERR_CL_UNKNOWN_FORMAT;
        
    }

    /* Get the field value(s) */
    
    if ((ret = cl_CertGetField(CLHandle, &pool, cert_ptr, *CertField, &results_ptr, NumberOfMatchedFields)) != CSSM_OK )
    {
        (*memory_pool_funcs.recycle)(&pool);
        return ret;
    }
    

    /* Update the query table */
    if ( *NumberOfMatchedFields > 1 )
    {
        /* Note: this guarantees that the OID is not one of the following:
            INT_IssuerNameCStruct
            INT_SubjectNameCStruct
            INT_SubjectPublicKeyCStruct
            INT_SignatureCStruct
            INT_IssuerNameLDAP
            INT_SubjectNameLDAP
            INT_SignedCertificateCStruct
            INT_CertificateCStruct
            INT_CertificateExtensionsCStruct
        */    
        ret = cl_QT_AddResults(contextPtr, results_ptr, *NumberOfMatchedFields, ResultsHandle);
        if(CSSM_OK != ret)
        {
            (*memory_pool_funcs.recycle)(&pool);
            return ret;
        }
    
        /* Return a the first array element */
        ret =  cl_QT_GetNextResult(contextPtr, CLHandle, *ResultsHandle, &datap);

        if(CSSM_OK != ret)
        {
            (*memory_pool_funcs.recycle)(&pool);
            return ret;
        }
        else
        {
            *result_val = datap;
            (*memory_pool_funcs.recycle)(&pool);
            return CSSM_OK;
        }
    }

    /* *NumberOfMatchedFields = 0 with no error or =1 */
    *ResultsHandle = 0;
    switch( x509_cl_CertOidToInt(*CertField))
    {
        case INT_IssuerNameCStruct:
        case INT_SubjectNameCStruct:
        case INT_SubjectPublicKeyCStruct:
        case INT_SignatureCStruct:
        case INT_IssuerNameLDAP:
        case INT_SubjectNameLDAP:
        case INT_SignedCertificateCStruct:
        case INT_CertificateCStruct:
        case INT_CertificateExtensionsCStruct:
            /* Data alloced from Heap memory and not from mempool */
            /* Just recycle mempool, assign result and return */
            *result_val = results_ptr;
            break;
            
        default:
            /* Any other struct is a flat struct and needs to be copied into
               application memory before returning */
            if (results_ptr)
            {
                *result_val = App_Calloc(CLHandle, 1, sizeof(CSSM_DATA));
                if (!(*result_val))
                {       
                   (*memory_pool_funcs.recycle)(&pool);
                   return CSSMERR_CL_MEMORY_ERROR;
                }
                (*result_val)->Length = results_ptr->Length;
                if ((*result_val)->Length)
                {
                    (*result_val)->Data = App_Malloc(CLHandle, (*result_val)->Length);
                    if (!(*result_val)->Data)
                    {
                       (*memory_pool_funcs.recycle)(&pool);
                       return CSSMERR_CL_MEMORY_ERROR;
                    }

                    port_memcpy((*result_val)->Data, results_ptr->Data, (*result_val)->Length);   
                }
            }
            break;
    }
    (*memory_pool_funcs.recycle)(&pool);
    return CSSM_OK;
}

