/*-----------------------------------------------------------------------
 * File: X_GETFLD.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
#include "maf_include.h"
/* CSSM includes */
#include "cssm.h"
#include "guids.h"
#include "cssmport.h"
#include "cssmutil.h"

/* Certificate-specific includes */
#include "x509defs.h"

#include "oidsx509.h"
#include "oidsalg.h"
#include "x_fndefs.h"
#include "x509hlpr.h"
#include "clutil.h"
#include "internal.h"

#include "api_shim.h"
#include "maf_util.h"

#include "clberder.h"
extern MEMORY_POOL_FUNCS memory_pool_funcs;

#include <cssmMemTrack.h>



/*                      ***MEMPOOL ALERT***
    All fields except the following are allocated from the mempool and will 
    need to be copied into regular memory before returning to the Application 
    (in the CSSMCLI function). 
    The following need not be copied and can directly be returned:
        INT_IssuerNameCStruct
        INT_SubjectNameCStruct
        INT_SubjectPublicKeyCStruct
        INT_SignatureCStruct
        INT_IssuerNameLDAP
        INT_SubjectNameLDAP
        INT_SignedCertificateCStruct
        INT_CertificateCStruct
        INT_CertificateExtensionsCStruct
*/


/*-----------------------------------------------------------------------------
 * Name: cl_CertGetField
 *
 * Description:
 * This function gets the value or values of the requested certificate OID.
 * It will allocate any memory necessary to hold the output values using
 * the application's memory allocation routines.
 *
 * Parameters: 
 * Cert  (input)   : The certificate whose field(s) will be returned
 * Oid   (input)   : The OID to be retrieved.
 * Results (output): An array of the retrieved field values.
 * NumberOfMatchedFields (output) : The length of the Results array
 *
 * Return value:
 * A CSSM_RETURN indicator of whether or not the field was retrieved.
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_CertGetField (CSSM_CL_HANDLE CLHandle, 
                             MEMORY_POOL_PTR poolptr,
                             CSSM_X509_SIGNED_CERTIFICATE_PTR Cert, 
                             CSSM_OID Oid,
                             CSSM_DATA_PTR *Results,
                             uint32 *NumberOfMatchedFields)
{
    /* Initializations */
    CSSM_DATA_PTR return_data_ptr = NULL;
    CSSM_DATA temp_data = {0, NULL};

    uint32 i=0;
    CSSM_RETURN ret = CSSM_OK;
    CSSM_CLUTIL_CONTEXT *contextPtr;
    
    ADDIN_ATTACH_DATA *pAttachData = NULL;

    /* Check to see that we know our execution environment */
    pAttachData = (ADDIN_ATTACH_DATA *)Addin_GetAttachData();
    if ( pAttachData == NULL )
    {
      return CSSMERR_CL_INTERNAL_ERROR;
  
    } 

    contextPtr = (CSSM_CLUTIL_CONTEXT *)pAttachData;

    *Results = NULL;
    *NumberOfMatchedFields = 0;
    /* Check that it is a CSSM certificate field */
    if (!IsX509V3CertOid(Oid))
    {
        return CSSMERR_CL_UNKNOWN_TAG;       
    }

    switch ( x509_cl_CertOidToInt(Oid) ) {
    case INT_SignedCertificate :
        ret = MP_DerEncodeSignedCertificate(poolptr, Cert, &return_data_ptr);
        if (ret != CSSM_OK)
        {
           return ret;
        }
        if (!return_data_ptr->Data)
        {
            *Results = NULL;
            *NumberOfMatchedFields = 0;
            return CSSM_OK;
        }
        *Results = return_data_ptr;
        *NumberOfMatchedFields = 1;
        return CSSM_OK;

    case INT_Certificate :
        ret = MP_DerEncodeTbsCertificate(poolptr, &Cert->certificate, &return_data_ptr );
        if (ret != CSSM_OK )
        {
            return ret;
        }
        if (!return_data_ptr->Data)
        {
            *Results = NULL;
            *NumberOfMatchedFields = 0;
            return CSSM_OK;
        }

        *Results = return_data_ptr;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;

    case INT_Version :
        /* Allocate space */
        return_data_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_DATA), poolptr);
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        return_data_ptr->Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
            Cert->certificate.version.Length, poolptr);

        if (!return_data_ptr->Data)
        {
            return CSSMERR_CL_MEMORY_ERROR;
        }

        /* Copy the data */
        return_data_ptr->Length = Cert->certificate.version.Length;
        port_memcpy( return_data_ptr->Data, Cert->certificate.version.Data, return_data_ptr->Length );

        *Results = return_data_ptr;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;


    case INT_SerialNumber :
        /* Allocate space */
        return_data_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_DATA), poolptr);
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        return_data_ptr->Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
                                                    Cert->certificate.serialNumber.Length, 
                                                    poolptr);
        
        if (!return_data_ptr->Data)
        {
            return CSSMERR_CL_MEMORY_ERROR;
        }

        /* Copy the data */
        return_data_ptr->Length = Cert->certificate.serialNumber.Length;
        port_memcpy( return_data_ptr->Data, Cert->certificate.serialNumber.Data, return_data_ptr->Length );

        *Results = return_data_ptr;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;


    case INT_IssuerName :
        /* If this field is not available, return 1 match, NULL data */
        if (!Cert->certificate.issuer.numberOfRDNs || !Cert->certificate.issuer.RelativeDistinguishedName)
        {
            return_data_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_DATA), poolptr);
            if (!return_data_ptr)
            {
                return CSSMERR_CL_MEMORY_ERROR;
               
            }
            return_data_ptr->Length = 0;
            return_data_ptr->Data   = NULL;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 1;
            return CSSM_OK;
        }
        

        ret = MP_DerEncodeName(poolptr, &Cert->certificate.issuer, Results);
        *NumberOfMatchedFields = 1;

        return CSSM_OK;

    case INT_ValidityNotBefore :
        /* Allocate space */
        return_data_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_DATA), poolptr);
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        /* If this field is not available, return 1 match, NULL data */
        if (!Cert->certificate.validity.notBefore.time.Data)
        {
            return_data_ptr->Length = 0;
            return_data_ptr->Data   = NULL;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 1;
            return CSSM_OK;
        }

        /* Allocate space */
        return_data_ptr->Length = strlen((char *)Cert->certificate.validity.notBefore.time.Data) +1;
        return_data_ptr->Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
                                                    return_data_ptr->Length, 
                                                    poolptr);
        
        if (!return_data_ptr->Data)
        {
            
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        /* Copy the data */
        port_memcpy( return_data_ptr->Data, Cert->certificate.validity.notBefore.time.Data, return_data_ptr->Length );

        *Results = return_data_ptr;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;


    case INT_ValidityNotAfter :
        /* Allocate space */
        return_data_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_DATA), poolptr);
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
        }

        /* If this field is not available, return 1 match, NULL data */
        if (!Cert->certificate.validity.notAfter.time.Data)
        {
            return_data_ptr->Length = 0;
            return_data_ptr->Data   = NULL;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 1;
            return CSSM_OK;
        }

        /* Allocate space */
        return_data_ptr->Length = strlen((char *)Cert->certificate.validity.notAfter.time.Data) +1;
        return_data_ptr->Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
                                                    return_data_ptr->Length, 
                                                    poolptr);
        if (!return_data_ptr->Data)
        {
            
            return CSSMERR_CL_MEMORY_ERROR;
          
        }

        /* Copy the data */
        port_memcpy( return_data_ptr->Data, Cert->certificate.validity.notAfter.time.Data, return_data_ptr->Length );

        *Results = return_data_ptr;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;

    case INT_SubjectName :
        /* If this field is not available, return no match, CSSM_OK */
        if (!Cert->certificate.subject.numberOfRDNs || !Cert->certificate.subject.RelativeDistinguishedName)
        {
            return_data_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_DATA), poolptr);
            if (!return_data_ptr)
            {
                return CSSMERR_CL_MEMORY_ERROR;
                
            }
            return_data_ptr->Data   = NULL;
            return_data_ptr->Length = 0;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 1;
            return CSSM_OK;
        }

        ret = MP_DerEncodeName(poolptr, &Cert->certificate.subject, Results);
        *NumberOfMatchedFields = 1;

        return ret;

    case INT_CSSMKeyStruct :
        *Results = NULL;
        *NumberOfMatchedFields = 0;
        return CSSM_OK;

    case INT_SubjectPublicKeyAlgorithm :
        /* Allocate space */
        return_data_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_DATA), poolptr);
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
            
        }

        /* If this field is not available, return 1 match, NULL data */
        if (!Cert->certificate.subjectPublicKeyInfo.algorithm.algorithm.Data || 
            !Cert->certificate.subjectPublicKeyInfo.algorithm.algorithm.Length)
        {
            return_data_ptr->Length = 0;
            return_data_ptr->Data   = NULL;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 1;
            return CSSM_OK;
        }

        return_data_ptr->Length = Cert->certificate.subjectPublicKeyInfo.algorithm.algorithm.Length;
        return_data_ptr->Data   = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
                                                    return_data_ptr->Length, 
                                                    poolptr);
        if (!return_data_ptr->Data )
        {
            
            return CSSMERR_CL_MEMORY_ERROR;
           
        }
        port_memcpy(return_data_ptr->Data, Cert->certificate.subjectPublicKeyInfo.algorithm.algorithm.Data, return_data_ptr->Length);

        *Results = return_data_ptr;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;


    case INT_SubjectPublicKeyAlgorithmParameters :
        /* Allocate space */
        return_data_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_DATA), poolptr);
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        /* If this field is not available, return 1 match, NULL data */
        if (!Cert->certificate.subjectPublicKeyInfo.algorithm.parameters.Data || 
            !Cert->certificate.subjectPublicKeyInfo.algorithm.parameters.Length)
        {
            return_data_ptr->Length = 0;
            return_data_ptr->Data   = NULL;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 1;
            return CSSM_OK;
        }

        return_data_ptr->Length = Cert->certificate.subjectPublicKeyInfo.algorithm.parameters.Length;
        return_data_ptr->Data   = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
                                                    return_data_ptr->Length, 
                                                    poolptr);
        if (!return_data_ptr->Data )
        {
            
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
        port_memcpy(return_data_ptr->Data, Cert->certificate.subjectPublicKeyInfo.algorithm.parameters.Data, return_data_ptr->Length);

        *Results = return_data_ptr;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;

    case INT_SubjectPublicKey :
        /* Allocate space */
        return_data_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_DATA), poolptr);
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
            
        }

        /* If this field is not available, return 1 match, NULL data */
        if (!Cert->certificate.subjectPublicKeyInfo.subjectPublicKey.Data ||
            !Cert->certificate.subjectPublicKeyInfo.subjectPublicKey.Length)
        {
            return_data_ptr->Data   = NULL;
            return_data_ptr->Length = 0;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 1;
            return CSSM_OK;
        }
        return_data_ptr->Length = Cert->certificate.subjectPublicKeyInfo.subjectPublicKey.Length;
        return_data_ptr->Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
                                                    return_data_ptr->Length, 
                                                    poolptr);
        if (!return_data_ptr->Data)
        {
            
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        /* Copy the data */
        port_memcpy( return_data_ptr->Data, Cert->certificate.subjectPublicKeyInfo.subjectPublicKey.Data, return_data_ptr->Length );

        *Results = return_data_ptr;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;

    case INT_SubjectPublicKeyCStruct :
        /* Allocate space */
        return_data_ptr = App_Malloc(CLHandle,sizeof(CSSM_DATA));
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        /* If this no subfield in SPKI is available, return 1 match, NULL data */
        if ((!Cert->certificate.subjectPublicKeyInfo.subjectPublicKey.Data ||
            !Cert->certificate.subjectPublicKeyInfo.subjectPublicKey.Length) &&
            (!Cert->certificate.subjectPublicKeyInfo.algorithm.algorithm.Data ||
            !Cert->certificate.subjectPublicKeyInfo.algorithm.algorithm.Length) &&
            (!Cert->certificate.subjectPublicKeyInfo.algorithm.parameters.Data ||
            !Cert->certificate.subjectPublicKeyInfo.algorithm.parameters.Length))
        {
            return_data_ptr->Data   = NULL;
            return_data_ptr->Length = 0;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 1;
            return CSSM_OK;
        }
        return_data_ptr->Length = (uint32)-1; 
        CSSM_CLUTIL_X509_DuplicateSPKI(contextPtr, CLHandle,
                                        &Cert->certificate.subjectPublicKeyInfo,
                                        (CSSM_X509_SUBJECT_PUBLIC_KEY_INFO_PTR *)&(return_data_ptr->Data));
        if (!return_data_ptr->Data)
        {
            App_Free(CLHandle, return_data_ptr);
            return CSSMERR_CL_MEMORY_ERROR;
            
        }

        *Results = return_data_ptr;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;

    case INT_CertificateIssuerUniqueId :
        /* Allocate space */
        return_data_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_DATA), poolptr);
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        return_data_ptr->Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
                                                    Cert->certificate.issuerUniqueIdentifier.Length, 
                                                    poolptr);
        if (!return_data_ptr->Data)
        {

            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        /* Copy the data */
        return_data_ptr->Length = Cert->certificate.issuerUniqueIdentifier.Length;
        port_memcpy( return_data_ptr->Data, Cert->certificate.issuerUniqueIdentifier.Data, return_data_ptr->Length );

        *Results = return_data_ptr;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;


    case INT_CertificateSubjectUniqueId :
        /* Allocate space */
        return_data_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_DATA), poolptr);
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        return_data_ptr->Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
                                                    Cert->certificate.subjectUniqueIdentifier.Length, 
                                                    poolptr);
        if (!return_data_ptr->Data)
        {
            
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        /* Copy the data */
        return_data_ptr->Length = Cert->certificate.subjectUniqueIdentifier.Length;
        port_memcpy( return_data_ptr->Data, Cert->certificate.subjectUniqueIdentifier.Data, return_data_ptr->Length );

        *Results = return_data_ptr;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;

    case INT_CertificateExtensionStruct :
        /* Get all of the certificate extensions */
        /* These should be packed as 1 extension per array element */
        if (!Cert->certificate.extensions.numberOfExtensions)
        {
            *Results = NULL;
            *NumberOfMatchedFields = 0;
            return CSSMERR_CL_NO_FIELD_VALUES;
            
        }

        return_data_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
                                                    Cert->certificate.extensions.numberOfExtensions * sizeof(CSSM_DATA), 
                                                    poolptr);
        if (!return_data_ptr)
        { 
            return CSSMERR_CL_MEMORY_ERROR;
            
        }

        

        for (i=0; i<Cert->certificate.extensions.numberOfExtensions; i++)
        {
            
            ret = MP_DerEncodeExtnValue(poolptr, 
                                        Cert->certificate.extensions.extensions[i],
                                        &temp_data);

            /* Error check */
            if (ret != CSSM_OK)
            {
                return ret;
            }

            /* Update the return data array */
            return_data_ptr[i].Length = temp_data.Length;
            return_data_ptr[i].Data = temp_data.Data;

        }

        *Results = return_data_ptr;
        *NumberOfMatchedFields = Cert->certificate.extensions.numberOfExtensions;

        return CSSM_OK;


    case INT_CertificateNumberOfExtensions :
        /* A certificate wil always have this field, even if it is 0 */
        /* Allocate space */
        return_data_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_DATA), poolptr);
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
            
        }

        return_data_ptr->Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(uint32), poolptr);
        if (!return_data_ptr->Data)
        {
        
            return CSSMERR_CL_MEMORY_ERROR;
            
        }

        /* Copy the data */
        return_data_ptr->Length = sizeof(uint32);
        port_memcpy( return_data_ptr->Data, 
            &Cert->certificate.extensions.numberOfExtensions, 
            sizeof(uint32) );

        *Results = return_data_ptr;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;


    case INT_CertificateExtensionId :
        /* This returns a list of the Extension IDs */
        /* If this field is not available, return no match, CSSM_OK */
        if (!Cert->certificate.extensions.numberOfExtensions)
        {
            *Results = NULL;
            *NumberOfMatchedFields = 0;
            return CSSMERR_CL_NO_FIELD_VALUES;
           
        }

        /* Allocate room for an array of the extension Ids as CSSM_DATA types */
        return_data_ptr = (CSSM_DATA_PTR) CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
                                                              Cert->certificate.extensions.numberOfExtensions*sizeof(CSSM_DATA),
                                                              poolptr); 
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        /* Create an array of the extension Ids */
        for (i=0; i < Cert->certificate.extensions.numberOfExtensions; i++)
        {
            /* Copy the data */
            return_data_ptr[i].Length = Cert->certificate.extensions.extensions[i].extnId.Length; 
            if ( return_data_ptr[i].Length )
            {
                /* Allocate space */
                return_data_ptr[i].Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
                                                              Cert->certificate.extensions.extensions[i].extnId.Length, 
                                                              poolptr);
                    
                if (!return_data_ptr[i].Data)
                {
                   
                    return CSSMERR_CL_MEMORY_ERROR;
                }
                
                port_memcpy( return_data_ptr[i].Data, Cert->certificate.extensions.extensions[i].extnId.Data, return_data_ptr[i].Length );
            }
            else 
                return_data_ptr[i].Data = NULL;
        }

        *Results = return_data_ptr;
        *NumberOfMatchedFields = Cert->certificate.extensions.numberOfExtensions;

        return CSSM_OK;

    case INT_CertificateExtensionCritical :
        /* This returns a list of the Extension Critical fields */
        /* If this field is not available, return no match, CSSM_OK */
        if (!Cert->certificate.extensions.numberOfExtensions)
        {
            *Results = NULL;
            *NumberOfMatchedFields = 0;
            return CSSMERR_CL_NO_FIELD_VALUES;
           
        }

        return_data_ptr = (CSSM_DATA_PTR) CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
                                                              Cert->certificate.extensions.numberOfExtensions*sizeof(CSSM_DATA),
                                                              poolptr); 
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        /* Create an array of the extension Ids */
        for (i=0; i < Cert->certificate.extensions.numberOfExtensions; i++)
        {
            /* Allocate space */
            return_data_ptr[i].Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(uint32), poolptr);
            if (!return_data_ptr[i].Data)
            {
                
                return CSSMERR_CL_MEMORY_ERROR;
            }

            /* Copy the data */
            port_memcpy( return_data_ptr[i].Data, &Cert->certificate.extensions.extensions[i].critical, sizeof(uint32) );
            return_data_ptr[i].Length = sizeof(uint32); 
        }

        *Results = return_data_ptr;
        *NumberOfMatchedFields = Cert->certificate.extensions.numberOfExtensions;

        return CSSM_OK;

    case INT_CertificateExtensionType :
        /* This returns a list of the Extension Type fields */
        /* If this field is not available, return no match, CSSM_OK */
        if (!Cert->certificate.extensions.numberOfExtensions)
        {
            *Results = NULL;
            *NumberOfMatchedFields = 0;
            return CSSMERR_CL_NO_FIELD_VALUES;
           
        }

        /* Allocate room for an array of the extension Types as CSSM_DATA types */
        return_data_ptr = (CSSM_DATA_PTR) CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
                                                              Cert->certificate.extensions.numberOfExtensions*sizeof(CSSM_DATA),
                                                              poolptr); 
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
        }

        /* Create an array of the extension Ids */
        for (i=0; i < Cert->certificate.extensions.numberOfExtensions; i++)
        {
            /* Allocate space */
            return_data_ptr[i].Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_BER_TAG), poolptr);
            if (!return_data_ptr[i].Data)
            {
                return CSSMERR_CL_MEMORY_ERROR;
            }

            /* Copy the data */
            port_memcpy( return_data_ptr[i].Data, 
                &Cert->certificate.extensions.extensions[i].value.tagAndValue->type, 
                sizeof(CSSM_BER_TAG) );
            return_data_ptr[i].Length = sizeof(CSSM_BER_TAG); 
        }

        *Results = return_data_ptr;
        *NumberOfMatchedFields = Cert->certificate.extensions.numberOfExtensions;

        return CSSM_OK;

    case INT_CertificateExtensionValue :
        /* This returns a list of the Extensions */
        /* If this field is not available, return no match, CSSM_OK */
        if (!Cert->certificate.extensions.numberOfExtensions)
        {
            *Results = NULL;
            *NumberOfMatchedFields = 0;
            return CSSMERR_CL_NO_FIELD_VALUES;
            
        }

        /* Allocate room for an array of the extensions as CSSM_DATA types */
        return_data_ptr = (CSSM_DATA_PTR) CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
                                                              Cert->certificate.extensions.numberOfExtensions*sizeof(CSSM_DATA),
                                                              poolptr); 
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        /* Create an array of the extensions */
        for (i=0; i < Cert->certificate.extensions.numberOfExtensions; i++)
        {
            /* Copy the data */
            if (CSSM_CLUTIL_X509_IsEqualOid(&Cert->certificate.extensions.extensions[i].extnId, 
                    &CSSMOID_X509ExtBasicConstraints) == CSSM_TRUE) 
            {
                return_data_ptr[i].Length = sizeof(CSSM_X509EXT_BASICCONSTRAINTS);
                if ( return_data_ptr[i].Length )
                {   
                    /* Allocate space */
                    return_data_ptr[i].Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, return_data_ptr[i].Length, poolptr);
                    
                    if (!return_data_ptr[i].Data)
                    {
                        return  CSSMERR_CL_MEMORY_ERROR;
                    }
                    port_memcpy( return_data_ptr[i].Data, 
                        &Cert->certificate.extensions.extensions[i].value.parsedValue, 
                        return_data_ptr[i].Length );
                }
                else
                    return_data_ptr[i].Data = NULL;
            }
            else
            {

                return_data_ptr[i].Length = 
                    Cert->certificate.extensions.extensions[i].value.tagAndValue->value.Length; 
                
                if ( return_data_ptr[i].Length )
                {
                    /* Allocate space */
                    return_data_ptr[i].Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, return_data_ptr[i].Length, poolptr);
                
                    if (!return_data_ptr[i].Data)
                    {
                        return CSSMERR_CL_MEMORY_ERROR;
                    }

                    port_memcpy( return_data_ptr[i].Data, 
                        Cert->certificate.extensions.extensions[i].value.tagAndValue->value.Data, 
                        return_data_ptr[i].Length );
                }
                else
                    return_data_ptr[i].Data = NULL;
            }

        }

        *Results = return_data_ptr;
        *NumberOfMatchedFields = Cert->certificate.extensions.numberOfExtensions;

        return CSSM_OK;


    case INT_SignatureStruct :
        /* If there is no signature, return NULL */
        if (!Cert->signature.encrypted.Data || !Cert->signature.encrypted.Length)
        {
            return_data_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
                sizeof(CSSM_DATA), poolptr);
            if (!return_data_ptr)
            {
                return CSSMERR_CL_MEMORY_ERROR;
                
            }
            return_data_ptr->Data = NULL;
            return_data_ptr->Length = 0;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 0;
            return CSSM_OK;
        }

        MP_DerEncodeSignature(poolptr, &Cert->signature, Results );
        *NumberOfMatchedFields = 1;

        return CSSM_OK;

    case INT_SignatureAlgorithm :
        /* Allocate space */
        return_data_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_DATA), poolptr);
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
            
        }

        /* If this field is not available, return 1 match, NULL data */
        if (!Cert->signature.algorithmIdentifier.algorithm.Data || 
            !Cert->signature.algorithmIdentifier.algorithm.Length)
        {
            return_data_ptr->Length = 0;
            return_data_ptr->Data   = NULL;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 1;
            return CSSM_OK;
        }

        return_data_ptr->Length = Cert->signature.algorithmIdentifier.algorithm.Length;
        return_data_ptr->Data   = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, return_data_ptr->Length, poolptr);
        if (!return_data_ptr->Data )
        {
            return CSSMERR_CL_MEMORY_ERROR;   
        }
        port_memcpy(return_data_ptr->Data, Cert->signature.algorithmIdentifier.algorithm.Data, return_data_ptr->Length);

        *Results = return_data_ptr;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;


    case INT_SignatureAlgorithmParameters :
        /* Allocate space */
        return_data_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_DATA), poolptr);
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
            
        }

        /* If this field is not available, return 1 match, NULL data */
        if (!Cert->signature.algorithmIdentifier.parameters.Data || 
            !Cert->signature.algorithmIdentifier.parameters.Length)
        {
            return_data_ptr->Length = 0;
            return_data_ptr->Data   = NULL;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 1;
            return CSSM_OK;
        }

        return_data_ptr->Length = Cert->signature.algorithmIdentifier.parameters.Length;
        return_data_ptr->Data   = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, return_data_ptr->Length, poolptr);
        
        if (!return_data_ptr->Data )
        {
        
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
        port_memcpy(return_data_ptr->Data, Cert->signature.algorithmIdentifier.parameters.Data, return_data_ptr->Length);

        *Results = return_data_ptr;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;


    case INT_Signature :
        return_data_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, sizeof(CSSM_DATA), poolptr);
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        /* If there is no signature, return NoMatch, CSSM_OK */
        if (!Cert->signature.encrypted.Data || !Cert->signature.encrypted.Length)
        {
            return_data_ptr->Data = NULL;
            return_data_ptr->Length = 0;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 0;
            return CSSM_OK;
        }

        /* Allocate memory */
        return_data_ptr->Length = Cert->signature.encrypted.Length;
        return_data_ptr->Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, return_data_ptr->Length, poolptr);
        if (!return_data_ptr->Data)
        {
            
            return CSSMERR_CL_MEMORY_ERROR;
          
        }
        port_memcpy(return_data_ptr->Data, Cert->signature.encrypted.Data, return_data_ptr->Length);

        *Results = return_data_ptr;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;

    case INT_SignatureCStruct :
        /* Allocate space */
        return_data_ptr = App_Malloc(CLHandle,sizeof(CSSM_DATA));
        if (!return_data_ptr)
        {
            return CSSMERR_CL_MEMORY_ERROR;
            
        }

        /* If no subfield in signature is available, return NoMatch, NULL data */
        if ((!Cert->signature.encrypted.Data ||
            !Cert->signature.encrypted.Length) &&
            (!Cert->signature.algorithmIdentifier.algorithm.Data ||
            !Cert->signature.algorithmIdentifier.algorithm.Length) &&
            (!Cert->signature.algorithmIdentifier.parameters.Data ||
            !Cert->signature.algorithmIdentifier.parameters.Length))
        {
            return_data_ptr->Data   = NULL;
            return_data_ptr->Length = 0;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 0;
            return CSSM_OK;
        }
        return_data_ptr->Length = (uint32) -1; 
        CSSM_CLUTIL_X509_DuplicateSignature(contextPtr, CLHandle, &Cert->signature,
                                            (CSSM_X509_SIGNATURE_PTR *)&(return_data_ptr->Data));
        if (!return_data_ptr->Data)
        {
            App_Free(CLHandle, return_data_ptr);
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        *Results = return_data_ptr;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;

    case INT_IssuerNameCStruct :
        /* If this field is not available, return no match, CSSM_OK */
        if(!Cert->certificate.issuer.numberOfRDNs || 
           !Cert->certificate.issuer.RelativeDistinguishedName)
        {
            return_data_ptr = App_Malloc(CLHandle,
                                                          sizeof(CSSM_DATA));
            if (!return_data_ptr)
            {
                return CSSMERR_CL_MEMORY_ERROR;
               
            }
            return_data_ptr->Data   = NULL;
            return_data_ptr->Length = 0;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 1;
            return CSSM_OK;
        }

        /* allocate a CSSM_DATA to hold our results */
        *Results = App_Calloc(CLHandle,
                                               1,
                                               sizeof(CSSM_DATA));

        if(!(*Results))
        {
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        ret = CSSM_CLUTIL_X509_DuplicateCNameStruct(contextPtr, CLHandle,
                                                    (CSSM_X509_NAME_PTR)(&Cert->certificate.issuer),
                                                    (CSSM_X509_NAME_PTR *)&((*Results)->Data));
        (*Results)->Length = (uint32)-1;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;


    case INT_SubjectNameCStruct :
        /* If this field is not available, return no match, CSSM_OK */
        if (!Cert->certificate.subject.numberOfRDNs ||
            !Cert->certificate.subject.RelativeDistinguishedName)
        {
            return_data_ptr = App_Malloc(CLHandle,
                                                          sizeof(CSSM_DATA));
            if (!return_data_ptr)
            {
                return CSSMERR_CL_MEMORY_ERROR;
               
            }
            return_data_ptr->Data   = NULL;
            return_data_ptr->Length = 0;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 1;
            return CSSM_OK;
        }

        /* allocate a CSSM_DATA to hold our results */
        *Results = App_Calloc(CLHandle,
                                               1,
                                               sizeof(CSSM_DATA));

        if(!(*Results))
        {
            return CSSMERR_CL_MEMORY_ERROR;
           
        }


        ret = CSSM_CLUTIL_X509_DuplicateCNameStruct(contextPtr, CLHandle,
                                              &Cert->certificate.subject,
                                              (CSSM_X509_NAME_PTR *)&((*Results)->Data));
        (*Results)->Length = (uint32)-1;
        *NumberOfMatchedFields = 1;

        return CSSM_OK;



    case INT_IssuerNameLDAP :
        /* If this field is not available, return no match, CSSM_OK */
        if (!Cert->certificate.issuer.numberOfRDNs || !Cert->certificate.issuer.RelativeDistinguishedName)
        {
            return_data_ptr = App_Malloc(CLHandle,sizeof(CSSM_DATA));
            if (!return_data_ptr)
            {
                return CSSMERR_CL_MEMORY_ERROR;
              
            }
            return_data_ptr->Data   = NULL;
            return_data_ptr->Length = 0;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 1;
            return CSSM_OK;
        }

        if((ret = CSSM_CLUTIL_X509_CNameStructToLDAPString(contextPtr, CLHandle,
                                                     &Cert->certificate.issuer,
                                                     Results)) != CSSM_OK)
        {
        
            
            if (ret != CSSMERR_CL_UNKNOWN_TAG) 
            {
                *NumberOfMatchedFields = 0;
                return ret;
            }

            
            return_data_ptr = App_Malloc(CLHandle,sizeof(CSSM_DATA));
            if (!return_data_ptr)
            {
                return CSSMERR_CL_MEMORY_ERROR;
                
            }
            return_data_ptr->Data   = NULL;
            return_data_ptr->Length = 0;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 1;
            return CSSM_OK;
        }

        *NumberOfMatchedFields = 1;

        return CSSM_OK;


   case INT_SubjectNameLDAP :
        /* If this field is not available, return no match, CSSM_OK */
        if (!Cert->certificate.subject.numberOfRDNs || !Cert->certificate.subject.RelativeDistinguishedName)
        {
            return_data_ptr = App_Malloc(CLHandle,sizeof(CSSM_DATA));
            if (!return_data_ptr)
            {
                return CSSMERR_CL_MEMORY_ERROR;
              
            }
            return_data_ptr->Data   = NULL;
            return_data_ptr->Length = 0;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 1;
            return CSSM_OK;
        }
        if((ret = CSSM_CLUTIL_X509_CNameStructToLDAPString(contextPtr, CLHandle,
                                                     &Cert->certificate.subject,
                                                     Results)) != CSSM_OK)
        {
            
            if (ret != CSSMERR_CL_UNKNOWN_TAG) 
            {
                *NumberOfMatchedFields = 0;
                return ret;
            }

            
            return_data_ptr = App_Malloc(CLHandle,sizeof(CSSM_DATA));
            if (!return_data_ptr)
            {
                return CSSMERR_CL_MEMORY_ERROR;
               
            }
            return_data_ptr->Data   = NULL;
            return_data_ptr->Length = 0;
            *Results = return_data_ptr;
            *NumberOfMatchedFields = 1;
            return CSSM_OK;
        }
        
        *NumberOfMatchedFields = 1;

        return CSSM_OK;

   
   case INT_SignedCertificateCStruct:
        return_data_ptr = get_signedcertificate_field(CLHandle, Cert);
        break;

    case INT_CertificateCStruct:
        return_data_ptr = get_certificate_field(CLHandle, Cert);
        break;

    case INT_CertificateExtensionsCStruct:
        return_data_ptr = get_certificate_extensions_field(CLHandle, Cert);
        break;

    case INT_ExtBasicConstraints :
    case INT_ExtPrivateUseExemptions :
    case INT_CertificateExtensionsStruct :
    case INT_CertificateExtensionCStruct:
    default:
        return CSSMERR_CL_FUNCTION_FAILED;
       
    }

    if (!return_data_ptr)
        return CSSMERR_CL_INTERNAL_ERROR;
    if (!return_data_ptr->Data)
    {
        App_Free(CLHandle, return_data_ptr);
        return_data_ptr = NULL;

        return CSSMERR_CL_INTERNAL_ERROR;
    }

    *Results = return_data_ptr;
    *NumberOfMatchedFields = 1;

    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: cl_CertGetManyFields
 *
 * Description:
 * This function returns the fields of the input certificate
 * which are given by the input array of OIDs.
 *
 * Parameters: 
 * Cert (input)           : The certificate whose fields will be returned.
 * OidsToRetrieve (input) : An array of the OIDs to be retrieved from the input cert
 * NumOidsToRetrieve (input) : The length of the array of OIDs to be retrieved
 * NumberOfFields (output): A pointer to the length of returned array of fields
 *
 * Return value:
 * An array of fields from the certificate.
 * 
 * Error Codes:
 * CSSM_INVALID_ADDIN_HANDLE
 * CSSMERR_CL_INVALID_CERT_POINTER
 * CSSM_CL_INVALID_DATA_POINTER
 * CSSMERR_CL_UNKNOWN_FORMAT
 * CSSMERR_CL_MEMORY_ERROR
 * CSSM_CL_UNSUPPORTED_OPERATION
 * CSSM_CL_GET_FIELDS_FAIL
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_CertGetManyFields (CSSM_CL_HANDLE CLHandle,
                                  MEMORY_POOL_PTR poolptr,
                                  const CSSM_DATA *Cert,
                                  const CSSM_OID **OidsToRetrieve,
                                  uint32 NumOidsToRetrieve,
                                  uint32 *NumberOfFields,
                                  CSSM_FIELD_PTR *fields_ptr)
{
    /* Initializations */
   
    CSSM_FIELD_PTR              temp_field_ptr = NULL;
    CSSM_X509_SIGNED_CERTIFICATE_PTR cert_ptr = NULL;
    CSSM_DATA_PTR               results_ptr = NULL;
    uint32                      num_results = 0;
    uint32                      num_fields = 0;
    uint32                      max_field_array_size = 0;
    uint32                      i=0, j=0;
    CSSM_RETURN                 rRtn = CSSM_OK;
    CSSM_RETURN                 ret = CSSM_OK;
    
    if (!fields_ptr)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    *fields_ptr = NULL;
       
        /* Check that Cert is a valid data pointer */
    if ( (Cert == NULL) || CL_IsBadReadPtr (Cert, sizeof(CSSM_DATA))) 
    {
        return CSSMERR_CL_INVALID_CERT_POINTER;
       
    }
    if ( CL_IsBadReadPtr (Cert->Data, Cert->Length) ||
         !Cert->Data || !Cert->Length) 
    {
       return CSSMERR_CL_INVALID_CERT_POINTER;
        
    }
        /* Check that OidsToRetrieve is a valid pointer */
    if ( CL_IsBadReadPtr (OidsToRetrieve, sizeof(CSSM_OID_PTR)) ) 
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
    }
        /* Check that NumberOfFields is a valid pointer */
    if ( port_IsBadWritePtr (NumberOfFields, sizeof(uint32)) ) 
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;       
    }
       /* Unpack the Cert */
    ret = MP_DerDecodeSignedCertificate(poolptr, Cert, &cert_ptr);

    if (ret != CSSM_OK)
    {
       
        if (ret != CSSMERR_CL_MEMORY_ERROR)
            return CSSMERR_CL_UNKNOWN_FORMAT;
        return ret;
    }
       /* Check that Cert is a CSSM certificate */
    if (CSSM_CLUTIL_X509_IsBadCertPtr(cert_ptr))
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;       
    }

    /* Allocate room for the fields */
    *fields_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
        NumOidsToRetrieve*sizeof(CSSM_FIELD), poolptr);
    
    
    if (!(*fields_ptr))
    {
        return CSSMERR_CL_MEMORY_ERROR;       
    }
    max_field_array_size = NumOidsToRetrieve;
    /* For each possible OID, */
    for(i=0; i<NumOidsToRetrieve; i++)
    {
        /* Get the value or array of values which are part of Cert */
        rRtn = cl_CertGetField(CLHandle, poolptr, cert_ptr, *OidsToRetrieve[i], &results_ptr, &num_results);
            /* If it was not possible to retrieve the field(s), abort */
        if (rRtn != CSSM_OK  && CSSMERR_CL_NO_FIELD_VALUES != rRtn)
        {
            goto CLEANUP_AND_EXIT;
        }

        /* Add the results to the FieldID list */
        if (num_results > 0 )
        {
            /* Allocate space, if necessary */
            while ( num_fields + num_results >= max_field_array_size )
            {
                temp_field_ptr = (*fields_ptr);
                (*fields_ptr) = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func,
                    (max_field_array_size + NumOidsToRetrieve), sizeof(CSSM_FIELD), poolptr);
                
                if (!(*fields_ptr))
                {
                    rRtn = CSSMERR_CL_MEMORY_ERROR;
                    goto CLEANUP_AND_EXIT;
                    
                }   
                port_memcpy((*fields_ptr), temp_field_ptr, num_fields*sizeof(CSSM_FIELD));
                max_field_array_size += NumOidsToRetrieve;
    
    
            }

            /* Add the new value(s) */
            for(j=0; j<num_results; j++)
            {
                (*fields_ptr)[num_fields].FieldOid.Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, OidsToRetrieve[i]->Length, poolptr);
                if (!((*fields_ptr)[num_fields].FieldOid.Data))
                {
                    rRtn = CSSMERR_CL_MEMORY_ERROR;
                    goto CLEANUP_AND_EXIT;
                }
                
                port_memcpy((*fields_ptr)[num_fields].FieldOid.Data, OidsToRetrieve[i]->Data, OidsToRetrieve[i]->Length);
                (*fields_ptr)[num_fields].FieldOid.Length = OidsToRetrieve[i]->Length;
                (*fields_ptr)[num_fields].FieldValue = results_ptr[j];
                switch (x509_cl_CertOidToInt((*fields_ptr)[num_fields].FieldOid))
                {
                    case INT_IssuerNameCStruct:
                    case INT_SubjectNameCStruct:
                    case INT_SubjectPublicKeyCStruct:
                    case INT_SignatureCStruct:
                    case INT_IssuerNameLDAP:
                    case INT_SubjectNameLDAP:
                    case INT_SignedCertificateCStruct:
                    case INT_CertificateCStruct:
                    case INT_CertificateExtensionsCStruct:
                        /* Free data since its not from the mempool */
                        App_Free(CLHandle, results_ptr); /* num_results will be 1 */
                    default:
                        /* Do nothing */
                        ;
                }

                num_fields++;
            }
            
        }
    }

    /* Update the outputs */
    
   
    *NumberOfFields = num_fields;
    
    return ret;

CLEANUP_AND_EXIT:
    for(i=0;i< num_fields; i++)
    {
        switch (x509_cl_CertOidToInt((*fields_ptr)[i].FieldOid))
        {
            case INT_IssuerNameCStruct:
            case INT_SubjectNameCStruct:
            case INT_SubjectPublicKeyCStruct:
            case INT_SignatureCStruct:
            case INT_IssuerNameLDAP:
            case INT_SubjectNameLDAP:
            case INT_SignedCertificateCStruct:
            case INT_CertificateCStruct:
            case INT_CertificateExtensionsCStruct:
                /* Free data since its not from the mempool */
                App_Free(CLHandle, (*fields_ptr)[i].FieldValue.Data);
            default:
                /* Do nothing */
                ;
        }
    }
    return rRtn;
}
                               

