/*-----------------------------------------------------------------------
 * File: ENCRYPT.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "cspinc.h"

typedef struct _alg_requirements
{
    CSSM_ALGORITHMS AlgId;
    CSSM_ENCRYPT_MODE AlgMode;
    CSSM_BOOL bRequiresIv;
    uint32 IVLength;
}
_ALG_REQUIREMENTS;

static _ALG_REQUIREMENTS AlgRequirements[] =
{
    { CSSM_ALGID_DES, CSSM_ALGMODE_ECBPad, CSSM_FALSE, 0 },
    { CSSM_ALGID_DES, CSSM_ALGMODE_CBCPadIV8, CSSM_TRUE, 8 },
};

static const uint32 NumAlgRequirements = sizeof(AlgRequirements) / 
                                            sizeof(_ALG_REQUIREMENTS);

CSSM_RETURN EAYCSP_ValidateAlgorithmParams(
                        CSSM_ALGORITHMS AlgId,
                        CSSM_ENCRYPT_MODE AlgMode,
                        CSSM_DATA * pIV )
{
    CSSM_RETURN rv = CSSMERR_CSP_INVALID_ALGORITHM;
    uint32 i;

    for ( i = 0; i < NumAlgRequirements; i++ )
    {
        if ( ( AlgId == AlgRequirements[i].AlgId ) &&
             ( AlgMode != AlgRequirements[i].AlgMode ) )
        {
            /* Set this error code in anticitation of failure */
            rv = CSSMERR_CSP_INVALID_ATTR_MODE;
        }
        else if ( ( AlgId == AlgRequirements[i].AlgId ) &&
                  ( AlgMode == AlgRequirements[i].AlgMode ) )
        {
            rv = CSSM_OK;
            break;
        }
    }

    ERR( rv );

    if ( rv == CSSM_OK )
    {
        if ( AlgRequirements[i].bRequiresIv )
        {
            if ( pIV == NULL )
            {
                ERR( rv = CSSMERR_CSP_MISSING_ATTR_INIT_VECTOR );
            }
            else
            {
                if ( ( pIV->Length != AlgRequirements[i].IVLength ) ||
                     port_IsBadReadPtr( pIV->Data, pIV->Length ) )
                {
                    ERR( rv = CSSMERR_CSP_INVALID_ATTR_INIT_VECTOR );
                }
            }
        }
    }

    return rv;
}

/* This function assumes EAYCSP_ValidateAlgorithmParams succeeded */
static CSSM_RETURN __ConstructEncryptCipher(
                        CSSM_ALGORITHMS Algorithm,
                        CSSM_ENCRYPT_MODE AlgMode,
                        CSSM_PADDING PadMode,
                        const CSSM_DATA *pKey,
                        const CSSM_DATA *pIV,
                        EVP_CIPHER_CTX *pCipherCtx )
{
    CSSM_RETURN rv = CSSM_OK;

    assert( pCipherCtx );

    /* Make sure the padding mode is acceptable. SSLeay crypto implements
     * PKCS #5/7 padding as part of the EVP abstraction. */
    if ( ( PadMode != CSSM_PADDING_PKCS5 ) &&
         ( PadMode != CSSM_PADDING_PKCS7 ) )
    {
        ERR( rv = CSSMERR_CSP_INVALID_ATTR_PADDING );
    }
    else
    {
        memset( pCipherCtx, 0, sizeof(*pCipherCtx) );
        switch ( Algorithm )
        {
        case CSSM_ALGID_DES:
            {
                switch ( AlgMode )
                {
                case CSSM_ALGMODE_ECBPad:
                    {
                        EVP_EncryptInit( pCipherCtx, EVP_des_ecb(), pKey->Data, NULL );
                        break;
                    }
                case CSSM_ALGMODE_CBCPadIV8:
                    {
                        EVP_EncryptInit( pCipherCtx, EVP_des_cbc(), pKey->Data, pIV->Data );
                        break;
                    }
                default:
                    {
                        ERR( rv = CSSMERR_CSP_INVALID_ATTR_MODE );
                    }
                }
                break;
            }
        default:
            {
                ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
            }
        } /* switch Algorithm */
    } /* Correct padding mode */

    return rv;
}

static CSSM_RETURN __EncryptDataInit(
                        CSM_SESSION *pSession,
                        CSSM_PRIVILEGE Privilege,
                        uint32 RequiredKeyUsage )
{
    CSSM_RETURN rv = CSSM_OK;
    EVP_CIPHER_CTX *pState = NULL;
    CSSM_DATA KeyData = { 0, NULL };
    CSSM_ALGORITHMS AlgId = CSSM_ALGID_NONE;
    CSSM_ENCRYPT_MODE AlgMode = CSSM_ALGMODE_NONE;
    CSSM_PADDING Padding = CSSM_PADDING_NONE;
    CSSM_DATA *pIV = NULL;
    const CSSM_CONTEXT_ATTRIBUTE *pIvAttr = NULL;
    const CSSM_CONTEXT_ATTRIBUTE *pModeAttr = NULL;
    const CSSM_CONTEXT_ATTRIBUTE *pPaddingAttr = NULL;

    /* Allocate a place to store the encryption state */
    if ( ( rv = CSM_AllocateSessionState(
                                    pSession,
                                    sizeof(EVP_CIPHER_CTX) ) ) == CSSM_OK )
    {
        pState = (EVP_CIPHER_CTX*)(pSession->pCryptoState);

        /* Fetch and validate the algorithm attributes */
        AlgId = pSession->pContext->AlgorithmType;
        pModeAttr = CSM_GetContextAttribute( pSession->pContext,
                                             CSSM_ATTRIBUTE_MODE );
        if ( pModeAttr != NULL )
        {
            AlgMode = pModeAttr->Attribute.Uint32;
        }
        pPaddingAttr = CSM_GetContextAttribute( pSession->pContext,
                                                CSSM_ATTRIBUTE_PADDING );
        if ( pPaddingAttr != NULL )
        {
            Padding = pPaddingAttr->Attribute.Uint32;
        }
        pIvAttr = CSM_GetContextAttribute( pSession->pContext,
                                           CSSM_ATTRIBUTE_INIT_VECTOR );
        if ( pIvAttr != NULL )
        {
            pIV = pIvAttr->Attribute.Data;
        }
        /* Fetch the signing key */
        if ( ( rv = EAYCSP_FetchSessionKey( pSession,
                                            RequiredKeyUsage,
                                            CSSM_TRUE,
                                            &KeyData ) ) == CSSM_OK )
        {
            if ( ( rv = EAYCSP_ValidateAlgorithmParams( AlgId,
                                                        AlgMode,
                                                        pIV ) ) == CSSM_OK )
            {
                /* Create the encryption object */
                rv = __ConstructEncryptCipher( AlgId,
                                               AlgMode,
                                               Padding,
                                               &KeyData,
                                               pIV,
                                               pState );

                /* Free the temporary key buffer */
                memset( KeyData.Data, 0, KeyData.Length );
                Addin_free( KeyData.Data, NULL );
            } /* Fetch and validate algorithm parameters */
        } /* Fetch key */
    } /* Allocate session state */

    return rv;
}

/* Pre-declared for use in single-stage encrypt case */
static CSSM_RETURN __EncryptDataFinal(
                        CSM_SESSION *pSession,
                        CSSM_DATA *RemData );

static CSSM_RETURN __EncryptDataUpdate(
                        CSM_SESSION *pSession,
                        const CSSM_DATA *ClearBufs,
                        uint32 ClearBufCount,
                        CSSM_DATA *CipherBufs,
                        uint32 CipherBufCount,
                        uint32 *bytesEncrypted,
                        CSSM_DATA *RemData,
                        CSSM_BOOL bStaged )
{
    CSSM_RETURN rv = CSSM_OK;
    EVP_CIPHER_CTX *pState = NULL;
    CSM_VECTOR InputVector;
    CSM_VECTOR OutputVector;
    CSSM_DATA OutputData = { 0, NULL };
    CSSM_DATA InputData = { 0, NULL };
    uint32 OutputLength = 0;


    /* Fetch the cipher object from the session */
    assert( pSession->pCryptoState &&
            ( pSession->uCryptoStateSize == sizeof(EVP_CIPHER_CTX) ) );
    pState = (EVP_CIPHER_CTX*)pSession->pCryptoState;
    assert( pState );

    /* Create the input and output vectors */
    if ( ( rv = CSM_CreateVector( pSession->hCsp,
                                  &InputVector,
                                  CSSM_TRUE,
                                  ClearBufs,
                                  ClearBufCount,
                                  NULL ) ) == CSSM_OK )
    {
        if ( ( rv = CSM_CreateVector( pSession->hCsp,
                                      &OutputVector,
                                      CSSM_FALSE,
                                      CipherBufs,
                                      CipherBufCount,
                                      RemData ) ) == CSSM_OK )
        {
            /* Calculate the output length and reserve the output space */
            CSM_GetVectorLength( &InputVector, &OutputLength );
            OutputLength = ( ( ( OutputLength + pState->buf_len ) /
                               EVP_CIPHER_CTX_block_size( pState )
                             ) * EVP_CIPHER_CTX_block_size( pState )
                           ) +
                           ( ( bStaged )
                                ? 0 
                                : EVP_CIPHER_CTX_block_size( pState ) );
            if ( ( rv = CSM_GuaranteeVectorSpace(
                                            &OutputVector,
                                            OutputLength ) ) == CSSM_OK )
            {
                /* Allocate a temporary buffer to hold the output */
                OutputData.Data = Addin_malloc( OutputLength, NULL );
                if ( OutputData.Data == NULL )
                {
                    ERR( rv = CSSMERR_CSP_MEMORY_ERROR );
                }
                else
                {
                    OutputData.Length = OutputLength;

                    /* Fetch the data to encrypt */
                    rv = CSM_GetBlockLengthVectorBuffer(
                                                &InputVector,
                                                1,
                                                &InputData );
                    if ( ( rv == CSSM_OK ) ||
                         ( rv == CSMR_NO_MORE_DATA ) )
                    {
                        /* Encrypt the data */
                        EVP_EncryptUpdate( pState,
                                           OutputData.Data,
                                           (int*)&OutputData.Length,
                                           InputData.Data,
                                           InputData.Length );

                        /* Add the output data to the vector */
                        CSM_PopulateVector( &OutputVector, &OutputData );

                        /* Is this the a single staged operation that needs
                         * to be finalized?
                         */
                        if ( !bStaged )
                        {
                            /* Reset the buffer size to its true length */
                            OutputData.Length = OutputLength;

                            /* Finalize the encryption */
                            if ( ( rv = __EncryptDataFinal( 
                                                pSession,
                                                &OutputData ) ) == CSSM_OK )
                            {
                                CSM_PopulateVector( &OutputVector,
                                                    &OutputData );
                            }
                        }

                        if ( rv == CSSM_OK )
                        {
                            *bytesEncrypted = OutputLength;
                        }
                    } /* Fetch contiguous buffer */

                    /* Release the temporary buffer */
                    memset( OutputData.Data, 0, OutputLength );
                    Addin_free( OutputData.Data, NULL );
                } /* Allocate temp buffer */
            } /* Allocate output space */

            /* Shutdown the output vector */
            CSM_DeleteVector( &OutputVector, ( rv != CSSM_OK ) );
        } /* Create output vector */

        /* Adjust error code to account for zero length input */
        if ( rv == CSMR_NO_MORE_DATA )
        {
            rv = CSSM_OK;
            *bytesEncrypted = 0;
        }

        /* Shutdown the input vector */
        CSM_DeleteVector( &InputVector, CSSM_FALSE );
    } /* Create input vector */

    return rv;
}

static CSSM_RETURN __EncryptDataFinal(
                        CSM_SESSION *pSession,
                        CSSM_DATA *RemData )
{
    CSSM_RETURN rv = CSSM_OK;
    uint32 OutputLength = 0;
    EVP_CIPHER_CTX *pState = NULL;
    CSSM_BOOL bAllocatedRemData = CSSM_FALSE;

    /* Fetch the cipher object from the session */
    assert( pSession->pCryptoState &&
            ( pSession->uCryptoStateSize == sizeof(EVP_CIPHER_CTX) ) );
    pState = (EVP_CIPHER_CTX*)pSession->pCryptoState;
    assert( pState );

    /* Get the length of the output */
    OutputLength = EVP_CIPHER_CTX_block_size( pState );

    /* Make sure there is enough space to return the final block */
    if ( ( rv = MAF_CreateOutputBuffer( pSession->hCsp,
                                        OutputLength, 
                                        RemData,
                                        &bAllocatedRemData ) ) == CSSM_OK )
    {
        /* Complete the encryption */
        EVP_EncryptFinal( pState,
                          RemData->Data,
                          (int*)&RemData->Length );
    }

    return rv;
}

void ADDIN_CSM_EncryptDataError( CSM_SESSION *pSession )
{
    EVP_CIPHER_CTX *pState;

    /* Just make sure the state is correct */
    assert( pSession->pCryptoState &&
            ( pSession->uCryptoStateSize == sizeof(EVP_CIPHER_CTX) ) );

    /* Clear the context */
    pState = (EVP_CIPHER_CTX*)pSession->pCryptoState;
    memset( pState, 0, sizeof(*pState) );
}

CSSM_RETURN CSSMAPI ADDIN_CSM_EncryptData( CSM_SESSION *pSession,
                                  const CSSM_DATA *ClearBufs,
                                  uint32 ClearBufCount,
                                  CSSM_DATA *CipherBufs,
                                  uint32 CipherBufCount,
                                  uint32 *bytesEncrypted,
                                  CSSM_DATA *RemData,
                                  CSSM_PRIVILEGE Privilege )
{
    CSSM_RETURN rv = CSSM_OK;

    /* Start the encryption operation */
    if ( ( rv = __EncryptDataInit( pSession,
                                   Privilege,
                                   CSSM_KEYUSE_ENCRYPT ) ) == CSSM_OK )
    {
        /* Perform the encryption operation. Since the "staged" parameter
         * to the service function is true, it will call the "final"
         * function internally to handle output buffers.
         */
        rv = __EncryptDataUpdate( pSession,
                                  ClearBufs,
                                  ClearBufCount,
                                  CipherBufs,
                                  CipherBufCount,
                                  bytesEncrypted,
                                  RemData,
                                  CSSM_FALSE );

        /* Make sure the session is cleared of the encryption operation */
        ADDIN_CSM_EncryptDataError( pSession );
    } /* Init */

    return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_EncryptDataInit(
                        CSM_SESSION *pSession,
                        CSSM_PRIVILEGE Privilege )
{
    CSSM_RETURN rv = CSSM_OK;

    /* Initialize the encryption operation */
    rv = __EncryptDataInit( pSession, Privilege, CSSM_KEYUSE_ENCRYPT );


    return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_EncryptDataUpdate(
                        CSM_SESSION *pSession,
                        const CSSM_DATA *ClearBufs,
                        uint32 ClearBufCount,
                        CSSM_DATA *CipherBufs,
                        uint32 CipherBufCount,
                        uint32 *bytesEncrypted )
{
    CSSM_RETURN rv = CSSM_OK;

    /* Feed data buffers into the encryption */
    rv = __EncryptDataUpdate( pSession,
                              ClearBufs,
                              ClearBufCount,
                              CipherBufs,
                              CipherBufCount,
                              bytesEncrypted,
                              NULL,
                              CSSM_TRUE );

    return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_EncryptDataFinal(
                        CSM_SESSION *pSession,
                        CSSM_DATA *RemData )
{
    CSSM_RETURN rv = CSSM_OK;

    /* Finalize the encryption operation */
    rv = __EncryptDataFinal( pSession,
                             RemData );

    /* Clear the encryption state */
    ADDIN_CSM_EncryptDataError( pSession );

    return rv;
}

CSSM_RETURN EAYCSP_EncryptBlock(
                        CSM_SESSION *pSession,
                        uint32 RequiredKeyUsage,
                        const CSSM_DATA *ClearText,
                        CSSM_DATA *CipherText )
{
    CSSM_RETURN rv = CSSM_OK;
    uint32 BytesEncrypted = 0;

    /* Start the decryption operation */
    if ( ( rv = __EncryptDataInit( pSession,
                                   CSSM_USEE_NONE,
                                   RequiredKeyUsage ) ) == CSSM_OK )
    {
        /* Perform the encryption operation. Since the "staged" parameter
         * to the service function is true, it will call the "final"
         * function internally to handle output buffers.
         */
        rv = __EncryptDataUpdate( pSession,
                                  ClearText,
                                  1,
                                  CipherText,
                                  1,
                                  &BytesEncrypted,
                                  NULL,
                                  CSSM_FALSE );

        /* Make sure the session is cleared of the decryption operation */
        ADDIN_CSM_EncryptDataError( pSession );
    } /* Init */

    return rv;
}
