/*-----------------------------------------------------------------------
 * File: KEYS.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "cspinc.h"

typedef struct _key_limits
{
    CSSM_ALGORITHMS AlgId;
    CSSM_RANGE LengthRange;
    uint32 LengthIncrement;
}
_KEY_LIMITS;

static _KEY_LIMITS KeyLimits[] =
{
    { CSSM_ALGID_DES, { 64, 64 }, 64 },
    { CSSM_ALGID_DSA, { 512, 4096 }, 64 },
    { CSSM_ALGID_GenericSecret, { 8, 1024 }, 8 }
};

static const uint32 NumKeyLimits = sizeof(KeyLimits) / sizeof(_KEY_LIMITS);

CSSM_RETURN EAYCSP_CheckKeyLimits(
                        CSSM_ALGORITHMS Algid,
                        uint32 Length )
{
    CSSM_RETURN rv = CSSM_OK;
    uint32 i;

    for ( i = 0; ( i < NumKeyLimits ) && 
                 ( Algid != KeyLimits[i].AlgId ); i++ );
    if ( i == NumKeyLimits )
    {
        ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
    }
    else
    {
        if ( ( Length < KeyLimits[i].LengthRange.Min ) ||
             ( Length > KeyLimits[i].LengthRange.Max ) ||
             ( ( Length % KeyLimits[i].LengthIncrement ) != 0 ) )
        {
            ERR( rv = CSSMERR_CSP_UNSUPPORTED_KEY_SIZE );
        }
    }

    return rv;
}

CSSM_RETURN EAYCSP_KMFormatCtxByAlgId(
                        CSSM_ALGORITHMS AlgId,
                        CSSM_KEYCLASS KeyClass,
                        CSM_KM_FORMAT_CONTEXT *pContext )
{
    CSSM_RETURN rv = CSSM_OK;

    pContext->SigAlg = CSSM_ALGID_NONE;

    switch ( AlgId )
    {
    case CSSM_ALGID_DSA:
    case CSSM_ALGID_SHA1WithDSA:
        {
            pContext->Format = CSSM_KEYBLOB_RAW_FORMAT_FIPS186;
            break;
        }
    case CSSM_ALGID_DES:
    case CSSM_ALGID_3DES_3KEY_EDE:
        {
            pContext->Format = CSSM_KEYBLOB_RAW_FORMAT_OCTET_STRING;
            break;
        }
    default:
        {
            ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
        }
    }

    return rv;
}

CSSM_RETURN EAYCSP_GetEAYKeyType(
                        CSSM_ALGORITHMS AlgId,
                        int *EayKeyType )
{
    CSSM_RETURN rv = CSSM_OK;

    switch ( AlgId )
    {
    case CSSM_ALGID_DSA:
    case CSSM_ALGID_SHA1WithDSA:
        {
            *EayKeyType = EVP_PKEY_DSA;
            break;
        }
    default:
        {
            ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
        }
    }

    return rv;
}

CSSM_RETURN EAYCSP_GetCdsaKeyType(
                        CSSM_ALGORITHMS AlgId,
                        CSSM_ALGORITHMS *CdsaKeyType )
{
    CSSM_RETURN rv = CSSM_OK;

    switch ( AlgId )
    {
    case CSSM_ALGID_DSA:
    case CSSM_ALGID_SHA1WithDSA:
        {
            *CdsaKeyType = CSSM_ALGID_DSA;
            break;
        }
    case CSSM_ALGID_DES:
        {
            *CdsaKeyType = CSSM_ALGID_DES;
            break;
        }
    default:
        {
            ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
        }
    }

    return rv;
}

CSSM_RETURN EAYCSP_NativeToEncodedKey(
                        CSSM_CSP_HANDLE CSPHandle,
                        EVP_PKEY *pNativeKey,
                        CSSM_KEY *pEncodedKey )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_DATA KeyParts[CSM_MAX_KEY_PARTS];
    uint32 NumKeyParts = CSM_MAX_KEY_PARTS;
    uint32 i;
    uint32 TotalLength = 0;
    uint8 *TempBuf = NULL;
    CSSM_BOOL bAllocatedKeyblob = CSSM_FALSE;
    /* EAY crypto types */
    BIGNUM *Numbers[CSM_MAX_KEY_PARTS];
    DSA *pDsaKey = NULL;

    if ( pNativeKey->pkey.dsa != NULL )
    {
        pDsaKey = pNativeKey->pkey.dsa;
        
        /* Calculate the combined length of the key values */
        Numbers[CSM_DSA_PRIME] = pDsaKey->p;
        Numbers[CSM_DSA_SUB_PRIME] = pDsaKey->q;
        Numbers[CSM_DSA_BASE] = pDsaKey->g;
        if ( pEncodedKey->KeyHeader.KeyClass == CSSM_KEYCLASS_PUBLIC_KEY )
        {
            Numbers[CSM_DSA_PUBLIC_VALUE] = pDsaKey->pub_key;
        }
        else
        {
            Numbers[CSM_DSA_PRIVATE_VALUE] = pDsaKey->priv_key;
        }
        for ( i = 0; i < CSM_NUMBER_DSA_PUBLIC_KEY_PARTS; i++ )
        {
            TotalLength += ( BN_num_bits( Numbers[i] ) / 8 ) + 1;
        }
        
        /* Allocate a temporary buffer */
        if ( ( TempBuf = Addin_malloc( TotalLength, NULL ) ) == NULL )
        {
            ERR( rv = CSSMERR_CSP_MEMORY_ERROR );
        }
        else
        {
            KeyParts[CSM_DSA_PRIME].Data = TempBuf;
            KeyParts[CSM_DSA_PRIME].Length = 
                                        BN_bn2bin( 
                                            Numbers[CSM_DSA_PRIME],
                                            KeyParts[CSM_DSA_PRIME].Data );
            TempBuf += KeyParts[CSM_DSA_PRIME].Length;
            KeyParts[CSM_DSA_SUB_PRIME].Data = TempBuf;
            KeyParts[CSM_DSA_SUB_PRIME].Length = 
                                        BN_bn2bin( 
                                            Numbers[CSM_DSA_SUB_PRIME],
                                            KeyParts[CSM_DSA_SUB_PRIME].Data );
            TempBuf += KeyParts[CSM_DSA_SUB_PRIME].Length;
            KeyParts[CSM_DSA_BASE].Data = TempBuf;
            KeyParts[CSM_DSA_BASE].Length = 
                                        BN_bn2bin( 
                                            Numbers[CSM_DSA_BASE],
                                            KeyParts[CSM_DSA_BASE].Data );
            TempBuf += KeyParts[CSM_DSA_BASE].Length;
            KeyParts[CSM_DSA_PUBLIC_VALUE].Data = TempBuf;
            KeyParts[CSM_DSA_PUBLIC_VALUE].Length = 
                                        BN_bn2bin( 
                                            Numbers[CSM_DSA_PUBLIC_VALUE],
                                            KeyParts[CSM_DSA_PUBLIC_VALUE].Data );
            NumKeyParts = CSM_NUMBER_DSA_PUBLIC_KEY_PARTS;

            /* Encode the data */
            pEncodedKey->KeyData.Data = NULL;
            pEncodedKey->KeyData.Length = 0;
            rv = CSM_CreateKeyblob( CSPHandle,
                                    CSSM_ALGID_DSA,
                                    pEncodedKey->KeyHeader.KeyClass,
                                    pEncodedKey->KeyHeader.Format,
                                    KeyParts,
                                    NumKeyParts,
                                    &pEncodedKey->KeyData,
                                    &bAllocatedKeyblob );

            /* Release the temporary buffer */
            memset( KeyParts[CSM_DSA_PRIME].Data, 0, TotalLength );
            Addin_free( KeyParts[CSM_DSA_PRIME].Data, NULL );
        }

    }
    else
    {
        ERR( rv = CSSMERR_CSP_INTERNAL_ERROR );
    }

    return rv;
}

CSSM_RETURN EAYCSP_EncodedToNativeKey(
                        const CSSM_KEY *pEncodedKey,
                        EVP_PKEY **ppNativeKey )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_DATA KeyParts[CSM_MAX_KEY_PARTS];
    uint32 NumKeyParts = CSM_MAX_KEY_PARTS;
    /* EAY crypto types */
    EVP_PKEY *pEvpKey = NULL;
    DSA *pDsaKey = NULL;

    /* Parse the key */
    if ( ( rv = CSM_ParseKeyblob( pEncodedKey->KeyHeader.AlgorithmId,
                                  pEncodedKey->KeyHeader.KeyClass,
                                  pEncodedKey->KeyHeader.Format,
                                  KeyParts,
                                  &NumKeyParts,
                                  &pEncodedKey->KeyData ) ) == CSSM_OK )
    {
        /* Allocate the high level key abstraction */
        if ( ( pEvpKey = EVP_PKEY_new() ) == NULL )
        {
            ERR( rv = CSSMERR_CSP_MEMORY_ERROR );
        }
        else
        {
            switch ( pEncodedKey->KeyHeader.AlgorithmId )
            {
            case CSSM_ALGID_DSA:
                {
                    pEvpKey->type = EVP_PKEY_DSA;

                    /* Allocate the DSA key structure */
                    if ( ( pDsaKey = DSA_new() ) == NULL )
                    {
                        ERR( rv = CSSMERR_CSP_MEMORY_ERROR );
                    }
                    else
                    {
                        pEvpKey->pkey.dsa = pDsaKey;
                        pDsaKey->version = 0;
                        /* Parse fields common to both key types */
                        if ( ( ( pDsaKey->p = BN_bin2bn( 
                                                KeyParts[CSM_DSA_PRIME].Data,
                                                KeyParts[CSM_DSA_PRIME].Length,
                                                pDsaKey->p ) ) == NULL ) || 
                             ( ( pDsaKey->q = BN_bin2bn(
                                                KeyParts[CSM_DSA_SUB_PRIME].Data,
                                                KeyParts[CSM_DSA_SUB_PRIME].Length,
                                                pDsaKey->q ) ) == NULL ) ||
                             ( ( pDsaKey->g = BN_bin2bn(
                                                KeyParts[CSM_DSA_BASE].Data,
                                                KeyParts[CSM_DSA_BASE].Length,
                                                pDsaKey->g ) ) == NULL ) )
                        {
                            ERR( rv = CSSMERR_CSP_INVALID_KEY );
                        }
                        else
                        {
                            /* Parse type specific fields */
                            if ( pEncodedKey->KeyHeader.KeyClass == 
                                                    CSSM_KEYCLASS_PUBLIC_KEY )
                            {
                                if ( ( pDsaKey->pub_key = BN_bin2bn(
                                                KeyParts[CSM_DSA_PUBLIC_VALUE].Data,
                                                KeyParts[CSM_DSA_PUBLIC_VALUE].Length,
                                                pDsaKey->pub_key) ) == NULL )
                                {
                                    ERR( rv = CSSMERR_CSP_INVALID_KEY );
                                }
                            }
                            else
                            {
                                if ( ( pDsaKey->priv_key = BN_bin2bn(
                                                KeyParts[CSM_DSA_PRIVATE_VALUE].Data,
                                                KeyParts[CSM_DSA_PRIVATE_VALUE].Length,
                                                pDsaKey->priv_key) ) == NULL )
                                {
                                    ERR( rv = CSSMERR_CSP_INVALID_KEY );
                                }
                            } /* Convert type specific fields */
                        } /* Parse common fields to both key types */
                    }
                    break;
                } /* case CSSM_ALGID_DSA */
            default:
                {
                    ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
                }
            } /* switch */

            if ( rv != CSSM_OK )
            {
                EVP_PKEY_free( pEvpKey );
                pEvpKey = NULL;
            }
        }
    }

    *ppNativeKey = pEvpKey;

    return rv;
}

CSSM_RETURN EAYCSP_FetchKeyData(
                        CSM_SESSION *pSession,
                        const CSSM_KEY *pSrcKey,
                        const CSSM_ACCESS_CREDENTIALS *pAccessCred,
                        const CSM_KM_FORMAT_CONTEXT *pFormatSpec,
                        CSSM_KEYCLASS ExpectedKeyClass,
                        CSSM_KEY *pDestKey )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_ALGORITHMS AlgId;
    const CSSM_CONTEXT_ATTRIBUTE *pKeyAttr = NULL;
    const CSSM_CONTEXT_ATTRIBUTE *pCredAttr = NULL;
    CSM_KM_KEY_HANDLE hKey = CSSM_INVALID_HANDLE;
    const ADDIN_LOAD_DATA *pLoadData = NULL;
    CSM_KM_FORMAT_CONTEXT KmFormatCtx;

    assert( pSession );
    assert( pDestKey );

    /* If the source key comes from a context, not directly from the
     * caller, then resolve the attributes.
     */
    if ( !pSrcKey )
    {
        /* Fetch the required context attributes */
        pKeyAttr = CSM_GetContextAttribute( pSession->pContext,
                                            CSSM_ATTRIBUTE_KEY );
        pCredAttr = CSM_GetContextAttribute( pSession->pContext,
                                             CSSM_ATTRIBUTE_ACCESS_CREDENTIALS );
        if ( pCredAttr != NULL )
        {
            pAccessCred = pCredAttr->Attribute.AccessCredentials;
        }
        if ( pKeyAttr == NULL )
        {
            ERR( rv = CSSMERR_CSP_MISSING_ATTR_KEY );
        }
        else
        {
            pSrcKey = pKeyAttr->Attribute.Key;
            if ( pSrcKey == NULL )
            {
                ERR( rv = CSSMERR_CSP_INVALID_ATTR_KEY );
            }
            else
            {
                if ( ( rv = CSM_ValidateInputKey( pSrcKey ) ) != CSSM_OK )
                {
                    ERR( rv = CSSMERR_CSP_INVALID_KEY );
                }
            }
        }
    }

    if ( rv == CSSM_OK )
    {
        /* Attempt to resolve the key */
        AlgId = pSession->pContext->AlgorithmType;
        if ( ( pSrcKey->KeyHeader.BlobType == CSSM_KEYBLOB_REFERENCE ) &&
             ( pSrcKey->KeyHeader.Format == CSSM_KEYBLOB_REF_FORMAT_INTEGER ) )
        {
            /* The key is really in the key manager. Do all of the remaining
             * checks on that one.
             */
            if ( ( pSrcKey->KeyData.Length != sizeof(CSM_KM_KEY_HANDLE) ) ||
                 port_IsBadReadPtr( pSrcKey->KeyData.Data,
                                    sizeof(CSM_KM_KEY_HANDLE) ) )
            {
                ERR( rv = CSSMERR_CSP_INVALID_KEY_REFERENCE );
            }
            else
            {
                pLoadData = Addin_GetLoadData();
                assert( pLoadData );
                hKey = *(CSM_KM_KEY_HANDLE*)pSrcKey->KeyData.Data;
                if ( pFormatSpec )
                {
                    KmFormatCtx = *pFormatSpec; 
                }
                else
                {
                    rv = EAYCSP_KMFormatCtxByAlgId( 
                                                AlgId,
                                                ExpectedKeyClass,
                                                &KmFormatCtx );
                }
                if ( rv == CSSM_OK )
                {
                    rv = CSM_KMFindKeyByHandle( pLoadData->hKeyFile,
                                                hKey,
                                                &KmFormatCtx,
                                                pAccessCred,
                                                pDestKey );

                    /* If they application passes in the wrong type of key,
                     * and the passed key requires access credentials (even if
                     * the correct key doesn't), the key manager may return
                     * CSSMERR_CSP_INVALID_ACCESS_CREDENTIALS instead of the
                     * _REAL_ error, CSSMERR_CSP_INVALID_KEY_CLASS.
                     */
                    if ( ( rv == CSSMERR_CSP_INVALID_ACCESS_CREDENTIALS ) &&
                         ( pSrcKey->KeyHeader.KeyClass != ExpectedKeyClass ) )
                    {
                        ERR( rv = CSSMERR_CSP_INVALID_KEY_CLASS );
                    }
                    else if ( rv == KMERR_INVALID_HANDLE )
                    {
                        ERR( rv = CSSMERR_CSP_INVALID_KEY_REFERENCE );
                    }
                    else if ( rv == KMERR_CAN_NOT_INTERPRET_KEY_DATA )
                    {
                        ERR( rv = CSSMERR_CSP_INVALID_KEY );
                    }
                }
            }
        } /* Reference key */
        else if ( pSrcKey->KeyHeader.BlobType == CSSM_KEYBLOB_RAW )
        { /* Raw Key */
            /* Reformat the key blob to match the format request. */
            memcpy( &pDestKey->KeyHeader,
                    &pSrcKey->KeyHeader,
                    sizeof(pSrcKey->KeyHeader) );
            pDestKey->KeyData.Data = Addin_malloc( pSrcKey->KeyData.Length, NULL );
            if ( pDestKey->KeyData.Data != NULL )
            {
                pDestKey->KeyData.Length = pSrcKey->KeyData.Length;
                memcpy( pDestKey->KeyData.Data,
                        pSrcKey->KeyData.Data,
                        pSrcKey->KeyData.Length );

                /* Reformat the key as requested */
                if ( pFormatSpec )
                {
                    rv = CSM_KMReformatKey( MAF_ALLOCATE_LOCAL,
                                            pDestKey,
                                            pFormatSpec );
                }

                if ( rv != CSSM_OK )
                {
                    memset( pDestKey->KeyData.Data,
                            0,
                            pDestKey->KeyData.Length );
                    Addin_free( pDestKey->KeyData.Data, NULL );
                }
            }
            else
            {
                ERR( rv = CSSMERR_CSP_MEMORY_ERROR );
            }
        } /* Raw Key */
        else
        {
            ERR( rv = CSSMERR_CSP_INVALID_KEY_FORMAT );
        }
    }

    return rv;
}

CSSM_RETURN EAYCSP_FetchPrivateKey(
                        CSM_SESSION *pSession,
                        uint32 RequiredKeyUse,
                        CSSM_BOOL bEnforceKeyUse,
                        EVP_PKEY **ppSigKey )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_ALGORITHMS RequiredKeyType;
    CSSM_KEY Key;

    /* Fetch the required context attributes */
    if ( ( rv = EAYCSP_FetchKeyData( pSession,
                                     NULL, NULL,
                                     NULL,
                                     CSSM_KEYCLASS_PRIVATE_KEY,
                                     &Key ) ) == CSSM_OK )
    {
        if ( ( rv = EAYCSP_GetCdsaKeyType( 
                                    pSession->pContext->AlgorithmType,
                                    &RequiredKeyType ) ) == CSSM_OK )
        {
            /* Do algorithm and usage checks */
            if ( Key.KeyHeader.KeyClass != CSSM_KEYCLASS_PRIVATE_KEY )
            {
                ERR( rv = CSSMERR_CSP_INVALID_KEY_CLASS );
            }
            else if ( Key.KeyHeader.AlgorithmId != RequiredKeyType )
            {
                ERR( rv = CSSMERR_CSP_ALGID_MISMATCH );
            }
            else if ( bEnforceKeyUse &
                      !CHECKBITS( Key.KeyHeader.KeyUsage,
                                  RequiredKeyUse ) &&
                      !CHECKBITS( Key.KeyHeader.KeyUsage,
                                  CSSM_KEYUSE_ANY ) )
            {
                ERR( rv = CSSMERR_CSP_KEY_USAGE_INCORRECT );
            }
            else
            {
                rv = EAYCSP_EncodedToNativeKey( &Key,
                                                ppSigKey );
            }
        } /* Determine required key type */

        /* Free the key temporary key data buffer */
        memset( Key.KeyData.Data, 0, Key.KeyData.Length );
        Addin_free( Key.KeyData.Data, NULL );
    } /* Key data fetched */

    return rv;
}

CSSM_RETURN EAYCSP_FetchPublicKey(
                        CSM_SESSION *pSession,
                        uint32 RequiredKeyUse,
                        CSSM_BOOL bEnforceKeyUse,
                        EVP_PKEY **ppSigKey )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_ALGORITHMS RequiredKeyType;
    CSSM_KEY Key;

    if ( ( rv = EAYCSP_FetchKeyData( pSession,
                                     NULL, NULL,
                                     NULL,
                                     CSSM_KEYCLASS_PUBLIC_KEY,
                                     &Key ) ) == CSSM_OK )
    {
        if ( ( rv = EAYCSP_GetCdsaKeyType( 
                                    pSession->pContext->AlgorithmType,
                                    &RequiredKeyType ) ) == CSSM_OK )
        {
            /* Do algorithm and usage checks */
            if ( Key.KeyHeader.KeyClass != CSSM_KEYCLASS_PUBLIC_KEY )
            {
                ERR( rv = CSSMERR_CSP_INVALID_KEY_CLASS );
            }
            else if ( Key.KeyHeader.AlgorithmId != RequiredKeyType )
            {
                ERR( rv = CSSMERR_CSP_ALGID_MISMATCH );
            }
            else if ( bEnforceKeyUse &
                      !CHECKBITS( Key.KeyHeader.KeyUsage,
                                  RequiredKeyUse ) &&
                      !CHECKBITS( Key.KeyHeader.KeyUsage,
                                  CSSM_KEYUSE_ANY ) )
            {
                ERR( rv = CSSMERR_CSP_KEY_USAGE_INCORRECT );
            }
            else
            {
                rv = EAYCSP_EncodedToNativeKey( &Key,
                                                ppSigKey );
            }
        } /* Determine required key type */

        /* Free the key temporary key data buffer */
        memset( Key.KeyData.Data, 0, Key.KeyData.Length );
        Addin_free( Key.KeyData.Data, NULL );
    }  /* Key data fetched */

    return rv;
}

CSSM_RETURN EAYCSP_FetchSessionKey(
                        CSM_SESSION *pSession,
                        uint32 RequiredKeyUse,
                        CSSM_BOOL bEnforceKeyUse,
                        CSSM_DATA *pSymKey )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_ALGORITHMS RequiredKeyType;
    CSSM_KEY Key;
    uint32 KeySize;
    CSSM_DATA KeyParts[CSM_MAX_KEY_PARTS];
    uint32 NumKeyParts = CSM_MAX_KEY_PARTS;

    if ( ( rv = EAYCSP_FetchKeyData( pSession,
                                     NULL, NULL,
                                     NULL,
                                     CSSM_KEYCLASS_SESSION_KEY,
                                     &Key ) ) == CSSM_OK )
    {
        if ( ( rv = EAYCSP_GetCdsaKeyType( 
                                    pSession->pContext->AlgorithmType,
                                    &RequiredKeyType ) ) == CSSM_OK )
        {
            KeySize = Key.KeyHeader.LogicalKeySizeInBits;

            /* Do algorithm and usage checks */
            if ( Key.KeyHeader.KeyClass != CSSM_KEYCLASS_SESSION_KEY )
            {
                ERR( rv = CSSMERR_CSP_INVALID_KEY_CLASS );
            }
            else if ( Key.KeyHeader.AlgorithmId != RequiredKeyType )
            {
                ERR( rv = CSSMERR_CSP_ALGID_MISMATCH );
            }
            else if ( bEnforceKeyUse &&
                      !CHECKBITS( Key.KeyHeader.KeyUsage,
                                  RequiredKeyUse ) &&
                      !CHECKBITS( Key.KeyHeader.KeyUsage,
                                  CSSM_KEYUSE_ANY ) )
            {
                ERR( rv = CSSMERR_CSP_KEY_USAGE_INCORRECT );
            }
            else if ( ( rv = EAYCSP_CheckKeyLimits( 
                                        Key.KeyHeader.AlgorithmId,
                                        KeySize ) ) == CSSM_OK )
            {
                /* Extract the key material from the key blob */
                if ( ( rv = CSM_ParseKeyblob( Key.KeyHeader.AlgorithmId,
                                              CSSM_KEYCLASS_SESSION_KEY,
                                              Key.KeyHeader.Format,
                                              KeyParts,
                                              &NumKeyParts,
                                              &Key.KeyData ) ) == CSSM_OK )
                {
                    /* Check key length against the key itself */
                    if ( ( KeyParts[0].Length * 8 ) != KeySize )
                    {
                        ERR( rv = CSSMERR_CSP_INVALID_ATTR_KEY );
                    }
                    else
                    {
                        /* Return the key to the caller */
                        pSymKey->Data = Addin_malloc( Key.KeyData.Length,
                                                      NULL );
                        if ( pSymKey->Data == NULL )
                        {
                            ERR( rv = CSSMERR_CSP_MEMORY_ERROR );
                        }
                        else
                        {
                            pSymKey->Length = Key.KeyData.Length;
                            memcpy( pSymKey->Data,
                                    Key.KeyData.Data,
                                    Key.KeyData.Length );
                        } /* Allocate and copy the key */
                    } /* Check the key size against the key itself */
                } /* Extract the key from the keyblob */
            } /* Check key parameters */
        } /* Determine required key type */

        /* Free the key buffer from the key manager */
        memset( Key.KeyData.Data, 0, Key.KeyData.Length );
        Addin_free( Key.KeyData.Data, NULL );
    } /* Key data fetched */

    return rv;
}

