/*-----------------------------------------------------------------------
 * File: QRYKSIZE.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "cspinc.h"

CSSM_RETURN CSSMAPI ADDIN_CSM_QueryKeySizeInBits(
                        CSSM_CSP_HANDLE CSPHandle,
                        CSSM_CC_HANDLE CCHandle,
                        const CSSM_CONTEXT *Context,
                        const CSSM_KEY *pKey,
                        CSSM_KEY_SIZE *KeySize )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_KEYHEADER KeyHeader;
    CSM_KM_KEY_HANDLE hKey = CSSM_INVALID_HANDLE;
    const ADDIN_LOAD_DATA *pLoadData = NULL;
    CSSM_DATA KeyPieces[CSM_MAX_KEY_PARTS];
    uint32 NumKeyPieces = CSM_MAX_KEY_PARTS;

    /* Determine how to query the key */
    switch ( pKey->KeyHeader.BlobType )
    {
    case CSSM_KEYBLOB_REFERENCE:
        {
            /* We only support integer references */
            if ( pKey->KeyHeader.Format != CSSM_KEYBLOB_REF_FORMAT_INTEGER )
            {
                ERR( rv = CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT );
            }
            else
            {
                /* Fetch the key header and use its value */
                pLoadData = Addin_GetLoadData();
                assert( pLoadData );
                if ( ( pKey->KeyData.Length != sizeof(CSM_KM_KEY_HANDLE) ) ||
                     port_IsBadReadPtr( pKey->KeyData.Data, sizeof(CSM_KM_KEY_HANDLE) ) )
                {
                    ERR( rv = CSSMERR_CSP_INVALID_KEY_REFERENCE );
                }
                else
                {
                    hKey = *(uint32*)pKey->KeyData.Data;
                    if ( ( rv = CSM_KMGetKeyHeader(
                                    pLoadData->hKeyFile,
                                    hKey,
                                    &KeyHeader ) ) == CSSM_OK )
                    {
                        KeySize->LogicalKeySizeInBits = 
                                            KeyHeader.LogicalKeySizeInBits;
                        KeySize->EffectiveKeySizeInBits = 
                                            KeyHeader.LogicalKeySizeInBits;
                    }
                }
            }
            break;
        } /* case CSSM_KEYBLOB_REFERENCE */
    case CSSM_KEYBLOB_RAW:
        {
            /* Attempt to parse the key */
            if ( ( rv = CSM_ParseKeyblob( pKey->KeyHeader.AlgorithmId,
                                          pKey->KeyHeader.KeyClass,
                                          pKey->KeyHeader.Format,
                                          KeyPieces,
                                          &NumKeyPieces,
                                          &pKey->KeyData ) ) == CSSM_OK )
            {
                switch ( pKey->KeyHeader.AlgorithmId )
                {
                case CSSM_ALGID_DSA:
                    {
                        /* Length is the size of the prime */
                        KeySize->LogicalKeySizeInBits =
                                    KeyPieces[CSM_DSA_PRIME].Length * 8;
                        KeySize->EffectiveKeySizeInBits =
                                    KeySize->LogicalKeySizeInBits;
                        break;
                    }
                case CSSM_ALGID_DES:
                case CSSM_ALGID_GenericSecret:
                    {
                        KeySize->LogicalKeySizeInBits =
                                        KeyPieces[0].Length * 8;
                        KeySize->EffectiveKeySizeInBits =
                                        KeySize->LogicalKeySizeInBits;
                        break;
                    }
                default:
                    {
                        ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
                    }
                }
            }
            break;
        } /* case CSSM_KEYBLOB_RAW */
    default:
        {
            /* Can not calculate a size for keys that are wrapped or
             * have an unknown blob type.
             */
            ERR( rv = CSSMERR_CSP_KEY_BLOB_TYPE_INCORRECT );
        }
    }

    /* Adjust the effective key size for DES variants */
    if ( rv == CSSM_OK )
    {
        if ( ( pKey->KeyHeader.AlgorithmId == CSSM_ALGID_DES ) )
        {
            KeySize->EffectiveKeySizeInBits = 
                            ( KeySize->EffectiveKeySizeInBits / 8 ) * 7;
        }
    }

    return rv;
}

