/*-----------------------------------------------------------------------
 * File: QRYSIZE.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "cspinc.h"

CSSM_RETURN CSSMAPI ADDIN_CSM_QuerySize(
                        CSM_SESSION *pSession,
                        CSSM_BOOL Encrypt,
                        uint32 QuerySizeCount, 
                        CSSM_QUERY_SIZE_DATA *DataBlock )
{
    CSSM_RETURN rv = CSSM_OK;
    uint32 i;
    const CSSM_CONTEXT_ATTRIBUTE *pAttr = NULL;

    switch ( pSession->SessionType )
    {
    case CSM_ST_VERIFY:
    case CSM_ST_VERIFY_MAC:
        {
            /* These functions product no output, just a result */
            for ( i = 0; i < QuerySizeCount; i++ )
            {
                DataBlock[i].SizeOutputBlock = 0;
            }
            break;
        } /* csae CSM_ST_VERIFY* */
    case CSM_ST_RANDOMGEN:
        {
            pAttr = CSM_GetContextAttribute( pSession->pContext,
                                             CSSM_ATTRIBUTE_OUTPUT_SIZE );
            for ( i = 0; i < QuerySizeCount; i++ )
            {
                if ( pAttr )
                {
                    DataBlock[i].SizeOutputBlock = pAttr->Attribute.Uint32;
                }
                else
                {
                    ERR( rv = CSSMERR_CSP_QUERY_SIZE_UNKNOWN );
                }
            }
            break;
        } /* case CSM_ST_RANDOMGEN */
    case CSM_ST_DIGEST:
        {
            for ( i = 0; ( i < QuerySizeCount ) && ( rv == CSSM_OK ); i++ )
            {
                switch ( pSession->pContext->AlgorithmType )
                {
                case CSSM_ALGID_MD5:
                    {
                        DataBlock[i].SizeOutputBlock = 
                                                EVP_MD_size( EVP_md5() );
                        break;
                    }
                case CSSM_ALGID_SHA1:
                    {
                        DataBlock[i].SizeOutputBlock = 
                                                EVP_MD_size( EVP_sha1() );
                        break;
                    }
                default:
                    {
                        ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
                    }
                }
            }
            break;
        } /* case CSM_ST_DIGEST */
    case CSM_ST_SIGN:
        {
            uint32 Length = 0;

            switch ( pSession->pContext->AlgorithmType )
            {
            case CSSM_ALGID_DSA:
            case CSSM_ALGID_SHA1WithDSA:
                {
                    Length = 48; /* Maximum length with DER encoding */
                    break;
                }
            default:
                {
                    ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
                }
            }

            for ( i = 0; ( i < QuerySizeCount ) && ( rv == CSSM_OK ); i++ )
            {
                DataBlock[i].SizeOutputBlock = Length;
            }
            break;
        } /* case CSM_ST_SIGN */
    case CSM_ST_ENCRYPT:
        {
            /* The lengths for the output blocks are calculated as follows:
             * 1) Determine how many bytes are in the incomplete block.
             * 2) Determine the output length for the number of complete
             *    block created using the incomplete block length and the
             *    new data length.
             * 3a) If there are more lengths to calculate, update the number
             *    of bytes in the incomplete block and goto step 1 for the
             *    next buffer.
             * 3b) If this is the last block, add the length of the final
             *    block including the pad amount.
             */
            EVP_CIPHER_CTX *pState = NULL;
            uint32 IncompleteLength = 0;
            uint32 BlockSize = 1;

            /* Step 1 */
            pState = (EVP_CIPHER_CTX*)pSession->pCryptoState;
            if ( pState == NULL )
            {
                IncompleteLength = 0; /* Assume starting state */
                switch ( pSession->pContext->AlgorithmType )
                {
                case CSSM_ALGID_DES:
                    {
                        /* Any mode will work since only block variants are
                         * exported.
                         */
                        BlockSize = EVP_CIPHER_block_size( EVP_des_ecb() );
                        break;
                    }
                default:
                    {
                        ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
                    }
                }
            }
            else
            {
                IncompleteLength = pState->buf_len;
                BlockSize = EVP_CIPHER_CTX_block_size( pState );
            }

            if ( rv == CSSM_OK )
            {
                for ( i = 0; i < QuerySizeCount; i++ )
                {
                    /* Step 2 */
                    DataBlock[i].SizeOutputBlock =  
                            ( ( DataBlock[i].SizeInputBlock + IncompleteLength ) /
                               BlockSize
                            ) * BlockSize;

                    /* Step 3 */
                    if ( i == ( QuerySizeCount -1 ) )
                    {
                        DataBlock[i].SizeOutputBlock += BlockSize;
                    }
                    else
                    {
                        IncompleteLength = 
                            ( IncompleteLength + DataBlock[i].SizeInputBlock )
                            % BlockSize;
                    }
                }
            }
            break;
        }
    case CSM_ST_DECRYPT:
        {
            /* The lengths for the output blocks are calculated as follows:
             * 1) Determine how many bytes are in the incomplete block.
             * 2) Determine the output length for the number of complete
             *    block created using the incomplete block length and the
             *    new data length.
             * 3) Update the number of bytes in the incomplete block and
             *    goto step 1 for the next buffer.
             */
            EVP_CIPHER_CTX *pState = NULL;
            uint32 IncompleteLength = 0;
            uint32 BlockSize = 1;

            /* Step 1 */
            pState = (EVP_CIPHER_CTX*)pSession->pCryptoState;
            if ( pState == NULL )
            {
                IncompleteLength = 0; /* Assume starting state */
                switch ( pSession->pContext->AlgorithmType )
                {
                case CSSM_ALGID_DES:
                    {
                        /* Any mode will work since only block variants are
                         * exported.
                         */
                        BlockSize = EVP_CIPHER_block_size( EVP_des_ecb() );
                        break;
                    }
                default:
                    {
                        ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
                    }
                }
            }
            else
            {
                IncompleteLength = pState->buf_len;
                BlockSize = EVP_CIPHER_CTX_block_size( pState );
            }

            if ( rv == CSSM_OK )
            {
                for ( i = 0; i < QuerySizeCount; i++ )
                {
                    /* Step 2 */
                    DataBlock[i].SizeOutputBlock =  
                            ( ( DataBlock[i].SizeInputBlock + IncompleteLength ) /
                               BlockSize
                            ) * BlockSize;

                    /* Step 3 */
                    IncompleteLength = 
                            ( IncompleteLength + DataBlock[i].SizeInputBlock )
                            % BlockSize;
                }
            }
            break;
        }
    case CSM_ST_NONE:
    case CSM_ST_UNIQUEGEN:
    case CSM_ST_KEYGEN:
    case CSM_ST_DERIVE:
    case CSM_ST_WRAP:
    case CSM_ST_UNWRAP:
    case CSM_ST_GEN_MAC:
    case CSM_ST_CUSTOM:
        {
            /* The CSP doesn't implement contexts of this type, or it
             * doesn't make sense to query this type of operation.
             */
            ERR( rv = CSSMERR_CSP_INVALID_CONTEXT );
            break;
        }
    default:
        {
            /* Something corrupted the tracking session */
            ERR( rv = CSSMERR_CSP_INTERNAL_ERROR );
        }
    }
    return rv;
}

