/*-----------------------------------------------------------------------
 * File: CSP_KGEN
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/*
 * File Description: Contains key generation operations for the CSP.
 */

#include "geninc.h"

extern B_ALGORITHM_OBJ  g_RandomNumGen;

/* Algorithm methods for key generation */
B_ALGORITHM_METHOD  *RSA_KEY_GEN_CHOOSER[] = {
    &AM_RSA_KEY_GEN,
    (B_ALGORITHM_METHOD*)NULL_PTR
};
B_ALGORITHM_METHOD  *DSA_KEY_GEN_CHOOSER[] = {
    &AM_DSA_PARAM_GEN,
    &AM_DSA_KEY_GEN,
    (B_ALGORITHM_METHOD*)NULL_PTR
};
B_ALGORITHM_METHOD  *DH_KEY_GEN_CHOOSER[] = {
  &AM_DH_KEY_AGREE,
  (B_ALGORITHM_METHOD *)NULL_PTR
};

/* Generate key pair function for RSA and DSA keys */
static CSSM_RETURN csp_GenerateKeyPair(CSM_SESSION *pSession,
                                    uint32 algid,
                                    uint32 nPublicKeyUsage,
                                    uint32 nPublicKeyAttr,
                                    const CSSM_DATA *pPublicKeyLabel,
                                    CSSM_KEY_PTR pPubKey,
                                    uint32 nPrivateKeyUsage,
                                    uint32 nPrivateKeyAttr,
                                    const CSSM_DATA *pPrivateKeyLabel,
                                    CSSM_KEY_PTR pPriKey,
                                    const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry );

/* Generate key pair function for Diffie Hellman */
static CSSM_RETURN csp_GenerateKeyPair_DH(CSM_SESSION *pSession, 
                                      uint32 nPublicKeyUsage,
                                      uint32 nPublicKeyAttr,
                                      const CSSM_DATA *pPublicKeyLabel,
                                      CSSM_KEY_PTR pPubKey,
                                      uint32 nPrivateKeyUsage,
                                      uint32 nPrivateKeyAttr,
                                      const CSSM_DATA *pPrivateKeyLabel,
                                      CSSM_KEY_PTR pPriKey,
                                      const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry );

CSSM_RETURN CSSMAPI ADDIN_CSM_GenerateKey( 
                         CSM_SESSION *pSession,
                         uint32 KeyUsage,
                         uint32 KeyAttr,
                         const CSSM_DATA *KeyLabel,
                         const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
                         CSSM_KEY_PTR Key,
						 CSSM_PRIVILEGE Privilege )
{
    CSSM_RETURN                      retval = CSSM_OK;
    CSSM_CONTEXT                    *pContext = pSession->pContext;
    const CSSM_CONTEXT_ATTRIBUTE    *pKeyBits = NULL;
    uint32                          NumBytesToGenerate = 0;
    uint32                          algType, keylength = 0;
    CSSM_BOOL                       bAllocatedOutput;
	const CSSM_CONTEXT_ATTRIBUTE    *contextAttr = NULL;
	uint32                          SeedLen = 0;
    CSSM_DATA                       seedData = {0, NULL};
	CSSM_BOOL                       bAllocatedSeed = CSSM_FALSE;

    /* Unreferenced formal parameter; suppress compiler warning */
    KeyLabel;
    
    algType = pContext->AlgorithmType;

    if ((algType == CSSM_ALGID_DES) ||
        (algType == CSSM_ALGID_RC2) ||
        (algType == CSSM_ALGID_RC4) ||
        (algType == CSSM_ALGID_RC5) ||
		(algType == CSSM_ALGID_DESX) ||
		(algType == CSSM_ALGID_3DES_3KEY))
    {
        /* Check the key attribute */
        if ((KeyAttr & CSSM_KEYATTR_RETURN_REF) ||
            (KeyAttr & CSSM_KEYATTR_RETURN_NONE) ||
            (KeyAttr & CSSM_KEYATTR_PERMANENT) ||
            (KeyAttr & CSSM_KEYATTR_MODIFIABLE) ||
            (KeyAttr & CSSM_KEYATTR_SENSITIVE))
        {
            return CSSMERR_CSP_UNSUPPORTED_KEYATTR_MASK;
        }

        pKeyBits = CSM_GetContextAttribute(pContext, CSSM_ATTRIBUTE_KEY_LENGTH);
        if (!pKeyBits)
        {
            return CSSMERR_CSP_INVALID_ATTR_KEY_LENGTH;
        }

        keylength = pKeyBits->Attribute.Uint32;

        if (algType == CSSM_ALGID_DES)
        {
            if (keylength != 64)
            {
                retval = CSSMERR_CSP_UNSUPPORTED_KEY_SIZE;
            }
        }
        else if ((algType == CSSM_ALGID_RC4) || (algType == CSSM_ALGID_RC5))
        {
            if ((keylength < 0) || (keylength > 128))
            {
                retval = CSSMERR_CSP_UNSUPPORTED_KEY_SIZE;
            }
        }
        else if (algType == CSSM_ALGID_RC2)
        {
            if ((keylength < 0) || (keylength > 1024))
            {
                retval = CSSMERR_CSP_UNSUPPORTED_KEY_SIZE;
            }
        }
		else if (algType == CSSM_ALGID_3DES_3KEY)
        {
            if (keylength != 192)
            {
                retval = CSSMERR_CSP_UNSUPPORTED_KEY_SIZE;
            }
        }
		else if (algType == CSSM_ALGID_DESX)
        {
            if (keylength != 128)
            {
                retval = CSSMERR_CSP_UNSUPPORTED_KEY_SIZE;
            }
        }

		if ( retval == CSSMERR_CSP_UNSUPPORTED_KEY_SIZE )
		{
			/* Adjust the err code if it indicates invalid key size
			 * since the size came from a context attribute.
			 */
			return CSSMERR_CSP_INVALID_ATTR_KEY_LENGTH;
		}

        if ((pKeyBits->Attribute.Uint32) % 8)
        {
            NumBytesToGenerate = pKeyBits->Attribute.Uint32 / 8 + 1;
        }
        else
        {
            NumBytesToGenerate = pKeyBits->Attribute.Uint32 / 8;
        }

        /* Set the KeyData to NULL so that the buffer will be allocated by the
         * CSP */
        Key->KeyData.Data = NULL;
        Key->KeyData.Length = 0;

		/* 
		 * Retrieve the seed and get internal seed from the caller seed
		 * currently this code is used to return the proper error code. The seed value
		 * is not used 
		 */
		if ((contextAttr = CSM_GetContextAttribute(pContext, 
						CSSM_ATTRIBUTE_SEED)) != NULL)
		{
			retval = MAF_ResolveCryptoData(
							contextAttr->Attribute.CryptoData,
							&seedData,
							SeedLen,
							&bAllocatedSeed );
			if ( retval == CSSMERR_CSP_INVALID_CRYPTO_DATA )            
                retval = CSSMERR_CSP_INVALID_ATTR_SEED;
			if (retval != CSSM_OK)
				return retval;

			if (bAllocatedSeed == CSSM_TRUE)
			{
				Addin_free(seedData.Data, NULL);
			}
		}
        
        /* Allocate space for the key */
        retval = MAF_CreateOutputBuffer(pSession->hCsp, NumBytesToGenerate,
                            &(Key->KeyData), &bAllocatedOutput);
        if (retval == CSSM_OK)
        {
            /* Generate random bits for symmetric key */
            if ((retval = csp_GenerateRandomBytes(NumBytesToGenerate,
                            Key->KeyData.Data)) != CSSM_OK)
            {
                MAF_DeleteOutputBuffer(pSession->hCsp, &(Key->KeyData), 
                            bAllocatedOutput); 
                return retval;
            }
            Key->KeyData.Length = NumBytesToGenerate;

            /* Set the key header */
            {
                uint32      nBlobType=0, nKeyFormat=0, nKeyClass=0,
                            nKeySizeInBits=0, nKeyAttr=0;
                const CSSM_CONTEXT_ATTRIBUTE    *pStart = NULL, *pEnd = NULL;
                CSSM_DATE   *pStartDate = NULL, *pEndDate = NULL;

				
                pStart = CSM_GetContextAttribute(pContext, CSSM_ATTRIBUTE_START_DATE);
                if (pStart)
                {
                    pStartDate = pStart->Attribute.Date;
                }

                pEnd = CSM_GetContextAttribute(pContext, CSSM_ATTRIBUTE_END_DATE);
                if (pEnd)
                {
                    pEndDate = pEnd->Attribute.Date;
				}


                nBlobType = CSSM_KEYBLOB_RAW;
                nKeyFormat= CSSM_KEYBLOB_RAW_FORMAT_NONE;
                nKeyClass = CSSM_KEYCLASS_SESSION_KEY;
                nKeyAttr = KeyAttr;
                nKeySizeInBits = pKeyBits->Attribute.Uint32;
                csp_KeyHeader_Fill(&Key->KeyHeader, nBlobType, nKeyFormat, 
                                    pSession->pContext->AlgorithmType,
                                    nKeyClass, nKeySizeInBits, nKeyAttr,
                                    KeyUsage, pStartDate, pEndDate,
                                    CSSM_ALGID_NONE, 0, 0);
            }
        }
    }
    else
    {
        retval = CSSMERR_CSP_INVALID_ALGORITHM;
    }

    return retval;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_GenerateKeyPair( 
                        CSM_SESSION *pSession,
                        uint32 PublicKeyUsage,
                        uint32 PublicKeyAttr,
                        const CSSM_DATA *PublicKeyLabel,
                        CSSM_KEY_PTR PublicKey,
                        uint32 PrivateKeyUsage,
                        uint32 PrivateKeyAttr,
                        const CSSM_DATA *PrivateKeyLabel,
                        const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
                        CSSM_KEY_PTR PrivateKey,
                        CSSM_PRIVILEGE Privilege )
{
    CSSM_RETURN         retval = CSSM_OK;
    uint32              algID = CSSM_ALGID_NONE;
    
    if (pSession)
    {
        if (pSession->pContext)
        {
            algID = pSession->pContext->AlgorithmType;
        }
        else
        {
            retval = CSSMERR_CSP_INTERNAL_ERROR;
        }
    }
    else
    {
        retval = CSSMERR_CSP_INTERNAL_ERROR;
    }

    /* Check the key attributes */
    if ((PublicKeyAttr & CSSM_KEYATTR_RETURN_REF) ||
        (PublicKeyAttr & CSSM_KEYATTR_PERMANENT) ||
        (PublicKeyAttr & CSSM_KEYATTR_PRIVATE) ||
        (PublicKeyAttr & CSSM_KEYATTR_MODIFIABLE) ||
        (PublicKeyAttr & CSSM_KEYATTR_SENSITIVE))
    {
        retval = CSSMERR_CSP_UNSUPPORTED_KEYATTR_MASK;
    }
    if ((PrivateKeyAttr & CSSM_KEYATTR_RETURN_DATA) ||
        (PrivateKeyAttr & CSSM_KEYATTR_MODIFIABLE))
    {
        retval = CSSMERR_CSP_UNSUPPORTED_KEYATTR_MASK;
    }

    if (retval == CSSM_OK)
    {
        switch (algID)
        {
            case CSSM_ALGID_RSA:
            case CSSM_ALGID_DSA:
                retval = csp_GenerateKeyPair(pSession, algID, PublicKeyUsage,
                            PublicKeyAttr, PublicKeyLabel, PublicKey,
                            PrivateKeyUsage, PrivateKeyAttr, PrivateKeyLabel,
                            PrivateKey,
                            CredAndAclEntry );
                break;

            case CSSM_ALGID_DH:
                retval = csp_GenerateKeyPair_DH(pSession, PublicKeyUsage,
                            PublicKeyAttr, PublicKeyLabel, PublicKey,
                            PrivateKeyUsage, PrivateKeyAttr, PrivateKeyLabel,
                            PrivateKey, CredAndAclEntry );
                break;

            default:
            {
                retval = CSSMERR_CSP_INVALID_ALGORITHM;
                break;
            }
        }
    }

    return retval;
}

/*****************************
 *        CSP_DSA/RSA        *
 *****************************/
/*---------------------------------------------------------------
 *Name: csp_GenerateKeyPair
 *
 *Description:
 * This function generates an asymmetric DSA or RSA key pair.
 *
 *Parameters: 
 * pSession(input)        - MAF maintained structure contains context, state,
 *                          and other information for the crypto operation.
 * algid(input)           - The algorithm type of the key to be generated.
 *                          Either RSA or DSA.
 * nPublicKeyUsage(input) - A bit mask specifying how the new public key can 
 *                          be used.
 * nPublicKeyAttr(input)  - A bit mask specifying other attributes to be 
 *                          associated with the new public key. 
 * pPublicKeyLabel(input) - Pointer to a byte string that will be used as the
 *                          label for the public key.
 * pPubKey(output)        - Pointer to CSSM_KEY structure used to obtain the 
 *                          public key.
 * nPrivateKeyUsage(input)- A bit mask specifying how the new private key can
 *                          be used.
 * nPrivateKeyAttr(input) - A bit mask specifying other attributes to be
 *                          associated with the new private key.
 * pPrivateKeyLabel(input)- Pointer to a byte string that will be used as the
 *                          label for the private key.
 * pPriKey(output)        - Pointer to CSSM_ KEY structure used to obtain the
 *                          private key.
 *
 *Returns:
 * CSSM_OK -   Key pair generation succeeded
 *-------------------------------------------------------------*/
static CSSM_RETURN csp_GenerateKeyPair(CSM_SESSION *pSession,
                                    uint32 algid,
                                    uint32 nPublicKeyUsage,
                                    uint32 nPublicKeyAttr,
                                    const CSSM_DATA *pPublicKeyLabel,
                                    CSSM_KEY_PTR pPubKey,
                                    uint32 nPrivateKeyUsage,
                                    uint32 nPrivateKeyAttr,
                                    const CSSM_DATA *pPrivateKeyLabel,
                                    CSSM_KEY_PTR pPriKey,
                                    const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry )
{
    CSSM_RETURN                     rtn = CSSM_OK;
	uint32                          algType;
    uint32                          nKeySizeInBits=0;
    CSSM_KEY                        TempPri, TempPub;
    B_ALGORITHM_OBJ                 keypairGenerator = (B_ALGORITHM_OBJ)NULL_PTR,
                                    dsaParamGenerator = (B_ALGORITHM_OBJ)NULL_PTR;
    int                             status;
    B_DSA_PARAM_GEN_PARAMS          dsaParams;
    A_RSA_KEY_GEN_PARAMS            rsaKeygenParams;
    static unsigned char            f4Data[3] = { 0x01, 0x00, 0x01 };
    B_KEY_OBJ                       pubKey = (B_KEY_OBJ)NULL_PTR,
                                    priKey = (B_KEY_OBJ)NULL_PTR;
    ITEM                            *pubKeyBER, *priKeyBER;
    int                             loopMultiple = CSSM_FALSE;
    CSSM_CONTEXT                    *pContext = pSession->pContext;
    CSSM_BOOL                       bAllocatedOutput;
    const CSSM_CONTEXT_ATTRIBUTE    *pKeyBits = NULL;
#if defined (COMPAQ_FIX) || defined (VMS)
    const CSSM_CONTEXT_ATTRIBUTE    *pPublicExponent = NULL;
#endif
    const ADDIN_LOAD_DATA           *loadData;
    CSM_KM_KEY_HANDLE               keyHandle;
    
    /* Get the KeySizeInBits from context attributes */
    pKeyBits = CSM_GetContextAttribute(pContext, CSSM_ATTRIBUTE_KEY_LENGTH);
    if (!pKeyBits)
    {
        return CSSMERR_CSP_INVALID_ATTR_KEY_LENGTH;
    }

    nKeySizeInBits = pKeyBits->Attribute.Uint32;

	algType = pContext->AlgorithmType;
    if (algType == CSSM_ALGID_DES)
    {
        if (nKeySizeInBits != 64)
        {
            rtn = CSSMERR_CSP_UNSUPPORTED_KEY_SIZE;
        }
    }
    else if ((algType == CSSM_ALGID_RC4) || (algType == CSSM_ALGID_RC5))
    {
        if ((nKeySizeInBits < 0) || (nKeySizeInBits > 128))
        {
            rtn = CSSMERR_CSP_UNSUPPORTED_KEY_SIZE;
        }
    }
    else if (algType == CSSM_ALGID_RC2)
    {
        if ((nKeySizeInBits < 0) || (nKeySizeInBits > 1024))
        {
            rtn = CSSMERR_CSP_UNSUPPORTED_KEY_SIZE;
        }
    }

	if ( rtn == CSSMERR_CSP_UNSUPPORTED_KEY_SIZE )
	{
		/* Adjust the err code if it indicates invalid key size
		 * since the size came from a context attribute.
		 */
		return CSSMERR_CSP_INVALID_ATTR_KEY_LENGTH;
	}

    memset(&TempPub, 0, sizeof(CSSM_KEY));
    memset(&TempPri, 0, sizeof(CSSM_KEY));
    TempPub.KeyData.Data = Addin_calloc(1, MAX_KEYBUF_LEN, NULL);
    if (TempPub.KeyData.Data == NULL)
    {
        return CSSMERR_CSP_MEMORY_ERROR;
    }
    TempPub.KeyData.Length = MAX_KEYBUF_LEN;
    TempPri.KeyData.Data = Addin_calloc(1, MAX_KEYBUF_LEN, NULL);
    if (TempPri.KeyData.Data == NULL)
    {
        Addin_free(TempPub.KeyData.Data, NULL);
        return CSSMERR_CSP_MEMORY_ERROR;
    }
    TempPri.KeyData.Length = MAX_KEYBUF_LEN;

    do {
        /* Create the keypair generation object */
        if ((status = B_CreateAlgorithmObject(&keypairGenerator)) != 0)
        {
            rtn = ConvertBsafeError(status);
            return rtn;
        }
        if ((status = B_CreateKeyObject(&pubKey)) != 0)
        {
            rtn = ConvertBsafeError(status);
            break;
        }
        if ((status = B_CreateKeyObject(&priKey)) != 0)
        {
            rtn = ConvertBsafeError(status);
            break;
        }

        if (algid == CSSM_ALGID_DSA)
        {
            if ((status = B_CreateAlgorithmObject (&dsaParamGenerator)) != 0)
            {
                rtn = ConvertBsafeError(status);
                break;
            }

            dsaParams.primeBits = nKeySizeInBits;
            if ((status = B_SetAlgorithmInfo
                 (dsaParamGenerator, AI_DSAParamGen, (POINTER)&dsaParams)) != 0)
            {
                rtn = ConvertBsafeError(status);
                break;
            }
            if ((status = B_GenerateInit
                 (dsaParamGenerator, DSA_KEY_GEN_CHOOSER, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
            {
                rtn = ConvertBsafeError(status);
                break;
            }
            if ((status = B_GenerateParameters
                (dsaParamGenerator, keypairGenerator, g_RandomNumGen,
                (A_SURRENDER_CTX *)NULL_PTR)) != 0)
            {
                rtn = ConvertBsafeError(status);
                break;
            }
        
            if ((status = B_GenerateInit
                (keypairGenerator, DSA_KEY_GEN_CHOOSER, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
            {
                rtn = ConvertBsafeError(status);
                break;
            }
        }
        else if (algid == CSSM_ALGID_RSA)
        {
            rsaKeygenParams.modulusBits = nKeySizeInBits;
#if defined (COMPAQ_FIX) || defined (VMS)
	    /* If a parameter was passed in, use it for the publicExponent */
	    pPublicExponent = CSM_GetContextAttribute(pContext,
						      CSSM_ATTRIBUTE_ALG_PARAMS);
	    if (pPublicExponent != NULL)
	    {
	      rsaKeygenParams.publicExponent.data =
		pPublicExponent->Attribute.Data->Data;
	      rsaKeygenParams.publicExponent.len =
		pPublicExponent->Attribute.Data->Length;
	    }
	    else
	    {
	      /* If a parameter was not passed in, use the default */
	      rsaKeygenParams.publicExponent.data = f4Data;
	      rsaKeygenParams.publicExponent.len = 3;
	    }
#else
            rsaKeygenParams.publicExponent.data = f4Data;
            rsaKeygenParams.publicExponent.len = 3;
#endif
            if ((status = B_SetAlgorithmInfo(keypairGenerator, AI_RSAKeyGen, (POINTER)&rsaKeygenParams)) != 0) {
                rtn = ConvertBsafeError(status);
                break;
            }
            /* Initialize keypair generation */
            if ((status = B_GenerateInit(keypairGenerator, RSA_KEY_GEN_CHOOSER, (A_SURRENDER_CTX*)NULL_PTR)) != 0) {
                rtn = ConvertBsafeError(status);
                break;
            }
        }

        /* Generate key */
        if ((status = B_GenerateKeypair(keypairGenerator, pubKey, priKey, 
                            g_RandomNumGen, (A_SURRENDER_CTX*)NULL_PTR)) != 0)
        {
            rtn = ConvertBsafeError(status);
            break;
        }

        if (algid == CSSM_ALGID_DSA)
        {
            if ((status = B_GetKeyInfo((POINTER*)&pubKeyBER, pubKey, KI_DSAPublicBER)) != 0)
            {
                rtn = ConvertBsafeError(status);
                break;
            }

            /* Convert the BSafe DSA Public Key to a CSSM DSA Public key */
            if ((rtn = Cryp_BSafeDSAPublicKey_to_CSSMDSAPublicKey(nKeySizeInBits,
                                nPublicKeyUsage, pubKeyBER->data,
                                &TempPub)) != CSSM_OK)
            {
                break;
            }
        }
        else if (algid == CSSM_ALGID_RSA)
        {
            if ((status = B_GetKeyInfo((POINTER*)&pubKeyBER, pubKey,
                                KI_RSAPublicBER)) != 0)
            {
                rtn = ConvertBsafeError(status);
                break;
            }

            /* Convert the BSafe Public Key to a PKCS1 Public key for CSSM */
            if ((rtn = Cryp_BSafeRSAPublicKey_to_CSSMRSAPublicKey(nKeySizeInBits,
                                nPublicKeyUsage, pubKeyBER->data,
                                &TempPub)) != CSSM_OK)
            {
                break;
            }
        }

        /* Check to see if the application wanted to receive the public key
         * information back.
         */
        if ((nPublicKeyAttr & CSSM_KEYATTR_RETURN_NONE) == 0)
        {
            if (pPubKey)
            {
                /* CSP is required to allocate space for the key, so ignore any
                 * values that were set in the KeyData.
                 */
                pPubKey->KeyData.Length = 0;
                pPubKey->KeyData.Data = NULL;

                if ((rtn = MAF_CreateOutputBuffer(pSession->hCsp, 
                            TempPub.KeyData.Length, &(pPubKey->KeyData),
                            &bAllocatedOutput)) != CSSM_OK)
                {
                    break;
                }
                memcpy(&pPubKey->KeyHeader, &TempPub.KeyHeader, sizeof(CSSM_KEYHEADER));
                memcpy(pPubKey->KeyData.Data, TempPub.KeyData.Data, TempPub.KeyData.Length);
                pPubKey->KeyData.Length = TempPub.KeyData.Length;
            }
            else
            {
                rtn = CSSMERR_CSP_INVALID_KEY_POINTER;
                break;
            }
        }

        if (algid == CSSM_ALGID_DSA)
        {
            if ((status = B_GetKeyInfo((POINTER*)&priKeyBER, priKey, 
                                        KI_DSAPrivateBER)) != 0)
            {
                rtn = ConvertBsafeError(status);
                break;
            }

            /* Convert the BSafe DSA Private Key to a CSSM DSA Private key */
            if ((rtn = Cryp_BSafeDSAPrivateKey_to_CSSMDSAPrivateKey(nKeySizeInBits,
                                 nPrivateKeyUsage, priKeyBER->data,
                                 &TempPri)) != CSSM_OK)
            {
                break;
            }
        }
        else if (algid == CSSM_ALGID_RSA)
        {
            if ((status = B_GetKeyInfo((POINTER*)&priKeyBER, priKey,
                                        KI_PKCS_RSAPrivateBER)) != 0)
            {
                rtn = ConvertBsafeError(status);
                break;
            }

            /* Convert the BSafe Private Key to a PKCS1 Private key for CSSM */
            if ((rtn = Cryp_BSafeRSAPrivateKey_to_CSSMRSAPrivateKey(nKeySizeInBits,
                                 nPrivateKeyUsage, priKeyBER->data,
                                 &TempPri)) != CSSM_OK)
            {
                break;
            }
        }

        /* Set the key attributes, as passed by the user */
#if defined (COMPAQ_FIX) || defined (VMS)
	/* making sure to filter out the API only attributes. If they did not
	 * passed in settable attributes use the attributes from the
	 * generation routine.
	 */
		TempPri.KeyHeader.KeyAttr = (nPrivateKeyAttr
					     & ~(CSSM_KEYATTR_RETURN_DEFAULT |
						CSSM_KEYATTR_RETURN_DATA |
						CSSM_KEYATTR_RETURN_REF |
						CSSM_KEYATTR_RETURN_NONE)) ?
		                             nPrivateKeyAttr
					     & ~(CSSM_KEYATTR_RETURN_DEFAULT |
						CSSM_KEYATTR_RETURN_DATA |
						CSSM_KEYATTR_RETURN_REF |
						CSSM_KEYATTR_RETURN_NONE) :
	                                     TempPri.KeyHeader.KeyAttr;
#else
		TempPri.KeyHeader.KeyAttr = nPrivateKeyAttr;
#endif

        /* Place the private key into the key manager */
        loadData = Addin_GetLoadData();

        if (loadData == NULL)
        {
            rtn = CSSMERR_CSP_INTERNAL_ERROR;
            break;
        }

        if ((rtn = CSM_KMInsertKey(loadData->hKeyFile, &TempPri, &TempPub, NULL,
                              CredAndAclEntry, &keyHandle)) != CSSM_OK)
        {
            break;
        }

        /* Only return the private key reference if it was asked for */
        if ((nPrivateKeyAttr & CSSM_KEYATTR_RETURN_NONE) == 0)
        {
            if (pPriKey)
            {
                /* CSP is required to allocate space for the key, so ignore any
                 * values that were set in the KeyData.
                 */
                pPriKey->KeyData.Length = 0;
                pPriKey->KeyData.Data = NULL;

                if ((rtn = MAF_CreateOutputBuffer(pSession->hCsp, 
                            sizeof(CSM_KM_KEY_HANDLE), &(pPriKey->KeyData),
                            &bAllocatedOutput)) != CSSM_OK)
                {
                    break;
                }
                memcpy(&pPriKey->KeyHeader, &TempPri.KeyHeader, sizeof(CSSM_KEYHEADER));

                /* Correct the private key header information */
                pPriKey->KeyHeader.BlobType = CSSM_KEYBLOB_REFERENCE;
                pPriKey->KeyHeader.Format = CSSM_KEYBLOB_REF_FORMAT_INTEGER;

                memcpy(pPriKey->KeyData.Data, &keyHandle, sizeof(CSM_KM_KEY_HANDLE));
                pPriKey->KeyData.Length = sizeof(CSM_KM_KEY_HANDLE);
            }
        }
    } while (loopMultiple); /* suppress compiler warning */


    if (TempPub.KeyData.Data)
    {
        Addin_free(TempPub.KeyData.Data, NULL);
    }
    if (TempPri.KeyData.Data)
    {
        /* Zero out the raw private key then free the memory */
        memset(TempPri.KeyData.Data, 0, TempPri.KeyData.Length);
        Addin_free(TempPri.KeyData.Data, NULL);
    }

    /* Destroy objects */
    B_DestroyAlgorithmObject(&keypairGenerator);
    
    if (algid == CSSM_ALGID_DSA)
    {
        B_DestroyAlgorithmObject(&dsaParamGenerator);
    }
    B_DestroyKeyObject(&pubKey);
    B_DestroyKeyObject(&priKey);

    return rtn;
}

/*****************************
 *          CSP_DH           *
 *****************************/
/*---------------------------------------------------------------
 *Name: csp_GenerateKeyPair_DH
 *
 *Description:
 * This function generates an asymmetric DH key pair.
 *
 *Parameters: 
 * Session(input) - MAF maintained structure contains context, state, and
 *                  other information for the crypto operation.
 * nPublicKeyUsage(input) - A bit mask specifying how the new public key can be
 *                          used.
 * nPublicKeyAttr(input) - A bit mask specifying other attributes to be
 *                         associated with the new public key. 
 * pPublicKeyLabel(input) - Pointer to a byte string that will be used as the
 *                          label for the public key.
 * pPubKey(output) - Pointer to CSSM_KEY structure used to obtain the public
 *                   key.
 * nPrivateKeyUsage(input) - A bit mask specifying how the new private key can
 *                           be used.
 * nPrivateKeyAttr(input) - A bit mask specifying other attributes to be
 *                          associated with the new private key.
 * pPrivateKeyLabel(input) - Pointer to a byte string that will be used as the
 *                           label for the private key.
 * pPriKey(output) - Pointer to CSSM_KEY structure used to obtain the private
 *                   key.
 *
 *Returns:
 * CSSM_OK -   Key pair generation succeeded
 *-------------------------------------------------------------*/
static CSSM_RETURN csp_GenerateKeyPair_DH(CSM_SESSION *pSession, 
                                      uint32 nPublicKeyUsage,
                                      uint32 nPublicKeyAttr,
                                      const CSSM_DATA *pPublicKeyLabel,
                                      CSSM_KEY_PTR pPubKey,
                                      uint32 nPrivateKeyUsage,
                                      uint32 nPrivateKeyAttr,
                                      const CSSM_DATA *pPrivateKeyLabel,
                                      CSSM_KEY_PTR pPriKey,
                                      const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry)
{
    CSSM_RETURN                     rtn = CSSM_OK;
    CSSM_KEY                        tempPriKey;
    CSM_KM_KEY_HANDLE               keyHandle;
    uint32                          status;
    CSSM_BOOL                       bPubAllocOutput = CSSM_FALSE,
                                    bPriAllocOutput = CSSM_FALSE;
    const CSSM_CONTEXT_ATTRIBUTE    *pAttribute = NULL;
    CSSM_DATA                       *pAlgParams = NULL;
    B_ALGORITHM_OBJ                 dhKeyAgreeAlg = (B_ALGORITHM_OBJ)NULL_PTR;
    B_ALGORITHM_OBJ                 randomNumGen = (B_ALGORITHM_OBJ)NULL_PTR;
    ITEM                            dhParametersBER;
    A_DH_KEY_AGREE_PARAMS           *getParams;
    const ADDIN_LOAD_DATA           *loadData;

    /* Initialization */
    memset(&tempPriKey, 0, sizeof(CSSM_KEY));

    if ((nPublicKeyAttr & CSSM_KEYATTR_RETURN_REF) || 
        (nPublicKeyAttr & CSSM_KEYATTR_RETURN_NONE))
    {
        return CSSMERR_CSP_INVALID_KEYATTR_MASK;
    }

    if ((nPrivateKeyAttr & CSSM_KEYATTR_RETURN_DATA) || 
        (nPrivateKeyAttr & CSSM_KEYATTR_RETURN_NONE) ||
        (nPrivateKeyAttr & CSSM_KEYATTR_PERMANENT))
    {
        return CSSMERR_CSP_INVALID_KEYATTR_MASK;
    }

    /* Insure that the private key usage allows deriving */
    if ((!(nPrivateKeyUsage & CSSM_KEYUSE_ANY)) &&
        (!(nPrivateKeyUsage & CSSM_KEYUSE_DERIVE)))
    {
        return CSSMERR_CSP_INVALID_KEYUSAGE_MASK;
    }

    /* Fetch the parameter block */
    pAttribute = CSM_GetContextAttribute( pSession->pContext,
                                          CSSM_ATTRIBUTE_ALG_PARAMS );
    if ( pAttribute == NULL )
    {
        return CSSMERR_CSP_INVALID_ATTR_ALG_PARAMS;
    }

    if ((pPubKey == NULL) || (pPriKey == NULL))
    {
        return CSSMERR_CSP_INVALID_KEY_POINTER;
    }

    pAlgParams = pAttribute->Attribute.Data;
    if ((pAlgParams->Data == NULL) || (pAlgParams->Length == 0))
    {
        return CSSMERR_CSP_INVALID_ATTR_ALG_PARAMS;
    }

#if BSAFE_VERSION > 3
    if ((rtn = csp_InitRandom (CSSM_ALGID_SHARandom, &randomNumGen)) != CSSM_OK)
#else
    if ((rtn = csp_InitRandom (CSSM_ALGID_MD5Random, &randomNumGen)) != CSSM_OK)
#endif
    {
        return rtn;
    }

    /* Create the algorithm object */
    if ((status = B_CreateAlgorithmObject(&dhKeyAgreeAlg)) != 0)
    {
        rtn = ConvertBsafeError(status);
        goto FAIL;
    }

    /* Set the parameters in the algorithm object */
    
    dhParametersBER.data = pAlgParams->Data;
    dhParametersBER.len = pAlgParams->Length;

    if ((status = B_SetAlgorithmInfo(dhKeyAgreeAlg, AI_DHKeyAgreeBER,
                      (POINTER)&dhParametersBER)) != 0)
    {
        rtn = ConvertBsafeError(status);
        goto FAIL;
    }

    if ((status = B_KeyAgreeInit(dhKeyAgreeAlg, (B_KEY_OBJ)NULL_PTR, 
                    DH_KEY_GEN_CHOOSER, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
    {
        rtn = ConvertBsafeError(status);
        goto FAIL;
    }

    /* Determine how much space to allocate for the public value */
    if ((status = B_GetAlgorithmInfo((POINTER *)&getParams, 
                dhKeyAgreeAlg, AI_DHKeyAgree)) != 0)
    {
        rtn = ConvertBsafeError(status);
        goto FAIL;
    }

    /* Make sure the output key buffer is there and has enough space */
    if ((rtn = MAF_CreateOutputBuffer(pSession->hCsp, getParams->prime.len,
                            &(pPubKey->KeyData), &bPubAllocOutput)) != CSSM_OK)
    {
        goto FAIL;
    }

    if ((status = B_KeyAgreePhase1(dhKeyAgreeAlg, pPubKey->KeyData.Data, 
                    (unsigned int *)&(pPubKey->KeyData.Length), getParams->prime.len,  /* ifdef VMS cast */
                    randomNumGen, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
    {
        rtn = ConvertBsafeError(status);
        goto FAIL;
    }

    /* Save the algorithm object for the derive key operation and pass back
     * a reference in the private key return parameter. */
    tempPriKey.KeyData.Length = sizeof(B_ALGORITHM_OBJ);
    tempPriKey.KeyData.Data = Addin_calloc(1, sizeof(B_ALGORITHM_OBJ), NULL);
    if (tempPriKey.KeyData.Data == NULL)
    {
        rtn = CSSMERR_CSP_MEMORY_ERROR;
        goto FAIL;
    }

    memcpy(tempPriKey.KeyData.Data, &dhKeyAgreeAlg, sizeof(B_ALGORITHM_OBJ));

    /* Set the key headers */
    csp_KeyHeader_Fill(&(pPubKey->KeyHeader),
                       CSSM_KEYBLOB_RAW,
                       CSSM_KEYBLOB_RAW_FORMAT_NONE,
                       CSSM_ALGID_DH,
                       CSSM_KEYCLASS_PUBLIC_KEY,
                       (getParams->prime.len * 8),
                       nPublicKeyAttr,
                       nPublicKeyUsage,
                       NULL,
                       NULL,
                       0,
                       0,
                       0);


    csp_KeyHeader_Fill(&(tempPriKey.KeyHeader),
                       CSSM_KEYBLOB_RAW,
                       CSSM_KEYBLOB_RAW_FORMAT_OTHER,
                       CSSM_ALGID_DH,
                       CSSM_KEYCLASS_PRIVATE_KEY,
                       (getParams->prime.len * 8),
                       nPrivateKeyAttr,
                       nPrivateKeyUsage,
                       NULL,
                       NULL,
                       0,
                       0,
                       0);

    /* Place the temporary key in the key manager as a memory only key */
    loadData = Addin_GetLoadData();

    if (loadData == NULL)
    {
        rtn = CSSMERR_CSP_INTERNAL_ERROR;
        goto FAIL;
    }

    rtn = CSM_KMInsertKey(loadData->hKeyFile,
                          &tempPriKey,
                          pPubKey,
                          NULL,
                          CredAndAclEntry,
                          &keyHandle);
    if (rtn != CSSM_OK)
    {
        goto FAIL;
    }

    if ((rtn = MAF_CreateOutputBuffer(pSession->hCsp, sizeof(CSM_KM_KEY_HANDLE),
                            &(pPriKey->KeyData), &bPriAllocOutput)) != CSSM_OK)
    {
        goto FAIL;
    }
    pPriKey->KeyData.Length = sizeof(CSM_KM_KEY_HANDLE);
    memcpy(pPriKey->KeyData.Data, &keyHandle, sizeof(CSM_KM_KEY_HANDLE));

    memcpy(&(pPriKey->KeyHeader), &(tempPriKey.KeyHeader), sizeof(CSSM_KEYHEADER));

    /* Correct the format settings in the key header */
    pPriKey->KeyHeader.BlobType = CSSM_KEYBLOB_REFERENCE;
    pPriKey->KeyHeader.Format = CSSM_KEYBLOB_REF_FORMAT_INTEGER;

    B_DestroyAlgorithmObject(&randomNumGen);

    if (tempPriKey.KeyData.Data)
    {
        Addin_free(tempPriKey.KeyData.Data, NULL);
    }

    return rtn;

FAIL:
    MAF_DeleteOutputBuffer(pSession->hCsp, &(pPubKey->KeyData), bPubAllocOutput);
    MAF_DeleteOutputBuffer(pSession->hCsp, &(pPriKey->KeyData), bPriAllocOutput);

    if (randomNumGen)
        B_DestroyAlgorithmObject(&randomNumGen);
    if (dhKeyAgreeAlg)
        B_DestroyAlgorithmObject(&dhKeyAgreeAlg);

    if (tempPriKey.KeyData.Data)
        Addin_free(tempPriKey.KeyData.Data, NULL);

    return rtn;
}
