/*-----------------------------------------------------------------------
 * File: CSP_VRFY.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */
/*
** OpenVMS History
**
 *    01-Nov-1999	Carol Kelly
 *		o Merge in changes from Unix code base
 *
 */
 
/*
 * File Description: Contains signature verification operations for the CSP.
 */

#include "geninc.h"

/* BSafe Signing Algorithm Choosers */
B_ALGORITHM_METHOD  *VERIFY_DSA_CHOOSER[] = {
    &AM_SHA,
    &AM_DSA_VERIFY,
    (B_ALGORITHM_METHOD*)NULL_PTR
};
B_ALGORITHM_METHOD  *VERIFY_MD5WithRSA_CHOOSER[] = {
    &AM_MD5,
    &AM_RSA_DECRYPT,
    (B_ALGORITHM_METHOD*)NULL_PTR
};
B_ALGORITHM_METHOD  *VERIFY_MD2WithRSA_CHOOSER[] = {
    &AM_MD2,
    &AM_RSA_DECRYPT,
    (B_ALGORITHM_METHOD*)NULL_PTR
};
B_ALGORITHM_METHOD  *VERIFY_SHA1WithRSA_CHOOSER[] = {
    &AM_SHA,
    &AM_RSA_DECRYPT,
    (B_ALGORITHM_METHOD*)NULL_PTR
};

extern CSSM_RETURN bsafe_ValidateSignVrfyAlg (CSSM_ALGORITHMS Alg);

static CSSM_RETURN __ValidateSignature(
						CSSM_ALGORITHMS AlgId,
						const CSSM_DATA *pSig )
{
	CSSM_RETURN rv = CSSM_OK;
	const uint8 *pSigData = NULL;
    BER_PARSED_ITEM BerData[4];
    sint32 SetItemCount;
	sint32 i;

	if ( !pSig ||
		 ( pSig->Length == 0 ) )
	{
		ERR( rv = CSSMERR_CSP_INVALID_SIGNATURE );
	}
	else
	{
		pSigData = pSig->Data;

		switch ( AlgId )
		{
		case CSSM_ALGID_MD5WithRSA:
		case CSSM_ALGID_MD2WithRSA:
		case CSSM_ALGID_SHA1WithRSA:
			{
				break;
			}
		case CSSM_ALGID_DSA:
		case CSSM_ALGID_SHA1WithDSA:
			{
				/* Do a quick expansion of the signature to make sure it's
				 * the right general format.
				 */
				pSigData = BER_ExpandItem( pSigData, pSig->Length, BerData );
				if ( ( pSigData != NULL) &&
					 ( *BerData[0].Tag == BER_CONSTRUCTED_SEQUENCE ) )
				{
					SetItemCount = BER_ExpandSet( BerData[0].Content,
												  BerData[0].ContentLength,
												  ( sizeof(BerData) / 
													sizeof(BER_PARSED_ITEM) ),
												  0,
												  NULL,
												  BerData );
					if ( SetItemCount != 2 )
					{
						ERR( rv = CSSMERR_CSP_INVALID_SIGNATURE );

						for ( i = 0; i < SetItemCount; i++ )
						{
							/* Check to make sure the lengths are reasonable */
							if ( ( BerData[i].ContentLength != 20 ) &&
								 ( ( BerData[i].ContentLength != 21 ) &&
								   ( BerData[i].Content != 0 ) ) )
							{
								ERR( rv = CSSMERR_CSP_INVALID_SIGNATURE );
							}
						}
					}
				}
				else
				{
					ERR( rv = CSSMERR_CSP_INVALID_SIGNATURE );
				}
				break;
			}
		default:
			{
				ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
			}
		}
	}

	return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_VerifyData( CSM_SESSION *pSession,
                                          const CSSM_DATA *DataBufs,
                                          uint32 DataBufCount,
                                          CSSM_ALGORITHMS DigestAlgorithm,
                                          const CSSM_DATA *Signature )
{
    CSSM_RETURN  rtn = CSSM_OK;

#ifdef VMS /* COMPAQ_FIX */
    const CSSM_CONTEXT_ATTRIBUTE *pKeyAttr = NULL;
    const CSSM_CONTEXT_ATTRIBUTE *pModeAttr = NULL;
    uint32                       decrypt_only = 0;
    CSSM_DATA                    ClearBufs;
    CSSM_DATA                    remData;
    CSSM_DATA                    CipherBufs;
    uint32                       bytesDecrypted;

    /*
     * Check for special case. If no digest is need, the signature algorithm
     * is RSA and the mode attribute is NONE just encrypt the data.
     */
    if (DigestAlgorithm == CSSM_ALGID_NONE)
    {
      if ((pSession->pContext->AlgorithmType == CSSM_ALGID_RSA))
      {
	pModeAttr = CSM_GetContextAttribute(pSession->pContext,
					    CSSM_ATTRIBUTE_MODE);
	if (pModeAttr != NULL &&
	    pModeAttr->Attribute.Uint32 == CSSM_ALGMODE_NONE)
        {
	  decrypt_only = 1;
        }
      }
    }

    if (decrypt_only == 1)
    {
      /* Only one buffer allowed for encryption only */
      if (DataBufCount != 1)
      {
	rtn = CSSMERR_CSP_VECTOR_OF_BUFS_UNSUPPORTED;
      }
      else
      {
	ClearBufs.Data = NULL;
        ClearBufs.Length = 0;
        CipherBufs.Data = Signature->Data;
        CipherBufs.Length = Signature->Length;
        remData.Data = NULL;
        remData.Length = 0;
        rtn = ADDIN_CSM_DecryptData(pSession,
                                    &CipherBufs,
                                    1,
                                    &ClearBufs,
                                    1,
                                    &bytesDecrypted,
                                    &remData,
                                    (unsigned __int64)NULL);
        if (rtn == CSSM_OK)
        {
          if(remData.Length !=0)
          {
            /* We should not have a remainder */
            rtn = CSSMERR_CSP_OUTPUT_LENGTH_ERROR;
 
            /* Free any allocated buffer */
            if (remData.Data != NULL)
            {
              App_Free(pSession->hCsp, remData.Data);
            }
          }
          else
          {
            /* Compare the decrypted data with the data passed in */
            if (ClearBufs.Length != DataBufs->Length ||
                memcmp(ClearBufs.Data,
                       DataBufs->Data,
                       ClearBufs.Length))
            {
              rtn = CSSMERR_CSP_VERIFY_FAILED;
            }
          }
 
          /* Free any allocated buffer */
          if (ClearBufs.Data != NULL)
          {
            App_Free(pSession->hCsp, ClearBufs.Data);
          }
        }
      }
    }
    else
    {
#endif
    rtn = ADDIN_CSM_VerifyDataInit(pSession);

    if (rtn == CSSM_OK)
    {
        rtn = ADDIN_CSM_VerifyDataUpdate(pSession, DataBufs, DataBufCount);
        if (rtn == CSSM_OK)
        {
            rtn = ADDIN_CSM_VerifyDataFinal(pSession, Signature);
        }
    }

    CleanState( (BSAFECSP_STATE_PTR)pSession->pCryptoState );

#ifdef VMS /* COMPAQ_FIX */
    }
#endif
    
    return rtn;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_VerifyDataInit( CSM_SESSION *pSession )
{
    int                             status;
    uint8                           pubkeybuf[MAX_KEYBUF_LEN];
    const CSSM_CONTEXT_ATTRIBUTE    *pKeyAttr = NULL;
    CSSM_KEY_PTR                    pKey;
    ITEM                            pubKeyBER;
    BSAFECSP_STATE_PTR              pState = NULL;
    CSSM_RETURN                     rtn = CSSM_OK;

    pubKeyBER.data = pubkeybuf;
    pubKeyBER.len = sizeof(pubkeybuf);

    /* Get the raw key from the context */
    pKeyAttr = CSM_GetContextAttribute(pSession->pContext, CSSM_ATTRIBUTE_KEY);
    if ((pKeyAttr == NULL) || (pKeyAttr->Attribute.Key == NULL))
    {
        rtn = CSSMERR_CSP_MISSING_ATTR_KEY;
        goto FAIL;
    }

    pKey = pKeyAttr->Attribute.Key;

	if ( pKey == NULL )
    {
        rtn = CSSMERR_CSP_INVALID_ATTR_KEY;
        goto FAIL;
    }
    else
    {
		if ((pKey->KeyData.Data == NULL) || (pKey->KeyData.Length == 0))
		{
			rtn = CSSMERR_CSP_INVALID_KEY;
			goto FAIL;
		}
	}

    /* Verify the correct key was passed in */
    if (pKey->KeyHeader.KeyClass != CSSM_KEYCLASS_PUBLIC_KEY)
    {
        rtn = CSSMERR_CSP_INVALID_KEY_CLASS;
        goto FAIL;
    }

	if ((rtn = bsafe_ValidateSignVrfyAlg (pSession->pContext->AlgorithmType)) != CSSM_OK)
	{                
		rtn = CSSMERR_CSP_INVALID_ALGORITHM;
		goto FAIL;
	}

    if (pKey->KeyHeader.AlgorithmId == CSSM_ALGID_RSA)
    {
        if ((pSession->pContext->AlgorithmType != CSSM_ALGID_MD2WithRSA) &&
            (pSession->pContext->AlgorithmType != CSSM_ALGID_MD5WithRSA) &&
            (pSession->pContext->AlgorithmType != CSSM_ALGID_SHA1WithRSA))
        {
            rtn = CSSMERR_CSP_ALGID_MISMATCH;
            goto FAIL;
        }
    }
    else if (pKey->KeyHeader.AlgorithmId == CSSM_ALGID_DSA)
    {
        if (pSession->pContext->AlgorithmType != CSSM_ALGID_SHA1WithDSA)
        {
            rtn = CSSMERR_CSP_ALGID_MISMATCH;
            goto FAIL;
        }
    }

	if (!CHECKBITS(pKey->KeyHeader.KeyUsage,
							  CSSM_KEYUSE_VERIFY ) &&
				  !CHECKBITS(pKey->KeyHeader.KeyUsage,
						  CSSM_KEYUSE_ANY ) )
	{                
		rtn = CSSMERR_CSP_KEY_USAGE_INCORRECT;
		goto FAIL;
	}


    /* Allocate space to store the algorithm map pointer */
    rtn = CSM_AllocateSessionState( pSession, sizeof(BSAFECSP_STATE) );

    if ( rtn != CSSM_OK )
    {
        rtn = CSSMERR_CSP_INTERNAL_ERROR;
    }
    else
    {
        pState = (BSAFECSP_STATE_PTR)pSession->pCryptoState;
        memset(pState, 0, sizeof(BSAFECSP_STATE));

        /* Create the verify object */
        if ((status = B_CreateAlgorithmObject(&pState->algObject)) != 0)
        {
            rtn = ConvertBsafeError(status);
        }

        if (rtn == CSSM_OK)
        {
            if ((status = B_CreateKeyObject(&pState->keyObject)) != 0)
            {
                rtn = ConvertBsafeError(status);
            }
        }

        if (rtn == CSSM_OK)
        {
            /* Set and Initialize the signing algorithm */
            switch (pSession->pContext->AlgorithmType) {
            case CSSM_ALGID_SHA1WithDSA:
                /* Convert the key to BSafe Key Blob format */
                if (CSSM_OK != Cryp_CSSMDSAPublicKey_to_BSafeDSAPublicKey(pKey,
                                    pubKeyBER.data, (uint32 *)&pubKeyBER.len))
                {
                    rtn = CSSMERR_CSP_INVALID_KEY;
                    goto FAIL;
                }

                if ((status = B_SetKeyInfo(pState->keyObject, KI_DSAPublicBER,
                                    (POINTER)&pubKeyBER)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                if ((status = B_SetAlgorithmInfo(pState->algObject,
                                    AI_DSAWithSHA1, NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                if ((status = B_VerifyInit(pState->algObject, 
                                    pState->keyObject, VERIFY_DSA_CHOOSER,
                                    (A_SURRENDER_CTX*)NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                break;
            case CSSM_ALGID_MD2WithRSA:
                /* Convert the key to BSafe Key Blob format */
                if (CSSM_OK != Cryp_CSSMRSAPublicKey_to_BSafeRSAPublicKey(pKey,
                                    pubKeyBER.data, (uint32 *)&pubKeyBER.len))
                {
                    rtn = CSSMERR_CSP_INVALID_KEY;
                    goto FAIL;
                }

                if ((status = B_SetKeyInfo(pState->keyObject, KI_RSAPublicBER,
                                    (POINTER)&pubKeyBER)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                if ((status = B_SetAlgorithmInfo(pState->algObject,
                                    AI_MD2WithRSAEncryption, NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                if ((status = B_VerifyInit(pState->algObject, 
                                    pState->keyObject,
                                    VERIFY_MD2WithRSA_CHOOSER,
                                    (A_SURRENDER_CTX*)NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                break;
            case CSSM_ALGID_MD5WithRSA:
                /* Convert the key to BSafe Key Blob format */
                if (CSSM_OK != Cryp_CSSMRSAPublicKey_to_BSafeRSAPublicKey(pKey,
                                    pubKeyBER.data, (uint32 *)&pubKeyBER.len))
                {
                    rtn = CSSMERR_CSP_INVALID_KEY;
                    goto FAIL;
                }

                if ((status = B_SetKeyInfo(pState->keyObject, KI_RSAPublicBER,
                                    (POINTER)&pubKeyBER)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                if ((status = B_SetAlgorithmInfo(pState->algObject,
                                    AI_MD5WithRSAEncryption, NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                if ((status = B_VerifyInit(pState->algObject,
                                    pState->keyObject,
                                    VERIFY_MD5WithRSA_CHOOSER,
                                    (A_SURRENDER_CTX*)NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                break;
            case CSSM_ALGID_SHA1WithRSA:
                /* Convert the key to BSafe Key Blob format */
                if (CSSM_OK != Cryp_CSSMRSAPublicKey_to_BSafeRSAPublicKey(pKey,
                                    pubKeyBER.data, (uint32 *)&pubKeyBER.len))
                {
                    rtn = CSSMERR_CSP_INVALID_KEY;
                    goto FAIL;
                }

                if ((status = B_SetKeyInfo(pState->keyObject, KI_RSAPublicBER,
                                    (POINTER)&pubKeyBER)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                if ((status = B_SetAlgorithmInfo(pState->algObject,
                                    AI_SHA1WithRSAEncryption, NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                if ((status = B_VerifyInit(pState->algObject, 
                                    pState->keyObject,
                                    VERIFY_SHA1WithRSA_CHOOSER,
                                    (A_SURRENDER_CTX*)NULL_PTR)) != 0)
                {
                    rtn = ConvertBsafeError(status);
                    goto FAIL;
                }
                break;
            default:
                rtn = CSSMERR_CSP_INVALID_ALGORITHM;
                goto FAIL;
            }
        }
    }

    return rtn;

FAIL:
    CleanState(pState);
    return rtn;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_VerifyDataUpdate( CSM_SESSION *pSession,
                                                const CSSM_DATA *DataBufs,
                                                uint32 DataBufCount )
{
    int                 status;
    uint32              i = 0;
    BSAFECSP_STATE_PTR  pState = NULL;
    CSSM_RETURN         rtn = CSSM_OK;

    /* Get the digital verifier object from the algorithm state */
    pState = (BSAFECSP_STATE_PTR)pSession->pCryptoState;
    if (!pState)
    {
        rtn = CSSMERR_CSP_INTERNAL_ERROR;
    }
    else
    {
        /* Update */
        for (i = 0; i < DataBufCount; i++)
        {
            if ((status = B_VerifyUpdate(pState->algObject, DataBufs[i].Data,
                                DataBufs[i].Length,
                                (A_SURRENDER_CTX*)NULL_PTR)) != 0)
            {
                rtn = ConvertBsafeError(status);
                CleanState(pState);
            }
        }
    }

    return rtn;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_VerifyDataFinal( CSM_SESSION *pSession,
                                               const CSSM_DATA *Signature )
{
    int                 status;
    BSAFECSP_STATE_PTR  pState = NULL;
    CSSM_RETURN         rtn = CSSM_OK;

    /* Get the digital verifier object from the algorithm state */
    pState = (BSAFECSP_STATE_PTR)pSession->pCryptoState;
    if (!pState)
    {
        rtn = CSSMERR_CSP_INTERNAL_ERROR;
    }
    else
    {
        if ((Signature == NULL) || (Signature->Length == 0) || 
            (Signature->Data == NULL))
        {
            rtn = CSSMERR_CSP_INVALID_SIGNATURE;
        }
        else
        {
			if ( ( rtn = __ValidateSignature( pSession->pContext->AlgorithmType,
											 Signature ) ) == CSSM_OK )
			{
				/* Final verify */
				if ((status = B_VerifyFinal(pState->algObject, Signature->Data,
									Signature->Length, (B_ALGORITHM_OBJ)NULL_PTR,
									(A_SURRENDER_CTX*)NULL_PTR)) != 0)
				{
					rtn = ConvertBsafeError(status);
				}
			}
        }
    }

    /* Destroy the signature object */
    CleanState(pState);
    return rtn;
}

void ADDIN_CSM_VerifyDataError( CSM_SESSION *pSession )
{
    BSAFECSP_STATE_PTR  pState = NULL;

    /* Get the pState->algObject object from the algorithm state */
    pState = (BSAFECSP_STATE_PTR)pSession->pCryptoState;
    if (pState)
    {
        CleanState(pState);
    }
}
