/*-----------------------------------------------------------------------
 * File: KEYCONV.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
#include "geninc.h"
#include "ber_der.h"
#include <cssmMemTrack.h>

// Converts BSafe Key Blobs (RSA and DSA) to CSSM Key Blobs and vice versa


/******************************************************************************/
/******************************** RSA Key Blobs *******************************/
/******************************************************************************/

/*---------------------------------------------------------------
 *Name: BSafe_GetRSAPublicKeyData
 *
 *Description:
 * This function strips the Bsafe header from a Bsafe RSA public key blob.
 *
 *Parameters: 
 * pBSafePubKeyBlob(input)  - Bsafe public key blob
 *
 *Returns:
 *-------------------------------------------------------------*/
static uint8* BSafe_GetRSAPublicKeyData(const uint8 *pBSafePubKeyBlob)
{
     BER_PARSED_ITEM item;
     const uint8 *position;
     uint32 length;

     length = BER_SizeofObject((uint8*)pBSafePubKeyBlob);
     position = BER_ExpandItem((uint8*)pBSafePubKeyBlob, length, &item);
     if (!position)
          return 0;
     length = BER_SizeofObject(item.Content);     /* skip algID and parameters */
     position = item.Content + length;
     position = BER_ExpandItem(position, item.ContentLength - length, &item);
     if (!position)
          return 0;
     if (item.Tag[0] != BER_BIT_STRING)
          return 0;
     return (uint8 *) (item.Content + 1);     /* skip the leading zero byte for bit string */
}

/*---------------------------------------------------------------
 *Name: BSafe_GetRSAPrivateKeyData
 *
 *Description:
 * This function strips the Bsafe header from a Bsafe RSA private key blob.
 *
 *Parameters: 
 * pBSafePriKeyBlob(input)  - Bsafe private key blob
 *
 *Returns:
 *-------------------------------------------------------------*/
static uint8* BSafe_GetRSAPrivateKeyData(const uint8 *pBSafePriKeyBlob)
{
     BER_PARSED_ITEM item;
     const uint8 *position;
     uint32 length;

     length = BER_SizeofObject((uint8*)pBSafePriKeyBlob);
     position = BER_ExpandItem((uint8*)pBSafePriKeyBlob, length, &item);
     if (!position)
          return 0;
     length = BER_SizeofObject(item.Content);     /* skip version in key header */
     position = item.Content + length;
     length = BER_SizeofObject(position);          /* skip algID and parameters */
     position += length;
     position = BER_ExpandItem(position, BER_SizeofObject(position), &item);
     if (!position)
         return 0;
     return (uint8 *) item.Content;
}

/*---------------------------------------------------------------
 *Name: BSafe_RSAKeyDataLength
 *
 *Description:
 * This function retrieves length of Bsafe key blob (BER-encoded public or private key).
 *
 *Parameters: 
 * pBSafeKeyData(input)  - Bsafe public or private key blob
 *
 *Returns:
 *-------------------------------------------------------------*/
static uint32 BSafe_RSAKeyDataLength(const uint8 *pBSafeKeyData)
{
    if (pBSafeKeyData == NULL)
        return  0;
    return BER_SizeofObject((uint8*)pBSafeKeyData);
}

CSSM_RETURN
Cryp_BSafeRSAPublicKey_to_CSSMRSAPublicKey(
    uint32                  nModulus,
    uint32                  nPublicKeyUsage,
    uint8                   *pBSafePubKey,
    CSSM_KEY_PTR            pCssmPubKey)
{
    CSSM_KEYHEADER  *pCssmPubKeyHdr=NULL;
    uint8           *pCssmPubKeyData=NULL;
    uint8           *pBSFPubKeyData=NULL;
    uint32          nBSFPubKeyDataLen=0;

    pBSFPubKeyData = BSafe_GetRSAPublicKeyData((uint8*)pBSafePubKey);
    if (pBSFPubKeyData == NULL)
        return CSSMERR_CSP_INVALID_KEY;
    nBSFPubKeyDataLen = BSafe_RSAKeyDataLength(pBSFPubKeyData);

    // CSSM public key (header and data)
    pCssmPubKeyHdr  = &pCssmPubKey->KeyHeader;
    pCssmPubKeyData = pCssmPubKey->KeyData.Data;

    // Intializes key headers
    {
        uint32  nBlobType=0, nKeyFormat=0, nKeyClass=0, nKeySizeInBits=0,
                nKeyAttr=0;

        /* public key */    
        nBlobType = CSSM_KEYBLOB_RAW;
        nKeyFormat= CSSM_KEYBLOB_RAW_FORMAT_PKCS1;
        nKeyClass = CSSM_KEYCLASS_PUBLIC_KEY;
        nKeySizeInBits = nModulus;
        nKeyAttr = CSSM_KEYATTR_EXTRACTABLE;
        
        csp_KeyHeader_Fill(pCssmPubKeyHdr, nBlobType, nKeyFormat, CSSM_ALGID_RSA,
                            nKeyClass, nKeySizeInBits, nKeyAttr,
                            nPublicKeyUsage, NULL/*&StartDate*/, NULL/*&EndDate*/, 
                            CSSM_ALGID_NONE, 0, 0);
    }

    memcpy(pCssmPubKeyData, pBSFPubKeyData, nBSFPubKeyDataLen);
    pCssmPubKey->KeyData.Length = nBSFPubKeyDataLen;

    return CSSM_OK;
}

CSSM_RETURN
Cryp_BSafeRSAPrivateKey_to_CSSMRSAPrivateKey(
    uint32                  nModulus,
    uint32                  nPrivateKeyUsage,
    uint8                   *pBSafePriKey,
    CSSM_KEY_PTR            pCssmPriKey)
{
    CSSM_KEYHEADER  *pCssmPriKeyHdr=NULL;
    uint8           *pCssmPriKeyData=NULL;
    uint8           *pBSFPriKeyData=NULL;
    uint32          nBSFPriKeyDataLen=0;

    pBSFPriKeyData = BSafe_GetRSAPrivateKeyData((uint8*)pBSafePriKey);
    if (pBSFPriKeyData == NULL)
        return CSSMERR_CSP_INVALID_KEY;
    nBSFPriKeyDataLen = BSafe_RSAKeyDataLength(pBSFPriKeyData);

    // CSSM private key (header and data)
    pCssmPriKeyHdr  = &pCssmPriKey->KeyHeader;
    pCssmPriKeyData = pCssmPriKey->KeyData.Data;

    // Intializes key headers
    {
        uint32  nBlobType=0, nKeyFormat=0, nKeyClass=0, nKeySizeInBits=0,
                nKeyAttr=0;

        /* private key */
        nBlobType = CSSM_KEYBLOB_RAW;
        nKeyFormat= CSSM_KEYBLOB_RAW_FORMAT_PKCS1;
        nKeyClass = CSSM_KEYCLASS_PRIVATE_KEY;
        nKeySizeInBits = nModulus;
        nKeyAttr = CSSM_KEYATTR_SENSITIVE|CSSM_KEYATTR_EXTRACTABLE|CSSM_KEYATTR_PERMANENT;
        
        csp_KeyHeader_Fill(pCssmPriKeyHdr, nBlobType, nKeyFormat, CSSM_ALGID_RSA,
                            nKeyClass, nKeySizeInBits, nKeyAttr,
                            nPrivateKeyUsage, NULL/*&StartDate*/, NULL/*&EndDate*/, 
                            CSSM_ALGID_NONE, 0, 0);
    }

    memcpy(pCssmPriKeyData, pBSFPriKeyData, nBSFPriKeyDataLen);
    pCssmPriKey->KeyData.Length = nBSFPriKeyDataLen;

    return CSSM_OK;
}

CSSM_RETURN
Cryp_CSSMRSAPublicKey_to_BSafeRSAPublicKey(
    CSSM_KEY_PTR            pCssmPubKey,
    uint8                   *pBSafePubKey,
    uint32                  *pnBSafePubLen)
{
    sint32 nPubDataLen=0, length=0;
    uint8 pkcs1rsaalgid[] =
            {
            0x06, 0x09, 0x2A, 0x86, 0x48, 0x86, 0xF7, 0x0D, 0x01, 0x01, 0x01,  /* PKCS1 RSA OID */
            0x05, 0x00  /* BER NULL parameters */
            };
    uint8 constructedSequence[] = { BER_CONSTRUCTED | BER_SEQUENCE };
    uint8 bitString[] = { BER_BIT_STRING };
    BER_PARSED_ITEM publicItems[2];
    uint8 keyBuffer[MAX_KEYBUF_LEN];
    uint8 *pCssmPubKeyData=NULL;

    if ((pCssmPubKey == NULL) || (pBSafePubKey == NULL) || (pnBSafePubLen == NULL))
        return CSSMERR_CSP_INTERNAL_ERROR;

    nPubDataLen = pCssmPubKey->KeyData.Length;
    if (nPubDataLen <= 0)
        return CSSMERR_CSP_INTERNAL_ERROR;

    pCssmPubKeyData = pCssmPubKey->KeyData.Data;

    /* reconstruct the original public key from the key block */
    /* append leading zero byte to public key block */
    keyBuffer[0] = 0;
    length =  BER_SizeofObject(pCssmPubKeyData);
    memcpy(keyBuffer + 1, pCssmPubKeyData, length);
    publicItems[0].Tag = constructedSequence;
    publicItems[0].ContentLength = sizeof(pkcs1rsaalgid);
    publicItems[0].Content = pkcs1rsaalgid;
    publicItems[1].Tag = bitString;
    publicItems[1].ContentLength = length + 1; /* we added extra leading byte */
    publicItems[1].Content = keyBuffer;
    *pnBSafePubLen = DER_ConstructItem(constructedSequence, 2, publicItems, pBSafePubKey);

    return CSSM_OK; 
}

CSSM_RETURN
Cryp_CSSMRSAPrivateKey_to_BSafeRSAPrivateKey(
    CSSM_KEY_PTR            pCssmPriKey,
    uint8                   *pBSafePriKey,
    uint32                  *pnBSafePriLen)
{
    sint32 nPriDataLen=0;
    uint8 pkcs1rsaalgid[] =
            {
            0x06, 0x09, 0x2A, 0x86, 0x48, 0x86, 0xF7, 0x0D, 0x01, 0x01, 0x01,  /* PKCS1 RSA OID */
            0x05, 0x00  /* BER NULL parameters */
            };
    uint8 zero[] = { 0 };
    uint8 constructedSequence[] = { BER_CONSTRUCTED | BER_SEQUENCE };
    uint8 integer[] = { BER_INTEGER };
    uint8 octetString[] = { BER_OCTET_STRING };
    BER_PARSED_ITEM privateItems[3];
    uint8 *pCssmPriKeyData=NULL;

    if ((pCssmPriKey == NULL) || (pBSafePriKey == NULL) || (pnBSafePriLen == NULL))
        return CSSMERR_CSP_INTERNAL_ERROR;

    nPriDataLen = pCssmPriKey->KeyData.Length;
    if (nPriDataLen <= 0)
        return CSSMERR_CSP_INTERNAL_ERROR;

    pCssmPriKeyData = pCssmPriKey->KeyData.Data;
    
    /* reconstruct the original private key from the key block */
    privateItems[0].Tag = integer;
    privateItems[0].ContentLength = sizeof(zero);
    privateItems[0].Content = zero;
    privateItems[1].Tag = constructedSequence;
    privateItems[1].ContentLength = sizeof(pkcs1rsaalgid);
    privateItems[1].Content = pkcs1rsaalgid;
    privateItems[2].Tag = octetString;
    privateItems[2].ContentLength = BER_SizeofObject(pCssmPriKeyData);
    if ( privateItems[2].ContentLength > pCssmPriKey->KeyData.Length )
    {
        return CSSMERR_CSP_INVALID_KEY;
    }
    privateItems[2].Content = pCssmPriKeyData;
    *pnBSafePriLen = DER_ConstructItem(constructedSequence, 3, privateItems, pBSafePriKey);

    return CSSM_OK;
}


/******************************************************************************/
/******************************** DSA Key Blobs *******************************/
/******************************************************************************/

/* 
 * allocate space for a DER_node with N children
 */
DER_NODE_PTR 
DER_AllocNode(int NumberOfChildren)
{
    DER_NODE_PTR result;
    result = (DER_NODE_PTR) local_malloc(sizeof(DER_NODE), NULL);
    if (result) {
        memset(result, 0, sizeof(DER_NODE));
        result->Count = (sint16)NumberOfChildren;   /* we just wiped it out--store again */
        result->Child = (struct der_node_child_struct *) 
                local_malloc(sizeof(struct der_node_child_struct) * NumberOfChildren, NULL);
        if (result->Child) {
            memset(result->Child, 0, sizeof(struct der_node_child_struct) * NumberOfChildren);
        } else {
            local_free (result, NULL);
            result = NULL;
        }
    }

    return result;
}

/* recycle parse tree */
void
DER_FreeParseTree(DER_NODE *Tree)
{
    int i;
    if (!Tree) 
        return;
    for (i = 0; i < Tree->Count; i++) {
        if (!Tree->Child[i].IsLeaf && Tree->Child[i].X.Node) 
            DER_FreeParseTree(Tree->Child[i].X.Node);
    }
    if (Tree->Child) local_free(Tree->Child, NULL);
    local_free(Tree, NULL);
}

/* simple untyped BER parsing -- returns parse tree or NULL, if error */
CSSM_RETURN
BER_ParseSimpleObject(BER_PARSED_ITEM *item, DER_NODE **pnode)
{
    DER_NODE *node = NULL;
    int i;
    const uint8 *position, *end;
    CSSM_RETURN retValue = CSSM_OK;

    *pnode = NULL;
    if ((item->Tag[0] & BER_CONSTRUCTED) == 0)
        return CSSMERR_CSP_INVALID_KEY; /* just a simple item */
    node = local_malloc(sizeof(DER_NODE), NULL);
    if (NULL == node)
        return CSSMERR_CSP_MEMORY_ERROR;
    node->Tag = node->OriginalTag = item->Tag;
    node->Type = 0;     /* no types (this a simple object) */
    node->Child = 0;    /* assume no children for now */
    node->Length = 0;   /* generally must be recalculated after changing */
    node->Count = BER_CountItems(item->Content, item->ContentLength);
    if (node->Count) {
        node->Child = local_malloc(sizeof(struct der_node_child_struct) * node->Count, NULL);
        if (NULL == node->Child) {
            local_free (node, NULL);
            return CSSMERR_CSP_MEMORY_ERROR;
        }
        position = item->Content;
        end = position + item->ContentLength;
        for (i = 0; i < node->Count; i++) {
            position = BER_ExpandItem(position, end - position, &node->Child[i].X.Input);
            if (!position) {
                retValue = CSSMERR_CSP_INVALID_KEY;
                break;  /* an error */
            }
            if ((node->Child[i].X.Input.Tag[0] & BER_CONSTRUCTED) != 0) {
                node->Child[i].IsLeaf = 0;
                retValue = BER_ParseSimpleObject(&node->Child[i].X.Input, &(node->Child[i].X.Node));
                if (CSSM_OK != retValue) {                  
                    break;  /* an error */
                }
            } else {
                node->Child[i].IsLeaf = 1;
                node->Child[i].X.Node = 0;
            }
        }
        if (i < node->Count) {  /* error cleanup */
            while (i > 0) 
                DER_FreeParseTree(node->Child[--i].X.Node);
            local_free(node->Child, NULL);
            local_free(node, NULL);         
            node = NULL;
        }
    }
    *pnode = node;
    return retValue;
}

static uint8 dsaVer512[] = { 2, 0 };
static uint8 dsaVer768[] = { 3, 0 };
static uint8 dsaVer1024[] = { 4, 0 };
static uint8 integerTag[] = { BER_INTEGER };
static uint8 oidTag[] = { BER_OBJECT_IDENTIFIER };
static uint8 bitstringTag[] = { BER_BIT_STRING };
static uint8 octetstringTag[] = { BER_OCTET_STRING };
static uint8 sequenceTag[] = { BER_CONSTRUCTED_SEQUENCE };

/* 
 *  convert a bsafe 3.0 DSA public key block to a CSSM DSA public key block 
 */
static CSSM_RETURN
BsafeToCSSMDSAPublicKeyBlock(
    uint8  *KeyBlock,                   /* pointer to bsafe keyblock for input */
    uint8 **CSSMDSAKey,                 /* pointer to CSSM DSA Key */
    uint32 *CSSMDSAKeyLength)           /* length of CSSM DSA Key */
{
    DER_NODE *root, *parameters, *newnode;
    BER_PARSED_ITEM item, Yitem;
    uint32 KeyBlockLength, contLength;
    int i;
    CSSM_RETURN retValue = CSSM_OK;

    /* in case of error, return NULL and zero lengths */
    *CSSMDSAKey = 0;
    *CSSMDSAKeyLength = 0;
    /* sample code to convert bsafe 3.0 dsa public key to X509 subjectPublicKeyInfo */
    KeyBlockLength = BER_SizeofObject((uint8*)KeyBlock);

    if (!BER_ExpandItem(KeyBlock, KeyBlockLength, &item)) 
        return CSSMERR_CSP_INVALID_KEY;

    retValue = BER_ParseSimpleObject(&item, &root); /* no fancy type */
    if (CSSM_OK != retValue)
        return retValue;

    /* Allocate new DER node */
    newnode = DER_AllocNode(4);
    if (NULL == newnode)
        return CSSMERR_CSP_MEMORY_ERROR;

    for (;;) {  /* loop body just for error cleanup */
        if (root->Count != 2)  /* seq (algid, keybits) */
            break;
        if (!root->Child[0].X.Node)
            break;
        /* find the parameters */
        parameters = root->Child[0].X.Node->Child[1].X.Node;
        if (!parameters) 
            break;
        if (parameters->Count != 4) /* bsafe 3.0 generates 3*256, p, q, g */
            break;

        newnode->Tag = newnode->OriginalTag = sequenceTag;

        newnode->Child[0].IsLeaf = 1;
        newnode->Child[0].X.Input.Tag = integerTag;
        contLength = parameters->Child[1].X.Input.ContentLength;
        newnode->Child[0].X.Input.Content = local_malloc(contLength, NULL);
        if (NULL == newnode->Child[0].X.Input.Content) {
            retValue = CSSMERR_CSP_MEMORY_ERROR;
            break;
        }
        memcpy((void *)newnode->Child[0].X.Input.Content, 
                    parameters->Child[1].X.Input.Content, contLength);
        newnode->Child[0].X.Input.ContentLength = contLength;

        newnode->Child[1].IsLeaf = 1;
        newnode->Child[1].X.Input.Tag = integerTag;
        contLength = parameters->Child[2].X.Input.ContentLength;
        newnode->Child[1].X.Input.Content = local_malloc(contLength, NULL);
        if (NULL == newnode->Child[1].X.Input.Content) {
            retValue = CSSMERR_CSP_MEMORY_ERROR;
            break;
        }
        memcpy((void *)newnode->Child[1].X.Input.Content, 
                    parameters->Child[2].X.Input.Content, contLength);
        newnode->Child[1].X.Input.ContentLength = contLength;

        newnode->Child[2].IsLeaf = 1;
        newnode->Child[2].X.Input.Tag = integerTag;
        contLength = parameters->Child[3].X.Input.ContentLength;
        newnode->Child[2].X.Input.Content = local_malloc(contLength, NULL);
        if (NULL == newnode->Child[1].X.Input.Content) {
            retValue = CSSMERR_CSP_MEMORY_ERROR;
            break;
        }
        memcpy((void *)newnode->Child[2].X.Input.Content, 
                    parameters->Child[3].X.Input.Content, contLength);
        newnode->Child[2].X.Input.ContentLength = contLength;

        if (!BER_ExpandItem(root->Child[1].X.Input.Content + 1, 
                            root->Child[1].X.Input.ContentLength - 1, &Yitem)) 
            break;

        newnode->Child[3].IsLeaf = 1;
        newnode->Child[3].X.Input.Tag = integerTag;
        newnode->Child[3].X.Input.Content = local_malloc(Yitem.ContentLength, NULL);
        if (NULL == newnode->Child[3].X.Input.Content) {
            retValue = CSSMERR_CSP_MEMORY_ERROR;
            break;
        }
        memcpy((void *)newnode->Child[3].X.Input.Content, Yitem.Content, Yitem.ContentLength);
        newnode->Child[3].X.Input.ContentLength = Yitem.ContentLength;

        /* flatten newnode which represents the CSSM DSA key */
        *CSSMDSAKeyLength = DER_SizeofTree(newnode);
        *CSSMDSAKey = local_malloc(*CSSMDSAKeyLength, NULL);
        if (!*CSSMDSAKey) {
            retValue = CSSMERR_CSP_MEMORY_ERROR;
            break;
        }
        *CSSMDSAKeyLength = DER_CopyTree(newnode, *CSSMDSAKey);  /* flatten  */
        break;
    }
    for (i = 0; i < 4; i++) {
        newnode->Child[i].X.Input.Tag = 0;
        if (newnode->Child[i].X.Input.Content) 
            local_free((void *)newnode->Child[i].X.Input.Content, NULL);
    }
    if (newnode) DER_FreeParseTree(newnode);
    DER_FreeParseTree(root);

    return retValue;
}

/* 
 *  convert a bsafe 3.0 DSA private key block to a CSSM DSA private key block 
 */
static CSSM_RETURN
BsafeToCSSMDSAPrivateKeyBlock(
    uint8  *KeyBlock,                   /* pointer to bsafe keyblock for input */
    uint8 **CSSMDSAKey,                 /* pointer to CSSM DSA Key */
    uint32 *CSSMDSAKeyLength)           /* length of CSSM DSA Key */
{
    DER_NODE *root, *parameters, *newnode;
    BER_PARSED_ITEM item, Tmpitem, Xitem;
    uint32 KeyBlockLength, contLength;
    int i;
    CSSM_RETURN retValue = CSSM_OK;

    /* in case of error, return NULL and zero lengths */
    *CSSMDSAKey = 0;
    *CSSMDSAKeyLength = 0;
    /* sample code to convert bsafe 3.0 dsa public key to X509 subjectPublicKeyInfo */
    KeyBlockLength = BER_SizeofObject((uint8*)KeyBlock);

    if (!BER_ExpandItem(KeyBlock, KeyBlockLength, &item)) 
        return CSSMERR_CSP_INVALID_KEY;

    retValue = BER_ParseSimpleObject(&item, &root); /* no fancy type */
    if (CSSM_OK != retValue)
        return retValue;

    /* Allocate new DER node */
    newnode = DER_AllocNode(4);
    if (NULL == newnode)
        return CSSMERR_CSP_MEMORY_ERROR;

    for (;;) {  /* loop body just for error cleanup */
        if (root->Count != 3)  {/* seq (0, algid, keybits) */
            retValue = CSSMERR_CSP_INVALID_KEY;
            break;
        }
        if (!root->Child[1].X.Node) {
            retValue = CSSMERR_CSP_INVALID_KEY;
            break;
        }
        /* find the parameters */
        parameters = root->Child[1].X.Node->Child[1].X.Node;
        if (!parameters)  {
            retValue = CSSMERR_CSP_INVALID_KEY;
            break;
        }
            
        if (parameters->Count != 4) {/* bsafe 3.0 generates 3*256, p, q, g */
            retValue = CSSMERR_CSP_INVALID_KEY;
            break;
        }

        newnode->Tag = newnode->OriginalTag = sequenceTag;

        newnode->Child[0].IsLeaf = 1;
        newnode->Child[0].X.Input.Tag = integerTag;
        contLength = parameters->Child[1].X.Input.ContentLength;
        newnode->Child[0].X.Input.Content = local_malloc(contLength, NULL);
        if (NULL == newnode->Child[0].X.Input.Content) {
            retValue = CSSMERR_CSP_MEMORY_ERROR;
            break;
        }
        memcpy((void *)newnode->Child[0].X.Input.Content, 
                    parameters->Child[1].X.Input.Content, contLength);
        newnode->Child[0].X.Input.ContentLength = contLength;

        newnode->Child[1].IsLeaf = 1;
        newnode->Child[1].X.Input.Tag = integerTag;
        contLength = parameters->Child[2].X.Input.ContentLength;
        newnode->Child[1].X.Input.Content = local_malloc(contLength, NULL);
        if (NULL == newnode->Child[1].X.Input.Content) {
            retValue = CSSMERR_CSP_MEMORY_ERROR;
            break;
        }
        memcpy((void *)newnode->Child[1].X.Input.Content, 
                    parameters->Child[2].X.Input.Content, contLength);
        newnode->Child[1].X.Input.ContentLength = contLength;

        newnode->Child[2].IsLeaf = 1;
        newnode->Child[2].X.Input.Tag = integerTag;
        contLength = parameters->Child[3].X.Input.ContentLength;
        newnode->Child[2].X.Input.Content = local_malloc(contLength, NULL);
        if (NULL == newnode->Child[2].X.Input.Content) {
            retValue = CSSMERR_CSP_MEMORY_ERROR;
            break;
        }
        memcpy((void *)newnode->Child[2].X.Input.Content, 
                    parameters->Child[3].X.Input.Content, contLength);
        newnode->Child[2].X.Input.ContentLength = contLength;

        if (!BER_ExpandItem(root->Child[2].X.Input.Content, 
                    root->Child[2].X.Input.ContentLength, &Tmpitem)) {
            retValue = CSSMERR_CSP_INVALID_KEY;
            break;
        }
        if (!BER_ExpandItem(Tmpitem.Content, Tmpitem.ContentLength, &Xitem)) {
            retValue = CSSMERR_CSP_INVALID_KEY;
            break;
        }

        newnode->Child[3].IsLeaf = 1;
        newnode->Child[3].X.Input.Tag = integerTag;
        newnode->Child[3].X.Input.Content = local_malloc(Xitem.ContentLength, NULL);
        if (NULL == newnode->Child[3].X.Input.Content) {
            retValue = CSSMERR_CSP_MEMORY_ERROR;
            break;
        }
        memcpy((void *)newnode->Child[3].X.Input.Content, Xitem.Content, Xitem.ContentLength);
        newnode->Child[3].X.Input.ContentLength = Xitem.ContentLength;

        /* flatten newnode which represents the CSSM DSA key */
        *CSSMDSAKeyLength = DER_SizeofTree(newnode);
        *CSSMDSAKey = local_malloc(*CSSMDSAKeyLength, NULL);
        if (!*CSSMDSAKey) {     
            retValue = CSSMERR_CSP_MEMORY_ERROR;
            break;
        }
        *CSSMDSAKeyLength = DER_CopyTree(newnode, *CSSMDSAKey);  /* flatten  */
        break;
    }
    for (i = 0; i < 4; i++) {
        newnode->Child[i].X.Input.Tag = 0;
        if (newnode->Child[i].X.Input.Content) 
            local_free((void *)newnode->Child[i].X.Input.Content, NULL);
    }
    if (newnode) DER_FreeParseTree(newnode);
    DER_FreeParseTree(root);

    return retValue;
}

CSSM_RETURN
Cryp_BSafeDSAPublicKey_to_CSSMDSAPublicKey(
    uint32                  nModulus,
    uint32                  nPublicKeyUsage,
    uint8                   *pBSafePubKey,
    CSSM_KEY_PTR            pCssmPubKey)
{
    CSSM_RETURN retValue;
    CSSM_KEYHEADER  *pCssmPubKeyHdr=NULL;
    uint8           *pCssmPubKeyData=NULL;
    uint8           *pBSFPubKeyData=NULL;
    uint32          nBSFPubKeyDataLen=0;

    retValue = BsafeToCSSMDSAPublicKeyBlock(pBSafePubKey, &pBSFPubKeyData, &nBSFPubKeyDataLen);
    if (CSSM_OK != retValue)
        return retValue;

    // CSSM public key (header and data)
    pCssmPubKeyHdr  = &pCssmPubKey->KeyHeader;
    pCssmPubKeyData = pCssmPubKey->KeyData.Data;

    // Intializes key headers
    {
        uint32  nBlobType=0, nKeyFormat=0, nKeyClass=0, nKeySizeInBits=0,
                nKeyAttr=0;

        /* public key */    
        nBlobType = CSSM_KEYBLOB_RAW;
        nKeyFormat= CSSM_KEYBLOB_RAW_FORMAT_FIPS186;
        nKeyClass = CSSM_KEYCLASS_PUBLIC_KEY;
        nKeySizeInBits = nModulus;
        nKeyAttr = CSSM_KEYATTR_EXTRACTABLE;
        
        csp_KeyHeader_Fill(pCssmPubKeyHdr, nBlobType, nKeyFormat, CSSM_ALGID_DSA,
                            nKeyClass, nKeySizeInBits, nKeyAttr,
                            nPublicKeyUsage, NULL/*&StartDate*/, NULL/*&EndDate*/, 
                            CSSM_ALGID_NONE, 0, 0);
    }

    memcpy(pCssmPubKeyData, pBSFPubKeyData, nBSFPubKeyDataLen);
    pCssmPubKey->KeyData.Length = nBSFPubKeyDataLen;
    local_free(pBSFPubKeyData, NULL);

    return CSSM_OK;
}

CSSM_RETURN
Cryp_BSafeDSAPrivateKey_to_CSSMDSAPrivateKey(
    uint32                  nModulus,
    uint32                  nPrivateKeyUsage,
    uint8                   *pBSafePriKey,
    CSSM_KEY_PTR            pCssmPriKey)
{
    CSSM_KEYHEADER  *pCssmPriKeyHdr=NULL;
    uint8           *pCssmPriKeyData=NULL;
    uint8           *pBSFPriKeyData=NULL;
    uint32          nBSFPriKeyDataLen=0;
    CSSM_RETURN retValue = CSSM_OK;

    retValue = BsafeToCSSMDSAPrivateKeyBlock(pBSafePriKey, &pBSFPriKeyData, &nBSFPriKeyDataLen);
    if (CSSM_OK != retValue)
        return retValue;

    // CSSM private key (header and data)
    pCssmPriKeyHdr  = &pCssmPriKey->KeyHeader;
    pCssmPriKeyData = pCssmPriKey->KeyData.Data;

    // Intializes key headers
    {
        uint32  nBlobType=0, nKeyFormat=0, nKeyClass=0, nKeySizeInBits=0,
                nKeyAttr=0;

        /* private key */
        nBlobType = CSSM_KEYBLOB_RAW;
        nKeyFormat= CSSM_KEYBLOB_RAW_FORMAT_FIPS186;
        nKeyClass = CSSM_KEYCLASS_PRIVATE_KEY;
        nKeySizeInBits = nModulus;
        nKeyAttr = CSSM_KEYATTR_SENSITIVE|CSSM_KEYATTR_EXTRACTABLE|CSSM_KEYATTR_PERMANENT;
        
        csp_KeyHeader_Fill(pCssmPriKeyHdr, nBlobType, nKeyFormat, CSSM_ALGID_DSA,
                            nKeyClass, nKeySizeInBits, nKeyAttr,
                            nPrivateKeyUsage, NULL/*&StartDate*/, NULL/*&EndDate*/, 
                            CSSM_ALGID_NONE, 0, 0);
    }

    memcpy(pCssmPriKeyData, pBSFPriKeyData, nBSFPriKeyDataLen);
    pCssmPriKey->KeyData.Length = nBSFPriKeyDataLen;
    local_free(pBSFPriKeyData, NULL);

    return CSSM_OK;
}

CSSM_RETURN
Cryp_CSSMDSAPublicKey_to_BSafeDSAPublicKey(
    CSSM_KEY_PTR            pCssmPubKey,
    uint8                   *pBSafePubKey,
    uint32                  *pnBSafePubLen)
{
    uint8 *pCssmPubKeyData=NULL;
    sint32 nPubDataLen=0;
    uint8 dsaoid[] = { 0x2b, 0x0e, 0x03, 0x02, 0x0c };
    uint8 yBuf[256];
    uint32 yLen = 0, contLength, tagLen, lenLen;
    DER_NODE *root, *newnode, *newnode0, *newnode01;
    BER_PARSED_ITEM item;
    int i;
    CSSM_RETURN retValue = CSSM_OK;

    if ((pCssmPubKey == NULL) || (pBSafePubKey == NULL) || (pnBSafePubLen == NULL))
        return CSSMERR_CSP_INTERNAL_ERROR;

    nPubDataLen = pCssmPubKey->KeyData.Length;
    if (nPubDataLen <= 0)
        return CSSMERR_CSP_INTERNAL_ERROR;

    pCssmPubKeyData = pCssmPubKey->KeyData.Data;

    /* parse it */
    if (!BER_ExpandItem(pCssmPubKeyData, nPubDataLen, &item)) 
        return CSSMERR_CSP_INTERNAL_ERROR;
    retValue = BER_ParseSimpleObject(&item, &root); /* no fancy type */
    if (CSSM_OK != retValue)
        return retValue;
    if (root->Count != 4)  /* seq (p, q, g, y) */
        return CSSMERR_CSP_INTERNAL_ERROR;

    /* Allocate new DER nodes */
    newnode = DER_AllocNode(2);
    if (NULL == newnode) {
        return CSSMERR_CSP_MEMORY_ERROR;
    }
    newnode0 = DER_AllocNode(2);
    if (NULL == newnode0) {
        DER_FreeParseTree(newnode);
        return CSSMERR_CSP_MEMORY_ERROR;
    }
    newnode01 = DER_AllocNode(4);
    if (NULL == newnode01) {
        DER_FreeParseTree(newnode);
        DER_FreeParseTree(newnode0);
        return CSSMERR_CSP_MEMORY_ERROR;
    }


    for (;;) {  /* loop body just for error cleanup */

        /* START: seq(3*256, p, q, g) */
        newnode01->Tag = newnode01->OriginalTag = sequenceTag;

        newnode01->Child[0].IsLeaf = 1;
        newnode01->Child[0].X.Input.Tag = integerTag;
        if (root->Child[0].X.Input.ContentLength >= 128) {
            newnode01->Child[0].X.Input.Content = dsaVer1024;
            newnode01->Child[0].X.Input.ContentLength = sizeof(dsaVer1024);
        } else if (root->Child[0].X.Input.ContentLength >= 96) {
            newnode01->Child[0].X.Input.Content = dsaVer768;
            newnode01->Child[0].X.Input.ContentLength = sizeof(dsaVer768);
        } else {
            newnode01->Child[0].X.Input.Content = dsaVer512;
            newnode01->Child[0].X.Input.ContentLength = sizeof(dsaVer512);
        }

        newnode01->Child[1].IsLeaf = 1;
        newnode01->Child[1].X.Input.Tag = integerTag;
        contLength = root->Child[0].X.Input.ContentLength;
        newnode01->Child[1].X.Input.Content = local_malloc(contLength, NULL);
        if (NULL == newnode01->Child[1].X.Input.Content) {
            retValue = CSSMERR_CSP_MEMORY_ERROR;
            break;
        }
        memcpy((void *)newnode01->Child[1].X.Input.Content, 
                    root->Child[0].X.Input.Content, contLength);
        newnode01->Child[1].X.Input.ContentLength = contLength;

        newnode01->Child[2].IsLeaf = 1;
        newnode01->Child[2].X.Input.Tag = integerTag;
        contLength = root->Child[1].X.Input.ContentLength;
        newnode01->Child[2].X.Input.Content = local_malloc(contLength, NULL);
        if (NULL == newnode01->Child[2].X.Input.Content) {
            retValue = CSSMERR_CSP_MEMORY_ERROR;
            break;
        }
        memcpy((void *)newnode01->Child[2].X.Input.Content, 
                    root->Child[1].X.Input.Content, contLength);
        newnode01->Child[2].X.Input.ContentLength = contLength;

        newnode01->Child[3].IsLeaf = 1;
        newnode01->Child[3].X.Input.Tag = integerTag;
        contLength = root->Child[2].X.Input.ContentLength;
        newnode01->Child[3].X.Input.Content = local_malloc(contLength, NULL);
        if (NULL == newnode01->Child[3].X.Input.Content) {
            retValue = CSSMERR_CSP_MEMORY_ERROR;
            break;
        }
        memcpy((void *)newnode01->Child[3].X.Input.Content, 
                    root->Child[2].X.Input.Content, contLength);
        newnode01->Child[3].X.Input.ContentLength = contLength;
        /* END: seq(3*256, p, q, g) */

        /* START: seq(OID, newnode01) */
        newnode0->Tag = newnode0->OriginalTag = sequenceTag;

        newnode0->Child[0].IsLeaf = 1;
        newnode0->Child[0].X.Input.Tag = oidTag;
        newnode0->Child[0].X.Input.Content = dsaoid;
        newnode0->Child[0].X.Input.ContentLength = sizeof(dsaoid);

        newnode0->Child[1].IsLeaf = 0;
        newnode0->Child[1].X.Node = newnode01;
        /* END: seq(OID, newnode01) */

        /* START: seq(newnode0, bitstring Y) */
        newnode->Tag = newnode->OriginalTag = sequenceTag;

        newnode->Child[0].IsLeaf = 0;
        newnode->Child[0].X.Node = newnode0;

        yBuf[0] = 0; 
        yLen++;
        tagLen = BER_OutputTag(root->Child[3].X.Input.Tag, yBuf+yLen); 
        yLen += tagLen;
        lenLen = BER_OutputLength(root->Child[3].X.Input.ContentLength, yBuf+yLen);
        yLen += lenLen;
        memcpy(yBuf+yLen, root->Child[3].X.Input.Content, 
                    root->Child[3].X.Input.ContentLength);
        yLen += root->Child[3].X.Input.ContentLength;

        newnode->Child[1].IsLeaf = 1;
        newnode->Child[1].X.Input.Tag = bitstringTag;
        newnode->Child[1].X.Input.Content = yBuf;
        newnode->Child[1].X.Input.ContentLength = yLen;
        /* END: seq(newnode0, bitstring Y) */

        /* flatten newnode which represents the CSSM DSA key */
        *pnBSafePubLen = DER_SizeofTree(newnode);       
        *pnBSafePubLen = DER_CopyTree(newnode, pBSafePubKey);  /* flatten  */
        break;
    }
    for (i = 1; i < 4; i++) {
        newnode01->Child[i].X.Input.Tag = 0;
        if (newnode01->Child[i].X.Input.Content) 
            local_free((void *)newnode01->Child[i].X.Input.Content, NULL);
    }
    if (newnode) DER_FreeParseTree(newnode);
    //if (newnode0) DER_FreeParseTree(newnode0);
    //if (newnode01) DER_FreeParseTree(newnode01);
    DER_FreeParseTree(root);

    return retValue;    
}


CSSM_RETURN
Cryp_CSSMDSAPrivateKey_to_BSafeDSAPrivateKey(
    CSSM_KEY_PTR            pCssmPriKey,
    uint8                   *pBSafePriKey,
    uint32                  *pnBSafePriLen)
{
    uint8 *pCssmPriKeyData=NULL;
    sint32 nPriDataLen=0;
    uint8 dsaoid[] = { 0x2b, 0x0e, 0x03, 0x02, 0x0c };
    uint8 zero[] = { 0 };
    uint8 xBuf[256];
    uint32 xLen, contLength;
    DER_NODE *root, *newnode, *newnode1, *newnode11;
    BER_PARSED_ITEM item;
    int i;
    CSSM_RETURN retValue = CSSM_OK;

    if ((pCssmPriKey == NULL) || (pBSafePriKey == NULL) || (pnBSafePriLen == NULL))
        return CSSMERR_CSP_INTERNAL_ERROR;

    nPriDataLen = pCssmPriKey->KeyData.Length;
    if (nPriDataLen <= 0)
        return CSSMERR_CSP_INTERNAL_ERROR;

    pCssmPriKeyData = pCssmPriKey->KeyData.Data;

    /* parse it */
    if (!BER_ExpandItem(pCssmPriKeyData, nPriDataLen, &item)) 
        return CSSMERR_CSP_INTERNAL_ERROR;
    retValue = BER_ParseSimpleObject(&item, &root); /* no fancy type */
    if (CSSM_OK != retValue)
        return retValue;
    if (root->Count != 4)  /* seq (p, q, g, x) */
        return CSSMERR_CSP_INTERNAL_ERROR;

    /* Allocate new DER nodes */
    newnode = DER_AllocNode(3);
    if (NULL == newnode) {
        return CSSMERR_CSP_MEMORY_ERROR;
    }
    newnode1 = DER_AllocNode(2);
    if (NULL == newnode1) {
        DER_FreeParseTree(newnode);
        return CSSMERR_CSP_MEMORY_ERROR;
    }
    newnode11 = DER_AllocNode(4);
    if (NULL == newnode11) {
        DER_FreeParseTree(newnode);
        DER_FreeParseTree(newnode1);
        return CSSMERR_CSP_MEMORY_ERROR;
    }

    for (;;) {  /* loop body just for error cleanup */

        /* START: seq(3*256, p, q, g) */
        newnode11->Tag = newnode11->OriginalTag = sequenceTag;

        newnode11->Child[0].IsLeaf = 1;
        newnode11->Child[0].X.Input.Tag = integerTag;
        if (root->Child[0].X.Input.ContentLength >= 128) {
            newnode11->Child[0].X.Input.Content = dsaVer1024;
            newnode11->Child[0].X.Input.ContentLength = sizeof(dsaVer1024);
        } else if (root->Child[0].X.Input.ContentLength >= 96) {
            newnode11->Child[0].X.Input.Content = dsaVer768;
            newnode11->Child[0].X.Input.ContentLength = sizeof(dsaVer768);
        } else {
            newnode11->Child[0].X.Input.Content = dsaVer512;
            newnode11->Child[0].X.Input.ContentLength = sizeof(dsaVer512);
        }

        newnode11->Child[1].IsLeaf = 1;
        newnode11->Child[1].X.Input.Tag = integerTag;
        contLength = root->Child[0].X.Input.ContentLength;
        newnode11->Child[1].X.Input.Content = local_malloc(contLength, NULL);
        if (NULL == newnode11->Child[1].X.Input.Content) {
            retValue = CSSMERR_CSP_MEMORY_ERROR;
            break;
        }
        memcpy((void *)newnode11->Child[1].X.Input.Content, 
                    root->Child[0].X.Input.Content, contLength);
        newnode11->Child[1].X.Input.ContentLength = contLength;

        newnode11->Child[2].IsLeaf = 1;
        newnode11->Child[2].X.Input.Tag = integerTag;
        contLength = root->Child[1].X.Input.ContentLength;
        newnode11->Child[2].X.Input.Content = local_malloc(contLength, NULL);
        if (NULL == newnode11->Child[2].X.Input.Content) {
            retValue = CSSMERR_CSP_MEMORY_ERROR;
            break;
        }
        memcpy((void *)newnode11->Child[2].X.Input.Content, 
                    root->Child[1].X.Input.Content, contLength);
        newnode11->Child[2].X.Input.ContentLength = contLength;

        newnode11->Child[3].IsLeaf = 1;
        newnode11->Child[3].X.Input.Tag = integerTag;
        contLength = root->Child[2].X.Input.ContentLength;
        newnode11->Child[3].X.Input.Content = local_malloc(contLength, NULL);
        if (NULL == newnode11->Child[3].X.Input.Content) {
            retValue = CSSMERR_CSP_MEMORY_ERROR;
            break;
        }
        memcpy((void *)newnode11->Child[3].X.Input.Content, 
                    root->Child[2].X.Input.Content, contLength);
        newnode11->Child[3].X.Input.ContentLength = contLength;
        /* END: seq(3*256, p, q, g) */

        /* START: seq(OID, newnode11) */
        newnode1->Tag = newnode1->OriginalTag = sequenceTag;

        newnode1->Child[0].IsLeaf = 1;
        newnode1->Child[0].X.Input.Tag = oidTag;
        newnode1->Child[0].X.Input.Content = dsaoid;
        newnode1->Child[0].X.Input.ContentLength = sizeof(dsaoid);

        newnode1->Child[1].IsLeaf = 0;
        newnode1->Child[1].X.Node = newnode11;
        /* END: seq(OID, newnode11) */

        /* START: seq(0, newnode1, octetstring X) */
        newnode->Tag = newnode->OriginalTag = sequenceTag;

        newnode->Child[0].IsLeaf = 1;
        newnode->Child[0].X.Input.Tag = integerTag;
        newnode->Child[0].X.Input.Content = zero;
        newnode->Child[0].X.Input.ContentLength = sizeof(zero);

        newnode->Child[1].IsLeaf = 0;
        newnode->Child[1].X.Node = newnode1;

        xLen = DER_ConstructItem(sequenceTag, 1, &root->Child[3].X.Input, xBuf);
        newnode->Child[2].IsLeaf = 1;
        newnode->Child[2].X.Input.Tag = octetstringTag;
        newnode->Child[2].X.Input.Content = xBuf;
        newnode->Child[2].X.Input.ContentLength = xLen;
        /* END: seq(newnode1, bitstring X) */

        /* flatten newnode which represents the CSSM DSA key */
        *pnBSafePriLen = DER_SizeofTree(newnode);
        *pnBSafePriLen = DER_CopyTree(newnode, pBSafePriKey);  /* flatten  */
        break;
    }
    for (i = 1; i < 4; i++) {
        newnode11->Child[i].X.Input.Tag = 0;
        if (newnode11->Child[i].X.Input.Content) 
            local_free((void *)newnode11->Child[i].X.Input.Content, NULL);
    }
    if (newnode) DER_FreeParseTree(newnode);
    //if (newnode1) DER_FreeParseTree(newnode1);
    //if (newnode11) DER_FreeParseTree(newnode11);
    DER_FreeParseTree(root);

    return retValue;    
}
