/*-----------------------------------------------------------------------
 * File: CHN_UTIL.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 

/* CSSM includes */
#include "cssmtype.h"         /* CSSM data types */
#include "cssmport.h"
#include "X509Time.h"
#include "cssmMemTrack.h"

/* X.509 includes */
#include "oidscert.h"         /* x.509 oid definitions */

/* local includes */
#include "Chn_util.h"         /* chn_util function defs */
#include "api_shim.h"       /* Wrappers around CSSM APIs */
#include "maf_include.h"
#include "inx509tp.h"


/*---------------------------------------------------------------
 *
 * Name: tp_CertGetFieldValue
 *
 * Description:
 *   Retrieves the Certificate field value associated with the oid
 *
 * Parameters: 
 *   hCL (input)          - Handle of the CL module
 *   oid (input)          - OID of the certificate field
 *   pCert (input)        - Pointer to the certificate to be 
 *                          operated on
 *
 * Returns:
 *   NULL - failed to retrieve field
 *   non NULL - pointer to the field
 *
 *-------------------------------------------------------------*/

CSSM_RETURN
tp_CertGetFieldValue( CSSM_CL_HANDLE CLhandle, 
                        CSSM_OID_PTR OID,
                        const CSSM_DATA *pCert,
                        CSSM_DATA_PTR *retVal)
{
  uint32 numberOfFields = 0;
  CSSM_HANDLE resultsHandle = 0;
  CSSM_RETURN Rtn = CSSM_OK;

/* --GET ENCODED ISSUER NAME FROM CERT-- */
  Rtn = addin_cl_CertGetFirstFieldValue( CLhandle, 
                                            pCert, 
                                            OID,
                                            &resultsHandle, 
                                            &numberOfFields,
                                            retVal);
   
  return( Rtn );
}

CSSM_RETURN
tp_CertGetIssuerX509Name( CSSM_CL_HANDLE CLhandle, 
                            const CSSM_DATA *pCert, 
                            CSSM_DATA_PTR *retVal )
{
/* --GET ENCODED ISSUER NAME FROM CERT-- */
  return ( tp_CertGetFieldValue(CLhandle, &CSSMOID_X509V1IssuerName, pCert, retVal) );
}

CSSM_RETURN
tp_CertGetSubjectX509Name( CSSM_CL_HANDLE CLhandle, 
                            const CSSM_DATA *pCert,
                            CSSM_DATA_PTR *retVal )
{
/* --GET ENCODED SUBJECT NAME FROM CERT-- */
  return ( tp_CertGetFieldValue(CLhandle, &CSSMOID_X509V1SubjectName, pCert, retVal) );
}


/*---------------------------------------------------------------
 *
 * Name: tp_IsSelfSignedCert
 *
 * Description:
 *   Determines if an X.509 certificate is self-signed 
 *
 * Parameters: 
 *   hTP (input)          - Handle of the TP module
 *   hCL (input)          - Handle of the CL module
 *   pCert (input)        - Pointer to the certificate to be 
 *                          operated on
 *
 * Returns:
 *   NULL - failed to retrieve issuer name from cert
 *   non NULL - pointer to the issuer name (decoded)
 *
 *-------------------------------------------------------------*/

/* determines if certificate self-signed */
CSSM_BOOL
tp_IsSelfSignedCert( CSSM_TP_HANDLE hTP, 
                        CSSM_CL_HANDLE hCL, 
                        const CSSM_DATA *pCert)
{
    CSSM_DATA_PTR pSubjectX509Name = NULL;
    CSSM_DATA_PTR pIssuerX509Name = NULL;
    CSSM_BOOL bResult = CSSM_FALSE;
    CSSM_RETURN Rtn = CSSM_OK;

    /* --CHECK INPUTS-- */
    if( pCert == NULL )
    {
        /* no cert to work on, therefor not self-signed */
        return( CSSM_FALSE );
    }

    /* --GET NECESSARY FIELDS FROM CERT-- */
    Rtn = tp_CertGetSubjectX509Name( hCL, pCert, &pSubjectX509Name );
    if (Rtn != CSSM_OK)
    {
        return( CSSM_FALSE );
    }

    Rtn = tp_CertGetIssuerX509Name( hCL, pCert, &pIssuerX509Name );
    if (Rtn != CSSM_OK)
    {
      App_Free( hTP, pSubjectX509Name->Data);
      App_Free( hTP, pSubjectX509Name );
      return( CSSM_FALSE );
    }
        
    /* --COMPARE FIELDS-- */
    bResult = tp_IsSameX509Name( pSubjectX509Name, pIssuerX509Name );
    
    /* --CLEANUP MEMORY-- */
    App_Free( hTP, pSubjectX509Name->Data );
    App_Free( hTP, pSubjectX509Name );
    App_Free( hTP, pIssuerX509Name->Data );
    App_Free( hTP, pIssuerX509Name );

    /* --RETURN-- */
    return( bResult );
}



/*---------------------------------------------------------------
 *
 * Name: tp_IsSameX509Name
 *
 * Description:
 *   Determines if two X.509 distinguished names are equivalent
 *
 * Parameters: 
 *   pName1 (input)       - Pointer to first X.509 name (decoded)
 *   pName2 (input)       - Pointer to second X.509 name (decoded)
 *
 * Returns:
 *   CSSM_TRUE, if both names are equivalent
 *   CSSM_FALSE, if an error occurs or the names are not 
 *     equivalent
 *
 *-------------------------------------------------------------*/

CSSM_BOOL
tp_IsSameX509Name( const CSSM_DATA *pName1, 
                   const CSSM_DATA *pName2 )
{
  if (pName1->Length != pName2->Length)
  {
    return CSSM_FALSE;
  }

  if (((pName1->Data == NULL) && (pName2->Data != NULL))
        || ((pName1->Data != NULL) && (pName2->Data == NULL)) )
  {
    return CSSM_FALSE;
  }

  if (pName1->Data == NULL)
  {
    return CSSM_TRUE;
  }

  return ((memcmp(pName1->Data, pName2->Data, pName1->Length) == 0) ? CSSM_TRUE : CSSM_FALSE);
}

CSSM_BOOL
tp_IsValidCrlType( CSSM_CRL_TYPE crlType )
{
	switch (crlType)
	{
	case CSSM_CRL_TYPE_UNKNOWN:
	case CSSM_CRL_TYPE_X_509v1:
    case CSSM_CRL_TYPE_X_509v2:
    case CSSM_CRL_TYPE_SPKI:
    case CSSM_CRL_TYPE_MULTIPLE:
		return CSSM_TRUE;
	
	default:
		return CSSM_FALSE;
	}
}

CSSM_BOOL
tp_IsValidCrlEncoding( CSSM_CRL_ENCODING crlEncoding )
{
	switch (crlEncoding)
	{
	case CSSM_CRL_ENCODING_UNKNOWN:
    case CSSM_CRL_ENCODING_CUSTOM:
    case CSSM_CRL_ENCODING_BER:
    case CSSM_CRL_ENCODING_DER:
    case CSSM_CRL_ENCODING_BLOOM:
    case CSSM_CRL_ENCODING_SEXPR:
    case CSSM_CRL_ENCODING_MULTIPLE:
		return CSSM_TRUE;
	
	default:
		return CSSM_FALSE;
	}
}

/*---------------------------------------------------------------
 *
 * Name: tp_GetCertFromDBList
 *
 * Description:
 *   Gets a certificate from the DBs in the DBList using the
 *   passed slection predicate as the query criteria
 *
 * Parameters: 
 *   hTP (input)          - Handle of the TP module
 *   hCL (input)          - Handle of the CL module
 *   pPredicate (input)   - Pointer to the selection predicate
 *                          to be used for the query
 *   NumberOfPredicates (input)
 *                        - Number of predicates passed in 
 *                          pPredicate
 *   pDBList (input)      - Pointer to the DB list to be searched
 *
 * Returns:
 *   NULL, if the certificate cannot be retrieved
 *   Pointer to the retrieved certificate (encoded) on success
 *
 *-------------------------------------------------------------*/

CSSM_RETURN 
tp_GetCertFromDBList( CSSM_TP_HANDLE hTP,
                      CSSM_CL_HANDLE hCL, 
                      CSSM_SELECTION_PREDICATE_PTR pPredicate,
                      uint32 NumberOfPredicates,
                      const CSSM_DL_DB_LIST *pDBList,
                      CSSM_DATA_PTR *ppCert)
{
    CSSM_DATA CurrentCert = {0, NULL}, *pReturnCert = NULL;
    CSSM_API_MEMORY_FUNCS DLMemFuncs = {NULL, NULL, NULL, NULL, NULL}, 
    CLMemFuncs = {NULL, NULL, NULL, NULL,NULL};
    CSSM_DL_DB_HANDLE_PTR /* pDBListIndex = NULL, */ pCurrentDLDB = NULL;
    uint32 counter1 = 0;
    CSSM_DL_HANDLE hCurrentDL = 0;
    CSSM_DB_HANDLE hCurrentDB = 0;
    CSSM_HANDLE hResults;
    CSSM_RETURN RetCode = CSSM_OK;
    CSSM_DB_UNIQUE_RECORD_PTR pDB_UR = NULL;

    /* DB query structures */
    CSSM_QUERY Query = { CSSM_DL_DB_RECORD_CERT,
                       CSSM_DB_AND,
                       0,
                       NULL,
                       CSSM_QUERY_TIMELIMIT_NONE,
                       CSSM_QUERY_SIZELIMIT_NONE,
                       0 };

	if ( pDBList->DLDBHandle == NULL )
	{
		return( CSSMERR_TP_INVALID_DB_LIST );
	}

    Query.SelectionPredicate = pPredicate;
    Query.NumSelectionPredicates = NumberOfPredicates;

  
    /* --GET CL MEM FUNCS-- */
    RetCode = addin_GetAPIMemoryFunctions( hCL, &CLMemFuncs );  
    if( RetCode != CSSM_OK )
    {
        return( RetCode );
    }
  
    
/* --ITERATE THROUGH DBLIST AND QUERY FOR CERT-- */
    for( counter1 = 0; counter1 < pDBList->NumHandles; counter1++ )
    {
        /* Get current handles and memory functions */
        pCurrentDLDB = &pDBList->DLDBHandle[counter1];
        hCurrentDL = pCurrentDLDB->DLHandle;
        if( hCurrentDL == 0 )
        {
            return( CSSMERR_TP_INVALID_DL_HANDLE );
        }

        hCurrentDB = pCurrentDLDB->DBHandle;
        if( hCurrentDB == 0 )
        {
            return( CSSMERR_TP_INVALID_DB_HANDLE );
        }

        /* --GET MEMORY FUNCTIONS CORRESPONDING TO THE CURRENT DL-- */
        RetCode = addin_GetAPIMemoryFunctions( hCurrentDL, &DLMemFuncs );  
        if( RetCode != CSSM_OK )
        {
            return( RetCode );
        }
    
        /* --QUERY FOR CERTIFICATE WHICH MEETS PREDICATE PROVIDED-- */
        RetCode = addin_dl_DataGetFirst( *pCurrentDLDB, &Query,
                            &hResults, NULL, &CurrentCert, &pDB_UR );

        /* cleanup query state */
        addin_dl_AbortQuery( *pCurrentDLDB, hResults );
        addin_dl_FreeUniqueRecord(*pCurrentDLDB, pDB_UR);
    
        /* --IF A CERT IS FOUND, COPY AND RETURN-- */
        if( CurrentCert.Data != NULL )
        {
            /* copy original cert into TP memory space */
            pReturnCert = App_Malloc( hTP, sizeof( CSSM_DATA ) );
            pReturnCert->Data = App_Malloc( hTP, CurrentCert.Length );
            pReturnCert->Length = CurrentCert.Length;
            memcpy( pReturnCert->Data, CurrentCert.Data, CurrentCert.Length );
            /* free original cert */
            CSSM_FREE_WRAPPER(DLMemFuncs.free_func,  CurrentCert.Data, DLMemFuncs.AllocRef );
            /* return copy */
            *ppCert = pReturnCert;
            return( RetCode );
        } /* A cert was found, copied, and returned to the caller */

        /* A cert was not found */

    } /* for */
  
    /* --CERT RETRIEVAL FAILED-- */
    return( RetCode );
}


/*---------------------------------------------------------------
 *
 * Name: tp_GetCertFromDBListBySName
 *
 * Description:
 *   Gets a certificate from the DBs in the DBList using the
 *   subject name as the query criteria
 *
 * Parameters: 
 *   hTP (input)          - Handle of the TP module
 *   hCL (input)          - Handle of the CL module
 *   pSubjectName (input) - Pointer to the subject name (encoded)
 *                          to be used as the query search string
 *   pDBList (input)      - Pointer to the DB list to be searched
 *
 * Returns:
 *   NULL, if the certificate cannot be retrieved
 *   Pointer to the retrieved certificate (encoded) on success
 *
 *-------------------------------------------------------------*/

 CSSM_RETURN
 tp_GetCertFromDBListBySName( CSSM_TP_HANDLE hTP, 
                              CSSM_CL_HANDLE hCL, 
                              const CSSM_DATA *pSubjectName,
                              const CSSM_DL_DB_LIST *pDBList,
                              CSSM_DATA_PTR *ppCert)
{
    CSSM_SELECTION_PREDICATE Predicate;
    CSSM_DB_ATTRIBUTE_DATA Attribute;
    CSSM_DB_ATTRIBUTE_INFO AttrInfo;
    CSSM_API_MEMORY_FUNCS CLMemFuncs = {NULL, NULL, NULL, NULL, NULL};
    CSSM_RETURN RetCode = CSSM_OK;

    /* --GET CL MEM FUNCS-- */
    RetCode = addin_GetAPIMemoryFunctions( hCL, &CLMemFuncs );  
    /* check return value */
    if( RetCode != CSSM_OK )
    {
        return( RetCode );
    }
  
    /* --BUILD SUBJECT NAME PREDICATE-- */
    /* build attribute info data structure */
    AttrInfo.AttributeNameFormat = CSSM_DB_ATTRIBUTE_NAME_AS_OID;
    AttrInfo.Label.AttributeOID = CSSMOID_X509V1SubjectName;
    AttrInfo.AttributeFormat = CSSM_DB_ATTRIBUTE_FORMAT_BLOB;

    /* build attribute data structure */
    Attribute.Info = AttrInfo;
    Attribute.Value = (CSSM_DATA_PTR)pSubjectName;
    Attribute.NumberOfValues = 1;

    /* build selection predicate data structure */
    Predicate.Attribute = Attribute;
    Predicate.DbOperator = CSSM_DB_EQUAL;

    /* --RETRIEVE CERT-- */
    /* call internal function to retrieve cert */
    RetCode = tp_GetCertFromDBList( hTP, hCL, &Predicate, 1, pDBList, ppCert );
  
    /* --RETURN-- */
    return( RetCode );
}


 /*---------------------------------------------------------------
 *
 * Name: tp_CheckCertValidityPeriodValid
 *
 * Description:
 *   Checks to see that the certificate's validity period is valid 
 *
 * Parameters: 
 *   hCL (input)          - Handle of the CL module
 *   pCert (input)        - Pointer to the certificate to be 
 *                          operated on
 *
 * Returns:
 *   CSSM_OK - passed validity period check
 *   otherwise - an error code
 *
 *-------------------------------------------------------------*/

CSSM_RETURN
tp_CheckCertValidityPeriodValid( CSSM_CL_HANDLE hCL, const CSSM_DATA *pCert )
{
    uint32 NumberOfFields = 0;
    CSSM_HANDLE ResultsHandle = 0;
    CSSM_DATA_PTR EncodedTime = NULL; 
    CSSM_API_MEMORY_FUNCS CLMemFuncs = {NULL, NULL, NULL, NULL, NULL};
    CSSM_RETURN RetCode = CSSM_OK;
    sint32 current_time, timeCompare;
    uint8 currentTimeStr[16];
    CSSM_DATA currentX509Time = {16, NULL};

    memset(currentTimeStr, 0, 16);
    currentX509Time.Data = currentTimeStr;

    /* --OBTAIN CL MEMORY FUNCTIONS-- */
    RetCode = addin_GetAPIMemoryFunctions( hCL, &CLMemFuncs ); 
    if( RetCode != CSSM_OK )
    { /* cssm should have set an error */
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* --GET CURRENT TIME-- */
    time((unsigned long *)&current_time);     /* ifdef VMS apply cast */
    if (TimeToGeneralizedTime(current_time, &currentX509Time) != CSSM_OK)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* --CHECK VALIDITYNOTBEFORE TIME-- */
    /* Get Validity NotBefore Time */
    RetCode = addin_cl_CertGetFirstFieldValue( hCL, 
                                                pCert, 
                                                &CSSMOID_X509V1ValidityNotBefore,
                                                &ResultsHandle, 
                                                &NumberOfFields, 
                                                &EncodedTime);
    if(RetCode != CSSM_OK)
    { /* CL should have set an error */
        return RetCode;
    }

    /* Check Validity */
    RetCode = X509TimeCompare(&currentX509Time, EncodedTime, &timeCompare);

    /* Free memory */  
    CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  (void *)EncodedTime->Data, CLMemFuncs.AllocRef );
    CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  (void *)EncodedTime, CLMemFuncs.AllocRef );

    if (RetCode != CSSM_OK)
        return RetCode;
    if (timeCompare < 0)
        return CSSMERR_TP_CERT_NOT_VALID_YET;

    /* --CHECK VALIDITYNOTAFTER TIME-- */  
    /* Get Validity NotAfter Time */
    RetCode = addin_cl_CertGetFirstFieldValue( hCL, 
                                                pCert, 
                                                &CSSMOID_X509V1ValidityNotAfter,
                                                &ResultsHandle, 
                                                &NumberOfFields,
                                                &EncodedTime);
    if(RetCode != CSSM_OK)
    { /* CL should have set an error */
        return RetCode;
    }

    /* Check Validity */
    RetCode = X509TimeCompare(EncodedTime, &currentX509Time, &timeCompare);

    /* Free memory */  
    CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  (void *)EncodedTime->Data, CLMemFuncs.AllocRef );
    CSSM_FREE_WRAPPER(CLMemFuncs.free_func,  (void *)EncodedTime, CLMemFuncs.AllocRef );

    if (RetCode != CSSM_OK)
        return CSSMERR_TP_INTERNAL_ERROR;
    if (timeCompare < 0)
        return CSSMERR_TP_CERT_EXPIRED;
          
    return CSSM_OK;
}


/*---------------------------------------------------------------
 *
 * Name: tp_IsCertRevoked
 *
 * Description:
 *   Checks the DBs to see if pCert is an entry in any CRL
 *
 * Parameters: 
 *   hTP (input)          - Handle of the TP module
 *   hCL (input)          - Handle of the CL module
 *   pDBList (input)      - Pointer to the DB list to be searched
 *                          to be used as the query search string
 *   pCert (input)        - Pointer to the Certificate to be 
 *                          searched for
 *
 * Returns:
 *   CSSM_TRUE if the certificate is revoked
 *   CSSM_FALSE if the certificate is not revoked
 *
 *-------------------------------------------------------------*/

CSSM_BOOL
tp_IsCertRevoked( CSSM_TP_HANDLE hTP, 
                 CSSM_CL_HANDLE hCL, 
                 const CSSM_DL_DB_LIST *pDBList,
                 CSSM_DATA_PTR pCert)
{
    CSSM_DATA Crl = {0, NULL};
    CSSM_RETURN RetCode = CSSM_OK;
    CSSM_QUERY Query;
    CSSM_HANDLE ResultsHandle = 0;
    CSSM_DB_UNIQUE_RECORD_PTR record_ptr = NULL; /* free after each use */
    unsigned int dldb_count = 0;
    CSSM_BOOL IsCertInCrl_flag = CSSM_FALSE;

    hTP;

    if (pDBList == NULL)
        return CSSM_FALSE;

    if (pDBList->NumHandles == 0)
        return CSSM_FALSE;

  
    /* --BUILD QUERY-- */
    memset( &Query, 0, sizeof(CSSM_QUERY));

    /* Search for all CRLs */
    Query.NumSelectionPredicates = 0;
    Query.SelectionPredicate = NULL;
    Query.RecordType = CSSM_DL_DB_RECORD_CRL;
    Query.Conjunctive = CSSM_DB_NONE;

    /* For each Database */
    for (dldb_count = 0 ; 
           (dldb_count < pDBList->NumHandles) && (IsCertInCrl_flag == CSSM_FALSE) ; 
           dldb_count++)
    {
        /* Get the first CRL */
        RetCode = addin_dl_DataGetFirst( pDBList->DLDBHandle[dldb_count], 
                        &Query, &ResultsHandle, NULL, &Crl, &record_ptr );
        if (record_ptr != NULL) 
        { /* Got one */
            addin_dl_FreeUniqueRecord(pDBList->DLDBHandle[dldb_count], record_ptr);
            record_ptr = NULL;
        } 
        else
        { /* No CRLs in this DB */
            addin_dl_AbortQuery(pDBList->DLDBHandle[dldb_count], ResultsHandle);
            if (Crl.Data)
            {
                App_Free( pDBList->DLDBHandle[dldb_count].DLHandle, Crl.Data );
            }
            Crl.Data = NULL;
            Crl.Length = 0;
            continue; /* iterate to next DB */
        }

        RetCode = addin_cl_IsCertInCrl( hCL, pCert, &Crl, &IsCertInCrl_flag );

        if (IsCertInCrl_flag == CSSM_TRUE)
        { /* found the Cert in the Crl - done! */
            addin_dl_AbortQuery(pDBList->DLDBHandle[dldb_count], ResultsHandle);
            if (Crl.Data)
            {
                App_Free( pDBList->DLDBHandle[dldb_count].DLHandle, Crl.Data );
            }
            Crl.Data = NULL;
            Crl.Length = 0;
            break; /* search no more */
        }

        if (Crl.Data)
        {
            App_Free( pDBList->DLDBHandle[dldb_count].DLHandle, Crl.Data );
        }
        Crl.Data = NULL;
        Crl.Length = 0;
        record_ptr = NULL;

        while( record_ptr != NULL ) 
        {
            RetCode = addin_dl_DataGetNext(pDBList->DLDBHandle[dldb_count],
                        ResultsHandle, NULL, &Crl, &record_ptr);
      
            if (record_ptr != NULL) 
            { /* Got one */
                addin_dl_FreeUniqueRecord(pDBList->DLDBHandle[dldb_count], record_ptr);
            } 
            else
            { /* No CRLs */
                /* clean up, check next DB */
                addin_dl_AbortQuery(pDBList->DLDBHandle[dldb_count], ResultsHandle);
                if (Crl.Data)
                {
                    App_Free( pDBList->DLDBHandle[dldb_count].DLHandle, Crl.Data );
                }
                Crl.Data = NULL;
                Crl.Length = 0;
                break;
            }

            RetCode = addin_cl_IsCertInCrl( hCL, pCert, &Crl, &IsCertInCrl_flag );

            if (Crl.Data)
            {
                App_Free( pDBList->DLDBHandle[dldb_count].DLHandle, Crl.Data );
            }
            Crl.Data = NULL;
            Crl.Length = 0;

            if (IsCertInCrl_flag == CSSM_TRUE)
            { /* found the Cert in the Crl - done! */
                addin_dl_AbortQuery(pDBList->DLDBHandle[dldb_count], ResultsHandle);
                break;
            }
        } /* while( EODS == CSSM_FALSE ) */ 
    } /* for */

    /* --RETURN-- */
    return IsCertInCrl_flag;
}

void 
free_certgroup(CSSM_CERTGROUP_PTR pCertGroup) 
{
    unsigned int i;

    if (pCertGroup == NULL)
    {
        return;
    }

    if (pCertGroup->GroupList.CertList != NULL)
    {
        for (i=0; i<pCertGroup->NumCerts; i++)
        {
            if (pCertGroup->GroupList.CertList[i].Data != NULL)
            {
                free(pCertGroup->GroupList.CertList[i].Data);
            }
        }
        free (pCertGroup->GroupList.CertList);
    }
#if 0   
    if (pCertGroup->Reserved != NULL)
    {
        free(pCertGroup->Reserved);
    }
#endif /* mhs */
    free (pCertGroup);
    return;
}

/*-----------------------------------------------------------------------------
 * Name: tp_CertGroupConstruct
 *
 * Description:
 *   This function constructs an x.509 cert chain ending in a self-signed root 
 *   cert from a fragment and a list of databases. 
 *
 * Parameters: 
 *   TPHandle (input):         handle of TP executing call
 *   CLHandle (input):         handle of CL to use with this call
 *   CSPHandle (input):        handle of CSP to use with this call
 *                             [unused - must be 0]
 *   CertGroupFrag (input):    the cert group to construct from
 *   DBList (input):           list of databases to be used to find certs in the
 *                             chain.  (Only these DB's will be accessed during
 *                             the construct.)
 *   
 * Return value:
 *   The constructed cert group on success.
 *   NULL on failure
 *
 * Error Codes:
 *   -TBD-
 *
 *---------------------------------------------------------------------------*/
static void 
crtgrpcnst_free(CSSM_TP_HANDLE TPHandle, void *ptr, CSSM_UPCALLS_PTR pMemFuncs)
{
    if (pMemFuncs == NULL)
    {
        App_Free( TPHandle, ptr );
    }
    else
    {
        CSSM_FREE_WRAPPER(pMemFuncs->free_func,  TPHandle, ptr );
    }

    return;
}

CSSM_RETURN CSSMAPI 
tp_CertGroupConstruct(  CSSM_TP_HANDLE TPHandle,
                        CSSM_CL_HANDLE CLHandle,
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_DL_DB_LIST *DBList,
                        const CSSM_CERTGROUP *CertGroupFrag,
                        CSSM_BOOL IsAppCall,
                        CSSM_CERTGROUP_PTR *ppCertGroup)
{
    CSSM_DATA_PTR pSubjectCertIssuerName = NULL;
    CSSM_DATA_PTR pEncSubjectCertIssuerName = NULL;
    CSSM_DATA_PTR pSubjectCert = NULL;
    CSSM_DATA_PTR pIssuerCert = NULL;

    CSSM_BOOL bRootFound = CSSM_FALSE; 
    CSSM_BOOL bDBCert = CSSM_FALSE;
    uint32 counter1 = 0;
    uint32 counter2 = 0;
    uint32 counter3 = 0;
    uint32 ReturnNumCerts = 0;
    uint32 RootIndex = 0;
    CSSM_DATA_PTR pTempEncodedCert = NULL;
    CSSM_DATA_PTR pReturnCertList = NULL;
    CSSM_CERTGROUP_PTR pReturnCertGroup = NULL;
#if 0 
    CSSM_TP_X509V3_GRP_INFO_PTR pReturnReserved = NULL;
#endif /* mhs */
    const ADDIN_LOAD_DATA *pLoadData = NULL;
    CSSM_RETURN RetCode = CSSM_OK;

    CSPHandle;

    /* Do we know our execution context? */
    pLoadData = Addin_GetLoadData();
    if ( pLoadData == NULL )
    {
        return( CSSMERR_TP_INTERNAL_ERROR );
    }

    RootIndex = ReturnNumCerts;

    /* --GET PRIMARY CERT-- */
    pSubjectCert = CertGroupFrag->GroupList.CertList;
    if( pSubjectCert == NULL )
    {
        return( CSSMERR_TP_INVALID_CERTGROUP_POINTER );
    }

    /* --SETUP RETURN STRUCTURE-- */
    /* using an arbitrary limit for initial setup */
    if (IsAppCall)
    {
        pReturnCertList = (CSSM_DATA_PTR) 
                            App_Malloc( TPHandle, MAX_CHAIN_DEPTH*sizeof(CSSM_DATA_PTR) ); 
    }
    else
    {
        pReturnCertList = (CSSM_DATA_PTR) 
                            Addin_malloc( MAX_CHAIN_DEPTH*sizeof(CSSM_DATA_PTR), NULL ); 
    }

    if(NULL == pReturnCertList)
    {
        return (CSSMERR_TP_MEMORY_ERROR);
    }

    /* --COPY PRIMARY CERT TO RETURN CERT GROUP-- */
    ReturnNumCerts = 0;
    pReturnCertList[ReturnNumCerts].Length = pSubjectCert->Length;
    if (IsAppCall)
    {
        pReturnCertList[ReturnNumCerts].Data = 
                            App_Malloc( TPHandle, pSubjectCert->Length );
    }
    else
    {
        pReturnCertList[ReturnNumCerts].Data = 
                            Addin_malloc( pSubjectCert->Length, NULL );
    }
    memcpy( pReturnCertList[ReturnNumCerts].Data, pSubjectCert->Data, 
               pSubjectCert->Length );
    ReturnNumCerts++;

    for( counter1 = 0; bRootFound == CSSM_FALSE; counter1++ )
    {
        /* --GET ISSUER NAME FROM SUBJECT CERT-- */
        RetCode = tp_CertGetIssuerX509Name( CLHandle, pSubjectCert, &pSubjectCertIssuerName );
        if( pSubjectCertIssuerName == NULL )
        {
            for( counter2 = 0; counter2 < ReturnNumCerts; counter2++)
            {
                if (IsAppCall)
                {
                    App_Free( TPHandle, pReturnCertList[counter2].Data );
                }
                else
                {
                    Addin_free( pReturnCertList[counter2].Data , NULL);
                }
            }
            if (IsAppCall)
            {
                App_Free( TPHandle, pReturnCertList );
            }
            else
            {
                Addin_free( pReturnCertList, NULL );
            }
            /* error should already be set by the CL module */
            return( RetCode );
        }
  
        /* --CHECK IF THE SUBJECT IS SELF-SIGNED-- */
        if( tp_IsSelfSignedCert( TPHandle, CLHandle, pSubjectCert ) == CSSM_TRUE )
        {
            bRootFound = CSSM_TRUE;
            pIssuerCert = pSubjectCert;          
        }

        /* --FIND ISSUER CERT-- */
        /* look for issuer cert in CertGroupFrag */
        if( pIssuerCert == NULL && bRootFound != CSSM_TRUE )
        {
            for( counter2 = 0; 
                    counter2 < CertGroupFrag->NumCerts && pIssuerCert == NULL; 
                    counter2++ )
            {
                pTempEncodedCert = &CertGroupFrag->GroupList.CertList[counter2];
                {
                    CSSM_DATA_PTR tempSubjectName;
                    RetCode = tp_CertGetSubjectX509Name( CLHandle, 
                                                pTempEncodedCert, &tempSubjectName);
					if (RetCode != CSSM_OK)
					{
						return( RetCode );
					}

                    if( tp_IsSameX509Name( tempSubjectName, pSubjectCertIssuerName ) == CSSM_TRUE )
                    {
                        pIssuerCert = pTempEncodedCert;
                    }
                
                    if (tempSubjectName != NULL)
                    {
                        if (tempSubjectName->Data != NULL)
                        {
                            crtgrpcnst_free(TPHandle, tempSubjectName->Data, NULL);
                        }
                        crtgrpcnst_free(TPHandle, tempSubjectName, NULL);
                    }                       
                }
            }
        }

        /* not in CertGroupFrag, check DB's in DBList */
        if( (pIssuerCert == NULL && bRootFound != CSSM_TRUE) && DBList != NULL )
        {
            /* db search set-up */
            pEncSubjectCertIssuerName = pSubjectCertIssuerName;
            if( pEncSubjectCertIssuerName == NULL )
            {
                for( counter2 = 0; counter2 < ReturnNumCerts; counter2++)
                {
                    if (IsAppCall)
                    {
                        App_Free( TPHandle, pReturnCertList[counter2].Data );
                    }
                    else
                    {
                        Addin_free( pReturnCertList[counter2].Data, NULL );
                    }
                }
                if (IsAppCall)
                {
                    App_Free( TPHandle, pReturnCertList );
                }
                else
                {
                    Addin_free( pReturnCertList, NULL );
                }

                return( RetCode );
            }

            RetCode = tp_GetCertFromDBListBySName( TPHandle,
                                                 CLHandle, 
                                                 pEncSubjectCertIssuerName, 
                                                 DBList,
                                                 &pIssuerCert);
			if (RetCode != CSSM_OK)
			{
				return( RetCode );
			}
      
            /* if a cert was retrieved from db set db flag */
            if( pIssuerCert != NULL )
                bDBCert = CSSM_TRUE;
        }
    
        /* done with SubjectCertIssuerName, so free it */
        App_Free( TPHandle, pSubjectCertIssuerName->Data );
        App_Free( TPHandle, pSubjectCertIssuerName );

        /* --CHECK FOR SUCCESS OF OBTAINING NEXT LINK-- */
        if( pIssuerCert == NULL && bRootFound != CSSM_TRUE )
        {
            /* could not find next link in cert chain */
            /* clean-up memory */
            for( counter3 = 0; counter3 < ReturnNumCerts; counter3++)
            {
                if (IsAppCall)
                {
                    App_Free( TPHandle, pReturnCertList[counter3].Data );
                }
                else
                {
                    Addin_free( pReturnCertList[counter3].Data, NULL );
                }
            }
            if (IsAppCall)
            {
                App_Free( TPHandle, pReturnCertList ); 
            }
            else
            {
                Addin_free( pReturnCertList, NULL ); 
            }
            return( CSSMERR_TP_CERTGROUP_INCOMPLETE );
        }
    
        /* --COPY ISSUER CERT INTO RETURN CERT LIST-- */
        if( bRootFound != CSSM_TRUE )
        {
            ReturnNumCerts++;  
      
            /* if inside original bounds, copy */
            if( ReturnNumCerts+1 < MAX_CHAIN_DEPTH ) 
            {
                pReturnCertList[ReturnNumCerts-1].Length = pIssuerCert->Length;

                if (IsAppCall)
                {
                    pReturnCertList[ReturnNumCerts-1].Data = 
                                App_Malloc( TPHandle, pIssuerCert->Length );
                }
                else 
                {
                    pReturnCertList[ReturnNumCerts-1].Data = 
                                Addin_malloc( pIssuerCert->Length, NULL );
                }
                /* mhs -- what if malloc returns NULL */
                memcpy( pReturnCertList[ReturnNumCerts-1].Data, pIssuerCert->Data, 
                     pIssuerCert->Length );
            }
            else /* else exceeded bounds of initial memory structure, so expand */
            {
                if (IsAppCall)
                {
                    pReturnCertList = (CSSM_DATA_PTR)
                            App_Realloc( TPHandle, pReturnCertList, 
                                    ReturnNumCerts*sizeof(CSSM_DATA_PTR) );
                }
                else
                {
                    pReturnCertList = (CSSM_DATA_PTR)
                            Addin_realloc( pReturnCertList, 
                                    ReturnNumCerts*sizeof(CSSM_DATA_PTR), NULL );
                }

                if( pReturnCertList == NULL )
                {
                    /* memory cleanup */
                    if( bDBCert == CSSM_TRUE )
                    {      
                        App_Free( TPHandle, pIssuerCert->Data );
                        App_Free( TPHandle, pIssuerCert );
                    }
                    for( counter3 = 0; counter3 < ReturnNumCerts; counter3++)
                    {
                        if (IsAppCall)
                        {
                            App_Free( TPHandle, pReturnCertList[counter3].Data );
                        }
                        else
                        {
                            Addin_free( pReturnCertList[counter3].Data, NULL );
                        }
                    }
                    if (IsAppCall)
                    {
                        App_Free( TPHandle, pReturnCertList ); 
                    }
                    else
                    {
                        Addin_free( pReturnCertList, NULL ); 
                    }
                    /* set error - memory error */
                    return( CSSMERR_TP_MEMORY_ERROR );
                }
  
                pReturnCertList[ReturnNumCerts-1].Length = pIssuerCert->Length;
                if (IsAppCall)
                {
                    pReturnCertList[ReturnNumCerts-1].Data = 
                            App_Malloc( TPHandle, pIssuerCert->Length );
                }
                else
                {
                    pReturnCertList[ReturnNumCerts-1].Data = 
                            Addin_malloc( pIssuerCert->Length, NULL );
                }
                memcpy( pReturnCertList[ReturnNumCerts].Data, pIssuerCert->Data, 
                     pIssuerCert->Length );
            }
        }

        /* memory cleanup */
        if( bDBCert == CSSM_TRUE && pIssuerCert != NULL )
        {
            App_Free( TPHandle, pIssuerCert->Data );
            App_Free( TPHandle, pIssuerCert );
        }
    
        /* setup for next iteration */
        bDBCert = CSSM_FALSE;
        /* subject is now the last cert in the return list */
        pSubjectCert = &pReturnCertList[ReturnNumCerts-1];
        pIssuerCert = NULL;
    }

    /* --CAPTURE INDEX TO ROOT-- */
    RootIndex = ReturnNumCerts-1;

    /* --TO DO-- */
    /* --COPY UNUSED CERTS TO THE END OF THE RETURN CERT LIST-- */

    /* --PACK RETURN CERTGROUP-- */
    if (IsAppCall)
    {
        pReturnCertGroup = (CSSM_CERTGROUP_PTR)App_Malloc( TPHandle, sizeof(CSSM_CERTGROUP));
#if 0
        pReturnReserved = (CSSM_TP_X509V3_GRP_INFO_PTR)App_Malloc( TPHandle, 
                                                    sizeof(CSSM_TP_X509V3_GRP_INFO_PTR));
#endif /* mhs */
        pReturnCertList = App_Realloc( TPHandle, pReturnCertList, sizeof(CSSM_DATA)*ReturnNumCerts);
        if( pReturnCertGroup == NULL || 
#if 0
            pReturnReserved == NULL || 
#endif /* mhs */
            pReturnCertList == NULL )
        {
        /* clean-up memory */
        for( counter3 = 0; counter3 < ReturnNumCerts; counter3++)
        {
            App_Free( TPHandle, pReturnCertList[counter3].Data );
        }
        App_Free( TPHandle, pReturnCertList ); 
#if 0
        App_Free( TPHandle, pReturnReserved ); 
#endif /* mhs */
        App_Free( TPHandle, pReturnCertGroup ); 
        return( CSSMERR_TP_MEMORY_ERROR );
        }
    }
    else
    {
        pReturnCertGroup = (CSSM_CERTGROUP_PTR)Addin_malloc( 
                                                    sizeof(CSSM_CERTGROUP), NULL );
#if 0
        pReturnReserved = (CSSM_TP_X509V3_GRP_INFO_PTR)Addin_malloc( 
                                                    sizeof(CSSM_TP_X509V3_GRP_INFO_PTR),
                                                    NULL);
#endif /* mhs */
        pReturnCertList = Addin_realloc( pReturnCertList, 
                                                    sizeof(CSSM_DATA)*ReturnNumCerts,
                                                    NULL);
        if( pReturnCertGroup == NULL || 
#if 0
            pReturnReserved == NULL || 
#endif /* mhs */
            pReturnCertList == NULL )
        {
            /* clean-up memory */
            for( counter3 = 0; counter3 < ReturnNumCerts; counter3++)
            {
                Addin_free( pReturnCertList[counter3].Data, NULL );
            }
            Addin_free( pReturnCertList, NULL );
#if 0            
            Addin_free( pReturnReserved, NULL ); 
#endif /* mhs */
            Addin_free( pReturnCertGroup, NULL ); 
            return( CSSMERR_TP_MEMORY_ERROR );
        }
    }

#if 0
    pReturnReserved->IndexToRootCert = RootIndex;
#endif /* mhs */
    pReturnCertGroup->CertType = CSSM_CERT_X_509v3;
    pReturnCertGroup->CertEncoding = CSSM_CERT_ENCODING_UNKNOWN;
    pReturnCertGroup->NumCerts = ReturnNumCerts;
    pReturnCertGroup->GroupList.CertList = pReturnCertList;
    pReturnCertGroup->CertGroupType = CSSM_CERTGROUP_DATA;
    *ppCertGroup = pReturnCertGroup;

    /* --RETURN CONSTRUCTED GROUP-- */
    return( RetCode );
}

