/*-----------------------------------------------------------------------
 * File: CRLVRFY.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 


// suppress warnings in the runtime headers
#ifndef VMS
#pragma warning (disable:4201 4514 4214 4115)
#endif
#ifdef WIN32
#include <windows.h>
#endif
#include <stdio.h>
#ifndef VMS
#pragma warning (default:4201 4514 4214 4115)
#endif

/* CDSA includes */
#include "cssm.h"               /* CSSM types, errors, and APIs */
#include "cssmspi.h"            /* CSSM service provider interface */
#include "cssmtpi.h"            /* CSSM trust policy interface */

/* Intel includes */
#include "oidsx509.h"           /* x.509 oid defs */
#include "oidsalg.h"            /* algorithm OID defs */
#include "algoid.h"
#include "maf_include.h"

/* local includes */
#include "inx509tp.h"           /* func and data definitions 
                                   specific to this module */
#include "api_shim.h"
#include "Chn_util.h"

#include <cssmMemTrack.h>

/*---------------------------------------------------------------
 *
 * Name: TP_CrlVerify
 *
 * Description:
 *   Verify the signtaure of the CRL given the inputs.  Maps 
 *   directly to a call to CSSM_CL_CrlVerify().
 *
 * Parameters: 
 *   TPHandle (input)       - Handle of the TP module
 *   CLHandle (input)       - Handle of the CL module
 *   CSPHandle (input)      - Handle of the CSP module
 *   DBList (input)         - List of DLDBHandles to databases
 *                            [unused - must be NULL]
 *   CrlToBeVerified (input)     
 *                          - pointer to the CRL that signature
 *                            needs to be verified.
 *   SignerCertGroup (input)     
 *                          - pointer to the certificate group that 
 *                            signed the SubjectCert certificate.
 *   VerifyScope (input)    - pointer to data that describes the 
 *                            verification scope.
 *                            [unused - must be NULL]
 *   ScopeSize (input)      - number of entries in VerifyScope
 *                            [unused - must be zero]
 *
 * Returns:
 *   CSSM_FALSE             - unable to verify signature
 *   CSSM_TRUE              - Crl verified successfully
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI 
CrlVerify ( CSSM_TP_HANDLE TPHandle,
                       CSSM_CL_HANDLE CLHandle,
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_ENCODED_CRL *CrlToBeVerified,
                        const CSSM_CERTGROUP *SignerCertGroup,
                        const CSSM_TP_VERIFY_CONTEXT *VerifyContext,
                        CSSM_TP_VERIFY_CONTEXT_RESULT_PTR RevokerVerifyResult )
{
    CSSM_DATA_PTR pSignerCert = NULL;
    CSSM_CC_HANDLE hCC = 0;
    CSSM_KEY_PTR pSignerCertPublicKey = NULL;
    CSSM_DATA_PTR pTempData = NULL;
    CSSM_HANDLE ResultsHandle = 0;
    uint32 SignatureAlgId = 0;
    uint32 NumberOfFields = 0;

    CSSM_RETURN Rtn = CSSM_OK;

    /* unused parameters */
    TPHandle;
    VerifyContext;
    RevokerVerifyResult;

    /* --GET SIGNER CERT FROM SIGNER CERT GROUP */
    pSignerCert = SignerCertGroup->GroupList.CertList;
	if( pSignerCert == NULL )
    {
        return( CSSMERR_TP_INVALID_CERTGROUP_POINTER );
    }

    /* --GET PUB KEY FOR CRL ISSUER-- */
    Rtn = addin_cl_CertGetKeyInfo( CLHandle, pSignerCert, &pSignerCertPublicKey );
    if( Rtn != CSSM_OK )
    { 
        /* error set by CL */
        return( Rtn );
    }

    /* --GET ALGID FOR CRL SIGNATURE-- */
    Rtn = addin_cl_CrlGetFirstFieldValue( CLHandle, &CrlToBeVerified->CrlBlob, 
                                               &CSSMOID_X509V1SignatureAlgorithm,
                                               &ResultsHandle, &NumberOfFields,
                                               &pTempData);
    if( Rtn != CSSM_OK )
    {
        App_Free( CLHandle, pSignerCertPublicKey->KeyData.Data );
        App_Free( CLHandle, pSignerCertPublicKey );
        /* error set by CL */
        return( Rtn);
    }
  
    SignatureAlgId = X509_AlgorithmOidToAlgId( (CSSM_OID_PTR)pTempData );
    App_Free( CLHandle, pTempData->Data );
    App_Free( CLHandle, pTempData );
    if( SignatureAlgId == 0 )
    {
        App_Free( CLHandle, pSignerCertPublicKey->KeyData.Data );
        App_Free( CLHandle, pSignerCertPublicKey );
        /* error set by X509_AlgorithmOidToAlgId */
        return( CSSMERR_TP_INVALID_SIGNATURE );
    }
  
    Rtn = addin_csp_CreateSignatureContext( CSPHandle, 
                                            SignatureAlgId, 
                                            NULL, 
                                            pSignerCertPublicKey, &hCC );
    /* cleanup unneeded memory */
    App_Free( CLHandle, pSignerCertPublicKey->KeyData.Data );
    App_Free( CLHandle, pSignerCertPublicKey );
    if( Rtn != CSSM_OK )
    {
        return( Rtn );
    }

    /* --CALL CSSM_CL_CrlVerify() AND STORE RESULT FOR RETURN-- */
    Rtn = addin_cl_CrlVerify( CLHandle, hCC, 
                                &CrlToBeVerified->CrlBlob, pSignerCert, NULL, 0 );

    /* delete context */
    addin_DeleteContext( hCC );

    return( Rtn );
}


CSSM_RETURN CSSMAPI 
TP_CrlVerify ( CSSM_TP_HANDLE TPHandle,
                       CSSM_CL_HANDLE CLHandle,
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_ENCODED_CRL *CrlToBeVerified,
                        const CSSM_CERTGROUP *SignerCertGroup,
                        const CSSM_TP_VERIFY_CONTEXT *VerifyContext,
                        CSSM_TP_VERIFY_CONTEXT_RESULT_PTR RevokerVerifyResult )
{
    const ADDIN_LOAD_DATA *pLoadData = NULL;

	if ( CrlToBeVerified == NULL )
	{
		return ( CSSMERR_TP_INVALID_CRL_POINTER );
	}
	if (!tp_IsValidCrlType(CrlToBeVerified->CrlType))
	{
		return ( CSSMERR_TP_INVALID_CRL_TYPE );
	}
	if (!tp_IsValidCrlEncoding(CrlToBeVerified->CrlEncoding))
	{
		return ( CSSMERR_TP_INVALID_CRL_ENCODING );
	}

    /* Do we know our execution context? */
    pLoadData = Addin_GetLoadData();
    if ( pLoadData == NULL )
    {
        return CSSM_FALSE;
    }

    /* --CHECK INPUTS-- */
    if (SignerCertGroup == NULL)
    { /* SignerCertGroup must be specified */
        return( CSSMERR_TP_INVALID_CERTGROUP_POINTER );
    }

    return CrlVerify(TPHandle, CLHandle, CSPHandle, CrlToBeVerified,
                        SignerCertGroup, VerifyContext, RevokerVerifyResult);
}
