/*-----------------------------------------------------------------------
 * File: MEF_INTEG.C
 *
 * Copyright (c) 1999-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

/*
 * This file implements EMM/CSSM interfaces.
 */

#include "precomp.h"

static cssm_SWMRLock_PTR g_IntegLock = NULL;


#ifdef ISL_INCLUDED
#include "callouts.h"
#ifdef VMS
#include "intel/rootkeys.h"
#include "intel/rootnames.h"
#include "intel/modselfkey.h"
#else
#include "rootkeys.h"
#include "rootnames.h"
#include "modselfkey.h"
#endif

/* Maximum length of name/key buffers */
#define MAX_ISSUER_INFO_BUFFER  1024
#define MAX_CERTS               8


static ISL_VERIFIED_MODULE_PTR* VerifiedCSSMModulePtrArr= NULL;
static uint32 NumCSSMSections = 0;

static ISL_VERIFIED_MODULE_PTR EMMVerifiedModulePtr = NULL;

#else

static CSSM_LIB_HANDLE EMMLibHandle = 0;

#endif  /* ISL_INCLUDED */



CSSM_RETURN
mef_InitIntegrity()
{
    CSSM_RETURN err = CSSM_OK;

    g_IntegLock = local_malloc(sizeof(cssm_SWMRLock),
                               NULL);

    if(NULL == g_IntegLock)
    {
        ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
    }
    else
    {
        ERR(err = cssm_SWMRLockCreate(g_IntegLock,
                                      NULL));
    }

    return err;
}



CSSM_RETURN
mef_FiniIntegrity()
{
    CSSM_RETURN err = CSSM_OK;

    if(NULL != g_IntegLock)
    {
        /* make sure no one else has the integrity lock */
        ERR(err = cssm_SWMRLockWaitToWrite(g_IntegLock,
                                           CSSM_INFINITE_WAIT));
        if(CSSM_OK == err)
        {
            ERR(err = cssm_SWMRLockDoneWriting(g_IntegLock));

            if(CSSM_OK == err)
            {
                /* release the system resources for the lock */
                ERR(err = cssm_SWMRLockDelete(g_IntegLock));

                if(CSSM_OK == err)
                {
                    /* free the memory used for the lock */
                    local_free(g_IntegLock,
                               NULL);

                    g_IntegLock = NULL;
                }
            }
        }
    }
    else
    {
        /* this shouldn't occur during correct operation */
        ERR(err = CSSMERR_CSSM_FUNCTION_FAILED);
    }

    return err;
}



#ifndef ISL_INCLUDED

CSSM_RETURN
mef_SelfIntegrityCheck()
{
    CSSM_RETURN err = CSSM_OK;
    CSSM_DATA Credential;
    CSSM_DATA ModuleName;
    CSSM_DATA ModulePath;

    ERR(err = cssm_SWMRLockWaitToWrite(g_IntegLock,
                                       CSSM_INFINITE_WAIT));
    if(CSSM_OK == err)
    {
        if(0 == EMMLibHandle)
        {
            /* we need to get a handle to this library */
            ERR(err = mef_MDSGetCredentialInfo(&EMM_GUID,
                                               MDS_CDSADIR_EMM_RECORDTYPE,
                                               &Credential,
                                               &ModuleName,
                                               &ModulePath));
            if(CSSM_OK == err)
            {
                /* from the module path/name we can get a lib handle */
                char FullModName[CSSM_MAX_PATH];
                memset(FullModName,
                       0,
                       CSSM_MAX_PATH);

                /* note that the lengths given here are data lengths, not
                   string lengths
                */
                if(CSSM_MAX_PATH >= (ModuleName.Length + ModulePath.Length + 1))
                {
                    if(ModulePath.Data[ModulePath.Length-2] == DIRECTORY_SEPARATOR)
                    {
                        sprintf(FullModName,
                                "%s%s",
                                ModulePath.Data,
                                ModuleName.Data);
                    }
                    else
                    {
                        sprintf(FullModName,
                                "%s%c%s",
                                ModulePath.Data,
                                DIRECTORY_SEPARATOR,
                                ModuleName.Data);
                    }

                    ERR(err = port_GetModuleHandle(FullModName,
                                                   &EMMLibHandle));
                }
                else
                {
                    /* full name is longer than legal */
                    ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
                }

                ERR(mef_MDSReleaseCredentialInfo(&Credential,
                                                 &ModuleName,
                                                 &ModulePath));
            }
        }

        cssm_SWMRLockDoneWriting(g_IntegLock);
    }

    return err;
}



CSSM_RETURN
mef_SelfIntegrityClean()
{
    CSSM_RETURN err = CSSM_OK;

    ERR(err = cssm_SWMRLockWaitToWrite(g_IntegLock,
                                       CSSM_INFINITE_WAIT));
    if(CSSM_OK == err)
    {
        EMMLibHandle = 0;

        cssm_SWMRLockDoneWriting(g_IntegLock);
    }

    return err;
}



CSSM_RETURN
mef_VerifyCSSM(const CSSM_GUID *CssmGuid,
               CSSM_PROC_ADDR CssmRetAddress)
{
    return CSSM_OK;
}



CSSM_RETURN
mef_VerifyCallerIsCSSM(CSSM_PROC_ADDR CssmRetAddress)
{
    return CSSM_OK;
}

CSSM_RETURN
mef_CleanVerifiedCSSM(CSSM_PROC_ADDR CssmRetAddress)
{
    return CSSM_OK;
}

CSSM_RETURN
mef_LocateProcAddrInSelf(CSSM_FUNC_NAME_ADDR_PTR FuncNameAddrPtr,
                         uint32 NumOfFuncNameAddr)
{
    CSSM_RETURN err = CSSM_OK;
    uint32 i;

    ERR(err = cssm_SWMRLockWaitToRead(g_IntegLock,
                                      CSSM_INFINITE_WAIT));
    if(CSSM_OK == err)
    {
        if(0 != EMMLibHandle)
        {
            i = 0;
            while(i < NumOfFuncNameAddr)
            {
                ERR(err = port_GetProcAddress(EMMLibHandle,
                                              FuncNameAddrPtr[i].Name,
                                              &(FuncNameAddrPtr[i].Address)));
                i++;
            }
        }
        else
        {
            ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
        }

        ERR(cssm_SWMRLockDoneReading(g_IntegLock));
    }

    return err;
}


#else /* ISL_INCLUDED */


static CSSM_RETURN
mef_VerifyCredentials(ISL_CONST_DATA Credentials,
                      ISL_CONST_DATA SearchPath,
                      CSSM_KEY_HIERARCHY KeyHierarchy,
                      ISL_VERIFIED_SIGNATURE_ROOT_PTR* pVerifiedRoot);


static CSSM_RETURN
mef_GetRootKeysAndNames(CSSM_KEY_HIERARCHY KeyHierarchy,
                        ISL_CONST_DATA_PTR *ppRootKeys,
                        ISL_CONST_DATA_PTR *ppRootNames,
                        uint32 *pNumItems);


static void
mef_FreeRootKeysAndNames(uint32 NumItems,
                         ISL_CONST_DATA_PTR pRootKeys,
                         ISL_CONST_DATA_PTR pRootNames);


static CSSM_RETURN
mef_VerifyLoadedModules(const CSSM_GUID *Guid,
                        MEF_MODULE_TYPE ModuleType,
                        ISL_VERIFIED_SIGNATURE_ROOT_PTR pVerifiedRoot,
                        ISL_VERIFIED_MODULE_PTR** VerModPtrArr,
                        uint32 *NumCreds);



CSSM_RETURN
mef_SelfIntegrityCheck()
{
    CSSM_RETURN err = CSSM_OK;

    ERR(err = cssm_SWMRLockWaitToWrite(g_IntegLock,
                                       CSSM_INFINITE_WAIT));
    if(CSSM_OK == err)
    {
        if(NULL == EMMVerifiedModulePtr)
        {
            EMMVerifiedModulePtr = EISL_SelfCheck();

            if(NULL == EMMVerifiedModulePtr)
            {
                ERR(err = CSSMERR_CSSM_SELF_CHECK_FAILED);
            }
        }
        else
        {
            if(0 == EISL_ContinueVerification(EMMVerifiedModulePtr,
                                              1))
            {
                ERR(err = CSSMERR_CSSM_SELF_CHECK_FAILED);
            }
        }

        cssm_SWMRLockDoneWriting(g_IntegLock);
    }

    return err;
}



CSSM_RETURN
mef_SelfIntegrityClean()
{
    CSSM_RETURN err = CSSM_OK;

    ERR(err = cssm_SWMRLockWaitToWrite(g_IntegLock,
                                       CSSM_INFINITE_WAIT));
    if(CSSM_OK == err)
    {
        if(NULL != EMMVerifiedModulePtr)
        {
            EISL_RecycleVerifiedModuleCredentials(EMMVerifiedModulePtr);

            EMMVerifiedModulePtr = NULL;
        }

        cssm_SWMRLockDoneWriting(g_IntegLock);
    }

    return err;
}



CSSM_RETURN
mef_VerifyCSSM(const CSSM_GUID *CssmGuid,
               CSSM_PROC_ADDR CssmRetAddress)
{
    CSSM_RETURN err = CSSM_OK;
    CSSM_BOOL AddrFound;
    CSSM_BOOL CredentialsCreated = CSSM_FALSE;
    uint32 i;
    CSSM_DATA Credential;
    CSSM_DATA ModuleName;
    CSSM_DATA ModulePath;

    ISL_VERIFIED_MODULE_PTR* tempVerModPtrArr = NULL;
    uint32 tempNumCreds = 0;

    ERR(err = cssm_SWMRLockWaitToWrite(g_IntegLock,
                                       CSSM_INFINITE_WAIT));
    if(CSSM_OK == err)
    {
        if(NULL == VerifiedCSSMModulePtrArr)
        {
            /* Fetch the information from MDS */
            ERR(err = mef_MDSGetCredentialInfo(CssmGuid,
                                               MDS_OBJECT_RECORDTYPE,
                                               &Credential,
                                               &ModuleName, 
                                               &ModulePath));
            if(CSSM_OK == err)
            {
                /* first we'll check CSSM on the integrity hierarchy */
                ERR(err = mef_VerifyModule(CssmGuid,
                                           &Credential,
                                           &ModuleName,
                                           &ModulePath,
                                           CSSM_KEY_HIERARCHY_INTEG,
                                           MEF_MT_CSSM,
                                           &tempVerModPtrArr,
                                           &tempNumCreds));
                if(CSSM_OK != err)
                {
                    /* We'll check it on the export key hierarchy if there was
                       an error.  All we really care is that the CSSM verifies
                       on some key that we trust.
                    */
                    ERR(err = mef_VerifyModule(CssmGuid,
                                               &Credential,
                                               &ModuleName,
                                               &ModulePath,
                                               CSSM_KEY_HIERARCHY_EXPORT,
                                               MEF_MT_CSSM,
                                               &tempVerModPtrArr,
                                               &tempNumCreds));
                }

                if(CSSM_OK == err)
                {
                    VerifiedCSSMModulePtrArr = tempVerModPtrArr;
                    NumCSSMSections = tempNumCreds;
                    CredentialsCreated = CSSM_TRUE;
                }

                ERR(mef_MDSReleaseCredentialInfo(&Credential,
                                                 &ModuleName,
                                                 &ModulePath));
            }
        }
        else
        {
            i = 0;

            while((CSSM_OK == err) &&
                  (i < NumCSSMSections))
            {
                if(0 == EISL_ContinueVerification(VerifiedCSSMModulePtrArr[i],
                                                  1))
                {
                    ERR(err = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);
                }
                i++;
            }
        }

        /* now we verify that the caller is actually the trusted CSSM */
        if(CSSM_OK == err)
        {
            i = 0;
            AddrFound = CSSM_FALSE;

            while((CSSM_FALSE == AddrFound) &&
                  (i < NumCSSMSections))
            {
                if(ISL_OK == EISL_CheckAddressWithinModule(VerifiedCSSMModulePtrArr[i],
                                                           (ISL_FUNCTION_PTR)CssmRetAddress))
                {
                    AddrFound = CSSM_TRUE;
                }

                i++;
            }

            if(CSSM_FALSE == AddrFound)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

        ERR(cssm_SWMRLockDoneWriting(g_IntegLock));

        /* if we shouldn't have added these credentials, we delete them here */
        if((CSSM_OK != err) &&
           (CSSM_TRUE == CredentialsCreated))
        {
            ERR(mef_CleanVerifiedCSSM(CssmRetAddress));
        }
    }

    return err;
}



CSSM_RETURN
mef_CleanVerifiedCSSM(CSSM_PROC_ADDR CssmRetAddress)
{
    CSSM_RETURN err = CSSM_OK;

    ERR(err = cssm_SWMRLockWaitToWrite(g_IntegLock,
                                       CSSM_INFINITE_WAIT));
    if(CSSM_OK == err)
    {
        if(NULL != VerifiedCSSMModulePtrArr)
        {
            EISL_RecycleVerifiedModuleCredentials(VerifiedCSSMModulePtrArr[0]);

            local_free(VerifiedCSSMModulePtrArr,
                       NULL);
            VerifiedCSSMModulePtrArr = NULL;
        }

        ERR(cssm_SWMRLockDoneWriting(g_IntegLock));
    }

    return err;
}



CSSM_RETURN
mef_VerifyCallerIsCSSM(CSSM_PROC_ADDR CssmRetAddress)
{
    CSSM_RETURN err = CSSM_OK;
    CSSM_BOOL AddrFound;
    uint32 i;

    ERR(err = cssm_SWMRLockWaitToRead(g_IntegLock,
                                      CSSM_INFINITE_WAIT));
    if(CSSM_OK == err)
    {
        i = 0;
        AddrFound = CSSM_FALSE;

        while((CSSM_FALSE == AddrFound) &&
              (i < NumCSSMSections))
        {
            if(ISL_OK == EISL_CheckAddressWithinModule(VerifiedCSSMModulePtrArr[i],
                                                       (ISL_FUNCTION_PTR)CssmRetAddress))
            {
                AddrFound = CSSM_TRUE;
            }

            i++;
        }

        if(CSSM_FALSE == AddrFound)
        {
            ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
        }

        ERR(cssm_SWMRLockDoneReading(g_IntegLock));
    }

    return err;
}



CSSM_RETURN
mef_LocateProcAddrInSelf(CSSM_FUNC_NAME_ADDR_PTR FuncNameAddrPtr,
                         uint32 NumOfFuncNameAddr)
{
    CSSM_RETURN err = CSSM_OK;
    ISL_CONST_DATA FuncName;
    ISL_FUNCTION_PTR FuncAddr;
    uint32 i;

    ERR(err = cssm_SWMRLockWaitToRead(g_IntegLock,
                                      CSSM_INFINITE_WAIT));
    if(CSSM_OK == err)
    {
        if(NULL != EMMVerifiedModulePtr)
        {
            i = 0;
            while(i < NumOfFuncNameAddr)
            {
                FuncName.Data = (const uint8*)FuncNameAddrPtr[i].Name;
                FuncName.Length = strlen(FuncNameAddrPtr[i].Name) + 1;
                FuncAddr = EISL_LocateProcedureAddress(EMMVerifiedModulePtr,
                                                       FuncName);
                if(NULL != FuncAddr)
                {
                    FuncNameAddrPtr[i].Address = (CSSM_PROC_ADDR)FuncAddr;
                }

                i++;
            }
        }
        else
        {
            ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
        }

        ERR(cssm_SWMRLockDoneReading(g_IntegLock));
    }

    return err;
}



CSSM_RETURN
mef_VerifyModule(const CSSM_GUID *Guid,
                 const CSSM_DATA *Credential,
                 const CSSM_DATA *ModuleName,
                 const CSSM_DATA *ModulePath,
                 CSSM_KEY_HIERARCHY KeyHierarchy,
                 MEF_MODULE_TYPE ModuleType,
                 ISL_VERIFIED_MODULE_PTR** VerModPtrArr,
                 uint32 *NumCreds)
{
    CSSM_RETURN err = CSSM_OK;
    ISL_VERIFIED_SIGNATURE_ROOT_PTR pVerifiedRoot = NULL;

    ISL_CONST_DATA tempCredentials;
    ISL_CONST_DATA tempSearchPath;

    tempCredentials.Data = Credential->Data;
    tempCredentials.Length = Credential->Length;
    tempSearchPath.Data = ModulePath->Data;
    tempSearchPath.Length = ModulePath->Length;

    ERR(err = mef_VerifyCredentials(tempCredentials,
                                    tempSearchPath,
                                    KeyHierarchy,
                                    &pVerifiedRoot));
    if(CSSM_OK == err)
    {
        ERR(err = mef_VerifyLoadedModules(Guid,
                                          ModuleType,
                                          pVerifiedRoot,
                                          VerModPtrArr,
                                          NumCreds));

        if(CSSM_OK != err)
        {
            EISL_RecycleVerifiedSignatureRoot( pVerifiedRoot );
        }
    }

    return err;
}



static CSSM_RETURN
mef_VerifyCredentials(ISL_CONST_DATA Credentials,
                      ISL_CONST_DATA SearchPath,
                      CSSM_KEY_HIERARCHY KeyHierarchy,
                      ISL_VERIFIED_SIGNATURE_ROOT_PTR* pVerifiedRoot)
{
    CSSM_RETURN err = CSSM_OK;
    ISL_VERIFIED_SIGNATURE_ROOT_PTR SigRoot = NULL;
    ISL_VERIFIED_CERTIFICATE_PTR    SigCert = NULL;
    uint32 NumCerts = 0;
    uint32 NumKeys=0;
    uint32 i;
    ISL_CONST_DATA_PTR RootKeys = NULL;
    ISL_CONST_DATA_PTR RootNames = NULL;
    ISL_VERIFIED_CERTIFICATE_CHAIN_PTR  CertChain;
    ISL_VERIFIED_CERTIFICATE_PTR    Certs[MAX_CERTS];

    ERR(err = mef_GetRootKeysAndNames(KeyHierarchy, 
                                      &RootKeys,
                                      &RootNames,
                                      &NumKeys));
    if(CSSM_OK == err)
    {
        i = 0;
        CertChain = NULL;

        /* Attempt to build a verified cert chain from one of the roots */
        while((i < NumKeys) &&
              (NULL == CertChain))
        {
            CertChain = EISL_CreateCertificateChainWithCredentialData(RootNames[i],
                                                                      RootKeys[i],
                                                                      Credentials,
                                                                      SearchPath);
            i++;
        }

        if(NULL == CertChain)
        {
            ERR(err = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);
        }
        else
        {
            /* Find the leaf cert by copying the chain and fetching the last
             * cert. We don't have to free the "copied" chain because it is
             * referenced by the first cert chain.
             */
            NumCerts = EISL_CopyCertificateChain(CertChain,
                                                 Certs,
                                                 MAX_CERTS);
            if(0 == NumCerts)
            {
                ERR(err = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);
            }
            else
            {
                /* The signer cert if the last in the list */
                SigCert = Certs[NumCerts - 1];

                /* Create the signature root */
                SigRoot = EISL_CreateVerifiedSignatureRootWithCredentialDataAndCertificate(Credentials,
                                                                                           SearchPath,
                                                                                           SigCert);
                if(NULL == SigRoot)
                {
                    ERR(err = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);
                }

            } /* EISL_CopyCertificateChain */

            /* Delete the certificate chain */
            EISL_RecycleVerifiedCertificateChain(CertChain);

        } /* EISL_CreateCertificateChain... */

        mef_FreeRootKeysAndNames(NumKeys,
                                 RootKeys,
                                 RootNames);
    }

    /* Set the return values */
    *pVerifiedRoot = SigRoot;

    return err;
}



static CSSM_RETURN
mef_GetRootKeysAndNames(CSSM_KEY_HIERARCHY KeyHierarchy,
                        ISL_CONST_DATA_PTR *ppRootKeys,
                        ISL_CONST_DATA_PTR *ppRootNames,
                        uint32 *pNumItems)
{
    CSSM_RETURN err = CSSM_OK;
    uint32 NumKeys = 0;
    ISL_CONST_DATA_PTR pRootKeys = NULL;
    ISL_CONST_DATA_PTR pRootNames = NULL;
    uint32 i;


    /* get the number of keys/names */
    if(CSSM_KEY_HIERARCHY_INTEG == KeyHierarchy)
    {
        NumKeys = cssm_GetIntegrityRootKeys(NULL);
    }
    else
    {
        if(CSSM_KEY_HIERARCHY_EXPORT == KeyHierarchy)
        {
            NumKeys = cssm_GetExportRootKeys(NULL);
        }
        else
        {
            ERR(err = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);
        }
    }

    if(CSSM_OK == err)
    {
        if(0 == NumKeys)
        {
            ERR(err = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);
        }
        else
        {
            pRootKeys = local_calloc(sizeof(ISL_CONST_DATA),
                                     NumKeys,
                                     NULL);
            pRootNames = local_calloc(sizeof(ISL_CONST_DATA),
                                      NumKeys,
                                      NULL);

            if((NULL == pRootNames) ||
               (NULL == pRootKeys))
            {
                ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
            }
            else
            {
                i = 0;

                while((i < NumKeys) &&
                      (CSSM_OK == err))
                {
                    pRootKeys[i].Length = MAX_ISSUER_INFO_BUFFER;
                    pRootKeys[i].Data = local_calloc(MAX_ISSUER_INFO_BUFFER,
                                                     1,
                                                     NULL);
                    pRootNames[i].Length = MAX_ISSUER_INFO_BUFFER;
                    pRootNames[i].Data = local_calloc(MAX_ISSUER_INFO_BUFFER,
                                                      1,
                                                      NULL);

                    if((NULL == pRootKeys[i].Data) ||
                       (NULL == pRootNames[i].Data))
                    {
                        ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                    }

                    i++;
                } /* for i = 0 to NumKeys-1 */

                if(CSSM_OK == err)
                {
                    if(CSSM_KEY_HIERARCHY_INTEG == KeyHierarchy)
                    {
                        cssm_GetIntegrityRootKeys((ISL_CALLOUT_DATA*)pRootKeys);
                        cssm_GetIntegrityRootNames((ISL_CALLOUT_DATA*)pRootNames);
                    }
                    else
                    {
                        cssm_GetExportRootKeys((ISL_CALLOUT_DATA*)pRootKeys);
                        cssm_GetExportRootNames((ISL_CALLOUT_DATA*)pRootNames);
                    }
                }
            } /* Buffers allocated */

            if(CSSM_OK != err)
            {
                /* Free the key buffers in an error case */
                mef_FreeRootKeysAndNames(NumKeys,
                                         pRootKeys,
                                         pRootNames);
                pRootNames = NULL;
                pRootKeys = NULL;
            }
        } /* NumKeys > 0 */
    } /* Valid key hierarchy */

    /* Set return values */
    *pNumItems = NumKeys;
    *ppRootKeys = pRootKeys;
    *ppRootNames = pRootNames;

    return err;
}



static void
mef_FreeRootKeysAndNames(uint32 NumItems,
                         ISL_CONST_DATA_PTR pRootKeys,
                         ISL_CONST_DATA_PTR pRootNames)
{
    uint32 i;

    if(NULL != pRootKeys)
    {
        for(i = 0; i < NumItems; i++)
        {
            if(NULL != pRootKeys[i].Data)
            {
                local_free((uint8*)pRootKeys[i].Data,
                           NULL);
            }
        }

        local_free(pRootKeys,
                   NULL);
    }

    if(NULL != pRootNames)
    {
        for(i = 0; i < NumItems; i++)
        {
            if(NULL != pRootNames[i].Data)
            {
                local_free((uint8*)pRootNames[i].Data,
                           NULL);
            }
        }

        local_free(pRootNames,
                   NULL);
    }

    return;
}



static CSSM_RETURN
mef_VerifyLoadedModules(const CSSM_GUID *Guid,
                        MEF_MODULE_TYPE ModuleType,
                        ISL_VERIFIED_SIGNATURE_ROOT_PTR pVerifiedRoot,
                        ISL_VERIFIED_MODULE_PTR** VerModPtrArr,
                        uint32 *NumCreds)
{
    CSSM_RETURN err = CSSM_OK;
    uint32 NumSections = 0;
    uint32 j;

    CSSM_BOOL Found;

    ISL_VERIFIED_MODULE_PTR* VerifiedModulePtrArr = NULL;
    ISL_MANIFEST_SECTION_PTR pSection = NULL;
    ISL_ITERATOR_PTR pSecIter = NULL;
    uint8 RequiredAttributeValue[MEF_PRINTABLE_GUID_LENGTH];
    uint8 SectionAttributeValue[2 * MEF_PRINTABLE_GUID_LENGTH];
    ISL_CONST_DATA ReqAttrValue;
    ISL_CONST_DATA SecAttrValue;
    ISL_CONST_DATA SecAttrName;

    if(NULL == (pSecIter = EISL_CreateManifestSectionEnumerator(pVerifiedRoot)))
    {
        ERR(err = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);
    }
    else
    {
        /* Build the section executable name */
        mef_GUIDToStr(Guid,
                      (char*)RequiredAttributeValue);
        ReqAttrValue.Length = strlen((char*)RequiredAttributeValue);
        ReqAttrValue.Data = (uint8*)RequiredAttributeValue;

        /* Build the section attribute name */
        SecAttrName.Data = (uint8*)"CDSA_GUID";
        SecAttrName.Length = strlen((char*)SecAttrName.Data);

        /* Find the section named concat("executable:",pModuleName) */
        Found = CSSM_FALSE;

        while((CSSM_FALSE == Found) &&
              (NULL != (pSection = EISL_GetNextManifestSection(pSecIter))))
        {
            SecAttrValue.Data = SectionAttributeValue;
            SecAttrValue.Length = sizeof(SectionAttributeValue);

            if(ISL_OK == EISL_FindManifestSectionAttribute(pSection,
                                                           SecAttrName,
                                                           &SecAttrValue))
            {
                if((ReqAttrValue.Length == SecAttrValue.Length) &&
                   (0 == memcmp(ReqAttrValue.Data,
                                SecAttrValue.Data,
                                ReqAttrValue.Length)))
                {
                    /* This is the correct section */
                    Found = CSSM_TRUE;
                }
            }
        }

        /* Make sure we found the section */
        if((CSSM_FALSE == Found) ||
           (NULL == pSection))
        {
            ERR(err = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);
        }
        else
        {
            /* See if we need to look for "CSSM" or some such tag in the
               manifest.  If so, go looking for it.
            */
            if(MEF_MT_OTHER != ModuleType)
            {
                /* Find the CDSA_MODULE attribute and make sure it
                 * matches the required type.
                 */
                SecAttrName.Data = (uint8*)"CDSA_MODULE";
                SecAttrName.Length = strlen( (char*)SecAttrName.Data );

                SecAttrValue.Data = (uint8*)SectionAttributeValue;
                SecAttrValue.Length = sizeof(SectionAttributeValue);

                /* Set the correct required value */
                if(MEF_MT_EMM != ModuleType)
                {
                    ReqAttrValue.Data = (uint8*)"CSSM";
                }
                else
                {
                    ReqAttrValue.Data = (uint8*)"EMM";
                }

                ReqAttrValue.Length = strlen((char*)ReqAttrValue.Data);

                if(ISL_OK == EISL_FindManifestSectionAttribute(pSection,
                                                               SecAttrName,
                                                               &SecAttrValue))
                {
                    if((SecAttrValue.Length != ReqAttrValue.Length) ||
                       (0 != memcmp(ReqAttrValue.Data,
                                    SecAttrValue.Data,
                                    ReqAttrValue.Length)))
                    {
                        ERR(err = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);
                    }
                }
                else
                {
                    ERR(err = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);
                }
            }
        }

        /* Delete the section iterator and get possibly get a new one for
           cycling on the actual verifications
        */
        EISL_RecycleManifestSectionEnumerator( pSecIter );

        if(CSSM_OK == err)
        {
            pSecIter = EISL_CreateManifestSectionEnumerator(pVerifiedRoot);

            if(NULL == pSecIter)
            {
                ERR(err = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);
            }
            else
            {
                NumSections = 0;
                while((CSSM_OK == err) &&
                      (NULL != (pSection = EISL_GetNextManifestSection(pSecIter))))
                {
                    NumSections++;

                    if(1 == NumSections)
                    {
                        VerifiedModulePtrArr = (ISL_VERIFIED_MODULE_PTR*)local_calloc((uint32)sizeof(ISL_VERIFIED_MODULE_PTR),
                                                                                      1,
                                                                                      NULL);
                    }
                    else
                    {
                        VerifiedModulePtrArr = (ISL_VERIFIED_MODULE_PTR*)local_realloc(VerifiedModulePtrArr,
                                                                                       (uint32)(sizeof(ISL_VERIFIED_MODULE_PTR) * NumSections),
                                                                                       NULL);
                    }

                    if(VerifiedModulePtrArr == NULL)
                    {
                        ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                    }
                    else
                    {
                        VerifiedModulePtrArr[NumSections-1] = EISL_VerifyLoadedModule(pSection);

                        if(NULL == VerifiedModulePtrArr[NumSections-1])
                        {
                            ERR(err = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);

                            for(j = 0; j < NumSections-1; j++)
                            {
                                EISL_RecycleVerifiedModuleCredentials(VerifiedModulePtrArr[j]);
                            }

                            local_free(VerifiedModulePtrArr,
                                       NULL);
                            VerifiedModulePtrArr = NULL;
                        }
                    }
                }

                if(CSSM_OK != err)
                {
                    NumSections = 0;
                }

                /* recycle the manifest section iterator */
                EISL_RecycleManifestSectionEnumerator(pSecIter);
            }
        }

        *VerModPtrArr = VerifiedModulePtrArr;
        *NumCreds = NumSections;

    } /* EISL_CreateManifestSectionEnumerator */

    return err;
}



/*************************************************************************/
/***** Functions Required by EISL VerifyAndLoad APIs *********************/
/*************************************************************************/


/* These functions return the default root name and key values. The MEF
 * always supplies these to the functions, so these are just place holders.
 */

int EISL_RetrieveRootIssuerKey( ISL_CALLOUT_DATA *Name)
{
    return 0;
}



int EISL_RetrieveRootIssuerName( ISL_CALLOUT_DATA *Name)
{
    return 0;   
}




/*************************************************************************/
/***** Functions Required by EISL Self-Check *****************************/
/*************************************************************************/

void
EISL_RetrieveSelfCheckSectionName(ISL_CALLOUT_DATA *Name)
{
    long len = strlen(EMM_SELF_CHECK_SECTION);

    if(len <= Name->length)
    {
        strcpy((char*)Name->value,
               EMM_SELF_CHECK_SECTION);

        Name->length = len;
    }
    else
    {
        Name->length = 0;
    }

    return; 
}



void
EISL_RetrieveSelfCheckCredentials(ISL_CALLOUT_DATA *Name,
                                  ISL_CALLOUT_DATA *Path)
{
    CSSM_RETURN err = CSSM_OK;
    CSSM_DATA Credential;
    CSSM_DATA ModulePath;
    CSSM_DATA ModuleName;

    if((NULL != Name) &&
       (NULL != Path))
    {
        if(CSSM_OK == ERR(err = mef_MDSGetCredentialInfo(&EMM_GUID,
                                                         MDS_CDSADIR_EMM_RECORDTYPE,
                                                         &Credential,
                                                         &ModuleName,
                                                         &ModulePath)))
        {
            memcpy(Name->value,
                   Credential.Data,
                   Credential.Length);
            Name->length = Credential.Length;

            memcpy(Path->value,
                   ModulePath.Data,
                   ModulePath.Length);
            Path->length = ModulePath.Length;

            ERR(err = mef_MDSReleaseCredentialInfo(&Credential,
                                                   &ModuleName,
                                                   &ModulePath));
        }

        if(CSSM_OK != err)
        {
            /* we've had an error, so we set the return values to indicate
               that EISL shouldn't trust the data
            */
            Name->length = 0;
            Path->length = 0;
        }
    }

    return;
}



void
EISL_RetrieveSelfCheckCredentialsSize(uint32 *Size)
{
    CSSM_RETURN err = CSSM_OK;
    CSSM_DATA Credential;
    CSSM_DATA ModulePath;
    CSSM_DATA ModuleName;


    if(NULL != Size)
    {
        if(CSSM_OK == ERR(err = mef_MDSGetCredentialInfo(&EMM_GUID,
                                                         MDS_CDSADIR_EMM_RECORDTYPE,
                                                         &Credential,
                                                         &ModuleName,
                                                         &ModulePath)))
        {
            *Size = Credential.Length;

            ERR(err = mef_MDSReleaseCredentialInfo(&Credential,
                                                   &ModuleName,
                                                   &ModulePath));
        }

        if(CSSM_OK != err)
        {
            /* we've had an error, so we set the return values to 0 */
            *Size = 0;
        }
    }

    return;
}

#endif /* ISL_INCLUDED */
