/*-----------------------------------------------------------------------
 * File: mod_manage.c
 *
 * Copyright (c) 1999-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

/*
 * INTEL CONFIDENTIAL 
 * This file, software, or program is supplied under the terms of a 
 * license agreement or nondisclosure agreement with Intel Corporation 
 * and may not be copied or disclosed except in accordance with the 
 * terms of that agreement. This file, software, or program contains 
 * copyrighted material and/or trade secret information of Intel 
 * Corporation, and must be treated as such. Intel reserves all rights 
 * in this material, except as the license agreement or nondisclosure 
 * agreement specifically indicate. 
 */


/*
 * This file implements the module management functions for an HRS addin based
 * on MAF.
 */


#include "precomp.h"
#include <time.h>
#include "mds_util_api.h"
#include "mds_util_helper.h"

/* Declare the custom attach functions for hrs service. 
 * Function is defined in hrspwbspi.c; however, the 
 * definition of Addin_callout_ModuleAttach below requires 
 * its declaration here. */
CSSM_RETURN HRS_ModuleAttach(CSSM_MODULE_FUNCS_PTR *FuncTbl);


/*-----------------------------------------------------------------------------
 * Name: Addin_callout_LoadDataConstruct
 *
 * Description:
 * Constructs and populates the custom ADDIN_LOAD_DATA struct provided by 
 * the MAF load tracker.  For this BSP, the function attaches a CSP and saves 
 * the handle for use later when hashing and verifying passwords.
 *
 * Parameters:
 * pAddinLoadData (output) : data struct to initialize
 *
 * Note:
 * If this function fails, Addin_callout_LoadDataDestroy will still be
 * called.
 *
 * Return Code:
 * CSSM_OK if successful, CSSM_FAIL otherwise. 
 *---------------------------------------------------------------------------*/
CSSM_RETURN Addin_callout_LoadDataConstruct( ADDIN_MODULE_HANDLE hCssm,
                                             ADDIN_LOAD_DATA *   pAddinLoadData )
{	
	CSSM_RETURN				 retVal;	
	CSSM_CRYPTO_DATA		 seed					= { {0, NULL}, NULL, NULL};
	time_t					 timeSeed;
	MDSUTIL_LIST			*pCSPModuleList			= NULL;
	MDSUTIL_MODULE_INFO_PTR  pInfo					= NULL;
	MDSUTIL_SERVICE_INFO	*pServiceInfo;
	MDSUTIL_CSPSUBSERVICE	*pSubservice;
	CSSM_CONTEXT			*pCapability;
	CSSM_CONTEXT			*capabilityList;
	CSSM_CONTEXT_ATTRIBUTE  *pAttributes;
	CSSM_CONTEXT_ATTRIBUTE	 curAttribute;
	CSSM_BOOL				 CSPFound				= CSSM_FALSE;
    uint32					 capabilityTestBitmap	= 0,
							 encryptionTestBitmap	= 0,
							 keyDeriveTestBitmap	= 0;
	uint32					 randomGenBlockSize		= 0;
    uint32	                 curModuleIndex,  curServiceIndex,    curSubserviceIndex, 
			                 numCapabilities, curCapabilityIndex, curAttributeIndex;	
    
    _CSSM_ModuleLoad_PTR                  ModuleLoad;
    _CSSM_ModuleAttach_PTR                ModuleAttach;
    _CSSM_CSP_CreateRandomGenContext_PTR  CreateRandomGenContext;
    _CSSM_CSP_CreateDigestContext_PTR     CreateDigestContext;

    /* Initialize the fields of the ADDIN_LOAD_DATA to zero */
    memset((void*)pAddinLoadData, 0, sizeof(ADDIN_LOAD_DATA));
   
    /*-------------------------------------------------------------------*
     * Fetch the API entry points that ensure we are calling to the CSSM *
     * that loaded the BSP and was checked for integrity and secure      *
     * linkage                                                           *
     *-------------------------------------------------------------------*/
    if ((CSSM_OK != (retVal = Addin_GetProcAddress(
                        hCssm,
                        "CSSM_ModuleLoad",
                        (CSSM_PROC_ADDR*)&ModuleLoad)))                             ||
        
        (CSSM_OK != (retVal = Addin_GetProcAddress(
                        hCssm,
                        "CSSM_ModuleAttach",
                        (CSSM_PROC_ADDR*)&ModuleAttach)))                           ||
        
        (CSSM_OK != (retVal = Addin_GetProcAddress(
                        hCssm,
                        "CSSM_CSP_CreateRandomGenContext",
                        (CSSM_PROC_ADDR*)&CreateRandomGenContext)))                 ||
        
        (CSSM_OK != (retVal = Addin_GetProcAddress(
                        hCssm,
                        "CSSM_CSP_CreateDigestContext",
                        (CSSM_PROC_ADDR*)&CreateDigestContext)))                    ||
        
        (CSSM_OK != (retVal = Addin_GetProcAddress(
                        hCssm,
                        "CSSM_ModuleUnload",
                        (CSSM_PROC_ADDR*)&pAddinLoadData->ModuleUnload)))           ||
        
        (CSSM_OK != (retVal = Addin_GetProcAddress(
                        hCssm,
                        "CSSM_ModuleDetach",
                        (CSSM_PROC_ADDR*)&pAddinLoadData->ModuleDetach)))           ||

        (CSSM_OK != (retVal = Addin_GetProcAddress(
                        hCssm,
                        "CSSM_GenerateRandom",
                        (CSSM_PROC_ADDR*)&pAddinLoadData->GenerateRandom)))         ||

        (CSSM_OK != (retVal = Addin_GetProcAddress(
                        hCssm,
                        "CSSM_CSP_CreateDeriveKeyContext",
                        (CSSM_PROC_ADDR*)&pAddinLoadData->CreateDeriveKeyContext))) ||

        (CSSM_OK != (retVal = Addin_GetProcAddress(
                        hCssm,
                        "CSSM_DeriveKey",
                        (CSSM_PROC_ADDR*)&pAddinLoadData->DeriveKey)))              ||

        (CSSM_OK != (retVal = Addin_GetProcAddress(
                        hCssm,
                        "CSSM_CSP_CreateSymmetricContext",
                        (CSSM_PROC_ADDR*)&pAddinLoadData->CreateSymmetricContext))) ||

        (CSSM_OK != (retVal = Addin_GetProcAddress(
                        hCssm,
                        "CSSM_EncryptData",
                        (CSSM_PROC_ADDR*)&pAddinLoadData->EncryptData)))            ||

        (CSSM_OK != (retVal = Addin_GetProcAddress(
                        hCssm,
                        "CSSM_DecryptData",
                        (CSSM_PROC_ADDR*)&pAddinLoadData->DecryptData)))            ||
    
        (CSSM_OK != (retVal = Addin_GetProcAddress(
                        hCssm,
                        "CSSM_DeleteContext",
                        (CSSM_PROC_ADDR*)&pAddinLoadData->DeleteContext)))          ||
    
        (CSSM_OK != (retVal = Addin_GetProcAddress(
                        hCssm,
                        "CSSM_FreeKey",
                        (CSSM_PROC_ADDR*)&pAddinLoadData->FreeKey)))                ||
    
        (CSSM_OK != (retVal = Addin_GetProcAddress(
                        hCssm,
                        "CSSM_DigestData",
                        (CSSM_PROC_ADDR*)&pAddinLoadData->DigestData)))  )
    {
        return retVal;
    }

	/*----------------------------------*
	 * MDS Discovery of an adequate CSP *
	 *----------------------------------*/

    /* bit flags to check encryption attributes */
    #define CORRECT_KEY_SIZE		0x0001
    #define CORRECT_IV_SIZE			0x0002
    #define CORRECT_MODE			0x0004
    #define CORRECT_PADDING			0x0008
    #define CORRECT_BLOCK_SIZE		0x0010

    /* bit flags to check key derivation attributes */
    /*#define CORRECT_KEY_SIZE		0x0001  defined above*/
    #define CORRECT_KEY_TYPE		0x0002

    /* bit flags to check if encryption, key derivation, and 
     * random number generation services are available and correct */
    #define ENCRYPTION_AVAILABLE	0x0001
    #define KEY_DERIVE_AVAILABLE	0x0002
    #define RANDOM_GEN_AVAILABLE	0x0004

    /* Values to verify that the attributes of the capabilities 
     * are all available and correct */
    #define CORRECT_ENCRYPTION		(CORRECT_KEY_SIZE  | CORRECT_IV_SIZE | \
								     CORRECT_MODE      | CORRECT_PADDING | \
								     CORRECT_BLOCK_SIZE)
    #define CORRECT_KEY_DERIVE		(CORRECT_KEY_SIZE  | CORRECT_KEY_TYPE)

    #define ALL_CAPABILITIES_AVAILABLE \
	    (ENCRYPTION_AVAILABLE | KEY_DERIVE_AVAILABLE | RANDOM_GEN_AVAILABLE)

	/* Initialize the MDS */
	MDSUTIL_Init();

	/* Retrieve a list of all CSP modules available to the CSSM */
	retVal = MDSUTIL_ListModules(	CSSM_SERVICE_CSP,
									CSSM_FALSE,
									&pCSPModuleList);
	if (CSSM_OK != retVal)
		return retVal;

	if (NULL == pCSPModuleList || 0 == pCSPModuleList->NumberItems)
		/* TODO: what do we return to say no CSP's? */
		return CSSMERR_HRS_INTERNAL_ERROR;

	/* Test each module to see if it is capable of the work necessary
	 * to the BSP */
	for (curModuleIndex = 0; 
		 curModuleIndex < pCSPModuleList->NumberItems && CSSM_FALSE == CSPFound; 
		 curModuleIndex++)
	{
		/* Retrieve the current module's service and identification information */
		retVal = MDSUTIL_GetModuleInfo(	
						&(pCSPModuleList->Items[curModuleIndex].SubserviceUid.Guid),
						CSSM_SERVICE_CSP,
						pCSPModuleList->Items[curModuleIndex].SubserviceUid.SubserviceId,
						CSSM_USEE_NONE,
						&pInfo);

		if (CSSM_OK != retVal)
        {
            MDSUTIL_FreeModuleList(pCSPModuleList);
			MDSUTIL_Term();
			return retVal;
        }

		/* Check each service to see if it provides the CSP subservices needed */
		for (curServiceIndex = 0;
			 curServiceIndex < pInfo->NumberOfServices && CSSM_FALSE == CSPFound; 
			 curServiceIndex++)
		{				
			pServiceInfo = &pInfo->ServiceList[curServiceIndex];

			/* Check each subservice to see if it provides the capabilities needed */
			for (curSubserviceIndex = 0; 
				 curSubserviceIndex < pServiceInfo->NumberOfSubServices &&
				 CSSM_FALSE == CSPFound;
				 curSubserviceIndex++)
			{				
				pSubservice = (MDSUTIL_CSPSUBSERVICE*)
					pServiceInfo->SubServiceList[curSubserviceIndex].SubServiceInfo;

				numCapabilities = 0;				

				if (pSubservice->CspType == CSSM_CSP_SOFTWARE)
				{
					numCapabilities = 
						pSubservice->SubServiceInfo.
									 SoftwareCspSubService.
									 NumberOfCapabilities;
					capabilityList  = 
						pSubservice->SubServiceInfo.
									 SoftwareCspSubService.
									 CapabilityList;										
				}
				else if (pSubservice->CspType == CSSM_CSP_HARDWARE)
				{
					numCapabilities = 
						pSubservice->SubServiceInfo.
									 HardwareCspSubService.
									 NumberOfCapabilities;
					capabilityList  = 
						pSubservice->SubServiceInfo.
									 HardwareCspSubService.
									 CapabilityList;
				}
				else
					/* TODO: return MDS error? */
					return CSSMERR_HRS_INTERNAL_ERROR;

				/* Reset test bitmaps for this subservice's test */
				capabilityTestBitmap = 0;
				encryptionTestBitmap = 0;
				keyDeriveTestBitmap  = 0;

				/* Check to see if this subservice is capable of doing 
				 * everything the BSP needs */
				for (curCapabilityIndex = 0; 
					 curCapabilityIndex < numCapabilities && CSSM_FALSE == CSPFound; 
					 curCapabilityIndex++)
				{					
					pCapability = &capabilityList[curCapabilityIndex];
					pAttributes = pCapability->ContextAttributes;

					/* Check to see if this capability is the necessary
					 * encryption algorithm */
					if ((pCapability->AlgorithmType == CSSM_ALGID_DES) &&
						(pCapability->ContextType   == CSSM_ALGCLASS_SYMMETRIC))
					{
						/* Check to see if this algorithm uses the proper 
						 * modes, padding, keys, etc */
						for (curAttributeIndex = 0; 
							 curAttributeIndex < pCapability->NumberOfAttributes &&
							 CSSM_FALSE == CSPFound;
							 curAttributeIndex++)
						{
							curAttribute = pAttributes[curAttributeIndex];

							if (curAttribute.AttributeType    == CSSM_ATTRIBUTE_KEY_LENGTH && 
								curAttribute.Attribute.Uint32 == 64)
								encryptionTestBitmap |= CORRECT_KEY_SIZE;
							else if
							   (curAttribute.AttributeType    == CSSM_ATTRIBUTE_IV_SIZE &&
							    curAttribute.Attribute.Uint32 == 8 /*INIT_VECTOR_SIZE*/)
								encryptionTestBitmap |= CORRECT_IV_SIZE;
							else if
							   (curAttribute.AttributeType    == CSSM_ATTRIBUTE_MODE &&
							    curAttribute.Attribute.Uint32 == CSSM_ALGMODE_CBCPadIV8)
								encryptionTestBitmap |= CORRECT_MODE;
							else if
							   (curAttribute.AttributeType    == CSSM_ATTRIBUTE_PADDING &&
							    curAttribute.Attribute.Uint32 == CSSM_PADDING_PKCS5)
								encryptionTestBitmap |= CORRECT_PADDING;
							else if 
							   (curAttribute.AttributeType    == CSSM_ATTRIBUTE_BLOCK_SIZE &&
							    curAttribute.Attribute.Uint32 == 8)
								encryptionTestBitmap |= CORRECT_BLOCK_SIZE;
							
							if (CORRECT_ENCRYPTION == encryptionTestBitmap)
							{
								capabilityTestBitmap |= ENCRYPTION_AVAILABLE;
								break;
							}
						}
					}
					else if 
					/* Check to see if this capability is the necessary
					 * key derivation algorithm */
					   ((pCapability->AlgorithmType == CSSM_ALGID_PKCS5_PBKDF1_SHA1) &&
						(pCapability->ContextType   == CSSM_ALGCLASS_DERIVEKEY))
					{

						/* Check to see if this algorithm generates the proper
						 * key with the proper size */
						for (curAttributeIndex = 0; 
							 curAttributeIndex < pCapability->NumberOfAttributes;
							 curAttributeIndex++)
						{
							curAttribute = pAttributes[curAttributeIndex];

							if (curAttribute.AttributeType    == CSSM_ATTRIBUTE_KEY_LENGTH &&
							    curAttribute.Attribute.Uint32 == 64)
								keyDeriveTestBitmap |= CORRECT_KEY_SIZE;
							else if
							   (curAttribute.AttributeType    == CSSM_ATTRIBUTE_KEY_TYPE &&
							    curAttribute.Attribute.Uint32 == CSSM_ALGID_DES)
								keyDeriveTestBitmap |= CORRECT_KEY_TYPE;
								 	
							if (CORRECT_KEY_DERIVE == keyDeriveTestBitmap)
							{
								capabilityTestBitmap |= KEY_DERIVE_AVAILABLE;
								break;
							}
						}
					}
					else if
					/* Check to see if this capability is random number 
					 * generation of any sort */
					   (pCapability->ContextType   == CSSM_ALGCLASS_RANDOMGEN)
					{
						/* Find out the block size that is returned as necessary 
						 * for generation of salts and IV vectors later */
						for (curAttributeIndex = 0; 
							 curAttributeIndex < pCapability->NumberOfAttributes;
							 curAttributeIndex++)
						{
							curAttribute = pAttributes[curAttributeIndex];

							if (curAttribute.AttributeType == CSSM_ATTRIBUTE_BLOCK_SIZE)								
							{
								randomGenBlockSize = curAttribute.Attribute.Uint32;

								capabilityTestBitmap |= RANDOM_GEN_AVAILABLE;
								break;
							}
						}
					}
				
					/* If this subservice has all the necessary capabilities,
					 * we can stop searching */
					if (ALL_CAPABILITIES_AVAILABLE == capabilityTestBitmap)
					{
						pAddinLoadData->CSPGuid = 
							pCSPModuleList->Items[curModuleIndex].SubserviceUid.Guid;

						/* Load the module */
						retVal = ModuleLoad(
									&pAddinLoadData->CSPGuid,
									CSSM_KEY_HIERARCHY_NONE,
									NULL,
									NULL);

						if (retVal != CSSM_OK)
						{
							MDSUTIL_FreeModuleInfo(pInfo);
							MDSUTIL_FreeModuleList(pCSPModuleList);
							MDSUTIL_Term();
						
							return retVal;
						}
						
						/* Attach to the module and store the handle in the  *
						 * BSP ADDIN_LOAD_DATA struct to enable crypto calls *
						 * in the SPI implementation.                        */
						retVal = ModuleAttach(
									&pAddinLoadData->CSPGuid,
									&pInfo->CompatibleCSSMVersion,
									&Addin_APIMemFuncs,
									pSubservice->SubServiceId,								
									CSSM_SERVICE_CSP,			
									/* Attach flags */ 0,
									CSSM_KEY_HIERARCHY_NONE,	
									/* Func table   */ NULL,
									/* Num funcs    */ 0,
									/* Reserved     */ NULL,
									&pAddinLoadData->CSPHandle);

						if (retVal != CSSM_OK)
						{
							MDSUTIL_FreeModuleInfo(pInfo);
							MDSUTIL_FreeModuleList(pCSPModuleList);
							MDSUTIL_Term();
						
							return retVal;
						}

						CSPFound = CSSM_TRUE;					
					}

				}	/* capability loop */
			
			}	/* subservice loop */

		}	/* service loop */

		/* Release the current module's information from memory */
		MDSUTIL_FreeModuleInfo(pInfo);

	}	/* module loop */

	/* Release the list of CSP modules from memory */
	MDSUTIL_FreeModuleList(pCSPModuleList);

	/* Terminate the MDS session and release the context 
	 * information from memory */
	MDSUTIL_Term();


	/*-------------------------------*
	 * Create cryptographic contexts *
	 *-------------------------------*/
	
	/*-----------------------------------------------------------------------*
	 * We'll use the SHA1 algorithm to sign.  We don't need any context info *
	 * for this, so we can create and store the digest context here.  There  *
	 * is also no need for context information for the random number         *
	 * generation contexts, so we cache those here as well.  Also note that  *
     * the potential to share contexts in multithreaded access to the BSP    *
     * that is a side effect of this does *not* cause any thread's data to   *
     * corrupted; none of these operations are staged, so no one thread's    *
     * calling of a given CSP operation will interfere with another thread's *
	 *-----------------------------------------------------------------------*/
	
	/* Create a digest context for hashing captured *
	 * passwords in the SPI implementation.         */
	retVal = CreateDigestContext(	pAddinLoadData->CSPHandle,
									CSSM_ALGID_SHA1,
									&pAddinLoadData->DigestContext);
	if (retVal != CSSM_OK) return retVal;

	/* Create contexts for random number generation for the 
	 * cryptographic salts and initialization vectors */
	time(&timeSeed);
	seed.Param.Length = sizeof(time_t);
	seed.Param.Data   = (uint8*)&timeSeed;	
	retVal = CreateRandomGenContext(	pAddinLoadData->CSPHandle,
									    CSSM_ALGID_MD5Random,
								    	&seed,
										HASH_SALT_SIZE,
										&pAddinLoadData->HashRandomContext);
	if (retVal != CSSM_OK) return retVal;

	time(&timeSeed);
	seed.Param.Length = sizeof(time_t);
	seed.Param.Data   = (uint8*)&timeSeed;
	retVal = CreateRandomGenContext(	pAddinLoadData->CSPHandle,
										CSSM_ALGID_MD5Random,
										&seed,
										PBE_SALT_SIZE,
										&pAddinLoadData->KeyDeriveRandomContext);
	if (retVal != CSSM_OK) return retVal;

	time(&timeSeed);
	seed.Param.Length = sizeof(time_t);
	seed.Param.Data   = (uint8*)&timeSeed;
	retVal = CreateRandomGenContext(	pAddinLoadData->CSPHandle,
										CSSM_ALGID_MD5Random,
										&seed,
										INIT_VECTOR_SIZE,
										&pAddinLoadData->InitRandomContext);
	if (retVal != CSSM_OK) return retVal;

	return CSSM_OK;
}



/*-------------------------------------------------------------------------
 * Name: Addin_callout_LoadDataDestroy
 *
 * Description:
 * Destructs a ADDIN_LOAD_DATA structure
 *
 * Parameters: 
 * pAddinLoadData (output) : ADDIN_LOAD_DATA structure to destroy
 * 
 * Return Code:
 * none
 *-------------------------------------------------------------------------
 */
void Addin_callout_LoadDataDestroy( ADDIN_LOAD_DATA * pAddinLoadData )
{
	/*---------------------------------------------*
	 * Detach and unload CSP, and release the CSSM *
	 *---------------------------------------------*/

	pAddinLoadData->DeleteContext(pAddinLoadData->DigestContext);

	pAddinLoadData->DeleteContext(pAddinLoadData->HashRandomContext);

	pAddinLoadData->DeleteContext(pAddinLoadData->KeyDeriveRandomContext);

	pAddinLoadData->DeleteContext(pAddinLoadData->InitRandomContext);

	pAddinLoadData->ModuleDetach( pAddinLoadData->CSPHandle);

	pAddinLoadData->ModuleUnload(&pAddinLoadData->CSPGuid, NULL, NULL);						
}



/*-------------------------------------------------------------------------
 * Name: Addin_callout_AttachDataConstruct
 *
 * Description:
 * Constructs and populates the custom ADDIN_ATTACH_DATA struct provided
 * by the MAF attach tracker.  For this BSP, the function initializes
 * a list of BIRs and a counter used to give the newly created BIRs their 
 * handles.
 *
 * Parameters: 
 * pAddinAttachData (output) : data struct to initialize
 * 
 * Note:
 * If this function fails, Addin_callout_AttachDataDestroy will still be 
 * called
 *
 * Return Code:
 * CSSM_OK if successful, otherwise CSSM_FAIL
 *-------------------------------------------------------------------------
 */
CSSM_RETURN 
Addin_callout_AttachDataConstruct(ADDIN_ATTACH_DATA * pAddinAttachData )
{	
	pAddinAttachData->BIRList     = NULL;	
	pAddinAttachData->HandleIndex = 1; 
		
	return CSSM_OK;
}


/*-------------------------------------------------------------------------
 * Name: Addin_callout_AttachDataDestroy
 *
 * Description:
 * Deconstructs the addin's attach data struct.  In this BSP, 
 * this deallocates memory used for the list of BIRs residing in the
 * addin's memory space.
 *
 * Parameters: 
 * pAddinAttachData (output) : ADDIN_ATTACH_DATA structure to destroy
 * 
 * Return Code:
 * none
 *-------------------------------------------------------------------------
 */
void Addin_callout_AttachDataDestroy(ADDIN_ATTACH_DATA * pAddinAttachData )
{
	BIR_LIST_NODE* pNodeToFree;

	while (pAddinAttachData->BIRList != NULL)
	{
		pNodeToFree = pAddinAttachData->BIRList;
		pAddinAttachData->BIRList = (BIR_LIST_NODE*)pNodeToFree->NextNode;
		Addin_free(pNodeToFree->BIR.BiometricData,	NULL);
		Addin_free(pNodeToFree,						NULL);
	}
}



/*-----------------------------------------------------------------------------
 * Name: Addin_callout_Initialize
 *
 * Description:
 * Preforms initialization that takes places when the module is loaded into
 * the process space
 *
 * Parameters: 
 * None
 * 
 * Return Code:
 * CSSM_OK if successful, otherwise CSSM_FAIL
 *---------------------------------------------------------------------------*/

CSSM_RETURN
Addin_callout_Initialize(void)
{
    return port_CreateMutex(NULL, &BSPMutex);   
}



/*-----------------------------------------------------------------------------
 * Name: Addin_callout_Terminate
 *
 * Description:
 * Terminates state before the module is unloaded from the process space
 *
 * Parameters: 
 * None.
 * 
 * Return Code:
 * CSSM_OK if successful, otherwise CSSM_FAIL
 *---------------------------------------------------------------------------*/

CSSM_RETURN
Addin_callout_Terminate(void)
{
    return port_CloseMutex(BSPMutex);
}



/*-------------------------------------------------------------------------
 * Name: Addin_callout_ModuleLoad
 *
 * Description:
 * Call used to indicate that the application has requested a module load
 *
 * Parameters: 
 * pLoadTracker (input/output) : The load tracker node that was created. 
 *  This node may be writen to.
 * CssmNotifyCallback (input) : Callback function to indicate to CSSM that
 *  a subservice is ready for attach (must be called before an attach can 
 *  occurr).
 * CssmNotifyCallbackCtx (input) : Opaque data to pass to the callback
 *  function.
 * 
 * Return Code:
 * CSSM_OK if successful, otherwise CSSM_FAIL
 *-------------------------------------------------------------------------
 */

CSSM_RETURN
Addin_callout_ModuleLoad(MAF_MODULE_LOAD_TRACKER_PTR pLoadTracker,
                         CSSM_SPI_ModuleEventHandler CssmNotifyCallback,
                         const void* CssmNotifyCallbackCtx)
{
    CSSM_RETURN ret = CSSM_OK;

    ERR(ret = CssmNotifyCallback(&ADDIN_GUID,
                                 (void *)CssmNotifyCallbackCtx,
                                 0,
                                 CSSM_SERVICE_HRS,
                                 CSSM_NOTIFY_INSERT));
    return ret;
}



/*-------------------------------------------------------------------------
 * Name: Addin_callout_ModuleUnload
 *
 * Description:
 * Function is called as part of the unload process to give indication to
 * the addin.
 *
 * Parameters: 
 * pLoadTracker (input/output) : The load tracker node that was created. 
 *  This node may be written to.
 * CssmNotifyCallback (input) : Identifies the corresponding ModuleLoad 
 *  call CssmNotifyCallbackCtx (input) : Identifies the corresponding 
 *  ModuleLoad call
 * 
 * Note: 
 * There is no need to call the callback function on unload.
 *
 * Return Code:
 * CSSM_OK if successful, otherwise CSSM_FAIL
 *-------------------------------------------------------------------------
 */

CSSM_RETURN Addin_callout_ModuleUnload(
                        MAF_MODULE_LOAD_TRACKER_PTR pLoadTracker,
                        CSSM_SPI_ModuleEventHandler CssmNotifyCallback,
                        const void* CssmNotifyCallbackCtx )
{
    return CSSM_OK;
}



/*-------------------------------------------------------------------------
 * Name: Addin_callout_ModuleAttach
 *
 * Description:
 * Functions is called to indicate that a module attach has been reqested. 
 *
 * Parameters: 
 * pLoadTracker (input/output) : An initialized load tracker, the 
 *  ADDIN_LOAD_DATA structure can be modified
 * pAttachTracker (input/output) : An construcuted attach tracker, the 
 *  ADDIN_ATTACH_DATA structure can be modified
 * FuncTbl (outpt) : Buffer to hold the SPI functions that the addin wants
 * to register with CSSM
 * 
 * Return Code:
 * CSSM_OK if successful, otherwise CSSM_FAIL
 *-------------------------------------------------------------------------
 */

CSSM_RETURN Addin_callout_ModuleAttach(
                        MAF_MODULE_LOAD_TRACKER_PTR pLoadTracker, 
                        MAF_MODULE_ATTACH_TRACKER_PTR pAttachTracker,
                        CSSM_MODULE_FUNCS_PTR *FuncTbl )
{
    if (!pAttachTracker)
    {
        return CSSMERR_CSSM_INVALID_POINTER;
    }

    switch (pAttachTracker->SubServiceType)
    {
        case CSSM_SERVICE_HRS: return HRS_ModuleAttach(FuncTbl);
        default:               return CSSMERR_CSSM_INVALID_SERVICE_MASK;
    }
}
