/*-----------------------------------------------------------------------
 * File: HMACSHA1.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

/* Implements the HMAC-SHA1 construct using EICL primitives.
 * Algorithm description is found in IETF RFC2104.
 */

#include "maf_include.h"

#include "eicl.h"

#include "hmacsha1.h"


CSSM_RETURN CSSMAPI HMAC_SHA1_Begin(
                        HMAC_SHA1_CONTEXT *pContext,
                        const CSSM_DATA *Key )
{
    CSSM_RETURN rv = CSSM_OK;
    uint32 i;
    EICLData HashData;

    assert( pContext && Key && Key->Data );

    /* Clear the context to automatically set the key padding */
    memset( pContext, 0, sizeof(*pContext) );

    /* Copy the key into the Opad and Ipad buffers. Reduce the length of
     * the key if it is longer than the block size.
     */
    if ( Key->Length > HMAC_SHA1_BLOCK_SIZE )
    {
        EICLSHADigest ReducedKey;

        EICL_SHABegin( &pContext->ShaState );
        HashData.value = (EICLByte*)Key->Data;
        HashData.length = (long)Key->Length;
        EICL_SHAProcess( &HashData, &pContext->ShaState );
        EICL_SHAEnd( &pContext->ShaState, ReducedKey );

        memcpy( pContext->kIpad, ReducedKey, sizeof(ReducedKey) );
        memcpy( pContext->kOpad, ReducedKey, sizeof(ReducedKey) );
    }
    else
    {
        memcpy( pContext->kIpad, Key->Data, Key->Length );
        memcpy( pContext->kOpad, Key->Data, Key->Length );
    }

    /* Perform Ipad and Opad XOR calculations */
    for ( i = 0; i < sizeof(pContext->kIpad); i++)
    {
        pContext->kIpad[i] ^= 0x36;
        pContext->kOpad[i] ^= 0x5C;
    }

    /* Perform the initial hash update of Ipad */
    EICL_SHABegin( &pContext->ShaState );
    HashData.value = pContext->kIpad;
    HashData.length = sizeof(pContext->kIpad);
    EICL_SHAProcess( &HashData, &pContext->ShaState );

    return rv;
}

CSSM_RETURN CSSMAPI HMAC_SHA1_Update(
                        HMAC_SHA1_CONTEXT *pContext,
                        const CSSM_DATA *Data )
{
    CSSM_RETURN rv = CSSM_OK;
    EICLData HashData;

    assert( pContext && Data && Data->Data );

    /* Add the input data to the HMAC hash */
    HashData.value = (EICLByte*)Data->Data;
    HashData.length = (long)Data->Length;
    EICL_SHAProcess( &HashData, &pContext->ShaState );

    return rv;
}

CSSM_RETURN CSSMAPI HMAC_SHA1_End(
                        HMAC_SHA1_CONTEXT *pContext,
                        CSSM_DATA *Result )
{
    CSSM_RETURN rv = CSSM_OK;
    EICLSHADigest HmacState;
    EICLData HashData;

    assert( pContext &&
            Result && 
            Result->Data && 
            ( Result->Length == sizeof(HmacState) ) );

    /* Finish out the inner hash */
    EICL_SHAEnd( &pContext->ShaState, HmacState );

    /* Hash Opad with the inner hash */
    EICL_SHABegin( &pContext->ShaState );
    HashData.value = pContext->kOpad;
    HashData.length = sizeof(pContext->kOpad);
    EICL_SHAProcess( &HashData, &pContext->ShaState );
    HashData.value = HmacState;
    HashData.length = sizeof(HmacState);
    EICL_SHAProcess( &HashData, &pContext->ShaState );
    EICL_SHAEnd( &pContext->ShaState, HmacState );

    /* Copy the result to the output buffer */
    memcpy( Result->Data, HmacState, sizeof(HmacState) );

    /* Wipe the context */
    memset( pContext, 0, sizeof(*pContext) );

    return rv;
}
