/*-----------------------------------------------------------------------
 * File: MAF_MONITOR.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "maf_include.h"

/* MAF Utility Headers */
#include "maf_monitor.h"

/**-------------------------------------------------------------------------
 ** MAF_StartMonitor
 **
 ** Description
 **  Used to spawn a thread in the current process that can be used to
 **  monitor the status of a dynamic addin. The function MonitorFunc is
 **  specified by the caller and will be called by the new thread. The
 **  addin must have a way to signal the thread to shut down. The new thread
 **  will be created with the lowest possible priority.
 **
 ** Parameters
 **  StackSize (input) - The size of the initial stack for the thread. A
 **   size of zero indicates the same size as the main process thread.
 **  MonitorFunc (input) - Address of the function to be called by the new
 **   thread.
 **  phMonitor (output) - Pointer to a monitor handle that will recieve a
 **   reference to the new monitor thread.
 **
 ** Return Value
 **  CSSM_INVALID_POINTER - The phMonitor or MonitorFunc pointers are
 **   invalid for writing or execution respectively.
 **  MAFR_THREAD_ERROR - Creation or manipulation of a new thread failed.
 **  CSSM_OK - The new monitor thread has been started.
 *--------------------------------------------------------------------------
 **/
CSSM_RETURN MAF_StartMonitor(
                        port_ThreadProc MonitorFunc,
                        void *MonitorFuncParam,
                        PORT_THREAD_HANDLE *phMonitor )
{
    CSSM_RETURN rv = CSSM_OK;
    PORT_THREAD_HANDLE hThread;
     

    /* Check the input */
    if ( ( port_IsBadWritePtr( phMonitor,
                               sizeof(PORT_THREAD_HANDLE) ) == CSSM_TRUE ) ||
         ( port_IsBadCodePtr( (CSSM_PROC_ADDR)MonitorFunc ) == CSSM_TRUE ) )
    {
        return CSSM_ERRCODE_INVALID_POINTER;
    }

    /* Spawn the thread */
    rv = port_CreateThread( MonitorFunc,
                            MonitorFuncParam,
                            CSSM_FALSE,
                            &hThread );
    if ( rv == CSSM_OK )
    {
        *phMonitor = hThread;
    }

    return rv;
}

/**-------------------------------------------------------------------------
 ** MAF_WaitOnMonitor
 **
 ** Description
 **  Given a monitor handle, this function will wait until the corresponding
 **  thread completes or TimeOut milliseconds have passed. A timeout of zero
 **  will cause the function to check whether or not the thread is still
 **  running. 
 **
 ** Parameters
 **  hMonitor (input) - Handle of the monitor thread on which to block..
 **  TimeOut (input) - Number of milliseconds to block. A timeout of zero
 **   indicates a "peek" operation. A timeout of
 **   MAF_MONITOR_TIMEOUT_INFINITE indicates an indefinite wait.
 **
 ** Return Value
 **  MAFR_TIMEOUT - The timeout period expired before the monitor thread
 **   terminated.
 **  MAFR_INVALID_HANDLE - The monitor handle has never been issued or the
 **   handle has been waited on already.
 **  CSSM_OK - The new monitor thread has terminated.
 *-------------------------------------------------------------------------
 **/
CSSM_RETURN MAF_WaitOnMonitor(
                        PORT_THREAD_HANDLE hMonitor,
                        uint32 *pReturnCode )
{
    CSSM_RETURN rv = CSSM_OK;
    uint32 RetCode;

    /* Wait for the thread handle to become signalled. This happens when
     * the thread terminates.
     */
    rv = port_JoinThread( (HANDLE)hMonitor, &RetCode );
    if ( rv == CSSM_OK )
    {
        if ( pReturnCode )
        {
            *pReturnCode = RetCode;
        }
    }

    return rv;
}


