/*-----------------------------------------------------------------------
 * File: KEYUTIL.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "ber_der.h"
#include "keyutil.h"
#include "CSM/csm_keyparse.h"
#include "util/maf_misc.h"

#include "cpsexp.h"
#include "clutil.h"
#include "mempool.h"
#define MAX_PUB_KEY_LEN 1024

extern MEMORY_POOL_FUNCS memory_pool_funcs;
extern MEMORY_POOL_PTR poolptr;

static CSSM_RETURN copyCSSMDATA(CSSM_DATA *dst, const CSSM_DATA *src)
{
    if (!src || ! dst)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    port_memset(dst, 0, sizeof(CSSM_DATA));
    if (!src->Length || ! src->Data)
    {
        return CSSM_OK;
    }
    dst->Length = src->Length;
    dst->Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func,
                    src->Length,
                    poolptr);
    if (!dst->Data)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    port_memcpy(dst->Data, src->Data, dst->Length);
    return CSSM_OK;
}
CSSM_RETURN PKListTOCssmKey(CSSM_CSP_HANDLE CSPHandle,
                            CSSM_LIST *PKList, 
                            CSSM_KEY_PTR keyptr)
{
    CSSM_LIST_ELEMENT_PTR w;
    CSSM_LIST keylist;
    CSSM_RETURN ret = CSSM_OK;

    if (!PKList)
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
    }
    if (!keyptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    
    keyptr->KeyHeader.HeaderVersion = CSSM_KEYHEADER_VERSION;
    /* keyptr->KeyHeader.CspId is unknown (remains 0) */
    keyptr->KeyHeader.BlobType = CSSM_KEYBLOB_RAW;
    keyptr->KeyHeader.KeyClass = CSSM_KEYCLASS_PUBLIC_KEY;
    /* keyptr->KeyHeader.KeyAttr remains 0 */
    /* keyptr->KeyHeader.KeyUsage is unknown  (set to any usage) */
    keyptr->KeyHeader.KeyUsage = CSSM_KEYUSE_ANY;
    /* keyptr->KeyHeader.StartDate is unknown  (remains 0) */
    /* keyptr->KeyHeader.EndDate is unknown  (remains 0) */
    keyptr->KeyHeader.WrapAlgorithmId = CSSM_ALGID_NONE;
    keyptr->KeyHeader.WrapMode = CSSM_ALGMODE_NONE;


    w = PKList->Head;
    if (!w)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    w = w->NextElement;
    if (!w || w->ElementType != CSSM_LIST_ELEMENT_SUBLIST)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    keylist = w->Element.Sublist;

    w = keylist.Head;
    if (!w)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    switch(w->WordID)
    {
        case CSSM_WORDID_RSA_PKCS1_MD5:
            
        case CSSM_WORDID_RSA_PKCS1_SHA1:
            
        case CSSM_WORDID_RSA_PKCS1:
            keyptr->KeyHeader.AlgorithmId = CSSM_ALGID_RSA;
            keyptr->KeyHeader.Format = CSSM_KEYBLOB_RAW_FORMAT_PKCS1;
            ret = RSAKeyListToCssmKey(CSPHandle, &keylist, keyptr);
            break;

        case CSSM_WORDID_DSA_SHA1:
            keyptr->KeyHeader.AlgorithmId = CSSM_ALGID_DSA;

            keyptr->KeyHeader.Format = CSSM_KEYBLOB_RAW_FORMAT_FIPS186;
            ret = DSAKeyListToCssmKey(CSPHandle, &keylist, keyptr);
            break;
        default:
            return CSSMERR_CL_INVALID_DATA;
    }
        
    if (CSSM_OK != ret)
    {
        return ret;
    }
    ret = ObtainEffectiveKeySize(keyptr, &keyptr->KeyHeader.LogicalKeySizeInBits);
    if (CSSM_OK != ret)
    {
        /* Free local memory */
        return ret;
    }

    return CSSM_OK;

}


/*-----------------------------------------------------------------------------
 * Name: ObtainEffectiveKeySize
 *
 * Description:
 * Obtains the effective keysize (in bits) of the input key
 * 
 * Parameters: 
 * Key (input): CSSM_KEY whose effective keysize has to be determined
 * EffKeySize (output): Effective Key size in bits
 *---------------------------------------------------------------------------*/
CSSM_RETURN ObtainEffectiveKeySize (CSSM_KEY_PTR Key,
                                    uint32 *EffKeySize)
{


    CSSM_RETURN ret = CSSM_OK;

    CSSM_DATA KeyParts[CSM_MAX_PUBLIC_KEY_PARTS];
    uint32 NumKeyParts = sizeof(KeyParts) / sizeof(CSSM_DATA);


    if (!EffKeySize)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    *EffKeySize = 0;

    /* Clear the key part array */
    port_memset( KeyParts, 0, sizeof(KeyParts) );

    /* Parse the key parts */
    ret = CSM_ParseKeyblob( Key->KeyHeader.AlgorithmId,
                                 Key->KeyHeader.KeyClass,
                                 Key->KeyHeader.Format,
                                 KeyParts,
                                 &NumKeyParts,
                                 &Key->KeyData );
    if ( ret == CSSM_OK )
    {
        switch ( Key->KeyHeader.AlgorithmId )
        {
        case CSSM_ALGID_RSA:
            {
                *EffKeySize = KeyParts[CSM_RSA_MODULUS].Length * 8;
                break;
            }
        case CSSM_ALGID_DSA:
            {
                *EffKeySize = KeyParts[CSM_DSA_PRIME].Length * 8;
                break;
            }
        default:
            {
                /* Assume the length of the first parameter */
                *EffKeySize = KeyParts[0].Length * 8;
                break;
            }
        }
    }
    else
    {
        *EffKeySize = 0;
    }

    
    return CSSM_OK;
}

CSSM_RETURN DSAKeyListToCssmKey(CSSM_CSP_HANDLE CSPHandle,
                                CSSM_LIST *keylist, 
                                CSSM_KEY_PTR keyptr)
{
    CSSM_LIST_ELEMENT_PTR w,y;
    CSSM_DATA KeyParts[CSM_NUMBER_DSA_PUBLIC_KEY_PARTS];
    uint32 NumKeyParts = CSM_NUMBER_DSA_PUBLIC_KEY_PARTS;
    

    CSSM_BOOL bAllocatedKeyblob = CSSM_FALSE;

    w = keylist->Head;
    if (!w)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    w = w->NextElement;
    /* p  CSM_DSA_PRIME */
    if (!w)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    y = w->Element.Sublist.Head;
    if (!y)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    y = y->NextElement;
    if (!y)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    KeyParts[CSM_DSA_PRIME] = y->Element.Word;

    /* g CSM_DSA_BASE */
    w = w->NextElement;
    if (!w)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    y = w->Element.Sublist.Head;
    if (!y)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    y = y->NextElement;
    if (!y)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    KeyParts[CSM_DSA_BASE] = y->Element.Word;

    /* q CSM_DSA_SUB_PRIME */
    w = w->NextElement;
    if (!w)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    y = w->Element.Sublist.Head;
    if (!y)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    y = y->NextElement;
    if (!y)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    KeyParts[CSM_DSA_SUB_PRIME] = y->Element.Word;

    /* pub_key y CSM_DSA_PUBLIC_VALUE */
    w = w->NextElement;
    if (!w)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    y = w->Element.Sublist.Head;
    if (!y)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    y = y->NextElement;
    if (!y)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    KeyParts[CSM_DSA_PUBLIC_VALUE] = y->Element.Word;

    

    keyptr->KeyData.Length = MAX_PUB_KEY_LEN;
    keyptr->KeyData.Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func,
                    keyptr->KeyData.Length,
                    poolptr);
    if (!keyptr->KeyData.Data)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
        return CSM_CreateKeyblob(MAF_ALLOCATE_LOCAL,
                            CSSM_ALGID_DSA,
                            keyptr->KeyHeader.KeyClass,
                            keyptr->KeyHeader.Format,
                            KeyParts,
                            NumKeyParts,
                            &keyptr->KeyData,
                            &bAllocatedKeyblob );




}

CSSM_RETURN RSAKeyListToCssmKey(CSSM_CSP_HANDLE CSPHandle,
                                CSSM_LIST *keylist, 
                                CSSM_KEY_PTR keyptr)
{
    CSSM_LIST_ELEMENT_PTR w,y;
    CSSM_DATA KeyParts[CSM_NUMBER_RSA_PUBLIC_KEY_PARTS];
    uint32 NumKeyParts = CSM_NUMBER_RSA_PUBLIC_KEY_PARTS;
    


    CSSM_BOOL bAllocatedKeyblob = CSSM_FALSE;


    w = keylist->Head;
    if (!w)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    w = w->NextElement;
    /* e CSM_RSA_PUBLIC_EXPONENT */
    if (!w)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    y = w->Element.Sublist.Head;
    if (!y)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    y = y->NextElement;
    if (!y)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    KeyParts[CSM_RSA_PUBLIC_EXPONENT] = y->Element.Word;


    w = w->NextElement;
    /* n CSM_RSA_MODULUS */
    if (!w)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    y = w->Element.Sublist.Head;
    if (!y)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    y = y->NextElement;
    if (!y)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    KeyParts[CSM_RSA_MODULUS] = y->Element.Word;


    keyptr->KeyData.Length = MAX_PUB_KEY_LEN;
    keyptr->KeyData.Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func,
                    keyptr->KeyData.Length,
                    poolptr);
    if (!keyptr->KeyData.Data)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    
    return  CSM_CreateKeyblob(MAF_ALLOCATE_LOCAL,
                        CSSM_ALGID_RSA,
                        keyptr->KeyHeader.KeyClass,
                        keyptr->KeyHeader.Format,
                        KeyParts,
                        NumKeyParts,
                        &keyptr->KeyData,
                        &bAllocatedKeyblob );
    




}

CSSM_RETURN CssmKeyToSexp(CSSM_KEY_PTR keyptr, 
                          CSSM_LIST_PTR KeyAsList,
                          CSSM_DATA *result)
{
    CSSM_DATA KeyParts[CSM_MAX_KEY_PARTS];
    uint32 NumKeyParts = sizeof(KeyParts) / sizeof(CSSM_DATA);
    CSSM_RETURN ret = CSSM_OK;
    CSSM_LIST_ELEMENT_PTR w;
    CSSM_LIST pub_key;
    CSSM_LIST params;
    CSSM_DATA_PTR tmp;

    CSSM_LIST p, g, q, y;
    CSSM_LIST e, n;

    if (!keyptr)
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
    }
    if (!result && !KeyAsList)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    if (result)
    {
        result->Length = 0;
        result->Data = NULL;
    }
    if (KeyAsList)
    {
        init_list(KeyAsList);
    }
    init_list(&p);
    init_list(&g);
    init_list(&q);
    init_list(&y);
    init_list(&e);
    init_list(&n);

    ret = CSM_ParseKeyblob( keyptr->KeyHeader.AlgorithmId,
                            keyptr->KeyHeader.KeyClass,
                            keyptr->KeyHeader.Format,
                            KeyParts,
                            &NumKeyParts,
                            &keyptr->KeyData );
    if (CSSM_OK != ret)
    {
        return ret;
    }
    if (keyptr->KeyHeader.AlgorithmId == CSSM_ALGID_SHA1WithDSA ||
        keyptr->KeyHeader.AlgorithmId == CSSM_ALGID_DSA)
    {
        /* DSA/SHA1 key */
                
        ret = cl_lit_sexp(&p, (uint8 *)"(1:p1:#)");
        if (CSSM_OK != ret)
        {
            return ret;
        }
        
        ret = cl_lit_sexp(&g, (uint8 *)"(1:g1:#)");
        if (CSSM_OK != ret)
        {
            drop_list(&p);
            return ret;
        }
        

        ret = cl_lit_sexp(&q, (uint8 *)"(1:q1:#)");
        if (CSSM_OK != ret)
        {
            drop_list(&p);
            drop_list(&g);
            return ret;
        }
        

        ret = cl_lit_sexp(&y, (uint8 *)"(1:y1:#)");
        if (CSSM_OK != ret)
        {
            drop_list(&p);
            drop_list(&g);
            drop_list(&q);
            return ret;
        }
        if (!p.Tail || !g.Tail || !q.Tail || !y.Tail)
        {
            drop_list(&p);
            drop_list(&g);
            drop_list(&q);
            drop_list(&y);
            return CSSMERR_CL_INTERNAL_ERROR;
        }

        p.Tail->ElementType = CSSM_LIST_ELEMENT_DATUM;
        if (p.Tail->Element.Word.Data)
        {
            CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func, p.Tail->Element.Word.Data, poolptr);
        }
        ret = copyCSSMDATA(&(p.Tail->Element.Word), &(KeyParts[CSM_DSA_PRIME]));
        if (CSSM_OK != ret)
        {
            
            drop_list(&p);
            drop_list(&g);
            drop_list(&q);
            drop_list(&y);
            return ret;
        }

        g.Tail->ElementType = CSSM_LIST_ELEMENT_DATUM;
        if (g.Tail->Element.Word.Data)
        {
            CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func,g.Tail->Element.Word.Data, poolptr);
        }
        ret = copyCSSMDATA(&(g.Tail->Element.Word), &(KeyParts[CSM_DSA_BASE]));
        if (CSSM_OK != ret)
        {
            
            drop_list(&p);
            drop_list(&g);
            drop_list(&q);
            drop_list(&y);
            return ret;
        }


        q.Tail->ElementType = CSSM_LIST_ELEMENT_DATUM;
        if (q.Tail->Element.Word.Data)
        {
            CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func,q.Tail->Element.Word.Data, poolptr);
        }
        ret = copyCSSMDATA(&(q.Tail->Element.Word), &(KeyParts[CSM_DSA_SUB_PRIME]));
        if (CSSM_OK != ret)
        {
            
            drop_list(&p);
            drop_list(&g);
            drop_list(&q);
            drop_list(&y);
            return ret;
        }

        y.Tail->ElementType = CSSM_LIST_ELEMENT_DATUM;
        if (y.Tail->Element.Word.Data)
        {
            CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func,y.Tail->Element.Word.Data, poolptr);
        }
        ret = copyCSSMDATA(&(y.Tail->Element.Word), &(KeyParts[CSM_DSA_PUBLIC_VALUE]));
        if (CSSM_OK != ret)
        {
            
            drop_list(&p);
            drop_list(&g);
            drop_list(&q);
            drop_list(&y);
            return ret;
        }

        /* Lets set up the params list */
        init_list(&params);
        w = new_element();
        if (!w)
        {
            drop_list(&p);
            drop_list(&g);
            drop_list(&q);
            drop_list(&y);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        w->WordID = CSSM_WORDID_DSA_SHA1;
        w->ElementType = CSSM_LIST_ELEMENT_DATUM;
        tmp = new_data(8, (uint8*)"dsa-sha1");
        if (!tmp)
        {
            drop_list(&p);
            drop_list(&g);
            drop_list(&q);
            drop_list(&y);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        w->Element.Word = *tmp;
        CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func,tmp, poolptr);
        
        params.Head = w;
        params.Tail = w;

        w = new_element();
        if (!w)
        {
            drop_list(&p);
            drop_list(&g);
            drop_list(&q);
            drop_list(&y);
            drop_list(&params);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        w->ElementType = CSSM_LIST_ELEMENT_SUBLIST;
        w->Element.Sublist = p;

        params.Tail->NextElement = w;
        params.Tail = w;

        w = new_element();
        if (!w)
        {
            drop_list(&g);
            drop_list(&q);
            drop_list(&y);
            drop_list(&params);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        w->ElementType = CSSM_LIST_ELEMENT_SUBLIST;
        w->Element.Sublist = g;

        params.Tail->NextElement = w;
        params.Tail = w;

        w = new_element();
        if (!w)
        {
            drop_list(&q);
            drop_list(&y);
            drop_list(&params);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        w->ElementType = CSSM_LIST_ELEMENT_SUBLIST;
        w->Element.Sublist = q;

        params.Tail->NextElement = w;
        params.Tail = w;

        w = new_element();
        if (!w)
        {
            drop_list(&y);
            drop_list(&params);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        w->ElementType = CSSM_LIST_ELEMENT_SUBLIST;
        w->Element.Sublist = y;

        params.Tail->NextElement = w;
        params.Tail = w;


        /* Lets set up the whole pub_key list */
        init_list(&pub_key);
        w = new_element();
        if (!w)
        {
            drop_list(&params);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        w->WordID = CSSM_WORDID_PUBLIC_KEY;
        w->ElementType = CSSM_LIST_ELEMENT_DATUM;
        tmp = new_data(10, (uint8*)"public-key");
        if (!tmp)
        {
            drop_list(&params);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        w->Element.Word = *tmp;
        CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func,tmp, poolptr);
        
        pub_key.Head = w;
        pub_key.Tail = w;

        w = new_element();
        if (!w)
        {
            drop_list(&params);
            drop_list(&pub_key);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        w->ElementType = CSSM_LIST_ELEMENT_SUBLIST;
        w->Element.Sublist = params;

        pub_key.Tail->NextElement = w;
        pub_key.Tail = w;
    }/* if (keyptr->KeyHeader.AlgorithmId == CSSM_ALGID_SHA1WithDSA) */
    else
    {
        
        CSSM_WORDID_TYPE WordID;
        ret = cl_lit_sexp(&e, (uint8 *)"(1:e1:#)");
        if (CSSM_OK != ret)
        {
            return ret;
        }
        
        
        ret = cl_lit_sexp(&n, (uint8 *)"(1:n1:#)");
        if (CSSM_OK != ret)
        {
            drop_list(&e);
            return ret;
        }
        if (!e.Tail || !n.Tail)
        {
            drop_list(&e);
            drop_list(&n);
            return CSSMERR_CL_INTERNAL_ERROR;
        }
        e.Tail->ElementType = CSSM_LIST_ELEMENT_DATUM;
        if (e.Tail->Element.Word.Data)
        {
            CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func,e.Tail->Element.Word.Data, poolptr);
        }
        ret = copyCSSMDATA(&(e.Tail->Element.Word), &(KeyParts[CSM_RSA_PUBLIC_EXPONENT]));
        if (CSSM_OK != ret)
        {
            drop_list(&e);
            drop_list(&n);
            return ret;
        }
        

        n.Tail->ElementType = CSSM_LIST_ELEMENT_DATUM;
        if (n.Tail->Element.Word.Data)
        {
            CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func,n.Tail->Element.Word.Data, poolptr);
        }
        ret = copyCSSMDATA(&(n.Tail->Element.Word), &(KeyParts[CSM_RSA_MODULUS]));
        if (CSSM_OK != ret)
        {
            drop_list(&e);
            drop_list(&n);
            return ret;
        }
        

        /* Create params list */
        init_list(&params);
        switch (keyptr->KeyHeader.AlgorithmId)
        {
            case CSSM_ALGID_MD5WithRSA:
                tmp = new_data(13, (uint8*)"rsa-pkcs1-md5");
                WordID = CSSM_WORDID_RSA_PKCS1_MD5;
                break;
            case CSSM_ALGID_SHA1WithRSA:
                tmp = new_data(14, (uint8*)"rsa-pkcs1-sha1");
                WordID = CSSM_WORDID_RSA_PKCS1_SHA1;
                break;
            case CSSM_ALGID_RSA:
                tmp = new_data(9, (uint8*)"rsa-pkcs1");
                WordID = CSSM_WORDID_RSA_PKCS1;
                break;
            default:
                drop_list(&e);
                drop_list(&n);
                return CSSMERR_CL_INTERNAL_ERROR;

        }
        if (!tmp)
        {
            drop_list(&e);
            drop_list(&n);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        w = new_element();
        if (!w)
        {
            drop_list(&e);
            drop_list(&n);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        w->WordID = WordID;
        w->ElementType = CSSM_LIST_ELEMENT_DATUM;
        w->Element.Word = *tmp;
        CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func,tmp, poolptr);
        
        params.Head = w;
        params.Tail = w;

        w = new_element();
        if (!w)
        {
            drop_list(&e);
            drop_list(&n);
            drop_list(&params);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        w->ElementType = CSSM_LIST_ELEMENT_SUBLIST;
        w->Element.Sublist = e;

        params.Tail->NextElement = w;
        params.Tail = w;

        w = new_element();
        if (!w)
        {
            drop_list(&n);
            drop_list(&params);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        w->ElementType = CSSM_LIST_ELEMENT_SUBLIST;
        w->Element.Sublist = n;

        params.Tail->NextElement = w;
        params.Tail = w;

        /* Lets set up the whole pub_key list */
        init_list(&pub_key);
        w = new_element();
        if (!w)
        {
            drop_list(&params);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        w->WordID = CSSM_WORDID_PUBLIC_KEY;
        w->ElementType = CSSM_LIST_ELEMENT_DATUM;
        tmp = new_data(10, (uint8*)"public-key");
        if (!tmp)
        {
            drop_list(&params);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        w->Element.Word = *tmp;
        CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func,tmp, poolptr);
        
        pub_key.Head = w;
        pub_key.Tail = w;

        w = new_element();
        if (!w)
        {
            drop_list(&params);
            drop_list(&pub_key);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        w->ElementType = CSSM_LIST_ELEMENT_SUBLIST;
        w->Element.Sublist = params;

        pub_key.Tail->NextElement = w;
        pub_key.Tail = w;
        
    }/* else  (keyptr->KeyHeader.AlgorithmId == CSSM_ALGID_SHA1WithDSA) */

    /* Now we have the pub_key CSSM_LIST populated */

    if (result)
    { 
        /* user wants the packed Sexpr */

        result->Length = size_list(&pub_key);
        result->Data = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func,
                        result->Length,
                        1,
                        poolptr);
        if (!result->Data)
        {
            drop_list(&pub_key);
            return CSSMERR_CL_MEMORY_ERROR;
        }
        pack_list(&(result->Data), &pub_key);
    }
    if (KeyAsList)
    {
        /* User wants a copy of the list. 
        Give her that and do not drop the list */
        *KeyAsList = pub_key;
    }
    else
    {
        
        drop_list(&pub_key);
    }
    return CSSM_OK;
}


CSSM_RETURN AddKeyToPrincipal(uint8 *Sexp,
							  CSSM_KEY_PTR keyptr,
							  CSSM_LIST_PTR result)
{
	CSSM_RETURN ret;
	CSSM_LIST_ELEMENT_PTR w;

	if (!Sexp || !keyptr)
	{
		return CSSMERR_CL_INVALID_INPUT_POINTER;
	}
	if (!result)
	{
		return CSSMERR_CL_INVALID_OUTPUT_POINTER;
	}
	
	ret = cl_lit_sexp(result, Sexp);
	if (CSSM_OK != ret)
	{
		return ret;
	}

	/* Lets walk to the public key */
	w = result->Head;
	if (!w)
	{
		drop_list(result);
		return CSSMERR_CL_INTERNAL_ERROR;
	}

	w = w->NextElement;
	/* at pub-key node. lets nix whats in there now and insert a true key */
	if (!w)
	{
		drop_list(result);
		return CSSMERR_CL_INTERNAL_ERROR;
	}
	CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func,w->Element.Word.Data, poolptr);
	w->Element.Word.Data = NULL;
	w->Element.Word.Length = 0;

	w->ElementType = CSSM_LIST_ELEMENT_SUBLIST;
	w->WordID = CSSM_WORDID_PUBLIC_KEY;
	ret = CssmKeyToSexp(keyptr, &(w->Element.Sublist), NULL);
	if (CSSM_OK != ret)
	{
		drop_list(result);
		return ret;
	}

	return CSSM_OK;
}


