/*-----------------------------------------------------------------------
 * File: AAL.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */
 
#include "aal.h"
#include "aalintrn.h"
#ifdef ISL_INCLUDED
#include "eislapi.h"
#endif


/*-----------------------------------------------------------------------------
 * Name: AAL_LoadCSSM
 *
 * Description:  
 *  This function searches thru the Windows registry and loads the version
 *   ( and optionally vendor) of CSSM that the application requested.
 *  It checks the integrity of CSSM before loading. Since different 
 * vendor cssms can be loaded, the application needs to pass the root key of
 * trust in a x509 certificate (CSSM_DATA) to this function. 
 * EISL_VerifyAndLoadModuleAndCredentialsUsingCertificate is used to verify the
 * loaded module using the key in the certificate. If this parameter is not
 * passed, the embedded root key in the shim is used to VerifyAndLoad cssm
 * After loading CSSM, the function address requested by the applications are
 * returned in the application supplied array of Func Name/Address, and
 * verifies it belongs in the loaded cssm address range.
 * 
 * Parameters: 
 * CssmFunctable(input/output)
 * NumCssmFuncTable(input)
 * Guid (input)
 * CssmHandle (input/output)
 * TrustedKeys (input)
 * NumTrustedKeys (input)
 *
 * Returns:
 *  AAL_RETURN_CODEs. AAL_OK for success and other values for failure
 * 
 * Error Codes:
 *  AAL_OK,
    AAL_CANT_GET_VALUE, 
    AAL_CANT_OPEN_KEY,
    AAL_CANT_ENUM_KEY,
    AAL_CANT_QUERY_KEY,
    AAL_AAL_VERSION_NOT_AVAILABLE,
    AAL_CSSM_LOAD_FAIL,
    AAL_CSSM_INTEGRITY_FAIL,
    AAL_CSSM_FUNC_ADDR_INTEGRITY_FAIL,
    AAL_MEMORY_ERROR,
    AAL_INVALID_PTR

 *---------------------------------------------------------------------------*/

AAL_RETURN_CODE AAL_FUNC
AAL_LoadCssm(AAL_FUNC_TABLE_PTR CssmFuncTable,
             uint32 NumCssmFuncTable,
             CSSM_GUID_PTR Guid,
             AAL_LIB_HANDLE_PTR CssmHandle,
             CSSM_DATA_PTR TrustedKeys,
             uint32 NumTrustedKeys)
{
    /* The application needs to pass the public key of root of trust of vendor.
    Or the embedded root key is used to verify and load cssm*/

    AAL_RETURN_CODE ret ;

    if(CssmFuncTable == NULL || Guid == NULL)
        return AAL_INVALID_PTR;
    
    if( (IsBadReadPtr(CssmFuncTable, sizeof(AAL_FUNC_TABLE)*NumCssmFuncTable)) || 
        (IsBadReadPtr(Guid, sizeof(CSSM_GUID))) )
        return AAL_INVALID_PTR;
    
    if(IsBadWritePtr(CssmFuncTable,sizeof(AAL_FUNC_TABLE)*NumCssmFuncTable))
        return AAL_INVALID_PTR;

    if(IsBadWritePtr(CssmHandle, sizeof(AAL_LIB_HANDLE)))
        return AAL_INVALID_PTR;

    ret = aal_LoadCSSM(CssmFuncTable,
                       NumCssmFuncTable,
                       Guid,
                       CssmHandle,
                       TrustedKeys,
                       NumTrustedKeys);
    return ret;
}

/*-----------------------------------------------------------------------------
 * Name: AAL_UnloadCSSM
 *
 * Description:  
 * Unloads the loaded CSSM. Frees the library
 * Parameters: 
 * Handle(input)
 *
 * Returns:
 *  None
 * 
 * Error Codes:
 *---------------------------------------------------------------------------*/


void  AAL_FUNC
AAL_UnloadCssm(AAL_LIB_HANDLE CssmHandle)
{
    if(CssmHandle)
    {
#ifdef ISL_INCLUDED
#ifdef WIN32
        FreeLibrary(EISL_GetLibHandle((ISL_VERIFIED_MODULE_PTR)CssmHandle) );
#else
        port_FreeLibrary(EISL_GetLibHandle((ISL_VERIFIED_MODULE_PTR)CssmHandle) );
#endif
        EISL_RecycleVerifiedModuleCredentials((ISL_VERIFIED_MODULE_PTR)CssmHandle);
#else
#ifdef WIN32
        FreeLibrary(CssmHandle);
#else
        port_FreeLibrary(CssmHandle);
#endif
#endif
    }
}

/*-----------------------------------------------------------------------------
 * Name: AAL_GetCSSMList
 *
 * Description:  
 * Gets a list of all CSSM version/vendor infos in the System registry
 * If the VersionVendorArray is not big enough, as indicated by
 * NumOfVersionVendorInfo then the NumOfVersionVendorInfo field is updated to
 * indicate the correct size of the array needed.
 *
 * Parameters: 
 *
 * VersionVendorPtr(input/output)
 * NumOfVersionVendorInfo(input/output)
 *
 * Returns:
 *  
 * 
 * Error Codes:
 *---------------------------------------------------------------------------*/
AAL_RETURN_CODE AAL_FUNC
AAL_GetCssmList(AAL_VERSION_VENDOR_PTR CssmInfo,
                uint32* NumOfVersionVendorInfo)
{
    AAL_RETURN_CODE ret;

    if((IsBadWritePtr(NumOfVersionVendorInfo,
                      sizeof(uint32))) ||
       ((NULL != CssmInfo) &&
        (IsBadWritePtr(CssmInfo,
                       sizeof(AAL_VERSION_VENDOR) * (*NumOfVersionVendorInfo)))))
    {
        ret = AAL_INVALID_PTR;
    }
    else
    {
#ifdef VMS
        ret = aalint_GetCSSMList(CssmInfo, NumOfVersionVendorInfo);
#else
        ret = aal_GetCSSMList(CssmInfo, NumOfVersionVendorInfo);
#endif
    }

    return ret;
}

AAL_RETURN_CODE AAL_FUNC
AAL_CheckAddressWithinVerifiedModule(AAL_LIB_HANDLE LibHandle,
                                     CSSM_PROC_ADDR Address)
{
#ifdef ISL_INCLUDED
        if(Address != NULL)
        {
            if(EISL_CheckAddressWithinModule((ISL_VERIFIED_MODULE_PTR)LibHandle,
                                             (ISL_FUNCTION_PTR)Address) != ISL_OK)
            {
                return AAL_CSSM_FUNC_ADDR_INTEGRITY_FAIL;
            }
            else
                return AAL_OK;
        }
        else
            return AAL_FAIL;

#else
    return AAL_OK;
#endif
}

/*-----------------------------------------------------------------------------
 * Name: AAL_GetModuleLocation
 *
 * Description:  
 *
 * Parameters: 
 *
 * ModuleGuid(input)
 * Location(output)
 * Returns:
 *  
 * 
 * Error Codes:
 *---------------------------------------------------------------------------*/

AAL_RETURN_CODE AAL_FUNC
AAL_GetModuleLocation(CSSM_GUID_PTR ModuleGuid,
                      AAL_PATHNAME Location)
{
    AAL_RETURN_CODE ret;
#ifdef VMS
    ret = aalint_GetModuleLocation( ModuleGuid, Location);
#else
    ret = aal_GetModuleLocation( ModuleGuid, Location);
#endif
    return ret;
}

/*-----------------------------------------------------------------------------
 * Name: AAL_GetManifestInfo
 *
 * Description:  
 * Generates the Credential patha nd section from the dllpath
 * For example, if the dll path is c:\winnt\system32\cssm32.dll,
 * then the credential path will be c:\winnt\system32\cssm32.esw and
 * section name is cssm32.
 *
 * Parameters: 
 *
 * szdllPath(input)
 * szCredentialPath(input/output)
 * szSection(input/output)
 * Returns:
 *  
 * 
 * Error Codes:
 *---------------------------------------------------------------------------*/

AAL_RETURN_CODE AAL_FUNC
AAL_GetManifestInfo(CSSM_GUID_PTR ModuleGuid,
                    CSSM_DATA_PTR Manifest,
                    CSSM_DATA_PTR ManifestSearchPath,
                    AAL_PATHNAME ManifestSection)
{
    return aal_GetCredentialAndSection(ModuleGuid,
                                       Manifest,
                                       ManifestSearchPath,
                                       ManifestSection);
}

