/*-----------------------------------------------------------------------
 * File: AALMDS.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include <sys/stat.h>
#ifdef VMS
#include <string.h>
#include <stdlib.h>
#endif
#include "cssmtype.h"
#include "aal.h"
#include "aalintrn.h"
#include "mds.h"
#include "mds_schema.h"
#include "cssmerr.h"

#ifdef _DEBUG
#include <crtdbg.h>
#endif


#ifdef CSSM_MEMTRACK_ON
void* CSSMAPI
_aal_malloc_func(uint32 size,
                 void *allocRef,
                 const char * szFilename,
                 uint32 u32LineNumber)
{
    return _malloc_dbg(size, _NORMAL_BLOCK, szFilename, u32LineNumber);
}

void* CSSMAPI
_aal_calloc_func(uint32 num_elements,
                 uint32 size_element, 
                 void *allocRef,
                 const char * szFilename,
                 uint32 u32LineNumber)
{
    return _calloc_dbg(num_elements, size_element, _NORMAL_BLOCK, szFilename, u32LineNumber);
}


void* CSSMAPI
_aal_realloc_func(void * ptr, 
                  uint32 new_size, 
                  void *allocRef,
                  const char * szFilename,
                  uint32 u32LineNumber)
{
    return _realloc_dbg(ptr, new_size, _NORMAL_BLOCK, szFilename, u32LineNumber);
}

void CSSMAPI
_aal_free_func(void * ptr,
               void *allocRef,
               const char * szFilename,
               uint32 u32LineNumber)
{
    _free_dbg(ptr, _NORMAL_BLOCK);
    return;
}
#else
void* CSSMAPI
_aal_malloc_func(uint32 size,
                 void *allocRef)
{
    return malloc(size);
}

void* CSSMAPI
_aal_calloc_func(uint32 num_elements,
                 uint32 size_element, 
                 void *allocRef)
{
    return calloc(num_elements, size_element);
}


void* CSSMAPI
_aal_realloc_func(void * ptr, 
                  uint32 new_size, 
                  void *allocRef)
{
    return realloc(ptr, new_size);
}

void CSSMAPI
_aal_free_func(void * ptr,
              void *allocRef)
{
    free(ptr);
    return;
}

#endif

CSSM_MEMORY_FUNCS aal_pMemoryFunctions = {  _aal_malloc_func,
                                            _aal_free_func,
                                            _aal_realloc_func,
                                            _aal_calloc_func,
                                            NULL };


#define MDS_MAX_ATTRIBUTES  5
CSSM_DB_ATTRIBUTE_DATA s_mdsAttributes[MDS_MAX_ATTRIBUTES];



void CSSMAPI cssm_GUIDToStr (const CSSM_GUID *GUID, char *ID)
{
    if ((GUID == NULL) || (ID == NULL))
        return;

    sprintf (ID, GUID_FORMAT_STRING, 
                     GUID->Data1, GUID->Data2, GUID->Data3, GUID->Data4[0], 
                     GUID->Data4[1], GUID->Data4[2], GUID->Data4[3], 
                     GUID->Data4[4], GUID->Data4[5], GUID->Data4[6],
                     GUID->Data4[7]);
}



void CSSMAPI cssm_StrToGUID (char *ID, CSSM_GUID_PTR GUID)
{
    uint32 i, Data1;
    uint32 Data2, Data3;
    uint32 Data4[8];

    if ((ID == NULL) || (GUID == NULL))
    {
        return;
    }


    sscanf (ID, GUID_FORMAT_STRING, 
                     &Data1, &Data2, &Data3, &Data4[0],
                     &Data4[1], &Data4[2], &Data4[3],
                     &Data4[4], &Data4[5], &Data4[6],
                     &Data4[7]);
        
    GUID->Data1 = Data1;
    GUID->Data2 = (uint16) Data2;
    GUID->Data3 = (uint16) Data3;
    for (i=0; i<8; i++)
      GUID->Data4[i] = (uint8) Data4[i];    
}



AAL_RETURN_CODE aal_GetRecord(
    MDS_FUNCS pDlFunctions,
    CSSM_DL_DB_HANDLE hDLDB,
    CSSM_GUID_PTR guid,
    CSSM_DB_RECORDTYPE RecordType,
    uint32 NumAttributes,
    CSSM_DB_ATTRIBUTE_DATA *OutputAttributeData)
{
    CSSM_QUERY Query;
    CSSM_SELECTION_PREDICATE Predicate;
    CSSM_DB_UNIQUE_RECORD_PTR pRecordId;
    CSSM_DB_RECORD_ATTRIBUTE_DATA OutputAttributes;
    CSSM_DATA OutputData;
    CSSM_HANDLE ResultsHandle;
    char GuidStr[AAL_MAX_REG];
    CSSM_RETURN ret = CSSM_OK;

    /*
     * Convert the Guid to a string.
     */
    cssm_GUIDToStr (guid, GuidStr);

    memset (&Query, 0, sizeof (CSSM_QUERY));
    memset (&Predicate, 0, sizeof (CSSM_SELECTION_PREDICATE));
    Query.RecordType = RecordType;
    Query.Conjunctive = CSSM_DB_NONE;
    Query.NumSelectionPredicates = 1;
    Query.SelectionPredicate = &Predicate;
    Query.QueryLimits.TimeLimit = CSSM_QUERY_TIMELIMIT_NONE;
    Query.QueryLimits.SizeLimit = 1;
    Query.QueryFlags = 0;

    Predicate.DbOperator = CSSM_DB_EQUAL;
    
    /* NOTE: The first attribute *NEEDS* to be an index */
    Predicate.Attribute.Info = s_MdsAttrInfo_ModuleId;
    Predicate.Attribute.NumberOfValues = 1;
    Predicate.Attribute.Value = (CSSM_DATA *) aal_malloc_func(sizeof(CSSM_DATA), NULL);
    if (Predicate.Attribute.Value == NULL)
        return (CSSMERR_DL_MEMORY_ERROR);
    Predicate.Attribute.Value->Data = (unsigned char *)GuidStr;
    Predicate.Attribute.Value->Length = strlen (GuidStr) + 1;
    
    memset (&OutputAttributes, 0, sizeof (CSSM_DB_RECORD_ATTRIBUTE_DATA));
    /* We want to get all the other attributes back */
    OutputAttributes.DataRecordType = RecordType;
    OutputAttributes.NumberOfAttributes = NumAttributes;
    OutputAttributes.AttributeData = OutputAttributeData;

    ret = pDlFunctions.DataGetFirst (hDLDB, /* DLDBHandle */
                                     &Query,
                                     &ResultsHandle,
                                     &OutputAttributes,
                                     &OutputData,
                                     &pRecordId);
    pDlFunctions.DataAbortQuery( hDLDB, ResultsHandle );    

    if (ret != CSSM_OK)
        return AAL_MDS_DATA_GET_ERROR;

    if (OutputData.Data)
        aal_free_func(OutputData.Data, NULL);

    pDlFunctions.FreeUniqueRecord(hDLDB, pRecordId);
    aal_free_func(Predicate.Attribute.Value, NULL);

    return AAL_OK;
}



AAL_RETURN_CODE
aal_GetModuleInfo (CSSM_GUID_PTR guid,
                   CSSM_DATA_PTR Credential,
                   CSSM_DATA_PTR ModuleName, 
                   CSSM_DATA_PTR ModulePath)
{
    CSSM_DL_HANDLE Handle;
    CSSM_DB_HANDLE dbHandle;
    CSSM_DL_DB_HANDLE hDLDB;
    CSSM_DB_ACCESS_TYPE access = CSSM_DB_ACCESS_READ;
    CSSM_DB_ATTRIBUTE_DATA OutputAttributeData[MDS_OBJECT_NUM_ATTRIBUTES];
    MDS_FUNCS pDlFunctions;
    AAL_RETURN_CODE err = AAL_OK;
    CSSM_RETURN ret = CSSM_OK;

    if (( ret = MDS_Initialize(
                    NULL,
                    NULL,
                    &aal_pMemoryFunctions,
                    &pDlFunctions,
                    &Handle)) != CSSM_OK)
    {
        err = AAL_MDS_INITIALISE_ERROR;
    }
    else
    {
        if (( ret = pDlFunctions.DbOpen (Handle,
                                         MDS_OBJECT_DIRECTORY,
                                         NULL,
                                         access,
                                         NULL,
                                         NULL,
                                         &dbHandle)) != CSSM_OK)
        {
            err = AAL_MDS_OPEN_ERROR;
        }
        else
        {
            hDLDB.DLHandle = Handle;
            hDLDB.DBHandle = dbHandle;

            OutputAttributeData[0].Info = s_MdsAttrInfo_ModuleId;
            OutputAttributeData[1].Info = s_MdsAttrInfo_Manifest;
            OutputAttributeData[2].Info = s_MdsAttrInfo_ModuleName;
            OutputAttributeData[3].Info = s_MdsAttrInfo_Path;
            OutputAttributeData[4].Info = s_MdsAttrInfo_ProductVersion;
            if ((err = aal_GetRecord(
                                    pDlFunctions,
                                    hDLDB,
                                    guid,
                                    MDS_OBJECT_RECORDTYPE,
                                    MDS_OBJECT_NUM_ATTRIBUTES,
                                    OutputAttributeData)) == AAL_OK)
            {
                if (Credential)
                    *Credential = *(OutputAttributeData[1].Value);
                if (ModuleName)
                    *ModuleName = *(OutputAttributeData[2].Value);
                if (ModulePath)
                    *ModulePath = *(OutputAttributeData[3].Value);

                /*
                 * Since no alloc ref is defined for the aal memory functions,
                 * NULL is used. Free all the unused pointers.
                 */
                aal_free_func(OutputAttributeData[0].Value->Data, NULL);                
                aal_free_func(OutputAttributeData[0].Value, NULL);
                aal_free_func(OutputAttributeData[1].Value, NULL);
                aal_free_func(OutputAttributeData[2].Value, NULL);
                aal_free_func(OutputAttributeData[3].Value, NULL);
                aal_free_func(OutputAttributeData[4].Value->Data, NULL);
                aal_free_func(OutputAttributeData[4].Value, NULL);

            }

            /* close the database */
             pDlFunctions.DbClose (hDLDB);
        }

        /* terminate MDS */
        MDS_Terminate(Handle);
    }

    return err;
}



AAL_RETURN_CODE
aal_GetModulePath(CSSM_DATA ModuleName,
                  CSSM_DATA ModuleSearchPath,
                  AAL_PATHNAME ModulePath)
{
    struct _stat st;
    char *path;
    const char *ptr;
    int len;

    ptr = (const char *)ModuleSearchPath.Data;
    len = ModuleSearchPath.Length;
    while (ptr && len) {
        path = strchr (ptr, ';');
    
        if (path != NULL) {
            *path = '\0';
            len -= path - ptr;
        }
        strcpy (ModulePath, ptr);
        strcat (ModulePath, DIRECTORY_SEPARATOR_STRING);
        strcat (ModulePath, (char *)ModuleName.Data);

        if ((_stat(ModulePath, &st)) == 0)
            return AAL_OK;
        ptr = path;
        if (path)
            ptr++;
    }
    return AAL_FAIL;
}



void
aal_StringToVersion (char *VersionStr, AAL_VERSION_PTR Version)
{
    uint32 Major;
    uint32 Minor;
    char* p;

    Major = atoi (VersionStr);
    p = strrchr (VersionStr, '.');
    p++;
    Minor = atoi (p);
    Version->Major = Major;
    Version->Minor = Minor;
}



AAL_RETURN_CODE
aal_GetListFromMDS(
    MDS_FUNCS pDlFunctions,
    CSSM_DL_DB_HANDLE hDLDB,
    AAL_VERSION_VENDOR_PTR VersionVendorPtr,
    uint32* NumOfVersionVendorInfo)
{
    uint32 i;
    AAL_RETURN_CODE err = AAL_OK;

    CSSM_QUERY Query;
    CSSM_DB_UNIQUE_RECORD RecordId;
    CSSM_DB_UNIQUE_RECORD_PTR pRecordId;
    CSSM_DB_RECORD_ATTRIBUTE_DATA OutputAttributes;
    CSSM_DB_ATTRIBUTE_DATA OutputAttributeData[MDS_CDSADIR_CSSM_NUM_ATTRIBUTES];
    CSSM_DATA OutputData;
    CSSM_HANDLE ResultsHandle;
    uint32 numRecords = 0;
    CSSM_RETURN ret = CSSM_OK;

    memset((void*)&RecordId,
           0,
           sizeof(CSSM_DB_UNIQUE_RECORD));
    memset (&Query, 0, sizeof (CSSM_QUERY));
    Query.RecordType = MDS_CDSADIR_CSSM_RECORDTYPE;
    Query.Conjunctive = CSSM_DB_NONE;
    Query.NumSelectionPredicates = 0;
    Query.SelectionPredicate = NULL;
    Query.QueryLimits.TimeLimit = CSSM_QUERY_TIMELIMIT_NONE;
    Query.QueryLimits.SizeLimit = 1;
    Query.QueryFlags = 0;


    OutputAttributes.DataRecordType = MDS_CDSADIR_CSSM_RECORDTYPE;
    OutputAttributes.NumberOfAttributes = MDS_CDSADIR_CSSM_NUM_ATTRIBUTES;
    OutputAttributes.AttributeData = (struct cssm_db_attribute_data *) &OutputAttributeData;
    OutputAttributeData[0].Info = s_MdsAttrInfo_ModuleId;
    OutputAttributeData[1].Info = s_MdsAttrInfo_CDSAVersion;
    OutputAttributeData[2].Info = s_MdsAttrInfo_Vendor;
    OutputAttributeData[3].Info = s_MdsAttrInfo_Desc;
    OutputAttributeData[4].Info = s_MdsAttrInfo_NativeServices;

    pRecordId = &RecordId;

    /* get the first record */
    ret = pDlFunctions.DataGetFirst(hDLDB,
                                    &Query,
                                    &ResultsHandle,
                                    &OutputAttributes,
                                    &OutputData,
                                    &pRecordId);

    /* We'll keep getting records as long as we don't reach the end or
       encounter an error.
    */
    while(CSSM_OK == ret)
    {
        numRecords++;

        /* make sure we have enough room to copy the new data */
        if ((numRecords-1) < *NumOfVersionVendorInfo)
        {
            aal_StringToVersion ((char *)OutputAttributeData[1].Value->Data,
                                 &VersionVendorPtr[numRecords-1].Version);
            strncpy (VersionVendorPtr[numRecords-1].Vendor,
                     (char *)OutputAttributeData[2].Value->Data,
                     OutputAttributeData[2].Value->Length);
            cssm_StrToGUID ((char *)OutputAttributeData[0].Value->Data,
                            &VersionVendorPtr[numRecords-1].Guid);  

            for (i = 0; i < MDS_CDSADIR_CSSM_NUM_ATTRIBUTES;i++)
            {
                if( OutputAttributeData[i].Value )
                {
                    if (OutputAttributeData[i].Value->Data)
                        aal_free_func(OutputAttributeData[i].Value->Data,NULL);
                    aal_free_func(OutputAttributeData[i].Value,NULL);
                }
            }
            if (OutputData.Data)
                aal_free_func(OutputData.Data, NULL);

            pDlFunctions.FreeUniqueRecord(hDLDB, pRecordId);

            ret = pDlFunctions.DataGetNext (hDLDB,
                                            ResultsHandle,
                                            &OutputAttributes,
                                            &OutputData,
                                            &pRecordId);
        }
        else
        {
            for (i = 0; i < MDS_CDSADIR_CSSM_NUM_ATTRIBUTES;i++)
            {
                if( OutputAttributeData[i].Value )
                {
                    if (OutputAttributeData[i].Value->Data)
                        aal_free_func(OutputAttributeData[i].Value->Data, NULL);
                    aal_free_func(OutputAttributeData[i].Value,NULL);
                }
            }
            if (OutputData.Data)
                aal_free_func(OutputData.Data, NULL);

            pDlFunctions.FreeUniqueRecord(hDLDB,
                                          pRecordId);


            while((ret = pDlFunctions.DataGetNext(hDLDB,
                                                  ResultsHandle,
                                                  &OutputAttributes,
                                                  &OutputData, 
                                                  &pRecordId )) == CSSM_OK)
            {
                numRecords++;

                for (i = 0; i < MDS_CDSADIR_CSSM_NUM_ATTRIBUTES;i++)
                {
                    if( OutputAttributeData[i].Value )
                    {
                        if (OutputAttributeData[i].Value->Data)
                            aal_free_func(OutputAttributeData[i].Value->Data, NULL);
                        aal_free_func(OutputAttributeData[i].Value, NULL);
                    }
                }
                if (OutputData.Data)
                    aal_free_func(OutputData.Data, NULL);

                pDlFunctions.FreeUniqueRecord(hDLDB,
                                              pRecordId);
            }

            err = AAL_NOT_ENOUGH_MEMORY;
        }
    }

    /* set the number of CSSMs present in the system */
    //if(CSSM_TRUE == EndOfDataStore)
    //{
        *NumOfVersionVendorInfo = numRecords;
    //}
    //else
    //{
     //   err = AAL_FAIL;
    //}

    pDlFunctions.DataAbortQuery( hDLDB, ResultsHandle );

    return err;
}



AAL_RETURN_CODE
aal_GetCSSMList(AAL_VERSION_VENDOR_PTR VersionVendorPtr,
                uint32* NumOfVersionVendorInfo)
{
    CSSM_DL_HANDLE Handle;
    CSSM_DB_HANDLE dbHandle;
    CSSM_DL_DB_HANDLE hDLDB;
    CSSM_DB_ACCESS_TYPE access = CSSM_DB_ACCESS_READ;
    MDS_FUNCS pDlFunctions;
    AAL_RETURN_CODE err = AAL_OK;
    CSSM_RETURN ret = CSSM_OK;

    if ((ret = MDS_Initialize( 
                    NULL,
                    NULL,
                    &aal_pMemoryFunctions,
                    &pDlFunctions,&Handle)) == CSSM_OK)
    {
        if ((ret = pDlFunctions.DbOpen(Handle,
                                            MDS_CDSA_DIRECTORY,
                                            NULL,
                                            access,
                                            NULL,
                                            NULL,&dbHandle)) == CSSM_OK)
        {
            hDLDB.DLHandle = Handle;
            hDLDB.DBHandle = dbHandle;

            err = aal_GetListFromMDS(pDlFunctions,
                                     hDLDB,
                                     VersionVendorPtr,
                                     NumOfVersionVendorInfo);

            pDlFunctions.DbClose(hDLDB);
        }
        else
        {
            err = AAL_FAIL;
        }

        MDS_Terminate(Handle);
    }
    else
    {
        err = AAL_FAIL;
    }

    return err;
}
