/*-----------------------------------------------------------------------
 * File: ADDMGR.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

/*
 * This file contains the code used to manage plug-in modules by the CSSM.
 */
#ifdef VMS
#include <string.h>
#endif
#include "precomp.h"

/*
 * cssm_NewModuleRecord
 *
 * Description: This function creates a new module record for a given GUID and
 *      places this record in the global module record list.  It is assumed
 *      that the caller does not hold a lock on the global list when this
 *      function is called.
 *
 *      The new module record is returned with a writer lock held.
 *
 *      This function considers it to be an error to call this function when
 *      a module with this GUID (same module) is already present in the global
 *      list.
 *
 * Parameters:
 *      Guid (input) : GUID of the module to be added to the record list.
 *      NewModuleRecord (output) : The new record with the writer lock held.
 *
 * Returns:
 *      CSSM_OK - success
 *      All other returns indicate error.
 */

CSSM_RETURN CSSMAPI
cssm_NewModuleRecord(const CSSM_GUID *Guid,
                     cssm_INTERNAL_MODULE_LIST_PTR *NewModuleRecord)
{
    cssm_INTERNAL_MODULE_LIST_PTR tempRecord = NULL;
    cssm_INTERNAL_MODULE_LIST_PTR HandleSearcher = NULL;
    uint32 HandleIndex = 0;
    CSSM_RETURN ret = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_cssm_NewModuleRecord);


    /* now we need to lock the global list */
    if( (ret = cssm_GetModuleListLock(CSSM_WRITER_LOCK)) == CSSM_OK)
    {
        /* We need to check to make sure this module hasn't already been
           loaded.
        */
        tempRecord = hModuleListHead;

        while((NULL != tempRecord) &&
              (CSSM_OK == ret))
        {
            /* Test to see if the GUIDs match */
            if(CSSM_TRUE == CSSM_CompareGuids((*Guid),
                                              tempRecord->GUID))
            {
                ERR(ret = CSSMERR_CSSM_INTERNAL_ERROR);
            }
            else
            {
                /* Move to the next record */
                tempRecord = tempRecord->Next;
            }
        }

        if(CSSM_OK == ret)
        {
            /* Allocate memory for the new record */
            if ((tempRecord = 
                (cssm_INTERNAL_MODULE_LIST_PTR)local_calloc((uint32)sizeof(cssm_INTERNAL_MODULE_LIST),
                                                            1,
                                                            0)) == NULL)
            {
                ERR(ret = CSSMERR_CSSM_MEMORY_ERROR);
            }
            else
            {
                /* copy the input GUID */
                memcpy(&(tempRecord->GUID),
                       Guid,
                       sizeof(CSSM_GUID));

                /* generate index for this new record */

                /* is this the first record */
                if(NULL == hModuleListTail)
                {
                    HandleSearcher = NULL;
                    HandleIndex = 0;
                }
                else
                {
                    /* if this isn't the rollover mark */
                    if(0xFF800000 > cssm_ExtractModuleIndex(hModuleListTail->ModuleHandleIndex))
                    {
                        HandleIndex = cssm_ExtractModuleIndex(hModuleListTail->ModuleHandleIndex) + 0x00800000;
                        HandleSearcher = hModuleListTail;
                    }
                    else
                    {
                        /* just making sure the first slot isn't available */
                        if(0 != cssm_ExtractModuleIndex(hModuleListHead->ModuleHandleIndex))
                        {
                            HandleSearcher = NULL;
                            HandleIndex = 0;
                        }
                        else
                        {
                            /* Now we need to search the list in hipes of
                               finding an emtpy slot to insert this record in.
                            */
                            HandleSearcher = hModuleListHead;
                            while((NULL != HandleSearcher->Next) &&
                                ((cssm_ExtractModuleIndex(HandleSearcher->ModuleHandleIndex) + 0x00800000) ==
                                (cssm_ExtractModuleIndex(HandleSearcher->Next->ModuleHandleIndex))))
                            {
                                HandleSearcher = HandleSearcher->Next;
                            }
                    
                            /* make sure we found a spot to add things */
                            if(HandleSearcher == hModuleListTail)
                            {
                                /* all 512 possible modules are loaded */
                                ERR(ret = CSSMERR_CSSM_INTERNAL_ERROR);
                            }
                            else
                            {
                                /* We are inserting after HandleSearcher,
                                so we know the handle value.
                                */
                                HandleIndex = cssm_ExtractModuleIndex(HandleSearcher->ModuleHandleIndex) + 0x00800000;
                            }
                        }
                    }
                }

                if(CSSM_OK == ret)
                {
                    /* assign the handle into the record entry */
                    tempRecord->ModuleHandleIndex = HandleIndex;
        
                    /* now we generate the SWMR lock for this record */
                    if(ERR(ret = cssm_SWMRInit(&tempRecord->hModuleSWMRLock)) == CSSM_OK)
                    {
                        /* Now that we successfully have a new record, we'll
                           lock it, set the global list, and then return the
                           record for updating.
                        */
                        if ((ret = cssm_WriterLock (tempRecord->hModuleSWMRLock)) == CSSM_OK)      
                        {
                            /* now we insert the record into the list */
                            if(NULL == HandleSearcher)
                            {
                                /* we're inserting at the head */
                                tempRecord->Next = hModuleListHead;
                                hModuleListHead = tempRecord;
                
                                /* if this is the only record */
                                if(NULL == hModuleListTail)
                                {
                                    hModuleListTail = tempRecord;
                                }
                            }
                            else
                            {
                                tempRecord->Next = HandleSearcher->Next;
                                HandleSearcher->Next = tempRecord;
                
                                /* check if we need to move the tail */
                                if(hModuleListTail == HandleSearcher)
                                {
                                    hModuleListTail = tempRecord;
                                }
                            }
                        }

                        if(CSSM_OK != ret)
                        {
                            /* free up the lock memory - we know we don't
                               hold the writer lock here
                            */
                            cssm_SWMRDelete(tempRecord->hModuleSWMRLock);
                        }
                    }
                }

                if(CSSM_OK != ret)
                {
                    local_free(tempRecord,
                               NULL);
                }
            }
        }

        /* unlock the global lock whether we had success or not */
        cssm_ReleaseModuleListLock(CSSM_WRITER_LOCK);
    }

    if(CSSM_OK == ret)
    {
        *NewModuleRecord = tempRecord;
    }

    return ret;
}


/*-----------------------------------------------------------------------
*
*Name: cssm_InsertAttachRecord
*
*Description:
* Used to generate an attach record and insert it in the attach list
* contained in ServiceRecord.  Function also generates the random handle
* used to refer to this attach record.  Writer lock needs to be set on
* ServiceRecord by the calling function.  The ModuleRecord needs to have
* at least a reader lock applied when this function is called.
*
*Parameters:
*   MemoryFunc: Application memory functions to be stored in the attach record
*   Version: CDSA version passed at attach time - to be stored in attach record
*   SubServiceType: type of subservice being attached to
*   AttachFlags: flags for attach - store in attach record
*   KeyHierarchy: key hierarchy of attach - stored in attach record
*   ModuleRecord: record for module, 
*   ServiceRecord: record for subservice
*   NewAttachHandle: new attach value which is to be used
*   NewAttachRecord: new attach record
*
*Returns:
*   CSSM_OK - Success
*   All other returns are errors
*---------------------------------------------------------------------*/
CSSM_RETURN
cssm_InsertAttachRecord(const CSSM_API_MEMORY_FUNCS *MemoryFunc,
                        const CSSM_VERSION *Version,
                        CSSM_SERVICE_TYPE SubServiceType,
                        CSSM_ATTACH_FLAGS AttachFlags,
                        CSSM_KEY_HIERARCHY KeyHierarchy,
                        const cssm_INTERNAL_MODULE_LIST *ModuleRecord,
                        cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord,
                        CSSM_HANDLE *NewAttachHandle,
                        cssm_INTERNAL_ATTACH_LIST_PTR *NewAttachRecord)
{
    cssm_INTERNAL_ATTACH_LIST_PTR TempRecord;
    cssm_INTERNAL_ATTACH_LIST_PTR ListIndexPtr;
    CSSM_RETURN ret = CSSM_OK;
    uint32 RandomBits = 0;
    uint32 tempNewHandle = 0;

    TRACE(cssm_trace_info, FNCODE_cssm_IncrementAddInAppMemory);


    /* allocate memory for attach record */
    TempRecord = (cssm_INTERNAL_ATTACH_LIST_PTR)local_calloc((uint32)sizeof(cssm_INTERNAL_ATTACH_LIST),
                                                             1,
                                                             0);
    if(TempRecord == NULL)
    {
        RETURN(CSSMERR_CSSM_MEMORY_ERROR)
    }
    
    TempRecord->MemoryFunctions = (CSSM_API_MEMORY_FUNCS_PTR)local_calloc((uint32)sizeof(CSSM_API_MEMORY_FUNCS),
                                                                          1,
                                                                          0);
    if(TempRecord->MemoryFunctions == NULL)
    {
        local_free (TempRecord, 0);
        return (CSSMERR_CSSM_MEMORY_ERROR);
    }
    
    /* Generate a handle for the library; make sure it is unique */
    while(0 == tempNewHandle)
    {
        RandomBits = cssm_Rand();
        ListIndexPtr = ServiceRecord->AttachList;
        
        /* strip the random bits to the 17 bits needed for the attach record */
        tempNewHandle = cssm_ExtractServiceIndex(RandomBits) ^
            cssm_ExtractAttachIndex(RandomBits);
        
        while(NULL != ListIndexPtr)
        {
            if ((ret = cssm_ReaderLock (ListIndexPtr->hAttachSWMRLock)) == CSSM_OK)      
            {
                
                if(tempNewHandle == cssm_ExtractAttachIndex(ListIndexPtr->AttachHandle))
                {
                    RandomBits = 0;
                    tempNewHandle = 0;
                }
                
                cssm_ReaderUnlock(ListIndexPtr->hAttachSWMRLock);
            }
            else
            {
                local_free(TempRecord->MemoryFunctions, 0);
                local_free (TempRecord, 0);
                RETURN(ret)
            }
            ListIndexPtr = ListIndexPtr->Next;
        }
    }
    
    /* copy the handle into the new record */
    TempRecord->AttachHandle = cssm_ExtractModuleIndex(ServiceRecord->ServiceHandleIndex) |
        cssm_ExtractServiceIndex(ServiceRecord->ServiceHandleIndex) |
        tempNewHandle;
    
    /* copy app memory functions for the new record */
    TempRecord->MemoryFunctions->malloc_func = MemoryFunc->malloc_func;
    TempRecord->MemoryFunctions->realloc_func = MemoryFunc->realloc_func;
    TempRecord->MemoryFunctions->free_func = MemoryFunc->free_func;
    TempRecord->MemoryFunctions->calloc_func = MemoryFunc->calloc_func;
    TempRecord->MemoryFunctions->AllocRef = MemoryFunc->AllocRef;
    

    /* copy attach record options */
    TempRecord->Version = *Version;
    TempRecord->AttachFlags = AttachFlags;
    TempRecord->KeyHierarchy = KeyHierarchy;
    TempRecord->ServiceType = SubServiceType;

    /* copy module level settings */
    TempRecord->MultiTaskMutex = ModuleRecord->MultiTaskMutex;
    TempRecord->ThreadSafe = ModuleRecord->ThreadSafe;

    /*
    * Create the multiple-reader-single-write synchronization object.
    */
    if((ret = cssm_SWMRInit (&TempRecord->hAttachSWMRLock)) != CSSM_OK)
    {
        local_free(TempRecord->MemoryFunctions, 0);
        local_free (TempRecord, 0);
        RETURN (ret)
    }
    
    if ((ret = cssm_WriterLock (TempRecord->hAttachSWMRLock)) != CSSM_OK)   
    {
        local_free(TempRecord->MemoryFunctions, 0);
        local_free(TempRecord, 0);
        RETURN (ret)
    }
    
    TempRecord->AttachComplete = CSSM_FALSE;
    
    *NewAttachHandle = TempRecord->AttachHandle;
    *NewAttachRecord = TempRecord;
    
    /* insert at the beginning of the attach list */
    if(NULL != ServiceRecord->AttachList)
        TempRecord->Next = ServiceRecord->AttachList;
    
    ServiceRecord->AttachList = TempRecord;
    
    
    RETURN(CSSM_OK)
        
}



/*-----------------------------------------------------------------------
*
*Name: cssm_GetAppMemoryFunctions
*
*Description:
*   This function gets the application's memory function out of the attach
*   record.
*
*Parameters:
*   AttachRecord: The attach record to get the functions from
*   MemFuncs: The memory functions that were given when hAddIn was generated
*
*Returns:
*   CSSM_OK: Success
*   CSSMERR_CSSM_INTERNAL_ERROR: There weren't memory functions in the attach
*       record.  This should've been caught earlier.
*---------------------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
cssm_GetAppMemoryFunctions(cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord,
                           CSSM_API_MEMORY_FUNCS_PTR *MemFuncs)
{
    CSSM_RETURN err = CSSM_OK;

    if(NULL != AttachRecord->MemoryFunctions)
    {
        *MemFuncs = AttachRecord->MemoryFunctions;
    }
    else
    {
        ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
    }

    return CSSM_OK;
}



/*-----------------------------------------------------------------------
*
*Name: cssm_GetAddInCallbackByType
*
*Description:
* Interface function that allows an API function to get the callback
* into an AddIn given its info pointer.
*
*Parameters:
* ModuleInfo(input) - Pointer to the info for the addin.
* AddInType(input) - mask specifying the addin type
*  the callback registered with the CSSM.
*
*Returns:
* Pointer to the the structure containg the callback functions for the addin.
* NULL - if invalid addin
*---------------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
cssm_GetAddInCallBackByType(cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord,
                            CSSM_SERVICE_MASK AddInType, void** CallBack)
{
    CSSM_RETURN ret = CSSM_OK;
    
    TRACE(cssm_trace_info, FNCODE_cssm_GetAddInCallBack);
    
    if( (NULL != AttachRecord) && (NULL != CallBack) )
    {
        if(NULL != AttachRecord->AddInFunctions)
        {
        if(AddInType == (CSSM_SERVICE_MASK)AttachRecord->ServiceType)
           {
           *CallBack = (void **)AttachRecord->AddInFunctions->ServiceFuncs; /* ifdef VMS cast */
           }
        }
        
        if(NULL == *CallBack)
        {
            ret = CSSMERR_CSSM_INVALID_HANDLE_USAGE;
        }
        
    }
    else
    {
        ret = CSSMERR_CSSM_INVALID_POINTER;
    }
    
    return ret;
}

CSSM_RETURN CSSMAPI cssm_AddEmmRecordToModuleRecord( cssm_INTERNALMANAGERLIST_PTR* pEMMRecord,
													 cssm_INTERNAL_MODULE_LIST_PTR ModuleRecord )
{
	CSSM_RETURN ret = CSSM_OK;
	uint32 i;
	if ( ModuleRecord->NumEmms != 0 )
	{
		CSSM_BOOL bFound = CSSM_FALSE;
		for( i=0; i< ModuleRecord->NumEmms; i++)
		{
			if( ModuleRecord->pEMMRecord[i] == *pEMMRecord )
			{
				bFound = CSSM_TRUE;
				break;
			}
		}	
		if( !bFound )
		{
			ModuleRecord->pEMMRecord = (cssm_INTERNALMANAGERLIST_PTR*) local_realloc(ModuleRecord->pEMMRecord, 
													(uint32)sizeof(cssm_INTERNALMANAGERLIST_PTR) * (ModuleRecord->NumEmms+1),
													NULL);
			if( ModuleRecord->pEMMRecord == NULL )
				ret = CSSMERR_CSSM_MEMORY_ERROR;
			else
			{
				ModuleRecord->NumEmms++;
				ModuleRecord->pEMMRecord[ModuleRecord->NumEmms] = *pEMMRecord;
				ret = CSSM_OK;
			}
		}
		
	}
	else
	{
		ModuleRecord->NumEmms = 1;
		ModuleRecord->pEMMRecord = (cssm_INTERNALMANAGERLIST_PTR*) local_calloc( 
													(uint32)sizeof(cssm_INTERNALMANAGERLIST_PTR) * (ModuleRecord->NumEmms),
													1,
													NULL);
		if( ModuleRecord->pEMMRecord == NULL )
			ret = CSSMERR_CSSM_MEMORY_ERROR;
		else
		{
			*ModuleRecord->pEMMRecord = *pEMMRecord;
			ret = CSSM_OK;
		}

	}
	RETURN ( ret )
}

CSSM_RETURN CSSMAPI cssm_AttachCommon( 
                         const CSSM_GUID *ModuleGuid,
                         const CSSM_VERSION *Version,
                         const CSSM_API_MEMORY_FUNCS *MemoryFuncs,
                         uint32 SubserviceID,
                         CSSM_SERVICE_TYPE SubServiceType,
                         CSSM_ATTACH_FLAGS AttachFlags,
                         CSSM_KEY_HIERARCHY KeyHierarchy,
						 const CSSM_GUID* CallerGuid,
                         CSSM_FUNC_NAME_ADDR *FunctionTable,
                         uint32 NumFunctionTable,
                         const void * Reserved,
                         CSSM_MODULE_HANDLE_PTR NewModuleHandle,
                         cssm_INTERNAL_MODULE_LIST_PTR *ModuleRecord,
                         cssm_INTERNAL_SERVICE_LIST_PTR *ServiceRecord,
                         cssm_INTERNAL_ATTACH_LIST_PTR *AttachRecord,
                         CSSM_GUID_PTR ModuleManagerGuid,
                         cssm_INTERNALMANAGERLIST_PTR *pEMMRecord)
{
    CSSM_RETURN dwRtn = CSSM_OK;


    /*
    * Get the writer lock on the correct service list entry
    */
    if((dwRtn = cssm_FindServiceByGUIDAndSSIDAndMultiLock(ModuleGuid,
        SubserviceID,
        CSSM_NO_LOCK,
        ModuleRecord,
        CSSM_READER_LOCK,
        ServiceRecord,
        CSSM_WRITER_LOCK)) != CSSM_OK)
    {
        RETURN (dwRtn)
    }
    
    /* Now we check to make SURE that the service type being requested is
    supported by this addin.
    */
    if(0 == (SubServiceType & (*ServiceRecord)->ServiceMask))
    {
        cssm_ReleaseModuleLock(*ModuleRecord,
            CSSM_READER_LOCK);
        cssm_ReleaseServiceLock(*ServiceRecord,
            CSSM_WRITER_LOCK);
        RETURN (CSSMERR_CSSM_INVALID_SERVICE_MASK)
    }
    port_memset(ModuleManagerGuid, 0, sizeof(CSSM_GUID));
    /* If this is module requires an EMM we need to load the EMM here.  If this
    isn't the first attach, the call will just bump the ref count on the EMM
    list entry.
    */
    
    if((dwRtn = cssm_checkEMMAndLoad(ModuleGuid,
        CallerGuid,
        ModuleManagerGuid,
        SubServiceType,
        pEMMRecord,
        Version,
        KeyHierarchy)) != CSSM_OK) 
    {
    /*
    * Release all locks and fail.
        */
        cssm_ReleaseModuleLock(*ModuleRecord,
            CSSM_READER_LOCK);
        cssm_ReleaseServiceLock(*ServiceRecord,
            CSSM_WRITER_LOCK);
        RETURN (dwRtn)
    }
	if( *pEMMRecord != NULL )
	{
		dwRtn = cssm_AddEmmRecordToModuleRecord( pEMMRecord, *ModuleRecord );
		if( CSSM_OK != dwRtn )
		{
		/*
		* Release all locks and fail.
			*/
			cssm_ReleaseModuleLock(*ModuleRecord,
				CSSM_READER_LOCK);
			cssm_ReleaseServiceLock(*ServiceRecord,
				CSSM_WRITER_LOCK);
			RETURN (dwRtn)

		}
	}
    /* create an entry in the attach list for this service record */
    if((dwRtn = cssm_InsertAttachRecord(MemoryFuncs,
        Version,
        SubServiceType,
        AttachFlags,
        KeyHierarchy,
        *ModuleRecord,
        *ServiceRecord,
        NewModuleHandle,
        AttachRecord)) != CSSM_OK)
    {
        cssm_ReleaseModuleLock(*ModuleRecord,
            CSSM_READER_LOCK);
        cssm_ReleaseServiceLock(*ServiceRecord,
            CSSM_WRITER_LOCK);
        RETURN (dwRtn)
    }

    (*AttachRecord)->pEMMRecord = *pEMMRecord;
#ifndef ISL_INCLUDED
    (*ServiceRecord)->ModuleLibHandle = (*ModuleRecord)->LibHandle;
#endif
    /* we're done with the module record now so we release the lock */
    cssm_ReleaseModuleLock(*ModuleRecord,
        CSSM_READER_LOCK);
    RETURN (dwRtn)
}

void CSSMAPI 
cssm_FailAttach( cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord,
                 cssm_INTERNAL_ATTACH_LIST_PTR* AttachRecord,
                 CSSM_MODULE_HANDLE_PTR NewModuleHandle)
{
    cssm_CleanInternalAttachRecord(ServiceRecord,
                                   AttachRecord);
    cssm_ReleaseServiceLock(ServiceRecord,
                            CSSM_WRITER_LOCK);
    *NewModuleHandle = 0;
}


#ifdef ISL_INCLUDED

CSSM_RETURN CSSMAPI 
cssm_RetrieveManifest( const CSSM_GUID* ModuleGuid,
                        CSSM_DATA_PTR Manifest,
                        CSSM_DATA_PTR ModulePath, 
                        CSSM_SERVICE_TYPE ServiceType,
                        CSSM_BOOL bUpdatePrimary,
						cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord
						)
{
    CSSM_RETURN dwRtn;

    dwRtn = mdsutil_GetPrimaryManifest( ModuleGuid, Manifest, ServiceType );
    if( dwRtn == CSSM_OK )
    {
		if( ( Manifest->Length == 0 ) && ( Manifest->Data == NULL ) )
        {
			dwRtn = mdsutil_GetCommonCredentialInfo (ModuleGuid,Manifest, NULL, ModulePath);
            if( bUpdatePrimary )
				ServiceRecord->bDiffPrimaryManifest = CSSM_FALSE;
        }
		else
		{
			dwRtn = mdsutil_GetCommonCredentialInfo (ModuleGuid, NULL, NULL, ModulePath);
		}
    }
    else
    {
        dwRtn = mdsutil_GetCommonCredentialInfo (ModuleGuid, Manifest, NULL, ModulePath);
        if( Manifest->Length == 0 && Manifest->Data == 0)
            dwRtn = CSSMERR_CSSM_MDS_ERROR;
    }
    
    RETURN (dwRtn)
}

CSSM_RETURN CSSMAPI
cssm_ModuleAttachInteg(const CSSM_GUID *ModuleGuid,
                  const CSSM_VERSION *Version,
                  const CSSM_API_MEMORY_FUNCS *MemoryFuncs,
                  uint32 SubserviceID,
                  CSSM_SERVICE_TYPE SubServiceType,
                  CSSM_ATTACH_FLAGS AttachFlags,
                  const CSSM_GUID *CallerGuid,
                  CSSM_KEY_HIERARCHY KeyHierarchy,
                  CSSM_FUNC_NAME_ADDR *FunctionTable,
                  uint32 NumFunctionTable,
                  const void * Reserved,
                  CSSM_MODULE_HANDLE_PTR NewModuleHandle)
{
    cssm_INTERNAL_MODULE_LIST_PTR ModuleRecord = NULL;
    cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_GUID ModuleManagerGuid;
    cssm_INTERNALMANAGERLIST_PTR pEMMRecord= NULL;
    CSSM_RETURN dwRtn = CSSM_OK;
    CSSM_DATA Manifest;
    CSSM_DATA ModulePath;
    CSSM_BOOL KeyHierarchyAttach = CSSM_FALSE;
    
   
    dwRtn = cssm_AttachCommon( ModuleGuid,
                         Version,
                         MemoryFuncs,
                         SubserviceID,
                         SubServiceType,
                         AttachFlags,
                         KeyHierarchy,
                         CallerGuid,
                         FunctionTable,
                         NumFunctionTable,
                         Reserved,
                         NewModuleHandle,
                         &ModuleRecord,
                         &ServiceRecord,
                         &AttachRecord,
                         &ModuleManagerGuid,
                         &pEMMRecord  );
    if( dwRtn == CSSM_OK )
    {
        Manifest.Length = 0;
        Manifest.Data = NULL;
        ModulePath.Length = 0;
        ModulePath.Data = NULL;
        /* check if we've verified before and we're not CSSM_KEY_HIERARCHY_NONE */

        if( NULL == ServiceRecord->SPCredentialPtr ) 
        {
            dwRtn = cssm_RetrieveManifest( ModuleGuid,
                                        &Manifest,
                                        &ModulePath, 
                                        SubServiceType,
                                        CSSM_TRUE,
										ServiceRecord);
            if( dwRtn == CSSM_OK )
            {
                dwRtn = cssm_AttachModuleUsingExportOrIntegrityKeys(ModuleGuid,
                                            Version,
                                            SubserviceID,
                                            SubServiceType,
                                            AttachFlags,
                                            *NewModuleHandle,
                                            KeyHierarchy,
                                            &AttachRecord->AddInFunctions,                                                    
                                            ServiceRecord,
                                            AttachRecord,
                                            CallerGuid,
                                            &Manifest,
                                            &ModulePath);

            }
            if( Manifest.Data != NULL )
                local_free ((char *)Manifest.Data, NULL);
            if( ModulePath.Data != NULL)
                local_free ((char *)ModulePath.Data, NULL);
        }
        else
        {
    		/* Checks if an attach has happened before using the key hierarchy*/
	    	/* we care only the KeyHierarchyAttach flag*/
            cssm_CheckIfAttachRecordWithMatchingKeyHierarchyExists(*NewModuleHandle,
                                                                    KeyHierarchy,
                                                                    ServiceRecord,
                                                                   &KeyHierarchyAttach);
            if( !KeyHierarchyAttach )
            {
                dwRtn = cssm_RetrieveManifest( ModuleGuid,
                                        &Manifest,
                                        &ModulePath, 
                                        SubServiceType,
                                        CSSM_FALSE,
										ServiceRecord);
                if( dwRtn == CSSM_OK )
                {
                     dwRtn = cssm_AttachModuleUsingExportOrIntegrityKeys(ModuleGuid,
                                                                    Version,
                                                                    SubserviceID,
                                                                    SubServiceType,
                                                                    AttachFlags,
                                                                    *NewModuleHandle,
                                                                    KeyHierarchy,
                                                                    &AttachRecord->AddInFunctions,
                                                                    ServiceRecord,
                                                                    AttachRecord,
                                                                    CallerGuid,
                                                                    &Manifest,
                                                                    &ModulePath);
                 }
                 if( Manifest.Data != NULL )
                     local_free ((char *)Manifest.Data, NULL);
                 if( ModulePath.Data != NULL)
                     local_free ((char *)ModulePath.Data, NULL);

            }
            else
            {
                dwRtn = cssm_AttachModuleNoKeys(ModuleGuid,
                                                Version,
                                                SubserviceID,
                                                SubServiceType,
                                                AttachFlags,
                                                *NewModuleHandle,
                                                KeyHierarchy,
                                                CallerGuid,
                                                &AttachRecord->AddInFunctions,
                                                AttachRecord,
				                                ServiceRecord);
            }        
        }
        if( dwRtn == CSSM_OK )
        {
            if(cssm_CheckFunctionTableInverifiedModules(ServiceRecord->NumSPCreds,
                                                        ServiceRecord->SPCredentialPtr,
                                                         AttachRecord->AddInFunctions))
            {
                /* At this point we know we're done with the service record, so we'll
                   unlock it here.
                   */
                cssm_ReleaseServiceLock(ServiceRecord,
                                      CSSM_WRITER_LOCK);
        
                        /* We've successfully completed all attach requirements - unless we use
                        KR.  In that case, the attach lock will have to be re-obtained.
                    */
        
                AttachRecord->AttachComplete = CSSM_TRUE;
                cssm_ReleaseAttachLock(AttachRecord,
                                    CSSM_WRITER_LOCK);
        
                        /* If the caller wants a function table back, this is where we set
                        it.
                    */
                if((NULL != FunctionTable) &&
                   (0 != NumFunctionTable))
                {
                    cssm_GetFunctionTable(FunctionTable,
                                NumFunctionTable, 
                                *NewModuleHandle);
                }

            }
            else
            {
                dwRtn = CSSMERR_CSSM_FUNCTION_INTEGRITY_FAIL;
            }
        }
        if( dwRtn != CSSM_OK )
            cssm_FailAttach( ServiceRecord, &AttachRecord, NewModuleHandle );
    }
    RETURN(dwRtn)
}
#else

CSSM_RETURN CSSMAPI
cssm_ModuleAttachNoInteg(const CSSM_GUID *ModuleGuid,
                         const CSSM_VERSION *Version,
                         const CSSM_API_MEMORY_FUNCS *MemoryFuncs,
                         uint32 SubserviceID,
                         CSSM_SERVICE_TYPE SubServiceType,
                         CSSM_ATTACH_FLAGS AttachFlags,
                         const CSSM_GUID *CallerGuid,
                         CSSM_KEY_HIERARCHY KeyHierarchy,
                         CSSM_FUNC_NAME_ADDR *FunctionTable,
                         uint32 NumFunctionTable,
                         const void * Reserved,
                         CSSM_MODULE_HANDLE_PTR NewModuleHandle)
{
    cssm_INTERNAL_MODULE_LIST_PTR ModuleRecord = NULL;
    cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_GUID ModuleManagerGuid;
    cssm_INTERNALMANAGERLIST_PTR pEMMRecord= NULL;
    CSSM_RETURN dwRtn = CSSM_OK;
    
    TRACE(cssm_trace_info, FNCODE_cssm_ModuleAttach);
    ERR(dwRtn = cssm_AttachCommon(ModuleGuid,
                                  Version,
                                  MemoryFuncs,
                                  SubserviceID,
                                  SubServiceType,
                                  AttachFlags,
                                  KeyHierarchy,
								  CallerGuid,
                                  FunctionTable,
                                  NumFunctionTable,
                                  Reserved,
                                  NewModuleHandle,
                                  &ModuleRecord,
                                  &ServiceRecord,
                                  &AttachRecord,
                                  &ModuleManagerGuid,
                                  &pEMMRecord));

    if(CSSM_OK == dwRtn)
    {
        ERR(dwRtn = cssm_AttachModuleNoKeys(ModuleGuid,
                                            Version,
                                            SubserviceID,
                                            SubServiceType,
                                            AttachFlags,
                                            *NewModuleHandle,
                                            KeyHierarchy,
											CallerGuid,
                                            &AttachRecord->AddInFunctions,
                                            AttachRecord,
											ServiceRecord));

        if(CSSM_OK != dwRtn)
        {
            /* We've failed the attach so we'll remove the attach record */
            cssm_CleanInternalAttachRecord(ServiceRecord,
                                           &AttachRecord);
        }

        /* At this point we know we're done with the service record, so we'll
           unlock it here.
        */
        cssm_ReleaseServiceLock(ServiceRecord,
                                CSSM_WRITER_LOCK);

        if(CSSM_OK == dwRtn)
        {
            /* We've successfully completed all attach requirements - unless we
               use KR.  In that case, the attach lock will have to be
               re-obtained.
            */

            AttachRecord->AttachComplete = CSSM_TRUE;
            cssm_ReleaseAttachLock(AttachRecord,
                                   CSSM_WRITER_LOCK);

            /* If the caller wants a function table back, this is where we set
               it.
            */
            if((NULL != FunctionTable) &&
               (0 != NumFunctionTable))
            {
                cssm_GetFunctionTable(FunctionTable,
                                      NumFunctionTable,
                                      *NewModuleHandle);
            }
        }
    }


    RETURN(dwRtn)
}

#endif

/*---------------------------------------------------------------
*
*Name: cssm_ModuleAttach
*
*Description:
*   This function is called from the CSSM_ModuleAttach api in cssm.c.
*   If it is the first attach, it will add a ned addin module record, do 
*   the integrity check of the module as well as cross check and do the 
*   initializtion of the module. If it is an emm addin module, the emm 
*    manager module is loaded before the addin module load.
*   If it is not th efirst time attach, then a new addin handle is generated
*   and a new HandleInfo is added to the internal add in module record list.
*   
*
*Parameters: 
*   pModuleGUID - guid of the add-in module
*   Version - version number of requested add-in
*   pMemoryFunc - application's memory functions
*   SubServiceID - SubServiceID
*   SubServiceFlags - SubServiceFlags
*   Application - Application context nonce for use with the callback
*   Notification - Callback used to tell the application when an
*                  add-in event takes place.
*   CallerGuid - Guid of the calling application
*   USEERequest - USEE exemption request from the app
*   eventType - Event code passed to the add-in after attach
*   detachEventType - Event code passed to the add-in when a load
*                     failure occures.
*
*Returns:
*  0 - unable to attach to the add-in module
*  non 0 - handle to the add-in module
*
*----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_ModuleAttach(const CSSM_GUID *ModuleGuid,
                  const CSSM_VERSION *Version,
                  const CSSM_API_MEMORY_FUNCS *MemoryFuncs,
                  uint32 SubserviceID,
                  CSSM_SERVICE_TYPE SubServiceType,
                  CSSM_ATTACH_FLAGS AttachFlags,
                  CSSM_KEY_HIERARCHY KeyHierarchy,
                  CSSM_FUNC_NAME_ADDR *FunctionTable,
                  uint32 NumFunctionTable,
                  const void * Reserved,
                  CSSM_MODULE_HANDLE_PTR NewModuleHandle,
				  const CSSM_GUID* CallerGuid
                  )
{
    CSSM_RETURN ret = CSSM_OK;
#ifdef ISL_INCLUDED
    ret = cssm_ModuleAttachInteg( ModuleGuid,
                                  Version,
                                  MemoryFuncs,
                                  SubserviceID,
                                  SubServiceType,
                                  AttachFlags,
                                  CallerGuid,
                                  KeyHierarchy,
                                  FunctionTable,
                                  NumFunctionTable,
                                  Reserved,
                                  NewModuleHandle);
#else

    ret = cssm_ModuleAttachNoInteg(   ModuleGuid,
                                      Version,
                                      MemoryFuncs,
                                      SubserviceID,
                                      SubServiceType,
                                      AttachFlags,
                                      CallerGuid,
                                      KeyHierarchy,
                                      FunctionTable,
                                      NumFunctionTable,
                                      Reserved,
                                      NewModuleHandle);

#endif
    RETURN (ret)
}


/*---------------------------------------------------------------
*
*Name: cssm_CleanInternalModuleRecord
*
*Description:
*   Deletes frees the internal structures under a module record.
*   This routine is called with a writer lock on the ModuleRecord.
*   The writer lock on the global module list should also be held.
*   This function will not release the writer lock on the ModuleRecord.
*
*Parameters: 
*   ModuleRecord - the record from which everything should be cleaned.
*
*Returns:
*  CSSM_FAIL - unable to delete internal add-in record
*  CSSM_OK - add-in record is removed
*
*----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_CleanInternalModuleRecord(cssm_INTERNAL_MODULE_LIST_PTR *ModuleRecord,
                               CSSM_LIB_HANDLE *MainLibHandle)
{
    cssm_INTERNAL_MODULE_LIST_PTR LeadSearcher = NULL;
    cssm_INTERNAL_MODULE_LIST_PTR TrailSearcher = NULL;
    cssm_INTERNAL_SERVICE_LIST_PTR TempService = NULL;
    CSSM_RETURN err = CSSM_OK;
    uint32 i;

#ifdef ISL_INCLUDED
    char szModuleUnload[] = "CSSM_SPI_ModuleUnload";
    CSSM_LIB_HANDLE tempHandle;
    ISL_CONST_DATA ProcedureName;
    ProcedureName.Length = (uint32)strlen(szModuleUnload);
    ProcedureName.Data = (const unsigned char*)szModuleUnload;
#endif


    if(NULL != ModuleRecord)
        LeadSearcher = hModuleListHead;

    /* first, let's take our entry out of the global list */
    while((NULL != LeadSearcher) && (ModuleRecord != NULL) &&
        (LeadSearcher->ModuleHandleIndex != (*ModuleRecord)->ModuleHandleIndex))
    {
        TrailSearcher = LeadSearcher;
        LeadSearcher = LeadSearcher->Next;
    }
    
    /* check if we've found the ModuleRecord */
    if(NULL != LeadSearcher)
    {
        /* update the global list */
        if(NULL == TrailSearcher)
        {
            hModuleListHead = LeadSearcher->Next;
            
            if(LeadSearcher == hModuleListTail)
            {
                hModuleListTail = NULL;
            }
        }
        else
        {
            TrailSearcher->Next = LeadSearcher->Next;
            
            if(LeadSearcher == hModuleListTail)
            {
                hModuleListTail = TrailSearcher;
            }
        }
        
        /* We've now remove this entry from the global list.  It's now time to,
        get rid of any and all service entries under this module record.
        */
        if(NULL != ModuleRecord)
        {
            TempService = (*ModuleRecord)->ServiceList;
            while(NULL != TempService)
            {
                cssm_WriterLock (TempService->hServiceSWMRLock);            
                if(ModuleRecord != NULL)
                {
                    cssm_CleanInternalServiceRecord((*ModuleRecord),
                                                    &TempService);
                    TempService = (*ModuleRecord)->ServiceList;
                }
            }

#ifdef ISL_INCLUDED

                if(CSSM_OK == ERR(cssm_IdentifyVerifiedModuleByGUID((*ModuleRecord)->NumSections,
                                                                    (*ModuleRecord)->ModuleCredentialPtr,
                                                                    &(*ModuleRecord)->GUID,
                                                                    &i)))
                {
                    *MainLibHandle = EISL_GetLibHandle((*ModuleRecord)->ModuleCredentialPtr[i]);
                }
#else
                *MainLibHandle = (*ModuleRecord)->LibHandle;
#endif

#ifdef ISL_INCLUDED

            /* Call free library on each library since we previously loaded
               each library.
            */
            i = 0;
            while(i < (*ModuleRecord)->NumSections)
           {
                tempHandle = EISL_GetLibHandle((*ModuleRecord)->ModuleCredentialPtr[i]);

                if(*MainLibHandle != tempHandle)
                {
                    ERR(port_FreeLibrary(tempHandle));
                }

                i++;
            }
			if( (*ModuleRecord)->ModuleCredentialPtr )
			{
				EISL_RecycleVerifiedModuleCredentials((*ModuleRecord)->ModuleCredentialPtr[0]);
				local_free( (*ModuleRecord)->ModuleCredentialPtr , NULL );
			}
#endif
			if( (*ModuleRecord)->pEMMRecord != NULL )
			{
				for ( i = 0; i < (*ModuleRecord)->NumEmms; i++ )
				{
					if (0 == (*ModuleRecord)->pEMMRecord[i]->attachRefCount)
					{
						cssm_EMMManagerUnload (&((*ModuleRecord)->pEMMRecord[i]->EMMGuid));	
					}
				}
			}
			local_free( (*ModuleRecord)->pEMMRecord, NULL );

            /* free locks */
            if(NULL != (*ModuleRecord)->hModuleSWMRLock)
            {
                cssm_WriterUnlock((*ModuleRecord)->hModuleSWMRLock);
                cssm_SWMRDelete((*ModuleRecord)->hModuleSWMRLock);
            }
        
            /* if we still have a mutex, we need to delete it now */
            if(CSSM_FALSE == (*ModuleRecord)->ThreadSafe)
            {
                port_CloseMutex((*ModuleRecord)->MultiTaskMutex);
            }

            /* free the last of the memory */
            local_free(*ModuleRecord,0);
            *ModuleRecord = NULL;
        }
    }
    else
    {
        ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
    }

    RETURN (err)
}

void CSSMAPI
cssm_CleanInternalServiceRecord(cssm_INTERNAL_MODULE_LIST_PTR ModuleRecord,
                                cssm_INTERNAL_SERVICE_LIST_PTR *ServiceRecord)
{
    cssm_INTERNAL_SERVICE_LIST_PTR LeadSearcher = NULL;
    cssm_INTERNAL_SERVICE_LIST_PTR TrailSearcher = NULL;
    cssm_INTERNAL_SERVICE_LIST_PTR TempService = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR TempAttach = NULL;
    cssm_VENDOR_PRIVILEGES_PTR tempVendorPrivs;

    if(NULL != ServiceRecord)
        LeadSearcher = ModuleRecord->ServiceList;
    
    /* first, let's take our entry out of the global list */
    while((NULL != LeadSearcher) && (ServiceRecord != NULL) &&
        (LeadSearcher->ServiceHandleIndex != (*ServiceRecord)->ServiceHandleIndex))
    {
        TrailSearcher = LeadSearcher;
        LeadSearcher = LeadSearcher->Next;
    }
    
    /* check if we've found the ModuleRecord */
    if(NULL != LeadSearcher)
    {
        /* update the global list */
        if(NULL == TrailSearcher)
        {
            ModuleRecord->ServiceList = LeadSearcher->Next;
            
            if(LeadSearcher == ModuleRecord->ServiceList)
            {
                ModuleRecord->ServiceList = NULL;
            }
        }
        else
        {
            TrailSearcher->Next = LeadSearcher->Next;
            
            if(LeadSearcher == TrailSearcher)
            {
                ModuleRecord->ServiceList = TrailSearcher;
            }
        }
    }
    if(ServiceRecord != NULL)
    {
        TempService = *ServiceRecord;
        if ( TempService != NULL )
            
            
        {
            TempAttach = TempService->AttachList;
            while(NULL != TempAttach)
            {
                cssm_WriterLock (TempAttach->hAttachSWMRLock);
                
                cssm_CleanInternalAttachRecord( TempService, 
                    &TempAttach);
                TempAttach = TempService->AttachList;
            }
#ifdef ISL_INCLUDED
            if( TempService->bDiffPrimaryManifest )
            {
                uint32 i = 0;
                while(i < TempService->NumSPCreds)
                {
                    ERR(port_FreeLibrary(EISL_GetLibHandle(TempService->SPCredentialPtr[i])));

                    i++;
                }
            }
			if( TempService->SPCredentialPtr )
			{
				EISL_RecycleVerifiedModuleCredentials(TempService->SPCredentialPtr[0]);
				local_free( TempService->SPCredentialPtr, NULL);
			}
                      
#endif

            if(NULL != TempService->SPTags)
            {
                local_free(TempService->SPTags,
                           NULL);
            }

            if(NULL != TempService->TOGPrivs)
            {
                local_free(TempService->SPTags,
                           NULL);
            }

            tempVendorPrivs = TempService->VendorPrivs;
            while(NULL != tempVendorPrivs)
            {
                TempService->VendorPrivs = tempVendorPrivs->Next;

                if(NULL != tempVendorPrivs->VendorPrivs)
                {
                    local_free(tempVendorPrivs->VendorPrivs,
                               NULL);
                }

                local_free(tempVendorPrivs,
                           NULL);

                tempVendorPrivs = TempService->VendorPrivs;
            }

            if(TempService->hServiceSWMRLock != NULL)
            {
                cssm_WriterUnlock(TempService->hServiceSWMRLock);
                cssm_SWMRDelete(TempService->hServiceSWMRLock);
            }
            local_free( TempService, NULL);
        }
    }
}


void CSSMAPI
cssm_CleanInternalAttachRecord(cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord,
                               cssm_INTERNAL_ATTACH_LIST_PTR *AttachRecord)
{
    cssm_INTERNAL_ATTACH_LIST_PTR LeadSearcher = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR TrailSearcher = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR TempAttach = NULL;
    cssm_INTERNAL_CONTEXT_PTR TempContext = NULL;
    if((NULL != AttachRecord) &&
       (NULL != ServiceRecord))
        LeadSearcher = ServiceRecord->AttachList;
    
    /* first, let's take our entry out of the global list */
    while((NULL != LeadSearcher) && (AttachRecord != NULL) &&
        (LeadSearcher->AttachHandle != (*AttachRecord)->AttachHandle))
    {
        TrailSearcher = LeadSearcher;
        LeadSearcher = LeadSearcher->Next;
    }
    
    /* check if we've found the ModuleRecord */
    if(NULL != LeadSearcher)
    {
        /* update the global list */
        if(NULL == TrailSearcher)
        {
            ServiceRecord->AttachList = LeadSearcher->Next;
            
            if(LeadSearcher == ServiceRecord->AttachList)
            {
                ServiceRecord->AttachList = NULL;
            }
        }
        else
        {
            TrailSearcher->Next = LeadSearcher->Next;
            
            if(LeadSearcher == TrailSearcher)
            {
                ServiceRecord->AttachList = TrailSearcher;
            }
        }
    }
    if( AttachRecord != NULL)
    {
        TempAttach = *AttachRecord;
        if ( TempAttach != NULL)
            
            
        {
            /* loop through and destroy any contexts we may have */
            while(NULL != TempAttach->ContextList)
            {
                TempContext = TempAttach->ContextList;
                cssm_DeleteContext(TempAttach, &TempContext);
            }
            
            /* now we call the SPI_ModuleDetach for this addin */
            if(CSSM_TRUE == TempAttach->AttachComplete)
                cssm_ModuleDetach(TempAttach,ServiceRecord);
            if( TempAttach->MemoryFunctions != NULL)
                local_free( TempAttach->MemoryFunctions, NULL);
            if( TempAttach->AddInFunctions != NULL)
		{
		if( TempAttach->AddInFunctions->ServiceFuncs != NULL)
                    local_free((char *)TempAttach->AddInFunctions->ServiceFuncs, NULL ); /* ifdef VMS cast */
                local_free( TempAttach->AddInFunctions, NULL);
			}
            if(NULL != TempAttach->hAttachSWMRLock)
            {
                cssm_WriterUnlock(TempAttach->hAttachSWMRLock);
                cssm_SWMRDelete(TempAttach->hAttachSWMRLock);
            }
            local_free( TempAttach, NULL);
        }
    }
}



CSSM_RETURN CSSMAPI 
cssm_GetAttachFunctions(CSSM_HANDLE hAddIn,
                        CSSM_SERVICE_MASK AddinType,
                        void **CallBack,
                        CSSM_GUID_PTR Guid)
{
    CSSM_RETURN Ret = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    cssm_INTERNAL_MODULE_LIST_PTR ModuleRecord;
    uint32 i;
    
    if(NULL == CallBack)
    {
        Ret = CSSMERR_CSSM_INVALID_POINTER;
    }
    else
    {
        if(( Ret = cssm_FindAttachAndMultiLock(hAddIn,
            CSSM_NO_LOCK,
            NULL,
            CSSM_NO_LOCK,
            NULL,
            CSSM_NO_LOCK,
            &AttachRecord,
            CSSM_READER_LOCK)) == CSSM_OK)
        {
            Ret = cssm_GetAddInCallBackByType(AttachRecord, AddinType, CallBack);
        }
        else
        {
            /* the search didn't find anything using this handle */
            Ret = CSSMERR_CSSM_INVALID_ADDIN_HANDLE;
        }
        
        /* If everything's ok and the user wants to know if it's threadsafe,
        we find out.
        */
        if( (CSSM_OK == Ret) &&  (NULL != Guid) )
        {
            if(CSSM_OK == cssm_FindModuleAndMultiLock(hAddIn,
                CSSM_NO_LOCK,
                &ModuleRecord,
                CSSM_READER_LOCK))
            {
                /* copy the GUID, if it's desired */
                if(NULL != Guid)
                {
                    /* copy the GUID */
                    Guid->Data1 = ModuleRecord->GUID.Data1;
                    Guid->Data2 = ModuleRecord->GUID.Data2;
                    Guid->Data3 = ModuleRecord->GUID.Data3;
                    
                    for (i=0; i<8; i++) 
                        Guid->Data4[i] = ModuleRecord->GUID.Data4[i];
                }
                
                cssm_ReleaseModuleLock(ModuleRecord,
                    CSSM_READER_LOCK);
            }
            else
            {
                cssm_ReleaseAttachLock(AttachRecord,
                    CSSM_READER_LOCK);
                Ret = CSSMERR_CSSM_INVALID_ADDIN_HANDLE;
            }
        }
    }
    
    return Ret;
}


void CSSMAPI
cssm_ReleaseAttachFunctions(CSSM_HANDLE hAddIn)
{
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    
    if(CSSM_OK == cssm_FindAttachAndMultiLock(hAddIn,
        CSSM_NO_LOCK,
        NULL,
        CSSM_NO_LOCK,
        NULL,
        CSSM_NO_LOCK,
        &AttachRecord,
        CSSM_READER_LOCK))
    {
    /* We now need to free 2 reader locks - the one we just got and the
    one we were initially called to free.
        */
        cssm_ReleaseAttachLock(AttachRecord,
            CSSM_READER_LOCK);
        cssm_ReleaseAttachLock(AttachRecord,
            CSSM_READER_LOCK);
    }
}






/*-----------------------------------------------------------------------
*
*Name: cssm_ModuleDetach
*
*Description:
* This function assumes a lock has been set on AttachRecord and calls the
* addin module's SPI_ModuleDetach.
*Parameters:
* AttachRecord (input/output): record for the attach instance which we
*                              are going to detach.
*
*Returns:
* CSSM_OK - Module detached.
* CSSM_FAIL - in case of failures.
*---------------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_ModuleDetach(cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord,cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord)
{
    CSSM_RETURN dwRtn = CSSM_OK;
    SPI_DETACH_FUNC_PTR DetachFunc = NULL;
    char szModuleDetach[] = "CSSM_SPI_ModuleDetach";
    
#ifdef ISL_INCLUDED
    uint32 i;
    ISL_CONST_DATA ProcedureName;
#endif  /* ISL_INCLUDED */
    
    /* Check if we have an attach record - if so, we assume it already has
    a writer lock set.
    */
    if(NULL != AttachRecord)
    {
    /* Make sure the attach completed and we actually have to call
    detach.  We return CSSM_OK even if this call was unecessary.
        */
        if(CSSM_TRUE == AttachRecord->AttachComplete)
        {
            /* get the function pointer */
#ifdef ISL_INCLUDED
            ProcedureName.Length = (uint32)strlen(szModuleDetach);
            ProcedureName.Data = (const unsigned char*)szModuleDetach;

            i = 0;
            while((NULL == DetachFunc) &&
                  (i < ServiceRecord->NumSPCreds))
            {
                DetachFunc = (SPI_DETACH_FUNC_PTR)EISL_LocateProcedureAddress(ServiceRecord->SPCredentialPtr[i],
                                                                              ProcedureName);
            }
#else
            dwRtn = port_GetProcAddress(ServiceRecord->ModuleLibHandle,
                                        szModuleDetach,
                                        (CSSM_PROC_ADDR *)&DetachFunc);
#endif  /* ISL_INCLUDED */
            
            /* make sure we actually got a function */
            if( (dwRtn == CSSM_OK) && (DetachFunc != NULL) )
            {
                dwRtn = DetachFunc(AttachRecord->AttachHandle);
                AttachRecord->AttachComplete = CSSM_FALSE;
            }
            else
            {
                dwRtn = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED;
            }
            /* if the module is an elective module, unload the EMM module/decrement the 
            reference count*/
            if(!cssm_IsCSSMMask(AttachRecord->ServiceType))
            {
                dwRtn = cssm_EMMManagerDetach(&AttachRecord->pEMMRecord->EMMGuid,
                                              &AttachRecord->pEMMRecord);
            }
        }
    }
    
    RETURN (dwRtn)
}



/*---------------------------------------------------------------
*
*Name: cssm_GetGUIDFromHandle
*
*Description:
*  Locate the GUID of the add-in module given a hanlde.
*
*Parameters: 
*  ModuleHandle - Handle to module assigned by CSSM
*
*Returns:
*  Non NULL - pointer to the GUID structure
*  NULL - error in retreiving the GUID 
*
*----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_GetGUIDFromHandle(CSSM_HANDLE ModuleHandle, CSSM_GUID_PTR GuidPtr)
{
    cssm_INTERNAL_MODULE_LIST_PTR ModuleRecord;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord;
    
    TRACE(cssm_trace_info, FNCODE_cssm_GetGUIDFromHandle);
    
    if(port_IsBadWritePtr(GuidPtr, sizeof(CSSM_GUID)))
    {
        RETURN (CSSMERR_CSSM_INVALID_POINTER)
    }
    
    /*
    * Get the ModuleInfo record associated with the handle.  We search down
    * to the attach record to make sure the handle is valid even though the
    * information we need is actually just in the ModuleRecord
    */
    if(CSSM_OK != cssm_FindAttachAndMultiLock(ModuleHandle,
                                              CSSM_NO_LOCK,
                                              &ModuleRecord,
                                              CSSM_READER_LOCK,
                                              NULL,
                                              CSSM_NO_LOCK,
                                              &AttachRecord,
                                              CSSM_NO_LOCK))
    {
        port_memset ( GuidPtr, 0, sizeof(CSSM_GUID));
        RETURN (CSSMERR_CSSM_INVALID_ADDIN_HANDLE)
    }
    else
    {
        /* copy the guid to return pointer */
        port_memcpy ( GuidPtr, &ModuleRecord->GUID, sizeof(CSSM_GUID));
        
        /*
        * Release the Module Record.
        */
        cssm_ReleaseModuleLock(ModuleRecord,
            CSSM_READER_LOCK);
    }
    
    RETURN(CSSM_OK)
}

/*---------------------------------------------------------------
*
*Name: cssm_GetSubserviceUIDFromHandle
*
*Description:
*  Return the unique identifier of the attached module subservice, as identified 
*   by the input handle.  .
*
*Parameters: 
*  ModuleHandle - Handle to module assigned by CSSM
*   MemoryFuncs - memory functions to be used for allocation.
*
*Returns:
*  Non NULL - pointer to the SUBSERVICE_UID structure
*  NULL - error in retreiving the subservice UID 
*
*----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
cssm_GetSubserviceUIDFromHandle(CSSM_HANDLE ModuleHandle,
                                CSSM_SUBSERVICE_UID_PTR UidPtr)
{
    cssm_INTERNAL_MODULE_LIST_PTR ModuleRecord = NULL;
    cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN rc = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_cssm_GetSubserviceUIDFromHandle);

    if((UidPtr == NULL ) ||
       (port_IsBadWritePtr( UidPtr, sizeof( CSSM_SUBSERVICE_UID))) )
    {
        ERR(rc = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        /*
         * Get the module list record associated with the handle.
         */
        if(ERR(rc = cssm_FindAttachAndMultiLock(ModuleHandle,
                                                CSSM_NO_LOCK,
                                                &ModuleRecord,
                                                CSSM_READER_LOCK,
                                                &ServiceRecord,
                                                CSSM_READER_LOCK,
                                                &AttachRecord,
                                                CSSM_READER_LOCK)) == CSSM_OK)
        {
            port_memcpy( UidPtr, &ModuleRecord->GUID, sizeof(CSSM_GUID));

            UidPtr->Version.Major = AttachRecord->Version.Major;
            UidPtr->Version.Minor = AttachRecord->Version.Minor;
            UidPtr->SubserviceId =  ServiceRecord->SubServiceID;
            UidPtr->SubserviceType = AttachRecord->ServiceType;

            /* release module record */
            cssm_ReleaseModuleLock(ModuleRecord,
                                   CSSM_READER_LOCK);

            /* release service record */
            cssm_ReleaseServiceLock(ServiceRecord,
                                    CSSM_READER_LOCK);

            /* release attach record */
            cssm_ReleaseAttachLock(AttachRecord,
                                   CSSM_READER_LOCK);
        }
        else
        {
            ERR(rc = CSSMERR_CSSM_INVALID_ADDIN_HANDLE);
        }
    }

    RETURN (rc)
}

/*---------------------------------------------------------------
*
*Name: cssm_CleanUp
*
*Description:
*  Clean up internal structures.  This function is called just
*  before CSSM is unloaded.  Fix memory leaks problems created
*  by ill behaved addins and apps.
*
*Parameters: 
*  none
*
*Returns:
*  none 
*
*----------------------------------------------------------------*/
CSSM_RETURN cssm_CleanUp (void)
{
    cssm_INTERNAL_MODULE_LIST_PTR TempModule = NULL;
    UNLOAD_MODULE_FUNC_PTR SPI_Unload_Func = NULL;
    char szModuleUnload[] = "CSSM_SPI_ModuleUnload";
    CSSM_RETURN rc = CSSM_OK;
    CSSM_LIB_HANDLE MainLibHandle;


    TRACE(cssm_trace_info, FNCODE_cssm_CleanUp);
    
    if(CSSM_OK == cssm_GetModuleListLock(CSSM_WRITER_LOCK))
    {
        while(NULL != hModuleListHead)
        {
            TempModule = hModuleListHead;
            
            /* lock the record we'r about to delete */
            if (cssm_WriterLock (TempModule->hModuleSWMRLock) == CSSM_OK)           
            {
				CSSM_GUID tmpGuid;
				
				memcpy(&tmpGuid, &TempModule->GUID, sizeof(CSSM_GUID));

                /* Delete everything under the record - the function will move
                   hModuleListHead for us.  We'll also get the lib handle for
                   the module that implements SPI_ModuleUnload
                */
                /**********************************************************/

                ERR(rc = cssm_CleanInternalModuleRecord(&TempModule,
                                                        &MainLibHandle));

                if(CSSM_OK == rc)
                {
                        rc = port_GetProcAddress(MainLibHandle,
                                                 szModuleUnload,
                                                 (CSSM_PROC_ADDR *)&SPI_Unload_Func);

                        if((CSSM_OK == rc) &&
                           (NULL != SPI_Unload_Func))
                        {
                            /* We've taken the record out of the internal
                               lists, now we'll call the addin to notify it
                               of the unload.
                            */
                            ERR(rc = SPI_Unload_Func(&cssm30_GUID,
                                                     &tmpGuid,
                                                     CSSM_ModuleEventHandler,
                                                     NULL));
                        }
                }
                else
                {
                    rc = CSSMERR_CSSM_ADDIN_UNLOAD_FAILED;
                }
                
                
                /************************************************************/
                ERR(port_FreeLibrary(MainLibHandle));
            }
        }
        
        /* delete the global module list lock */
        cssm_WriterUnlock(hModuleListSWMRLock);
        cssm_SWMRDelete(hModuleListSWMRLock);
    }
    
    /* re-init these variables in case they are somehow used */
    hModuleListSWMRLock = NULL;
    hModuleListTail = NULL;
    
    RETURN (rc)
}



CSSM_RETURN CSSMAPI
cssm_LockMultiTaskMutex(cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord)
{
    CSSM_RETURN Ret = CSSM_OK;
    
    if(NULL != AttachRecord)
    {
        if(CSSM_FALSE == AttachRecord->ThreadSafe)
        {
            /* try to lock the mutex */
            if((Ret = port_LockMutex(AttachRecord->MultiTaskMutex,
                CSSM_INFINITE_WAIT)) != CSSM_OK)
            {
                /* it's not locked, so return is false */
                RETURN (Ret)
            }
        }
    }
    else
    {
        
        RETURN (CSSMERR_CSSM_INVALID_POINTER)
    }
    
    RETURN (Ret)
}



CSSM_RETURN CSSMAPI
cssm_UnlockMultiTaskMutex(cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord)
{
    if(NULL != AttachRecord)
    {
        if(CSSM_FALSE == AttachRecord->ThreadSafe)
        {
            /* unlock the mutex */
            port_UnlockMutex(AttachRecord->MultiTaskMutex);
        }
    }
    else
    {
        /* doubtful it'll get checked, but it's an error */
        
        RETURN (CSSMERR_CSSM_INVALID_POINTER)
    }
    RETURN(CSSM_OK)
        
}



/*-----------------------------------------------------------------------
*
*Name: cssm_GetAttachRecordAndLockMultiTask
*
*Description:
*   This function gets the attach infomation record given the addin handle.
*   It locks the record for reading and returns a pointer to the attach info
*   and callback functions for the module. It also locks the multitask 
*   mutex for the addin.
*
*Parameters:
* hAddIn(input) - AddIn handle of library to query
* AddinType(input) - type of the addin
* CallBack(output) - Pointer to a function pointer that will hold
*  the callback registered with the CSSM.
*
*Returns:
* NULL - In case of failures.
* NON-NULL - Pointer to the module information for the addin.
*---------------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
cssm_GetAttachRecordAndLockMultiTask(CSSM_HANDLE hAddIn,
                                     CSSM_SERVICE_TYPE AddinType,
                                     void **CallBack,
                                     cssm_INTERNAL_ATTACH_LIST_PTR *AttachInfo)
{
    CSSM_RETURN err = CSSM_OK;
    
    TRACE(cssm_trace_info, FNCODE_cssm_GetModuleRecordAndLockMultiTask);
    
    if((NULL == AttachInfo) ||
       (port_IsBadWritePtr(AttachInfo,
                           (uint32)sizeof(cssm_INTERNAL_ATTACH_LIST_PTR))) ||
       (NULL == CallBack))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
        
        /*
         * Get the moduleinfo record for the addin.
         */

        if(ERR(err = cssm_FindAttachAndMultiLock(hAddIn,
                                                 CSSM_NO_LOCK,
                                                 NULL,
                                                 CSSM_NO_LOCK,
                                                 NULL,
                                                 CSSM_NO_LOCK,
                                                 AttachInfo,
                                                 CSSM_READER_LOCK)) == CSSM_OK)
        {
            /*
             * Get the call back for this record using the AddinType.
             */
            ERR(err  = cssm_GetAddInCallBackByType(*AttachInfo,
                                                   AddinType,
                                                   CallBack));

            if(CSSM_OK == err)
            {
                /* If this module isn't threadsafe, we'll lock the module
                   mutex.  Note that this isn't totally effective since it
                   only provides safety from this CSSM.
                */
                if(CSSM_FALSE == (*AttachInfo)->ThreadSafe)
                {
                    ERR(err = port_LockMutex((*AttachInfo)->MultiTaskMutex,
                                             CSSM_INFINITE_WAIT));
                }
            }

            if(CSSM_OK != err)
            {
                /* error case, so we don't return values to caller */
                cssm_ReleaseAttachLock(*AttachInfo,
                                       CSSM_READER_LOCK);
                *AttachInfo = NULL;
                *CallBack = NULL;
            }
        }
    }

    RETURN (err)
}



CSSM_RETURN CSSMAPI
cssm_ReleaseAttachRecordAndUnlockMultiTask(cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord)
{
    CSSM_RETURN rc = CSSM_OK;
    /* just to make sure we don't walk a NULL pointer */
    if(NULL != AttachRecord)
    {
        if(CSSM_FALSE == AttachRecord->ThreadSafe)
        {
            rc = port_UnlockMutex(AttachRecord->MultiTaskMutex);
        }
        cssm_ReleaseAttachLock(AttachRecord,
            CSSM_READER_LOCK);
        RETURN ( rc )
    }
    else
        RETURN (CSSMERR_CSSM_INVALID_POINTER )
}

/*-----------------------------------------------------------------------
*
*Name: cssm_ReaderLock
*
*Description:
*   This routine locks the SWMRLock lock for reading.
*
*Parameters:
* SWMRLock(input) - Single-writer-multiple-reader lock to be locked for reading.
*
*Returns:
* CSSM_TRUE - SWMRLock is locked.
* CSSM_FALSE - Unable to lock SWMRLock.
*---------------------------------------------------------------------*/

CSSM_RETURN CSSMAPI cssm_ReaderLock (cssm_SWMRLock_PTR SWMRLock) 
{
    CSSM_RETURN rc = CSSM_OK;
    
    TRACE(cssm_trace_info, FNCODE_cssm_ReaderLock);
    
    /*
    * Get the Reader lock on the Module Record.
    */
    if (SWMRLock != NULL) 
    {
        rc = cssm_SWMRLockWaitToRead (SWMRLock, CSSM_INFINITE_WAIT);
        RETURN (rc)
    } else 
    {
            RETURN (CSSMERR_CSSM_INVALID_POINTER)
        }
}

/*-----------------------------------------------------------------------
*
*Name: cssm_ReaderUnlock
*
*Description:
*   This routine locks the SWMRLock lock for reading.
*
*Parameters:
* SWMRLock(input) - Single-writer-multiple-reader lock to be unlocked.
*
*Returns:
* void
*---------------------------------------------------------------------*/

CSSM_RETURN CSSMAPI cssm_ReaderUnlock (cssm_SWMRLock_PTR SWMRLock) 
{
    CSSM_RETURN rc = CSSM_OK;
    
    TRACE(cssm_trace_info, FNCODE_cssm_ReaderUnlock);
    /*
    * Release the reader lock.
    */
    if (SWMRLock != NULL) 
        rc = cssm_SWMRLockDoneReading (SWMRLock);
    else
        rc = CSSMERR_CSSM_INVALID_POINTER;
    RETURN (rc)
}


/*-----------------------------------------------------------------------
*
*Name: cssm_WriterLock
*
*Description:
*   This routine locks the SWMRLock lock for writing.
*
*Parameters:
* SWMRLock(input) - Single-writer-multiple-reader lock to be locked for writing.
*
*Returns:
* CSSM_TRUE - SWMRLock is locked.
* CSSM_FALSE - Unable to lock SWMRLock.
*---------------------------------------------------------------------*/

CSSM_RETURN CSSMAPI cssm_WriterLock (cssm_SWMRLock_PTR SWMRLock) 
{
    CSSM_RETURN rc = CSSM_OK;
    
    TRACE(cssm_trace_info, FNCODE_cssm_WriterLock);
    
    /*
    * Get the Writer lock on the Module Record.
    */
    if (SWMRLock != NULL) 
    {
        rc = cssm_SWMRLockWaitToWrite (SWMRLock, CSSM_INFINITE_WAIT);
        RETURN(rc)
    } 
    else 
    {
        RETURN(CSSMERR_CSSM_INVALID_POINTER)
    }
}

/*-----------------------------------------------------------------------
*
*Name: cssm_WriterUnlock
*
*Description:
*   This routine locks the SWMRLock lock for writing.
*
*Parameters:
* SWMRLock(input) - Single-writer-multiple-reader lock to be unlocked.
*
*Returns:
* void
*---------------------------------------------------------------------*/

CSSM_RETURN CSSMAPI cssm_WriterUnlock (cssm_SWMRLock_PTR SWMRLock) 
{
    CSSM_RETURN rc = CSSM_OK;
    TRACE(cssm_trace_info, FNCODE_cssm_WriterUnlock);
    
    /*
    * Release the writer lock.
    */
    if (SWMRLock != NULL) 
        rc = cssm_SWMRLockDoneWriting (SWMRLock);
    else
        rc = CSSMERR_CSSM_INVALID_POINTER;
    RETURN (rc)
}

/*-----------------------------------------------------------------------
*
*Name: cssm_WriterToReader
*
*Description:
*   This routine changes a writer lock to the reader lock, so that no other
*   writers can get in before getting the reader lock,
*
*Parameters:
* SWMRLock(input) - Single-writer-multiple-reader lock to be locked for writing.
*
*Returns:
* CSSM_TRUE - SWMRLock is locked.
* CSSM_FALSE - Unable to lock SWMRLock.
*---------------------------------------------------------------------*/

CSSM_RETURN CSSMAPI cssm_WriterToReader (cssm_SWMRLock_PTR SWMRLock) 
{
    CSSM_RETURN ret = CSSM_OK;
    
    TRACE(cssm_trace_info, FNCODE_cssm_WriterToReader);
    
    /*
    * Get the Writer lock on the Module Record.
    */
    if (SWMRLock != NULL) 
    {
        ret = cssm_SWMRLockWriteToRead (SWMRLock);
        RETURN(ret)
    } 
    else 
    {
        RETURN( CSSMERR_CSSM_INVALID_POINTER )
    }
    
}

/*-----------------------------------------------------------------------
*
*Name: cssm_SWMRInit
*
*Description:
*   This routine creates and intializes the SWMRLock.
*
*Parameters:
* SWMRLock(input) - Single-writer-multiple-reader lock to be creates/intialized.
*
*Returns:
* CSSM_TRUE - On succesful intialization
* CSSM_FAIL - In case of failures.
*---------------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
cssm_SWMRInit(cssm_SWMRLock_PTR *SWMRLock)
{
    CSSM_RETURN err = CSSM_OK;

    /*
    * Create the multiple-reader-single-write synchronization object.
    */
    if(NULL != SWMRLock)
    {
        if(((*SWMRLock) = (cssm_SWMRLock_PTR)local_malloc((uint32)sizeof(cssm_SWMRLock),
                                                          NULL)) == NULL)
        {
            ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
        }
        else
        {
            /* create the lock */
            if(ERR(err = cssm_SWMRLockCreate(*SWMRLock,
                                             NULL)) != CSSM_OK)
            {
                local_free((*SWMRLock), NULL);
            }
        }
    }
    RETURN (err)
}


/*-----------------------------------------------------------------------
*
*Name: cssm_SWMRDelete
*
*Description:
*   This routine creates and intializes the SWMRLock.
*
*Parameters:
* SWMRLock(input) - Single-writer-multiple-reader lock to be deleted.
*
*Returns:
* none
*---------------------------------------------------------------------*/

CSSM_RETURN CSSMAPI cssm_SWMRDelete ( cssm_SWMRLock_PTR SWMRLock )

{
/*
* Delete the multiple-reader-single-write synchronization object.
    */
    if ( SWMRLock != NULL )
    {
        /* we want unnamed locks */
        cssm_SWMRLockDelete( SWMRLock );
        local_free( SWMRLock, NULL );
        RETURN (CSSM_OK)
    }
    else
        RETURN (CSSMERR_CSSM_INVALID_POINTER)
}




/*-----------------------------------------------------------------------
*
*Name: cssm_CheckIfAttachRecordWithMatchingUSEEExists
*
*Description:
* Function checks if an attach record with the USEE tag exists in the servirecord's
* attcah list
*
*Parameters:
* pService(input) - ServiceRecord
* USEERequest(input) - USEE Request
*
*Returns:
* CSSM_BOOL - TRUE or FALSE
*---------------------------------------------------------------------*/

/* Expects the ServiceRecord to be reader locked before calling this function*/

CSSM_RETURN
cssm_CheckIfAttachRecordWithMatchingKeyHierarchyExists(CSSM_MODULE_HANDLE hAttach,
                                                       CSSM_KEY_HIERARCHY KeyHierarchy,
                                                       cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord,
                                                       CSSM_BOOL *Exists)
{
    CSSM_RETURN err = CSSM_OK;


    if((NULL == Exists) ||
       (NULL == ServiceRecord))
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }
    else
    {
	    if( ServiceRecord->KeyHierarchy & KeyHierarchy )
	    {
		    *Exists = CSSM_TRUE;
	    }
        else
        {
            *Exists = CSSM_FALSE;
        }
    }

    RETURN(err)
}

CSSM_RETURN CSSMAPI cssm_CheckAddress( cssm_INTERNAL_SERVICE_LIST_PTR pService,
                                      cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord,
									  CSSM_PROC_ADDR Address, 
									  CSSM_BOOL* IsOK, 
									  CSSM_PRIVILEGE Priv)
{
    CSSM_RETURN rc = CSSM_OK;
#ifdef ISL_INCLUDED
    uint32 i;

    if( IsOK == NULL )
    {
        RETURN ( CSSMERR_CSSM_INVALID_POINTER )
    }
    *IsOK = CSSM_FALSE;
    for( i=0; i< pService->NumSPCreds; i++)
	{
	    if (EISL_CheckAddressWithinModule(pService->SPCredentialPtr[i], 
				    (ISL_FUNCTION_PTR)Address) == ISL_OK)
	    {
		    *IsOK = CSSM_TRUE;
		    break;
	    }
	}
    if( !(*IsOK) )
        rc = CSSMERR_CSSM_FUNCTION_INTEGRITY_FAIL;
    if( rc == CSSM_OK)
    {
        if( Priv != 0 )
        {
	        uint32 reqUseeTag = (CSSM_UINT64_HIGH(Priv)) & CSSM_USEE_MASK;
	        uint32 reqTOGPriv = (CSSM_UINT64_LOW(Priv)) & CSSM_TOGVENDOR_MASK;
            CSSM_BOOL bUseeFound = CSSM_FALSE;
            CSSM_BOOL bTOGFound = CSSM_FALSE;
            CSSM_BOOL bVendorFound = CSSM_TRUE; 
           	uint32 i = 0;

		    while ( ( !bUseeFound ) && ( i < pService->NumSPTags ) )
		    {
			    if( reqUseeTag != pService->SPTags[i] )
				    i++;
			    else
				    bUseeFound = CSSM_TRUE;
		    }
			if( reqTOGPriv != 0 )
			{
				if( reqTOGPriv < 0x8000000 )
				{
					while ( ( !bTOGFound ) && ( i < pService->NumTOGPrivs ) )
					{
						if( reqTOGPriv != pService->TOGPrivs[i] )
							i++;
						else
							bTOGFound = CSSM_TRUE;
					}
				}
				else
				{
					bVendorFound = CSSM_FALSE;
					while ( ( !bVendorFound ) && ( i < AttachRecord->NumVendorPrivs ) )
					{
						if( reqTOGPriv != AttachRecord->VendorPrivs[i] )
							i++;
						else
							bVendorFound = CSSM_TRUE;
					}
				}
			}
			else
			{
				bTOGFound = CSSM_TRUE;
			}
            if( bUseeFound && bTOGFound && bVendorFound )
            {
                rc = CSSM_OK;
            }
            else
            {
                rc = CSSMERR_CSSM_PRIVILEGE_NOT_GRANTED;
            }
        }
    }
#endif
    RETURN (rc)
}

CSSM_RETURN CSSMAPI cssm_PrivilegeCheckCaller (cssm_CALLER_LIST_ENTRY_PTR CallerListEntry,
		                                       cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord,
											   CSSM_PRIVILEGE InPriv )
{
    uint32 reqUseeTag = (CSSM_UINT64_HIGH(InPriv)) & CSSM_USEE_MASK;
	uint32 reqTOGPriv = (CSSM_UINT64_LOW(InPriv)) & CSSM_TOGVENDOR_MASK;
    CSSM_BOOL bUseeFound = CSSM_FALSE;
    CSSM_BOOL bTOGFound = CSSM_FALSE;
    CSSM_BOOL bVendorFound = CSSM_TRUE;
    uint32 i = 0;
	CSSM_RETURN rc = CSSM_OK;
	i = 0;
	if( InPriv != 0)
	{
		while ( ( !bUseeFound ) && ( i < CallerListEntry->NumUSEETags ) )
		{
			if( reqUseeTag != CallerListEntry->USEETags[i] )
				i++;
			else
				bUseeFound = CSSM_TRUE;
		}
		i = 0;
		if( reqTOGPriv != 0)
		{
			if( reqTOGPriv < 0x8000000 )
			{
				while ( ( !bTOGFound ) && ( i < CallerListEntry->NumTOGPrivs ) )
				{
					if( reqTOGPriv != CallerListEntry->TOGPrivs[i] )
						i++;
					else
						bTOGFound = CSSM_TRUE;
				}
			}
			else
			{
				if( CallerListEntry->KeyHierarchy & AttachRecord->KeyHierarchy )
				{
					cssm_VENDOR_PRIVILEGES_PTR tempPtr = CallerListEntry->VendorPrivs;
					CSSM_BOOL bKHFound = CSSM_FALSE;
					while ( tempPtr != NULL )
					{
						if( tempPtr->KeyHierarchy & AttachRecord->KeyHierarchy )
						{
							bKHFound = CSSM_TRUE;
							break;
						}
						else
							tempPtr = tempPtr->Next;
					}
					if( bKHFound )
					{
						bVendorFound = CSSM_FALSE;
						while ( ( !bVendorFound ) && ( i < tempPtr->NumVendorPrivs ) )
						{
							if( reqTOGPriv != tempPtr->VendorPrivs[i] )
								i++;
							else
								bVendorFound = CSSM_TRUE;
						}
					}
				}
			}
		}
		else
		{
			bTOGFound = CSSM_TRUE;
		}
		if( bUseeFound &&  bTOGFound && bVendorFound ) 
		{
			rc = CSSM_OK;
		}
		else
		{
			rc = CSSMERR_CSSM_PRIVILEGE_NOT_GRANTED;
		}
	}
	RETURN (rc)
}


CSSM_RETURN CSSMAPI 
cssm_IsFuncCallOK(CSSM_MODULE_HANDLE hAddin,
                  CSSM_PROC_ADDR SrcAddress, /* application */
                  CSSM_PROC_ADDR DestAddress,
                  CSSM_PRIVILEGE InPriv,
                  CSSM_PRIVILEGE *OutPriv,
                  CSSM_BITMASK Hints,
                  CSSM_BOOL *IsOK) /* addin */
{
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    CSSM_BOOL CheckApp = CSSM_FALSE;
    CSSM_BOOL CheckSP = CSSM_FALSE;

    cssm_INTERNAL_SERVICE_LIST_PTR pService = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
	cssm_CALLER_LIST_ENTRY_PTR CallerListEntry = NULL;
	cssm_CALLER_LIST_PTR CallerList = NULL;

    CSSM_PROC_ADDR AddressToCheck;
#endif

    if(CSSM_FALSE == port_IsBadWritePtr(IsOK,
                                        sizeof(CSSM_BOOL)))
	{

#ifdef ISL_INCLUDED

        if((CSSM_FALSE == port_IsBadCodePtr(SrcAddress)) &&
           (CSSM_FALSE == port_IsBadCodePtr(DestAddress)))
        {
            /* determine what mix of App/SP we should verify */
            if((0 != InPriv) ||
               (cssm_GetPVCPolicy() & CSSM_PVC_APP))
            {
                CheckApp = CSSM_TRUE;
            }

            if((0 != InPriv) ||
               (cssm_GetPVCPolicy() & CSSM_PVC_SP))
            {
                CheckSP = CSSM_TRUE;
            }


            if((CSSM_TRUE == CheckApp) ||
               (CSSM_TRUE == CheckSP))
            {
                /* if CheckSP is true, we'll need the service record and we
                   need the attach record if there's any privilege
                */
                ERR(err = cssm_FindAttachAndMultiLock(hAddin,
                                                      CSSM_NO_LOCK,
                                                      NULL,
                                                      CSSM_NO_LOCK,
                                                      &pService,
                                                      CSSM_READER_LOCK,
                                                      &AttachRecord,
                                                      CSSM_READER_LOCK));
                if(CSSM_OK == err)
                {
                    if(CSSM_TRUE == CheckApp)
		            {
                        if(CSSM_HINT_ADDRESS_SP == Hints)
                        {
                            AddressToCheck = DestAddress;
                        }
                        else
                        {
                            AddressToCheck = SrcAddress;
                        }

                        ERR(err = cssm_FindCallerListEntryAndMultiLock(AddressToCheck,
                                                                       CSSM_NO_LOCK,
                                                                       NULL,
                                                                       CSSM_NO_LOCK,
                                                                       &CallerList,
                                                                       CSSM_READER_LOCK,
                                                                       &CallerListEntry));
                        if(CSSM_OK == err)
                        {
                            if(0 != InPriv)
                            {
                                /* since the caller is known, we check privilege if we
                                   need to
                                */
                                ERR(err = cssm_PrivilegeCheckCaller(CallerListEntry,
                                                                    AttachRecord,
                                                                    InPriv));
                            }
                        }
                    }

                    if(CSSM_TRUE == CheckSP)
                    {
                        if(CSSM_HINT_ADDRESS_SP == Hints)
                        {
                            AddressToCheck = SrcAddress;
                        }
                        else
                        {
                            AddressToCheck = DestAddress;
                        }

                        ERR(err = cssm_CheckAddress(pService,
                                                    AttachRecord,
                                                    AddressToCheck,
                                                    IsOK,
                                                    InPriv));
                    }

                    cssm_ReleaseServiceLock(pService,
                                            CSSM_READER_LOCK);
                    cssm_ReleaseAttachLock(AttachRecord,
                                           CSSM_READER_LOCK);
                }
            }

            /* if required and everything went correctly, set the OutPriv
               to the privilege this call should have
            */
            if((CSSM_OK == err) &&
               (CSSM_FALSE == port_IsBadWritePtr(OutPriv,
                                                 sizeof(CSSM_PRIVILEGE))))
            {
                if(0 == InPriv)
                {
                    if(CSSM_PRIVILEGE_SCOPE_PROCESS == PrivilegeScope)
                    {
                        ERR(err = cssm_GetGlobalPrivilege(OutPriv));
                    }
                    else
                    {
                        if(CSSM_PRIVILEGE_SCOPE_THREAD == PrivilegeScope)
                        {
                            ERR(err = cssm_GetThreadPrivilege(OutPriv));
                        }
                    }
                }
                else
                {
                    *OutPriv = InPriv;
                }
            }
        }
        else
        {
            ERR(err = CSSMERR_CSSM_INVALID_POINTER);
        }

#endif /* ISL_INCLUDED */

        if(CSSM_OK == err)
        {
            *IsOK = CSSM_TRUE;
        }
        else
        {
            *IsOK = CSSM_FALSE;
        }
    }
    else
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }

    return err;
}



CSSM_RETURN CSSMAPI cssm_FailModuleLoad(const CSSM_GUID *ModuleGuid,
                                        CSSM_API_ModuleEventHandler AppNotifyCallback,
                                        const void* AppNotifyCallbackCtx)
{
    CSSM_RETURN Ret = CSSM_OK;
    cssm_INTERNAL_MODULE_LIST_PTR ModuleRecord = NULL;
    
    if( (Ret = cssm_FindModuleByGUIDAndMultiLock((CSSM_GUID_PTR)ModuleGuid,
        CSSM_NO_LOCK,
        &ModuleRecord,
        CSSM_WRITER_LOCK)) == CSSM_OK)
    {
        if(NULL == ModuleRecord)
        {
            Ret = CSSMERR_CSSM_INVALID_GUID;
        }
        else
        {
            Ret = cssm_PruneCallbackList(ModuleRecord,
                AppNotifyCallback,
                AppNotifyCallbackCtx);
        }
        cssm_ReleaseModuleLock(ModuleRecord,
            CSSM_WRITER_LOCK);
    }
    RETURN (Ret)
}


CSSM_RETURN CSSMAPI
cssm_DeleteModuleRecordonLoadFail(const CSSM_GUID *ModuleGuid,
                                  CSSM_API_ModuleEventHandler AppNotifyCallback,
                                  const void* AppNotifyCallbackCtx)
{
    CSSM_RETURN Ret = CSSM_OK;
    cssm_INTERNAL_MODULE_LIST_PTR ModuleRecord = NULL;
    CSSM_LIB_HANDLE MainLibHandle;


    if(NULL != ModuleGuid)
    {
        /* Even if the module is not found in the caller list, it is OK, because introduce may not
        have been called at this point.
        So we don't have to check the return code here*/
#ifdef ISL_INCLUDED
        cssm_ModuleImplicitUnintroduce( ModuleGuid );
#endif
        /* get global lock because we might need it */
        if((Ret = cssm_GetModuleListLock(CSSM_WRITER_LOCK)) == CSSM_OK)
        {
            /* Iterate down the list and find the record with a matching GUID */
            ModuleRecord = hModuleListHead;
            if(NULL != ModuleRecord)
            {
                while((NULL != ModuleRecord) &&
                    (CSSM_TRUE != CSSM_CompareGuids((*ModuleGuid),
                    ModuleRecord->GUID)))
                {
                    ModuleRecord = ModuleRecord->Next;
                }
                
                if(NULL != ModuleRecord)
                {
                    if (cssm_WriterLock (ModuleRecord->hModuleSWMRLock) != CSSM_OK)                    
                    {
                        ModuleRecord = NULL;
                    }
                    else
                    {
                    /* Make sure another thread didn't change these values
                    on us.
                        */
                        if(CSSM_TRUE != CSSM_CompareGuids((*ModuleGuid),
                            ModuleRecord->GUID))
                        {
                            cssm_ReleaseModuleLock(ModuleRecord,
                                CSSM_WRITER_LOCK);
                            ModuleRecord = NULL;
                        }
                    }
                }
                
                
                /* ModuleRecord should now either have a writer lock on our
                record or have determined that our record doesn't exist.
                */
                if(NULL == ModuleRecord)
                {
                    cssm_ReleaseModuleListLock(CSSM_WRITER_LOCK);
                    Ret = CSSMERR_CSSM_ADDIN_LOAD_FAILED;
                }
                else
                {
                    Ret = cssm_PruneCallbackList(ModuleRecord,
                        AppNotifyCallback,
                        AppNotifyCallbackCtx);
                    if(CSSM_OK != Ret)
                    {
                        cssm_ReleaseModuleLock(ModuleRecord,
                            CSSM_WRITER_LOCK);
                        cssm_ReleaseModuleListLock(CSSM_WRITER_LOCK);
                        Ret = CSSMERR_CSSM_ADDIN_UNLOAD_FAILED;
                    }
                }
            }
            else
            {
                cssm_ReleaseModuleListLock(CSSM_WRITER_LOCK);
                Ret = CSSMERR_CSSM_MODULE_NOT_LOADED;
            }
        }
        
        /* If everything's ok here, then we just need to perform the unload.
        After that we only have to free the removed record - if it's to
        be freed.
        */
        if(CSSM_OK == Ret)
        {
            /* Check if we should free the record - we need to at least unlock
               it.  If we should free it, we'll free memory associated with
               the record and unload the library.
            */
            if((NULL != ModuleRecord) &&
               (NULL == ModuleRecord->EventNotificationCallbackList))
            {
                /* If that's all the load calls we've gotten then
                   it's time to remove this module from our list.
                */
                ERR(Ret = cssm_CleanInternalModuleRecord(&ModuleRecord,
                                                         &MainLibHandle));
                if(CSSM_OK == Ret)
                {
                    ERR(port_FreeLibrary(MainLibHandle));
                }
            }
            else
            {
                cssm_ReleaseModuleLock(ModuleRecord,
                    CSSM_WRITER_LOCK);
            }
            
            /* release the module list lock in all cases */
            cssm_ReleaseModuleListLock(CSSM_WRITER_LOCK);
        }
    }
    RETURN(Ret)
        
}

#ifdef ISL_INCLUDED


CSSM_RETURN CSSMAPI
cssm_ModuleImplicitUnintroduce( const CSSM_GUID* ModuleGuid )
{
    CSSM_RETURN err = CSSM_OK;
	cssm_CALLER_LIST_PTR CallerList = NULL;
	cssm_CALLER_LIST_ENTRY_PTR CallerEntry = NULL;
	CSSM_PROC_ADDR RetAddressToRemove = NULL;
	/* now we go searching to find the correct record to remove */
	ERR(err = cssm_FindCallerListEntryByGuidAndMultiLock(ModuleGuid,
														 CSSM_NO_LOCK,
														 NULL,
														 CSSM_NO_LOCK,
														 &CallerList,
														 CSSM_READER_LOCK,
														 &CallerEntry));
	if(CSSM_OK == err)
	{
		/* get the start address - anything works as long as it's
		   within the module's address range
		*/
		RetAddressToRemove = CallerEntry->ModuleStartAddr;

		/* release our lock on this record */
		cssm_ReleaseCallerListEntry(CallerList,
									CSSM_READER_LOCK);
	}
	else
	{
		ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
	}
	if(CSSM_OK == err)
	{
		CallerEntry->RefCount--;
		if( 0 == CallerEntry->RefCount)
		{
			/* now we actually remove the entry from the caller list */
			ERR(err = cssm_RemoveCallerListEntry(RetAddressToRemove,
												 &CallerEntry));
			if(CSSM_OK == err)
			{
				/* now we free the memory that was in the entry */
				ERR(err = cssm_DestroyCallerListEntry(&CallerEntry));
			}
		}
	}
    RETURN (err)
}

CSSM_RETURN CSSMAPI 
cssm_ModuleImplicitIntroduce( const CSSM_GUID *ModuleGuid, 
                              CSSM_KEY_HIERARCHY KeyHierarchy,
                              uint32 NumSections,
                              ISL_VERIFIED_MODULE_PTR* VerifiedModulePtr)
{
    uint32 i = 0;
    CSSM_RETURN err;
    cssm_CALLER_LIST_ENTRY_PTR CallerRecord = NULL;
    cssm_CALLER_LIST_PTR CallerList = NULL;
    cssm_MODULE_COMPONENTS_PTR tempComponentPtr = NULL;

    ERR(err = cssm_FindCallerListEntryByGuidAndMultiLock(ModuleGuid,
                                                         CSSM_NO_LOCK,
                                                         NULL,
                                                         CSSM_NO_LOCK,
                                                         &CallerList,
                                                         CSSM_WRITER_LOCK,
                                                         &CallerRecord));
    if( CSSM_OK == err )
    {
        CallerRecord->RefCount++;

        cssm_ReleaseCallerListEntry(CallerList,
                                    CSSM_WRITER_LOCK);

    }
    else
    {
        /* clear the error set by failing to find the caller */
        err = CSSM_OK;

        /* allocate caller list entry and initialize record */
        CallerRecord = local_calloc(sizeof(cssm_CALLER_LIST_ENTRY),
                                    1,
                                    NULL);
        if(NULL == CallerRecord)
        {
            ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
        }
        else
        {
            memcpy(&CallerRecord->CallerGuid,
                   ModuleGuid,
                   sizeof(CSSM_GUID));

            CallerRecord->KeyHierarchy = KeyHierarchy;
            CallerRecord->bAddin = CSSM_TRUE;
            ERR(err = cssm_IdentifyVerifiedModuleByGUID(NumSections,
                                                    VerifiedModulePtr,
                                                    ModuleGuid,
                                                    &i));
            if( CSSM_OK == err )
            {
                port_ComputeAddressInModule(EISL_GetLibHandle(VerifiedModulePtr[i]),
                                                            (void**)(&CallerRecord->ModuleStartAddr),
                                                            (void**)(&CallerRecord->ModuleEndAddr));
            
                tempComponentPtr = (cssm_MODULE_COMPONENTS_PTR)local_calloc(sizeof(cssm_MODULE_COMPONENTS),
                                                                                                    1,
                                                                                                    NULL);
                if(NULL != tempComponentPtr)
                {
                    tempComponentPtr->StartAddr = CallerRecord->ModuleStartAddr;
                    tempComponentPtr->EndAddr = CallerRecord->ModuleEndAddr;
                    tempComponentPtr->VerifiedModule = (ISL_VERIFIED_MODULE_PTR*)local_calloc( 
                                                                       sizeof( ISL_VERIFIED_MODULE_PTR) * NumSections, 
                                                                       1,
                                                                       NULL);
                    if( tempComponentPtr->VerifiedModule == NULL )
                    {
                        local_free( CallerRecord, NULL );
                        ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                    }
                    else
                    {
                        tempComponentPtr->NumComponents = NumSections;
                        tempComponentPtr->Next = NULL;
                        for( i=0; i < NumSections; i++)
                        {
                            tempComponentPtr->VerifiedModule[i] = EISL_DuplicateVerifiedModulePtr( VerifiedModulePtr[i]);
                        }
                        CallerRecord->ModuleComponents = tempComponentPtr;

                    }
                }
                else
                {
                    /* error out so we release the memory we have
                       for this caller entry
                    */
                    ERR(err = CSSMERR_CSSM_MEMORY_ERROR);
                }
                if( CSSM_OK == err )
                {
                    CallerRecord->RefCount++;
                    ERR(err = cssm_InsertCallerListEntry(CallerRecord));
                }
            }
        }
    }
    RETURN (err)
}


#endif
CSSM_RETURN CSSMAPI
cssm_ModuleLoad(const CSSM_GUID *ModuleGuid,
		CSSM_KEY_HIERARCHY KeyHierarchy,
                CSSM_API_ModuleEventHandler AppNotifyCallback,
                const void* AppNotifyCallbackCtx)
{
    CSSM_RETURN Ret = CSSM_OK;
    cssm_INTERNAL_MODULE_LIST_PTR ModuleRecord = NULL;
    LOAD_MODULE_FUNC_PTR SPI_Load_Func = NULL;
    char szModuleLoad[] = "CSSM_SPI_ModuleLoad";
    CSSM_DATA tmpPath;

#ifdef ISL_INCLUDED
    CSSM_KEY_HIERARCHY tempKeyHierarchy;

    uint32 i;
    CSSM_DATA tmpCredential;
    ISL_CONST_DATA ProcedureName;
    uint32 NumExpSections = 0;
    ISL_VERIFIED_MODULE_PTR* DockingModuleExpCredentialPtr = NULL;
	uint32 NumIntegSections = 0;
    ISL_VERIFIED_MODULE_PTR* DockingModuleIntegCredentialPtr = NULL;
    ISL_CONST_DATA DMCredentialPath;
    ISL_CONST_DATA DMSearchPath;
    CSSM_API_MEMORY_FUNCS MemoryFuncs;
    
    
    ProcedureName.Length = (uint32)strlen(szModuleLoad);
    ProcedureName.Data = (const unsigned char*)szModuleLoad;
    
#else
    CSSM_DATA ModulePath;
    CSSM_DATA tmpModuleName;
    CSSM_LIB_HANDLE LibHandle = NULL;
#endif  /* ISL_INCLUDED */


    /* First, we check if this module is already loaded */
    if(ERR(Ret = cssm_FindModuleByGUIDAndMultiLock((CSSM_GUID_PTR)ModuleGuid,
                                                   CSSM_NO_LOCK,
                                                   &ModuleRecord,
                                                   CSSM_WRITER_LOCK)) != CSSM_OK)
    {
        /* Since this is the first call to load, we may need to
           authenticate the module.
        */

#ifdef ISL_INCLUDED
        /*
        * Get the credential, search path and section name from MDS
        */
        if(ERR(Ret = mdsutil_GetModuleCredentialInfo((CSSM_GUID_PTR)ModuleGuid,
                                                     &tmpCredential,
                                                     NULL,
                                                     &tmpPath)) != CSSM_OK)
        {
            RETURN (CSSMERR_CSSM_MDS_ERROR)
        }

        DMCredentialPath.Length = tmpCredential.Length;
        DMCredentialPath.Data = tmpCredential.Data;
        DMSearchPath.Length = tmpPath.Length;
        DMSearchPath.Data =  tmpPath.Data;

        MemoryFuncs.malloc_func = internal_malloc;
        MemoryFuncs.calloc_func = internal_calloc;
        MemoryFuncs.free_func = internal_free;
        MemoryFuncs.realloc_func = internal_realloc;

        if(CSSM_KEY_HIERARCHY_NONE == KeyHierarchy)
        {
            tempKeyHierarchy = CSSM_KEY_HIERARCHY_INTEG;
        }
        else
        {
            tempKeyHierarchy = KeyHierarchy;
        }

        if(CSSM_KEY_HIERARCHY_INTEG & tempKeyHierarchy)
        {
            /* verify the docking module */
            DockingModuleIntegCredentialPtr = cssm_VerifyCredentialsAndLoadModules(DMCredentialPath,
                                                            DMSearchPath,
                                                            tempKeyHierarchy,
                                                            CSSM_TRUE,
                                                            &NumIntegSections,
                                                            &MemoryFuncs);
            if(NULL == DockingModuleIntegCredentialPtr)
            {
                ERR(Ret = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);
            }
		}

        if(CSSM_KEY_HIERARCHY_EXPORT & tempKeyHierarchy)
		{
            /* see if the module has already been loaded into memory */
            if(CSSM_KEY_HIERARCHY_INTEG & tempKeyHierarchy)
            {
                /* if the module failed one integrity check, we won't bother
                   with a second one
                */
                if(CSSM_OK == Ret)
                {
        			/* verify the docking module */
        			DockingModuleExpCredentialPtr = cssm_VerifyCredentialsAndLoadModules(DMCredentialPath,
                                            				DMSearchPath,
                                                            tempKeyHierarchy,
                                                            CSSM_FALSE,
                                                            &NumExpSections,
                                                            &MemoryFuncs);

                    if(NULL == DockingModuleExpCredentialPtr)
	                {
		                Ret = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED;
	                }
                }
            }
            else
            {
        		/* verify the docking module */
        		DockingModuleExpCredentialPtr = cssm_VerifyCredentialsAndLoadModules(DMCredentialPath,
                                            			DMSearchPath,
                                                        tempKeyHierarchy,
                                                        CSSM_TRUE,
                                                        &NumExpSections,
                                                        &MemoryFuncs);

                if(NULL == DockingModuleExpCredentialPtr)
	            {
		            Ret = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED;
	            }
            }
		}

        local_free ((char *)DMCredentialPath.Data, NULL);
        local_free ((char *)DMSearchPath.Data, NULL);

        if(Ret != CSSM_OK)
        {
            /* we only need to check about freeing the integrity credentials,
               since those can be the only ones successfully created while
               still ending up in this error case.
            */
            if(NULL != DockingModuleIntegCredentialPtr)
            {
                i = 0;
                while(i < NumIntegSections)
                {
                    port_FreeLibrary(EISL_GetLibHandle(DockingModuleIntegCredentialPtr[i]));
                    i++;
                }

				ERR(cssm_FreeVerifiedCredentials(DockingModuleIntegCredentialPtr,
												 NumIntegSections));
				local_free(DockingModuleIntegCredentialPtr,
                           NULL);
            }

            RETURN (Ret)
        }
#else
        /*
        * Get the credential, search path and section name from MDS
        */
        if ( (Ret = mdsutil_GetModuleCredentialInfo ((CSSM_GUID_PTR)ModuleGuid, NULL, &tmpModuleName,
                                                            &tmpPath)) != CSSM_OK)
        {
            RETURN (CSSMERR_CSSM_MDS_ERROR)
        }
        
        Ret = mdsutil_GetModulePath (tmpModuleName, tmpPath, &ModulePath);
        local_free (tmpModuleName.Data, NULL);
        local_free (tmpPath.Data, NULL);
        if (Ret != CSSM_OK)
        {
            RETURN (CSSMERR_CSSM_MDS_ERROR)
        }
        /* Load the library directly */
        Ret = port_LoadLibrary((const char *)ModulePath.Data, &LibHandle);
        local_free (ModulePath.Data, NULL); 
        if ( (Ret != CSSM_OK) && (LibHandle == NULL) )
            RETURN (Ret)                           
            
#endif  /* ISL_INCLUDED */
            
        /* If all is ok, we've now loaded the module. */
        if( Ret == CSSM_OK )
        {
            /* We now make an entry in the module list */
            if(CSSM_OK != cssm_NewModuleRecord(ModuleGuid,
                &ModuleRecord))
            {
                /* we may've failed because another thread loaded the module */
                if(CSSM_OK == cssm_FindModuleByGUIDAndMultiLock((CSSM_GUID_PTR)ModuleGuid,
                    CSSM_NO_LOCK,
                    &ModuleRecord,
                    CSSM_WRITER_LOCK))
                {
                    Ret = cssm_UpdateCallbackList(ModuleRecord,
                        AppNotifyCallback,
                        AppNotifyCallbackCtx);
                    if(CSSM_OK != Ret)
                    {
                        cssm_ReleaseModuleLock(ModuleRecord,
                            CSSM_WRITER_LOCK);
                    }
                }
                else
                {
                    Ret = CSSMERR_CSSM_INVALID_GUID;
                }
            }
            else
            {
            /* Our addition worked.  Now we register the callback and
            either the new lib handle or the verified module ptr.
            After that call the SPI function.  Remember that the
            creation left the record in a writer_lock state.
                */
#ifdef ISL_INCLUDED                 
				if( NumIntegSections != 0)
					ModuleRecord->NumSections = NumIntegSections;
				else if( NumExpSections != 0)
					ModuleRecord->NumSections = NumExpSections;

				if( DockingModuleIntegCredentialPtr != NULL )
				{
					ModuleRecord->ModuleCredentialPtr = DockingModuleIntegCredentialPtr;
					if( DockingModuleExpCredentialPtr != NULL )
					{
						cssm_FreeVerifiedCredentials( DockingModuleExpCredentialPtr, NumExpSections);
						local_free( DockingModuleExpCredentialPtr, NULL);
						DockingModuleExpCredentialPtr = NULL;
						NumExpSections = 0;
					}
				}
				else if( DockingModuleExpCredentialPtr != NULL )
					ModuleRecord->ModuleCredentialPtr = DockingModuleExpCredentialPtr;
				
#else                  
                ModuleRecord->LibHandle = LibHandle;

#endif  /* ISL_INCLUDED */
                
                Ret = cssm_UpdateCallbackList(ModuleRecord,
                    AppNotifyCallback,
                    AppNotifyCallbackCtx);
                
                if(CSSM_OK == Ret)
                {               
                /*
                * The final thing we must do to finish the module record
                * is to set the usage mask and the threadsafe flag.
                    */
                    ERR(Ret = mdsutil_RegReadUsage(ModuleGuid,
                                                   &ModuleRecord->ServiceSummary));
                    if(0 == ModuleRecord->ServiceSummary)
                    {
                        cssm_ReleaseModuleLock(ModuleRecord,
                            CSSM_WRITER_LOCK);
                        cssm_DeleteModuleRecordonLoadFail ( ModuleGuid,
                            AppNotifyCallback,
                            AppNotifyCallbackCtx);
                        Ret = CSSMERR_CSSM_INVALID_SERVICE_MASK;                        
                    }
                    else
                    {
                        ERR(Ret = mdsutil_RegGetThreadSafe(ModuleGuid,
                                                           &ModuleRecord->ThreadSafe));

                        /* see if we should create the multitask mutex */
                        if(CSSM_OK == Ret)
                        {
                            if(CSSM_FALSE == ModuleRecord->ThreadSafe)
                            {
                                ERR(Ret = port_CreateMutex(NULL,
                                                           &ModuleRecord->MultiTaskMutex));

                                if(CSSM_OK != Ret)
                                {
                                    cssm_ReleaseModuleLock(ModuleRecord,
                                                           CSSM_WRITER_LOCK);
                                    cssm_DeleteModuleRecordonLoadFail(ModuleGuid,
                                                                      AppNotifyCallback,
                                                                      AppNotifyCallbackCtx);
                                }
                            }
                        }
                        else
                        {
                            cssm_ReleaseModuleLock(ModuleRecord,
                                CSSM_WRITER_LOCK);
                            cssm_DeleteModuleRecordonLoadFail ( ModuleGuid,
                                AppNotifyCallback,
                                AppNotifyCallbackCtx);
                            ERR(Ret = CSSMERR_CSSM_MDS_ERROR);
                        }
                    }                   
                }
                else
                {
                    cssm_ReleaseModuleLock(ModuleRecord,
                        CSSM_WRITER_LOCK);
                    cssm_DeleteModuleRecordonLoadFail ( ModuleGuid,
                        AppNotifyCallback,
                        AppNotifyCallbackCtx);
                }
            }
        }
    }
    else
    {
        Ret = cssm_UpdateCallbackList(ModuleRecord,
            AppNotifyCallback,
            AppNotifyCallbackCtx);
        
        if(CSSM_OK != Ret)
        {
            cssm_ReleaseModuleLock(ModuleRecord,
                CSSM_WRITER_LOCK);
        }
    }
    
    if(CSSM_OK == Ret)
    {
        /* Now we get the address for the function we're about to call. */
#ifdef ISL_INCLUDED

        /* find the correct module */
        if(CSSM_OK == ERR(cssm_IdentifyVerifiedModuleByGUID(ModuleRecord->NumSections,
                                                            ModuleRecord->ModuleCredentialPtr,
                                                            &ModuleRecord->GUID,
                                                            &i)))
        {
            SPI_Load_Func = (LOAD_MODULE_FUNC_PTR)EISL_LocateProcedureAddress(ModuleRecord->ModuleCredentialPtr[i],
                                                                              ProcedureName);
        }
#else
          Ret = port_GetProcAddress(ModuleRecord->LibHandle, szModuleLoad, (CSSM_PROC_ADDR*)&SPI_Load_Func);
#endif
        if( CSSM_OK == Ret )
        {
#ifdef ISL_INCLUDED
            ERR(Ret = cssm_ModuleImplicitIntroduce( ModuleGuid, KeyHierarchy,
                                                ModuleRecord->NumSections,
                                                ModuleRecord->ModuleCredentialPtr));
#endif
            /* Unlock the lock before the SPI call - if we don't we'll
            likely deadlock if the module is static.
            */
        }
        cssm_ReleaseModuleLock(ModuleRecord,
            CSSM_WRITER_LOCK);
        
        
        if((CSSM_OK == Ret) &&
           (NULL != SPI_Load_Func))
        {
            /* Now we call the addin's load function */
            Ret = SPI_Load_Func(&cssm30_GUID,
                ModuleGuid,
                CSSM_ModuleEventHandler, /* CSSM Notify Callback */
                NULL);                       /* Callback context */
            
            if(CSSM_OK != Ret)
            {
                cssm_DeleteModuleRecordonLoadFail(ModuleGuid,
                                                  AppNotifyCallback,
                                                  AppNotifyCallbackCtx);
            }
        }
        else
        {
            cssm_DeleteModuleRecordonLoadFail(ModuleGuid,
                                              AppNotifyCallback,
                                              AppNotifyCallbackCtx);
            ERR(Ret = CSSMERR_CSSM_ADDIN_LOAD_FAILED);
        }
    }
    RETURN(Ret)
}



CSSM_RETURN CSSMAPI
cssm_ModuleUnload(const CSSM_GUID *ModuleGuid,
                  CSSM_API_ModuleEventHandler AppNotifyCallback,
                  const void* AppNotifyCallbackCtx)
{
    CSSM_RETURN Ret = CSSM_OK;
    cssm_INTERNAL_MODULE_LIST_PTR ModuleRecord = NULL;
    UNLOAD_MODULE_FUNC_PTR SPI_Unload_Func = NULL;
    char szModuleUnload[] = "CSSM_SPI_ModuleUnload";
    CSSM_LIB_HANDLE MainLibHandle = 0;
    CSSM_BOOL LastUnload = CSSM_FALSE;

#ifdef ISL_INCLUDED
    ISL_CONST_DATA FuncName;
    uint32 i;

    FuncName.Data = (uint8*) szModuleUnload;
    FuncName.Length = strlen(szModuleUnload);
#endif

    if(NULL != ModuleGuid)
    {
        /* get global lock because we might need it */
        if((Ret = cssm_GetModuleListLock(CSSM_WRITER_LOCK)) == CSSM_OK)
        {
            /* Iterate down the list and find the record with a matching GUID */
            ModuleRecord = hModuleListHead;
            if(NULL != ModuleRecord)
            {
                while((NULL != ModuleRecord) &&
                    (CSSM_TRUE != CSSM_CompareGuids((*ModuleGuid),
                    ModuleRecord->GUID)))
                {
                    ModuleRecord = ModuleRecord->Next;
                }
                
                if(NULL != ModuleRecord)
                {
                    if (cssm_WriterLock (ModuleRecord->hModuleSWMRLock) != CSSM_OK)                    
                    {
                        ModuleRecord = NULL;
                    }
                    else
                    {
                    /* Make sure another thread didn't change these values
                    on us.
                        */
                        if(CSSM_TRUE != CSSM_CompareGuids((*ModuleGuid),
                            ModuleRecord->GUID))
                        {
                            cssm_ReleaseModuleLock(ModuleRecord,
                                CSSM_WRITER_LOCK);
                            ModuleRecord = NULL;
                        }
                    }
                }
                
                /* ModuleRecord should now either have a writer lock on our
                record or have determined that our record doesn't exist.
                */
                if(NULL == ModuleRecord)
                {
                    cssm_ReleaseModuleListLock(CSSM_WRITER_LOCK);
                    Ret = CSSMERR_CSSM_MODULE_NOT_LOADED;
                }
                else
                {
                    Ret = cssm_PruneCallbackList(ModuleRecord,
                        AppNotifyCallback,
                        AppNotifyCallbackCtx);
                    if(CSSM_OK != Ret)
                    {
                        cssm_ReleaseModuleLock(ModuleRecord,
                            CSSM_WRITER_LOCK);
                        cssm_ReleaseModuleListLock(CSSM_WRITER_LOCK);
                    }
                }
            }
            else
            {
                cssm_ReleaseModuleListLock(CSSM_WRITER_LOCK);
                Ret = CSSMERR_CSSM_ADDIN_UNLOAD_FAILED;
            }
        }
        
        /* If everything's ok here, then we just need to perform the unload.
        After that we only have to free the removed record - if it's to
        be freed.
        */
        if(CSSM_OK == Ret)
        {
        /* We need to find and locate the unload function and call it,
        whether or not this is the final unload for this module.
            */
            if(ModuleRecord != NULL)
            {

#ifdef ISL_INCLUDED

                if(CSSM_OK == ERR(cssm_IdentifyVerifiedModuleByGUID(ModuleRecord->NumSections,
                                                                    ModuleRecord->ModuleCredentialPtr,
                                                                    &ModuleRecord->GUID,
                                                                    &i)))
                {
                    SPI_Unload_Func = (UNLOAD_MODULE_FUNC_PTR)EISL_LocateProcedureAddress(ModuleRecord->ModuleCredentialPtr[i],
                                                                                          FuncName);
                }
#else
                Ret = port_GetProcAddress(ModuleRecord->LibHandle,
                                          szModuleUnload,
                                          (CSSM_PROC_ADDR *)&SPI_Unload_Func);
#endif
            }

            /* Check if we should free the record - we need to at least unlock
            it.  If we should free it, we'll free memory associated with
            the record and unload the library.
            */
            if((NULL != ModuleRecord) &&
               (NULL == ModuleRecord->EventNotificationCallbackList))
            {
                /* If that's all the load calls we've gotten then
                   it's time to remove this module from our list.
                */
                ERR(Ret = cssm_CleanInternalModuleRecord(&ModuleRecord,
                                                         &MainLibHandle));
                LastUnload = CSSM_TRUE;
            }
            else
            {
                cssm_ReleaseModuleLock(ModuleRecord,
                    CSSM_WRITER_LOCK);
            }
            
            /* release the module list lock in all cases */
            cssm_ReleaseModuleListLock(CSSM_WRITER_LOCK);
            
            /* We've taken the record out of the internal lists, now we'll call
            the addin to notify it of the unload.
            */
            if((CSSM_OK == Ret) &&
               (NULL != SPI_Unload_Func))
            {
                ERR(Ret = SPI_Unload_Func(&cssm30_GUID,
                                          ModuleGuid,
                                          CSSM_ModuleEventHandler,
                                          NULL));
            }
            else
            {
                Ret = CSSMERR_CSSM_ADDIN_UNLOAD_FAILED;
            }
#ifdef ISL_INCLUDED
            if( CSSM_OK == Ret )
                Ret = cssm_ModuleImplicitUnintroduce( ModuleGuid );
#endif
            if(CSSM_TRUE == LastUnload)
            {
                ERR(port_FreeLibrary(MainLibHandle));
            }
        }
    }
    RETURN(Ret)
}


/*****************************************************************************
*   Name: cssm_SetupStandardSPICall
*
*   Description: This function performs the standard initialization and return
*   address checks performed by the majority of the CSSM APIs.  It acquires
*   all locks necessary for a safe call to a service provider, and acquires
*   the SPI function table for the call.  Note that the ModuleHandle must be
*   to a service provider as indicated by ModuleType.
*
*   Parameters:
*       retAddress (input) : This is the return address from the API caller.
*           This return address will be checked against the caller list to make
*           sure this caller is authorized.
*       ModuleHandle (input) : This handle is used to select a particular
*           attach record, ensuring that the attach will be around for the
*           duration of the SPI call.
*       ModuleType (input) : This is the module type (CL, DL, CSP, etc.) that
*           the API function is about to call.  This is checked against the
*           attach record to make sure a call will be made with the correct
*           function table.
*       CallBack (output) : The function table associated with the ModuleHandle
*           and the ModuleType.
*       AttachRecord (output) : The attach record referenced by ModuleHandle.
*
*   WARNING: This function should not be modified very often, as it is used by
*       almost all API functions which route to a service provider.  Functions
*       like those in context.c and cspapi.c which require more to be done per
*       API call should either not use this function or use it in addition to
*       other code.  In short, don't change this function unless you absolutely
*       have to and know exactly what you're doing.
*
*****************************************************************************/

CSSM_RETURN
cssm_SetupStandardSPICall(CSSM_PROC_ADDR retAddress,
                          CSSM_HANDLE ModuleHandle,
                          CSSM_SERVICE_TYPE ModuleType,
                          void** CallBack,
                          cssm_INTERNAL_ATTACH_LIST_PTR *AttachRecord)
{
    CSSM_RETURN err = CSSM_OK;

    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK == ERR(err = cssm_CheckInit()))
    {

#ifdef ISL_INCLUDED

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }

#endif

        if(CSSM_OK == err)
        {
            /*
             * Get the AttachRecord record and the callback pointers
             * associated with the specified handle and the service mask.
             */
            ERR(err = cssm_GetAttachRecordAndLockMultiTask(ModuleHandle,
                                                           ModuleType,
                                                           CallBack,
                                                           AttachRecord));
        }
    }

    return err;
}



/*****************************************************************************
*   Name: cssm_CleanupStandardSPICall
*
*   Description: This function performs cleanup performed by the majority of
*   the CSSM APIs.  It releases the necessary locks.  It is assumed that this
*   function will be called after a successful call to
*   cssm_SetupStandardSPICall.
*
*   Parameters:
*       AttachRecord (input) : The attach record associated with the handle
*           used to make the recent SPI call.  This should be returned by
*           cssm_SetupStandardAPICall.
*
*   WARNING: This function should not be modified very often, as it is used by
*       almost all API functions which route to a service provider.  Functions
*       like those in context.c and cspapi.c which require more to be done per
*       API call should either not use this function or use it in addition to
*       other code.  In short, don't change this function unless you absolutely
*       have to and know exactly what you're doing.
*
*****************************************************************************/

CSSM_RETURN
cssm_CleanupStandardSPICall(cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord)
{
    CSSM_RETURN err = CSSM_OK;

    /* Release the AttachRecord Record.*/
    ERR(err = cssm_ReleaseAttachRecordAndUnlockMultiTask(AttachRecord));

    return err;
}
