/*-----------------------------------------------------------------------
 * File: EMMAPI.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "precomp.h"



/*---------------------------------------------------------------
 *
 *Name: CSSM_ListAttachedModuleManagers
 *
 *Description:
 *  This function returns a list of GUIDs for the currently attached/active
 *  module managers in the CSSM environment.
 *
 *Parameters: 
 *  ModuleManagerGuids (output) - List in which guids will be returned.
 *  NumberOfModuleManagers (input/output) - Pointer in which number of module
 *                          managers will be returned.
 *
 *Returns:
 *   CSSM_OK - List returned.
 *   All other returns indicate errors.
 *
 *-------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_ListAttachedModuleManagers(uint32 *NumberOfModuleManagers,
                                CSSM_GUID_PTR ModuleManagerGuids)
{
    CSSM_RETURN err = CSSM_OK;
    uint32 num;
    cssm_INTERNALMANAGERLIST_PTR tempPtr;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_ListAttachedModuleManagers);


    /* make sure that app as done a CSSM_Init */
    if(CSSM_OK != cssm_CheckInit())
    {
        ERR(err = CSSMERR_CSSM_NOT_INITIALIZED);
    }
    else
    {
        /*
         * Check for valid paramerers 
         */
        if((CSSM_TRUE == port_IsBadWritePtr(NumberOfModuleManagers,
                                           (uint32)sizeof(uint32))) ||
           ((NULL != ModuleManagerGuids) &&
            (CSSM_TRUE == port_IsBadWritePtr((void*)ModuleManagerGuids,
                                             (uint32)((*NumberOfModuleManagers) * sizeof(CSSM_GUID))))))
        {
            ERR(err = CSSMERR_CSSM_INVALID_POINTER);
        }
        else
        {

#ifdef ISL_INCLUDED
            /* if we have to, we check the application here */
            if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
            {
                ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL));
                if(CSSM_OK != err)
                {
                    ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
            }
#endif

        }
    }


    if(CSSM_OK == err)
    {
        /* If the ModuleManagerGuids parameter is NULL, we initialize
           NumberOfModuleManagerGuids to 0
        */
        if(NULL == ModuleManagerGuids)
        {
            *NumberOfModuleManagers = 0;
        }

        /*
         * Get the global reader lock.
         */
        if(CSSM_OK == cssm_ReaderLock(hEMMListSWMRLock))
        {
            tempPtr = EMMManagerListHead;
            num = 0;

            while ( (NULL != tempPtr) && (tempPtr->attachRefCount != 0) )
            {
                /* make sure we can copy into the output memory */
                if(num < *NumberOfModuleManagers)
                {
                    port_memcpy(&(ModuleManagerGuids[num]),
                                &tempPtr->EMMGuid,
                                sizeof(CSSM_GUID));
                }
                else
                {
                    ERR(err = CSSMERR_CSSM_BUFFER_TOO_SMALL);
                }

                num++;

                /* advance to the next record */
                tempPtr = tempPtr->Next;
            }

            /*
             * Release the global reader lock.
             */
            cssm_ReaderUnlock(hEMMListSWMRLock);

           *NumberOfModuleManagers = num;
        }
        else
        {
            ERR(err = CSSMERR_CSSM_INTERNAL_ERROR);
        }
    }

    return err;
}
