/*-----------------------------------------------------------------------
 * File: INTCONST.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "precomp.h"



/*
 * cssm_GetFunctionTable
 *
 * Description: This function takes a function table (FuncNameAddrPtr) and
 *      fills the addresses with the API level function address corresponding
 *      to the string name of the function.  For standard CDSA functions, this
 *      will be the address of the CSSM API level function.  For EMM functions,
 *      this function will call the EMM to fill in the table.
 *
 * Parameters:
 *      FuncNameAddrPtr - the function table to be filled with function
 *                          addresses
 *      NumOfFuncNameAddr - the number of entries in FuncNameAddrPtr
 *      hAddIn - Attach handle for the function table.  This is used to
 *                  determine if we'll need to contact an EMM for filling the
 *                  table.
 *
 * Returns:
 *      CSSM_OK - Everything functioned correctly.
 *      CSSMERR_CSSM_INVALID_SERVICE_MASK - The service type is unknown at this
 *                                          time.
 *      Other error codes come from other functions and may or may not be
 *      fatal.
 */

CSSM_RETURN
cssm_GetFunctionTable(CSSM_FUNC_NAME_ADDR_PTR FuncNameAddrPtr, 
                      uint32 NumOfFuncNameAddr, 
                      CSSM_HANDLE hAddIn)
{
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    uint32 i;
    CSSM_RETURN Ret = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_cssm_GetFunctionTable);

    ERR(Ret = cssm_FindAttachAndMultiLock(hAddIn,
                                          CSSM_NO_LOCK,   /* global lock */
                                          NULL,         /* module record */
                                          CSSM_NO_LOCK,   /* module lock */
                                          NULL,        /* service record */
                                          CSSM_NO_LOCK, /* service lock  */
                                          &AttachRecord,
                                          CSSM_READER_LOCK));
    if(CSSM_OK == Ret)
    {
        /* see if this is an EMM or a regular service */
        if(!cssm_IsCSSMMask(AttachRecord->ServiceType))
        {
            if(NULL != AttachRecord->pEMMRecord)
            {
                ERR(Ret = cssm_ReaderLock(AttachRecord->pEMMRecord->hEMMRecordSWMRLock));

                if(CSSM_OK == Ret)
                {
                    ERR(Ret = AttachRecord->pEMMRecord->EMMJT->RefreshFunctionTable(FuncNameAddrPtr,
                                                                                    NumOfFuncNameAddr));
                    cssm_ReaderUnlock( AttachRecord->pEMMRecord->hEMMRecordSWMRLock);
                }
            }
            else
            {
                ERR(Ret = CSSMERR_CSSM_INVALID_SERVICE_MASK);
            }
        }
        else
        {
            /* since this is a regular addin, we can just give the function
               pointers from this library
            */
            for(i = 0; i < NumOfFuncNameAddr; i++)
            {
#ifdef ISL_INCLUDED
                ISL_CONST_DATA FuncName;
                if(NULL != FuncNameAddrPtr[i].Name)
                {
                    FuncName.Data = (const uint8*)FuncNameAddrPtr[i].Name;
                    FuncName.Length = strlen((char*)FuncNameAddrPtr[i].Name);

                    FuncNameAddrPtr[i].Address = (CSSM_PROC_ADDR)EISL_LocateProcedureAddress(*(cssm30_Credentials),
                                                                                             FuncName);
                }
                else
                {
                    FuncNameAddrPtr[i].Address = NULL;
                }
#else
                ERR(port_GetProcAddress(cssm30_Library,
                                        FuncNameAddrPtr[i].Name,
                                        &(FuncNameAddrPtr[i].Address)));
#endif
            }
        }

        cssm_ReleaseAttachLock(AttachRecord,
                               CSSM_READER_LOCK);
    }

    return Ret;
}
