/*-----------------------------------------------------------------------
 * File: MODINSTALL.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include <sys/types.h>
#include <sys/stat.h>
#ifdef VMS
#include <string.h>
#endif

#include "cssm.h"
#include "cssmerr.h"
#include "cssmport.h"
#include "cssmlock.h"
#include "cssmtype.h"
#include "maf_collectn.h"
#include "maf_util.h"
#include "guids.h"
#include "mds.h"
#include "mds_schema.h"
#include "mds_util.h"
#ifdef VMS
#include "installdefs.h"
/* special case the installation pgm name for cssm build*/
#ifdef INSTALL_ENTRY_NAME
#undef INSTALL_ENTRY_NAME
#undef INSTALL_ENTRY_NAME_STRING
#define INSTALL_ENTRY_NAME         RegisterCDSAModuleCSSM
#define INSTALL_ENTRY_NAME_STRING "RegisterCDSAModuleCSSM"
#endif
#else
#include "intel/installdefs.h"
#endif



/* Number of MDS update operations that will be remembered by mds_util for
 * error clean-up during installation.
 */
#define CSSM_MDS_COMMIT_SIZE  150

const CSSM_GUID CORE_GUID = intel_cssm_guid_init;
const CSSM_API_MEMORY_FUNCS MemoryFuncs = {internal_malloc, 
                                           internal_free,
                                           internal_realloc,
                                           internal_calloc,
                                           NULL};

static CSSM_RETURN __CleanMDSCdsadir( MDSU_CONTEXT *pContext )
{
    CSSM_RETURN rv = CSSM_OK;
    MDSU_CDSA_CSSM_SCHEMA CssmTemplate;
    MDSU_CDSA_CSSM_SCHEMA CssmData;

    /* Execute a search on the CSSM record type to get a handle of all
     * records and then delete them.
     */
    /* CSSM Schema */
    CssmTemplate.ModuleID = CORE_GUID;
    rv = MDSU_FindFirst( pContext,
                         &IfiMdsuCdsadirCSSMSchema,
                         &CssmTemplate,
                         MDSU_CSSM_MODULE_ID,
                         &CssmData,
                         NULL );
    while ( rv == CSSM_OK )
    {
        rv = MDSU_FindNext( pContext,
                            &CssmData,
                            NULL );
    }
    if ( rv == CSSMERR_DL_ENDOFDATA )
    {
        /* Flush everything from the MDS */
        rv = MDSU_Revert( pContext );
    }

    /* Make sure nothing gets removed if there is an error */
    if ( ( rv != CSSMERR_DL_ENDOFDATA ) &&
         ( rv != CSSM_OK ) ) 
    {
        MDSU_Commit( pContext );
    }

    return rv;
}

static CSSM_RETURN __CleanMDSObjectdir( MDSU_CONTEXT *pContext )
{
    CSSM_RETURN rv = CSSM_OK;
    MDSU_OBJECT_SCHEMA CssmObjectTemplate;
    MDSU_OBJECT_SCHEMA CssmObjectData;

    /* Execute a search on the CSSM record type to get a handle of all
     * records and then delete them.
     */
    /* Object Schema */
    CssmObjectTemplate.ModuleID = CORE_GUID;
    rv = MDSU_FindFirst( pContext,
                         &IfiMdsuObjectdirSchema,
                         &CssmObjectTemplate,
                         MDSU_OBJ_MODULE_ID,
                         &CssmObjectData,
                         NULL );
    while ( rv == CSSM_OK )
    {
        if ( CssmObjectData.Manifest.Data )
        {
            MDSU_free( pContext, CssmObjectData.Manifest.Data );
        }

        rv = MDSU_FindNext( pContext,
                            &CssmObjectData,
                            NULL );
    }
    if ( rv == CSSMERR_DL_ENDOFDATA )
    {
        /* Flush everything from the MDS */
        rv = MDSU_Revert( pContext );
    }
    /* Make sure nothing gets removed if there is an error */
    if ( ( rv != CSSMERR_DL_ENDOFDATA ) &&
         ( rv != CSSM_OK ) ) 
    {
        MDSU_Commit( pContext );
    }

    return rv;
}



CSSM_RETURN CSSMAPI INSTALL_ENTRY_NAME(
                        const char *szModuleName,
                        const char *szSearchPath,
                        CSSM_DATA *pCredential,
                        INSTALL_ACTION Action,
                        INSTALL_ERROR *pError )
{
    CSSM_RETURN rv = CSSM_OK;
    MDSU_CONTEXT MDSUContext;
    MDSU_CDSA_CSSM_SCHEMA CssmTemplate;
    MDSU_OBJECT_SCHEMA CssmObjectTemplate;

    /* Initialize the MDS update */
    if ( ( rv = MDSU_Init( &MDSUContext, 
                           &MemoryFuncs,
                           &CORE_GUID,
                           MDS_CDSA_DIRECTORY,
                           ( CSSM_DB_ACCESS_READ | 
                             CSSM_DB_ACCESS_WRITE ),
                           CSSM_MDS_COMMIT_SIZE ) ) != CSSM_OK )
    {
        sprintf( pError->ErrorString, "MDS Error (Init): %X\n", rv );
        pError->ErrorCode = rv;
        return rv;
    }

    /* Place code to remove existing entries here */
    if ( ( rv = __CleanMDSCdsadir( &MDSUContext ) ) != CSSM_OK )
    {
        sprintf( pError->ErrorString, "MDS Error (Clean): %X\n", rv );
        pError->ErrorCode = rv;
        MDSU_Term( &MDSUContext );
        //return rv;
    }

    if (( Action == INSTALL_ACTION_INSTALL) || ( Action == INSTALL_ACTION_REFRESH ))
    {
        /* Fill in the cssm schema structure */
        memcpy( &CssmTemplate.ModuleID, &CORE_GUID, sizeof(CSSM_GUID) );
        CssmTemplate.CDSAVersion.Major = CSSM_MAJOR;
        CssmTemplate.CDSAVersion.Minor = CSSM_MINOR;
#if defined (__digital__) || defined (VMS)
        strcpy( CssmTemplate.CssmVendor, "Compaq Computer Corporation" );
#else
        strcpy( CssmTemplate.CssmVendor, "Intel Corporation" );
#endif
        strcpy( CssmTemplate.Desc, "CSSM Module" );
        CssmTemplate.NativeServices = CSSM_SERVICE_CSSM | CSSM_SERVICE_CL | 
            CSSM_SERVICE_TP   | CSSM_SERVICE_DL |
            CSSM_SERVICE_CSP  | CSSM_SERVICE_AC;
        
        /* Insert the CSSM schema record into the MDS */
        rv = MDSU_UpdateSchema( &MDSUContext,
            &IfiMdsuCdsadirCSSMSchema,
            &CssmTemplate,
            NULL );
        if ( rv != CSSM_OK )
        {
            sprintf( pError->ErrorString, "MDS Error (CSSM): %X\n", rv );
            pError->ErrorCode = rv;
            MDSU_Term( &MDSUContext );
            return rv;
        }
        
        /* Commit the records */
        MDSU_Commit( &MDSUContext );
    }

    /* Shut down the MDS connection */
    MDSU_Term( &MDSUContext );

    /* Initialize the MDS update */
    if ( ( rv = MDSU_Init( &MDSUContext, 
                           &MemoryFuncs,
                           &CORE_GUID,
                           MDS_OBJECT_DIRECTORY,
                           ( CSSM_DB_ACCESS_READ | 
                             CSSM_DB_ACCESS_WRITE ),
                           CSSM_MDS_COMMIT_SIZE ) ) != CSSM_OK )
    {
        sprintf( pError->ErrorString, "MDS Error (Init): %X\n", rv );
        pError->ErrorCode = rv;
        return rv;
    }

    /* Place code to remove existing entries here */
    if ( ( rv = __CleanMDSObjectdir( &MDSUContext ) ) != CSSM_OK )
    {
        sprintf( pError->ErrorString, "MDS Error (Clean): %X\n", rv );
        pError->ErrorCode = rv;
        MDSU_Term( &MDSUContext );
    }
    if ( Action == INSTALL_ACTION_UNINSTALL )
    {
        MDSU_Term( &MDSUContext );
        return rv;
    }

    /* Fill in the cssm object schema structure */
    memcpy( &CssmObjectTemplate.ModuleID, &CORE_GUID, sizeof(CSSM_GUID) );
    CssmObjectTemplate.Manifest = *pCredential;
    strcpy( CssmObjectTemplate.ModuleName, szModuleName );
    strcpy( CssmObjectTemplate.Path, szSearchPath );
    CssmObjectTemplate.ProductVersion.Major = 3;
    CssmObjectTemplate.ProductVersion.Minor = 1;

    /* Insert the CSSM schema record into the MDS */
    rv = MDSU_UpdateSchema( &MDSUContext,
                            &IfiMdsuObjectdirSchema,
                            &CssmObjectTemplate,
                            NULL );
    if ( rv != CSSM_OK )
    {
        sprintf( pError->ErrorString, "MDS Error (CSSM ObjectDir): %X\n", rv );
        pError->ErrorCode = rv;
        MDSU_Term( &MDSUContext );
        return rv;
    }

    /* Commit the records */
    MDSU_Commit( &MDSUContext );

    /* Shut down the MDS connection */
    MDSU_Term( &MDSUContext );

    return CSSM_OK;
}
