/*-----------------------------------------------------------------------
 * File: CDSACOMMON.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */


#include <assert.h>
#include <string.h>

#include "cssmtype.h"
#include "cssmerr.h"
#ifdef VMS
#include "cssmport.h"
#else
#include "port/cssmport.h"
#endif
#include "mds_schema.h"

#include "mds.h"
#include "mds_util.h"

#include "convert.h"
#include "internal.h"

#define __ModuleID          ( 0 )
#define __Manifest          ( 1 )
#define __ModuleName        ( 2 )
#define __Path              ( 3 )
#define __CdsaVersion       ( 4 )
#define __Desc              ( 5 )
#define __DynamicFlag       ( 6 )
#define __MultiThreadFlag   ( 7 )
#define __ServiceMask       ( 8 )

typedef struct _mdsu_cdsa_common_attr_state
{
    CSSM_DB_ATTRIBUTE_DATA DbRecordAttributes[MDS_CDSADIR_COMMON_NUM_ATTRIBUTES];
    CSSM_DATA AttributeValues[MDS_CDSADIR_COMMON_NUM_ATTRIBUTES];
    uint8 PrintableGuid[_MDSU_PRINTABLE_GUID_LENGTH];
    uint8 PrintableVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
}
_MDSU_CDSA_COMMON_ATTR_STATE;

static CSSM_RETURN CSSMAPI __MDSU_ConstructCDSACommonAttributes(
                            MDSU_CONTEXT *pContext,
                            const void *pSchemaData,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            CSSM_BOOL bFillValues,
                            void **ppAttrState )
{
    CSSM_RETURN rv = CSSM_OK;
    const MDSU_CDSA_COMMON_SCHEMA *pCommon = 
                            (const MDSU_CDSA_COMMON_SCHEMA*)pSchemaData;
    _MDSU_CDSA_COMMON_ATTR_STATE *pState = NULL;
    CSSM_DB_ATTRIBUTE_DATA *DbRecordAttributes;
    uint32 i;

    assert( pContext && DbRecord && ppAttrState );
    assert( ( bFillValues == CSSM_FALSE ) ||
            ( ( bFillValues == CSSM_TRUE ) && pSchemaData ) );

    /* Allocate the memory for the attributes */
    pState = MDSU_malloc( pContext, sizeof(*pState) );
    if ( pState == NULL )
    {
        ERR( rv = CSSMERR_CSSM_MEMORY_ERROR );
    }
    else
    {
        /* Clear the structure */
        memset( pState, 0, sizeof(*pState) );
        DbRecordAttributes = pState->DbRecordAttributes;

        /* Fill in the record information */
        DbRecord->DataRecordType = MDS_CDSADIR_COMMON_RECORDTYPE;
        DbRecord->SemanticInformation = 0;
        DbRecord->NumberOfAttributes = MDS_CDSADIR_COMMON_NUM_ATTRIBUTES;
        DbRecord->AttributeData = DbRecordAttributes;

        /* Fill in attribute info */
        DbRecordAttributes[__ModuleID].Info = s_MdsAttrInfo_ModuleId;
        DbRecordAttributes[__Manifest].Info = s_MdsAttrInfo_Manifest;
        DbRecordAttributes[__ModuleName].Info = s_MdsAttrInfo_ModuleName;
        DbRecordAttributes[__Path].Info = s_MdsAttrInfo_Path;
        DbRecordAttributes[__CdsaVersion].Info = s_MdsAttrInfo_CDSAVersion;
        DbRecordAttributes[__Desc].Info = s_MdsAttrInfo_Desc;
        DbRecordAttributes[__DynamicFlag].Info = s_MdsAttrInfo_DynamicFlag;
        DbRecordAttributes[__MultiThreadFlag].Info = 
                                                s_MdsAttrInfo_MultiThreadFlag;
        DbRecordAttributes[__ServiceMask].Info = 
                                                s_MdsAttrInfo_ServiceMaskFlag;

        if ( bFillValues == CSSM_TRUE )
        {
            /* Connect the attribute value array to the attribute array */
            for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
            {
                DbRecordAttributes[i].Value = &pState->AttributeValues[i];
                DbRecordAttributes[i].NumberOfValues = 1;
            }

            /* Fill in the record contents */
            /* ModuleID */
            _MDSU_GetPrintableGUID( &pCommon->ModuleID,
                                    (char*)pState->PrintableGuid );
            DbRecordAttributes[__ModuleID].Value->Data = 
                                                    pState->PrintableGuid;
            DbRecordAttributes[__ModuleID].Value->Length = 
                                strlen( (char*)pState->PrintableGuid ) + 1;

            /* Manifest */
            (*DbRecordAttributes[__Manifest].Value) = pCommon->Manifest;

            /* ModuleName */
            DbRecordAttributes[__ModuleName].Value->Length = 
                                        strlen( pCommon->ModuleName ) + 1 ;
            DbRecordAttributes[__ModuleName].Value->Data =
                                                (uint8*)pCommon->ModuleName;

            /* Path */
            DbRecordAttributes[__Path].Value->Length = 
                                        strlen( pCommon->Path ) + 1 ;
            DbRecordAttributes[__Path].Value->Data =
                                                (uint8*)pCommon->Path;

            /* Version */
            _MDSU_GetPrintableVersion( &pCommon->CDSAVersion,
                                       (char*)pState->PrintableVersion );
            DbRecordAttributes[__CdsaVersion].Value->Data = 
                                                    pState->PrintableVersion;
            DbRecordAttributes[__CdsaVersion].Value->Length = 
                                strlen( (char*)pState->PrintableVersion ) + 1;

            /* Desc */
            DbRecordAttributes[__Desc].Value->Length = 
                                        strlen( pCommon->Desc ) + 1 ;
            DbRecordAttributes[__Desc].Value->Data =
                                                (uint8*)pCommon->Desc;

            /* DynamicFlag */
            DbRecordAttributes[__DynamicFlag].Value->Length = sizeof(uint32);
            DbRecordAttributes[__DynamicFlag].Value->Data = 
                                                    (uint8*)&pCommon->DynamicFlag;

            /* MultiThreadFlag */
            DbRecordAttributes[__MultiThreadFlag].Value->Length = sizeof(uint32);
            DbRecordAttributes[__MultiThreadFlag].Value->Data =
                                                (uint8*)&pCommon->MultiThreadFlag;

            /* ServiceMask */
            DbRecordAttributes[__ServiceMask].Value->Length = sizeof(uint32);
            DbRecordAttributes[__ServiceMask].Value->Data =
                                                    (uint8*)&pCommon->ServiceMask;

        } /* bFillValues */
        else
        {
            for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
            {
                DbRecordAttributes[i].Value = NULL;
                DbRecordAttributes[i].NumberOfValues = 0;
            }
        }

        if ( rv != CSSM_OK )
        {
            /* Deallocate memory if there was an error */
            MDSU_free( pContext,
                       pState );
            pState = NULL;
        }
    } /* allocate DbRecordAttributes */

    *ppAttrState = (void*)pState;

    return rv;
}

static CSSM_RETURN CSSMAPI __MDSU_DestructCDSACommonAttributes(
                            MDSU_CONTEXT *pContext,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            CSSM_BOOL bFreeAll,
                            void *pAttrState )
{
    CSSM_DB_ATTRIBUTE_DATA *DbRecordAttributes;
    uint32 i;

    assert( pContext && DbRecord && pAttrState );

    /* Free the buffers allocated for the attribute values */
    DbRecordAttributes = DbRecord->AttributeData;
    if ( bFreeAll )
    {
        for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
        {
            if ( DbRecordAttributes[i].Value )
            {
                if ( DbRecordAttributes[i].Value->Data )
                {
                    MDSU_free( pContext,
                               DbRecordAttributes[i].Value->Data );
                }
                MDSU_free( pContext,
                           DbRecordAttributes[i].Value );
            }
        }
    }

    /* Free the state structure */
    MDSU_free( pContext,
               pAttrState );
    DbRecord->AttributeData = NULL;

    return CSSM_OK;
}

typedef struct _mdsu_cdsa_common_pred_state
{
    CSSM_SELECTION_PREDICATE Predicates[MDS_CDSADIR_COMMON_NUM_ATTRIBUTES];
    CSSM_DATA PredicateValues[MDS_CDSADIR_COMMON_NUM_ATTRIBUTES];
    uint8 PrintableGuid[_MDSU_PRINTABLE_GUID_LENGTH];
    uint8 PrintableVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
}
_MDSU_CDSA_COMMON_PRED_STATE;

static CSSM_RETURN CSSMAPI __MDSU_ConstructCDSACommonPredicate(
                            MDSU_CONTEXT *pContext,
                            const void *pSchemaTemplate,
                            MDSU_TEMPLATE_MASK ValidTemplates,
                            CSSM_QUERY *pQuery,
                            void **ppPredState )
{
    CSSM_RETURN rv = CSSM_OK;
    const MDSU_CDSA_COMMON_SCHEMA *pCommon = 
                            (const MDSU_CDSA_COMMON_SCHEMA*)pSchemaTemplate;
    _MDSU_CDSA_COMMON_PRED_STATE *pState = NULL;
    CSSM_SELECTION_PREDICATE *pPredicates = NULL;
    uint32 uNumPredicates;
    uint32 uIndex = 0;
    uint32 i;

    assert( pContext && 
            pQuery &&
            ppPredState &&
            ( ( pSchemaTemplate && ValidTemplates ) ||
              ( !pSchemaTemplate && !ValidTemplates ) ) );

    /* Count the predicates */
    /* Clear the manifest bit. Searching on manifest is not supported */
    ValidTemplates &= (~MDSU_COM_MANIFEST);
    _MDSU_CountBits( ValidTemplates, uNumPredicates );

    /* Allocate the predicate state and link the value structures */
    if ( uNumPredicates != 0 )
    {
        pState = MDSU_malloc( pContext, sizeof(*pState) );
        if ( pState == NULL )
        {
            ERR( rv = CSSMERR_CSSM_MEMORY_ERROR );
        }
        else
        {
            memset( pState,
                    0,
                    sizeof(*pState) );
            pPredicates = pState->Predicates;

            /* Allocate a single attribute value struct per attribute */
            for ( i = 0; i < uNumPredicates; i++ )
            {
                pPredicates[i].Attribute.Value =
                                            &pState->PredicateValues[i];
                pPredicates[i].Attribute.NumberOfValues = 1;
            }
        } /* Allocate and link predicate values */
    } /* uNumPredicates != 0 */

    if ( rv == CSSM_OK )
    {
        /* Fill in the query parameters */
        pQuery->RecordType = MDS_CDSADIR_COMMON_RECORDTYPE;
        pQuery->Conjunctive = CSSM_DB_AND;
        pQuery->NumSelectionPredicates = uNumPredicates;
        pQuery->QueryLimits.TimeLimit = CSSM_QUERY_TIMELIMIT_NONE;
        pQuery->QueryLimits.SizeLimit = CSSM_QUERY_SIZELIMIT_NONE;
        pQuery->QueryFlags = 0;
        pQuery->SelectionPredicate = pPredicates;
        for ( uIndex = 0; uIndex < uNumPredicates; uIndex++ )
        {
            pPredicates[uIndex].DbOperator = CSSM_DB_EQUAL;
        }

        /* Fill in the predicate data */
        uIndex = 0;
        if ( ValidTemplates & MDSU_COM_MODULE_ID )
        {
            /* ModuleID */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ModuleId;
            _MDSU_GetPrintableGUID( &pCommon->ModuleID,
                                    (char*)pState->PrintableGuid );
            pPredicates[uIndex].Attribute.Value->Data = pState->PrintableGuid;
            pPredicates[uIndex].Attribute.Value->Length = 
                                strlen( (char*)pState->PrintableGuid ) + 1;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_COM_MODULE_NAME )
        {
            /* ModuleName */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ModuleName;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pCommon->ModuleName ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pCommon->ModuleName;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_COM_PATH )
        {
            /* Path */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_Path;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pCommon->Path ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pCommon->Path;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_COM_CDSA_VERSION )
        {
            /* Version */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_Version;
            _MDSU_GetPrintableVersion( &pCommon->CDSAVersion,
                                       (char*)pState->PrintableVersion );
            pPredicates[uIndex].Attribute.Value->Data =
                                                pState->PrintableVersion;
            pPredicates[uIndex].Attribute.Value->Length = 
                            strlen( (char*)pState->PrintableVersion ) + 1;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_COM_DESC )
        {
            /* Desc */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_Desc;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pCommon->Desc ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pCommon->Desc;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_COM_DYNAMIC_FLAG )
        {
            /* DynamicFlag */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_DynamicFlag;
            pPredicates[uIndex].Attribute.Value->Length = sizeof(uint32);
            pPredicates[uIndex].Attribute.Value->Data = 
                                                    (uint8*)&pCommon->DynamicFlag;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_COM_MULTI_THREAD_FLAG )
        {
            /* MultiThreadFlag */
            pPredicates[uIndex].Attribute.Info = 
                                                s_MdsAttrInfo_MultiThreadFlag;
            pPredicates[uIndex].Attribute.Value->Length = sizeof(uint32);
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)&pCommon->MultiThreadFlag;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_COM_SERVICE_MASK )
        {
            /* ServiceMask */
            pPredicates[uIndex].Attribute.Info = 
                                                s_MdsAttrInfo_ServiceMaskFlag;
            pPredicates[uIndex].Attribute.Value->Length = sizeof(uint32);
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)&pCommon->ServiceMask;
            uIndex++;
        }

    } /* Initial allocation successful */

    *ppPredState = (void*)pState;

    return rv;
}

static CSSM_RETURN CSSMAPI __MDSU_DestructCDSACommonPredicate(
                            MDSU_CONTEXT *pContext,
                            CSSM_QUERY *pQuery,
                            void *pPredState )
{
    assert( pContext && pQuery );

    /* Free the attribute array */
    if ( pPredState != NULL )
    {
        MDSU_free( pContext, pPredState );
    }

    return CSSM_OK;
}

static CSSM_RETURN CSSMAPI __MDSU_ConvertCDSACommonAttributes(
                            MDSU_CONTEXT *pContext,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            void *pSchemaData )
{
    CSSM_RETURN crReturn;
    MDSU_CDSA_COMMON_SCHEMA *pCommon = (MDSU_CDSA_COMMON_SCHEMA*)pSchemaData;

    const CSSM_DB_ATTRIBUTE_DATA *DbAttr = NULL;

    assert( pSchemaData && DbRecord && pContext );

    DbAttr = DbRecord->AttributeData;

    /* ModuleID */
    crReturn = _MDSU_GetStructuredGUID( 
                                (char*)DbAttr[__ModuleID].Value->Data,
                                &pCommon->ModuleID );
    if ( crReturn != CSSM_OK )
    {
        return crReturn;
    }

    /* Manifest */
    if ( DbAttr[__Manifest].Value->Data != NULL )
    {
        pCommon->Manifest.Data = MDSU_malloc( 
                                        pContext,
                                        DbAttr[__Manifest].Value->Length );
        if ( pCommon->Manifest.Data == NULL )
        {
            return CSSMERR_CSSM_MEMORY_ERROR;
        }
        pCommon->Manifest.Length = DbAttr[__Manifest].Value->Length;
        memcpy( pCommon->Manifest.Data,
                DbAttr[__Manifest].Value->Data,
                DbAttr[__Manifest].Value->Length );
    }
    else
    {
        pCommon->Manifest.Data = NULL;
        pCommon->Manifest.Length = 0;
    }
    
    /* ModuleName */
	if ( DbAttr[__ModuleName].Value->Data != NULL )
        strncpy( pCommon->ModuleName,
                 (char*)DbAttr[__ModuleName].Value->Data,
				 DbAttr[__ModuleName].Value->Length );
	else
		strcpy(pCommon->ModuleName, "" );

    /* Path */
	if ( DbAttr[__Path].Value->Data != NULL )
        strncpy( pCommon->Path,
                 (char*)DbAttr[__Path].Value->Data,
				 DbAttr[__Path].Value->Length );
	else
		strcpy(pCommon->Path, "" );

    /* Version */
    crReturn = _MDSU_GetStructuredVersion( 
                                    (char*)DbAttr[__CdsaVersion].Value->Data,
                                    &pCommon->CDSAVersion );
    if ( crReturn != CSSM_OK )
    {
        MDSU_free( pContext, pCommon->Manifest.Data );
        pCommon->Manifest.Data = NULL;
    }

    /* Desc */
	if ( DbAttr[__Desc].Value->Data != NULL )
        strncpy( pCommon->Desc,
                 (char*)DbAttr[__Desc].Value->Data,
				 DbAttr[__Desc].Value->Length );
	else
		strcpy( pCommon->Desc, "" );

    /* DynamicFlag */
    pCommon->DynamicFlag = 
        ( *(uint32*)DbAttr[__DynamicFlag].Value->Data != 0 ) ? 
            CSSM_TRUE : CSSM_FALSE;

    /* MultiThreadFlag */
    pCommon->MultiThreadFlag = 
        ( *(uint32*)DbAttr[__MultiThreadFlag].Value->Data != 0 ) ? 
            CSSM_TRUE : CSSM_FALSE;

    /* ServiceMask */
    pCommon->ServiceMask = 
                    *(CSSM_SERVICE_TYPE*)DbAttr[__ServiceMask].Value->Data;

    return CSSM_OK;
}

const IfMdsuSchemaManipulator IfiMdsuCdsadirCommonSchema =
{
    __MDSU_ConstructCDSACommonAttributes,
    __MDSU_DestructCDSACommonAttributes,
    __MDSU_ConstructCDSACommonPredicate,
    __MDSU_DestructCDSACommonPredicate,
    __MDSU_ConvertCDSACommonAttributes,
    MDS_CDSADIR_COMMON_RECORDTYPE
};
