/*-----------------------------------------------------------------------
 * File: CDSACSPCAPABILITY.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */


#include <assert.h>
#include <string.h>
#ifdef VMS
#include "cssmport.h"
#else
#include "port/cssmport.h"
#endif
#include "cssmtype.h"
#include "cssmerr.h"
#include "mds_schema.h"

#include "mds.h"
#include "mds_util.h"
#include "convert.h"
#include "internal.h"

#define __ModuleID          ( 0 )
#define __SSID              ( 1 )
#define __UseeTag           ( 2 )
#define __GroupId           ( 3 )
#define __ContextType       ( 4 )
#define __AlgType           ( 5 )
#define __AttributeType     ( 6 )
#define __AttributeValue    ( 7 )
#define __Description       ( 8 )

typedef struct _mdsu_cdsa_csp_capability_attr_state
{
    CSSM_DB_ATTRIBUTE_DATA DbRecordAttributes[MDS_CDSADIR_CSP_CAPABILITY_NUM_ATTRIBUTES];
    CSSM_DATA AttributeValues[MDS_CDSADIR_CSP_CAPABILITY_NUM_ATTRIBUTES];
    uint8 PrintableGuid[_MDSU_PRINTABLE_GUID_LENGTH];
}
_MDSU_CDSA_CSP_CAPABILITY_ATTR_STATE;

static CSSM_RETURN CSSMAPI __MDSU_ConstructCDSACSPCapabilityAttributes(
                            MDSU_CONTEXT *pContext,
                            const void *pSchemaData,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            CSSM_BOOL bFillValues,
                            void **ppAttrState )
{
    CSSM_RETURN rv = CSSM_OK;
    const MDSU_CDSA_CSP_CAPABILITY_SCHEMA *pCapability = 
                        (const MDSU_CDSA_CSP_CAPABILITY_SCHEMA*)pSchemaData;
    _MDSU_CDSA_CSP_CAPABILITY_ATTR_STATE *pState = NULL;
    CSSM_DB_ATTRIBUTE_DATA *DbRecordAttributes;
    uint32 i;

    assert( pContext && ppAttrState && DbRecord );
    assert( ( bFillValues == CSSM_FALSE ) ||
            ( ( bFillValues == CSSM_TRUE ) && pSchemaData ) );

    /* Allocate the memory for the attributes */
    pState = MDSU_malloc( pContext,
                          sizeof(_MDSU_CDSA_CSP_CAPABILITY_ATTR_STATE) );
    if ( pState == NULL )
    {
        ERR( rv = CSSMERR_CSSM_MEMORY_ERROR );
    }
    else
    {
        /* Clear the structure */
        memset( pState, 
                0,
                sizeof(*pState) );
        DbRecordAttributes = pState->DbRecordAttributes;

        /* Fill in the record information */
        DbRecord->DataRecordType = MDS_CDSADIR_CSP_CAPABILITY_RECORDTYPE;
        DbRecord->SemanticInformation = 0;
        DbRecord->NumberOfAttributes = MDS_CDSADIR_CSP_CAPABILITY_NUM_ATTRIBUTES;
        DbRecord->AttributeData = DbRecordAttributes;

        /* Fill in the attribute info */
        DbRecordAttributes[__ModuleID].Info = s_MdsAttrInfo_ModuleId;
        DbRecordAttributes[__SSID].Info = s_MdsAttrInfo_SSID;
        DbRecordAttributes[__UseeTag].Info = s_MdsAttrInfo_UseeTag;
        DbRecordAttributes[__GroupId].Info = s_MdsAttrInfo_GroupId;
        DbRecordAttributes[__ContextType].Info = s_MdsAttrInfo_ContextType;
        DbRecordAttributes[__AlgType].Info = s_MdsAttrInfo_AlgType;
        DbRecordAttributes[__AttributeType].Info = s_MdsAttrInfo_AttributeType;
        DbRecordAttributes[__AttributeValue].Info = s_MdsAttrInfo_AttributeValue;
        DbRecordAttributes[__Description].Info = s_MdsAttrInfo_Description;

        if ( bFillValues == CSSM_TRUE )
        {
            /* Connect the attribute value array to the attribute array */
            for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
            {
                DbRecordAttributes[i].Value = &pState->AttributeValues[i];
                DbRecordAttributes[i].NumberOfValues = 1;
            }

            /* Fill in the record contents */
            /* ModuleID */
            _MDSU_GetPrintableGUID( &pCapability->ModuleID,
                                    (char*)pState->PrintableGuid );
            DbRecordAttributes[__ModuleID].Value->Data =
                                                    pState->PrintableGuid;
            DbRecordAttributes[__ModuleID].Value->Length = 
                                strlen( (char*)pState->PrintableGuid ) + 1;

            /* SSID */
            DbRecordAttributes[__SSID].Value->Length = sizeof(uint32);
            DbRecordAttributes[__SSID].Value->Data = (uint8*)&pCapability->SSID;

            /* UseeTag */
            DbRecordAttributes[__UseeTag].Value->Length = sizeof(uint32);
            DbRecordAttributes[__UseeTag].Value->Data = (uint8*)&pCapability->UseeTag;

            /* GroupId */
            DbRecordAttributes[__GroupId].Value->Length = sizeof(uint32);
            DbRecordAttributes[__GroupId].Value->Data = (uint8*)&pCapability->GroupId;

            /* ContextType */
            DbRecordAttributes[__ContextType].Value->Length = sizeof(uint32);
            DbRecordAttributes[__ContextType].Value->Data =
                                                (uint8*)&pCapability->ContextType;

            /* AlgType */
            DbRecordAttributes[__AlgType].Value->Length = sizeof(uint32);
            DbRecordAttributes[__AlgType].Value->Data =
                                                (uint8*)&pCapability->AlgType;

            /* AttributeType */
            DbRecordAttributes[__AttributeType].Value->Length = sizeof(uint32);
            DbRecordAttributes[__AttributeType].Value->Data =
                                                (uint8*)&pCapability->AttributeType;

            /* AttributeValue */
            DbRecordAttributes[__AttributeValue].Value->Length = 
                                pCapability->NumAttributeValues * sizeof(uint32);
            DbRecordAttributes[__AttributeValue].Value->Data =
                                            (uint8*)pCapability->AttributeValue;

            /* Description */
            DbRecordAttributes[__Description].Value->Length = 
                                        strlen( pCapability->Description ) + 1;
            DbRecordAttributes[__Description].Value->Data = 
                                            (uint8*)pCapability->Description;
                
        } /* bFillValues */
        else
        {
            for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
            {
                DbRecordAttributes[i].Value = NULL;
                DbRecordAttributes[i].NumberOfValues = 0;
            }
        }

        if ( rv != CSSM_OK )
        {
            /* Deallocate memory if there was an error */
            MDSU_free( pContext,
                       pState );
            pState = NULL;
        }
    } /* attribute array allocated successfully */

    *ppAttrState = (void*)pState;

    return CSSM_OK;
}

static CSSM_RETURN CSSMAPI __MDSU_DestructCDSACSPCapabilityAttributes(
                            MDSU_CONTEXT *pContext,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            CSSM_BOOL bFreeAll,
                            void *pAttrState )
{
    CSSM_DB_ATTRIBUTE_DATA *DbRecordAttributes;
    uint32 i;

    assert( pContext && DbRecord && pAttrState );

    /* Free the buffers allocated for the attribute values */
    DbRecordAttributes = DbRecord->AttributeData;
    if ( bFreeAll == CSSM_TRUE )
    {
        for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
        {
            if ( DbRecordAttributes[i].Value )
            {
                if ( DbRecordAttributes[i].Value->Data )
                {
                    MDSU_free( pContext,
                               DbRecordAttributes[i].Value->Data );
                }
                MDSU_free( pContext,
                           DbRecordAttributes[i].Value );
            }
        }
    }

    /* Free the attribute array */
    MDSU_free( pContext, pAttrState );

    return CSSM_OK;
}

typedef struct _mdsu_cdsa_csp_capability_pred_state
{
    CSSM_SELECTION_PREDICATE Predicates[MDS_CDSADIR_CSP_CAPABILITY_NUM_ATTRIBUTES];
    CSSM_DATA PredicateValues[MDS_CDSADIR_CSP_CAPABILITY_NUM_ATTRIBUTES];
    uint8 PrintableGuid[_MDSU_PRINTABLE_GUID_LENGTH];
}
_MDSU_CDSA_CSP_CAPABILITY_PRED_STATE;

static CSSM_RETURN CSSMAPI __MDSU_ConstructCDSACSPCapabilityPredicate(
                            MDSU_CONTEXT *pContext,
                            const void *pSchemaTemplate,
                            MDSU_TEMPLATE_MASK ValidTemplates,
                            CSSM_QUERY *pQuery,
                            void **ppPredState )
{
    CSSM_RETURN rv = CSSM_OK;
    const MDSU_CDSA_CSP_CAPABILITY_SCHEMA *pCapability = 
                (const MDSU_CDSA_CSP_CAPABILITY_SCHEMA*)pSchemaTemplate;
    _MDSU_CDSA_CSP_CAPABILITY_PRED_STATE *pState = NULL;
    CSSM_SELECTION_PREDICATE *pPredicates = NULL;
    uint32 uNumPredicates;
    uint32 uIndex = 0;
    uint32 i;

    assert( pContext && 
            pQuery &&
            ppPredState &&
            ( ( pSchemaTemplate && ValidTemplates ) ||
              ( !pSchemaTemplate && !ValidTemplates ) ) );

    /* Count the predicates */
    _MDSU_CountBits( ValidTemplates, uNumPredicates );

    /* Allocate the predicate state and link the value structures */
    if ( uNumPredicates != 0 )
    {
        pState = MDSU_malloc( pContext, sizeof(_MDSU_CDSA_CSP_CAPABILITY_PRED_STATE) );
        if ( pState == NULL )
        {
            ERR( rv = CSSMERR_CSSM_MEMORY_ERROR );
        }
        else
        {
            memset( pState,
                    0,
                    sizeof(*pState) );
            pPredicates = pState->Predicates;

            /* Allocate a single attribute value struct per attribute */
            for ( i = 0; i < uNumPredicates; i++ )
            {
                pPredicates[i].Attribute.Value =
                                            &pState->PredicateValues[i];
                pPredicates[i].Attribute.NumberOfValues = 1;
            }
        } /* Allocate predicate values */
    } /* uNumPredicates != 0 */

    if ( rv == CSSM_OK )
    {
        /* Fill in the query parameters */
        pQuery->RecordType = MDS_CDSADIR_CSP_CAPABILITY_RECORDTYPE;
        pQuery->Conjunctive = CSSM_DB_AND;
        pQuery->NumSelectionPredicates = uNumPredicates;
        pQuery->QueryLimits.TimeLimit = CSSM_QUERY_TIMELIMIT_NONE;
        pQuery->QueryLimits.SizeLimit = CSSM_QUERY_SIZELIMIT_NONE;
        pQuery->QueryFlags = 0;
        pQuery->SelectionPredicate = pPredicates;
        for ( uIndex = 0; uIndex < uNumPredicates; uIndex++ )
        {
            pPredicates[uIndex].DbOperator = CSSM_DB_EQUAL;
        }

        /* Fill in the predicate data */
        uIndex = 0;
        if ( ValidTemplates & MDSU_CSPCAP_MODULE_ID )
        {
            /* ModuleID */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ModuleId;
            _MDSU_GetPrintableGUID( &pCapability->ModuleID,
                                    (char*)pState->PrintableGuid );
            pPredicates[uIndex].Attribute.Value->Data =
                                                    pState->PrintableGuid;
            pPredicates[uIndex].Attribute.Value->Length = 
                                strlen( (char*)pState->PrintableGuid ) + 1;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_CSPCAP_SSID )
        {
            /* SSID */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_SSID;
            pPredicates[uIndex].Attribute.Value->Data = 
                                               (uint8*)&pCapability->SSID;
            pPredicates[uIndex].Attribute.Value->Length = sizeof(uint32);
            uIndex++;
        }
        if ( ValidTemplates & MDSU_CSPCAP_USEE_TAG )
        {
            /* UseeTag */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_UseeTag;
            pPredicates[uIndex].Attribute.Value->Length = sizeof(uint32);
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)&pCapability->UseeTag;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_CSPCAP_GROUP_ID )
        {
            /* UseeTag */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_GroupId;
            pPredicates[uIndex].Attribute.Value->Length = sizeof(uint32);
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)&pCapability->GroupId;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_CSPCAP_CONTEXT_TYPE )
        {
            /* ContextType */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ContextType;
            pPredicates[uIndex].Attribute.Value->Length = sizeof(uint32);
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)&pCapability->ContextType;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_CSPCAP_ALG_TYPE )
        {
            /* AlgType */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_AlgType;
            pPredicates[uIndex].Attribute.Value->Length = sizeof(uint32);
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)&pCapability->AlgType;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_CSPCAP_ATTRIBUTE_TYPE )
        {
            /* AttributeType */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_AttributeType;
            pPredicates[uIndex].Attribute.Value->Length = sizeof(uint32);
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)&pCapability->AttributeType;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_CSPCAP_ATTRIBUTE_VALUE )
        {
            /* AttributeValue */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_AttributeValue;
            pPredicates[uIndex].Attribute.Value->Length = 
                                pCapability->NumAttributeValues * sizeof(uint32);
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pCapability->AttributeValue;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_CSPCAP_DESCRIPTION )
        {
            /* Description */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_Description;
            pPredicates[uIndex].Attribute.Value->Length = 
                                            strlen( pCapability->Description ) + 1;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pCapability->Description;
            uIndex++;
        }

    } /* Initial allocation successful */
        
    *ppPredState = (void*)pState;

    return rv;
}

static CSSM_RETURN CSSMAPI __MDSU_DestructCDSACSPCapabilityPredicate(
                            MDSU_CONTEXT *pContext,
                            CSSM_QUERY *pQuery,
                            void *pPredState )
{
    assert( pContext && pQuery );

    /* Free the attribute array */
    if ( pPredState != NULL )
    {
        MDSU_free( pContext, pPredState );
    }

    return CSSM_OK;
}

static CSSM_RETURN CSSMAPI __MDSU_ConvertCDSACSPCapabilityAttributes(
                            MDSU_CONTEXT *pContext,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            void *pSchemaData )
{
    CSSM_RETURN crReturn;
    MDSU_CDSA_CSP_CAPABILITY_SCHEMA *pCapability = 
                            (MDSU_CDSA_CSP_CAPABILITY_SCHEMA*)pSchemaData;

    const CSSM_DB_ATTRIBUTE_DATA *DbAttr = NULL;

    assert( pSchemaData && DbRecord && pContext );

    DbAttr = DbRecord->AttributeData;

    /* ModuleID */
    crReturn = _MDSU_GetStructuredGUID( 
                                (char*)DbAttr[__ModuleID].Value->Data,
                                &pCapability->ModuleID );
    if ( crReturn != CSSM_OK )
    {
        return crReturn;
    }

    /* SSID */
    pCapability->SSID = *(uint32*)DbAttr[__SSID].Value->Data;
    
    /* UseeTag */
    pCapability->UseeTag = *(CSSM_USEE_TAG*)DbAttr[__UseeTag].Value->Data;

    /* GroupId */
    pCapability->GroupId = *(uint32*)DbAttr[__GroupId].Value->Data;

    /* ContextType */
    pCapability->ContextType = 
                    *(CSSM_CONTEXT_TYPE*)DbAttr[__ContextType].Value->Data;

    /* AlgType */
    pCapability->AlgType = *(CSSM_ALGORITHMS*)DbAttr[__AlgType].Value->Data;

    /* AttributeType */
    pCapability->AttributeType = 
                *(CSSM_ATTRIBUTE_TYPE*)DbAttr[__AttributeType].Value->Data;

    /* AttributeValue */
    memcpy( pCapability->AttributeValue,
            DbAttr[__AttributeValue].Value->Data,
            DbAttr[__AttributeValue].Value->Length );
    pCapability->NumAttributeValues = 
                    DbAttr[__AttributeValue].Value->Length / sizeof(uint32);

    /* Description */
    if ( DbAttr[__Description].Value->Data != NULL )
    {
        strncpy( pCapability->Description, 
                (char*)DbAttr[__Description].Value->Data,
				DbAttr[__Description].Value->Length );
    }
    else
    {
        strcpy( pCapability->Description, "" );
    }

    return CSSM_OK;
}

    /* MDS_CDSADIR_CSP_CAPABILITY_RECORDTYPE */
const IfMdsuSchemaManipulator IfiMdsuCdsadirCSPCapabilitySchema =
{
    __MDSU_ConstructCDSACSPCapabilityAttributes,
    __MDSU_DestructCDSACSPCapabilityAttributes,
    __MDSU_ConstructCDSACSPCapabilityPredicate,
    __MDSU_DestructCDSACSPCapabilityPredicate,
    __MDSU_ConvertCDSACSPCapabilityAttributes,
    MDS_CDSADIR_CSP_CAPABILITY_RECORDTYPE
};

