/*-----------------------------------------------------------------------
 * File: CDSACSPSCINFO.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include <assert.h>
#include <string.h>
#ifdef VMS
#include "cssmport.h"
#else
#include "port/cssmport.h"
#endif
#include "cssmtype.h"
#include "cssmerr.h"
#include "mds_schema.h"

#include "mds.h"
#include "mds_util.h"
#include "convert.h"
#include "internal.h"

#define __ModuleID          ( 0 )
#define __SSID              ( 1 )
#define __ScDesc            ( 2 )
#define __ScVendor          ( 3 )
#define __ScVersion         ( 4 )
#define __ScFirmwareVersion ( 5 )
#define __ScFlags           ( 6 )
#define __ScCustomFlags     ( 7 )
#define __ScSerialNumber    ( 8 )

typedef struct _mdsu_cdsa_csp_scinfo_attr_state
{
    CSSM_DB_ATTRIBUTE_DATA DbRecordAttributes[MDS_CDSADIR_CSP_SC_INFO_NUM_ATTRIBUTES];
    CSSM_DATA AttributeValues[MDS_CDSADIR_CSP_SC_INFO_NUM_ATTRIBUTES];
    uint8 PrintableGuid[_MDSU_PRINTABLE_GUID_LENGTH];
    uint8 PrintScVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
    uint8 PrintFirmwareVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
}
_MDSU_CDSA_CSP_SCINFO_ATTR_STATE;

static CSSM_RETURN CSSMAPI __MDSU_ConstructCDSACSPSCInfoAttributes(
                            MDSU_CONTEXT *pContext,
                            const void *pSchemaData,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            CSSM_BOOL bFillValues,
                            void **ppAttrState )
{
    CSSM_RETURN rv = CSSM_OK;
    const MDSU_CDSA_CSP_SMARTCARD_INFO_SCHEMA *pSmartcard = 
                        (const MDSU_CDSA_CSP_SMARTCARD_INFO_SCHEMA*)pSchemaData;
    _MDSU_CDSA_CSP_SCINFO_ATTR_STATE *pState = NULL;
    CSSM_DB_ATTRIBUTE_DATA *DbRecordAttributes;
    uint32 i;

    assert( pContext && ppAttrState && DbRecord );
    assert( ( bFillValues == CSSM_FALSE ) ||
            ( ( bFillValues == CSSM_TRUE ) && pSchemaData ) );

    /* Allocate the memory for the attributes */
    pState = MDSU_malloc( pContext,
                          sizeof(*pState) );
    if ( pState == NULL )
    {
        ERR( rv = CSSMERR_CSSM_MEMORY_ERROR );
    }
    else
    {
        /* Clear the structure */
        memset( pState,
                0,
                sizeof(*pState) );
        DbRecordAttributes = pState->DbRecordAttributes;

        /* Fill in the record information */
        DbRecord->DataRecordType = 
                            MDS_CDSADIR_CSP_SC_INFO_RECORDTYPE;
        DbRecord->SemanticInformation = 0;
        DbRecord->NumberOfAttributes = 
                            MDS_CDSADIR_CSP_SC_INFO_NUM_ATTRIBUTES;
        DbRecord->AttributeData = DbRecordAttributes;

        /* Fill in attribute info */
        DbRecordAttributes[__ModuleID].Info = s_MdsAttrInfo_ModuleId;
        DbRecordAttributes[__SSID].Info = s_MdsAttrInfo_SSID;
        DbRecordAttributes[__ScDesc].Info = s_MdsAttrInfo_SCDesc;
        DbRecordAttributes[__ScVendor].Info = s_MdsAttrInfo_SCVendor;
        DbRecordAttributes[__ScVersion].Info =
                                                s_MdsAttrInfo_SCVersion;
        DbRecordAttributes[__ScFirmwareVersion].Info = s_MdsAttrInfo_SCFirmwareVersion;
        DbRecordAttributes[__ScFlags].Info = s_MdsAttrInfo_SCFlags;
        DbRecordAttributes[__ScCustomFlags].Info = s_MdsAttrInfo_SCCustomFlags;
        DbRecordAttributes[__ScSerialNumber].Info = s_MdsAttrInfo_SCSerialNumber;

        if ( bFillValues == CSSM_TRUE )
        {
            /* Connect the attribute value array to the attribute array */
            for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
            {
                DbRecordAttributes[i].Value = &pState->AttributeValues[i];
                DbRecordAttributes[i].NumberOfValues = 1;
            }

            /* Fill in the record contents */
            /* ModuleID */
            _MDSU_GetPrintableGUID( &pSmartcard->ModuleID,
                                    (char*)pState->PrintableGuid );
            DbRecordAttributes[__ModuleID].Value->Data =
                                                    pState->PrintableGuid;
            DbRecordAttributes[__ModuleID].Value->Length = 
                                strlen( (char*)pState->PrintableGuid ) + 1;

            /* SSID */
            DbRecordAttributes[__SSID].Value->Length = sizeof(uint32);
            DbRecordAttributes[__SSID].Value->Data = (uint8*)&pSmartcard->SSID;

            /* ScDesc */
            DbRecordAttributes[__ScDesc].Value->Data = 
                                                    (uint8*)pSmartcard->ScDesc;
            DbRecordAttributes[__ScDesc].Value->Length =
                                            strlen( pSmartcard->ScDesc ) + 1;

            /* ScVendor */
            DbRecordAttributes[__ScVendor].Value->Length = 
                                            strlen( pSmartcard->ScVendor ) + 1;
            DbRecordAttributes[__ScVendor].Value->Data =
                                                (uint8*)pSmartcard->ScVendor;

            /* ScVersion */
            _MDSU_GetPrintableVersion(
                                &pSmartcard->ScVersion,
                                (char*)pState->PrintScVersion );
            DbRecordAttributes[__ScVersion].Value->Data =
                                                pState->PrintScVersion;
            DbRecordAttributes[__ScVersion].Value->Length = 
                            strlen( (char*)pState->PrintScVersion ) + 1;

            /* ScFirmwareVersion */
            _MDSU_GetPrintableVersion(
                                &pSmartcard->ScFirmwareVersion,
                                (char*)pState->PrintFirmwareVersion );
            DbRecordAttributes[__ScFirmwareVersion].Value->Data =
                                                pState->PrintFirmwareVersion;
            DbRecordAttributes[__ScFirmwareVersion].Value->Length = 
                        strlen( (char*)pState->PrintFirmwareVersion ) + 1;

            /* ScFlags */
            DbRecordAttributes[__ScFlags].Value->Length = sizeof(uint32);
            DbRecordAttributes[__ScFlags].Value->Data =
                                                (uint8*)&pSmartcard->ScFlags;

            /* ScCustomFlags */
            DbRecordAttributes[__ScCustomFlags].Value->Length = sizeof(uint32);
            DbRecordAttributes[__ScCustomFlags].Value->Data =
                                                (uint8*)&pSmartcard->ScCustomFlags;
            /* ScSerialNumber */
            DbRecordAttributes[__ScSerialNumber].Value->Length = 
                                            strlen( pSmartcard->ScSerialNumber ) + 1;
            DbRecordAttributes[__ScSerialNumber].Value->Data =
                                                (uint8*)pSmartcard->ScSerialNumber;
        } /* bFillValues */
        else
        {
            for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
            {
                DbRecordAttributes[i].Value = NULL;
                DbRecordAttributes[i].NumberOfValues = 0;
            }
        }
    }

    *ppAttrState = (void*)pState;

    return rv;
}

static CSSM_RETURN CSSMAPI __MDSU_DestructCDSACSPSCInfoAttributes(
                            MDSU_CONTEXT *pContext,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            CSSM_BOOL bFreeAll,
                            void *pAttrState )
{
    CSSM_DB_ATTRIBUTE_DATA *DbRecordAttributes;
    uint32 i;

    assert( pContext && DbRecord );

    /* Free the buffers allocated for the attribute values */
    DbRecordAttributes = DbRecord->AttributeData;
    if ( bFreeAll == CSSM_TRUE )
    {
        for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
        {
            if ( DbRecordAttributes[i].Value )
            {
                if ( DbRecordAttributes[i].Value->Data )
                {
                    MDSU_free( pContext,
                               DbRecordAttributes[i].Value->Data );
                }
                MDSU_free( pContext,
                           DbRecordAttributes[i].Value );
            }
        }
    }

    /* Free the attribute array */
    MDSU_free( pContext, pAttrState );

    return CSSM_OK;
}

typedef struct _mdsu_cdsa_csp_scinfo_pred_state
{
    CSSM_SELECTION_PREDICATE Predicates[MDS_CDSADIR_CSP_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES];
    CSSM_DATA PredicateValues[MDS_CDSADIR_CSP_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES];
    uint8 PrintableGuid[_MDSU_PRINTABLE_GUID_LENGTH];
    uint8 PrintScVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
    uint8 PrintFirmwareVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
}
_MDSU_CDSA_CSP_SCINFO_PRED_STATE;

static CSSM_RETURN CSSMAPI __MDSU_ConstructCDSACSPSCInfoPredicate(
                            MDSU_CONTEXT *pContext,
                            const void *pSchemaTemplate,
                            MDSU_TEMPLATE_MASK ValidTemplates,
                            CSSM_QUERY *pQuery,
                            void **ppPredState )
{
    CSSM_RETURN rv = CSSM_OK;
    const MDSU_CDSA_CSP_SMARTCARD_INFO_SCHEMA *pSmartcard = 
                            (const MDSU_CDSA_CSP_SMARTCARD_INFO_SCHEMA*)pSchemaTemplate;
    _MDSU_CDSA_CSP_SCINFO_PRED_STATE *pState = NULL;
    CSSM_SELECTION_PREDICATE *pPredicates = NULL;
    uint32 uNumPredicates;
    uint32 uIndex = 0;
    uint32 i;

    assert( pContext && 
            pQuery &&
            ppPredState &&
            ( ( pSchemaTemplate && ValidTemplates ) ||
              ( !pSchemaTemplate && !ValidTemplates ) ) );

    /* Count the predicates */
    _MDSU_CountBits( ValidTemplates, uNumPredicates );

    /* Allocate the predicate state and link the value structures */
    if ( uNumPredicates != 0 )
    {
        pState = MDSU_malloc( pContext, sizeof(*pState) );
        if ( pState == NULL )
        {
            ERR( rv = CSSMERR_CSSM_MEMORY_ERROR );
        }
        else
        {
            memset( pState,
                    0,
                    sizeof(*pState) );
            pPredicates = pState->Predicates;

            /* Allocate a single attribute value struct per attribute */
            for ( i = 0; i < uNumPredicates; i++ )
            {
                pPredicates[i].Attribute.Value =
                                            &pState->PredicateValues[i];
                pPredicates[i].Attribute.NumberOfValues = 1;
            }
        } /* Allocate and link predicate values */
    } /* uNumPredicates != 0 */

    if ( rv == CSSM_OK )
    {
        /* Fill in the query parameters */
        pQuery->RecordType = MDS_CDSADIR_CSP_SC_INFO_RECORDTYPE;
        pQuery->Conjunctive = CSSM_DB_AND;
        pQuery->NumSelectionPredicates = uNumPredicates;
        pQuery->QueryLimits.TimeLimit = CSSM_QUERY_TIMELIMIT_NONE;
        pQuery->QueryLimits.SizeLimit = CSSM_QUERY_SIZELIMIT_NONE;
        pQuery->QueryFlags = 0;
        pQuery->SelectionPredicate = pPredicates;
        for ( uIndex = 0; uIndex < uNumPredicates; uIndex++ )
        {
            pPredicates[uIndex].DbOperator = CSSM_DB_EQUAL;
        }

        /* Fill in the predicate data */
        uIndex = 0;
        if ( ValidTemplates & MDSU_CSPSC_MODULE_ID )
        {
            /* ModuleID */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ModuleId;
            _MDSU_GetPrintableGUID( &pSmartcard->ModuleID,
                                    (char*)pState->PrintableGuid );
            pPredicates[uIndex].Attribute.Value->Data = pState->PrintableGuid;
            pPredicates[uIndex].Attribute.Value->Length = 
                                strlen( (char*)pState->PrintableGuid ) + 1;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_CSPSC_SSID )
        {
            /* SSID */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_SSID;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        sizeof(pSmartcard->SSID) ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)&pSmartcard->SSID;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_CSPSC_DESC )
        {
            /* ScDesc */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_SCDesc;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pSmartcard->ScDesc ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pSmartcard->ScDesc;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_CSPSC_VENDOR )
        {
            /* ScVendor */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_SCVendor;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pSmartcard->ScVendor ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pSmartcard->ScVendor;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_CSPSC_VERSION )
        {
            /* ScVersion */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_SCVersion;
            _MDSU_GetPrintableVersion( &pSmartcard->ScVersion,
                                       (char*)pState->PrintScVersion );
            pPredicates[uIndex].Attribute.Value->Data =
                                                pState->PrintScVersion;
            pPredicates[uIndex].Attribute.Value->Length = 
                            strlen( (char*)pState->PrintScVersion ) + 1;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_CSPSC_FW_VERSION )
        {
            /* ScFirmwareVersion */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_SCFirmwareVersion;
            _MDSU_GetPrintableVersion( &pSmartcard->ScFirmwareVersion,
                                       (char*)pState->PrintFirmwareVersion );
            pPredicates[uIndex].Attribute.Value->Data =
                                                pState->PrintFirmwareVersion;
            pPredicates[uIndex].Attribute.Value->Length = 
                            strlen( (char*)pState->PrintFirmwareVersion ) + 1;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_CSPSC_FLAGS )
        {
            /* ScFlags */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_SCFlags;
            pPredicates[uIndex].Attribute.Value->Length = sizeof(uint32);
            pPredicates[uIndex].Attribute.Value->Data = 
                                                    (uint8*)&pSmartcard->ScFlags;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_CSPSC_CUSTOM_FLAGS )
        {
            /* ScCustomFlags */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_SCCustomFlags;
            pPredicates[uIndex].Attribute.Value->Length = sizeof(uint32);
            pPredicates[uIndex].Attribute.Value->Data = 
                                                    (uint8*)&pSmartcard->ScCustomFlags;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_CSPSC_SERIAL_NUMBER )
        {
            /* ScSerialNumber */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_SCSerialNumber;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pSmartcard->ScSerialNumber ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pSmartcard->ScSerialNumber;
            uIndex++;
        }

    } /* Initial allocation successful */

    *ppPredState = (void*)pState;

    return rv;
}

static CSSM_RETURN CSSMAPI __MDSU_DestructCDSACSPSCInfoPredicate(
                            MDSU_CONTEXT *pContext,
                            CSSM_QUERY *pQuery,
                            void *pPredState )
{
    assert( pContext && pQuery );

    /* Free the attribute array */
    if ( pPredState != NULL )
    {
        MDSU_free( pContext, pPredState );
    }

    return CSSM_OK;
}

static CSSM_RETURN CSSMAPI __MDSU_ConvertCDSACSPSCInfoAttributes(
                            MDSU_CONTEXT *pContext,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            void *pSchemaData )
{
    CSSM_RETURN crReturn;
    MDSU_CDSA_CSP_SMARTCARD_INFO_SCHEMA *pSmartcard = 
                            (MDSU_CDSA_CSP_SMARTCARD_INFO_SCHEMA*)pSchemaData;
    const CSSM_DB_ATTRIBUTE_DATA *DbAttr = NULL;

    assert( pSchemaData && DbRecord && pContext );

    DbAttr = DbRecord->AttributeData;

    /* ModuleID */
    crReturn = _MDSU_GetStructuredGUID( 
                                (char*)DbAttr[__ModuleID].Value->Data,
                                &pSmartcard->ModuleID );

    /* SSID */
    pSmartcard->SSID = *(uint32*)DbAttr[__SSID].Value->Data;

    /* ScDesc */
	if ( DbAttr[__ScDesc].Value->Data != NULL )
        strncpy( pSmartcard->ScDesc,
                 (char*)DbAttr[__ScDesc].Value->Data,
			     DbAttr[__ScDesc].Value->Length );
	else
		strcpy( pSmartcard->ScDesc, "" );

    /* ScVendor */
	if ( DbAttr[__ScVendor].Value->Data != NULL )
        strncpy( pSmartcard->ScVendor,
                 (char*)DbAttr[__ScVendor].Value->Data,
			     DbAttr[__ScVendor].Value->Length );
	else
		strcpy( pSmartcard->ScVendor, "" );

    /* ScVersion */
    _MDSU_GetStructuredVersion( (char*)DbAttr[__ScVersion].Value->Data,
                                &pSmartcard->ScVersion );

    /* ScFirmwareVersion */
    _MDSU_GetStructuredVersion( (char*)DbAttr[__ScFirmwareVersion].Value->Data,
                                &pSmartcard->ScFirmwareVersion );

    /* ScFlags */
    pSmartcard->ScFlags = *(uint32*)DbAttr[__ScFlags].Value->Data;

    /* ScCustomFlags */
    pSmartcard->ScCustomFlags = *(uint32*)DbAttr[__ScCustomFlags].Value->Data;

    /* ScSerialNumber */
 	if ( DbAttr[__ScSerialNumber].Value->Data != NULL )
        strncpy( pSmartcard->ScSerialNumber,
                 (char*)DbAttr[__ScSerialNumber].Value->Data,
			     DbAttr[__ScSerialNumber].Value->Length );
	else
		strcpy( pSmartcard->ScSerialNumber, "" );

    return CSSM_OK;
}

const IfMdsuSchemaManipulator IfiMdsuCdsadirCSPScInfoSchema =
{
    __MDSU_ConstructCDSACSPSCInfoAttributes,
    __MDSU_DestructCDSACSPSCInfoAttributes,
    __MDSU_ConstructCDSACSPSCInfoPredicate,
    __MDSU_DestructCDSACSPSCInfoPredicate,
    __MDSU_ConvertCDSACSPSCInfoAttributes,
    MDS_CDSADIR_CSP_SC_INFO_RECORDTYPE
};

