/*-----------------------------------------------------------------------
 * File: CDSAEMM.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */


#include <assert.h>
#include <string.h>

#include "cssmtype.h"
#include "cssmerr.h"
#ifdef VMS
#include "cssmport.h"
#else
#include "port/cssmport.h"
#endif
#include "mds_schema.h"

#include "mds.h"
#include "mds_util.h"

#include "convert.h"
#include "internal.h"

#define __ModuleID          ( 0 )
#define __Manifest          ( 1 )
#define __ModuleName        ( 2 )
#define __Path              ( 3 )
#define __CdsaVersion       ( 4 )
#define __EmmSpecVersion    ( 5 )
#define __Desc              ( 6 )
#define __PolicyStmt        ( 7 )
#define __EmmVersion        ( 8 )
#define __EmmVendor         ( 9 )
#define __EmmType           ( 10 )

typedef struct _mdsu_cdsa_emm_attr_state
{
    CSSM_DB_ATTRIBUTE_DATA DbRecordAttributes[MDS_CDSADIR_EMM_NUM_ATTRIBUTES];
    CSSM_DATA AttributeValues[MDS_CDSADIR_EMM_NUM_ATTRIBUTES];
    uint8 PrintableGuid[_MDSU_PRINTABLE_GUID_LENGTH];
    uint8 PrintableCDSAVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
    uint8 PrintableEMMVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
    uint8 PrintableEMMSpecVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
}
_MDSU_CDSA_EMM_ATTR_STATE;

static CSSM_RETURN CSSMAPI __MDSU_ConstructCDSAEMMAttributes(
                            MDSU_CONTEXT *pContext,
                            const void *pSchemaData,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            CSSM_BOOL bFillValues,
                            void **ppAttrState )
{
    CSSM_RETURN rv = CSSM_OK;
    const MDSU_CDSA_EMM_SCHEMA *pEMM = 
                            (const MDSU_CDSA_EMM_SCHEMA*)pSchemaData;
    _MDSU_CDSA_EMM_ATTR_STATE *pState = NULL;
    CSSM_DB_ATTRIBUTE_DATA *DbRecordAttributes;
    uint32 i;

    assert( pContext && DbRecord && ppAttrState );
    assert( ( bFillValues == CSSM_FALSE ) ||
            ( ( bFillValues == CSSM_TRUE ) && pSchemaData ) );

    /* Allocate the memory for the attributes */
    pState = MDSU_malloc( pContext, sizeof(_MDSU_CDSA_EMM_ATTR_STATE) );
    if ( pState == NULL )
    {
        ERR( rv = CSSMERR_CSSM_MEMORY_ERROR );
    }
    else
    {
        /* Clear the structure */
        memset( pState, 0, sizeof(_MDSU_CDSA_EMM_ATTR_STATE) );
        DbRecordAttributes = pState->DbRecordAttributes;

        /* Fill in the record information */
        DbRecord->DataRecordType = MDS_CDSADIR_EMM_RECORDTYPE;
        DbRecord->SemanticInformation = 0;
        DbRecord->NumberOfAttributes = MDS_CDSADIR_EMM_NUM_ATTRIBUTES;
        DbRecord->AttributeData = DbRecordAttributes;

        /* Fill in attribute info */
        DbRecordAttributes[__ModuleID].Info = s_MdsAttrInfo_ModuleId;
        DbRecordAttributes[__Manifest].Info = s_MdsAttrInfo_Manifest;
        DbRecordAttributes[__ModuleName].Info = s_MdsAttrInfo_ModuleName;
        DbRecordAttributes[__Path].Info = s_MdsAttrInfo_Path;
        DbRecordAttributes[__CdsaVersion].Info = s_MdsAttrInfo_CDSAVersion;
        DbRecordAttributes[__EmmSpecVersion].Info = s_MdsAttrInfo_EMMSpecVersion;
        DbRecordAttributes[__Desc].Info = s_MdsAttrInfo_Desc;
        DbRecordAttributes[__PolicyStmt].Info = s_MdsAttrInfo_PolicyStmt;
        DbRecordAttributes[__EmmVersion].Info = s_MdsAttrInfo_EmmVersion;
        DbRecordAttributes[__EmmVendor].Info = s_MdsAttrInfo_EmmVendor;
        DbRecordAttributes[__EmmType].Info = s_MdsAttrInfo_EmmType;

        if ( bFillValues == CSSM_TRUE )
        {
            /* Connect the attribute value array to the attribute array */
            for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
            {
                DbRecordAttributes[i].Value = &pState->AttributeValues[i];
                DbRecordAttributes[i].NumberOfValues = 1;
            }

            /* Fill in the record contents */
            /* ModuleID */
            _MDSU_GetPrintableGUID( &pEMM->ModuleID,
                                    (char*)pState->PrintableGuid );
            DbRecordAttributes[__ModuleID].Value->Data = 
                                                    pState->PrintableGuid;
            DbRecordAttributes[__ModuleID].Value->Length = 
                                strlen( (char*)pState->PrintableGuid ) + 1;

            /* Manifest */
            (*DbRecordAttributes[__Manifest].Value) = pEMM->Manifest;

            /* ModuleName */
            DbRecordAttributes[__ModuleName].Value->Length = 
                                        strlen( pEMM->ModuleName ) + 1 ;
            DbRecordAttributes[__ModuleName].Value->Data =
                                                (uint8*)pEMM->ModuleName;

            /* Path */
            DbRecordAttributes[__Path].Value->Length = 
                                        strlen( pEMM->Path ) + 1 ;
            DbRecordAttributes[__Path].Value->Data =
                                                (uint8*)pEMM->Path;

            /* CDSA Version */
            _MDSU_GetPrintableVersion( &pEMM->CDSAVersion,
                                       (char*)pState->PrintableCDSAVersion );
            DbRecordAttributes[__CdsaVersion].Value->Data = 
                                                    pState->PrintableCDSAVersion;
            DbRecordAttributes[__CdsaVersion].Value->Length = 
                                strlen( (char*)pState->PrintableCDSAVersion ) + 1;

            /* EMM Spec Version */
            _MDSU_GetPrintableVersion( &pEMM->EmmSpecVersion,
                                       (char*)pState->PrintableEMMSpecVersion );
            DbRecordAttributes[__EmmSpecVersion].Value->Data = 
                                                    pState->PrintableEMMSpecVersion;
            DbRecordAttributes[__EmmSpecVersion].Value->Length = 
                                strlen( (char*)pState->PrintableEMMSpecVersion ) + 1;

            /* Desc */
            DbRecordAttributes[__Desc].Value->Length = 
                                        strlen( pEMM->Desc ) + 1 ;
            DbRecordAttributes[__Desc].Value->Data =
                                                (uint8*)pEMM->Desc;

            /* PolicyStmt */
            (*DbRecordAttributes[__PolicyStmt].Value) = pEMM->PolicyStmt;

            /* EmmVersion */
            _MDSU_GetPrintableVersion( &pEMM->EmmVersion,
                                       (char*)pState->PrintableEMMVersion );
            DbRecordAttributes[__EmmVersion].Value->Data = 
                                                    pState->PrintableEMMVersion;
            DbRecordAttributes[__EmmVersion].Value->Length = 
                                strlen( (char*)pState->PrintableEMMVersion ) + 1;

            /* EmmVendor */
            DbRecordAttributes[__EmmVendor].Value->Length =
                                            strlen( pEMM->EmmVendor ) + 1;
            DbRecordAttributes[__EmmVendor].Value->Data =
                                            (uint8*)pEMM->EmmVendor;

            /* EmmType */
            DbRecordAttributes[__EmmType].Value->Length = sizeof(uint32);
            DbRecordAttributes[__EmmType].Value->Data = (uint8*)&pEMM->EmmType;

        } /* bFillValues */
        else
        {
            for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
            {
                DbRecordAttributes[i].Value = NULL;
                DbRecordAttributes[i].NumberOfValues = 0;
            }
        }

        if ( rv != CSSM_OK )
        {
            /* Deallocate memory if there was an error */
            MDSU_free( pContext,
                       pState );
            pState = NULL;
        }
    } /* allocate DbRecordAttributes */

    *ppAttrState = (void*)pState;

    return rv;
}

static CSSM_RETURN CSSMAPI __MDSU_DestructCDSAEMMAttributes(
                            MDSU_CONTEXT *pContext,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            CSSM_BOOL bFreeAll,
                            void *pAttrState )
{
    CSSM_DB_ATTRIBUTE_DATA *DbRecordAttributes;
    uint32 i;

    assert( pContext && DbRecord && pAttrState );

    /* Free the buffers allocated for the attribute values */
    DbRecordAttributes = DbRecord->AttributeData;
    if ( bFreeAll == CSSM_TRUE )
    {
        for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
        {
            if ( DbRecordAttributes[i].Value )
            {
                if ( DbRecordAttributes[i].Value->Data )
                {
                    MDSU_free( pContext,
                               DbRecordAttributes[i].Value->Data );
                }
                MDSU_free( pContext,
                           DbRecordAttributes[i].Value );
            }
        }
    }

    /* Free the state structure */
    MDSU_free( pContext,
               pAttrState );
    DbRecord->AttributeData = NULL;

    return CSSM_OK;
}

typedef struct _mdsu_cdsa_emm_pred_state
{
    CSSM_SELECTION_PREDICATE Predicates[MDS_CDSADIR_EMM_NUM_ATTRIBUTES];
    CSSM_DATA PredicateValues[MDS_CDSADIR_EMM_NUM_ATTRIBUTES];
    uint8 PrintableGuid[_MDSU_PRINTABLE_GUID_LENGTH];
    uint8 PrintableCDSAVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
    uint8 PrintableEMMVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
    uint8 PrintableEMMSpecVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
}
_MDSU_CDSA_EMM_PRED_STATE;

static CSSM_RETURN CSSMAPI __MDSU_ConstructCDSAEMMPredicate(
                            MDSU_CONTEXT *pContext,
                            const void *pSchemaTemplate,
                            MDSU_TEMPLATE_MASK ValidTemplates,
                            CSSM_QUERY *pQuery,
                            void **ppPredState )
{
    CSSM_RETURN rv = CSSM_OK;
    const MDSU_CDSA_EMM_SCHEMA *pEMM = 
                            (const MDSU_CDSA_EMM_SCHEMA*)pSchemaTemplate;
    _MDSU_CDSA_EMM_PRED_STATE *pState = NULL;
    CSSM_SELECTION_PREDICATE *pPredicates = NULL;
    uint32 uNumPredicates;
    uint32 uIndex = 0;
    uint32 i;

    assert( pContext && 
            pQuery &&
            ppPredState &&
            ( ( pSchemaTemplate && ValidTemplates ) ||
              ( !pSchemaTemplate && !ValidTemplates ) ) );

    /* Count the predicates */
    /* Clear the manifest bit. Searching on manifest is not supported */
    ValidTemplates &= (~MDSU_EMM_MANIFEST);
    _MDSU_CountBits( ValidTemplates, uNumPredicates );

    /* Allocate the predicate state and link the value structures */
    if ( uNumPredicates != 0 )
    {
        pState = MDSU_malloc( pContext, sizeof(_MDSU_CDSA_EMM_PRED_STATE) );
        if ( pState == NULL )
        {
            ERR( rv = CSSMERR_CSSM_MEMORY_ERROR );
        }
        else
        {
            memset( pState,
                    0,
                    sizeof(_MDSU_CDSA_EMM_PRED_STATE) );
            pPredicates = pState->Predicates;

            /* Allocate a single attribute value struct per attribute */
            for ( i = 0; i < uNumPredicates; i++ )
            {
                pPredicates[i].Attribute.Value = &pState->PredicateValues[i];
                pPredicates[i].Attribute.NumberOfValues = 1;
            }
        } /* Allocate and link predicate values */
    } /* uNumPredicates != 0 */

    if ( rv == CSSM_OK )
    {
        /* Fill in the query parameters */
        pQuery->RecordType = MDS_CDSADIR_EMM_RECORDTYPE;
        pQuery->Conjunctive = CSSM_DB_AND;
        pQuery->NumSelectionPredicates = uNumPredicates;
        pQuery->QueryLimits.TimeLimit = CSSM_QUERY_TIMELIMIT_NONE;
        pQuery->QueryLimits.SizeLimit = CSSM_QUERY_SIZELIMIT_NONE;
        pQuery->QueryFlags = 0;
        pQuery->SelectionPredicate = pPredicates;
        for ( uIndex = 0; uIndex < uNumPredicates; uIndex++ )
        {
            pPredicates[uIndex].DbOperator = CSSM_DB_EQUAL;
        }

        /* Fill in the predicate data */
        uIndex = 0;
        if ( ValidTemplates & MDSU_EMM_MODULE_ID )
        {
            /* ModuleID */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ModuleId;
            _MDSU_GetPrintableGUID( &pEMM->ModuleID,
                                    (char*)pState->PrintableGuid );
            pPredicates[uIndex].Attribute.Value->Data = pState->PrintableGuid;
            pPredicates[uIndex].Attribute.Value->Length = 
                                strlen( (char*)pState->PrintableGuid ) + 1;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMM_MODULE_NAME )
        {
            /* ModuleName */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ModuleName;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pEMM->ModuleName ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pEMM->ModuleName;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMM_PATH )
        {
            /* Path */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_Path;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pEMM->Path ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pEMM->Path;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMM_CDSA_VERSION )
        {
            /* CDSA Version */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_CDSAVersion;
            _MDSU_GetPrintableVersion( &pEMM->CDSAVersion,
                                       (char*)pState->PrintableCDSAVersion );
            pPredicates[uIndex].Attribute.Value->Data =
                                                pState->PrintableCDSAVersion;
            pPredicates[uIndex].Attribute.Value->Length = 
                            strlen( (char*)pState->PrintableCDSAVersion ) + 1;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMM_EMM_SPEC_VERSION )
        {
            /* EMM Spec Version */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_EMMSpecVersion;
            _MDSU_GetPrintableVersion( &pEMM->EmmSpecVersion,
                                       (char*)pState->PrintableEMMSpecVersion );
            pPredicates[uIndex].Attribute.Value->Data =
                                                pState->PrintableEMMSpecVersion;
            pPredicates[uIndex].Attribute.Value->Length = 
                            strlen( (char*)pState->PrintableEMMSpecVersion ) + 1;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMM_DESC )
        {
            /* Desc */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_Desc;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pEMM->Desc ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pEMM->Desc;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMM_POLICY_STMT )
        {
            /* PolicyStmt */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_PolicyStmt;
            (*pPredicates[uIndex].Attribute.Value) = pEMM->PolicyStmt;

            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMM_EMM_VERSION )
        {
            /* EmmVersion */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_EmmVersion;
            _MDSU_GetPrintableVersion( &pEMM->EmmVersion,
                                       (char*)pState->PrintableEMMVersion );
            pPredicates[uIndex].Attribute.Value->Data =
                                                pState->PrintableEMMVersion;
            pPredicates[uIndex].Attribute.Value->Length = 
                            strlen( (char*)pState->PrintableEMMVersion ) + 1;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMM_EMM_VENDOR )
        {
            /* EmmVendor */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_EmmVendor;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pEMM->EmmVendor ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pEMM->EmmVendor;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMM_EMM_TYPE )
        {
            /* EmmType - a ServiceMask */
            pPredicates[uIndex].Attribute.Info = 
                                                s_MdsAttrInfo_EmmType;
            pPredicates[uIndex].Attribute.Value->Length = sizeof(uint32);
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)&pEMM->EmmType;
            uIndex++;
        }

    } /* Initial allocation successful */

    *ppPredState = (void*)pState;

    return rv;
}

static CSSM_RETURN CSSMAPI __MDSU_DestructCDSAEMMPredicate(
                            MDSU_CONTEXT *pContext,
                            CSSM_QUERY *pQuery,
                            void *pPredState )
{
    assert( pContext && pQuery );

    /* Free the attribute array */
    if ( pPredState != NULL )
    {
        MDSU_free( pContext, pPredState );
    }

    return CSSM_OK;
}

static CSSM_RETURN CSSMAPI __MDSU_ConvertCDSAEMMAttributes(
                            MDSU_CONTEXT *pContext,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            void *pSchemaData )
{
    CSSM_RETURN crReturn;
    MDSU_CDSA_EMM_SCHEMA *pEMM = (MDSU_CDSA_EMM_SCHEMA*)pSchemaData;

    const CSSM_DB_ATTRIBUTE_DATA *DbAttr = NULL;

    assert( pSchemaData && DbRecord && pContext );

    DbAttr = DbRecord->AttributeData;

    /* ModuleID */
    crReturn = _MDSU_GetStructuredGUID( 
                                (char*)DbAttr[__ModuleID].Value->Data,
                                &pEMM->ModuleID );
    if ( crReturn != CSSM_OK )
    {
        return crReturn;
    }

    /* Manifest */
    if ( DbAttr[__Manifest].Value->Data != NULL )
    {
        pEMM->Manifest.Data = MDSU_malloc( 
                                        pContext,
                                        DbAttr[__Manifest].Value->Length );
        if ( pEMM->Manifest.Data == NULL )
        {
            return CSSMERR_CSSM_MEMORY_ERROR;
        }
        pEMM->Manifest.Length = DbAttr[__Manifest].Value->Length;
        memcpy( pEMM->Manifest.Data,
                DbAttr[__Manifest].Value->Data,
                DbAttr[__Manifest].Value->Length );
    }
    else
    {
        pEMM->Manifest.Data = NULL;
        pEMM->Manifest.Length = 0;
    }
    
    /* ModuleName */
	if ( DbAttr[__ModuleName].Value->Data != NULL )
        strncpy( pEMM->ModuleName,
                 (char*)DbAttr[__ModuleName].Value->Data,
				 DbAttr[__ModuleName].Value->Length );
	else
		strcpy( pEMM->ModuleName, "" );

    /* Path */
	if ( DbAttr[__Path].Value->Data != NULL )
        strncpy( pEMM->Path,
                 (char*)DbAttr[__Path].Value->Data,
				 DbAttr[__Path].Value->Length );
	else
		strcpy( pEMM->Path, "" );

    /* CDSA Version */
    crReturn = _MDSU_GetStructuredVersion( 
                                    (char*)DbAttr[__CdsaVersion].Value->Data,
                                    &pEMM->CDSAVersion );
    if ( crReturn != CSSM_OK )
    {
        if(NULL != pEMM->Manifest.Data)
        {
            MDSU_free( pContext, pEMM->Manifest.Data );
            pEMM->Manifest.Data = NULL;
        }
    }

    /* EMM Spec Version */
    crReturn = _MDSU_GetStructuredVersion( 
                                    (char*)DbAttr[__EmmSpecVersion].Value->Data,
                                    &pEMM->EmmSpecVersion );
    if ( crReturn != CSSM_OK )
    {
        if(NULL != pEMM->Manifest.Data)
        {
            MDSU_free( pContext, pEMM->Manifest.Data );
            pEMM->Manifest.Data = NULL;
        }
    }

    /* Desc */
 	if ( DbAttr[__Desc].Value->Data != NULL )
        strncpy( pEMM->Desc,
                 (char*)DbAttr[__Desc].Value->Data,
				 DbAttr[__Desc].Value->Length );
	else
		strcpy( pEMM->Desc, "" );

    /* PolicyStmt */
    if ( DbAttr[__PolicyStmt].Value->Data != NULL )
    {
        pEMM->PolicyStmt.Data = MDSU_malloc( 
                                        pContext,
                                        DbAttr[__PolicyStmt].Value->Length );
        if ( pEMM->PolicyStmt.Data == NULL )
        {
            return CSSMERR_CSSM_MEMORY_ERROR;
        }
        pEMM->PolicyStmt.Length = DbAttr[__PolicyStmt].Value->Length;
        memcpy( pEMM->PolicyStmt.Data,
                DbAttr[__PolicyStmt].Value->Data,
                DbAttr[__PolicyStmt].Value->Length );
    }
    else
    {
        pEMM->PolicyStmt.Data = NULL;
        pEMM->PolicyStmt.Length = 0;
    }

    /* EmmVersion */
    crReturn = _MDSU_GetStructuredVersion( 
                                    (char*)DbAttr[__EmmVersion].Value->Data,
                                    &pEMM->EmmVersion );
    if ( crReturn != CSSM_OK )
    {
        if(NULL != pEMM->PolicyStmt.Data)
        {
            MDSU_free( pContext, pEMM->PolicyStmt.Data );
            pEMM->PolicyStmt.Data = NULL;
        }
        if(NULL != pEMM->Manifest.Data)
        {
            MDSU_free( pContext, pEMM->Manifest.Data );
            pEMM->Manifest.Data = NULL;
        }
    }

    /* EmmVendor */
	if ( DbAttr[__EmmVendor].Value->Data != NULL )
        strncpy( pEMM->EmmVendor,
                 (char*)DbAttr[__EmmVendor].Value->Data,
				 DbAttr[__EmmVendor].Value->Length );
	else
		strcpy( pEMM->EmmVendor, "" );

    /* EmmType - a ServiceMask */
    pEMM->EmmType = *(uint32*)DbAttr[__EmmType].Value->Data;

    return CSSM_OK;
}

const IfMdsuSchemaManipulator IfiMdsuCdsadirEMMSchema =
{
    __MDSU_ConstructCDSAEMMAttributes,
    __MDSU_DestructCDSAEMMAttributes,
    __MDSU_ConstructCDSAEMMPredicate,
    __MDSU_DestructCDSAEMMPredicate,
    __MDSU_ConvertCDSAEMMAttributes,
    MDS_CDSADIR_EMM_RECORDTYPE
};
