/*-----------------------------------------------------------------------
 * File: MDS_APP_UTIL.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */


#include <assert.h>
#ifdef VMS
#include <string.h>
#endif
#include "cssmtype.h"
#include "cssmerr.h"

#include "mds_util_intrn.h"  /* IsBadCodePtr(), IsBadReadPtr() */
#include "mds_app_util.h"
#include "mds_schema.h"
#include "mds_util_type.h"
#include "mds_util_api.h"

MDS_FUNCS MDSFuncs;
CSSM_DL_DB_HANDLE hDLDBObject;
CSSM_DL_DB_HANDLE hDLDBCdsa;
CSSM_BOOL Initialised = CSSM_FALSE;




#ifndef CSSM_MEMTRACK_ON

void* CSSMAPI
_MDSUTIL_malloc(uint32  mem_size,
                void* allocRef)
{
    //allocRef ignored

    return malloc( mem_size );
}

void* CSSMAPI
_MDSUTIL_calloc(uint32  num_elem,
                uint32  num_bytes,
                void* allocRef)
{
    //allocRef ignored

    return calloc( num_elem, num_bytes );
}

void * CSSMAPI
_MDSUTIL_realloc(void *old_ptr,
                 uint32  num_bytes,
                 void* allocRef)
{
    //allocRef ignored

    return realloc( old_ptr, num_bytes );
}

void CSSMAPI
_MDSUTIL_free(void *mem_ptr,
              void* allocRef)
{
    //allocRef ignored

    if(NULL != mem_ptr)
    {
        free( mem_ptr );
    }
}

#else   /*  CSSM_MEMTRACK_ON */

#if defined (WIN32) && defined(_DEBUG)
#include <crtdbg.h>
#endif

void* CSSMAPI
_MDSUTIL_malloc(uint32  mem_size,
                void* allocRef,
                const char * szFilename,
                uint32 u32LineNumber)
{
    //allocRef ignored

#ifdef _DEBUG
    return _malloc_dbg( mem_size, _NORMAL_BLOCK, szFilename, u32LineNumber );
#else
    return malloc( mem_size );
#endif
}

void* CSSMAPI
_MDSUTIL_calloc(uint32  num_elem,
                uint32  num_bytes,
                void* allocRef,
                const char *szFilename,
                uint32 u32LineNumber)
{
    //allocRef ignored

#ifdef _DEBUG
    return _calloc_dbg( num_elem, num_bytes, _NORMAL_BLOCK, szFilename, u32LineNumber );
#else
    return calloc( num_elem, num_bytes );
#endif
}

void * CSSMAPI
_MDSUTIL_realloc(void *old_ptr,
                 uint32  num_bytes,
                 void* allocRef,
                 const char * szFilename,
                 uint32 u32LineNumber)
{
    //allocRef ignored

#ifdef _DEBUG
    return _realloc_dbg( old_ptr, num_bytes, _NORMAL_BLOCK, szFilename, u32LineNumber );
#else
    return realloc( old_ptr, num_bytes );
#endif
}

void CSSMAPI
_MDSUTIL_free(void *mem_ptr,
              void* allocRef,
              const char *File,
              uint32 Line)
{
    //allocRef ignored

    if (mem_ptr)
    {
#ifdef _DEBUG
    _free_dbg( mem_ptr, _NORMAL_BLOCK );
#else
    free( mem_ptr );
#endif
    }
}

#endif  /* CSSM_MEMTRACK_ON */



/* Memory function table used by all MDS contexts. */
static const CSSM_MEMORY_FUNCS sc_MDSUMemoryFuncs = { _MDSUTIL_malloc,
                                                      _MDSUTIL_free,
                                                      _MDSUTIL_realloc,
                                                      _MDSUTIL_calloc,
                                                      NULL };



CSSM_RETURN CSSMAPI MDSUTIL_InitEx(CSSM_BOOL ReadWrite)
{
    CSSM_DL_HANDLE Handle = 0;
    CSSM_DB_HANDLE dbHandle = 0;
    CSSM_DB_ACCESS_TYPE AccessType = CSSM_DB_ACCESS_READ;
    CSSM_RETURN retValue = CSSM_OK;
    CSSM_GUID MDSUTIL_Guid = intel_mds_guid_init;

    /* Attempt to initialize the MDS and get an MDS handle */
    if ((retValue = MDS_Initialize(&MDSUTIL_Guid,
                   NULL,
                   &sc_MDSUMemoryFuncs,
                   &MDSFuncs,
                   &Handle)) != CSSM_OK)
        return retValue;

    if ( ReadWrite == CSSM_TRUE )
        AccessType |= CSSM_DB_ACCESS_WRITE;

    if ((retValue = MDSFuncs.DbOpen(Handle, MDS_OBJECT_DIRECTORY, NULL, AccessType, NULL, NULL, &dbHandle)) != CSSM_OK)
    {
        MDS_Terminate (Handle);
        return (retValue);
    }
    
    hDLDBObject.DLHandle = Handle;
    hDLDBObject.DBHandle = dbHandle;

    if ((retValue = MDSFuncs.DbOpen (Handle, MDS_CDSA_DIRECTORY, NULL, AccessType, NULL, NULL, &dbHandle)) != CSSM_OK)
    {
        MDSFuncs.DbClose (hDLDBObject);
        MDS_Terminate (Handle);
        return (retValue);
    }

    hDLDBCdsa.DLHandle = Handle;
    hDLDBCdsa.DBHandle = dbHandle;
    Initialised = CSSM_TRUE;
    return (CSSM_OK);
}

void CSSMAPI MDSUTIL_Term()
{
    if (!IsBadCodePtr((CSSM_PROC_ADDR)MDSFuncs.DbClose)) 
    {
        MDSFuncs.DbClose (hDLDBObject);
        MDSFuncs.DbClose (hDLDBCdsa);
        MDS_Terminate (hDLDBCdsa.DLHandle);
        memset (&MDSFuncs, 0, sizeof (MDS_FUNCS));
    }
    Initialised = CSSM_FALSE;
    return;
}

CSSM_RETURN CSSMAPI mdsutil_CheckInit(void)
{
    if(Initialised)
        return CSSM_OK;
    else
        return CSSM_ERRCODE_MDS_ERROR;
}

/*----------------------------------------------------------
 *Name: cssm_CheckAPIMemFuncs
 *
 *Description:
 * Check to make sure that memory functions are initialize
 *
 *Parameters: 
 * MemoryFunctions (input) - pointer to memory function structure
 *
 *Returns: 
 * CSSM_FAIL - bad pointer somewhere
 * CSSM_OK - all pointers OK
 *
 *--------------------------------------------------------*/
CSSM_RETURN mdsutil_CheckAPIMemFuncs (
                           const CSSM_API_MEMORY_FUNCS *MemoryFunctions)
{
    /* Check validity of pointers passed in */
    if (IsBadReadPtr (MemoryFunctions, (uint32)sizeof (CSSM_API_MEMORY_FUNCS)))
        return CSSMERR_CSSM_INVALID_POINTER;

    /* Check validity of callbacks */
    if (IsBadCodePtr((CSSM_PROC_ADDR)MemoryFunctions->malloc_func))
        return CSSMERR_CSSM_INVALID_POINTER;

    if (IsBadCodePtr((CSSM_PROC_ADDR)MemoryFunctions->realloc_func))
        return CSSMERR_CSSM_INVALID_POINTER;

    if (IsBadCodePtr((CSSM_PROC_ADDR)MemoryFunctions->free_func))
        return CSSMERR_CSSM_INVALID_POINTER;

    if (IsBadCodePtr((CSSM_PROC_ADDR)MemoryFunctions->calloc_func))
        return CSSMERR_CSSM_INVALID_POINTER;

    return(CSSM_OK);
}
