/*-----------------------------------------------------------------------
 * File: PORTDLL.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

/* Portions:
 * (C) COPYRIGHT International Business Machines Corp. 1996, 1997
 * All Rights Reserved
 *
 * US Government Users Restricted Rights - Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 */
/*
 * OpenVMS History
 *
 * 29-Nov-2001	Carol Kelly
 *   - Implement port_ComputeAddressInModule for OpenVMS
 *
 */

#include "internal.h"
#ifdef VMS
#include <descrip.h>
#include <lib$routines.h>
#include <jpidef.h>
#include <imcbdef.h>
extern IMCB **CTL$GL_IMGLSTPTR;
#define VMS_STRING_TO_DESC(str, desc) \
{ \
    desc.dsc$w_length = strlen (str); \
    desc.dsc$a_pointer = str; \
    desc.dsc$b_dtype = DSC$K_DTYPE_T; \
    desc.dsc$b_class = DSC$K_CLASS_S; \
}
#endif

/* Dynamic Library Loading routines */

/*-----------------------------------------------------------------------------
 * Name: port_LoadLibrary
 *
 * Description:  Loads a dynamically linked library
 * 
 * Parameters: 
 * LibraryPathName (input)      :  Path name of library
 *
 * Returns:
 * Handle to the loaded library
 * NULL if an error occurs
 * 
 * Error Codes:
 * CSSMERR_CSSM_OS_ACCESS_DENIED - Unable to load the specified library
 * CSSMERR_CSSM_INVALID_POINTER - LibraryPathName pointer is not valid
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSMAPI 
port_LoadLibrary(
    const char *LibraryPathName,
    CSSM_LIB_HANDLE *lib_handle)
{
    CSSM_LIB_HANDLE tmp;

#ifdef VMS
    char *ModuleLogicalName;
    char *pChar;
#endif

    TRACE(port_trace_info, FNCODE_cssm_LoadLibrary);
    if (lib_handle == NULL)
    {
        RETURN(CSSMERR_CSSM_INVALID_POINTER);
    }
    *lib_handle = NULL;
    
    if (NULL == LibraryPathName || port_IsBadReadPtr(LibraryPathName, 1))
    {
        RETURN(CSSMERR_CSSM_INVALID_POINTER);
    }
    if (port_IsBadReadPtr(LibraryPathName, strlen(LibraryPathName)))
    {
        RETURN(CSSMERR_CSSM_INVALID_POINTER);
    }

#ifdef VMS
    /*
    ** If there is a path name passed in strip off everything except
    ** the filename with out extension so that a logical name will be
    ** left to pass into port_GetProcAddress.
    ** There is no telling what happens to 'LibraryPathName' after here
    ** so we must get some space that will not go away, until we are done. 
    ** e.g. cssm_ModuleLoad saves it in ModuleRecord to be used by cssm_ModuleAttach
    */
    pChar = NULL;
    pChar = strrchr( LibraryPathName, '/');
    if( pChar++ != NULL ){
       ModuleLogicalName = local_malloc(strlen((const char *)pChar),NULL);
       strcpy( ModuleLogicalName, pChar );
    }
    else {
       ModuleLogicalName = local_malloc(strlen((const char *)LibraryPathName),NULL);
       strcpy( ModuleLogicalName, (const char *)LibraryPathName );
    }
    pChar = NULL;
    pChar = strchr( ModuleLogicalName, '.' );
    if( pChar != NULL )
       *pChar = '\0';
    else
       RETURN (CSSMERR_CSSM_MDS_ERROR);
         
    *lib_handle = ModuleLogicalName;

    RETURN(CSSM_OK);
#endif

#if defined (WIN32)
    /* LoadLibrary will call DllMain() to init */
    if ((tmp = LoadLibrary(LibraryPathName)) == NULL)
	{
		RETURN(CSSMERR_CSSM_OS_ACCESS_DENIED);
    }
#elif defined (UNIX)
    /* dlopen will call _init() to init */
#if defined (LINUX)
	tmp = dlopen(LibraryPathName, RTLD_NOW);
#else
	tmp = dlopen(LibraryPathName, RTLD_LAZY);
#endif
    if (tmp == NULL)
    {
#ifdef _DBG
        printf ("%s", dlerror());
#endif
#ifdef LINUX
      port_syslog(CSSM_LOG_PORT, CSSM_LOG_ERR, 
		  "port_LoadLibrary: dlopen(%s) failed, error = %s\n", 
		  LibraryPathName, dlerror());
#endif
        RETURN(CSSMERR_CSSM_OS_ACCESS_DENIED);
    }
#endif
    *lib_handle = tmp;

    RETURN(CSSM_OK);
}



/*-----------------------------------------------------------------------------
 * Name: port_FreeLibrary
 *
 * Description:  Frees a dynamically linked library
 * 
 * Parameters: 
 * handle (input)   :  Handle of loaded library 
 *
 * Returns:
 * CSSM_OK if the operation is successful
 * CSSM_FAIL if there is an error
 * 
 * Error Codes:
 * CSSMERR_CSSM_OS_ACCESS_DENIED - Unable to free the specified library
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSMAPI port_FreeLibrary(CSSM_LIB_HANDLE handle)
{
    TRACE(port_trace_info, FNCODE_cssm_FreeLibrary);

    if (handle == NULL)
    {
        RETURN(CSSMERR_CSSM_INVALID_POINTER);
    }

#if defined(WIN32)
    if (FreeLibrary((HMODULE)handle) != TRUE)
    {
        RETURN(CSSMERR_CSSM_OS_ACCESS_DENIED);
    }
#elif defined (UNIX)
    if (dlclose(handle))
    {
#ifdef _DBG
        printf ("%s", dlerror());
#endif
#ifdef LINUX
	port_syslog(CSSM_LOG_PORT, CSSM_LOG_ERR,
		    "FreeLibrary(): dlclose(%x) failed, error = %s\n",
		    handle, dlerror());
#endif
        RETURN(CSSMERR_CSSM_OS_ACCESS_DENIED);
    }
#endif
    RETURN(CSSM_OK);
}




/*-----------------------------------------------------------------------------
 * Name: port_GetProcAddress
 *
 * Description:  Returns the address of the specified function
 * 
 * Parameters: 
 * handle (input)       :  Handle for a loaded library
 * func_name (input)    :  Name of function to get the address for
 *
 * Returns:
 * Address of the specified function
 * NULL if an error occurs
 * 
 * Error Codes:
 * CSSMERR_CSSM_CANT_GET_PROC_ADDR - Unable to get the procedure address
 * CSSMERR_CSSM_OS_ACCESS_DENIED - handle is invalid
 * CSSMERR_CSSM_INVALID_POINTER - func_name ptr is invalid
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSMAPI 
port_GetProcAddress(
    CSSM_LIB_HANDLE handle, 
    const char * func_name,
    CSSM_PROC_ADDR *proc)
{
    CSSM_PROC_ADDR tmp;
#ifdef VMS
    struct dsc$descriptor_s image_name_desc;
    struct dsc$descriptor_s routine_name_desc;
    long                    routine_address;
    int                     status = 0;
    char                    *image_name = handle;
#endif

    TRACE(port_trace_info, FNCODE_cssm_GetProcAddress);
    if (proc == NULL)
    {
        RETURN(CSSMERR_CSSM_INVALID_POINTER);
    }
    *proc = NULL;
#ifndef VMS
    if (handle == NULL)
    {
        RETURN(CSSMERR_CSSM_OS_ACCESS_DENIED);
    }
#endif
    if (NULL == func_name || port_IsBadReadPtr(func_name, 1))
    {
        RETURN(CSSMERR_CSSM_INVALID_POINTER);
    }
    if (port_IsBadReadPtr(func_name, strlen(func_name)))
    {
        RETURN(CSSMERR_CSSM_INVALID_POINTER);
    }

#ifdef VMS
    VMS_STRING_TO_DESC( image_name, image_name_desc );
    VMS_STRING_TO_DESC( (char *)func_name, routine_name_desc );
    status = lib$find_image_symbol( &image_name_desc, 
                                    &routine_name_desc,
                                    &routine_address );
    if( status != SS$_NORMAL ){
        RETURN(CSSMERR_CSSM_CANT_GET_PROC_ADDR);
    }
    else {
        *proc = (CSSM_PROC_ADDR)routine_address;
        RETURN(CSSM_OK);
    }
#endif

#if defined (WIN32)
    tmp = GetProcAddress(handle, func_name);    
    if (tmp == NULL)
    {
        RETURN(CSSMERR_CSSM_CANT_GET_PROC_ADDR);
    }
#elif defined (UNIX)
    tmp = (void *) dlsym(handle, func_name);
    if (tmp == NULL)
    {
#ifdef _DBG
        printf ("%s", dlerror());
#endif
#ifdef LINUX
	port_syslog(CSSM_LOG_PORT, CSSM_LOG_ERR, 
		    "port_GetProcAddress(): dlsym(%s) failed, error = %s\n",
		    func_name, dlerror());
#endif
        RETURN(CSSMERR_CSSM_CANT_GET_PROC_ADDR);
    }
#endif
    *proc = tmp;

    RETURN(CSSM_OK);
}



/*-----------------------------------------------------------------------------
 * Name: port_GetModuleHandle
 *
 * Description:  Returns the handle of the specified module
 * 
 * Parameters: 
 * module_name - Name of file to return a handle to
 *
 * Returns:
 * Handle to the specified module
 * NULL if an error occurs
 * 
 * Error Codes:
 * CSSMERR_CSSM_OS_ACCESS_DENIED - Unable to load the specified library
 * CSSMERR_CSSM_INVALID_POINTER - module_name pointer is not valid
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSMAPI 
port_GetModuleHandle(
    const char *module_name,
    CSSM_LIB_HANDLE *m_handle)
{
#if defined (WIN32)
    HMODULE module;
#endif

    TRACE(port_trace_info, FNCODE_cssm_GetModuleHandle);
    
    if (m_handle == NULL)
    {
        RETURN(CSSMERR_CSSM_INVALID_POINTER);
    }
    *m_handle = NULL;

#if defined (WIN32)
    if (NULL == module_name || port_IsBadReadPtr(module_name, 1))
    {
        RETURN(CSSMERR_CSSM_INVALID_POINTER);
    }

    if (port_IsBadReadPtr(module_name, strlen(module_name)))
    {
        RETURN(CSSMERR_CSSM_INVALID_POINTER);
    }

    module = GetModuleHandle(module_name);
    if (module == NULL)
    {
        RETURN(CSSMERR_CSSM_OS_ACCESS_DENIED);
    }
    *m_handle = module;
#elif defined (UNIX) || defined(VMS)
	if (port_LoadLibrary(module_name, m_handle) != CSSM_OK)
		RETURN(CSSMERR_CSSM_OS_ACCESS_DENIED);
#endif

    RETURN(CSSM_OK);
}

#ifdef LINUX
extern int _dlerror_run (void (*operate) (void *), void *args);
extern void _dl_signal_error (int errcode,const char *objname,const char *errstring);

int     
dlnam(void *handle, char *filename, int len)
{
  struct link_map *map = handle;
  
  void doit (void *a)
  {
    if (handle == NULL)
      _dl_signal_error (0, NULL, "Null dll handle is invalid");
    else
    {
      if ( (strlen(map->l_name) + 1) <= len )
	strcpy(filename, map->l_name);
      else
	_dl_signal_error (0, NULL, "Not enough memory to return the dll name");
    }
  }
  
  return _dlerror_run (doit,(void *)NULL) ? -1 : strlen(map->l_name);
}
void *
dlladdr(void *handle)
{
  struct link_map *map = handle;
  
  void doit (void *a)
  {
    if (handle == NULL)
      _dl_signal_error (0, NULL, "Null as dll handle is invalid");
  }
  
  return _dlerror_run (doit,(void *)NULL) ? NULL : (void *)(map->l_addr);
}
#endif

#if defined (__digital__)
#include <sys/types.h>
#include <loader.h>
/*----------------------------------------------------------------------------
 * Name: dlnam
 *
 * Description:  Given a handle to a shared object from dlopen(), gets
 *               the filename of the object.
 *
 * Parameters:
 * handle (input)   : Handle to shared object
 * buf (output)     : Filename buffer
 * buflen (input)   : Length in bytes of filename buffer
 *
 * Returns:
 *  0 if successful
 * -1 if an error occured
 *---------------------------------------------------------------------------*/
static int dlnam(void *handle, char *buf, int buflen)
{
    void *sym;
    ldr_process_t process;
    ldr_module_t mod_id;
    ldr_region_t reg_id;
    ldr_module_info_t mod_info;
    ldr_region_info_t reg_info;
    size_t size;
    void *endaddr;

    if ((sym = dlsym(handle, ".text")) == NULL)
    return -1;

    process = ldr_my_process();
    mod_id  = LDR_NULL_MODULE;

    for (;;)
    {
        if (ldr_next_module(process, &mod_id))
        return -1;
    if (mod_id == LDR_NULL_MODULE)
        break;

        if (ldr_inq_module(process, mod_id, &mod_info,
               sizeof mod_info, &size))
        return -1;
        if (size < sizeof mod_info)
        continue;

        for (reg_id = 0; reg_id < mod_info.lmi_nregion; reg_id++)
    {
            if (ldr_inq_region(process, mod_id, reg_id, &reg_info,
                               sizeof reg_info, &size))
        return -1;
        if (size < sizeof reg_info)
        continue;

        endaddr = (char *) reg_info.lri_mapaddr + reg_info.lri_size;
        if (sym >= reg_info.lri_mapaddr && sym < endaddr)
        {
        if (strlen(mod_info.lmi_name) >= buflen)
        {
            errno = ERANGE;
            return -1;
        }

        strcpy(buf, mod_info.lmi_name);
        return 0;
        }
        }
    }

    return -1;
}
#endif

/*-----------------------------------------------------------------------------
 * Name: port_GetModuleFileName
 *
 * Description:  Gets the filename of the specifed module
 * 
 * Parameters: 
 * module (input)   : Handle of a module
 * filename (output)    : Name of file
 * buf_length (input)   : Lenght of filename buffer
 *
 * Returns:
 * CSSM_OK if successful
 * CSSM_FAIL if an error occured
 * 
 * Error Codes:
 * CSSMERR_CSSM_OS_ACCESS_DENIED - Unable to get the module filename
 * CSSMERR_CSSM_INVALID_POINTER - LibraryPathName pointer is not valid
 * CSSMERR_CSSM_OS_ACCESS_DENIED - module handle is invalid
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSMAPI 
port_GetModuleFileName(
    CSSM_LIB_HANDLE module, 
    char *filename, 
    sint32 buf_length)
{
#if defined (WIN32)
    unsigned long  ret_val;
#endif

    TRACE(port_trace_info, FNCODE_cssm_GetModuleFileName);

#if defined (WIN32)
    if (module == NULL)
    {
        RETURN(CSSMERR_CSSM_OS_ACCESS_DENIED);
    }
    if (NULL == filename || port_IsBadReadPtr(filename, 1))
    {
        RETURN(CSSMERR_CSSM_INVALID_POINTER);
    }
    if (port_IsBadReadPtr(filename, strlen(filename)))
    {
        RETURN(CSSMERR_CSSM_INVALID_POINTER);
    }

    ret_val = GetModuleFileName(module, filename, (unsigned long)buf_length);

    if (ret_val == 0)
    {
        RETURN(CSSMERR_CSSM_OS_ACCESS_DENIED);
    }

#else  /* IBM_CHANGE */
#ifndef VMS
    if (dlnam(module, filename, buf_length)<0)
    {
        RETURN(CSSMERR_CSSM_OS_ACCESS_DENIED);
    }
#endif
#endif

    RETURN(CSSM_OK);
}


/*
 * getNTHeader
 *
 */
#if defined (WIN32)
PIMAGE_NT_HEADERS
port_getNTHeader(
    const unsigned char *image)
{
      PIMAGE_DOS_HEADER dosHeader;
      PIMAGE_NT_HEADERS pNTHeader;

      dosHeader = (PIMAGE_DOS_HEADER) image;
      if ( dosHeader->e_magic != IMAGE_DOS_SIGNATURE ){
            return NULL;
      }

      pNTHeader = (PIMAGE_NT_HEADERS)((DWORD) dosHeader + 
                  (DWORD) dosHeader->e_lfanew);
      if (pNTHeader->Signature != IMAGE_NT_SIGNATURE ){
            /* well, we can't figure this out */
            return NULL;
      }

      return pNTHeader;
}
#endif

CSSM_RETURN CSSMAPI
port_ComputeAddressInModule(CSSM_LIB_HANDLE LibHandle,
                            void* *StartAddress,
                            void* *EndAddress)
{
#if defined (VMS)
    IMCB *imcb_ppp = (void*)&CTL$GL_IMGLSTPTR;    /* A pointer to a pointer to the IMCB */
    IMCB *imcb_pp;
    IMCB *imcb_p;				
    char image_name[40];
    char LibHandleStr[40];
    char *ptr,*local_ptr;
    int local_len;
    unsigned char len;
    int found = CSSM_FALSE;

    if( imcb_ppp != NULL )                        /* pick up the next pointer */ 
       imcb_pp = imcb_ppp->imcb$l_flink;
    else
       return(!CSSM_OK);

     if( imcb_pp != NULL )                        /* pick up the pointer to the first IMCB of interest */
       imcb_p = imcb_pp->imcb$l_flink;
    else
       return(!CSSM_OK);

    /*
     * The module name in the imcb is uppercase, so upcase the input 
     */
    local_len = 0;
    local_ptr = (char*)&LibHandleStr;
    ptr = (char*)LibHandle;
    while( *ptr ){
      *local_ptr++ = toupper( (int)*ptr++ );
       local_len++;
    }
    
    /* the IMCBs are circlularly linked */
    while( (imcb_p != NULL) && (imcb_p != imcb_pp) && !found ){
       len = imcb_p->imcb$t_image_name[0];	/* imcb$t_image_name is a counted string */
       if( len == local_len ){
          memcpy( image_name, (imcb_p->imcb$t_image_name+1), len ); /* copy it into a temp for comparison*/
          if( strncmp( image_name, LibHandleStr, len ) == 0 ){
             *StartAddress = imcb_p->imcb$l_starting_address;
             *EndAddress   = imcb_p->imcb$l_end_address;
             found = CSSM_TRUE;
       }
       }
       imcb_p = imcb_p->imcb$l_flink;
    }

   /* Do we have any KFE resident sections which would be segmented? 
    * Only if we install images /resident which we don't
    */
   
    if( found )
       return(CSSM_OK);
    else 
       return(CSSMERR_CSSM_INTERNAL_ERROR);

#elif defined (WIN32)
    unsigned char *         pImage = NULL;
    PIMAGE_NT_HEADERS       pNTHeader = NULL;
    PIMAGE_SECTION_HEADER   SecHead = NULL;
    uint32                  SectionCount = 0;
    unsigned char *start = NULL;
    unsigned char *temp = NULL;
    unsigned char *end   = NULL;

    if( LibHandle == 0)
        return CSSMERR_CSSM_INTERNAL_ERROR;

    if (!StartAddress || !EndAddress)
        return CSSMERR_CSSM_INTERNAL_ERROR;

    pImage = (unsigned char *)LibHandle;
    if (!pImage)
        return CSSMERR_CSSM_INTERNAL_ERROR;

    pNTHeader = port_getNTHeader(pImage);
    if (!pNTHeader)
        return CSSMERR_CSSM_INTERNAL_ERROR;

    SecHead = IMAGE_FIRST_SECTION(pNTHeader);
    if (!SecHead)
        return CSSMERR_CSSM_INTERNAL_ERROR;

    start = pImage + SecHead->VirtualAddress;
    for(SectionCount = 0;
        SectionCount< (DWORD) pNTHeader->FileHeader.NumberOfSections;
        SectionCount++, SecHead++)
    {   
        temp = pImage + SecHead->VirtualAddress;       
        end   = temp +  SecHead->Misc.VirtualSize;
    }
    *StartAddress = start;
    *EndAddress = end;
    return CSSM_OK;
#else
    return (CSSM_OK);
#endif
}

