/*-----------------------------------------------------------------------
 * File: MDS_SHIM.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include <stdio.h>

#include "mds_shim.h"
#include "mds_dbfuncs.h"
#include "mds_shimutil.h"
#include "mds_schema.h"
#include "cssmport.h"
#include "cssmtype.h"

#ifdef VMS
#include <internal.h>
#endif

#if defined (WIN32)
#define MDS_BASENAME "mds300"
#else
#define MDS_BASENAME "libmds"
#endif

CSSM_MEMORY_FUNCS MemFuncs = { mds_local_malloc,
                               mds_local_free,
                               mds_local_realloc,
                               mds_local_calloc,
                               NULL};


CSSM_RETURN
MDS_RegisterModule( CSSM_GUID Guid,
                    const char *ModuleRoot,
                    const char *ModuleExt,
                    const char *ModulePath,
                    const char *ManifestPath,
                    MDS_MODULE_TYPE ModuleType )
{
    CSSM_RETURN RetVal = CSSM_OK;
    CSSM_LIB_HANDLE hMDSInst;
    CSSM_PROC_ADDR pfInit, pfTerminate;
    CSSM_DB_ACCESS_TYPE access = CSSM_DB_ACCESS_READ | CSSM_DB_ACCESS_WRITE;  // 3
    CSSM_DL_HANDLE Handle;
    CSSM_DB_HANDLE dbHandle = 0;
    CSSM_DL_DB_HANDLE hDLDB;
    CSSM_DATA Manifest, ModName, ModPath;
    char sysPath[_MAX_PATH], mdsPath[_MAX_PATH], ModuleName[_MAX_PATH];
#ifdef ISL_INCLUDED
    char credFile[_MAX_PATH];

    memset( credFile, 0, _MAX_PATH );
#endif
    memset( sysPath, 0, _MAX_PATH );
    memset( mdsPath, 0, _MAX_PATH );
    memset( ModuleName, 0, _MAX_PATH );

    GetSystemDirectory( sysPath, _MAX_PATH );
    sprintf( mdsPath, "%s%c%s%s", sysPath, DIRECTORY_SEPARATOR, 
		MDS_BASENAME, DLL_EXTENSION );
#ifdef WIN32
    hMDSInst = LoadLibrary( mdsPath );
#else
	port_LoadLibrary( mdsPath, &hMDSInst );
#endif
    if( hMDSInst == NULL )
    {
        return( CSSMERR_CSSM_INTERNAL_ERROR );
    }
    /* needed entry points from CSSM */
#ifdef WIN32
    pfInit = GetProcAddress( hMDSInst, "MDS_Initialize" );
    pfTerminate = GetProcAddress( hMDSInst, "MDS_Terminate" );

    if( pfInit == NULL )
    {
        FreeLibrary( hMDSInst );
#else
    if( port_GetProcAddress( hMDSInst, "MDS_Initialize" , &pfInit) != CSSM_OK )
	{
        port_FreeLibrary( hMDSInst );
#endif
        return( CSSMERR_CSSM_INTERNAL_ERROR );
    }

#ifdef WIN32
    if( pfTerminate == NULL )
	{
		FreeLibrary( hMDSInst );
#else
    if( port_GetProcAddress( hMDSInst, "MDS_Terminate" , &pfTerminate) != CSSM_OK )
    {
        port_FreeLibrary( hMDSInst );
#endif
        return( CSSMERR_CSSM_INTERNAL_ERROR );
    }

    if ((((MDSInitialize)pfInit)(NULL, 
                                 NULL,
                                 &MemFuncs,
                                 &MdsFuncs,
                                 &Handle)) != CSSM_OK)
    {
        printf ("MDS_Initialize failed.\n");
        return( CSSMERR_CSSM_INTERNAL_ERROR );
    }

    hDLDB.DLHandle = Handle;

    switch(ModuleType)
    {
        case MDS_CSSM:
        break;

        case MDS_ADDIN:
        break;

        case MDS_APPLICATION:
            sprintf( ModuleName, "%s%s", ModuleRoot, ModuleExt );
#ifdef ISL_INCLUDED
            sprintf( credFile, "%s%c%s%s", ManifestPath, 
				DIRECTORY_SEPARATOR, ModuleRoot, ".esw" );
            if( mds_LoadFile( credFile, &Manifest ) != CSSM_OK )
            {
                printf("mds_LoadFile Error: Couldn't read manifest...\n");
                RetVal = CSSMERR_CSSM_INTERNAL_ERROR;
            }
#else
            memset( &Manifest, 0, sizeof( Manifest ) );
#endif            
            ModName.Data = (unsigned char *)ModuleName;
            ModName.Length = strlen (ModuleName) + 1;
            ModPath.Data = (unsigned char *)ModulePath;
            ModPath.Length = strlen (ModulePath) + 1;
            if((MdsFuncs.DbOpen(Handle,
                                MDS_OBJECT_DIRECTORY,
                                NULL,
                                access,
                                NULL,
                                NULL,
                                &dbHandle)) != CSSM_OK)
            {
                printf ("Database open failed\n");
                ((MDSTerminate)pfTerminate)(Handle);
                return( CSSMERR_CSSM_INTERNAL_ERROR );
            } 

            hDLDB.DBHandle = dbHandle;
			
			/*
			 * Clean existing MDS record, if present
			 */
			mds_DeleteRecordByGuid(hDLDB,
                        &Guid,
                        MDS_OBJECT_RECORDTYPE);

			/*
			 * Insert the new record.
			 */
            RetVal = mds_InsertObjectRecord(hDLDB, 
                                            &Guid,
                                            Manifest,
                                            ModName,
                                            ModPath);
            MdsFuncs.DbClose (hDLDB);
        break;

        case MDS_MODULE_MANAGER:
        break;

        default:
            RetVal = CSSMERR_CSSM_INTERNAL_ERROR;
        break;
    }
    MdsFuncs.DbClose (hDLDB);

    ((MDSTerminate)pfTerminate)(Handle);

    return( RetVal );
}


CSSM_RETURN
MDS_DeRegisterModule( CSSM_GUID Guid,
                      MDS_MODULE_TYPE ModuleType )
{
    
    CSSM_RETURN RetVal = CSSM_OK;
    CSSM_LIB_HANDLE hMDSInst;
    CSSM_PROC_ADDR pfInit, pfTerminate;
    CSSM_DB_ACCESS_TYPE access = CSSM_DB_ACCESS_READ | CSSM_DB_ACCESS_WRITE;  // 3
    CSSM_DL_HANDLE Handle;
    CSSM_DB_HANDLE dbHandle;
    CSSM_DL_DB_HANDLE hDLDB;
    char sysPath[_MAX_PATH], mdsPath[_MAX_PATH];
    
    GetSystemDirectory( sysPath, _MAX_PATH );
    sprintf( mdsPath, "%s%c%s%s", sysPath, DIRECTORY_SEPARATOR, 
		MDS_BASENAME, DLL_EXTENSION );
#ifdef WIN32
    hMDSInst = LoadLibrary( mdsPath );
#else
	port_LoadLibrary( mdsPath, &hMDSInst );
#endif
    if( hMDSInst == NULL )
    {
        return( CSSMERR_CSSM_INTERNAL_ERROR );
    }

    /* needed entry points from CSSM */
#ifdef WIN32
    pfInit = GetProcAddress( hMDSInst, "MDS_Initialize" );
    pfTerminate = GetProcAddress( hMDSInst, "MDS_Terminate" );
 
    if( pfInit == NULL )
    {
        FreeLibrary( hMDSInst );
#else
	if( port_GetProcAddress( hMDSInst, "MDS_Initialize" , &pfInit) != CSSM_OK )
	{
        port_FreeLibrary( hMDSInst );
#endif
        return( CSSMERR_CSSM_INTERNAL_ERROR);
    }
 
#ifdef WIN32
    if( pfTerminate == NULL )
	{
        FreeLibrary( hMDSInst );
#else
    if( port_GetProcAddress( hMDSInst, "MDS_Terminate" , &pfTerminate) != CSSM_OK )
    {
        port_FreeLibrary( hMDSInst );
#endif
        return( CSSMERR_CSSM_INTERNAL_ERROR );
    }

    if ((RetVal = ((MDSInitialize)pfInit)( NULL, 
                                           NULL,
                                           &MemFuncs,
                                           &MdsFuncs,
                                           &Handle)) != CSSM_OK)
    {
        printf ("MDS_Initialize failed.\n");
        return( RetVal );
    }

    hDLDB.DLHandle = Handle;

    switch(ModuleType)
    {
        case MDS_CSSM:
        break;

        case MDS_ADDIN:
        break;

        case MDS_APPLICATION:
            if(( RetVal = MdsFuncs.DbOpen( Handle,
                                           MDS_OBJECT_DIRECTORY,
                                           NULL,
                                           access,
                                           NULL,
                                           NULL,
                                           &dbHandle)) != CSSM_OK)
            {
                printf ("Database open failed\n");
                ((MDSTerminate)pfTerminate)(Handle);
                return( RetVal );
            }

            hDLDB.DBHandle = dbHandle;

            RetVal = mds_DeleteRecordByGuid(hDLDB,
                                            &Guid,
                                            MDS_OBJECT_RECORDTYPE);

            MdsFuncs.DbClose (hDLDB);
        break;

        case MDS_MODULE_MANAGER:
        break;

        default:
            RetVal = CSSMERR_CSSM_INTERNAL_ERROR;
        break;
    }
    MdsFuncs.DbClose (hDLDB);

    ((MDSTerminate)pfTerminate)(Handle);

    return( RetVal );
}
