/*-----------------------------------------------------------------------
 * File: DSA.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
#include "integapi.h"
#include "cssm.h"
#include "ber_der.h"
#include "ber_bn.h"
#include "cssmtype.h"
#include "isl_internal.h"
#include <isl_memtrack.h>
extern ISL_DIGEST_METHODS PKCS7SHA1Methods;
extern ISL_SIGN_VERIFY_METHODS PKCS7DSAMethods;
extern CSSM_CSP_HANDLE gCSPHandle;

/* DSA algorithm glue */
static const unsigned char dsaoid[] = {
    0x06, 0x05, 0x2b, 0x0e, 0x03, 0x02, 27};  /* OIW combination dsa/sha-1 with parameters */
static const unsigned char pkcs7dsaoid[] = {
    0x06, 0x05, 0x2b, 0x0e, 0x03, 0x02, 12 };    /* OIW standalone dsa with parameters */
    
/* CSSM version of DSA sign and verify */
static void id(
    ISL_SERVICE_CLASS *algClass,                
    ISL_CONST_DATA *algID,              
    ISL_CONST_DATA *serviceName)
{
    *algClass = ISL_ServiceSignVerify;
    algID->Data = dsaoid;
    algID->Length = sizeof(dsaoid);
    serviceName->Data = (const unsigned char *)"DSA/SHA1";
    serviceName->Length = strlen((const char *)serviceName->Data);
}

#ifndef _IBM
#ifdef VMS
static ISL_STATUS sign(
#else
static ISL_ERROR_CODE sign(
#endif
    void *appContext,
    ISL_CONST_DATA parameters,
    const ISL_KEY_BLOCK *key,
    const ISL_DIGEST_METHODS     *digestMethods,
    ISL_CONST_DATA digestParameters,
    const ISL_GET_DATA_METHODS *getDataMethods,
    ISL_CONST_DATA getDataParameters,
    ISL_DATA *signature)
{
    return INTEGRITY_UNIMPLEMENTED;
}
#else // _IBM
static ISL_ERROR_CODE sign(
    void *appContext,
    ISL_CONST_DATA parameters,
    const ISL_KEY_BLOCK *key,
    const ISL_DIGEST_METHODS     *digestMethods,
    ISL_CONST_DATA digestParameters,
    const ISL_GET_DATA_METHODS *getDataMethods,
    ISL_CONST_DATA getDataParameters,
    ISL_DATA *signature)
{
    CSSM_RETURN cssmstatus;
    CSSM_CC_HANDLE hCC;
    CSSM_KEY cssmKey;
    CSSM_CRYPTO_DATA_PTR pass;

    if (digestMethods != &PKCS7SHA1Methods) 
        return ISL_UNSUPPORTED_ALGORITHM;
    
    pass = (CSSM_CRYPTO_DATA *)appContext;

    cssmKey.KeyHeader.BlobType = CSSM_KEYBLOB_RAW;
    cssmKey.KeyHeader.Format = CSSM_KEYBLOB_RAW_FORMAT_PKCS1;
    cssmKey.KeyHeader.HeaderVersion = CSSM_KEYHEADER_VERSION;   
    cssmKey.KeyHeader.AlgorithmId = CSSM_ALGID_DSA; 
    cssmKey.KeyHeader.KeyUsage = 
        CSSM_KEYUSE_ENCRYPT | CSSM_KEYUSE_DECRYPT | CSSM_KEYUSE_SIGN | CSSM_KEYUSE_VERIFY |
        CSSM_KEYUSE_WRAP | CSSM_KEYUSE_UNWRAP | CSSM_KEYUSE_DERIVE;
    cssmKey.KeyHeader.KeySizeInBits = 1024;
    cssmKey.KeyHeader.KeyClass = CSSM_KEYCLASS_PRIVATE_KEY;
    cssmKey.KeyHeader.WrapMode = 0;
    cssmKey.KeyHeader.WrapAlgorithmId = 0;
    cssmKey.KeyHeader.Reserved = 0;
    cssmKey.KeyData.Data = key->Data;
    cssmKey.KeyData.Length = key->Length;

    hCC = CSSM_CSP_CreateSignatureContext(gCSPHandle, CSSM_ALGID_DSA, (CSSM_CRYPTO_DATA *)appContext, &cssmKey);
    if (hCC == 0) 
        return INTEGRITY_INVALID_INPUT;

    cssmstatus = CSSM_SignData(hCC, (CSSM_DATA_PTR)&getDataParameters, 1,(CSSM_DATA_PTR)signature);

    CSSM_DeleteContext(hCC);

    if (cssmstatus != CSSM_OK)
        return INTEGRITY_INVALID_INPUT;

    return INTEGRITY_OK;
}
#endif // _IBM

static ISL_ERROR_CODE verify(
    ISL_CONST_DATA parameters,
    const ISL_KEY_BLOCK *key,
    ISL_CONST_DATA encryptedDigest,
    const ISL_DIGEST_METHODS *digestMethods,
    ISL_CONST_DATA digestParameters,
    const ISL_GET_DATA_METHODS *getDataMethods,
    ISL_CONST_DATA getDataParameters)
{
    return INTEGRITY_UNIMPLEMENTED;
}

static ISL_STATUS split(                       /* separate algorithms for PKCS#7 */
    const ISL_DIGEST_METHODS **DigestMethods,      /* digest algorithm */
    const ISL_SIGN_VERIFY_METHODS **SignMethods)  /* signature algorithm */
{
    return ISL_FAIL;
}

static ISL_STATUS SignWithClass(
    void *AppContext,
    ISL_CONST_DATA Parameters,
    const ISL_KEY_BLOCK *Key,
    const ISL_DIGEST_METHODS *DigestMethods,
    ISL_CONST_DATA DigestParameters,
    const ISL_GET_DATA_METHODS *GetDataMethods,
    ISL_CONST_DATA GetDataParameters,
    ISL_DATA_PTR EncryptedDigest,
    ISL_CLASS_PTR Class)
{
    CSSM_RETURN cssmstatus;
    CSSM_CC_HANDLE hCC;
    CSSM_KEY pubKey, privateKey;
    CSSM_CRYPTO_DATA_PTR pass;
    BER_PARSED_ITEM KeyNumbers[4];
    const uint8 *position, *end;
    uint32 i;
    uint8 keyBuff[1024];
    uint32 count;
    ISL_CSSM_MODULE_HANDLE *hCSP;
    ISL_CSSM_METHODS *pCssmMethods;

    if ((!Key) || (!EncryptedDigest) || (!AppContext))
        return ISL_FAIL;

    if (!Class)
        return ISL_FAIL;
    else if (!Class->ClassContext)
        return ISL_FAIL;
    else if(!((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->CSPList.ModuleHandles)
        return ISL_FAIL;
    else if(!((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->CSPList.NumberOfHandles)
        return ISL_FAIL;
    else if(!((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->MemoryFuncs)
        return ISL_FAIL;
    else if(!((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->CssmMethods)
        return ISL_FAIL;

        
    pass = (CSSM_CRYPTO_DATA *)AppContext;

    /* convert ISL_KEY to CSSM_KEY */
    memset((void*)&pubKey.KeyHeader, 0, sizeof(CSSM_KEYHEADER));
    pubKey.KeyHeader.HeaderVersion = CSSM_KEYHEADER_VERSION;
    pubKey.KeyHeader.BlobType = CSSM_KEYBLOB_RAW;
    pubKey.KeyHeader.Format = CSSM_KEYBLOB_RAW_FORMAT_FIPS186;
    pubKey.KeyHeader.KeyClass = CSSM_KEYCLASS_PUBLIC_KEY;
    pubKey.KeyHeader.AlgorithmId = CSSM_ALGID_DSA; 
    pubKey.KeyHeader.KeyUsage =  CSSM_KEYUSE_SIGN | CSSM_KEYUSE_VERIFY ;
    
    /* accept both (p,q,g,y) key or (p,q,g) parameters plus y key */
    if (Key->Data[0] == (BER_SEQUENCE | BER_CONSTRUCTED)) {
        pubKey.KeyData.Data =  (uint8*)Key->Data;
        pubKey.KeyData.Length = Key->Length;
    } else {
        /* collect the numbers, repackage as a sequence of 4 numbers */
        end = BER_ExpandItem((uint8*)Parameters.Data, Parameters.Length, &KeyNumbers[0]); /* strip sequence */
        if (!end) 
            return ISL_INVALID_INPUT;
        position = KeyNumbers[0].Content;
        for (i = 0; i < 3; i++) {
            position = BER_ExpandItem(position, end - position, &KeyNumbers[i]);
            if (!position)
                return ISL_INVALID_INPUT;
        }
        position = BER_ExpandItem((uint8 *)Key->Data, Key->Length, &KeyNumbers[3]);
        if (!position) 
            return ISL_INVALID_INPUT;
        pubKey.KeyData.Data = &keyBuff[0];
        pubKey.KeyData.Length = DER_ConstructItem((uint8*)Parameters.Data, 4, KeyNumbers, pubKey.KeyData.Data);         
    }
    
    pubKey.KeyHeader.LogicalKeySizeInBits = isl_GetKeyModulus(pubKey.KeyData.Data, pubKey.KeyData.Length);
    
    //initialize private key buffer
    memset((void*)&privateKey, 0, sizeof(CSSM_KEY));
    
    count = ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->CSPList.NumberOfHandles;
    hCSP = ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->CSPList.ModuleHandles;
    pCssmMethods = ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->CssmMethods;

    //find out which csp contains the corresponding private key
    for (i=0; i<count; i++) {
        //try to get the private key 
        if ( pCssmMethods->CSSM_CSP_ObtainPrivateKeyFromPublicKey (hCSP[i].Handle,
                                                     &pubKey,
                                                     &privateKey) == CSSM_OK)
        {

            //found the private key
            if (pCssmMethods->CSSM_CSP_CreateSignatureContext(hCSP[i].Handle, 
                                                        CSSM_ALGID_SHA1WithDSA, 
                                                        (CSSM_ACCESS_CREDENTIALS *)AppContext, 
                                                        &privateKey,
                                                        &hCC) != CSSM_OK)
            {
                return INTEGRITY_INVALID_INPUT;
            }
            cssmstatus = pCssmMethods->CSSM_SignData(hCC, 
                (CSSM_DATA_PTR)&GetDataParameters, 
                1,
                CSSM_ALGID_NONE,
                (CSSM_DATA_PTR)EncryptedDigest);
        
            
            //delete signature context
            pCssmMethods->CSSM_DeleteContext(hCC);

            //free the private key
            ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->MemoryFuncs->free_func(privateKey.KeyData.Data, 0);
            
            if (cssmstatus != CSSM_OK)
                return ISL_FAIL;
            else
                return ISL_OK;
        }
    }

    return ISL_FAIL;
}

static ISL_ERROR_CODE VerifyWithClass(
    ISL_CONST_DATA Parameters,
    const ISL_KEY_BLOCK *Key,
    ISL_CONST_DATA Signature,
    const ISL_DIGEST_METHODS *DigestMethods,
    ISL_CONST_DATA DigestParameters,
    const ISL_GET_DATA_METHODS *GetDataMethods,
    ISL_CONST_DATA GetDataParameters,
    ISL_CLASS_PTR Class)
{
    CSSM_CC_HANDLE hCC;
    CSSM_BOOL cssmStatus;
    CSSM_KEY pubKey;
    BER_PARSED_ITEM KeyNumbers[4];
    const uint8 *position, *end;
    uint8 keyBuff[1024];
    ISL_CSSM_MODULE_HANDLE *hCSP;
    ISL_CSSM_METHODS *pCssmMethods;
    uint32 count, i;
    CSSM_RETURN cssmret = CSSM_OK;

    if ((!Key) || (!Signature.Data) || (!Signature.Length) ||
        (!GetDataParameters.Data) || (!GetDataParameters.Length))
        return ISL_INVALID_INPUT;


    if (!Class)
        return ISL_FAIL;
    else if (!Class->ClassContext)
        return ISL_FAIL;
    else if(!((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->CSPList.ModuleHandles)
        return ISL_FAIL;
    else if(!((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->CSPList.NumberOfHandles)
        return ISL_FAIL;
    else if(!((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->CssmMethods)
        return ISL_FAIL;

    /* convert ISL_KEY to CSSM_KEY */
    memset((void*)&pubKey.KeyHeader,0, sizeof(CSSM_KEYHEADER));
    pubKey.KeyHeader.HeaderVersion = CSSM_KEYHEADER_VERSION;
    pubKey.KeyHeader.BlobType = CSSM_KEYBLOB_RAW;
    pubKey.KeyHeader.Format = CSSM_KEYBLOB_RAW_FORMAT_FIPS186;
    pubKey.KeyHeader.KeyClass = CSSM_KEYCLASS_PUBLIC_KEY;
    pubKey.KeyHeader.AlgorithmId = CSSM_ALGID_DSA; 
    pubKey.KeyHeader.KeyUsage = CSSM_KEYUSE_SIGN | CSSM_KEYUSE_VERIFY;
    ;

    /* accept both (p,q,g,y) key or (p,q,g) parameters plus y key */
    if (Key->Data[0] == (BER_SEQUENCE | BER_CONSTRUCTED)) {
        pubKey.KeyData.Data =  (uint8*)Key->Data;
        pubKey.KeyData.Length = Key->Length;
    } else {
        /* collect the numbers, repackage as a sequence of 4 numbers */
        end = BER_ExpandItem((uint8*)Parameters.Data, Parameters.Length, &KeyNumbers[0]); /* strip sequence */
        if (!end) 
            return ISL_INVALID_INPUT;
        position = KeyNumbers[0].Content;
        for (i = 0; i < 3; i++) {
            position = BER_ExpandItem(position, end - position, &KeyNumbers[i]);
            if (!position)
                return ISL_INVALID_INPUT;
        }
        position = BER_ExpandItem((uint8 *)Key->Data, Key->Length, &KeyNumbers[3]);
        if (!position) 
            return ISL_INVALID_INPUT;
        pubKey.KeyData.Data = &keyBuff[0];
        pubKey.KeyData.Length = DER_ConstructItem((uint8*)Parameters.Data, 4, KeyNumbers,pubKey.KeyData.Data);          
    }

    pubKey.KeyHeader.LogicalKeySizeInBits = isl_GetKeyModulus(pubKey.KeyData.Data, pubKey.KeyData.Length);

    count = ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->CSPList.NumberOfHandles;
    hCSP = ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->CSPList.ModuleHandles;
    pCssmMethods = ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->CssmMethods;

    //use any of the csp in the list to perform the verification
    for (i=0; i<count; i++) {
         cssmret = pCssmMethods->CSSM_CSP_CreateSignatureContext(hCSP[i].Handle, 
                                                            CSSM_ALGID_SHA1WithDSA,
                                                            NULL, 
                                                            &pubKey,
                                                            &hCC);
        if (CSSM_OK == cssmret) {
            cssmStatus =  pCssmMethods->CSSM_VerifyData(hCC, 
                                (CSSM_DATA_PTR)&GetDataParameters,
                                1, 
                                CSSM_ALGID_NONE,
                                (CSSM_DATA_PTR)&Signature);
                        
            pCssmMethods->CSSM_DeleteContext(hCC);
            if (cssmStatus == CSSM_OK)
                return ISL_OK;
            else
                return ISL_FAIL;
        }
    }
    return ISL_INVALID_INPUT;
}

static ISL_STATUS SplitCombinationWithClass(            /* separate algorithms for PKCS#7 */
    const ISL_DIGEST_METHODS **DigestMethods,      /* digest algorithm */
    const ISL_SIGN_VERIFY_METHODS **SignMethods,  /* signature algorithm */
    ISL_CLASS_PTR Class)
{
    if (!Class ||
        !Class->ClassContext)
        return ISL_FAIL;
        
    *DigestMethods = ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->SignatureMethodList.Methods[0].DigestMethods;
    *SignMethods = ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->SignatureMethodList.Methods[0].SignMethods;
    return ISL_OK;
}
    
ISL_SIGN_VERIFY_METHODS DSAMethods = {
    id,
    0, /* temp */
    sign,
    verify,
    split,
    SignWithClass, 
    VerifyWithClass, 
    SplitCombinationWithClass
};

/* non-combination algorithm DSA for PKCS7 signatures */

static void id7(
    ISL_SERVICE_CLASS *algClass,                
    ISL_CONST_DATA *algID,              
    ISL_CONST_DATA *serviceName)
{
    *algClass = ISL_ServiceSignVerify;
    algID->Data = pkcs7dsaoid;
    algID->Length = sizeof(pkcs7dsaoid);
    serviceName->Data = (const unsigned char *)"DSA";
    serviceName->Length = strlen((const char *)serviceName->Data);
}

/*
 *  The RSA sign is not implemented in 1.1 (especially the staged version), so we use
 *  the encrypt API after computing the hash function and padding.
 *  In the future, if RSA encrypt adds PCKS#1 padding, we may have to disable the padding
 *  (or perhaps even use the sign API) 
 */
#ifdef VMS
static ISL_STATUS sign7(
#else
static ISL_ERROR_CODE sign7(
#endif
    void *appContext,
    ISL_CONST_DATA parameters,
    const ISL_KEY_BLOCK *key,
    const ISL_DIGEST_METHODS     *digestMethods,
    ISL_CONST_DATA digestParameters,
    const ISL_GET_DATA_METHODS *getDataMethods,
    ISL_CONST_DATA getDataParameters,
    ISL_DATA *encryptedDigest)
{
    return INTEGRITY_UNIMPLEMENTED;
}

static ISL_ERROR_CODE verify7(
    ISL_CONST_DATA parameters,
    const ISL_KEY_BLOCK *key,
    ISL_CONST_DATA encryptedDigest,
    const ISL_DIGEST_METHODS *digestMethods,
    ISL_CONST_DATA digestParameters,
    const ISL_GET_DATA_METHODS *getDataMethods,
    ISL_CONST_DATA getDataParameters)
{
    return INTEGRITY_UNIMPLEMENTED;
}

static ISL_STATUS split7(                      /* separate algorithms for PKCS#7 */
        const ISL_DIGEST_METHODS **DigestMethods,      /* digest algorithm */
        const ISL_SIGN_VERIFY_METHODS **SignMethods)   /* signature algorithm */
{
    return ISL_FAIL;        /* not a combination algorithm */
}

static ISL_STATUS Sign7WithClass(
        void *AppContext,
        ISL_CONST_DATA Parameters,
        const ISL_KEY_BLOCK *Key,
        const ISL_DIGEST_METHODS *DigestMethods,
        ISL_CONST_DATA DigestParameters,
        const ISL_GET_DATA_METHODS *GetDataMethods,
        ISL_CONST_DATA GetDataParameters,
        ISL_DATA_PTR EncryptedDigest,
        ISL_CLASS_PTR Class){

    ISL_SIGN_VERIFY_METHODS *signMethods;
    uint32 i;

    i = 0;

    //use the corresponding signing method to sign the data
    while (i<((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->SignatureMethodList.NumberOfMethods) {
        if (DigestMethods == ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->SignatureMethodList.Methods[i].DigestMethods) {
            signMethods = ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->SignatureMethodList.Methods[i].SignMethods;
            return (*signMethods->SignWithClass)(AppContext, 
                Parameters, Key, 
                DigestMethods, DigestParameters, 
                GetDataMethods, GetDataParameters, 
                EncryptedDigest,
                signMethods->ServiceMethods.Class);
        }
        i++;
    } 
    return ISL_FAIL;

}

static ISL_ERROR_CODE Verify7WithClass(
        ISL_CONST_DATA Parameters,
        const ISL_KEY_BLOCK *Key,
        ISL_CONST_DATA Signature,
        const ISL_DIGEST_METHODS *DigestMethods,
        ISL_CONST_DATA DigestParameters,
        const ISL_GET_DATA_METHODS *GetDataMethods,
        ISL_CONST_DATA GetDataParameters,
        ISL_CLASS_PTR Class){

    ISL_SIGN_VERIFY_METHODS *signMethods;
    uint32 i = 0;

    while (i<((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->SignatureMethodList.NumberOfMethods) {
        if (DigestMethods == ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->SignatureMethodList.Methods[0].DigestMethods) {
            signMethods = ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->SignatureMethodList.Methods[i].SignMethods;
            return (*signMethods->VerifyWithClass)( 
                Parameters, Key, Signature, 
                DigestMethods, DigestParameters, 
                GetDataMethods, GetDataParameters,
                signMethods->ServiceMethods.Class);
        }
        i++;
    } 
    return ISL_FAIL;
}

static ISL_STATUS Split7CombinationWithClass(           /* separate algorithms for PKCS#7 */
        const ISL_DIGEST_METHODS **DigestMethods,      /* digest algorithm */
        const ISL_SIGN_VERIFY_METHODS **SignMethods,  /* signature algorithm */
        ISL_CLASS_PTR Class){
        return split7(DigestMethods, SignMethods);
}

ISL_SIGN_VERIFY_METHODS PKCS7DSAMethods = {
    id7,
    0,
    sign7,
    verify7,
    split7,
    Sign7WithClass, 
    Verify7WithClass, 
    Split7CombinationWithClass
};
