/*-----------------------------------------------------------------------
 * File: EISLARCHIVE.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/*
**	Implementation of ISL_ARCHIVE_CONTEXT class 
*/

#include "isl_internal.h"
#include "islutil.h"
#include "isl_parse.h"
#include <isl_memtrack.h>

#ifndef EISL
extern struct isl_archive_methods ArchiveMethods;
extern ISL_SIGNATURE_METHODS			PKCS7SignatureMethods;
extern ISL_CERTIFICATE_METHODS		X509CertMethods;
#endif

static
ISL_STATUS
BuildManifestTemplate(
	ISL_ARCHIVE_CONTEXT_PTR	Context,
	ISL_CONST_DATA_PTR	Template)
{
    uint8   *ptr = NULL;
    uint8   len = 11;
    
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_ARCHIVE_TAG,0x0f);
	
	/* allocate memory */
	Template->Data = isl_AllocateMemory(Context, len);
    if (!Template->Data) {
		REPORT_ALLOCATE_MEMORY_ERROR(Context, len);
	}

    ptr = (uint8 *)Template->Data;

	/* build up template */
    *(ptr++) = '<';
    *(ptr++) = 'm';
    *(ptr++) = 'a';
    *(ptr++) = 'n';
    *(ptr++) = 'i';
    *(ptr++) = 'f';
    *(ptr++) = 'e';
    *(ptr++) = 's';
    *(ptr++) = 't';
    *(ptr++) = '>';

    Template->Length = ptr - Template->Data;
    *(ptr++) = '\0';
    return ISL_OK;
}



static ISL_STATUS
BuildManifest(
	ISL_ARCHIVE_CONTEXT_PTR pArchive)
{
	ISL_CONST_DATA Template;
	ISL_DATA_CHAIN_PTR Results[MANIFEST_LAST_INDEX];
	ISL_DATA_CHAIN_PTR section;
	uint32 numSections;
	ISL_MANIFEST_SECTION_PTR ManifestSectionPtr;
	uint32 i;
	uint32 size;
	
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MANIFEST_TAG,0x0a);

	if (pArchive == NULL)
		return ISL_FAIL;

	if (BuildManifestTemplate (pArchive, &Template) != ISL_OK)
		return ISL_FAIL;

	/*
	 * Parse the manifest files to break into various sections.
	 */
	if (isl_ParseMemoryWithTemplate(
							pArchive,
							pArchive->Manifest.Image, 
							Template,
							Results) != ISL_OK)
		return ISL_FAIL;

	/*
	 * Initialize the header info.
	 */
	if (isl_BuildHeaderInfo (&pArchive->Manifest.Header,
									Results,
									pArchive) != ISL_OK)
		return ISL_FAIL;

	/*
	 * Count the number of sections in signer_info.
	 */
	section = Results [SECTION_INDEX];
	numSections = 0;
	while (section) {
		numSections++;
		section = section->next;
	}
	
	pArchive->Manifest.SectionCount = numSections;

	size = (pArchive->Manifest.SectionCount) * sizeof(ISL_MANIFEST_SECTION);
	if ((pArchive->Manifest.Section = isl_AllocateMemory(pArchive, size)) == NULL)
		return ISL_FAIL;


	memset(pArchive->Manifest.Section, 0, size);
	section = Results [SECTION_INDEX];
	if (BuildSectionTemplate (pArchive, &Template) != ISL_OK)
		return ISL_FAIL;
	for(i = 0; i < numSections; i++)
	{	
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MANIFEST_TAG,0x0b);
		/*
		 * Parse each section to break into attributes.
		 */
		if (isl_ParseMemoryWithTemplate(
								pArchive,
								*(ISL_CONST_DATA_PTR)section->ptr,
								Template,
								Results) != ISL_OK)
			return ISL_FAIL;

		ManifestSectionPtr = &(pArchive->Manifest.Section[i]);
		ManifestSectionPtr->Parent = pArchive;
		ManifestSectionPtr->Next = ManifestSectionPtr + 1;
		Results[0] = section;
		if (isl_BuildManifestSection(
					pArchive->Manifest.Section + (i), 
					Results) != ISL_OK)
			return ISL_FAIL;

		section = section->next;
	}
	ManifestSectionPtr->Next = NULL;

	return ISL_OK;
}

static ISL_STATUS
BuildSignableListPath(
	ISL_ARCHIVE_CONTEXT_PTR	Context,
	PK_CONST_DATA_PTR		Filename)
{
    uint8   *ptr = NULL;
    uint8   len = 22;
    
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_ARCHIVE_TAG,0x0f);
	
	/* allocate memory */
	Filename->Data = isl_AllocateMemory(Context, len);
    if (!Filename->Data) {
		REPORT_ALLOCATE_MEMORY_ERROR(Context, len);
	}

    ptr = (uint8 *)Filename->Data;

	/* build up path */
    *(ptr++) = 'M';
    *(ptr++) = 'E';
    *(ptr++) = 'T';
    *(ptr++) = 'A';
    *(ptr++) = '-';
    *(ptr++) = 'I';
    *(ptr++) = 'N';
    *(ptr++) = 'F';
    *(ptr++) = '/';
    *(ptr++) = 'M';
    *(ptr++) = 'A';
    *(ptr++) = 'N';
    *(ptr++) = 'I';
    *(ptr++) = 'F';
    *(ptr++) = 'E';
    *(ptr++) = 'S';
    *(ptr++) = 'T';
    *(ptr++) = '.';
    *(ptr++) = 'M';
    *(ptr++) = 'F';
    Filename->Length = ptr - Filename->Data;
    *(ptr++) = '\0';
    return ISL_OK;
}

/* object methods */
ISL_STATUS 
Archive_InitializeNew(			/* constructor for new archive */
		ISL_ARCHIVE_CONTEXT *result,			/* memory allocated for this instance */
		ISL_ARCHIVE_CONFIG_PTR configContext, 	/* my algorithm<->code extension configuration */
		ISL_CONST_DATA Path) 				/* where this archive is being constructed */
{
	NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_ARCHIVE_TAG,0x03);

	if (NULL == result)
		return ISL_FAIL;

#ifndef EISL
	if (NULL == configContext)
		return ISL_FAIL;
#endif
	
	memset(result, 0, sizeof(ISL_ARCHIVE_CONTEXT));
	result->Config =  configContext;

	if (NULL == Path.Data || Path.Length > 1024) {
#if (!defined(EISL) || defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_ARCHIVE_INITIALIZE_NEW);
		sprintf(gErrBuf, gErrStr, result, configContext, Path.Data);
		isl_ReportError(result, ISL_INVALID_PARAMETER, gErrBuf);
#endif
		return ISL_FAIL;
	}

#ifndef EISL
	result->Methods = &ArchiveMethods;
#endif
	result->PathName.Length = Path.Length;
	result->PathName.Data = isl_CopyMemory(result, (void *)Path.Data, Path.Length);
	if (NULL == result->PathName.Data) {
		REPORT_COPY_MEMORY_ERROR(result, (void *)Path.Data, Path.Length);
		return ISL_FAIL;
	}

	isl_InitHeaderInfo (&result->Manifest.Header);

	return ISL_OK;
}



static
ISL_STATUS
AddOldSignature(
	ISL_ARCHIVE_CONTEXT_PTR Context,
	ISL_CONST_DATA Name,
	ISL_CONST_DATA InMemoryImage)
{
	ISL_STATUS status;
	ISL_SIGNATURE_CONTEXT   *pSigContext = NULL;
	ISL_SIGNATURE_LIST		*pSignatureList = NULL;
	ISL_SIGNATURE_CONFIG    *pSigConfig = NULL;

	if (NULL == Context)
		return ISL_FAIL;

#ifdef EISL
	pSigContext = isl_AllocateMemory(Context, sizeof(ISL_SIGNATURE_CONTEXT));
    if (!pSigContext) {
		REPORT_ALLOCATE_MEMORY_ERROR(Context, sizeof(ISL_SIGNATURE_CONTEXT));
		return ISL_FAIL;
	}
	status = isl_InitializeSignatureFromImage(pSigContext,
						   					  pSigConfig,
						 					  Context,
											  Name,
											  InMemoryImage);
#else
	if (NULL == Context->Config)
		return ISL_FAIL;

 	pSigConfig = Context->Config->SignatureConfig;
    pSigContext = isl_AllocateMemory(Context, 
                                     PKCS7SignatureMethods.SizeofObject());
    if (!pSigContext) {
		REPORT_ALLOCATE_MEMORY_ERROR(Context, 
                                     PKCS7SignatureMethods.SizeofObject());
		return ISL_FAIL;
	}
	status = PKCS7SignatureMethods.InitializeFromImage(pSigContext,
													   pSigConfig,
													   Context,
													   Name,
													   InMemoryImage);
#endif
	if (status != ISL_OK) {
#if (!defined(EISL) || defined(DEBUG)) 
		gErrStr = isl_GetErrorStr(ISL_SIGNATURE_INITIALIZE_FROM_IMAGE);
		sprintf(gErrBuf, gErrStr, pSigContext, pSigConfig, Context, Name, InMemoryImage);
		isl_ReportError(Context, ISL_NO_MEMORY, gErrBuf);
#endif
		return ISL_FAIL;
	}
	pSignatureList = isl_AllocateMemory(Context, sizeof(ISL_SIGNATURE_LIST));
	if (!pSignatureList) {
		REPORT_ALLOCATE_MEMORY_ERROR(Context, sizeof(ISL_SIGNATURE_LIST));
		return ISL_FAIL;
	}
	pSignatureList->Signature = pSigContext;
	pSignatureList->Next = Context->Signatures;

	pSignatureList->Name.Data = isl_AllocateMemory(Context, Name.Length);
	if (!pSignatureList->Name.Data){ 
		REPORT_ALLOCATE_MEMORY_ERROR(Context, pSignatureList->Name.Length);
		return ISL_FAIL;
	}
	memcpy((void*)pSignatureList->Name.Data, (void*)Name.Data, Name.Length);
	pSignatureList->Name.Length = Name.Length;
	Context->Signatures = pSignatureList;
    return ISL_OK;

}

static
ISL_STATUS
AddOldCertificate(
	ISL_ARCHIVE_CONTEXT_PTR Context,
	ISL_CONST_DATA Name,
	ISL_CONST_DATA InMemoryImage)
{
	ISL_CERTIFICATE			*pCert;
	ISL_CERTIFICATE_LIST	*pCertList;

	if (NULL == Context)
		return ISL_FAIL;

	pCert = isl_AllocateMemory(Context, sizeof(ISL_CERTIFICATE));
	if (!pCert) {
		REPORT_ALLOCATE_MEMORY_ERROR(Context, sizeof(ISL_CERTIFICATE));
		return ISL_FAIL;
	}
#ifdef EISL
    if (ISL_OK != isl_X509Cert_Initialize(
				pCert,
                NULL,
                NULL,
                InMemoryImage))
#else
	if (NULL == Context->Config)
		return ISL_FAIL;
    if (ISL_OK != X509CertMethods.Initialize(
                pCert,
                Context->Config->MemoryContext,
                Context->Config->MemoryMethods,
                InMemoryImage) )
#endif
	{
#if (defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_CERT_INITIALIZE);
		sprintf(gErrBuf, 
				gErrStr, 
				pCert,
			    NULL,
				NULL,
				InMemoryImage);
		isl_ReportError(Context, ISL_INVALID_CERTIFICATE, gErrBuf);
#elif (!defined(EISL))
		gErrStr = isl_GetErrorStr(ISL_CERT_INITIALIZE);
		sprintf(gErrBuf, 
				gErrStr, 
				pCert,
				Context->Config->MemoryContext,
				Context->Config->MemoryMethods,
				InMemoryImage);
		isl_ReportError(Context, ISL_INVALID_CERTIFICATE, gErrBuf);
#endif
		return ISL_FAIL;
	}

    pCertList = isl_AllocateMemory(Context, sizeof(ISL_CERTIFICATE_LIST));
    if (!pCertList) {
		REPORT_ALLOCATE_MEMORY_ERROR(Context, sizeof(ISL_CERTIFICATE_LIST));
		return ISL_FAIL;
	}
    pCertList->Cert = pCert;
    pCertList->Next = Context->Certs;
	pCertList->Name.Data = isl_AllocateMemory(Context, Name.Length);
	if (!pCertList->Name.Data){ 
		REPORT_ALLOCATE_MEMORY_ERROR(Context, Name.Length);
		return ISL_FAIL;
	}
	memcpy((void*)pCertList->Name.Data, (void*)Name.Data, Name.Length);
	pCertList->Name.Length = Name.Length;
    Context->Certs = pCertList;

	return ISL_OK;
}
#ifndef VMS
#pragma warning (disable : 4102)
#endif
ISL_STATUS 
InitializeFromMemory(		            /* constructor for existing archive */
		ISL_ARCHIVE_CONTEXT *result,			/* memory allocated for this instance */
		ISL_ARCHIVE_CONFIG_PTR configContext, 	/* my algorithm<->code extension configuration */
		ISL_CONST_DATA FileImage)				/* buffer containing the credentials */
{   
	PK_ARCHIVE_PTR		    pPKArchive = NULL;
	PK_ITERATOR_PTR		    pkIterator = NULL;
	PK_CONST_DATA           Filename;
	PK_CONST_DATA           Filedata;
	uint8					dsa[3];
	uint8					cer[3];

	//configContext;

#ifndef EISL
	uint8					rsa[3];
	rsa[0] = 'R'; rsa[1] = 'S'; rsa[2] = 'A';
#endif

	NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_ARCHIVE_TAG,0x04);
	
	if (NULL == result) 
		return ISL_FAIL;
	
#ifndef EISL
	if (NULL == configContext || NULL == configContext->SignatureConfig)
		return ISL_FAIL;
#endif

	dsa[0] = 'D'; dsa[1] = 'S'; dsa[2] = 'A';
	cer[0] = 'C'; cer[1] = 'E'; cer[2] = 'R';
	
	
	/*
	 * Unzip the manifest archive.
	 */
	Filedata.Data = FileImage.Data;
	Filedata.Length = FileImage.Length;
	pPKArchive = pk_InitializeFromMemory (result, Filedata);
	if (NULL == pPKArchive) {
#if (!defined(EISL) || defined(DEBUG))
		gErrStr = isl_GetErrorStr(PK_LOAD_ARCHIVE);
		sprintf(gErrBuf, gErrStr, result, Filedata.Data);
		isl_ReportError(result, ISL_IO_ERROR, gErrBuf);
#endif
		return ISL_FAIL;
	}

	/* 
	 * load the manifest file 
	 */
	if (ISL_OK != BuildSignableListPath(result, &Filename))
	{
#if (!defined(EISL) || defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_BUILD_SIGNABLE_LIST_PATH);
		sprintf(gErrBuf, gErrStr, result, Filename.Data);
		isl_ReportError(result, ISL_INVALID_INPUT, gErrBuf);
#endif
		return ISL_FAIL;
	}

	if (PK_OK != pk_FindFile(pPKArchive, Filename, &Filedata))
	{
#if (!defined(EISL) || defined(DEBUG))
		gErrStr = isl_GetErrorStr(PK_FIND_FILE);
		sprintf(gErrBuf, gErrStr, pPKArchive, Filename.Data, &Filedata);
		isl_ReportError(result, ISL_IO_ERROR, gErrBuf);
#endif
		return ISL_FAIL;
	}
	/* parse the manifest file */
	result->Manifest.Image.Data = Filedata.Data;
	result->Manifest.Image.Length = Filedata.Length;
	if (ISL_OK != BuildManifest(result))
	{
#if (!defined(EISL) || defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_BUILD_MANIFEST);
		sprintf(gErrBuf, gErrStr, result);
		isl_ReportError(result, ISL_INVALID_INPUT, gErrBuf);
#endif
		return ISL_FAIL;
	}

	/* parse the dsa files */
	pkIterator = pk_CreateFileEnumerator(pPKArchive);
	if (!pkIterator) {
#if (!defined(EISL) || defined(DEBUG)) 
		gErrStr = isl_GetErrorStr(PK_CREATE_FILE_ENUM);
		sprintf(gErrBuf, gErrStr, pPKArchive);
		isl_ReportError(result, ISL_IO_ERROR, gErrBuf);
#endif
		return ISL_FAIL;
	}

	while (*pkIterator)
	{
        ISL_CONST_DATA          FilePath;
        ISL_CONST_DATA          FileName;
        ISL_CONST_DATA          FileTitle;
        ISL_CONST_DATA          FileExt;
        ISL_CONST_DATA          Image;

		NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_ARCHIVE_TAG,0x06);
		if (PK_OK != pk_GetNextFile(pkIterator, &Filename, &Filedata))
		{
#if (!defined(EISL) || defined(DEBUG)) 
			gErrStr = isl_GetErrorStr(PK_GET_NEXT_FILE);
			sprintf(gErrBuf, gErrStr, pkIterator, &Filename, &Filedata);
			isl_ReportError(result, ISL_IO_ERROR, gErrBuf);
#endif
			(void)pk_RecycleFileEnumerator(pkIterator);
			return ISL_FAIL;
		}
		ASSERT(pkIterator);
        FilePath.Data = Filename.Data;
        FilePath.Length = Filename.Length;
		if (ISL_OK != isl_ParseFilePath(FilePath, NULL, &FileName, &FileTitle, &FileExt))
		{
#if (!defined(EISL) || defined(DEBUG))
			gErrStr = isl_GetErrorStr(ISL_PARSE_FILE_PATH);
			sprintf(gErrBuf, gErrStr, FilePath.Data, NULL, &FileName, &FileTitle, &FileExt);
			isl_ReportError(result, ISL_INVALID_INPUT, gErrBuf);
#endif
			(void)pk_RecycleFileEnumerator(pkIterator);
			return ISL_FAIL;
		}
		Image.Data = Filedata.Data;
        Image.Length = Filedata.Length;

        if (
                (FileExt.Length == 3) &&
			    (
#ifndef EISL
                    (memcmp(FileExt.Data,rsa,3) == 0) ||
#endif
                    (memcmp(FileExt.Data,dsa,3) == 0)
               )
		   ) {
			if ((AddOldSignature(
						 result,
						 FileTitle,
						 Image)) != ISL_OK) {
				(void)pk_RecycleFileEnumerator(pkIterator);
				return ISL_FAIL; 
			}
        } else if (
                    (FileExt.Length == 3) &&
                    (memcmp(FileExt.Data,cer,3) == 0)
				  ) {
            
			if ((AddOldCertificate(
					 result,
					 FileTitle,
					 Image)) != ISL_OK) {
				(void)pk_RecycleFileEnumerator(pkIterator);
				return ISL_FAIL; 
			}
		}
       
		continue;
	}

	(void)pk_RecycleFileEnumerator(pkIterator);
	return ISL_OK;
}


/*-----------------------------------------------------------------------------
 * Class: Archive
 * Name:  Archive_InitializeFromMemory
 *
 * Description:
 * Constructor for existing in-memory archive
 *
 * Parameters: 
 * Context (input/output)   : this pointer
 * Config (input)   : algorithm<->code extension
 * Path (input)   :		path to the external representation of manifest
 *
 * Return value:
 * status of operation
 * 
 * Error Codes:
 *
 * Notes:
 * InMemoryImage is assumed to be persistent for the life of the archive object
 *---------------------------------------------------------------------------*/
ISL_STATUS 
Archive_InitializeFromMemory(		            /* constructor for existing archive */
		ISL_ARCHIVE_CONTEXT *result,			/* memory allocated for this instance */
		ISL_ARCHIVE_CONFIG_PTR configContext, 	/* my algorithm<->code extension configuration */
		ISL_CONST_DATA FileImage,				/* buffer containing the credentials */
		ISL_CONST_DATA SearchPath) 				/* path for disk directory/file of external representation */
{
	ISL_CONST_DATA localFileImage;

	NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_ARCHIVE_TAG,0x04);
	
	if (NULL == result) 
		return ISL_FAIL;
	
#ifndef EISL
	if (NULL == configContext || NULL == configContext->SignatureConfig)
		return ISL_FAIL;
#endif

	memset(result, 0, sizeof(ISL_ARCHIVE_CONTEXT));
	result->Config =  configContext;

#ifndef EISL
	result->Methods = &ArchiveMethods;
#endif
	result->PathName.Length = 0;
	result->PathName.Data = NULL;

	if ((SearchPath.Length == 0) || (SearchPath.Data == NULL))
		return ISL_FAIL;

	result->SearchPath.Length = SearchPath.Length;
	result->SearchPath.Data = isl_CopyMemory(result, (void *)SearchPath.Data, SearchPath.Length);
	if (NULL == result->SearchPath.Data) {
		REPORT_COPY_MEMORY_ERROR(result, (void *)SearchPath.Data, SearchPath.Length);
		return ISL_FAIL;
	}

	isl_InitHeaderInfo (&result->Manifest.Header);

	localFileImage.Length = FileImage.Length;
	localFileImage.Data = isl_CopyMemory(result, (void *)FileImage.Data, FileImage.Length);
	
	return InitializeFromMemory (result, configContext, localFileImage);

}
static ISL_STATUS
GetImageAndPathFromCredentials(
	ISL_ARCHIVE_CONTEXT_PTR pArchive, 
	ISL_CONST_DATA			Credentials,
	ISL_CONST_DATA_PTR	    Fileimage,
	ISL_CONST_DATA_PTR		SearchPath)
	
{
	const uint8				*ptr;

	NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_ARCHIVE_TAG,0x05);
	ASSERT(pArchive);

	
	if (ISL_OK != isl_LoadFile(pArchive, Credentials, Fileimage, 0) ||
											!Fileimage->Length )
		return ISL_FAIL;

	/*
	 * get the search path from the path to the manifest.
	 */
	ptr = Credentials.Data + Credentials.Length;
	while ((ptr > Credentials.Data) &&				/* input to scan */
			!((*ptr == '/') || (*ptr == '\\'))) 	/* not a slash */
		ptr--;

	SearchPath->Data = Credentials.Data;
	SearchPath->Length = (ptr - Credentials.Data);
	
	return ISL_OK;
}


/*-----------------------------------------------------------------------------
 * Class: Archive
 * Name:  Archive_InitializeFromPath
 *
 * Description:
 * Constructor for existing in-memory archive
 *
 * Parameters: 
 * Context (input/output)   : this pointer
 * Config (input)   : algorithm<->code extension
 * Path (input)   :		path to the external representation of manifest
 *
 * Return value:
 * status of operation
 * 
 * Error Codes:
 *
 * Notes:
 * InMemoryImage is assumed to be persistent for the life of the archive object
 *---------------------------------------------------------------------------*/
ISL_STATUS 
Archive_InitializeFromPath(		                        /* constructor for existing archive */
		ISL_ARCHIVE_CONTEXT *result,			/* memory allocated for this instance */
		ISL_ARCHIVE_CONFIG_PTR configContext, 	/* my algorithm<->code extension configuration */
		ISL_CONST_DATA Path) 				/* disk directory/file of external representation */
{
	ISL_CONST_DATA		    Fileimage;
	ISL_CONST_DATA			SearchPath;

	NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_ARCHIVE_TAG,0x04);
	
	if (NULL == result) 
		return ISL_FAIL;

#ifndef EISL
	if (NULL == configContext || NULL == configContext->SignatureConfig)
		return ISL_FAIL;
#endif
	
	if (ISL_OK != Archive_InitializeNew(result, configContext, Path)) {
#if (!defined(EISL) || defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_ARCHIVE_INITIALIZE_NEW);
		sprintf(gErrBuf, gErrStr, Path.Data, Path.Length);
		isl_ReportError(result, ISL_INVALID_INPUT, gErrBuf);
#endif
		return ISL_FAIL;
	}

	if (GetImageAndPathFromCredentials(
								result, 
								Path,
								&Fileimage,
								&SearchPath) != ISL_OK)
		return ISL_FAIL;

	/*
	 * Setup the search path in archive context.
	 */
	result->SearchPath.Length = SearchPath.Length;
	result->SearchPath.Data = isl_CopyMemory(result, (void *)SearchPath.Data, SearchPath.Length);
	if (NULL == result->SearchPath.Data) {
		REPORT_COPY_MEMORY_ERROR(result, (void *)SearchPath.Data, SearchPath.Length);
		return ISL_FAIL;
	}

	return InitializeFromMemory (
							result,
							configContext,
							Fileimage);
}
#ifndef VMS
#pragma warning (default : 4102)
#endif



static ISL_STATUS
RecycleCertList(ISL_CERTIFICATE_LIST *CertList)
{
	ISL_CERTIFICATE_LIST  *Item;
	ISL_STATUS status;

	status = ISL_OK;
	for(Item = CertList;
		Item;
		Item = Item->Next)
	{
#ifdef EISL
        if (ISL_OK != isl_X509Cert_Recycle(Item->Cert))
#else
        if (ISL_OK != X509CertMethods.Recycle(Item->Cert))
#endif
		{
			status = ISL_FAIL;
		}
	}
	return status;
}

ISL_STATUS
isl_recycleSignature (ISL_ARCHIVE_CONTEXT_PTR pArchive,
					  ISL_SIGNATURE_CONTEXT_PTR pSignatureContext)

{
	ISL_SIGNATURE_CONTEXT_PTR pTmpSigContext;
	ISL_SIGNATURE_LIST_PTR pSigList;
	ISL_SIGNATURE_LIST_PTR SavedpSigList = NULL;

	pSigList = pArchive->Signatures;

	while (pSigList) {
		pTmpSigContext = pSigList->Signature;
		if (pSignatureContext != pTmpSigContext) {
				/* Recycle Signature Certificate Objects */
			if (ISL_OK != RecycleCertList(pTmpSigContext->Certificates))
				return ISL_FAIL;
		} else {
			SavedpSigList = pSigList;
		}
		pSigList = pSigList->Next;
	}
	pArchive->Signatures = SavedpSigList;
	pArchive->Signatures->Next = NULL;
	return ISL_OK;
}


ISL_STATUS
isl_RecycleArchiveContext(
	ISL_ARCHIVE_CONTEXT *Context)
{
	ISL_MEMORY_BUFFER_PTR  item;
	ISL_MEMORY_BUFFER_PTR  temp;
	ISL_SIGNATURE_LIST_PTR Signatures;
	ISL_STATUS status;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_ARCHIVE_TAG,0x01);
	if (!Context) return ISL_FAIL;

	status = ISL_OK;
	/* Recycle Archive Certificate Objects */
	if (ISL_OK != RecycleCertList(Context->Certs))
	{
		status = ISL_FAIL;
	}

	/* Recycle Signature Certificate Objects */
	for(Signatures = Context->Signatures;
		Signatures;
		Signatures = Signatures->Next)
	{
		if (ISL_OK != RecycleCertList(Signatures->Signature->Certificates))
		{
			status = ISL_FAIL;
		}
	}

	item = Context->Buffers;
	while(item)
	{
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_ARCHIVE_TAG,0x02);
#ifdef EISL
		free(item->Buffer);					/* free buffer */
#else
		ISL_FREE_WRAPPER(
			Context->Config->MemoryMethods->Free,
			Context->Config->MemoryContext,
			item->Buffer);
#endif
		temp = item;						/* temp pointer to node */
		item = item->Next;					/* get next node pointer */
#ifdef EISL
		free(temp);							/* now free old node pointer */
#else
		ISL_FREE_WRAPPER(
			Context->Config->MemoryMethods->Free,
			Context->Config->MemoryContext,
			temp);
#endif
	}

	memset(Context, 0, sizeof(ISL_ARCHIVE_CONTEXT));	/* zero out memory */
	return status;
}
