/*-----------------------------------------------------------------------
 * File: EISLREFLIST.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/*
**	ISL_SIG_SECTION_LIST class public methods (EISL/ISL common routines)
*/

#include <stdio.h>
#include "isl_internal.h"
#include "islutil.h"
#include <isl_memtrack.h>


#ifdef EISL
static ISL_STATUS
VerifyManSect(
	ISL_SIG_SECTION_PTR pSigSect)
{
	EICLSHAState			state;
	EICLSHADigest		digest; 
	EICLData				message;
	ISL_ALG_INFO_PTR	pNode = NULL;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_REFLIST_TAG, 0x14);
	pNode = isl_FindSHA(pSigSect->SectionInfo.Algorithm);
	if (!pNode)
		return ISL_FAIL;

	message.value = (EICLByte *)pSigSect->ManSect->SectionInfo.Image.Data;
	message.length = (long) pSigSect->ManSect->SectionInfo.Image.Length;

	(void)EICL_SHABegin(&state);
	(void)EICL_SHAProcess(&message, &state);
	(void)EICL_SHAEnd(&state, digest);

	if (pNode->Hash->Value.Length != 20 ||
		memcmp(pNode->Hash->Value.Data, digest, 20) != 0)
	{
		return ISL_FAIL;
	}
	
	return ISL_OK;
}
#endif

/*
 * EISL-ISL Common functions.
 */

/*
 * isl_InitializeSignedListFromImage
 *
 * This function parses a verified signature file (SF) image.
 * As signature sections are parsed, the corresponding manifest sections
 * are verified by computing a hash of the manifest section and comparing it
 * to the hash value in the signature section of that manifest section.
 */

#ifndef EISL
extern ISL_SIG_SECTION_METHODS SignedListMethods;
extern ISL_STATUS VerifyManSectWithAllDigestValues(
				ISL_SIG_SECTION_PTR pSigSect);
#endif

static
ISL_STATUS
BuildSignerInfoTemplate(
	ISL_ARCHIVE_CONTEXT_PTR	Context,
	ISL_CONST_DATA_PTR	Template)
{
    uint8   *ptr = NULL;
    uint8   len = 14;
    
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_ARCHIVE_TAG,0x0f);
	
	/* allocate memory */
	Template->Data = isl_AllocateMemory(Context, len);
    if (!Template->Data) {
		REPORT_ALLOCATE_MEMORY_ERROR(Context, len);
	}

    ptr = (uint8 *)Template->Data;

	/* build up template */
    *(ptr++) = '<';
    *(ptr++) = 's';
    *(ptr++) = 'i';
    *(ptr++) = 'g';
    *(ptr++) = 'n';
    *(ptr++) = 'e';
    *(ptr++) = 'r';
    *(ptr++) = '_';
    *(ptr++) = 'i';
    *(ptr++) = 'n';
	*(ptr++) = 'f';
	*(ptr++) = 'o';
	*(ptr++) = '>';

    Template->Length = ptr - Template->Data;
    *(ptr++) = '\0';
    return ISL_OK;
}

ISL_STATUS isl_InitializeSignedListFromImage(	/* constructor for existing object list */
		ISL_SIG_SECTION_LIST *memory,			/* memory allocated for this instance */
		ISL_SIGNATURE_CONTEXT *Context,		    /* archive which contains signature */
		ISL_CONST_DATA Name,					/* disk file of external representation */
        ISL_CONST_DATA Image)
{
	uint32 i;
	ISL_ARCHIVE_CONTEXT_PTR ArchivePtr;
	ISL_SIG_SECTION_PTR		sigSectionPtr;
	ISL_CONST_DATA Template;
	ISL_DATA_CHAIN_PTR Results[MANIFEST_LAST_INDEX];
	ISL_DATA_CHAIN_PTR section;
	uint32 numSections;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_REFLIST_TAG,0x02);
    ASSERT(Context);

	if (!memory || !Context || !Context->Archive) 
		return ISL_FAIL;

    ArchivePtr = Context->Archive,
    memset(memory, 0, sizeof(ISL_SIG_SECTION_LIST));
  	memory->Parent = Context;

	if (!Name.Data || !Image.Data) {
#if (!defined(EISL) || defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_SIGNED_LIST_INITIALIZE_FROM_IMAGE);
		sprintf(gErrBuf, gErrStr, Context, Name.Data, Image.Data);
		isl_ReportError(Context->Archive, ISL_INVALID_PARAMETER, gErrBuf);
#endif
		return ISL_FAIL;
	}

    /* Initializing the Signed List Header Information */
	memory->Name.Length = Name.Length;
	memory->Name.Data = isl_CopyMemory(ArchivePtr, 
									   (void*)Name.Data,
									   Name.Length);
	if (!memory->Name.Data) {
		REPORT_COPY_MEMORY_ERROR(ArchivePtr, (void*)Name.Data, Name.Length);
	}
    memory->Image.Length = Image.Length;
    memory->Image.Data = isl_CopyMemory(ArchivePtr,
                                        (void*)Image.Data,
                                        Image.Length);
	if (!memory->Image.Data) {
		REPORT_COPY_MEMORY_ERROR(ArchivePtr, (void*)Image.Data, Image.Length);
	}

	if (BuildSignerInfoTemplate (ArchivePtr, &Template) != ISL_OK)
		return ISL_FAIL;

	/*
	 * Parse the signer_info file to break into various sections.
	 */
	if (isl_ParseMemoryWithTemplate(
							ArchivePtr,
							memory->Image, 
							Template,
							Results) != ISL_OK)
		return ISL_FAIL;

	/*
	 * Initialize the header info from the first section.
	 */
	if (isl_BuildHeaderInfo (&memory->HeaderInfo,
									Results,
									ArchivePtr) != ISL_OK) {
#if (!defined(EISL) || defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_GET_SIGNED_LIST_HEADER);
		sprintf(gErrBuf, gErrStr, Context, Results, memory);
		isl_ReportError(Context->Archive, ISL_INVALID_SIGNATURE_FILE, gErrBuf);
#endif
		return ISL_FAIL;
	}

	/*
	 * Count the number of sections in signer_info.
	 */
	section = Results [SECTION_INDEX];
	numSections = 0;
	while (section) {
		numSections++;
		section = section->next;
	}

	memory->SectionCount = numSections;
	memory->Section = isl_AllocateMemory(
						ArchivePtr, 
						memory->SectionCount * sizeof(ISL_MANIFEST_SECTION));
	
	if (memory->Section == NULL) 
		return ISL_FAIL;

	memset(memory->Section, 0, 
				memory->SectionCount * sizeof(ISL_MANIFEST_SECTION));

	/* Parsing the Signed List Sections */
	section = Results [SECTION_INDEX];
	if (BuildSectionTemplate (ArchivePtr, &Template) != ISL_OK)
		return ISL_FAIL;

	for(i=0; i < memory->SectionCount; i++)
	{	
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MANIFEST_TAG,0x0b);

		/*
		 * Parse each section to break into attributes.
		 */
		if (isl_ParseMemoryWithTemplate(
								ArchivePtr,
								*(ISL_CONST_DATA_PTR)section->ptr,
								Template,
								Results) != ISL_OK)
			return ISL_FAIL;

		sigSectionPtr = memory->Section + i;
		sigSectionPtr->Parent = Context;
#ifndef EISL
		sigSectionPtr->Methods = &SignedListMethods;
#endif
		Results[0] = section;
		if (isl_BuildSigSection(
				sigSectionPtr, 
				Results) != ISL_OK)
			return ISL_FAIL;
		sigSectionPtr->Next = sigSectionPtr + 1;
			
		
		/* Verify the Corresponding Manifest Section */
#ifdef EISL
		/* verify "SHA" hash value only because EISL only support SHA algorithm */
		if (VerifyManSect(sigSectionPtr) != ISL_OK)
			return ISL_FAIL;
#else
		/* verify all hash value in the signature section */
		if (VerifyManSectWithAllDigestValues(sigSectionPtr) != ISL_OK) {
#if (!defined(EISL) || defined(DEBUG))
			gErrStr = isl_GetErrorStr(ISL_BUILD_SIGNED_OBJECT);
			sprintf(gErrBuf, gErrStr, Context, sigSectionPtr);
			isl_ReportError(Context->Archive, ISL_FAILED_TO_VERIFY_MANIFEST_SECTION, gErrBuf);
#endif
			return ISL_FAIL;
		}
#endif
		section = section->next;
	

	}
	sigSectionPtr->Next = NULL;

	return ISL_OK;
}

