/*-----------------------------------------------------------------------
 * File: PARSER.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */



#include "isl_internal.h"
#include <isl_memtrack.h>

struct parseEl;

struct va_st {
  int lastFoff;
  union {
    long foff;
    long length;
    struct parseEl *pNode;
    int var;
    int byte;
  } u;
  char type;
};
    
typedef struct va_st stackEl;

/* the stack types: we can have 256...but I doubt if we'll break 8 */
#define END       0
#define VAR_BEGIN 1
#define VAR_END   2
#define VAR_DEPTH 3
#define BRANCH    4
#define FOFFSET   5
#define REGSTACK  6
#define BBYTE     7
#define SFBYTE    8
#define VERROR    9

/* parse tree nodes */
#define NOP       0
#define EXACT     1
#define ICASE     2
#define VAR_B     3
#define VAR_E     4
#define RANGE     5
#define REG_B     6
#define REG_E     7
#define FBYTE     8
#define PEND      9
#define PERROR   10

#define MAXNAME 16

#define PARSE_OK 1
#define PARSE_FAIL 0

#define ESCAPE_BLOCK 32
#define REG_STACK_DEPTH 512

struct parseEl {
  char  type;
#ifndef NODBG
  char  visited;
  int id;
#endif
  union {
    /* for REG_B, REG_E */
    char name[MAXNAME];
    /* for VAR_B, VAR_E */
    int var;
    /* for EXACT */
    int byte;
    /* for ICASE */
    char iBytes[2];
    /* for FBYTE */
    char *fbytes;
    /* for range */
    unsigned long *rbits;
  } u;
  /* for FBYTE */
  long nextByte;
  struct parseEl *choice0;
  struct parseEl *choice1;
};

/*
 * We're going to go with the assumption that a depth of 256 is *plenty* 
 * for this.  If this assumption fails, this can be bumped up.
 */
struct regStack {
  struct parseEl *stack[REG_STACK_DEPTH];
  int regEnd;
};

/* some stuff to unglop the "special" chars */
#define BSLASH 0x5c
#define SQUOTE 0x27
#define DQUOTE 0x22

#if defined (LINE_MAX) && defined (VMS)
#undef LINE_MAX
#endif
#define LINE_MAX 512

#ifdef VMS
#pragma intrinsic(strlen,memset,strcpy,memcpy)
#else
#pragma intrinsic(strlen,strcmp,memcmp,memset,strcpy,memcpy)
#endif

struct escape {
  char name[MAXNAME];
  struct parseEl *tree;
  int isLambda;
};

struct escapeTable {
  struct escape *table;
  int escapesUsed;
  int escapesAvail;
};

struct valueStack {
  stackEl *stack;
  int valTop;
  int lastFoff;
};

struct contentData {
  ISL_CONST_DATA_PTR ptr;
  unsigned long offset;
};

/* forward decls */
static int
parseRegString(ISL_MEMORY_BUFFER_PTR *mem,
	       char **s, 
#ifndef NODBG
	       int debugLevel,
	       char origLine[LINE_MAX],
	       char **defPtr,
	       int *id,
#endif
	       struct escapeTable *esc,
	       struct parseEl **head, 
	       struct parseEl **end, 
	       int inAlt,
	       int iCase,
	       int *isLambda);
#ifndef NODBG
static int dumpEscape();
#endif
static void clearVisited();

#define valPop(X) ((X)->stack[--((X)->valTop)])
#define valPPop(X) (&((X)->stack[--((X)->valTop)]))
#define valPush(X,Y) (Y)->stack[(Y)->valTop++] = (X)

#define fakeGetc(X) \
   ((X)->offset >= (X)->ptr->Length ? EOF : (X)->ptr->Data[(X)->offset++])
#define fakeFtell(X) (X)->offset
#define fakeFseek(X,Y) (X)->offset = (Y)


static void
recycleMemory(ISL_MEMORY_BUFFER_PTR mem)
{
  ISL_MEMORY_BUFFER_PTR memnext;
  while (mem){
    memnext = mem->Next;
    free(mem->Buffer);
    free(mem);
    mem = memnext;
  }
}

static void
*allocateMemory(ISL_MEMORY_BUFFER_PTR *mem, size_t size)
{
  ISL_MEMORY_BUFFER_PTR pMemNode = NULL;
  assert(mem);

  pMemNode = (ISL_MEMORY_BUFFER_PTR) malloc(sizeof(ISL_MEMORY_BUFFER));
  if (!pMemNode)
    return NULL;
  pMemNode->Buffer = malloc(size);
  if (!pMemNode->Buffer){
    free(pMemNode);
    return NULL;
  }
  pMemNode->Next = *mem;
  *mem = pMemNode;
  return pMemNode->Buffer;
}


static int
newParseEl(ISL_MEMORY_BUFFER_PTR *mem,
#ifndef NODBG
	   int *id,
#endif
	   struct parseEl **ret
	   )
{
  assert(ret);
#ifndef NODBG
  assert(id);
#endif

  *ret = (struct parseEl *) 
    allocateMemory(mem, (unsigned)(sizeof(struct parseEl)));
  assert(*ret);

  (*ret)->type = PERROR;
  (*ret)->choice0 = (*ret)->choice1 = NULL;

#ifndef NODBG
  (*ret)->visited = 0;
  (*ret)->id = (*id)++;
#endif
  return PARSE_OK;
}

static struct parseEl
*lookupEscape(char *s,
	      struct escapeTable *esc)
{
  /* returns the tree for name, NULL on failure */
  if (s == NULL || esc == NULL || esc->table == NULL){
    return(NULL);
  } else {
    int i;
    for (i = 0; i < esc->escapesUsed; i++){
      if (!(strcmp(s,esc->table[i].name))){
	return(esc->table[i].tree);
      }
    }
  }
  return(NULL);
}

static int
escapeIsLambda(char *s,
	       struct escapeTable *esc)
{
  /* returns 1 if escape s is defined and can evaluate to an empty
   * expression.
   */
  if (s == NULL || esc == NULL || esc->table == NULL){
    return(0);
  } else {
    int i;
    for (i = 0; i < esc->escapesUsed; i++){
      if (!(strcmp(s,esc->table[i].name))){
	return(esc->table[i].isLambda);
      }
    }
  }
  return(0);
}

static int
addEscape(ISL_MEMORY_BUFFER_PTR *mem,
	  char *s,
#ifndef NODBG
	  int debugLevel,
#endif
	  struct escapeTable *esc,
	  struct parseEl *tree,
	  int isLambda)
{
#ifndef NODBG
  struct parseEl *ptr;
#endif
  assert(s);
  assert(esc);

#ifndef NODBG
  /* make sure not already defined */
  ptr = lookupEscape(s,esc);
  assert(!ptr);
#endif

  assert(tree);
  assert((strlen(s)) < MAXNAME);

  if (esc->escapesUsed >= esc->escapesAvail){
    /* need more space */
    struct escape *newEsc;
    unsigned size, osize;
    osize = esc->escapesAvail * sizeof(struct parseEl);
    esc->escapesAvail += ESCAPE_BLOCK;
    size = esc->escapesAvail * sizeof(struct parseEl);
    newEsc = (struct escape *) allocateMemory(mem,size);
    assert(newEsc);

    if (esc->table){
      memcpy(newEsc,esc->table,osize);
    }
    esc->table = newEsc;
  }
  strcpy(esc->table[esc->escapesUsed].name,s);
  esc->table[esc->escapesUsed].isLambda = isLambda;
  esc->table[esc->escapesUsed++].tree = tree;
  return PARSE_OK;
}


/**************************************************************************/
/*****************************  line acquisition  *************************/
/**************************************************************************/
static int
getline(struct contentData *cD, 
#ifndef NODBG
	int debugLevel,
#endif
	char *buff,
	char **ret)
{
  /* get a new line from "cD".  The line we
   * return will be one that must be parsed, and well-formed.
   */
  int i, ch;
  i = 0;
  ch = ' ';
  while ((ch = fakeGetc(cD)) != EOF){ 
    if (ch != EOF){
      if (ch == '\r' || ch == '\n'){
	if (ch == '\r'){
	  int offset;
	  /* remember where we were */
	  offset = fakeFtell(cD);
	  ch = fakeGetc(cD);
	  if (ch != '\n'){
	    /* push ch back into the "stream" */
	    fakeFseek(cD,offset-1);
	  }
	}
	/* a completed line \r,\n, or \r\n */
	/* if next to last char is ^A, put the null there, not after the
	 * newline. This is to allow a macro def that is totally optional
	 * to be placed in the body section.  Otherwise, the newline that
	 * follows it bites us.
	 */
	if (i > 0 && buff[i-1] == 1){
	  buff[i-1] = '\0';
	} else {
	  buff[i++] = (char) ch;
	  buff[i] = '\0';
	}
	*ret = buff;
#ifndef NODBG
	if (debugLevel > 2){
	  fprintf(stdout,"DEBUG: TEMPFILE: %s\n", buff);
	}
#endif
	return PARSE_OK;
      } else {
	buff[i++] = (char) ch;
      }
    }
  }
  if (i > 0){
    /* ran out of file before newline.  Null-terminate and hope for best */
    buff[i] = '\0';
    *ret = buff;
  } else {
    *ret = NULL;
  }
#ifndef NODBG
  if (debugLevel > 2){
    fprintf(stdout,"DEBUG: TEMPFILE: %s\n", buff);
  }
#endif
  return PARSE_OK;
}

static int
addByte(ISL_MEMORY_BUFFER_PTR *mem,
	struct parseEl *p,
	char c,
	int l)
{
  assert(p);

  if (p->u.fbytes == NULL){
    p->nextByte = 0;
    p->u.fbytes = (char *) allocateMemory(mem,(unsigned)l);
    assert(p->u.fbytes);
  }
  p->u.fbytes[p->nextByte++] = c;
  return PARSE_OK;
}

#define SETBITS 	(CHAR_BIT * sizeof(unsigned long))
#define word(e)		((e)/SETBITS)
#define bit(e)		((e)%SETBITS)
#define ShiftBit(e)	(1UL << (e))

/* Create a Set with room for 128 elements */
static int
CreateSet(ISL_MEMORY_BUFFER_PTR *mem,
	  unsigned long **set)    
{
  assert(set);

  *set = (unsigned long *) allocateMemory(mem,16);
  assert(*set);

  memset(*set,0,16);
  return PARSE_OK;
}

#define IsElement(E,S) (((E) < 0 || (E) >= 128) ? 0 : \
                        ((S)[word(E)] & ShiftBit(bit(E))) != 0UL)

static int
Insert(register e, 
       unsigned long *S)			/* Insert e into S */
{
  assert(S);
  assert(e >= 0 && e < 128);
  S[word(e)] |= ShiftBit(bit(e));
  return PARSE_OK;
}

static int
DelMember(register int e, 
	  unsigned long *S)			/* Delete e from S */
{
  /*	This operation won't expand the set */
  assert(S);
  assert(e >= 0 && e < 128);
  S[word(e)] &= ~ShiftBit(bit(e));
  return PARSE_OK;
}

static int
BlockSet(register unsigned long *S)
{
  assert(S);
  memset(S,0xff,16);
  return PARSE_OK;
}

/****************************************************************************/
/************    little routines to manage the REGEX stack     **************/
/****************************************************************************/
#define regPush(X,Y) (X)->stack[(X)->regEnd++] = (Y)
#define regPop(X) (X)->stack[--((X)->regEnd)]
#define regClear(X) (X)->regEnd = 0;
#define regDepth(X) ((X)->regEnd)

#ifndef NODBG
struct regNumStack {
  int stack[REG_STACK_DEPTH];
  int regNumEnd;
};

static int
regNumPush(struct regNumStack *rnS, int num)
{
  assert(rnS);
  assert(rnS->regNumEnd < REG_STACK_DEPTH);
  rnS->stack[rnS->regNumEnd++] = num;
  return PARSE_OK;
}

static int
regNumPop(struct regNumStack *rnS)
{
  assert(rnS);
  assert(rnS->regNumEnd > 0);
  rnS->regNumEnd--;
  return PARSE_OK;
}

static int 
regNumPeek(struct regNumStack *rnS, int i, int *ret)
{
  assert(rnS);
  if (i < rnS->regNumEnd){
    *ret = rnS->stack[i];
    return PARSE_OK;
  }
  return PARSE_FAIL;
}

static int regNumZero(struct regNumStack *rnS)
{
  assert(rnS);
  rnS->regNumEnd = 0;
  return PARSE_OK;
}

static int regNumEnd(struct regNumStack *rnS, int *ret)
{
  assert(rnS);
  *ret = rnS->regNumEnd;
  return PARSE_OK;
}
#endif

/****************************************************************************/
/************   end of routines to manage the REGEX stack     ***************/
/****************************************************************************/
#ifndef NODBG
extern char *typeArray[];
#endif

static int
parseContent(ISL_MEMORY_BUFFER_PTR *mem,
	     struct contentData *cD,
#ifndef NODBG
	     int debugLevel,
#endif
	     struct parseEl *tree,
	     struct valueStack *vS)
{
  struct parseEl *walker;
  int varDepth = 0;
  int match = 0, ch, matchVal;
  stackEl value, *valP;
  struct regStack rS;

#ifndef NODBG
  struct regNumStack rnS;
  
  if ((regNumZero(&rnS)) != PARSE_OK)
    return PARSE_FAIL;
#endif

  rS.regEnd = 0;

  assert(tree);
  assert(cD);

  vS->valTop = 0;
  vS->lastFoff = 0;
  /* we'll estimate a max value stack size of 5x the length of the
   * content.  That should always be plenty
   */
  vS->stack = 
    (stackEl *) 
    allocateMemory(mem, (unsigned)(5 * cD->ptr->Length * sizeof(stackEl)));
  assert(vS->stack);

  /*
   * we should be able to walk the parse tree, consuming the file as
   * we go.  When we exhaust the parse tree, we should be at the end of
   * file, and vice versa.
   */

  walker = tree;

  while (walker){
#ifndef NODBG
    if (debugLevel > 2 && walker->type != NOP){
      int i, end;
      (void)fprintf(stdout,"DEBUG: CMP ");
      if ((regNumEnd(&rnS,&end)) != PARSE_OK){
	return PARSE_FAIL;
      }
      for (i = 0; i < end; i++){
	int peek;
	if ((regNumPeek(&rnS,i,&peek)) != PARSE_OK){
	  return PARSE_FAIL;
	}
	(void)fprintf(stdout,"N%d ", peek);
      }
      (void)fprintf(stdout,"N%d (%s) & byte %d\n",
	      walker->id,typeArray[walker->type],fakeFtell(cD));
    }
#endif
    matchVal = 1;

    if (walker->type == NOP){
      match = 1;
    } else {
      if (walker->type == RANGE){
	ch = fakeGetc(cD);
	match = IsElement((int)ch,walker->u.rbits);
	if (match){
	  value.type = BBYTE;
	  value.u.byte = ch;
	  valPush(value,vS);
#ifndef NODBG
	} else {
	  if (debugLevel > 2)
	    (void) fprintf(stdout,"DEBUG: RANGE mismatch (fi)<%d>\n", ch);
#endif
	}
      } else {
	if (walker->type == ICASE){
	  ch = fakeGetc(cD);
	  match = (ch == walker->u.iBytes[0] || ch == walker->u.iBytes[1]);
	  if (match){
	    value.type = BBYTE;
	    value.u.byte = ch;
	    valPush(value,vS);
#ifndef NODBG
	  } else {
	    if (debugLevel > 2){
	      (void) fprintf(stdout,
			     "DEBUG: ICASE mismatch (tmp)<%d> vs. (fi)<%d>\n",
			     walker->u.byte,ch);
	    }
#endif
	  }
	} else {
	  if (walker->type == EXACT){
	    ch = fakeGetc(cD);
	    match = (ch == walker->u.byte);
	    if (match){
	      value.type = BBYTE;
	      value.u.byte = ch;
	      valPush(value,vS);
#ifndef NODBG
	    } else {
	      if (debugLevel > 2){
		(void) fprintf(stdout,
			      "DEBUG: EXACT mismatch (tmp)<%d> vs. (fi)<%d>\n",
			       walker->u.byte,ch);
	      }
#endif
	    }
	  } else {
	    if (walker->type == FBYTE){
	      /* FBYTE--we should be able to do a strncmp for the length of
	       * "fbytes".  It's worth noting that these FBYTE guys should 
	       * only occur when the regex stack is empty and we are 
	       * tracking *no* variables...hence, they should never cause 
	       * any net change to the value stack, save for the addition
	       * of SFBYTE guys
	       */
	      int i,l;

	      assert(varDepth == 0);
	      assert((regDepth(&rS)) == 0);
	      assert(walker->u.fbytes != NULL);

	      i = 0;
	      l = strlen(walker->u.fbytes);
	      for (i = 0; i < l; i++ ){
		ch = fakeGetc(cD);
		if (ch != walker->u.fbytes[i])
		  break;
	      }
	      match = i == l;
	      if (match){
		value.type = SFBYTE;
		value.u.length = l;
		valPush(value,vS);
#ifndef NODBG
	      } else {
		if (debugLevel > 2 )
		  (void) fprintf(stdout,
			      "DEBUG: FBYTE mismatch (tmp)<%d> vs. (fi)<%d>\n",
				 walker->u.fbytes[i],ch);
#endif
	      }
	    } else {
	      if (walker->type == REG_B) {
		/* push the "return" address onto the regex stack */
		assert(walker->choice0 != NULL);
		regPush(&rS,walker->choice0);
#ifndef NODBG
		if ((regNumPush(&rnS,walker->id)) != PARSE_OK)
		  return PARSE_FAIL;
#endif
		/* descend down into the parse subtree */
		assert(walker->choice1 != NULL);
		walker = walker->choice1;
		matchVal = 0;
	      } else {
		if (walker->type == REG_E){
		  assert(walker->choice0 == NULL);
		  assert(walker->choice1 == NULL);
		  walker = regPop(&rS);
#ifndef NODBG
		  if ((regNumPop(&rnS)) != PARSE_OK)
		    return PARSE_FAIL;
#endif
		  matchVal = 0;
		} else {
		  if (walker->type == VAR_B){
		    /* add a var_beg marker on the end of the value stack */
		    value.type = VAR_BEGIN;
		    value.u.var = walker->u.var;
		    valPush(value,vS);
		    varDepth++;
		    assert(walker->choice1 == NULL);
		    walker = walker->choice0;
		    matchVal = 0;
		  } else {
		    if (walker->type == VAR_E){
		      /* add a var_end marker on the end of the value stack */
		      value.type = VAR_END;
		      value.u.var = walker->u.var;
		      valPush(value,vS);
		      varDepth--;
		      assert(walker->choice1 == NULL);
		      walker = walker->choice0;
		      matchVal = 0;
		    } else {
		      /* PEND: this is the last node in the parse tree.  If 
		       * the next char gotten is EOF, we win!  Otherwise, we 
		       * need to rewind just like elsewhere.
		       */
		      ch = fakeGetc(cD);
		      if (ch != EOF){
			match = 0;
		      } else {
			value.type = END;
			valPush(value,vS);
			return PARSE_OK;
		      }
		    }
		  }
		}
	      }
	    }
	  }
	}
      }
    }

    if (matchVal){
      if (match == 0){
	/* we need to rewind the parse tree */
	/* things that should happen...
	 * 2) backtrack to the last FOFF in the value stack, and seek to that
	 *    point
	 * 3) clear the REGEX stack, and pop all the REGSTACK entries from the
	 *    the value stack and push them onto the REGEX stack.
	 * 4) pop varDepth off the value stack
	 * 5) pop the BRANCH value off the stack, and branch to it
	 */
	if (vS->lastFoff >= 0){
	  value = vS->stack[vS->lastFoff];
	  vS->valTop = vS->lastFoff;
	  vS->lastFoff = value.lastFoff;
	} else {
	  /* this means we have searched the stack and found no prior point 
	   * appropriate for undoing a decision.  We're done, and the only
	   * copy(s) of the value stack end in ERROR.
	   */
#ifndef NODBG
	  if (debugLevel > 2){
	    (void)fprintf(stdout,"DEBUG: giving up at byte %d of content\n", 
			  fakeFtell(cD));
	  }
#endif
	  value.type = VERROR;
	  valPush(value,vS);
	  return PARSE_FAIL;
	}
	if (vS->valTop <= 0){
	  /* likewise, we've pooped out, as there's nothing to unwind.  This
	   * happens when the very first thing in the file was a FBYTE.  Yes,
	   * we rewound to an offset, but there may be nothing else to do.
	   */
#ifndef NODBG
	  if (debugLevel > 2){
	    (void)fprintf(stdout,"DEBUG: giving up at byte %d of content\n", 
			  fakeFtell(cD));
	  }
#endif
	  value.type = VERROR;
	  valPush(value,vS);
	  return PARSE_FAIL;
	}

	/* if we're still here, we need to finish unwinding the parse state */
	fakeFseek(cD,value.u.foff);
	regClear(&rS);
#ifndef NODBG
	if ((regNumZero(&rnS)) != PARSE_OK)
	  return PARSE_FAIL;
#endif
	valP = valPPop(vS);
	while (valP->type == REGSTACK){
	  regPush(&rS,valP->u.pNode);
#ifndef NODBG
	  if ((regNumPush(&rnS, valP->u.pNode->id)) != PARSE_OK)
	    return PARSE_FAIL;
	  if (debugLevel > 3)
	    (void)fprintf(stdout, 
			  "DEBUG: UNWIND pushed N%d onto regStack\n",
			  valP->u.pNode->id);
#endif
	  valP = valPPop(vS);
	}
	assert(valP->type == VAR_DEPTH);
	varDepth = valP->u.var;
	value = valPop(vS);
	assert(value.type == BRANCH);
#ifndef NODBG
	if (debugLevel > 3)
	  (void)
	    fprintf(stdout, 
		  "DEBUG: UNWIND to N%d (%s)->choice0: N%d (%s) next\n",
		  value.u.pNode->id,
		  typeArray[value.u.pNode->type],
		  value.u.pNode->choice0->id,
		  typeArray[value.u.pNode->choice0->type]);
#endif
	walker = value.u.pNode->choice0;
      } else {
	/* these are the only types of node that can have a "normal" non-NULL
	 * choice1.  We have to check to see if we have a choice.
	 */
	if (walker->choice1 != NULL){
	  /* we always take choice1 while we're successful.  But we need to
	   * leave enough state on the value stack to allow a backtrack to
	   * take the other fork from this point.  This consists of:
	   * 1) our offset in the "tree"
	   * 2) the value of varDepth
	   * 3) the elements that are on the regStack
	   * 4) the current file offset for "outfi"
	   */
	  int i;
	  value.type = BRANCH;
	  value.u.pNode = walker;
	  valPush(value,vS);

	  value.type = VAR_DEPTH;
	  value.u.var = varDepth;
	  valPush(value,vS);

	  value.type = REGSTACK;
	  i = regDepth(&rS) - 1;
	  for (; i >= 0; i--){
	    value.u.pNode = rS.stack[i];
	    valPush(value,vS);
	  }
	  value.type = FOFFSET;
	  value.u.foff = fakeFtell(cD);
	  value.lastFoff = vS->lastFoff;
	  vS->lastFoff = vS->valTop;
	  valPush(value,vS);
	  walker = walker->choice1;
	} else {
	  assert(walker->choice0);
	  walker = walker->choice0;
	}
      }
    }
  }
#ifndef NODBG
  (void)
    fprintf(stdout,"FATAL: reached non-END parse node with no child\n");
#endif
  return PARSE_FAIL;
}
/****************************************************************************/
/*******************  routines to manage the values stack(s)  ***************/
/****************************************************************************/

static int
populateResults(ISL_ARCHIVE_CONTEXT_PTR ArchivePtr,
		struct valueStack *vS,
		struct contentData *cD,
		ISL_DATA_CHAIN_PTR *results, 
		int maxResult)
{
  /* the values we have to emit, if any, are on bestStack.
   * So far as values go, the occurence of any VAR_BEGIN node starts
   * a value for the variable index indicated.  The only things we
   * care about besides those, VAR_END (you can guess, right?) and
   * END/ERROR, which determine the value emitted for $text_error,
   * are BYTE nodes.  A single value for any VAR will consist of any
   * BYTES found between a VAR_BEGIN and VAR_END pair.  Notice that
   * more than one variable can be accumulated at once, but these
   * cases should always be properly nested.
   */
  
  stackEl *ptr = vS->stack;
  ISL_DATA_CHAIN_PTR resultsInProgress = NULL;
  ISL_DATA_CHAIN_PTR dcPtr;
  const unsigned char *contPtr;

  assert(cD);
  assert(results);
  assert(vS);
  
  contPtr = cD->ptr->Data;
	 
  while (ptr != NULL){
    if (ptr->type == END){
      /* there should be nothing on the resultsInProgress */
      assert(resultsInProgress == NULL);
      return PARSE_OK;
    }
    if (ptr->type == VERROR){
      /* anything on the liveVarStack is incomplete, so we'll ignore */
      return PARSE_FAIL;
    }
    if (ptr->type == VAR_BEGIN){
      /* make a new ISL_DATA_CHAIN item */
      dcPtr = (ISL_DATA_CHAIN_PTR)
	isl_AllocateMemory(ArchivePtr, sizeof(ISL_DATA_CHAIN));
      assert(dcPtr);

      /* insert at list head */
      dcPtr->next = resultsInProgress;
      resultsInProgress = dcPtr;

      /* snag an ISL_DATA guy for this to point to */
      dcPtr->ptr = (ISL_DATA *)
	isl_AllocateMemory(ArchivePtr, sizeof(ISL_DATA));
      assert(dcPtr->ptr);
	
      /* set the pointer for this ISL_DATA to the current offset and set
       * the length to 0
       */
      dcPtr->ptr->Data = (unsigned char *)contPtr;
      dcPtr->ptr->Length = 0;
    } else {
      if (ptr->type == VAR_END){
	/* pop the var off the progress list, and attach it to its owner
	 * on the varList.
	 */
	assert(resultsInProgress);
	dcPtr = resultsInProgress;
	resultsInProgress = resultsInProgress->next;
	assert(ptr->u.var < maxResult);

	/* append it to the results list */
	{
	  ISL_DATA_CHAIN_PTR dcPtr2 = results[ptr->u.var];
	  if (! dcPtr2){
	    results[ptr->u.var] = dcPtr;
	  } else {
	    while (dcPtr2->next)
	      dcPtr2 = dcPtr2->next;
	    dcPtr2->next = dcPtr;
	  }
	  dcPtr->next = NULL;
	  dcPtr->ptr->Length = contPtr - dcPtr->ptr->Data;
	}
      } else {
	if (ptr->type == BBYTE){
	  /* wow, a byte. */
	  contPtr++;
	  assert(contPtr <= cD->ptr->Data + cD->ptr->Length);
	} else {
	  if (ptr->type == SFBYTE){
	    /* wow, lots of bytes. */
	    contPtr += ptr->u.length;
	    assert(contPtr <= cD->ptr->Data + cD->ptr->Length);
	  }
	}
      }
    }
    ptr++;
  }
  /* if we're here and didn't hit an END, things are hosed */
  return PARSE_FAIL;
}

/**************************************************************************/
/*************************** reading the defs section *********************/
/**************************************************************************/

static int
explicitRange(ISL_MEMORY_BUFFER_PTR *mem,
	      char ch,
#ifndef NODBG
	      int *id,
#endif
	      struct parseEl **ret)
{
  int i;
  if ((newParseEl(mem,
#ifndef NODBG
		 id,
#endif
		 ret)) != PARSE_OK){
    return PARSE_FAIL;
  }
  (*ret)->type = RANGE;
  if ((CreateSet(mem,&((*ret)->u.rbits))) != PARSE_OK){
    return PARSE_FAIL;
  }
  if (ch == 'd'){
    for (i = '0'; i <= '9'; i++)
      Insert(i,(*ret)->u.rbits);
  } else {
    if (ch == 'D'){
      BlockSet((*ret)->u.rbits);
      for (i = '0'; i <= '9'; i++)
	DelMember(i,(*ret)->u.rbits);
    } else {
      if (ch == '$'){
	BlockSet((*ret)->u.rbits);
      } else {
        if (ch == 'w'){
	  for (i = '0'; i <= '9'; i++)
	    Insert(i,(*ret)->u.rbits);
	  for (i = 'a'; i <= 'z'; i++)
	    Insert(i,(*ret)->u.rbits);
	  for (i = 'A'; i <= 'Z'; i++)
	    Insert(i,(*ret)->u.rbits);
	  Insert(i,(*ret)->u.rbits);
	} else {
	  if (ch == 'W'){
	    BlockSet((*ret)->u.rbits);
	    for (i = '0'; i <= '9'; i++)
	      DelMember(i,(*ret)->u.rbits);
	    for (i = 'a'; i <= 'z'; i++)
	      DelMember(i,(*ret)->u.rbits);
	    for (i = 'A'; i <= 'Z'; i++)
	      DelMember(i,(*ret)->u.rbits);
	    DelMember(i,(*ret)->u.rbits);
	  } else {
	    if (ch == 's'){
	      Insert(' ',(*ret)->u.rbits);
	      Insert('\t',(*ret)->u.rbits);
	      Insert('\f',(*ret)->u.rbits);
	      Insert('\n',(*ret)->u.rbits);
	    } else {
	      if (ch == 'S'){
		BlockSet((*ret)->u.rbits);
		DelMember(' ',(*ret)->u.rbits);
		DelMember('\t',(*ret)->u.rbits);
		DelMember('\f',(*ret)->u.rbits);
		DelMember('\n',(*ret)->u.rbits);
	      } else {
		if (ch == 'N'){
		  Insert(' ',(*ret)->u.rbits);
		  Insert('\t',(*ret)->u.rbits);
		  Insert('\f',(*ret)->u.rbits);
		} else {
#ifndef NODBG
		  (void)
		    fprintf(stdout,
			    "FATAL: unknown char %c to explicitRange()\n", ch);
#endif
		  return PARSE_FAIL;
		}
	      }
	    }
	  }
	}
      }
    }
  }
  return PARSE_OK;
}

static int
makeExact(ISL_MEMORY_BUFFER_PTR *mem,
	  char ch,
#ifndef NODBG
	  int *id,
#endif
	  int iCase,
	  struct parseEl **ret
	  )
{
  if ((newParseEl(mem,
#ifndef NODBG
		  id,
#endif
		  ret)) != PARSE_OK){
    return PARSE_FAIL;
  }
  if (iCase){
    (*ret)->type = ICASE;
    (*ret)->u.iBytes[0] = (char) _tolower(ch);
    (*ret)->u.iBytes[1] = (char) _toupper(ch);
  } else {
    (*ret)->type = EXACT;
    (*ret)->u.byte = ch;
  }
  return PARSE_OK;
}

static int
parseEscape(ISL_MEMORY_BUFFER_PTR *mem,
	    char **sptr,
#ifndef NODBG
	    int debugLevel,
	    char **defPtr,
	    int *id,
#endif
	    struct parseEl **pfront,
	    struct parseEl **pback)
{ 
#ifndef NODBG
  char *addr = *sptr;
#endif
  assert(**sptr == BSLASH);
  (*sptr)++;
  if (**sptr == 'n'){
    if ((makeExact(mem,'\n',
#ifndef NODBG
		   id,
#endif
		   0,pfront)) != PARSE_OK){
      return PARSE_FAIL;
    }
    *pback = *pfront;
  } else {
    if (**sptr == 't'){
      if ((makeExact(mem,'\t',
#ifndef NODBG
		     id,
#endif
		     0,pfront)) != PARSE_OK){
	return PARSE_FAIL;
      }
      *pback = *pfront;
    } else {
      if (**sptr == 'v'){
	if ((makeExact(mem,'\v',
#ifndef NODBG
		       id,
#endif
		       0,pfront)) != PARSE_OK){
	  return PARSE_FAIL;
	}
	*pback = *pfront;
      } else {
	if (**sptr == 'b'){
	  if ((makeExact(mem,'\b',
#ifndef NODBG
			 id,
#endif
			 0,pfront)) != PARSE_OK){
	    return PARSE_FAIL;
	  }
	  *pback = *pfront;
	} else {
	  if (**sptr == 'r'){
	    if ((makeExact(mem, '\r',
#ifndef NODBG
			   id,
#endif
			   0,pfront)) != PARSE_OK){
	      return PARSE_FAIL;
	    }
	    *pback = *pfront;
	  } else {
	    if (**sptr == 'f'){
	      if ((makeExact(mem, '\f',
#ifndef NODBG
			     id,
#endif
			     0,pfront)) != PARSE_OK){
		return PARSE_FAIL;
	      }
	      *pback = *pfront;
	    } else {
	      if (**sptr == 'a'){
		if ((makeExact(mem, '\a',
#ifndef NODBG
			       id,
#endif
			       0,pfront)) != PARSE_OK){
		  return PARSE_FAIL;
		}
		*pback = *pfront;
	      } else {
		if (**sptr == '0'){
		  /* octal */
		  int i, value;
		  value = 0;
		  (*sptr)++;
		  for (i = 0; i < 3; i++){
		    if (!ISDIGIT(**sptr)){
		      break;
		    } else {
		      value = value * 8 + **sptr - '\0';
		      (*sptr)++;
		    }
		  }
		  if (value > 127){
#ifndef NODBG
		    (void)
		      fprintf(stdout,"WARNING: octal value %d > 127\n", value);
#endif
		    value %= 128;
		  }
		  if ((makeExact(mem,
				 (char)value,
#ifndef NODBG
				 id,
#endif
				 0, pfront)) != PARSE_OK){
		    return PARSE_FAIL;
		  }
		  *pback = *pfront;
		} else {
		  /* Perl-style metacharacters */
		  if (**sptr == 'd' ||   /* digit, [0-9] */
		      **sptr == 'D' ||   /* non-digit, [^0-9] */
		      **sptr == 'w' ||   /* word char, [0-9a-z_A-Z] */
		      **sptr == 'W' ||   /* nonword char, [^0-9a-z_A-Z] */
		      **sptr == 'N' ||   /* any space but \n */
		      **sptr == 's' ||   /* whitespace */
		      **sptr == 'S' ||   /* non-whitespace */
		      **sptr == '$'      /* anything, newlines included */){
		    if ((explicitRange(mem, **sptr,
#ifndef NODBG
				       id,
#endif
				       pfront)) != PARSE_OK){
		      return PARSE_FAIL;
		    }
		    *pback = *pfront;
		  } else {
		    /* another BSLASH, or anything else */
		    if ((makeExact(mem,
				   **sptr,
#ifndef NODBG
				   id,
#endif
				   0,pfront)) != PARSE_OK){
		      return PARSE_FAIL;
		    }
		    *pback = *pfront;
		  }
		}
	      }
	    }
	  }
	}
      }
    }
  }
  (*sptr)++;
#ifndef NODBG
  if (debugLevel > 2){
    fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
	    (*pfront)->id, addr - *defPtr);
  }
#endif
  return PARSE_OK;
}

static int
parseCharAlt(ISL_MEMORY_BUFFER_PTR *mem,
	     char **sptr,
#ifndef NODBG
	     int debugLevel,
	     char *origLine,
	     char **defPtr,
	     int *id,
#endif
	     struct parseEl **pfront,
	     struct parseEl **pback)
{  
#ifndef NODBG
  char *addr = *sptr;
#endif
  if (**sptr == '.'){
    /* this should match anything but a newline */
    if ((newParseEl(mem,
#ifndef NODBG
		    id,
#endif
		    pfront)) != PARSE_OK){
      return PARSE_FAIL;
    }
    (*pfront)->type = RANGE;
    if ((CreateSet(mem,&((*pfront)->u.rbits))) != PARSE_OK){
      return PARSE_FAIL;
    }
    BlockSet((*pfront)->u.rbits);
    DelMember('\n',(*pfront)->u.rbits);
    *pback = *pfront;
    (*sptr)++;
  } else {
    int reverse = 0;

    assert(**sptr == '[');

    if ((newParseEl(mem,
#ifndef NODBG
		    id,
#endif
		    pfront)) != PARSE_OK){
      return PARSE_FAIL;
    }
    (*pfront)->type = RANGE;
    if ((CreateSet(mem,&((*pfront)->u.rbits))) != PARSE_OK){
      return PARSE_FAIL;
    }
    (*sptr)++;
    if (**sptr == '^'){
      reverse = 1;
      (*sptr)++;
      BlockSet((*pfront)->u.rbits);
    }
#define SET_ELEMENT(e,S) if (reverse) DelMember((e),(S)); else Insert((e),(S));
 
    /* check for leading '-' now */
    if (**sptr == '-'){
      SET_ELEMENT('-',(*pfront)->u.rbits);
      (*sptr)++;
    }
    while (**sptr != '\0' && **sptr != ']'){
      if (**sptr == '-'){
#ifndef NODBG
	(void) fprintf(stdout,
		       "ERROR: poorly formed character range with - :\n\t%s",
		       origLine);
#endif
	return PARSE_FAIL;
      }
      if (*(*sptr+1) == '-'){
	int i;
	if (*(*sptr+2) == '\0' || *(*sptr+2) == ']'){
#ifndef NODBG
	  (void) fprintf(stdout,
			 "FATAL: poorly formed character range with - :\n\t%s",
			 origLine);
#endif
	  return PARSE_FAIL;
	}

	/* it's a range */
	for (i = **sptr; i <= *(*sptr+2); i++)
	  SET_ELEMENT(i,(*pfront)->u.rbits);
	(*sptr) += 3;
      } else {
	/* single char */
	SET_ELEMENT(**sptr,(*pfront)->u.rbits);
	(*sptr)++;
      }
    }
    if (**sptr != ']'){
#ifndef NODBG
      (void)
	fprintf(stdout,
	      "FATAL: missing close-bracket in character alternation:\n\t%s",
	      origLine);
#endif
      return PARSE_FAIL;
    }
    (*sptr)++;
    *pback = *pfront;
  }
#ifndef NODBG
  if (debugLevel > 2){
    fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
	    (*pfront)->id, addr - *defPtr);
  }
#endif
  return PARSE_OK;
}

static int
parseAlt(ISL_MEMORY_BUFFER_PTR *mem,
	 char **sptr,
#ifndef NODBG
	 int debugLevel,
	 char origLine[LINE_MAX],
	 char **defPtr,
	 int *id,
#endif
	 struct escapeTable *esc,
	 struct parseEl **pfront,
	 struct parseEl **pback,
	 int iCase,
	 int *isLambda)
{
  struct parseEl *head, *end, *connector;
  int localLambda; 
#ifndef NODBG
  char *addr = *sptr;
#endif
  assert(**sptr == '{');
  (*sptr)++;

  /* get the NOP nodes for the front and back of the whole schmear */
  if ((newParseEl(mem,
#ifndef NODBG
		  id,
#endif
		  pfront)) != PARSE_OK){
    return PARSE_FAIL;
  }
  if ((newParseEl(mem,
#ifndef NODBG
		  id,
#endif
		  pback)) != PARSE_OK){
    return PARSE_FAIL;
  }
  (*pfront)->type = NOP;
  (*pback)->type = NOP;
#ifndef NODBG
  if (debugLevel > 2){
    fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
	    (*pfront)->id, addr - *defPtr);
    fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
	    (*pback)->id, addr - *defPtr);
  }
#endif
  connector = *pfront;

  /* an alternation is a lambda if *any* of the alternates are lambdas */
  *isLambda = 0;

  while (**sptr != '\0' && **sptr != '}'){

#ifndef NODBG
    head = end = NULL;
#endif

    parseRegString(mem, sptr,
#ifndef NODBG
		   debugLevel,
		   origLine,
		   defPtr,
		   id,
#endif
		   esc, &head,&end,1,iCase,&localLambda);

    assert(head != NULL);
    assert(end != NULL);

    if (localLambda)
      *isLambda = 1;

    if (**sptr == '|'){
      /* we've got another choice coming */
      connector->choice1 = head;
      end->choice0 = *pback;
      if ((newParseEl(mem,
#ifndef NODBG
		      id,
#endif
		      &(connector->choice0))) != PARSE_OK){
	return PARSE_FAIL;
      }
      connector = connector->choice0;
      connector->type = NOP;
#ifndef NODBG
      if (debugLevel > 2){
	fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
		connector->id, *sptr - *defPtr);
      }
#endif
      (*sptr)++;
      /* handle the "|}" case */
      if (**sptr == '}'){
	connector->choice0 = *pback;
	*isLambda = 1;
      }
    } else {
      if (**sptr != '}'){
#ifndef NODBG
	(void) fprintf(stdout,
		       "FATAL: missing close-brace on alternation:\n\t%s",
		       origLine);
#endif
	return PARSE_FAIL;
      }
      connector->choice0 = head;
      end->choice0 = *pback;
    }
  }
#ifndef NODBG
  if (debugLevel > 2){
    fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
	    (*pfront)->id, addr - *defPtr);
  }
#endif
  (*sptr)++;
  return PARSE_OK;
}

static int
parseDefinedName(ISL_MEMORY_BUFFER_PTR *mem,
		 char **sptr,
#ifndef NODBG
		 int debugLevel,
		 char **defPtr,
		 int *id,
#endif
		 struct escapeTable *esc,
		 struct parseEl **pfront,
		 struct parseEl **pback,
		 int *isLambda, int *result)
{
  /* allowable format is either:
   * <name> where name is a predefined escape.
   * <var=name> where name is predefined escape and var is any variable number
   * anything else, which fails to match, and we return a 0.
   * allowed chars are a-zA-Z_0-9 in both names, and they must start with a
   * alpha.
   * for the "name-only" case we will set pfront and pback to point to
   * a REG_B node that points (through its choice1) to the parse tree defined
   * for "name".  For var=name, pfront will point to a VAR_B node that points
   * to a REG_B as above, the REG_B will have its choice0 pointing to a
   * VAR_E node that will be pointed to by pback.
   * for the failure case (<whatever>), we have to return with sptr unaltered
   * and the string it points to intact.
   */
  char *p, *p2, *addr;
  struct parseEl *n1, *n2;

  addr = p = *sptr;
  assert(*p == '<');

  p++;
  if (!(ISALPHA(*p))){
    /* first char's not an alpha */
    if (!(ISDIGIT(*p))){
      /* first char's not a digit either, Uncle! */
      *result = 0;
      return PARSE_OK;
    }
    p++;
    while (ISDIGIT(*p)){
      p++;
    }
  } else {
    p++;
    while (ISALPHA(*p) || *p == '_'){
      p++;
    }
  }
  if (*p == '>'){
    /* it's just a name */
    *p = '\0';
    n1 = lookupEscape(*sptr + 1,esc);
    if (n1 == NULL){
      /* lookup failed, it's not defined */
      *p = '>';
      *result = 0;
      return PARSE_OK;
    } else {
      if ((newParseEl(mem,
#ifndef NODBG
		      id,
#endif
		      &n2)) != PARSE_OK){
	*p = '>';
	return PARSE_FAIL;
      }
      n2->type = REG_B;
      n2->choice1 = n1;

      /* check to see if this escape is a lambda */
      *isLambda = escapeIsLambda(*sptr + 1,esc);

      if (((strlen(*sptr + 1)) >= MAXNAME)){
	*p = '>';
	assert(0);
      }
      strcpy(n2->u.name,*sptr + 1);
      *p = '>';
      *pfront = *pback = n2;
      *sptr = p + 1;
#ifndef NODBG
      if (debugLevel > 2){
	fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
		(*pfront)->id, addr - *defPtr);
      }
#endif
      *result = 1;
      return PARSE_OK;
    }
  } else {
    if (*p == '='){
      /* it may be a var followed by an "=name>" */
      p2 = p + 1;
      if (!(ISALPHA(*p2))){
	/* first char's not an alpha */
	*result = 0;
	return PARSE_OK;
      }
      p2++;
      while (ISALPHA(*p2) || ISDIGIT(*p2) || *p2 == '_'){
	p2++;
      }

      if (*p2 == '>'){
	/* okay, this is a name, so we had <var=name>, see if "name" is
	 * valid.
	 */
	*p2 = '\0';
	n1 = lookupEscape(p + 1,esc);
	if (n1 == NULL){
	  /* lookup failed, it's not defined */
	  *p2 = '>';
	  *result = 0;
	  return PARSE_OK;
	} else {
	  if ((newParseEl(mem,
#ifndef NODBG
			  id,
#endif
			  pfront)) != PARSE_OK){
	    return PARSE_FAIL;
	  }
	  if ((newParseEl(mem,
#ifndef NODBG
			  id,
#endif
			  pback)) != PARSE_OK){
	    return PARSE_FAIL;
	  }
	  if ((newParseEl(mem,
#ifndef NODBG
			  id,
#endif
			  &n2)) != PARSE_OK){
	    return PARSE_FAIL;
	  }

	  /* get the nodes for the var+expr */
	  (*pfront)->type = VAR_B;
	  (*pback)->type  = VAR_E;
	  n2->type = REG_B;
	
	  /* get an index for the variable name */
	  {
	    /* poor man's strtol, base 10 */
	    char *pr = *sptr + 1;
	    long index = 0;
	    while (pr != p){
	      index *= 10;
	      index += *pr - '0';
	      pr++;
	    }
	    (*pfront)->u.var = (*pback)->u.var = index;
	  }

	  /* link up the nodes */
	  (*pfront)->choice0 = n2;
	  n2->choice0 = (*pback);
	  n2->choice1 = n1;

	  /* check to see if this escape is a lambda */
	  *isLambda = escapeIsLambda(p + 1,esc);

	  assert((strlen(p + 1)) < MAXNAME);
	  strcpy(n2->u.name,(p + 1));
	
	  /* adjust the string pointer */
	  *sptr = p2 + 1;
#ifndef NODBG
	  if (debugLevel > 2){
	    fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
		    (*pfront)->id, addr - *defPtr);
	    fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
		    n2->id, addr - *defPtr);
	    fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
		    (*pback)->id, addr - *defPtr);
	  }
#endif
	  *result = 1;
	  return PARSE_OK;
	}
      } else {
	*result = 0;
	return PARSE_OK;
      }
    } else {
      *result = 0;
      return PARSE_OK;
    }
  }
  return PARSE_FAIL;
}

static int
parseAtom(ISL_MEMORY_BUFFER_PTR *mem,
	  char **sptr,
#ifndef NODBG
	  int debugLevel,
	  char origLine[LINE_MAX],
	  char **defPtr,
	  int *id,
#endif
	  struct escapeTable *esc,
	  struct parseEl **pfront,
	  struct parseEl **pback,
	  int iCase, int isLit, int *isLambda)
{
#ifndef NODBG
  char *addr = *sptr;
#endif
  int result;
  if (isLit){
    /* we're in a quoted string */
    if ((makeExact(mem, **sptr,
#ifndef NODBG
		   id,
#endif
		   iCase,pfront)) != PARSE_OK){
      return PARSE_FAIL;
    }
    *pback = *pfront;
    *isLambda = 0;
    (*sptr)++;
  } else {
    if (**sptr == BSLASH){
      *isLambda = 0;
      parseEscape(mem, sptr,
#ifndef NODBG
		  debugLevel,
		  defPtr,
		  id,
#endif
		  pfront,pback);
    } else {
      if (**sptr == '[' || **sptr == '.'){
	*isLambda = 0;
	parseCharAlt(mem, sptr,
#ifndef NODBG
		     debugLevel,
		     origLine,
		     defPtr,
		     id,
#endif
		     pfront,pback);
      } else {
	if (**sptr == '{'){
	  parseAlt(mem,sptr,
#ifndef NODBG
		   debugLevel,
		   origLine,
		   defPtr,
		   id,
#endif
		   esc,pfront,pback,iCase,isLambda);
	} else {
	  if (**sptr == '<'){
	    if ((parseDefinedName(mem, sptr,
#ifndef NODBG
				  debugLevel,
				  defPtr,
				  id,
#endif
				  esc,pfront,pback,
				  isLambda,&result)) != PARSE_OK)
	      return PARSE_FAIL;
	    if (!result){
	      /* if call result zero, we need to parse the '<' and
	       * subsequent chars as atoms.
	       */
	      if ((makeExact(mem,**sptr,
#ifndef NODBG
			     id,
#endif
			     iCase,pfront)) != PARSE_OK){
		return PARSE_FAIL;
	      }
	      *pback = *pfront;
	      (*sptr)++;
	    }
	  } else {
	    *isLambda = 0;
	    if ((makeExact(mem, **sptr,
#ifndef NODBG
			   id,
#endif
			   iCase,pfront)) != PARSE_OK){
	      return PARSE_FAIL;
	    }
	    *pback = *pfront;
	    (*sptr)++;
	  }
	}
      }
    }
  }
#ifndef NODBG
  if (debugLevel > 2){
    fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
	    (*pfront)->id, addr - *defPtr);
  }
#endif
  return PARSE_OK;
}

static int
parseRegex(ISL_MEMORY_BUFFER_PTR *mem,
	   char **sptr,
#ifndef NODBG
	   int debugLevel,
	   char *origLine,
	   char **defPtr,
	   int *id,
#endif
	   struct escapeTable *esc,
	   struct parseEl **pfront,
	   struct parseEl **pback,
	   int iCase, int isLit, int *isLambda)
{
  struct parseEl *front = NULL;
  struct parseEl *back = NULL;  
#ifndef NODBG
  char *addr = *sptr;
#endif
  back = front = NULL;
  parseAtom(mem, sptr,
#ifndef NODBG
	    debugLevel,
	    origLine,
	    defPtr,
	    id,
#endif
	    esc,&front,&back,iCase,isLit,isLambda);
  assert(back != NULL);
  assert(front != NULL);

  /* handle repetition cases */
  if (!isLit && **sptr == '+'){
    /* this node's "choice1" may be spoken for, and we have to append a
     * NOP to the choice0 to make this stuff work.
     */
    if (*isLambda){
      /* user has asked for a repeated lambda expr */
#ifndef NODBG
      (void) fprintf(stdout, 
		     "ERROR: + modifying potentially empty expr:\n\t%s\n",
		     origLine);
#endif
      return PARSE_FAIL;
    }
    if (back->choice1 != NULL){
      struct parseEl *nop;
      if ((newParseEl(mem,
#ifndef NODBG
                      id,
#endif
		      &nop)) != PARSE_OK){
	return PARSE_FAIL;
      }
      nop->type = NOP;
      back->choice0 = nop;
      back = nop;
#ifndef NODBG
      if (debugLevel > 2){
	fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
		nop->id, *sptr - *defPtr);
      }
#endif
    }
    {
      /* shortest match first mods */
      struct parseEl *nop;
      if ((newParseEl(mem,
#ifndef NODBG
                      id,
#endif
		      &nop)) != PARSE_OK){
	return PARSE_FAIL;
      }
      nop->type = NOP;
      back->choice1 = nop;
      back->choice0 = front;
#ifndef NODBG
      if (debugLevel > 2){
	fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
		back->id, *sptr - *defPtr);
      }
#endif
      back = nop;
#ifndef NODBG
      if (debugLevel > 2){
	fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
		nop->id, *sptr - *defPtr);
      }
#endif
    }
    (*sptr)++;
    *pfront = front;
    *pback = back;
  } else {
    if (!isLit && **sptr == '*'){
      /* need to backwire "back" to point to front, but also need to
       * add NOPs so we can potentially jump around this atom altogether
       */
      struct parseEl *t1, *t2, *t3;
      if (*isLambda){
	/* user has asked for a repeated lambda expr */
#ifndef NODBG
	(void) fprintf(stdout, 
		       "FATAL: * modifying potentially empty expr:\n\t%s\n",
		       origLine);
#endif
	return PARSE_FAIL;
      }
      if ((newParseEl(mem,
#ifndef NODBG
                      id,
#endif
		      &t1)) != PARSE_OK){
	return PARSE_FAIL;
      }
      if ((newParseEl(mem,
#ifndef NODBG
                      id,
#endif
		      &t2)) != PARSE_OK){
	return PARSE_FAIL;
      }

      /* shortest match first mods */
      if ((newParseEl(mem,
#ifndef NODBG
                      id,
#endif
		      &t3)) != PARSE_OK){
	return PARSE_FAIL;
      }
      t1->type = t2->type = t3->type = NOP;
      t1->choice0 = front;
      t1->choice1 = t3;
      back->choice0 = t2;
      t2->choice0 = front;
      t2->choice1 = t3;
      *pfront = t1;
      *pback = t3;
#ifndef NODBG
      if (debugLevel > 2){
	fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
		t2->id, *sptr - *defPtr);
      }
#endif
      (*sptr)++;
      /* note that a * always changes the thing it modifies into a
       * potential lambda expression.
       */
      *isLambda = 1;
    } else {
      *pfront = front;
      *pback = back;
    }
  }
#ifndef NODBG
  if (debugLevel > 2){
    fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
	    (*pfront)->id, addr - *defPtr);
    fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
	    (*pback)->id, addr - *defPtr);
  }
#endif
  return PARSE_OK;
}

static int
parseRegString(ISL_MEMORY_BUFFER_PTR *mem,
	       char **s, 
#ifndef NODBG
	       int debugLevel,
	       char origLine[LINE_MAX],
	       char **defPtr,
	       int *id,
#endif
	       struct escapeTable *esc,
	       struct parseEl **head, 
	       struct parseEl **end, 
	       int inAlt, int iCase, int *isLambda)
{
  /* 
   * s is pointer to pointer to string to parse
   */
  struct parseEl *front, *back;
  int isLit = 0;
  int localLambda = 0;  
#ifndef NODBG
  char *addr = *s;
#endif

  *end = *head = NULL;

  /* assume the worst until proven wrong */
  *isLambda = 1;

  if (**s == '^'){
    iCase = 1;
    (*s)++;
  }

#ifndef NODBG
  front = back = NULL;
#endif

  if (!inAlt){
    /* toplevel, go to newline */
    while (**s != '\0' && **s != '\n'){

#ifndef NODBG
      front = back = NULL;
#endif

      /* if it's a double quote, toggle isLit */
      if (**s == DQUOTE){
        isLit = !isLit;
        (*s)++;
      } else {

	if (isLit){
	  if ((parseRegex(mem,s,
#ifndef NODBG
			  debugLevel,
			  origLine,
			  defPtr,
			  id,
#endif
			  esc,&front,&back,0,isLit,&localLambda)) != PARSE_OK)
	    return PARSE_FAIL;
	} else {
	  if ((parseRegex(mem,s,
#ifndef NODBG
			  debugLevel,
			  origLine,
			  defPtr,
			  id,
#endif
			  esc,&front,&back,iCase,isLit,&localLambda)) !=
	      PARSE_OK)
	    return PARSE_FAIL;
	}

	/* anything in the list that isn't a lambda means the whole
	 * expression can't be empty and isn't a lambda.
	 */
	if (!localLambda)
	  *isLambda = 0;

	assert(front != NULL);
	assert(back != NULL);

	if (*head == NULL){
	  *head = front;
	  *end  = back;
	} else {
	  /* hook new stuff onto the tree */
	  (*end)->choice0 = front;
	  *end = back;
	}
      }
    }

    if ((newParseEl(mem,
#ifndef NODBG
		    id,
#endif
		    &back)) != PARSE_OK){
      return PARSE_FAIL;
    }
    back->type = REG_E;
    if (*head == NULL){
      /* this "tree" is empty */
      assert(*end == NULL);
      *head = *end = back;
    } else {
      assert(*end != NULL);
      (*end)->choice0 = back;
      *end = back;
    }
  } else {
    /* we're in an alternation, and either a '|' or a '}' pops us out */
    while (**s != '\0' && **s != '|' && **s != '}'){

#ifndef NODBG
      front = back = NULL;
#endif

      /* if it's a double quote, toggle isLit */
      if (**s == DQUOTE){
	isLit = !isLit;
	(*s)++;
      } else {

	if ((parseRegex(mem, s,
#ifndef NODBG
			debugLevel,
			origLine,
			defPtr,
			id,
#endif
			esc,&front,&back,iCase,isLit,&localLambda)) !=
	    PARSE_OK)
	  return PARSE_FAIL;

	/* anything in the list that isn't a lambda means the whole
	 * expression can't be empty and isn't a lambda.
	 */
	if (!localLambda)
	  *isLambda = 0;

	assert(front != NULL);
	assert(back != NULL);

	if (*head == NULL){
	  *head = front;
	  *end  = back;
	} else {
	  /* hook new stuff onto the tree */
	  (*end)->choice0 = front;
	  *end = back;
	}
      }
    }
    if (*head == NULL){
      /* this resulted in an empty tree...fake a NOP so the caller
       * doesn't have to special case for this....
       */
      if ((newParseEl(mem,
#ifndef NODBG
                      id,
#endif
		      &back)) != PARSE_OK){
	return PARSE_FAIL;
      }
      back->type = NOP;
      *head = *end = back;
      *isLambda = 1;
    } else {
      assert(*end != NULL);
    }
  }
  assert(*head != NULL);
  assert(*end != NULL);

#ifndef NODBG
  if (debugLevel > 2){
    /* *head, *end can't be NULL */
    fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
	    (*head)->id, addr - *defPtr);
    fprintf(stdout,"DEBUG: N%d begins at offset %d\n",
	    (*end)->id, addr - *defPtr);
  }
#endif
  return PARSE_OK;
}

#ifndef NODBG
char *typeArray[] = {
  "PEND  ", "EXACT ", "ICASE ", "VAR_B ", "VAR_E ", "RANGE ",
  "REG_B ", "REG_E ", "FBYTE ", "NOP   ", "PERROR" };

static int
dumpEscape(p,level)
struct parseEl *p;
{
  int i;
  char buff0[16],buff1[16];
  if (p == NULL || p->visited)
    return PARSE_OK;

  fputs("DEBUG: ",stdout);

  p->visited = 1;
  for (i=0; i<level;i++){
    fputc(' ',stdout);
  }
  if (p->choice0){
    sprintf(buff0,"%04d",p->choice0->id);
  } else {
    strcpy(buff0,"----");
  }
  if (p->choice1){
    sprintf(buff1,"%04d",p->choice1->id);
  } else {
    strcpy(buff1,"----");
  }
  (void)
    fprintf(stdout,
	  "N%04d: 0:%s 1:%s %s =>",
	  p->id, buff0, buff1, typeArray[p->type]);
  if (p->type == EXACT || p->type == ICASE){
    (void) fprintf(stdout,"char = %d\n",p->u.byte);
  } else {
    if (p->type == VAR_B || p->type == VAR_E){
      (void) fprintf(stdout,"varindex = %d\n",p->u.var);
    } else {
      if (p->type == REG_B || p->type == REG_E){
	(void) fprintf(stdout,"escname = %s\n", p->u.name);
      } else {
	if (p->type == RANGE || p->type == NOP){
	  (void) fputc('\n',stdout);
	} else {
	  if (p->type == PEND){
	    /* no action */
	  } else {
	    (void) fprintf(stdout, 
			   "unknown type %d found in dumpEscape()\n",p->type);
	    return PARSE_FAIL;
	  }
	}
      }
    }
  }
  if (p->type != REG_B){
    if ((dumpEscape(p->choice1,level+1)) != PARSE_OK)
      return PARSE_FAIL;
  }
  return(dumpEscape(p->choice0,level+1));
}

static void
clearVisited(struct parseEl *p)
{
  if (p == NULL || p->visited == 0)
    return;
  p->visited = 0;
  clearVisited(p->choice1);
  clearVisited(p->choice0);
}
#endif
/**************************************************************************/
/*************************** reading the body section *********************/
/**************************************************************************/

static int
parseBodyLine(ISL_MEMORY_BUFFER_PTR *mem, char *s,
#ifndef NODBG
	      int debugLevel,
	      char **defPtr,
	      int *id,
#endif
	      struct escapeTable *esc,
	      struct parseEl **pfront,
	      struct parseEl **pback
	      )
{
  /* we have a line in hand.  We want to make a series of parse nodes
   * out of it.
   */
  struct parseEl *fbyte = NULL;
  struct parseEl *end = NULL;
  char buff[LINE_MAX];
  int result;

  assert(*pfront == NULL);

  buff[0] = '\0';

  while (*s != '\0'){  
#ifndef NODBG
	char *startOfCur = s;
#endif
    if (*s == '<'){
      /* might be an escape */
      struct parseEl *efront, *eback;
      int dummy;
      if ((parseDefinedName(mem, &s,
#ifndef NODBG
			    debugLevel,
			    defPtr,
			    id,
#endif
			    esc,&efront,&eback,&dummy,&result)) != PARSE_OK)
	return PARSE_FAIL;
      if (!result){
	/* call failed...just add the next byte */
	if (fbyte != NULL){
	  if ((addByte(mem, fbyte,*s,0)) != PARSE_OK)
	    return PARSE_FAIL;
	} else {
	  if ((newParseEl(mem,
#ifndef NODBG
			  id,
#endif
			  &fbyte)) != PARSE_OK){
	    return PARSE_FAIL;
	  }
	  fbyte->type = FBYTE;
	  fbyte->u.fbytes = NULL;
#ifndef NODBG
	  if (debugLevel > 2){
	    fprintf(stdout,
		    "DEBUG: N%d begins at offset %d\n",
		    fbyte->id, startOfCur - *defPtr);
	  }
#endif
	  if ((addByte(mem,fbyte,*s,strlen(s)+1)) != PARSE_OK)
	    return PARSE_FAIL;
	  if (*pfront == NULL){
	    assert(end == NULL);
	    *pfront = fbyte;
	  } else {
	    assert(end != NULL);
	    end->choice0 = fbyte;
	  }
	  end = fbyte;
	}
	s++;
      } else {
	/* call succeeded */
	if (*pfront == NULL){
	  assert(end == NULL);
	  *pfront = efront;
	} else {
	  assert(end != NULL);
	  end->choice0 = efront;
	}
#ifndef NODBG
	if (debugLevel > 2){
	  fprintf(stdout,
		  "DEBUG: N%d begins at offset %d\n",
		  efront->id, startOfCur - *defPtr);
	}
#endif
	end = eback;
	/* whatever "fbyte" we were working on is dead */
	if (fbyte){
	  if ((addByte(mem,fbyte,'\0',0)) != PARSE_OK)
	    return PARSE_FAIL;
	  fbyte = NULL;
	}
      }
    } else {
      if (fbyte != NULL){
	if ((addByte(mem,fbyte,*s,0)) != PARSE_OK)
	  return PARSE_FAIL;
      } else {
	if ((newParseEl(mem,
#ifndef NODBG
			id,
#endif
			&fbyte)) != PARSE_OK){
	  return PARSE_FAIL;
	}
	fbyte->type = FBYTE;
	fbyte->u.fbytes = NULL;
#ifndef NODBG
	if (debugLevel > 2){
	  fprintf(stdout,
		  "DEBUG: N%d begins at offset %d\n",
		  fbyte->id, startOfCur - *defPtr);
	}
#endif
	if ((addByte(mem,fbyte,*s,strlen(s)+1)) != PARSE_OK)
	  return PARSE_FAIL;
	if (*pfront == NULL){
	  assert(end == NULL);
	  *pfront = fbyte;
	} else {
	  assert(end != NULL);
	  end->choice0 = fbyte;
	}
	end = fbyte;
      }
      s++;
    }
  }
  if (fbyte)
    if ((addByte(mem,fbyte,'\0',0)) != PARSE_OK)
      return PARSE_FAIL;
  *pback = end;
  return PARSE_OK;
}

static int
parseBody(ISL_MEMORY_BUFFER_PTR *mem,
	  struct contentData *body,
#ifndef NODBG
	  int debugLevel,
	  int *id,
#endif
	  struct escapeTable *esc,
	  struct parseEl **ret)
{
  struct parseEl *head, *end,*front,*back;
#ifndef NODBG
  char *defPtr;
#endif 
  char *ptr;
  char buff[LINE_MAX];
  head = end = NULL;

  if ((getline(body,
#ifndef NODBG
	       debugLevel,
#endif
	       buff, &ptr)) != PARSE_OK){
    return PARSE_FAIL;
  }
  while (ptr != NULL){
    front = back = NULL;
#ifndef NODBG
    defPtr = ptr;
#endif
    if ((parseBodyLine(mem, ptr,
#ifndef NODBG
		       debugLevel,
		       &defPtr,
		       id,
#endif
		       esc,&front,&back)) != PARSE_OK)
      return PARSE_FAIL;
    assert(front != NULL);
    assert(back != NULL);

    if (head == NULL){
      assert(end == NULL);
      head = front;
      end = back;
    } else {
      assert(end != NULL);
      end->choice0 = front;
      end = back;
    }
    if ((getline(body,
#ifndef NODBG
		 debugLevel,
#endif
		 buff, &ptr)) != PARSE_OK){
      return PARSE_FAIL;
    }
  }
  assert(head != NULL);
  assert(end->choice0 == NULL);

  /* and an END node */
  if ((newParseEl(mem,
#ifndef NODBG
		  id,
#endif
		  &(end->choice0)
		  )) != PARSE_OK){
    return PARSE_FAIL;
  }
  end->choice0->type = PEND;
  end->choice0->choice0 = end->choice0->choice1 = NULL;
  *ret = head;
  return PARSE_OK;
}

static int
parseDefines(ISL_MEMORY_BUFFER_PTR *mem,
	     struct contentData *defines,
#ifndef NODBG
	     int debugLevel,
	     char origLine[LINE_MAX],
	     int *id,
#endif
	     struct escapeTable *esc)
{
  char *p2;
  struct parseEl *tree, *end;
  int isLambda;
  char buff[LINE_MAX];
  char *defPtr;

  /* zero the escapeTable */
  esc->escapesUsed = esc->escapesAvail = 0;
  esc->table = NULL;
  
  if ((getline(defines,
#ifndef NODBG
	       debugLevel,
#endif
	       buff, &defPtr)) != PARSE_OK){
    return PARSE_FAIL;
  }
  while (defPtr != NULL){
#ifndef NODBG
    /* preserve the line in case we have to cite it */
    if (strlen(defPtr) >= LINE_MAX - 1){
      (void) fprintf(stdout,
		     "line longer than %d bytes cannot be dealt with\n",
		     LINE_MAX - 1);
      return PARSE_FAIL;
    }
    strcpy(origLine,defPtr);
#endif
    /* line should be of the form: "name:REGEX\n" */
    p2 = buff;
    while (*p2 != '\0' && (ISALPHA(*p2) || ISDIGIT(*p2) || (*p2 == '_')))
      p2++;
    if (*p2 != ':'){
#ifndef NODBG
      (void) fprintf(stdout,"FATAL: found unexpected character %c ",*p2);
      *p2 = '\0';
      (void) fprintf(stdout,"after name %s\n", buff);
#endif
      return PARSE_FAIL;
    }
    *p2 = '\0';
    p2++;
    tree = lookupEscape(buff,esc);
    if (tree != NULL){
#ifndef NODBG
      (void)
	fprintf(stdout,"FATAL: escape name %s already defined\n", *defPtr);
#endif
      return PARSE_FAIL;
    }
    /* if we're still here, we're ready to try to build the regex that
     * started after the colon.
     */
    if ((parseRegString(mem, &p2,
#ifndef NODBG
			debugLevel,
			origLine,
			&defPtr,
			id,
#endif
			esc,&tree,&end,0,0,&isLambda)) != PARSE_OK)
      return PARSE_FAIL;

    assert(end->type == REG_E);
    assert((strlen(buff)) < MAXNAME);

    strcpy(end->u.name,buff);
    if ((addEscape(mem, buff,
#ifndef NODBG
		   debugLevel,
#endif
		   esc,tree,isLambda)) != PARSE_OK)
      return PARSE_FAIL;
#ifndef NODBG
    if (debugLevel > 2){
      (void)
	fprintf(stdout,"DEBUG: dumping tree for escape <%s>\n", buff);
      if ((dumpEscape(tree,0)) != PARSE_OK){
	return PARSE_FAIL;
      }
      clearVisited(tree);
    }
#endif
    if ((getline(defines,
#ifndef NODBG
		 debugLevel,
#endif
		 buff, &defPtr)) != PARSE_OK){
      return PARSE_FAIL;
    }
  }
  return PARSE_OK;
}


int
parseMain(ISL_ARCHIVE_CONTEXT_PTR ArchivePtr,
	  ISL_CONST_DATA_PTR defines,
	  ISL_CONST_DATA_PTR dataTemplate,
	  ISL_CONST_DATA_PTR content,
#ifndef NODBG
	  int debugLevel,
#endif
	  ISL_DATA_CHAIN_PTR *results,
	  int maxResult
	  )
{
  ISL_MEMORY_BUFFER_PTR lmem = NULL;
  struct parseEl *tree;
  struct escapeTable escapes; 
  struct valueStack vS;
  struct contentData cD;
#ifndef NODBG
  char origLine[LINE_MAX];
  int id = 0;
#endif
  int i;

  /* make sure the results chain is currently empty */
  for (i = 0; i < maxResult; i++){
    results[i] = NULL;
  }

  escapes.table = NULL;
  escapes.escapesUsed = escapes.escapesAvail = 0;
  if (defines){
    cD.ptr = defines;
    cD.offset = 0;

  /* parse the "definitions" we were handed */
    if ((parseDefines(&lmem, &cD,
#ifndef NODBG
		      debugLevel,
		      origLine,
		      &id,
#endif
		      &escapes)) != PARSE_OK)
      goto CLEANUP_AND_FAIL;
  }

  /* using the definitions created, build a parse tree from the dataTemplate. */
  if (!dataTemplate){
    goto CLEANUP_AND_FAIL;
  }
  cD.ptr = dataTemplate;
  cD.offset = 0;
  if ((parseBody(&lmem, &cD,
#ifndef NODBG
		 debugLevel,
		 &id,
#endif
		 &escapes,
		 &tree)) != PARSE_OK){
    goto CLEANUP_AND_FAIL;
  }

  /* now, use the tree to parse the content and to build up a value stack. */
  if (!content){
    goto CLEANUP_AND_FAIL;
  }
  cD.ptr = content;
  cD.offset = 0;
  if ((parseContent(&lmem, &cD,
#ifndef NODBG
		    debugLevel,
#endif
		    tree,&vS)) != PARSE_OK)
    goto CLEANUP_AND_FAIL;

  /* walk the value stack, populating our results */
  if (!results){
    goto CLEANUP_AND_FAIL;
  }
  /* note that we allocate the structures we're going to return with
   * the caller's mem pointer
   */
  if ((populateResults(ArchivePtr, &vS,&cD,results,maxResult)) != PARSE_OK)
    goto CLEANUP_AND_FAIL;

  /* we're done with our "local" memory */
  recycleMemory(lmem);
  return PARSE_OK;

CLEANUP_AND_FAIL:
  recycleMemory(lmem);
  return(0);
}
