/*-----------------------------------------------------------------------
 * File: SIGROOT.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 

/* 
 * Signature Root Public Methods
 *
 * EISL_CreateVerifiedSignatureRoot
 * EISL_CreateVerifiedSignatureRootWithCertificate
 * EISL_GetCertificateChain
 * EISL_FindManifestSection
 * EISL_CreateManifestSectionEnumerator
 * EISL_GetNextManifestSection
 * EISL_RecycleManifestSectionEnumerator
 * EISL_CreateSignatureAttributeEnumerator
 * EISL_GetNextSignatureAttribute
 * EISL_RecycleSignatureAttributeEnumerator
 * EISL_FindSignatureAttribute
 * EISL_RecycleVerifiedSignatureRoot
 */

#include <stdio.h>
#include "eislapi.h"
#include "isl_internal.h"
#include "ber_bn.h"
#include "callouts.h"
#include <isl_memtrack.h>


/* declaration of private methods */


static ISL_VERIFIED_SIGNATURE_ROOT_PTR
VerifySignatureRootWithCert(
	ISL_SIGNATURE_CONTEXT_PTR pSignatureContext, 
	ISL_VERIFIED_CERTIFICATE_PTR Cert);


/* Methods for ISL_SIGNATURE_CONTEXT */

/*
 * EISL_CreateVerifiedSignatureRoot
 *
 * This function directly verifies the digital signature specified by the credentials 
 * with the given public key.  It does NOT construct certificate chains, but must use 
 * the key directly.  If the credentials support multiple signers, the Signer parameter 
 * can be used to determine which signer is verified.
 *
 * The manifest sections may be enumerated using the object created by 
 * EISL_CreateManifestSectionEnumerator.
 *
 * The function does NOT verify the objects referenced in the manifest sections.  
 * However, the manifest sections are verified, and the attributes in the sections 
 * may be trusted.  
 */
ISL_VERIFIED_SIGNATURE_ROOT_PTR 
EISL_CreateVerifiedSignatureRoot(
	const ISL_CONST_DATA Credentials,
	const ISL_CONST_DATA Signer,
	const ISL_CONST_DATA PublicKey)
{
	ISL_ARCHIVE_CONTEXT_PTR		pArchive  = NULL;	
	ISL_VERIFIED_SIGNATURE_ROOT_PTR pVSigRoot = NULL;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x00);

	pArchive = (ISL_ARCHIVE_CONTEXT_PTR) malloc(sizeof(ISL_ARCHIVE_CONTEXT));
	if (!pArchive)
		return NULL;

	if (ISL_OK != isl_InitArchiveContextFromCredentials(pArchive, Credentials))
		goto CLEANUP_AND_FAIL;

	pVSigRoot = isl_CreateVerifiedSignatureRoot(pArchive, Signer, PublicKey);
	if (!pVSigRoot) goto CLEANUP_AND_FAIL;
	return pVSigRoot;

CLEANUP_AND_FAIL:
	(void)isl_RecycleArchiveContext(pArchive);
	free(pArchive);
	return NULL;
}

/*
 * The following API accepts manifest in a buffer and manifest path 
 * as parameters instead of manifest file.
 */ 
ISL_VERIFIED_SIGNATURE_ROOT_PTR 
EISL_CreateVerifiedSignatureRootWithCredentialData(
	const ISL_CONST_DATA CredentialsImage,
	const ISL_CONST_DATA ModuleSearchPath,
	const ISL_CONST_DATA Signer,
	const ISL_CONST_DATA PublicKey)
{
	ISL_ARCHIVE_CONTEXT_PTR		pArchive  = NULL;	
	ISL_VERIFIED_SIGNATURE_ROOT_PTR pVSigRoot = NULL;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x00);

	pArchive = (ISL_ARCHIVE_CONTEXT_PTR) malloc(sizeof(ISL_ARCHIVE_CONTEXT));
	if (NULL == pArchive)
		return NULL;

	if (ISL_OK != isl_InitArchiveContextFromCredentialsBuffer(pArchive, CredentialsImage, ModuleSearchPath))
		goto CLEANUP_AND_FAIL;

	pVSigRoot = isl_CreateVerifiedSignatureRoot(pArchive, Signer, PublicKey);
	if (!pVSigRoot) goto CLEANUP_AND_FAIL;
	return pVSigRoot;

CLEANUP_AND_FAIL:
	(void)isl_RecycleArchiveContext(pArchive);
	free(pArchive);
	return NULL;
}


/*
 * EISL_CreateVerifiedSignatureRootWithCertificate
 *
 * This function directly verifies the digital signature specified by the credentials 
 * with the given public key.  It does NOT construct certificate chains, but must uses 
 * the signer identification and public key in the certificate directly.  
 *
 * The manifest sections may be enumerated using the object created by 
 * EISL_CreateManifestSectionEnumerator.
 *
 * The function does NOT verify the objects referenced in the manifest sections.  
 * However, the manifest sections are verified, and the attributes in the sections 
 * may be trusted.  
 */
ISL_VERIFIED_SIGNATURE_ROOT_PTR 
EISL_CreateVerifiedSignatureRootWithCertificate(
	const ISL_CONST_DATA Credentials,
	const ISL_VERIFIED_CERTIFICATE_PTR Cert)
{
	ISL_ARCHIVE_CONTEXT_PTR			pArchive  = NULL;
	ISL_VERIFIED_SIGNATURE_ROOT_PTR pVSigRoot;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x01);

    pArchive = (ISL_ARCHIVE_CONTEXT_PTR) malloc(sizeof(ISL_ARCHIVE_CONTEXT));
	if (!pArchive) return NULL;
	if (ISL_OK != isl_InitArchiveContextFromCredentials(pArchive, Credentials))
		goto CLEANUP_AND_FAIL;
	pVSigRoot = isl_CreateVerifiedSignatureRootWithCertificate(pArchive, Cert);
	if (!pVSigRoot) goto CLEANUP_AND_FAIL;
	return pVSigRoot;

CLEANUP_AND_FAIL:
	(void)isl_RecycleArchiveContext(pArchive);
	free(pArchive);
	return NULL;
}

ISL_VERIFIED_SIGNATURE_ROOT_PTR 
EISL_CreateVerifiedSignatureRootWithCredentialDataAndCertificate(
	const ISL_CONST_DATA CredentialsImage,
	const ISL_CONST_DATA ModuleSearchPath,
	const ISL_VERIFIED_CERTIFICATE_PTR Cert)
{
	ISL_ARCHIVE_CONTEXT_PTR			pArchive  = NULL;
	ISL_VERIFIED_SIGNATURE_ROOT_PTR pVSigRoot;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x01);

    pArchive = (ISL_ARCHIVE_CONTEXT_PTR) malloc(sizeof(ISL_ARCHIVE_CONTEXT));
	if (!pArchive) return NULL;
	if (ISL_OK != isl_InitArchiveContextFromCredentialsBuffer(pArchive, CredentialsImage, ModuleSearchPath))
		goto CLEANUP_AND_FAIL;
	pVSigRoot = isl_CreateVerifiedSignatureRootWithCertificate(pArchive, Cert);
	if (!pVSigRoot) goto CLEANUP_AND_FAIL;
	return pVSigRoot;

CLEANUP_AND_FAIL:
	(void)isl_RecycleArchiveContext(pArchive);
	free(pArchive);
	return NULL;
}

ISL_VERIFIED_CERTIFICATE_CHAIN_PTR 
EISL_GetCertificateChain(
	const ISL_VERIFIED_MODULE_PTR Module)
{
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x02);
	if (!Module)
		return NULL;

	return Module->ManSect->Parent->Signatures->Signature->Signers->CertChain;
}

ISL_MANIFEST_SECTION_PTR 
EISL_FindManifestSection(
	const ISL_VERIFIED_SIGNATURE_ROOT_PTR Root,
	const ISL_CONST_DATA Name)
{
	ISL_MANIFEST_SECTION_PTR pManSect = NULL;
	ISL_ARCHIVE_CONTEXT_PTR pArchive = NULL;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x03);
	if (!Root)
		return NULL;

	pArchive = Root->Parent->Archive;

	if (!pArchive)
		return NULL;

	for(pManSect = pArchive->Manifest.Section;
		pManSect;
		pManSect = pManSect->Next)
	{
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x04);
		if (IS_EQUAL(Name, pManSect->SectionInfo.Name) ||
			IS_EQUAL(Name, pManSect->SectionInfo.SectionName) )
		{
			return pManSect;
		}
	}
	return NULL;
}

ISL_ITERATOR_PTR 
EISL_CreateManifestSectionEnumerator(
	const ISL_VERIFIED_SIGNATURE_ROOT_PTR Root)
{
	ISL_ITERATOR_PTR		pItr = NULL;
	ISL_ARCHIVE_CONTEXT_PTR pArchive = NULL;
	ISL_DATA				ManifestSectionPointerArray = {0, NULL};
	ISL_MANIFEST_SECTION_PTR pManSect = NULL;
	ISL_MANIFEST_SECTION_PTR *ManifestSections = NULL;
	unsigned long i = 0;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x05);
	if (!Root)
		return NULL;

	pArchive = Root->Parent->Archive;

	if (!pArchive)
		return NULL;

	if (!pArchive->Manifest.SectionCount)
		return NULL;

	ManifestSectionPointerArray.Length = sizeof(ISL_MANIFEST_SECTION_PTR) *
										 pArchive->Manifest.SectionCount;

	ManifestSections = isl_AllocateMemory(pArchive,
										  ManifestSectionPointerArray.Length);
	if (!ManifestSections) {
		REPORT_ALLOCATE_MEMORY_ERROR(pArchive,
									 ManifestSectionPointerArray.Length);
		return NULL;
	}
	for(pManSect = pArchive->Manifest.Section, i = 0;
		pManSect;
		pManSect = pManSect->Next, i++)
	{
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x06);

		ManifestSections[i] = pManSect;
	}

	ManifestSectionPointerArray.Data = (void*)ManifestSections;

	pItr = isl_CreateIterator(pArchive, &ManifestSectionPointerArray);
	isl_FreeMemory(pArchive, ManifestSections);

	if (!pItr) {
		REPORT_CREATE_ITERATOR_ERROR(pArchive, &ManifestSectionPointerArray);
		return NULL;
	}
	return pItr;
}

ISL_MANIFEST_SECTION_PTR 
EISL_GetNextManifestSection(
	const ISL_ITERATOR_PTR Iterator)
{
	void * node = isl_GetNextNode(Iterator);
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x07);
	return (ISL_MANIFEST_SECTION_PTR) node;
}

ISL_STATUS 
EISL_RecycleManifestSectionEnumerator(
	const ISL_ITERATOR_PTR Iterator)
{
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x08);
	return isl_RecycleIterator(Iterator);
}

ISL_ITERATOR_PTR 
EISL_CreateSignatureAttributeEnumerator(
	const ISL_VERIFIED_SIGNATURE_ROOT_PTR Verification)
{
	ISL_ITERATOR_PTR			pIterator = NULL;
	ISL_ARCHIVE_CONTEXT_PTR		pArchive = NULL;
	ISL_NAME_VALUE_PTR			*Attributes = NULL;
	ISL_NAME_VALUE_PTR			pAttribute = NULL;
	ISL_DATA					AttributePtrArray = {0, NULL};
	uint32						i = 0, AttCount = 0;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x09);
	if (!Verification)
		return NULL;

	pArchive = Verification->Parent->Archive;

	/* Count up the authenticated attributes */
	pAttribute = Verification->Attributes;
	while (pAttribute) {
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x09);
		AttCount++;
		pAttribute = pAttribute->Next;
	}

	AttributePtrArray.Length = sizeof(ISL_NAME_VALUE_PTR) * AttCount;
	Attributes = isl_AllocateMemory(pArchive, AttributePtrArray.Length);
	if (!Attributes) {
		REPORT_ALLOCATE_MEMORY_ERROR(pArchive, AttributePtrArray.Length);
		return NULL;
	}
	pAttribute = Verification->Attributes;
	for (i=0; i<AttCount; i++)
	{
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x0a);
		Attributes[i] = pAttribute;
		pAttribute = pAttribute->Next;
	}

	AttributePtrArray.Data = (void*)Attributes;

	pIterator = isl_CreateIterator(pArchive, &AttributePtrArray);
	isl_FreeMemory(pArchive, Attributes);

	if (!pIterator) {
		REPORT_CREATE_ITERATOR_ERROR(pArchive, &AttributePtrArray);
		return NULL;
	}

	return pIterator;
}

ISL_STATUS 
EISL_GetNextSignatureAttribute(
	const ISL_ITERATOR_PTR Iterator,
	ISL_CONST_DATA_PTR Name,
	ISL_CONST_DATA_PTR Value)
{
	ISL_NAME_VALUE_PTR	node = NULL;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x0b);
	if (!Iterator || !Name || !Value)
		return ISL_FAIL;
	
	node = isl_GetNextNode(Iterator);
	if (!node)
		return ISL_FAIL;

	memcpy(Name, &node->Name, sizeof(ISL_DATA));
	memcpy(Value, &node->Value, sizeof(ISL_DATA));
	return ISL_OK;
}

ISL_STATUS 
EISL_RecycleSignatureAttributeEnumerator(
	const ISL_ITERATOR_PTR Iterator)
{
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x0c);
	return isl_RecycleIterator(Iterator);
}

ISL_STATUS 
EISL_FindSignatureAttribute(
	ISL_VERIFIED_SIGNATURE_ROOT_PTR Root,
	ISL_CONST_DATA Name,
	ISL_CONST_DATA_PTR Value)
{
	ISL_ITERATOR_PTR	pIterator = NULL;
	ISL_CONST_DATA		testName = {0, NULL}, testValue = {0, NULL};

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x0d);
	if (!Root || !Value || !Name.Length)
		return ISL_FAIL;

	pIterator = EISL_CreateSignatureAttributeEnumerator(Root);
	if (!pIterator)
		return ISL_FAIL;

	while (ISL_OK == EISL_GetNextSignatureAttribute(pIterator, &testName, &testValue))
	{
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x0e);
		/* Compare the names */
		if ( IS_EQUAL(testName, Name) ) 
		{
			memcpy(Value, &testValue, sizeof(ISL_CONST_DATA));
			(void)EISL_RecycleSignatureAttributeEnumerator(pIterator);
			return ISL_OK;
		}
	}

	(void)EISL_RecycleSignatureAttributeEnumerator(pIterator);
	return ISL_FAIL;
}

ISL_STATUS 
EISL_RecycleVerifiedSignatureRoot(
	const ISL_VERIFIED_SIGNATURE_ROOT_PTR Root)
{
	ISL_ARCHIVE_CONTEXT_PTR pArchive;
	ISL_STATUS status = ISL_FAIL;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x0f);

	if (Root)
	{
		if (NULL == Root->Parent)
			return ISL_FAIL;
		pArchive = Root->Parent->Archive;
		status = isl_RecycleArchiveContext(pArchive);
		free (pArchive);
	}

	return status;
}

/* protected methods for ISL_SIGNATURE_CONTEXT */
ISL_STATUS
isl_InitSignerContext(
	ISL_SIGNATURE_CONTEXT_PTR	pSignatureContext,
	ISL_SIGNER_CONTEXT_PTR		pSignerContext)
{
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x10);
	if (!pSignerContext) return ISL_FAIL;
    memset(pSignerContext, 0, sizeof(ISL_SIGNATURE_CONTEXT));
    pSignerContext->Parent = pSignatureContext;
	return ISL_OK;
}


ISL_VERIFIED_SIGNATURE_ROOT_PTR
isl_CreateVerifiedSignatureRoot(
	const ISL_ARCHIVE_CONTEXT_PTR	Archive,
	const ISL_CONST_DATA			Signer,
	const ISL_CONST_DATA			PublicKey)
{
	ISL_SIGNER_CONTEXT_PTR pSignerContext = NULL;
	ISL_SIGNATURE_LIST_PTR pSigList = NULL;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x17);

	if (!Archive->Signatures) return NULL;

	pSigList = Archive->Signatures;

	if (!Signer.Data)
	{
		while (pSigList) {
			if (!pSigList->Signature || !pSigList->Signature->Signers)
				return NULL;

			pSignerContext = pSigList->Signature->Signers;

			if (isl_VerifySignature(pSignerContext, PublicKey) == ISL_OK) {
				/* Recycle all signers, except for the verified one. */
				if (isl_recycleSignature (Archive, pSigList->Signature) != ISL_OK)
					return NULL;

				return pSignerContext;
			}
			pSigList = pSigList->Next;
		}
	} else {
		/*
		 * If signer id is specified, match the signer id.
		 */
		while (pSigList) {
			pSignerContext = pSigList->Signature->Signers;

			while (pSignerContext) {
				if (IS_EQUAL(pSignerContext->SignerID, Signer)) {
					if (isl_VerifySignature(pSignerContext, PublicKey) == ISL_OK) {
						/* Recycle all signers, except for the verified one. */
						if (isl_recycleSignature (Archive, pSigList->Signature) != ISL_OK)
							return NULL;

						return pSignerContext;
					} else {
						return NULL;
					}
				}
				pSignerContext = pSignerContext->Next;
			}
			pSigList = pSigList->Next;
		}
	}

	return NULL;
}

ISL_VERIFIED_SIGNATURE_ROOT_PTR
isl_CreateVerifiedSignatureRootWithCertificate(
	const ISL_ARCHIVE_CONTEXT_PTR		Archive,
	const ISL_VERIFIED_CERTIFICATE_PTR	Cert)
{
	ISL_SIGNATURE_LIST_PTR pSigList = NULL;
	ISL_SIGNER_CONTEXT_PTR	pSignerContext;
	ISL_CONST_DATA			PublicKey;
	ISL_CONST_DATA			Signer;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x18);	
	
	if (!Archive->Signatures) 
		return NULL;

	if (isl_X509Cert_GetID(Cert, &Signer) != ISL_OK)
		return NULL;

	if (isl_X509Cert_GetKey(Cert, &PublicKey) != ISL_OK) 
		return NULL;

	pSigList = Archive->Signatures;

	while (pSigList) {
		if (!pSigList->Signature || !pSigList->Signature->Signers)
			return NULL;
		pSignerContext = pSigList->Signature->Signers;

		while (pSignerContext) {
			if (IS_EQUAL(pSignerContext->SignerID, Signer)) {
				if (isl_VerifySignature(pSignerContext, PublicKey) == ISL_OK) {
					/* Recycle all signers, except for the verified one. */
					if (isl_recycleSignature (Archive, pSigList->Signature) != ISL_OK)
							return NULL;
					return pSignerContext;
				} else {
					return NULL;
				}
			}
			pSignerContext = pSignerContext->Next;
		}
		pSigList = pSigList->Next;
	}
	return NULL;
}

ISL_STATUS
isl_VerifySignature(
	ISL_SIGNER_CONTEXT_PTR	pSignerContext,
	ISL_CONST_DATA			PublicKey)
{
	ISL_SIGNATURE_CONTEXT_PTR pSignatureContext;
	ISL_CONST_DATA			AlgParams;
	ISL_KEY_BLOCK			KeyBlock;
	EICLData					Message;
    EICLData					rs[2];
    EICLPublicKey			Key;
	uint8					work[6][4*MODULUS];
	ISL_KEY_BLOCK	*KeyToVerify;


    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x19);
	pSignatureContext = pSignerContext->Parent;
	rs[0].value = &work[0][0];		rs[0].length = MODULUS * 4;
	rs[1].value = &work[1][0];		rs[1].length = MODULUS * 4;

	if (BER_Decode((uint8 *)pSignerContext->Signature.Data, pSignerContext->Signature.Length, rs, 2) != 2) 
		return ISL_FAIL;

	Message.value = (uint8 *)pSignatureContext->SigSections->Image.Data;
	Message.length = (long) pSignatureContext->SigSections->Image.Length;

		/* Assign Storage for DSS Key */
	Key.PrimeModulus.value = &work[2][0];    /* p */
	Key.PrimeModulus.length = MODULUS * 4;   /* p */

	Key.PrimeDivisor.value = &work[3][0];    /* q */
	Key.PrimeDivisor.length = MODULUS * 4;   /* q */

	Key.OrderQ.value = &work[4][0];          /* g */
	Key.OrderQ.length = MODULUS * 4;         /* g */

	Key.PublicKey.value = &work[5][0];       /* y */
	Key.PublicKey.length = MODULUS * 4;      /* y */

	if (!PublicKey.Length) {

		ISL_CONST_DATA tmpKey;
		uint8 szPublicKey[1024];

		tmpKey.Length = 1024;
		tmpKey.Data = szPublicKey;

		(void) EISL_RetrieveRootIssuerKey((ISL_CALLOUT_DATA*)&tmpKey);

		KeyToVerify = (ISL_KEY_BLOCK*)&tmpKey;
	} else {
		KeyToVerify = (ISL_KEY_BLOCK*)&PublicKey;
	}


	/* Decode Key */
	if (isl_ParseKey(KeyToVerify, NULL, &AlgParams, &KeyBlock) != ISL_OK)
		return ISL_FAIL;

	if (isl_ParseKeyBlock(KeyBlock, &AlgParams, (EICLData *)&Key) != ISL_OK)
		return ISL_FAIL;

	/*Verify the List of Signable Objects */
	if (EICL_DSSVerify(&Message, &rs[0], &rs[1], &Key) != 0) return ISL_FAIL;

	return ISL_OK;
}


ISL_ITERATOR_PTR 
EISL_CreateSignatureInfoSectionAttributeEnumerator(
				const ISL_SIG_SECTION_PTR Section)
{
	ISL_ITERATOR_PTR			pIterator = NULL;
	ISL_ARCHIVE_CONTEXT_PTR		pArchive = NULL;
	ISL_NAME_VALUE_PTR			*Attributes = NULL;
	ISL_NAME_VALUE_PTR			pAttribute = NULL;
	ISL_DATA					AttributePtrArray = {0, NULL};
	uint32						i = 0, AttCount = 0;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x09);
	if (!Section)
		return NULL;

	pArchive = Section->Parent->Archive;

	/* Count up the authenticated attributes */
	pAttribute = Section->SectionInfo.Attribute;
	while (pAttribute) {
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x09);
		AttCount++;
		pAttribute = pAttribute->Next;
	}

	AttributePtrArray.Length = sizeof(ISL_NAME_VALUE_PTR) * AttCount;
	Attributes = isl_AllocateMemory(pArchive, AttributePtrArray.Length);
	if (!Attributes) {
		REPORT_ALLOCATE_MEMORY_ERROR(pArchive, AttributePtrArray.Length);
		return NULL;
	}
	pAttribute = Section->SectionInfo.Attribute;
	for (i=0; i<AttCount; i++)
	{
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x0a);
		Attributes[i] = pAttribute;
		pAttribute = pAttribute->Next;
	}

	AttributePtrArray.Data = (void*)Attributes;

	pIterator = isl_CreateIterator(pArchive, &AttributePtrArray);
	isl_FreeMemory(pArchive, Attributes);

	if (!pIterator) {
		REPORT_CREATE_ITERATOR_ERROR(pArchive, &AttributePtrArray);
		return NULL;
	}

	return pIterator;
}

ISL_STATUS 
EISL_GetNextSignatureInfoSectionAttribute(
	const ISL_ITERATOR_PTR Iterator,
	ISL_CONST_DATA_PTR Name,
	ISL_CONST_DATA_PTR Value)
{
	ISL_NAME_VALUE_PTR	node = NULL;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x0b);
	if (!Iterator || !Name || !Value)
		return ISL_FAIL;
	
	node = isl_GetNextNode(Iterator);
	if (!node)
		return ISL_FAIL;

	memcpy(Name, &node->Name, sizeof(ISL_DATA));
	memcpy(Value, &node->Value, sizeof(ISL_DATA));
	return ISL_OK;
}

ISL_STATUS 
EISL_RecycleSignatureInfoSectionAttributeEnumerator(
	const ISL_ITERATOR_PTR Iterator)
{
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x0c);
	return isl_RecycleIterator(Iterator);
}


ISL_STATUS 
EISL_FindSignatureInfoSectionAttribute(
	ISL_SIG_SECTION_PTR Section,
	ISL_CONST_DATA Name,
	ISL_CONST_DATA_PTR Value)
{
	ISL_ITERATOR_PTR	pIterator = NULL;
	ISL_CONST_DATA		testName = {0, NULL}, testValue = {0, NULL};

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x0d);
	if (!Section || !Value || !Name.Length)
		return ISL_FAIL;

	pIterator = EISL_CreateSignatureInfoSectionAttributeEnumerator(Section);
	if (!pIterator)
		return ISL_FAIL;

	while (ISL_OK == EISL_GetNextSignatureInfoSectionAttribute(pIterator, &testName, &testValue))
	{
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_SIGROOT_TAG,0x0e);
		/* Compare the names */
		if ( IS_EQUAL(testName, Name) ) 
		{
			memcpy(Value, &testValue, sizeof(ISL_CONST_DATA));
			(void)EISL_RecycleSignatureInfoSectionAttributeEnumerator(pIterator);
			return ISL_OK;
		}
	}

	(void)EISL_RecycleSignatureInfoSectionAttributeEnumerator(pIterator);
	return ISL_FAIL;
}


