/*-----------------------------------------------------------------------
 * File: INTEGAPI.H
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

/* Integrity verification tool and library API
 */

#ifndef ISL_API_H
#define ISL_API_H
#include "isltype.h"

#include "isl_cssm.h"
#include "aal.h"
					   
#include <stdlib.h>

/*
 * Algid's for signature/hashing algorithms.
 */
#define ISL_ALGID_MD5				"MD5"
#define ISL_ALGID_MD2				"MD2"
#define ISL_ALGID_SHA1				"SHA-1"
#define ISL_ALGID_DSA				"DSA"
#define ISL_ALGID_DSASHA1			"DSA/SHA1"
#define ISL_ALGID_PKCS1_RSA			"PKCS1 RSA"
#define ISL_ALGID_PKCS1_RSAMD2		"PKCS1 RSA/MD2"
#define ISL_ALGID_PKCS1_RSAMD5		"PKCS1 RSA/MD5";
#define ISL_ALGID_PKCS1_RSASHA1		"PKCS1 RSA/SHA-1"


/* error codes */
#define ISL_BASE_ERROR				(0UL )    /* Defined by CSSM */
#define ISL_NO_ERROR				(0UL )
#define ISL_INVALID_INPUT			(ISL_BASE_ERROR +1L )
#define ISL_NO_MEMORY				(ISL_BASE_ERROR +2L )
#define ISL_FORGERY					(ISL_BASE_ERROR +3L )
#define ISL_MISSING_PART			(ISL_BASE_ERROR +4L )
#define ISL_UNSUPPORTED_ALGORITHM	(ISL_BASE_ERROR +5L )
#define ISL_SIGNER_NOT_FOUND		(ISL_BASE_ERROR +6L )
#define ISL_IO_ERROR				(ISL_BASE_ERROR +7L )
#define ISL_ARCHIVE_INIT_FAILED		(ISL_BASE_ERROR +8L )
#define ISL_SIGNATURE_NOT_FOUND		(ISL_BASE_ERROR +9L )
#define ISL_INVALID_PARAMETER		(ISL_BASE_ERROR +10L )
#define ISL_ALGORITHM_NOT_FOUND		(ISL_BASE_ERROR +11L )
#define ISL_OBJECT_NOT_FOUND		(ISL_BASE_ERROR +12L )
#define ISL_PATH_TOO_LONG			(ISL_BASE_ERROR +13L )
#define ISL_DIGEST_NOT_FOUND		(ISL_BASE_ERROR +14L )
#define ISL_INVALID_SIGNATURE_FILE  (ISL_BASE_ERROR +15L )
#define ISL_UNABLE_TO_OPEN_FILE		(ISL_BASE_ERROR +16L )
#define ISL_INVALID_SIGNATURE		(ISL_BASE_ERROR +17L )
#define ISL_INVALID_CERTIFICATE		(ISL_BASE_ERROR +18L )
#define ISL_INVALID_BER_ENCODING	(ISL_BASE_ERROR +19L )
#define ISL_OBJECT_LIST_ALREADY_DEFINED			(ISL_BASE_ERROR +20L )
#define ISL_SIGNED_LIST_FAILED_TO_INITIALIZE	(ISL_BASE_ERROR +21L )
#define ISL_FAILED_TO_SET_SIGNED_LIST			(ISL_BASE_ERROR +22L )
#define ISL_CERTIFICATE_FAILED_TO_INITIALIZE	(ISL_BASE_ERROR +23L )
#define ISL_INVALID_DIGEST_CONFIG				(ISL_BASE_ERROR +24L )
#define ISL_INVALID_SIGNER_CONTEXT				(ISL_BASE_ERROR +25L )
#define ISL_NO_SIGNED_LIST						(ISL_BASE_ERROR +26L )
#define ISL_INVALID_SIGNED_LIST					(ISL_BASE_ERROR +27L )
#define ISL_INVALID_ARCHIVE						(ISL_BASE_ERROR +28L )
#define ISL_FAILED_TO_AUTOCONFIGURE				(ISL_BASE_ERROR +29L )
#define ISL_FAILED_TO_VERIFY_MANIFEST_SECTION	(ISL_BASE_ERROR +30L )
#define ISL_UNIMPLEMENTED						(ISL_BASE_ERROR +999L )

/* temporary definitions for backwards compatibility with previous work */
#define INTEGRITY_OK				ISL_OK						
#define INTEGRITY_INVALID_INPUT		ISL_INVALID_INPUT			
#define INTEGRITY_NO_MEMORY			ISL_NO_MEMORY				
#define INTEGRITY_FORGERY			ISL_FORGERY					
#define INTEGRITY_MISSING_PART		ISL_MISSING_PART			
#define INTEGRITY_UNSUPPORTED_ALGORITHM	ISL_UNSUPPORTED_ALGORITHM
#define INTEGRITY_SIGNER_NOT_FOUND	ISL_SIGNER_NOT_FOUND	
#define INTEGRITY_IO_ERROR			ISL_IO_ERROR			
#define INTEGRITY_UNIMPLEMENTED		ISL_UNIMPLEMENTED
/* end temporary definitions */

/*
**	General methods for any object (not in the methods vector--we may not have the methods!)
*/
ISL_CLASS_PTR ISL_GetClass(void * Context);		/* returns class structure for any class */

/*
**	all services have a common routine to identify themselves,
**	providing a human readable string as well as their signature format-specific algorithm ID
*/
typedef struct isl_service_class_methods {
	void (*id)(
		ISL_SERVICE_CLASS *AlgClass,		/* return service type code */
		ISL_CONST_DATA_PTR AlgID,			/* return archive-specific encoding */
		ISL_CONST_DATA_PTR ServiceName);	/* return human-readable description */
	ISL_CLASS_PTR Class;					/* pointer to class structure */
} ISL_SERVICE_CLASS_METHODS;

/*
**	Error Reporting object methods
*/
typedef struct isl_error_methods {
	ISL_ITERATOR_PTR (*CreateErrorEnumerator)(	 /* create iterator to list errors */
		ISL_ERROR_CONTEXT *Context);

	ISL_STATUS (*GetNextError)(				/* get next error code and message, if any */
		ISL_ITERATOR_PTR Context,
		ISL_ERROR_CODE *Code,
		ISL_CONST_DATA *Message);

	ISL_STATUS (*RecycleErrorEnumerator)(	/* recycle memory for error message iterator */
		ISL_ITERATOR_PTR Context);

	ISL_STATUS (*ClearErrors)(				/* clear error list for this object */
		ISL_ERROR_CONTEXT *Context);
} ISL_ERROR_METHODS;

/* 
**	ISL_ARCHIVE_CONFIGuration class public methods (configuration)
*/
typedef struct isl_archive_config_methods{
	/* class methods */
	ISL_SIZE (*SizeofObject)();				/* returns sizeof object */
	/* object methods */
	ISL_STATUS (*Initialize)(				/* ISL_ARCHIVE_CONTEXT constructor */
		ISL_ARCHIVE_CONFIG *Context,		/* pointer to configuration context */
		ISL_MEMORY_CONTEXT *MemoryContext,	/* memory for IntegrityContext */	
		ISL_MEMORY_METHODS *MemoryMethods);	/* malloc, free methods */

	ISL_STATUS (*AutoConfigure)(
		ISL_ARCHIVE_CONFIG *Context,		/* ISL_SIGNATURE_CONTEXT constructor */
		AAL_FUNC_TABLE_PTR CssmFuncTable,	/* Array specifying the names of CSSM Api 
											   functions and fucntion pointers to be 
											   used by ISL */
		uint32 NumCssmFuncTable,			/* Number of elements in the array */
		CSP_LIST_PTR CspList);				/* Array specifying a list of CSP
											   to be loaded by ISL */
	
	ISL_STATUS (*AddAlgorithm)(
		ISL_ARCHIVE_CONFIG *Context,			/* configuration context */
		ISL_SERVICE_CLASS AlgClass,			/* service type performed by this algorithm */
		ISL_CONST_DATA AlgID,				/* archive-specific-encoded algorithm ID */
		ISL_CONST_DATA HumanName,			/* human-readable algorithm name */
		const ISL_SERVICE_CLASS_METHODS *Methods);	/* vector of methods */

	ISL_SERVICE_CLASS_METHODS *(*FindAlgorithm)(	/* return a pointer to the service methods vector */
		ISL_ARCHIVE_CONFIG *Context,		/* pointer to configuration context */
		ISL_CONST_DATA AlgID);				/* given an algorithm ID or HumanName */

	ISL_STATUS (*AddAlgorithmService)(
		ISL_ARCHIVE_CONFIG *Context,		/* pointer to configuration context */
		char *(*GetExtension)(		    	/* pointer to function returning code extension name  */
			ISL_CONST_DATA AlgID));			/* given an algorithm ID */
	
	ISL_STATUS (*Recycle)(					/* destructor for IntegrityContext */
		ISL_ARCHIVE_CONFIG *Context);		/* configuration context */

	ISL_ITERATOR_PTR (*CreateAlgorithmEnumerator)(
		ISL_ARCHIVE_CONFIG *Context,			/* pointer to configuration context */
		ISL_SERVICE_CLASS ServiceClass);		/* algorithm type */

	ISL_SERVICE_METHODS *(*GetNextAlgorithm)(
		ISL_ITERATOR_PTR Context);

	ISL_STATUS (*RecycleAlgorithmEnumerator)(
		ISL_ITERATOR_PTR Context);

	ISL_STATUS (*GetErrorObject)(			/* return context and methods for error reporting */
		ISL_ERROR_CONTEXT **Context,		/* context of error reporting object */
		const ISL_ERROR_METHODS **Methods);	/* methods for error reporting */

	ISL_CLASS_PTR *Class;					/* pointer to class structure */

} ISL_ARCHIVE_CONFIG_METHODS;
/*
**	ISL_ARCHIVE_CONTEXT class public methods
*/
typedef struct isl_archive_methods{	
	/* class methods */
	ISL_SIZE (*SizeofObject)();			/* returns sizeof object */
	
	/* object methods */
	ISL_STATUS (*InitializeNew)(			/* constructor for new archive */
		ISL_ARCHIVE_CONTEXT *Memory,		/* memory allocated for this instance */
		ISL_ARCHIVE_CONFIG *ConfigContext, 	/* my algorithm<->code extension configuration */
		ISL_CONST_DATA Path);				/* where this archive is being constructed */
	
	ISL_STATUS (*InitializeFromPath)(		/* constructor for existing archive */
		ISL_ARCHIVE_CONTEXT *Memory,		/* memory allocated for this instance */
		ISL_ARCHIVE_CONFIG *ConfigContext, 	/* my algorithm<->code extension configuration */
		ISL_CONST_DATA Path);				/* disk directory/file of external representation */
		
	ISL_STATUS (*InitializeFromMemory)(		/* constructor for existing archive */
		ISL_ARCHIVE_CONTEXT *Memory,		/* memory allocated for this instance */
		ISL_ARCHIVE_CONFIG *ConfigContext, 	/* my algorithm<->code extension configuration */
		ISL_CONST_DATA FileImage,			/* memory image of external representation */
		ISL_CONST_DATA SearchPath);			/* Search path for the items in the manifest */
	
	ISL_STATUS (*Externalize)(				/* create exteral rep. of archive */
		ISL_ARCHIVE_CONTEXT *Context);		/* context */
		
	ISL_STATUS (*Recycle)(					/* destructor for archive */
		ISL_ARCHIVE_CONTEXT *Context);	    /* archive context */

	ISL_MANIFEST_SECTION_PTR (*CreateNewSignableObject)(/* create a new object that can be signed */
		ISL_ARCHIVE_CONTEXT *Context,			/* context */
		ISL_CONST_DATA Name,					/* name of object in manifest (optional) */
		ISL_CONST_DATA GetServiceName,			/* service to get data (in config) */
		ISL_CONST_DATA GetParameters);			/* get data parameters */
	
	ISL_STATUS (*RemoveSignableObject)(			/* remove an object that can be signed */
		ISL_ARCHIVE_CONTEXT *Context,			/* arhive context */
		ISL_MANIFEST_SECTION_PTR Section);		/* pointer to manifest section */
	
	ISL_MANIFEST_SECTION_PTR (*FindSignableObject)(		/* return found object or NULL */
		ISL_ARCHIVE_CONTEXT *Context,					/* context */
		ISL_CONST_DATA Name);							/* archive-unique name */
	
	ISL_ITERATOR_PTR (*CreateSignableObjectEnumerator)(
		ISL_ARCHIVE_CONTEXT *Context);		/* pointer to configuration context */

	ISL_MANIFEST_SECTION_PTR (*GetNextSignableObject)(
		ISL_ITERATOR_PTR Iterator);

	ISL_STATUS (*RecycleSignableObjectEnumerator)(
		ISL_ITERATOR_PTR Iterator);

	ISL_SIGNATURE_CONTEXT* (*CreateNewSignature)(	/* add PGP/PKCS#7 signature to archive */
		ISL_ARCHIVE_CONTEXT *Context,				/* archive context */
		ISL_SIGNATURE_CONFIG *SigConfigContext, /* signature config context */
		ISL_CONST_DATA Name);						/* file name within archive for signature */

	ISL_STATUS (*RemoveSignature)(			/* finish component list and enclosing signature */
		ISL_ARCHIVE_CONTEXT *Context,		/* archive context */
		ISL_SIGNATURE_CONTEXT *Signature);	/* signature within the archive */
	
	ISL_SIGNATURE_CONTEXT *(*FindSignature)(/* return found object or NULL */
		ISL_ARCHIVE_CONTEXT *Context,		/* context */
		ISL_CONST_DATA Name);				/* archive-unique name */
	
	ISL_ITERATOR_PTR (*CreateSignatureEnumerator)(
		ISL_ARCHIVE_CONTEXT *Context);		/* pointer to configuration context */

	ISL_SIGNATURE_CONTEXT *(*GetNextSignature)(
		ISL_ITERATOR_PTR Iterator);

	ISL_STATUS (*RecycleSignatureEnumerator)(
		ISL_ITERATOR_PTR Iterator);

	ISL_STATUS (*AddCertificate)(			/* add certificate file to archive */
		ISL_ARCHIVE_CONTEXT *Context,		/* archive context */
		ISL_CONST_DATA Name,				/* archive-unique name */
		ISL_CERTIFICATE *Certificate);		/* certificate */
	
	ISL_STATUS (*RemoveCertificate)(		/* remove certificate file from archive */
		ISL_ARCHIVE_CONTEXT *Context,		/* archive context */
		ISL_CERTIFICATE *Certificate);		/* certificate */
	
	ISL_CERTIFICATE *(*FindCertificate)(	/* return found object or NULL */
		ISL_ARCHIVE_CONTEXT *Context,		/* context */
		ISL_CONST_DATA Name);				/* archive-unique name */
	
	ISL_ITERATOR_PTR (*CreateCertificateEnumerator)(
		ISL_ARCHIVE_CONTEXT *Context);		/* pointer to archive context */

	ISL_CERTIFICATE *(*GetNextCertificate)(
		ISL_ITERATOR_PTR Context);

	ISL_STATUS (*RecycleCertificateEnumerator)(
		ISL_ITERATOR_PTR Context);

	ISL_STATUS (*GetArchivePath)(	/* return the path of the archive file */ 
		ISL_ARCHIVE_CONTEXT *Context,
		ISL_CONST_DATA_PTR PathName);

	ISL_STATUS (*GetArchiveDirectory)( /* return the directory of the archive file */
		ISL_ARCHIVE_CONTEXT *Context,	
		ISL_CONST_DATA_PTR DirName);	

	ISL_STATUS (*AddAttribute)(				/* add metadata to manifest header */
		ISL_ARCHIVE_CONTEXT_PTR Context,	/* archive context */
		ISL_CONST_DATA Attribute,			/* attribute "name" */
		ISL_CONST_DATA Value);				/* attribute value */

	ISL_STATUS (*RemoveAttribute)(			/* remove metadata from manifest header */
		ISL_ARCHIVE_CONTEXT_PTR Context,	/* archive context */
		ISL_CONST_DATA Attribute);			/* attribute "name" */

	ISL_STATUS (*FindAttributeValue)(		/* find metadata in manifest header */
		ISL_ARCHIVE_CONTEXT_PTR Context,	/* archive context */
		ISL_CONST_DATA Attribute,			/* attribute "name" */
		ISL_CONST_DATA_PTR Value);			/* fill in Value and Length if found */

	ISL_ITERATOR_PTR (*CreateAttributeEnumerator)(
		ISL_ARCHIVE_CONTEXT_PTR Context);		/* pointer to signer context */

	ISL_STATUS (*GetNextAttribute)(
		ISL_ITERATOR_PTR Context,
		ISL_CONST_DATA_PTR Attribute,
		ISL_CONST_DATA_PTR Value);

	ISL_STATUS (*RecycleAttributeEnumerator)(
		ISL_ITERATOR_PTR Context);

	ISL_STATUS (*GetErrorObject)(			/* return context and methods for error reporting */
		ISL_ERROR_CONTEXT **Context,		/* context of error reporting object */
		const ISL_ERROR_METHODS **Methods);	/* methods for error reporting */

	ISL_CLASS_PTR *Class;					/* pointer to class structure */

} ISL_ARCHIVE_METHODS;
/*
**	Signature Format Configuration
**
**	The signature format configuration provides collections of various services needed
**	to construct or verify a message-level signature object (PGP, PKCS#7).
**	The services conspire to use the same representation for such things as keys,
**	algorithms and parameters, signer and issuer names, etc.
**	In short, one can construct or verify particular instances of this signature
**	by constructing a Signature Format Object using the Signature Format Configuration
**	in the construtor call.
**	
**	The difference between an archive and message-level signature is that the 
**	message level signatures contain only references to the manifest sections
**	(signable objects). Archives can contain the objects directly.
**
**	Message level signatures may also be used as standalone envelopes for signed data.
*/
typedef struct isl_signature_config_methods{
	/* class methods */
	ISL_SIZE (*SizeofObject)();				/* returns sizeof object */

	/* object methods */
	ISL_STATUS (*Initialize)(				/* ISL_SIGNATURE_CONTEXT constructor */
		ISL_SIGNATURE_CONFIG *Context,		/* pointer to configuration context */
		ISL_ARCHIVE_CONFIG *ArchiveConfig);	/* pointer to archive configuration */
				
	ISL_STATUS (*AutoConfigure)(
		ISL_SIGNATURE_CONFIG *Context,		/* ISL_SIGNATURE_CONTEXT constructor */
		AAL_FUNC_TABLE_PTR CssmFuncTable,	/* Array specifying the names of CSSM Api 
											   functions and fucntion pointers to be 
											   used by ISL */
		uint32 NumCssmFuncTable,			/* Number of elements in the array */
		CSP_LIST_PTR CspList);				/* Array specifying a list of CSP
											   to be loaded by ISL */

	ISL_STATUS (*AddAlgorithm)(
		ISL_SIGNATURE_CONFIG *Context,		/* configuration context */
		ISL_SERVICE_CLASS AlgClass,			/* service type performed by this algorithm */
		ISL_CONST_DATA AlgID,				/* archive-specific-encoded algorithm ID */
		ISL_CONST_DATA HumanName,			/* human-readable algorithm name */
		ISL_SERVICE_CLASS_METHODS *Methods);		/* vector of methods */
		// void *ClassContext);				/* class context */

	ISL_SERVICE_CLASS_METHODS *(*FindAlgorithm)(	/* return a pointer to the service methods vector */
		ISL_SIGNATURE_CONFIG *Context,		/* pointer to configuration context */
		ISL_CONST_DATA AlgID);				/* given an algorithm ID */

	ISL_STATUS (*AddAlgorithmService)(
		ISL_SIGNATURE_CONFIG *Context,		/* pointer to configuration context */
		char *(*GgetExtension)(		    	/* pointer to function returning code extension name  */
			ISL_CONST_DATA AlgID));			/* given an algorithm ID */	

	ISL_STATUS (*Recycle)(					/* destructor for IntegrityContext */
		ISL_SIGNATURE_CONFIG *Context);		/* configuration context */

	ISL_ITERATOR_PTR (*CreateAlgorithmEnumerator)(
		ISL_SIGNATURE_CONFIG *Context,		/* pointer to configuration context */
		ISL_SERVICE_CLASS ServiceClass);	/* algorithm type */

	ISL_SERVICE_METHODS *(*GetNextAlgorithm)(
		ISL_ITERATOR_PTR Context);

	ISL_STATUS (*RecycleAlgorithmEnumerator)(
		ISL_ITERATOR_PTR Context);

	ISL_CLASS_PTR *Class;					/* pointer to class structure */

} ISL_SIGNATURE_CONFIG_METHODS;

/*
**	ISL_SIG_SECTION_LIST class public methods
*/
typedef struct isl_sig_section_methods {	
	/* class methods */
	ISL_SIZE (*SizeofObject)();				/* returns sizeof object */
	
	/* object methods */
	ISL_STATUS (*Initialize)(				/* constructor for new object list */
		ISL_SIG_SECTION_LIST *Memory,			/* memory allocated for this instance */
		ISL_SIGNATURE_CONTEXT *Signature,	/* signature that signs this list */
		ISL_CONST_DATA Name);				/* where this object list is being constructed */
	
	ISL_STATUS (*Flatten)(					/* create exteral rep. of signature block */
		ISL_SIG_SECTION_LIST *Context,			/* context */
		ISL_CONST_DATA_PTR Result);			/* flattened external representation output */
	
	ISL_STATUS (*Externalize)(				/* create exteral rep. of signed object list */
		ISL_SIG_SECTION_LIST *Context);			/* context */

	ISL_STATUS (*Recycle)(					/* destructor for signed object list */
		ISL_SIG_SECTION_LIST *Context);			/* context */

	ISL_STATUS (*AddSignedObject)(			/* add a new object that can be signed */
		ISL_SIG_SECTION_LIST *Context,			/* context */
		ISL_MANIFEST_SECTION_PTR Section);	/* signable object (manifest section) */
	
	ISL_STATUS (*RemoveSignedObject)(		/* remove an object that can be signed */
		ISL_SIG_SECTION_LIST *Context,			/* context */
		ISL_MANIFEST_SECTION_PTR Section);	/* signable object (manifest section) */
	
	ISL_MANIFEST_SECTION_PTR (*FindSignedObject)(	/* return found object or NULL */
		ISL_SIG_SECTION_LIST *Context,					/* context */
		ISL_CONST_DATA Name);						/* name of section to be found */

	ISL_ITERATOR_PTR (*CreateSignedObjectEnumerator)(
		ISL_SIG_SECTION_LIST *Context);		/* pointer to signed object list */

	ISL_MANIFEST_SECTION_PTR (*GetNextSignedObject)(
		ISL_ITERATOR_PTR Context);

	ISL_STATUS (*RecycleSignedObjectEnumerator)(
		ISL_ITERATOR_PTR Context);

	ISL_STATUS (*AddAttribute)(			/* add metadata to manifest header */
		ISL_SIG_SECTION_LIST *Context,	/* context(SignerInfoList) */
		ISL_CONST_DATA Attribute,		/* attribute "name" */
		ISL_CONST_DATA Value);			/* attribute value */

	ISL_STATUS (*RemoveAttribute)(		/* remove metadata from manifest header */
		ISL_SIG_SECTION_LIST *Context,	/* context(SignerInfoList) */
		ISL_CONST_DATA Attribute);		/* attribute "name" */

	ISL_STATUS (*FindAttributeValue)(	/* find metadata in manifest header */
		ISL_SIG_SECTION_LIST *Context,	/* context(SignerInfoList) */
		ISL_CONST_DATA Attribute,		/* attribute "name" */
		ISL_CONST_DATA_PTR Value);		/* fill in Value and Length if found */

	ISL_ITERATOR_PTR (*CreateAttributeEnumerator)(
		ISL_SIG_SECTION_LIST *Context);	/* pointer to signer context */

	ISL_STATUS (*GetNextAttribute)(
		ISL_ITERATOR_PTR Context,
		ISL_CONST_DATA_PTR Attribute,
		ISL_CONST_DATA_PTR Value);

	ISL_STATUS (*RecycleAttributeEnumerator)(
		ISL_ITERATOR_PTR Context);


	ISL_CLASS_PTR *Class;					/* pointer to class structure */


} ISL_SIG_SECTION_METHODS;

/*
**	ISL_SIGNATURE_CONTEXT class public methods
*/
typedef struct isl_signature_methods{	
	/* class methods */
	ISL_SIZE (*SizeofObject)();					/* returns sizeof object */
	
	/* object methods */
	ISL_STATUS (*InitializeNew)(				/* constructor for new signature */
		ISL_SIGNATURE_CONTEXT *Memory,			/* memory allocated for this instance */
		ISL_SIGNATURE_CONFIG *ConfigContext,	/* my algorithm<->code extension configuration */
		ISL_ARCHIVE_CONTEXT *Archive,			/* archive which will contain signature */
		ISL_CONST_DATA Name);					/* where this signature is being constructed */
	
	ISL_STATUS (*InitializeOld)(				/* constructor for existing archive */
		ISL_SIGNATURE_CONTEXT *Memory,			/* memory allocated for this instance */
		ISL_SIGNATURE_CONFIG *ConfigContext,	/* my algorithm<->code extension configuration */
		ISL_ARCHIVE_CONTEXT *Archive,			/* archive which contains signature */
		ISL_CONST_DATA Name);					/* disk file of external representation */

    ISL_STATUS (*InitializeFromImage)(			/* constructor for existing archive */
		ISL_SIGNATURE_CONTEXT   *Memory,		/* memory allocated for this instance */
		ISL_SIGNATURE_CONFIG    *ConfigContext,	/* my algorithm<->code extension configuration */
		ISL_ARCHIVE_CONTEXT     *Archive,		/* archive which contains signature */
		ISL_CONST_DATA          Name,			/* disk file of external representation */
	    ISL_CONST_DATA          Image);         /* image of external representation */

	ISL_STATUS (*Externalize)(					/* create exteral rep. of signature */
		ISL_SIGNATURE_CONTEXT *Context);		/* context */
		
	ISL_STATUS (*Recycle)(						/* destructor for signature */
		ISL_SIGNATURE_CONTEXT *Context);		/* signature context */

	ISL_STATUS (*SetObjectList)(				/* define list of sections to be signed */
		ISL_SIGNATURE_CONTEXT *Context,
		ISL_SIG_SECTION_LIST *ObjectList);

	ISL_SIG_SECTION_LIST_PTR (*CreateNewSignedObjectList)(
		ISL_SIGNATURE_CONTEXT *SigContext,		/* archive which contains signature */
		ISL_CONST_DATA Name);			

	ISL_STATUS (*SetContent)(					/* defines enveloped content to be signed */
		ISL_SIGNATURE_CONTEXT *Context,			/* for standalone signature */
		ISL_CONST_DATA Content);				/* arbitrary content to be signed */

	ISL_SIG_SECTION_LIST *(*GetObjectList)(		/* return list of signed objects */
		ISL_SIGNATURE_CONTEXT *Context);		/* context */

	ISL_SIGNER_CONTEXT_PTR (*CreateNewSigner)(	/* add a signer to the signature */
		ISL_SIGNATURE_CONTEXT *Context,			/* archive context */
		ISL_CONST_DATA SignerID);				/* signature-specific name */

	ISL_STATUS (*RemoveSigner)(					/* finish component list and enclosing signature */
		ISL_SIGNATURE_CONTEXT *Context,			/* archive context */
		ISL_SIGNER_CONTEXT_PTR Signer);			/* signer context */
	
	ISL_ITERATOR_PTR (*CreateSignerEnumerator)(
		ISL_SIGNATURE_CONTEXT *Context);		/* pointer to signature context */

	ISL_SIGNER_CONTEXT_PTR (*GetNextSigner)(
		ISL_ITERATOR_PTR Iterator);

	ISL_STATUS (*RecycleSignerEnumerator)(
		ISL_ITERATOR_PTR Iterator);

	ISL_STATUS (*AddCertificate)(				/* add certificate to signature */
		ISL_SIGNATURE_CONTEXT *Context,			/* archive context */
		ISL_CERTIFICATE *Certificate);			/* certificate */
	
	ISL_STATUS (*RemoveCertificate)(			/* add certificate to archive */
		ISL_SIGNATURE_CONTEXT *Context,			/* archive context */
		ISL_CERTIFICATE *Certificate);			/* certificate */
	
	ISL_ITERATOR_PTR (*CreateCertificateEnumerator)(
		ISL_SIGNATURE_CONTEXT *Context);			/* pointer to signature context */

	ISL_CERTIFICATE *(*GetNextCertificate)(
		ISL_ITERATOR_PTR Context);

	ISL_STATUS (*RecycleCertificateEnumerator)(
		ISL_ITERATOR_PTR Context);

	ISL_CERTIFICATE *(*FindCertificate)(		/* return found object or NULL */
		ISL_SIGNATURE_CONTEXT *Context,			/* context */
		ISL_CONST_DATA SubjectName);			/* subject distinguished name */

	ISL_STATUS (*GetArchiveName)(				/* return signature file name */
		ISL_SIGNATURE_CONTEXT *Context,
		ISL_CONST_DATA_PTR Value);
	
	ISL_STATUS (*Flatten)(					/* create exteral rep. of signed object list */
		ISL_SIGNATURE_CONTEXT *Context,		/* context */
		ISL_CONST_DATA_PTR Result);			/* flattened external representation output */
	
	 ISL_CLASS_PTR *Class;					/* pointer to class structure */

} ISL_SIGNATURE_METHODS;
/*
**	Signer Class
*/
typedef struct isl_signer_methods {

	ISL_STATUS (*GetSignerID)(
		ISL_SIGNER_CONTEXT_PTR Context,			/* signer context */
		ISL_CONST_DATA *SignerID);				/* updated issuer and serial number */

#ifndef _IBM
	ISL_STATUS (*Sign)(							/* sign */
		ISL_SIGNER_CONTEXT_PTR Context,			/* signer context */
		void *ApplicationContext,				/* context used by low level signature algorithm */
		ISL_CONST_DATA Key,						/* public key for signature */
		const ISL_DIGEST_METHODS *DigestMethods,		/* configuration methods for digest algorithm */
		const ISL_SIGN_VERIFY_METHODS *SignMethods);	/* configuration methods for signature algorithm */
#else
	ISL_STATUS (*Sign)(							/* sign */
		ISL_SIGNER_CONTEXT_PTR Context,			/* signer context */
		void *ApplicationContext,				/* context used by low level signature algorithm */
		ISL_CONST_DATA SubjectPublicKey,		/* public key for signature */
		ISL_CONST_DATA SubjectPrivateKey,		/* private key for signature - modified by Scott Shorter */
		const ISL_DIGEST_METHODS *DigestMethods,		/* configuration methods for digest algorithm */
		const ISL_SIGN_VERIFY_METHODS *SignMethods);	/* configuration methods for signature algorithm */
#endif

	ISL_STATUS (*Verify)(
		ISL_SIGNER_CONTEXT_PTR		Context,	/* signer context */
		const ISL_CONST_DATA_PTR	Key);		/* public key for signature */

#ifndef _IBM
	ISL_STATUS (*SignUsingCert)(
		ISL_SIGNER_CONTEXT_PTR Context,				/* signer context */
		void *ApplicationContext,					/* context used by low level signature algorithm */
		const ISL_CERTIFICATE_METHODS *CertMethods,	/* configuration methods for certificate format */
		const ISL_CERTIFICATE *Cert, 				/* the certificate */
		const ISL_DIGEST_METHODS *DigestMethods,	/* configuration methods for digest algorithm */
		const ISL_SIGN_VERIFY_METHODS *SignMethods);/* configuration methods for signature algorithm */
#else
	ISL_STATUS (*SignUsingCert)(
		ISL_SIGNER_CONTEXT_PTR Context,				/* signer context */
		void *ApplicationContext,					/* context used by low level signature algorithm */
		const ISL_CERTIFICATE_METHODS *CertMethods,	/* configuration methods for certificate format */
		const ISL_CERTIFICATE *Cert, 				/* the certificate */
		ISL_CONST_DATA SubjectPrivateKey,		/* private key for signature - modified by Scott Shorter */
		const ISL_DIGEST_METHODS *DigestMethods,	/* configuration methods for digest algorithm */
		const ISL_SIGN_VERIFY_METHODS *SignMethods);/* configuration methods for signature algorithm */
#endif
	ISL_STATUS (*VerifyUsingCert)(
		ISL_SIGNER_CONTEXT_PTR Context,				/* signer context */
		const ISL_CERTIFICATE_METHODS *CertMethods,	/* configuration methods for certificate format */
		const ISL_CERTIFICATE *Cert);				/* the certificate */

	ISL_STATUS (*AddAttribute)(					/* add metadata to section */
		ISL_SIGNER_CONTEXT_PTR Context,			/* signer context */
		ISL_CONST_DATA Attribute,				/* attribute "name" */
		ISL_CONST_DATA Value);					/* attribute value */

	ISL_STATUS (*RemoveAttribute)(				/* remove metadata from section */
		ISL_SIGNER_CONTEXT_PTR Context,			/* signer context */
		ISL_CONST_DATA Attribute);				/* attribute "name" */

	ISL_STATUS (*FindAttributeValue)(			/* find metadata in section */
		ISL_SIGNER_CONTEXT_PTR Context,			/* signer context */
		ISL_CONST_DATA Attribute,				/* attribute "name" */
		ISL_CONST_DATA_PTR Value);				/* fill in Value and Length if found */

	ISL_ITERATOR_PTR (*CreateAttributeEnumerator)(
		ISL_SIGNER_CONTEXT_PTR Context);		/* pointer to signer context */

	ISL_STATUS (*GetNextAttribute)(
		ISL_ITERATOR_PTR Context,
		ISL_CONST_DATA_PTR Attribute,
		ISL_CONST_DATA_PTR Value);

	ISL_STATUS (*RecycleAttributeEnumerator)(
		ISL_ITERATOR_PTR Context);

	ISL_CERTIFICATE *(*FindSignerCertificate)(	/* return found object or NULL */
		ISL_SIGNER_CONTEXT_PTR Context);

	ISL_SIGN_VERIFY_METHODS *(*GetSignMethod) (
		ISL_SIGNER_CONTEXT_PTR Context);

	ISL_DIGEST_METHODS *(*GetDigestMethod) (
		ISL_SIGNER_CONTEXT_PTR Context);

	ISL_CLASS_PTR *Class;					/* pointer to class structure */

} ISL_SIGNER_METHODS;
/*
**	Signable Object (Manifest Section) Class
*/
typedef struct isl_manifest_section_methods {
	/* class methods */
	ISL_SIZE (*SizeofObject)();			/* returns sizeof object */
	
	/* object methods */
	
	ISL_STATUS (*Recycle)(						/* destructor */
		ISL_MANIFEST_SECTION_PTR Context);		/* context */

	ISL_ARCHIVE_CONTEXT *(*GetArchive)(			/* get the archive */
		ISL_MANIFEST_SECTION_PTR Context);		/* the object */

	ISL_STATUS (*AddDigest)(					/* add digest algorithm to section */
		ISL_MANIFEST_SECTION_PTR Context,		/* context */
		const ISL_DIGEST_METHODS *Digest);		/* digest methods from configuration */
	
	ISL_STATUS (*RemoveDigest)(					/* add digest algorithm to section */
		ISL_MANIFEST_SECTION_PTR Context,		/* archive context */
		const ISL_DIGEST_METHODS *Digest);		/* digest methods from configuration */
	
	ISL_ITERATOR_PTR (*CreateDigestEnumerator)(
		ISL_MANIFEST_SECTION_PTR Context);		/* pointer to signature context */

	const ISL_DIGEST_METHODS *(*GetNextDigest)(
		ISL_ITERATOR_PTR Iterator);

	ISL_STATUS (*RecycleDigestEnumerator)(
		ISL_ITERATOR_PTR Iterator);

	ISL_STATUS (*Verify)(
		ISL_MANIFEST_SECTION_PTR Context);		/* context */

	ISL_STATUS (*ExtractAndVerify)(
		ISL_MANIFEST_SECTION_PTR Context,		/* context */
		void *OutputContext,					/* object to update */
		ISL_STATUS (*Update)(					/* object's update method */
			void *UpdateContext,				/* (place OutputContext here) */
			ISL_CONST_DATA Buffer));			/* Length and Value of update */

	ISL_STATUS (*AddDigestValue)(				/* add new digest value for all digests */
		ISL_MANIFEST_SECTION_PTR Context,		/* context */
		ISL_CONST_DATA VersionParameters);		/* extra parameters for get method */
	
	ISL_STATUS (*AddAttribute)(					/* add metadata to section */
		ISL_MANIFEST_SECTION_PTR Context,		/* archive context */
		ISL_CONST_DATA Attribute,				/* attribute "name" */
		ISL_CONST_DATA Value);					/* attribute value */

	ISL_STATUS (*RemoveAttribute)(				/* remove metadata from section */
		ISL_MANIFEST_SECTION_PTR Context,		/* archive context */
		ISL_CONST_DATA Attribute);				/* attribute "name" */

	ISL_STATUS (*FindAttributeValue)(			/* find metadata in section */
		ISL_MANIFEST_SECTION_PTR Context,		/* archive context */
		ISL_CONST_DATA Attribute,				/* attribute "name" */
		ISL_CONST_DATA_PTR Value);				/* fill in Value and Length if found */

	ISL_ITERATOR_PTR (*CreateAttributeEnumerator)(
		ISL_MANIFEST_SECTION_PTR Section);		/* pointer to section */

	ISL_STATUS (*GetNextAttribute)(
		ISL_ITERATOR_PTR Iterator,
		ISL_CONST_DATA_PTR Name,
		ISL_CONST_DATA_PTR Value);

	ISL_STATUS (*RecycleAttributeEnumerator)(
		ISL_ITERATOR_PTR Iterator);

	ISL_STATUS (*GetServiceName)(
		ISL_MANIFEST_SECTION_PTR Context,
		ISL_CONST_DATA_PTR Value);				/* get service name */

	ISL_STATUS (*GetParameters)(
		ISL_MANIFEST_SECTION_PTR Context,
		ISL_CONST_DATA_PTR Value);				/* get parameters */

	ISL_STATUS (*GetSectionName)(
		ISL_MANIFEST_SECTION_PTR Context,
		ISL_CONST_DATA_PTR Value);				/* get sectionname */

	ISL_CLASS_PTR *Class;					/* pointer to class structure */


} ISL_MANIFEST_SECTION_METHODS;
/*
**	Certificate object methods
*/
typedef struct isl_certificate_methods {
	/* class methods */
	ISL_SIZE (*SizeofObject)();			/* returns sizeof object */
	
	/* object methods */
	ISL_STATUS (*Initialize)(
		ISL_CERTIFICATE *Memory,				/* memory for internal rep of cert */
		ISL_MEMORY_CONTEXT *MemoryContext,		/* memory allocation object */
		const ISL_MEMORY_METHODS *MemoryMethods,/* memory allocation methods */
		ISL_CONST_DATA Cert);					/* external rep of certificate */

	ISL_STATUS (*Flatten)(
		ISL_CERTIFICATE *Cert,
		ISL_CONST_DATA_PTR ExternalCert);

	ISL_STATUS (*Recycle)(
		ISL_CERTIFICATE *Cert);

	ISL_STATUS (*GetID)(
		ISL_CERTIFICATE *Cert,
		ISL_CONST_DATA_PTR SignerID);

	ISL_STATUS (*GetKey)(
		ISL_CERTIFICATE *Cert,
		ISL_CONST_DATA_PTR Key);

	ISL_STATUS (*GetSubject)(
		ISL_CERTIFICATE *Cert,
		ISL_CONST_DATA_PTR SubjectID);

	ISL_STATUS (*GetIssuer)(
		ISL_CERTIFICATE *Cert,
		ISL_CONST_DATA_PTR IssuerID);

	ISL_STATUS (*GetAlgID)(
		ISL_CERTIFICATE *Cert,
		ISL_CONST_DATA *AlgID);	

	ISL_STATUS (*Verify)(
		ISL_CERTIFICATE *Cert,
		ISL_SIGNATURE_CONFIG *Config,
		ISL_KEY_BLOCK *Key);

	ISL_ITERATOR_PTR (*CreateAttributeEnumerator)(
		ISL_CERTIFICATE *Cert);		/* pointer to certificate */

	ISL_STATUS (*GetNextAttribute)(
		ISL_ITERATOR_PTR Context,
		ISL_CONST_DATA_PTR Attribute,
		ISL_CONST_DATA_PTR Value);

	ISL_STATUS (*RecycleAttributeEnumerator)(
		ISL_ITERATOR_PTR Context);

	ISL_STATUS (*GetErrorObject)(			/* return context and methods for error reporting */
		ISL_ERROR_CONTEXT **Context,		/* context of error reporting object */
		const ISL_ERROR_METHODS **Methods);	/* methods for error reporting */

	ISL_CLASS_PTR *Class;					/* pointer to class structure */

} ISL_CERTIFICATE_METHODS;


/* 
**	get signed data service
*/
typedef struct isl_get_data_methods{
	ISL_SERVICE_CLASS_METHODS ServiceMethods;

	ISL_SIZE (*SizeofObject)();

	ISL_STATUS (*Initialize)(
		ISL_GET_DATA_SERVICE_CONTEXT *Memory,
		ISL_CONST_DATA Parameters);				/* archive-specific-encoded parameters */
		// void *ClassContext,							/* class context */
		// ISL_MANIFEST_SECTION_PTR *Section);			/* manifest section, if any */	

	ISL_ERROR_CODE (*Update)(
		ISL_GET_DATA_SERVICE_CONTEXT *Context,
		ISL_CONST_DATA_PTR Data);					/* location, length of data returned */ 

	ISL_ERROR_CODE (*Recycle)(
		ISL_GET_DATA_SERVICE_CONTEXT *Context);

	ISL_STATUS (*InitializeWithClass)(
		ISL_GET_DATA_SERVICE_CONTEXT *Memory,
		ISL_CONST_DATA Parameters,					/* archive-specific-encoded parameters */
		ISL_CLASS_PTR Class,						/* class structure */
		ISL_MANIFEST_SECTION_PTR Section);			/* manifest section, if any */
		
} ISL_GET_DATA_METHODS;
/*
**	certificate services are now subsumed by the certificate object
*/
typedef struct isl_parse_cert_methods{
	ISL_SERVICE_CLASS_METHODS		ServiceMethods;
	ISL_CERTIFICATE_METHODS			CertMethods;
} ISL_PARSE_CERT_METHODS;

/*
**	digest function service
*/
typedef struct isl_digest_methods {
	ISL_SERVICE_CLASS_METHODS ServiceMethods;

	ISL_SIZE (*SizeofObject)() ;

	ISL_STATUS (*Initialize)(
		ISL_DIGEST_SERVICE_CONTEXT *Memory,
		ISL_CONST_DATA Parameters,
		ISL_SIZE *SizeofDigest);
		// , void *ClassContext);						/* class context from AddAlgorithm */

	ISL_ERROR_CODE (*Update)(
		ISL_DIGEST_SERVICE_CONTEXT *Context,
		ISL_CONST_DATA Data);

	ISL_ERROR_CODE (*Recycle)(
		ISL_DIGEST_SERVICE_CONTEXT *Context,
		ISL_DATA_PTR Digest);

	ISL_STATUS (*InitializeWithClass)(
		ISL_DIGEST_SERVICE_CONTEXT *Memory,
		ISL_CONST_DATA Parameters,
		ISL_SIZE *SizeofDigest,
		ISL_CLASS_PTR Class);						/* class structure */
} ISL_DIGEST_METHODS;


/*
**	signature service: digest encrypt/decrypt, no context
*/
typedef struct isl_sign_verify_methods {
	ISL_SERVICE_CLASS_METHODS ServiceMethods;


	ISL_STATUS (*Sign)(
		void *AppContext,
		ISL_CONST_DATA Parameters,
		const ISL_KEY_BLOCK *Key,
		const ISL_DIGEST_METHODS *DigestMethods,
		ISL_CONST_DATA DigestParameters,
		const ISL_GET_DATA_METHODS *GetDataMethods,
		ISL_CONST_DATA GetDataParameters,
		ISL_DATA_PTR EncryptedDigest);

	ISL_ERROR_CODE (*Verify)(
		ISL_CONST_DATA Parameters,
		const ISL_KEY_BLOCK *Key,
		ISL_CONST_DATA Signature,
		const ISL_DIGEST_METHODS *DigestMethods,
		ISL_CONST_DATA DigestParameters,
		const ISL_GET_DATA_METHODS *GetDataMethods,
		ISL_CONST_DATA GetDataParameters);


	ISL_STATUS (*SplitCombination)(					   /* separate algorithms for PKCS#7 */
		const ISL_DIGEST_METHODS **DigestMethods,	   /* digest algorithm */
		const ISL_SIGN_VERIFY_METHODS **SignMethods);  /* signature algorithm */

	ISL_STATUS (*SignWithClass)(
		void *AppContext,
		ISL_CONST_DATA Parameters,
		const ISL_KEY_BLOCK *Key,
		const ISL_DIGEST_METHODS *DigestMethods,
		ISL_CONST_DATA DigestParameters,
		const ISL_GET_DATA_METHODS *GetDataMethods,
		ISL_CONST_DATA GetDataParameters,
		ISL_DATA_PTR EncryptedDigest,
		ISL_CLASS_PTR Class);

	ISL_ERROR_CODE (*VerifyWithClass)(
		ISL_CONST_DATA Parameters,
		const ISL_KEY_BLOCK *Key,
		ISL_CONST_DATA Signature,
		const ISL_DIGEST_METHODS *DigestMethods,
		ISL_CONST_DATA DigestParameters,
		const ISL_GET_DATA_METHODS *GetDataMethods,
		ISL_CONST_DATA GetDataParameters,
		ISL_CLASS_PTR Class);

	ISL_STATUS (*SplitCombinationWithClass)(			/* separate algorithms for PKCS#7 */
		const ISL_DIGEST_METHODS **DigestMethods,	   /* digest algorithm */
		const ISL_SIGN_VERIFY_METHODS **SignMethods,  /* signature algorithm */
		ISL_CLASS_PTR Class);

} ISL_SIGN_VERIFY_METHODS;

/*
**	high level verify service (exactly our own verify API)
*/
typedef struct isl_parse_verify_methods{
	ISL_SERVICE_CLASS_METHODS ServiceMethods;

	ISL_ERROR_CODE (*Verify)(
		ISL_ARCHIVE_CONTEXT *Context,		/* configuration context */
		ISL_DATA SignedChecklist,			/* signed component list */
		ISL_DATA SignerID,					/* distinguished ID of root signer (from cert) */
		ISL_KEY_BLOCK PublicKey,			/* root public key for signed component list */
		void *reportContext,				/* context for progress report callback */
		void (*reportUpdate)(				/* callback function for reporting progress */
			void *CallBackContext,			/* reportContext */
			ISL_ERROR_CODE Report,			/* report error/report code */
			ISL_DATA Component,				/* component or null if outer signature */
			ISL_DATA Signer),				/* distinguished ID of signer */
		void *TrustContext,					/* trust model context */
		ISL_ERROR_CODE (*TrustUpdate)(		/* update method for trust model */
			void *UpdateContext,			/* trustContext */
			ISL_DATA Certificate,			/* NULL if a signature fails to verify */
			int ForYourInfo,				/* true:  this cert is just FYI, false: affects trust value */ 
			ISL_DATA CertType));			/* ID for certificate format */

	ISL_ERROR_CODE (*Find)(
		ISL_ARCHIVE_CONTEXT *Context,		/* configuration context */
		ISL_GET_DATA_METHODS *GetObject,	/* get data methods */
		ISL_DATA GetObjectParameters,		/* get data parameters */
		ISL_DATA *SignatureFormat,			/* if found, return ID for signature format, else NULL */
		ISL_DATA *CertificateFormat,		/* if needed, return ID for certificate format, else NULL */
		ISL_DATA *GetCertAlgorithm,			/* if needed, return algorithm to get certificate, else NULL */
		ISL_DATA *GetCertParameters);		/* if needed, return parameters to get cert alg */

	ISL_ERROR_CODE (*VerifyWithClass)(
		ISL_ARCHIVE_CONTEXT *Context,		/* configuration context */
		ISL_DATA SignedChecklist,			/* signed component list */
		ISL_DATA SignerID,					/* distinguished ID of root signer (from cert) */
		ISL_KEY_BLOCK PublicKey,			/* root public key for signed component list */
		void *reportContext,				/* context for progress report callback */
		void (*reportUpdate)(				/* callback function for reporting progress */
			void *CallBackContext,			/* reportContext */
			ISL_ERROR_CODE Report,			/* report error/report code */
			ISL_DATA Component,				/* component or null if outer signature */
			ISL_DATA Signer),				/* distinguished ID of signer */
		void *TrustContext,					/* trust model context */
		ISL_ERROR_CODE (*TrustUpdate)(		/* update method for trust model */
			void *UpdateContext,			/* trustContext */
			ISL_DATA Certificate,			/* NULL if a signature fails to verify */
			int ForYourInfo,				/* true:  this cert is just FYI, false: affects trust value */ 
			ISL_DATA CertType),				/* ID for certificate format */
		ISL_CLASS_PTR Class);		

	ISL_ERROR_CODE (*FindWithClass)(
		ISL_ARCHIVE_CONTEXT *Context,		/* configuration context */
		ISL_GET_DATA_METHODS *GetObject,	/* get data methods */
		ISL_DATA GetObjectParameters,		/* get data parameters */
		ISL_DATA *SignatureFormat,			/* if found, return ID for signature format, else NULL */
		ISL_DATA *CertificateFormat,		/* if needed, return ID for certificate format, else NULL */
		ISL_DATA *GetCertAlgorithm,			/* if needed, return algorithm to get certificate, else NULL */
		ISL_DATA *GetCertParameters,		/* if needed, return parameters to get cert alg */
		ISL_CLASS_PTR Class);

} ISL_PARSE_VERIFY_METHODS;

#endif /* ifndef ISL_API_H */
