/*-----------------------------------------------------------------------
 * File: ISSUER.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#ifdef WIN32
#include <windows.h>
#endif
#include <stdio.h>
#include <sys/stat.h>

#include <cssmtype.h>
#include "isl_internal.h"

/* given the certificate blob "certBlob", generate a function named
 * "funcname" in a file named "codefile" that will populate the four
 * arrays handed it with the pieces of a DSA public key.
 */
static int
makeIssuerFromBlob(FILE *pub, unsigned char *certBlob, int certLength,
									char *codefile, char *funcname, int index)
{
	ISL_CERTIFICATE Cert;
	ISL_CONST_DATA  myData;
	ISL_CONST_DATA Issuer;
	uint32 i;

	myData.Data = certBlob;
	myData.Length = certLength;

	if (isl_X509Cert_Initialize(&Cert, 
								  NULL,
								  NULL,
								  myData) != ISL_OK){
		printf("FATAL: Cert Initialize failed\n");
		return ISL_FAIL;
	}
	if (isl_X509Cert_GetIssuer(&Cert, &Issuer) != ISL_OK){
		printf("FATAL: Get Issuer failed\n");
		return ISL_FAIL;
	}

	fprintf (pub, "\t{\n");
	fprintf(pub,"\tconst unsigned char rootIssuerName%d[] = {", index);
	for (i = 0; i < Issuer.Length - 1; i++){
		if (i % 8 == 0)
			fprintf(pub,"\n\t");
		fprintf(pub,"0x%x, ", (unsigned char) Issuer.Data[i]);
	}
	fprintf(pub,"0x%x};\n", Issuer.Data[Issuer.Length-1]);
	fprintf(pub,"\tint len = %d;\n",Issuer.Length); 
	fprintf(pub,"\tif (Name) {\n");
	fprintf(pub,"\t\tif (len > Name[index].length) {\n");
	fprintf(pub,"\t\t\tName[index].length = 0;\n");
	fprintf(pub,"\t\t\treturn 0;\n");
	fprintf(pub,"\t\t}\n");

	fprintf(pub,"\t\tName[index].length = len;\n");
	fprintf(pub,"\t\tmemcpy(Name[index].value, rootIssuerName%d, Name[index].length);\n", index);
	fprintf(pub,"\t}\n");
	fprintf(pub,"\tindex++;\n");
	fprintf(pub,"\t}\n");
	return ISL_OK;
}

static int
makeSignerFromBlob(FILE *pub, unsigned char *certBlob, int certLength,
									char *codefile, char *funcname, int index)
{
	ISL_CERTIFICATE Cert;
	ISL_CONST_DATA  myData;
	ISL_CONST_DATA Signer;
	uint32 i;

	myData.Data = certBlob;
	myData.Length = certLength;

	if (isl_X509Cert_Initialize(&Cert, 
								  NULL,
								  NULL,
								  myData) != ISL_OK){
		printf("FATAL: Cert Initialize failed\n");
		return ISL_FAIL;
	}
	if (isl_X509Cert_GetID(&Cert, &Signer) != ISL_OK){
		printf("FATAL: Get Signer failed\n");
		return ISL_FAIL;
	}

	fprintf (pub, "\t{\n");
	fprintf(pub,"\tconst unsigned char rootSignerName%d[] = {", index);
	for (i = 0; i < Signer.Length - 1; i++){
		if (i % 8 == 0)
			fprintf(pub,"\n\t");
		fprintf(pub,"0x%x, ", (unsigned char) Signer.Data[i]);
	}
	fprintf(pub,"0x%x};\n", Signer.Data[Signer.Length-1]);
	fprintf(pub,"\tint len = %d;\n",Signer.Length); 
	fprintf(pub,"\tif (Name) {\n");
	fprintf(pub,"\t\tif (len > Name[index].length) {\n");
	fprintf(pub,"\t\t\tName[index].length = 0;\n");
	fprintf(pub,"\t\t\treturn 0;\n");
	fprintf(pub,"\t\t}\n");

	fprintf(pub,"\t\tName[index].length = len;\n");
	fprintf(pub,"\t\tmemcpy(Name[index].value, rootSignerName%d, Name[index].length);\n", index);
	fprintf(pub,"\t}\n");
	fprintf(pub,"\tindex++;\n");
	fprintf(pub,"\t}\n");
	return ISL_OK;
}


static int
makeKeyFromBlob(FILE *pub, unsigned char *certBlob, int certLength,
								char *codefile, char *funcname, int index)
{
	ISL_CERTIFICATE Cert;
	ISL_CONST_DATA  myData;
	ISL_CONST_DATA Key;
	uint32 i;

	myData.Data = certBlob;
	myData.Length = certLength;

	if (isl_X509Cert_Initialize(&Cert, 
								  NULL, 
								  NULL, 
								  myData) != ISL_OK){
		printf("FATAL: Cert Initialize failed\n");
		return ISL_FAIL;
	}
	if (isl_X509Cert_GetKey(&Cert, &Key) != ISL_OK){
		printf("FATAL: Get ID failed\n");
		return ISL_FAIL;
	}

	fprintf (pub, "\t{\n");
	fprintf(pub,"\tconst unsigned char Key%d[] = {", index);
	for (i = 0; i < Key.Length - 1; i++){
		if (i % 8 == 0)
			fprintf(pub,"\n\t");
		fprintf(pub,"0x%x, ", (unsigned char) Key.Data[i]);
	}
	fprintf(pub,"0x%x};\n", Key.Data[Key.Length-1]);
	fprintf(pub,"\tint len = %d;\n",Key.Length); 
	fprintf(pub,"\tif (Name) {\n");
	fprintf(pub,"\t\tif (len > Name[index].length) {\n");
	fprintf(pub,"\t\t\tName[index].length = 0;\n");
	fprintf(pub,"\t\t\treturn 0;\n");
	fprintf(pub,"\t\t}\n");

	fprintf(pub,"\t\tName[index].length = len;\n");
	fprintf(pub,"\t\tmemcpy(Name[index].value, Key%d, Name[index].length);\n", index);
	fprintf(pub,"\t}\n");
	fprintf(pub,"\tindex++;\n");
	fprintf(pub,"\t}\n");
	return ISL_OK;
}


static FILE *
makeHeader(char *codefile, char *funcname)
{
	FILE *Handle;

	/*
	 * Open the file for writing destroying any existing contents.
	 */
	if ((Handle = fopen(codefile, "w")) == NULL) {
		printf("could not open codefile %s\n", codefile);
		return NULL;
	}

	/* headers */
	fprintf(Handle,"#include \"callouts.h\"\n\n");

	fprintf(Handle,"int %s(ISL_CALLOUT_DATA *Name)\n{\n", funcname);
	fprintf(Handle,"\tint index = 0;\n");

	return (Handle);
}

static int
makeTrailer(FILE *Handle)
{
	fprintf(Handle,"\treturn index;\n");

	fprintf(Handle,"}\n");
	fclose(Handle);
	return ISL_OK;
}


int
main(int argc, char *argv[])

{
	FILE *fi;
	FILE *ci;
	FILE *issuerFileHandle;

	char *certinfofile;
	char *codefile;
	char *funcname;
	char *options;
	unsigned char *certBlob;
	struct _stat st;
	char certfile[64];
	int index = 0;

	if (argc < 5){
		printf("usage is %s <-iks> <certfile> <codefile> <funcname>\n", argv[0]);
		return 1;
	}
	options  = argv[1];
	certinfofile = argv[2];
	codefile = argv[3];
	funcname = argv[4];

	ci = fopen(certinfofile, "rb");
	if (!ci){
		printf("FATAL: could not open certfile %s for reading\n", certfile);
		return 1;
	}

	while (fgets (certfile, 64, ci) != NULL) {
		int len = 0;

		len = strlen(certfile);
		len--;
		while (certfile[len] == 0xd || certfile[len] == 0xa) {
			certfile[len--] = '\0';
		}
		printf ("certfile = %s len = %d\n", certfile, len);
		fi = fopen(certfile, "rb");
		if (!fi){
			perror ("fopen");
			printf("FATAL: could not open certfile %s for reading\n", certfile);
			return 1;
		}

		/* stat the file and find out how much memory we need */
		if ((_stat(certfile,&st)) != 0){
			printf("FATAL: _stat failed on %s\n", certfile);
			return 1;
		}

		/* grab the memory */
		certBlob = (unsigned char *) malloc(st.st_size);
		if (!certBlob){
			printf("FATAL: failed to malloc %d bytes of storage for certBlob\n",
			 st.st_size);
		return 1;
		}

		/* read in the file */
		if ((fread(certBlob,st.st_size,1,fi)) != 1){
			printf("FATAL: fread failed for %d bytes on %s\n", st.st_size, certfile);
			return 1;
		}

		fclose(fi);

		if (options[0] != '-')
		  return(ISL_FAIL);

		if (0 == index) {
			if ((issuerFileHandle = makeHeader(codefile,funcname)) == NULL)
				return (ISL_FAIL);
		}
		index++;

		switch(options[1]) {
		case 'i':
		case 'I':
			if (makeIssuerFromBlob(issuerFileHandle, certBlob,st.st_size, 
												codefile, funcname, index) != ISL_OK)
				return (ISL_FAIL);
			break;
		case 's':
		case 'S':
			if (makeSignerFromBlob(issuerFileHandle, certBlob,st.st_size, 
												codefile, funcname, index) != ISL_OK)
				return (ISL_FAIL);
			break;
		case 'k':
		case 'K':
			if (makeKeyFromBlob(issuerFileHandle, certBlob,st.st_size,
												codefile,funcname, index) != ISL_OK)
				return (ISL_FAIL);
			break;
		default:
			printf("usage is %s <-ik> <certinfofile> <codefile> <funcname>\n", argv[0]);
			return ISL_FAIL;
		}

		free(certBlob);
	}

	return (makeTrailer(issuerFileHandle));
}


/*
 * This function compares the two strings. This function returns TRUE even if one
 * of the strings does not include the NULL character.
 */
BOOL
is_equal (ISL_CONST_DATA Name, ISL_CONST_DATA String)

{
	if (String.Length == Name.Length) {
		if (!memcmp(Name.Data, String.Data, String.Length))
			return CSSM_TRUE;
	} else if (String.Length == Name.Length-1) {
		if ((Name.Data[Name.Length] == '\0') && (!memcmp(Name.Data, String.Data, String.Length)))
			return CSSM_TRUE;
	} else if (Name.Length == String.Length-1) {
		if ((String.Data[String.Length] == '\0') && (!memcmp(Name.Data, String.Data, Name.Length)))
		return CSSM_TRUE;
	}

	return CSSM_FALSE;
}
