/*-----------------------------------------------------------------------
 * File: MRM.H
 *
 * Copyright (c) 1998-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

// Mrm.h
//
// This file contains the class definition for the CMrm class.  The 
//  CMrm class implements a Manifest Request Message object, which
//  is made up of Signed Manifests (archived CManifest objects) and 
//  data files, in its external archived configuration.  The MRM is 
//  intended for use in requesting a signature or signatures over 
//  a set of data objects by some high-value key based on signatures 
//  collected from zero or more non-high-value key holders acting as 
//  reviewers.
//
//////////////////////////////////////////////////////////////////////
#ifndef MRM_H
#define MRM_H

// mrm type definitions
#include "wf_type.h"

// embedded object definitions
#include "ObjectList.h"
#include "MRMVersion.h"
#include "Manifest.h"

//////////////////////////////////////////////////////////////////////
// Class definition:
//  Members:
//	    mContentList(public)
//          list of data objects referenced by MRM
//	    mAppID(private)
//          AppID associated with this MRM
//	    mContextID(private)
//          ContextID associated with this MRM
//	    mVersion(private)
//          version of MRM
//	    mRequest(private)
//          CManifest object which represents the request
//	    mRequestCredential(private)
//          CManifest object which represents the request credential
//	    mSignedManifest(private)
//          CManifest object which represents the signed manifest
//	    mExternalFilePath(private)
//          external filepath where the MRM is to be stored persistently
//  Methods:
//      Init(public)
//          initializes a constructed MRM
//      SignRequest(public)
//          signs the request credential manifest for the MRM
//      CounterSignRequest(public)
//          counter-signs a signer in the request credential manifest 
//          for the MRM
//      SignManifest(public)
//          signs the signed manifest portion of the MRM
//      CounterSignManifest(public)
//          counter-signs a signer in the request credential manifest 
//          for the MRM
//      Verify(public)
//          verifies both integrity and content for all signers and all
//          sections in each contained manifest.  This verification 
//          will check that the hash values contained in each section 
//          match the actual hash values for the referant object.  It 
//          will also walk the manifests for each signer contained in it.
//      VerifyIntegrity(public)
//          verifies all signers in each manifest object.  This 
//          verification will check that the internal structure of 
//          each manifest is intact by walking from each signer to all
//          of the nodes signed by that signer
//      VerifyContent(public)
//          verifies all sections in each manifest object.  This 
//          verification will check that the hash values contained in 
//          each section match the actual hash values for the referant 
//          object
//      GetSignerGraph(public)
//          retrieves the graph of all signatures applied to the signed 
//          manifest portion of the MRM
//      GetReviewerGraph(public)
//          retrieves the graph of all signatures applied to the request 
//          credential manifest in the MRM
//      GetVersion(public)
//          retrieves the version object associated with the MRM
//      GetAppID(public)
//          retrieves the AppID associated with the MRM
//      GetContextID(public)
//          retrieves the context ID associated with the MRM
//      Commit(public)
//          commits the operational representation of the mrm object 
//          to its persistent archive representation
//      ExtractSignedManifest(public)
//          extracts the signed manifest from the MRM and archives it 
//      GetPackedSize(public)
//          gets the size of the MRM in its packed memory form
//      Pack(public)
//          packs the operational image of the MRM into the provided 
//          buffer
//      Recycle(private)
//          cleans up all state for current MRM and zero-izes it
/////////////////////////////////////////////////////////////////////
class DllExport CMrm  
{
public:
	// public members
	CObjectList *mContentList;
	
	// constructors
	CMrm();
	~CMrm();

	// initializers
    WF_RETURN Init(char *WFFilePath);
    WF_RETURN Init(char *WFFilePath,
                   WF_GUID *CSPGuid,
                   WF_CSSM_METHODS *CSSMMethods);
    WF_RETURN Init(CObjectList *List, 
                   char *MRMFilePath,
                   char *MRMFileName,
                   const WF_GUID *AppID, 
                   const WF_GUID *ContextID,
                   WF_CONST_DATA *Certificate, 
                   WF_CONST_DATA *Password, 
                   WF_CONST_DATA *SignerName, 
                   WF_CERTGROUP_PTR CertGroup,
                   CAttributeList *SignerAttributes);
    WF_RETURN Init(CObjectList *List, 
                   char *MRMFilePath,
                   char *MRMFileName,
                   const WF_GUID *AppID, 
                   const WF_GUID *ContextID,
                   WF_CONST_DATA *Certificate, 
                   WF_CONST_DATA *Password, 
                   WF_CONST_DATA *SignerName, 
                   WF_CERTGROUP_PTR CertGroup,
                   CAttributeList *SignerAttributes,
                   WF_GUID *CSPGuid,
                   WF_CSSM_METHODS *CSSMMethods);
	WF_RETURN Init(WF_CONST_DATA_PTR PackedMrm,
                   char *MRMFilePath,
                   char *MRMFileName);
	WF_RETURN Init(WF_CONST_DATA_PTR PackedMrm,
                   char *MRMFilePath,
                   char *MRMFileName,
                   WF_GUID *CSPGuid,
                   WF_CSSM_METHODS *CSMMMethods);

    // public methods
	WF_RETURN SignRequest(WF_CONST_DATA *Certificate, 
                          WF_CONST_DATA *Password,
                          WF_CONST_DATA *SignerName,
                          WF_CERTGROUP_PTR CertGroup,
                          CAttributeList *SignerAttributes,
                          MRM_SIGN_FLAG Flag );
	WF_RETURN CounterSignRequest(WF_CONST_DATA *Certificate, 
                                 WF_CONST_DATA *Password,
                                 WF_CONST_DATA *CounterSignerName,
                                 WF_CONST_DATA *SignerName,
                                 WF_CERTGROUP_PTR CertGroup,
                                 CAttributeList *SignerAttributes,
                                 MRM_SIGN_FLAG Flag );
	WF_RETURN SignManifest(WF_CONST_DATA *Certificate, 
                           WF_CONST_DATA *Password,
                           WF_CONST_DATA *SignerName,
                           WF_CERTGROUP_PTR CertGroup,
                           CAttributeList *SignerAttributes,
                           MRM_SIGN_FLAG Flag );
	WF_RETURN CounterSignManifest(WF_CONST_DATA *Certificate, 
                                  WF_CONST_DATA *Password,
                                  WF_CONST_DATA *CounterSignerName,
                                  WF_CONST_DATA *SignerName,
                                  WF_CERTGROUP_PTR CertGroup,
                                  CAttributeList *SignerAttributes,
                                  MRM_SIGN_FLAG Flag );
	WF_RETURN Verify();
	WF_RETURN VerifyIntegrity();
	WF_RETURN VerifyContent();
	WF_RETURN GetSignerGraph(CSignatureGraph* SigGraph);
	WF_RETURN GetReviewerGraph(CSignatureGraph* SigGraph);
	CMRMVersion GetVersion();
	WF_GUID* GetAppID();
	WF_GUID* GetContextID();
	WF_RETURN Commit();
    WF_RETURN Commit(WF_DATA *PrincipalID);
    WF_RETURN ExtractSignedManifest( char *OutputFilePath );
	WF_SIZE GetPackedSize();
	WF_RETURN Pack(WF_CONST_DATA_PTR PackedMrm, 
                   WF_DATA *PrincipalID);
	CAttributeList* GetAttributes();

private:
    // private members
	WF_GUID *mAppID;
	WF_GUID *mContextID;
	CMRMVersion *mVersion;
	CManifest *mRequest;
	CManifest *mRequestCredential;
	CManifest *mSignedManifest;
    char *mExternalFilePath;

    // private methods
    void Recycle();

};

#endif // !def MRM_H
