/*-----------------------------------------------------------------------
 * File: MRM.CPP
 *
 * Copyright (c) 1998-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

// Mrm.cpp
//
// This file contains the class implementation for the CMrm class.  
//  The CMrm class implements a Manifest Request Message object, 
//  which is made up of Signed Manifests (archived CManifest objects) 
//  and data files, in its external archived configuration.  The MRM 
//  is intended for use in requesting a signature or signatures over 
//  a set of data objects by some high-value key based on signatures 
//  collected from zero or more non-high-value key holders acting as 
//  reviewers.
//
// Copyright(c) 1998, Intel Corporation.  All rights reserved.
//
//////////////////////////////////////////////////////////////////////
#ifdef CSSM_MEMTRACK
#undef new
#include <crtdbg.h>
static char SZ_FILE_NAME[] = __FILE__;
#define new new(_NORMAL_BLOCK, SZ_FILE_NAME, __LINE__)  /*need space*/
#endif /* ifdef VMS comment this out - CSSM_MEMTRACK */

// mrm objects
#include "AttributeList.h"
#include "Mrm.h"

// pkware/guid functions
#include "wf_Util.h"

// system headers
#include <sys/stat.h>
#include <scMemTrack.h>
#include "cssmport.h"

//////////////////////////////////////////////////////////////////////
// CMrm::CMrm()
//  constructs a zero-ized CMrm object
//
// Parameters:
//  none
// Returns:
//  none
//////////////////////////////////////////////////////////////////////
CMrm::CMrm()
{
	mContentList = NULL;
	mAppID = NULL; 
	mContextID = NULL; 
	mVersion = NULL;
	mRequest = NULL;
	mRequestCredential = NULL;
	mSignedManifest = NULL;
    mExternalFilePath = NULL;
}

//////////////////////////////////////////////////////////////////////
// CMrm::Init(char* MRMFilePath)
//  initializes a CMrm object based on an archived MRM at MRMFilePath
//
// Parameters:
//  MRMFilePath(input)     
//      filepath to the MRM (.mrm) file to be used to initialize the 
//      object
// Returns:
//  WF_OK              initialization succeeded
//  WF_FAIL            initialization failed
// Side effects:
//  CSSM & ISL will be loaded implicitly
//  internally calls Init(char* ManifestPath, WF_HANDLE hCSP)
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::Init(char* MRMFilePath)
{
    return(Init(MRMFilePath, NULL, NULL));
}

//////////////////////////////////////////////////////////////////////
// CMrm::Init(char* MRMFilePath,
//            WF_GUID *CSPGuid)
//  initializes a CMrm object based on an archived MRM at MRMFilePath
//
// Parameters:
//  MRMFilePath(input)     
//      filepath to the MRM (.mrm) file to be used to initialize the 
//      object
//  CSPGuid(input)
//      guid of the CDSA CSP to be used for all cryptographic 
//      operations WRT manifests
// Returns:
//  WF_OK              initialization succeeded
//  WF_FAIL            initialization failed
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::Init(char* MRMFilePath,
                     WF_GUID *CSPGuid,
                     WF_CSSM_METHODS *CSSMMethods)
{
    FILE* fh;
    PK_ARCHIVE_PTR pArchive = NULL;
    PK_ITERATOR_PTR pIterator;
    WF_CONST_DATA Filename, Filedata, ArchiveName;
    int CredFound = 0, SMFound = 0;

    ArchiveName.Length = strlen(MRMFilePath);
    ArchiveName.Data = (unsigned char*)MRMFilePath;

    // extract mrm
    pArchive = mrm_pk_LoadArchive(ArchiveName);
    if( !pArchive )
        return WF_FAIL;

	// get fields from comment
	mVersion = new CMRMVersion();
	memcpy(&(mVersion->mMajor), 
           (pArchive->DirEnd->ZipFileComment.Data)+2, 
           1);
	memcpy(&(mVersion->mMinor), 
           (pArchive->DirEnd->ZipFileComment.Data)+3, 
           1);
    
    pIterator = mrm_pk_CreateFileEnumerator(pArchive);
    if( !pIterator )
    {
        // pk file error
        Recycle();
        return  WF_FAIL;
    }

	while (mrm_pk_GetNextFile(pIterator, &Filename, &Filedata) == PK_OK)
	{
        char Name[WF_MAX_PATH];
        memset(Name, 0, WF_MAX_PATH);
        memcpy(Name, Filename.Data, Filename.Length);
        // keep track of whether the cred or SM was found
        if( !strcmp(Name, MRM_REQUEST_CREDENTIAL_FILENAME)) CredFound = 1;
        if( !strcmp(Name, MRM_SIGNED_MANIFEST_FILENAME)) SMFound = 1;
        fh = fopen(Name, "wb");
        if(!fh)
        {
            // file error
            Recycle();
            mrm_pk_RecycleFileEnumerator(pIterator);
            mrm_pk_RecycleArchive(pArchive);
            return WF_FAIL;
        }
        if(fwrite(Filedata.Data, 1, Filedata.Length, fh) != Filedata.Length)
        {
            // file error
            fclose(fh);
        	Recycle();
            mrm_pk_RecycleFileEnumerator(pIterator);
            mrm_pk_RecycleArchive(pArchive);
            return WF_FAIL;
        }
        fclose(fh);
	}
	mrm_pk_RecycleFileEnumerator(pIterator);
    mrm_pk_RecycleArchive(pArchive);

    // initialize manifests in memory
    mRequest = new CManifest();
    if(mRequest->Init(MRM_REQUEST_FILEPATH) != WF_OK)
    {
        // unable to initialize request manifest
        // invalid operational representation of MRM
        Recycle();
        return WF_FAIL;
    }

    mAppID = (WF_GUID*)calloc(sizeof(WF_GUID), 1);
    mContextID = (WF_GUID*)calloc(sizeof(WF_GUID), 1);
    
    CObjectList* Sections = mRequest->GetSectionList();
    uint32 NumOfSections = Sections->GetNumberOfNodes();
    CObjectListNode* currRefNode = Sections->mHead;
    int i;
    for(i=0; i<NumOfSections; i++)
    {   
        if(!strcmp(currRefNode->mRef->mName, "MetaData"))
        {
            uint32 NumOfAttr = currRefNode->mRef->mAttributes->GetNumberOfNodes();
            CAttributeListNode* currAttr = currRefNode->mRef->mAttributes->mHead;
            for(unsigned int j=0; j<NumOfAttr; j++)
            {
                char Tmp[32];
                memset(Tmp, 0, 32);
                if(currAttr->mValue.Length <32)
                    memcpy(Tmp, currAttr->mValue.Data, currAttr->mValue.Length); 
                else
                    memcpy(Tmp, currAttr->mValue.Data, 31);
                
                if(!memcmp(currAttr->mAttribute.Data,
                           MRM_APP_ID_ATTR,
                           strlen(MRM_APP_ID_ATTR)))
                    wf_StrToGuid(Tmp, mAppID);

                if(!memcmp(currAttr->mAttribute.Data,
                           MRM_CTXT_ID_ATTR,
                           strlen(MRM_CTXT_ID_ATTR)))
                    wf_StrToGuid(Tmp, mContextID);

                currAttr = currAttr->mNext;
            }
        }
        currRefNode = currRefNode->mNext;
    }
    delete Sections;

    if( CredFound )
    {
        mRequestCredential = new CManifest();
        if(CSPGuid && CSSMMethods)
        {
            if(mRequestCredential->Init(MRM_REQUEST_CREDENTIAL_FILEPATH, 
                                        CSPGuid,
                                        CSSMMethods) != WF_OK)
            {
                // unable to initialize request credential manifest
                // operational MRM image incomplete
                Recycle();
                return WF_FAIL;
            }
        }
        else
        {
            if(mRequestCredential->Init(MRM_REQUEST_CREDENTIAL_FILEPATH) != WF_OK)
            {
                // unable to initialize request credential manifest
                // operational MRM image incomplete
                Recycle();
                return WF_FAIL;
            }
        }
    }
    else 
        mRequestCredential = NULL;
    if( SMFound )
    {
        mSignedManifest = new CManifest();
        if(CSPGuid && CSSMMethods)
        {
            if(mSignedManifest->Init(MRM_SIGNED_MANIFEST_FILEPATH, 
                                     CSPGuid,
                                     CSSMMethods) != WF_OK)
            {
                // unable to initialize signing manifest
                // operational MRM image incomplete
                Recycle();
                return WF_FAIL;
            }
        }
        else
        {
            if(mSignedManifest->Init(MRM_SIGNED_MANIFEST_FILEPATH) != WF_OK)
            {
                // unable to initialize signing manifest
                // operational MRM image incomplete
                Recycle();
                return WF_FAIL;
            }
        }
    }
    else 
        mSignedManifest = NULL;

    // initialize object list from request object list
    CObjectList *list = mRequest->GetSectionList();
    if(!list)
    {
        // unable to get request's object list
        // operational image of MRM invalid
        Recycle();
        return WF_FAIL;
    }
    mContentList = new CObjectList();
    // remove DONOTSIGN sections
    uint32 NumObjNodes = list->GetNumberOfNodes();
    CObjectListNode* currNode = list->mHead;
    for(i=0; i<NumObjNodes; i++)
    {
        int MetaFlag = 0;
        CAttributeList* AttrList = currNode->mRef->mAttributes;
        CAttributeListNode* currAttr = currNode->mRef->mAttributes->mHead;
        uint32 NumAttrNodes = currNode->mRef->mAttributes->GetNumberOfNodes();
        for(uint32 j=0; j<NumAttrNodes; j++)
        {
            char name[WF_MAX_STRING], value[WF_MAX_STRING];
            memset( name, 0, WF_MAX_STRING );
            memset( value, 0, WF_MAX_STRING );
        
            if(currAttr->mAttribute.Length < WF_MAX_STRING)
                memcpy(name, currAttr->mAttribute.Data, 
                       currAttr->mAttribute.Length );
            else // truncate because strings are too long
                memcpy(name, currAttr->mAttribute.Data, WF_MAX_STRING-1);
        
            if(currAttr->mValue.Length < WF_MAX_STRING)
                memcpy(value, currAttr->mValue.Data, 
                       currAttr->mValue.Length );
            else // truncate because strings are too long
                memcpy( value, currAttr->mValue.Data, WF_MAX_STRING-1);

            if(!strcmp(name, MRM_META_ATTR) && !strcmp(value, MRM_META_VALUE))
                MetaFlag = 1;

            currAttr = currAttr->mNext;
        }

        if(!MetaFlag)
            mContentList->AddNode(currNode); 
        currNode = currNode->mNext;
    }
    delete list;
    	
    // initialize remaining members
    mExternalFilePath = (char*)calloc(strlen(MRMFilePath)+1, 1);
    if( !mExternalFilePath )
    {
        // memory error
        Recycle();
        return WF_FAIL;
    }
    memcpy( mExternalFilePath, MRMFilePath, strlen(MRMFilePath));
    
    return WF_OK;
}

//////////////////////////////////////////////////////////////////////
// CMrm::Init(WF_CONST_DATA_PTR PackedMrm,
//            char* MRMFilePath,
//            char* MRMFileName)
//  initializes a CMrm object based on a packed MRM in a memory buffer
//  The CMrm object is initialized with the contents of the buffer, 
//  with the FilePath and FileName used for its external archive 
//  representation, should the object ever call Commit().
//
// Parameters:
//  PackedMrm(input)
//      buffer containing a packed MRM to be instantiated as a CMrm
//      object in operational representation
//  MRMFilePath(input)     
//      filepath for the MRM (.mrm) file to be used if the archive is
//      to be commited to persistant representation
//  MRMFileName(input)
//      filename for the MRM file (w/o extension) to be used if the 
//      archive is to be commited to persistant representation
// Returns:
//  WF_OK              initialization succeeded
//  WF_FAIL            initialization failed
// Side effects:
//  CSSM & ISL will be loaded implicitly
//  internally calls Init(WF_CONST_DATA_PTR PackedMrm,
//    char* MRMFilePath,char* MRMFileName)
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::Init(WF_CONST_DATA_PTR PackedMrm,
                     char* MRMFilePath,
                     char* MRMFileName)
{
    return( Init(PackedMrm, MRMFilePath, MRMFileName, NULL, NULL));
}


//////////////////////////////////////////////////////////////////////
// CMrm::Init(WF_CONST_DATA_PTR PackedMrm,
//            char* MRMFilePath,
//            char* MRMFileName,
//            WF_GUID *CSPGuid)
//  initializes a CMrm object based on a packed MRM in a memory buffer
//  The CMrm object is initialized with the contents of the buffer, 
//  with the FilePath and FileName used for its external archive 
//  representation, should the object ever call Commit().
//
// Parameters:
//  PackedMrm(input)
//      buffer containing a packed MRM to be instantiated as a CMrm
//      object in operational representation
//  MRMFilePath(input)     
//      filepath for the MRM (.mrm) file to be used if the archive is
//      to be commited to persistant representation
//  MRMFileName(input)
//      filename for the MRM file (w/o extension) to be used if the 
//      archive is to be commited to persistant representation
//  CSPGuid(input)
//      guid of the CDSA CSP to be used for all cryptographic 
//      operations WRT manifests
// Returns:
//  WF_OK              initialization succeeded
//  WF_FAIL            initialization failed
// Side effects:
//  CSSM & ISL will be loaded implicitly
//  All manifests enclosed in MRM will be commited to disk, then 
//   opened/locked and constructed into their operational 
//   representation
//  MRMFilePath\MRMFilename will be opened and locked for exclusive 
//   writing
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::Init(WF_CONST_DATA_PTR PackedMrm,
                     char* MRMFilePath,
                     char* MRMFileName,
                     WF_GUID *CSPGuid,
                     WF_CSSM_METHODS *CSSMMethods)
{
    FILE* fh = 0;
    PK_ARCHIVE_PTR pArchive = NULL;
    PK_ITERATOR_PTR pIterator = NULL;
    WF_CONST_DATA Filename, Filedata;
    int CredFound = 0, SMFound = 0;

    // unpack mrm
    pArchive = mrm_pk_BuildArchiveFromFileimage(*PackedMrm);
    if( !pArchive )
    {
        // pk file error
        return WF_FAIL;
    }

	mVersion = new CMRMVersion();
	memcpy(&(mVersion->mMajor), 
           (pArchive->DirEnd->ZipFileComment.Data)+2, 
           1);
	memcpy(&(mVersion->mMinor), 
           (pArchive->DirEnd->ZipFileComment.Data)+3, 
           1);

	pIterator = mrm_pk_CreateFileEnumerator(pArchive);
    if( !pIterator )
    {
        // pk file error
        Recycle();
        return WF_FAIL;
    }

	while (mrm_pk_GetNextFile(pIterator, &Filename, &Filedata) == PK_OK)
	{
        char Name[WF_MAX_PATH];
        memset(Name, 0, WF_MAX_PATH);
        if( Filename.Length < WF_MAX_PATH )
            memcpy(Name, Filename.Data, Filename.Length);
        else
        {
            // pk file error - invalid filename
            Recycle();
            mrm_pk_RecycleFileEnumerator(pIterator);
            mrm_pk_RecycleArchive(pArchive);
            return WF_FAIL;
        }
        // keep track of whether the cred or SM was found
        if( strcmp(Name, MRM_REQUEST_CREDENTIAL_FILENAME) == 0 ) CredFound = 1;
        if( strcmp(Name, MRM_SIGNED_MANIFEST_FILENAME) == 0 ) SMFound = 1;
        fh = fopen(Name, "wb");
        if(!fh)
        {
            // file error
            Recycle();
            mrm_pk_RecycleFileEnumerator(pIterator);
            mrm_pk_RecycleArchive(pArchive);
            return WF_FAIL;
        }
        if(fwrite(Filedata.Data, 1, Filedata.Length, fh) != Filedata.Length)
        {
            // file error
            Recycle();
            fclose(fh);
        	mrm_pk_RecycleFileEnumerator(pIterator);
            mrm_pk_RecycleArchive(pArchive);
            return WF_FAIL;
        }
        fclose(fh);
	}
	mrm_pk_RecycleFileEnumerator(pIterator);
    mrm_pk_RecycleArchive(pArchive);

    // initialize manifests in memory
    mRequest = new CManifest();
    if(CSPGuid)
    {
        if(mRequest->Init(MRM_REQUEST_FILEPATH, 
                          CSPGuid,
                          CSSMMethods) != WF_OK)
        {
            // unable to initialize request manifest
            // operational representation of MRM invalid
            Recycle();
            return WF_FAIL;
        }
    }
    else
    {
        if(mRequest->Init(MRM_REQUEST_FILEPATH) != WF_OK)
        {
            // unable to initialize request manifest
            // operational representation of MRM invalid
            Recycle();
            return WF_FAIL;
        }
    }

    mAppID = (WF_GUID*)calloc(sizeof(WF_GUID), 1);
    mContextID = (WF_GUID*)calloc(sizeof(WF_GUID), 1);
    
    CObjectList* Sections = mRequest->GetSectionList();
    uint32 NumOfSections = Sections->GetNumberOfNodes();
    CObjectListNode* currRefNode = Sections->mHead;
    int i;
    for(i=0; i<NumOfSections; i++)
    {   
        if(!strcmp(currRefNode->mRef->mName, "MetaData"))
        {
            uint32 NumOfAttr = currRefNode->mRef->mAttributes->GetNumberOfNodes();
            CAttributeListNode* currAttr = currRefNode->mRef->mAttributes->mHead;
            for(unsigned int j=0; j<NumOfAttr; j++)
            {
                char Tmp[40];
                memset(Tmp, 0, 40);
                if(currAttr->mValue.Length <40)
                    memcpy(Tmp, currAttr->mValue.Data, currAttr->mValue.Length); 
                else
                    memcpy(Tmp, currAttr->mValue.Data, 39);

                if(!memcmp(currAttr->mAttribute.Data,
                           MRM_APP_ID_ATTR,
                           strlen(MRM_APP_ID_ATTR)))
                    wf_StrToGuid(Tmp, mAppID);
 
                if(!memcmp(currAttr->mAttribute.Data,
                           MRM_CTXT_ID_ATTR,
                           strlen(MRM_CTXT_ID_ATTR)))
                    wf_StrToGuid(Tmp, mContextID);
                
                currAttr = currAttr->mNext;
            }
        }
        currRefNode = currRefNode->mNext;
    }
    delete Sections;

    if( CredFound )
    {
        mRequestCredential = new CManifest();
        if(CSPGuid)
        {
            if(mRequestCredential->Init(MRM_REQUEST_CREDENTIAL_FILEPATH, 
                                        CSPGuid,
                                        CSSMMethods) != WF_OK)
            {
                // unable to initialize request credential manifest
                // operational representation of MRM incomplete
                Recycle();
                return WF_FAIL;
            }
        }
        else
        {
            if(mRequestCredential->Init(MRM_REQUEST_CREDENTIAL_FILEPATH) != WF_OK)
            {
                // unable to initialize request credential manifest
                // operational representation of MRM incomplete
                Recycle();
                return WF_FAIL;
            }
        }
    }
    else 
        mRequestCredential = NULL;
    if( SMFound )
    {
        mSignedManifest = new CManifest();
        if(CSPGuid)
        {
            if(mSignedManifest->Init(MRM_SIGNED_MANIFEST_FILEPATH, 
                                     CSPGuid,
                                     CSSMMethods) != WF_OK)
            {
                // unable to initialize signed manifest
                // operational representation of MRM incomplete
                Recycle();
                return WF_FAIL;
            }
        }
        else
        {
            if(mSignedManifest->Init(MRM_SIGNED_MANIFEST_FILEPATH) != WF_OK)
            {
                // unable to initialize request credential manifest
                // operational representation of MRM incomplete
                Recycle();
                return WF_FAIL;
            }
        }
    }
    else 
        mSignedManifest = NULL;

    // initialize object list from request object list
    CObjectList *list = mRequest->GetSectionList();
    if(!list)
    {
        // unable to get object list from request
        // invalid op rep for MRM
        Recycle();
        return WF_FAIL;
    }
    
    mContentList = new CObjectList();
    // remove DONOTSIGN sections
    uint32 NumObjNodes = list->GetNumberOfNodes();
    CObjectListNode* currNode = list->mHead;
    for(i=0; i<NumObjNodes; i++)
    {
        int MetaFlag = 0;
        CAttributeList* AttrList = currNode->mRef->mAttributes;
        CAttributeListNode* currAttr = currNode->mRef->mAttributes->mHead;
        uint32 NumAttrNodes = currNode->mRef->mAttributes->GetNumberOfNodes();
        for(uint32 j=0; j<NumAttrNodes; j++)
        {
            char name[WF_MAX_STRING], value[WF_MAX_STRING];
            memset( name, 0, WF_MAX_STRING );
            memset( value, 0, WF_MAX_STRING );
        
            if(currAttr->mAttribute.Length < WF_MAX_STRING)
                memcpy(name, currAttr->mAttribute.Data, 
                       currAttr->mAttribute.Length );
            else // truncate because strings are too long
                memcpy(name, currAttr->mAttribute.Data, WF_MAX_STRING-1);
        
            if(currAttr->mValue.Length < WF_MAX_STRING)
                memcpy(value, currAttr->mValue.Data, 
                       currAttr->mValue.Length );
            else // truncate because strings are too long
                memcpy( value, currAttr->mValue.Data, WF_MAX_STRING-1);

            if(!strcmp(name, MRM_META_ATTR) && !strcmp(value, MRM_META_VALUE))
                MetaFlag = 1;

            currAttr = currAttr->mNext;
        }

        if(!MetaFlag)
            mContentList->AddNode(currNode); 
        currNode = currNode->mNext;
    }
	delete list;

    // initialize remaining members
    mExternalFilePath = (char*)calloc(strlen(MRMFilePath)+strlen(MRMFileName)+6, 1);
    sprintf( mExternalFilePath, "%s\\%s.mrm", MRMFilePath, MRMFileName );

    return WF_OK;
}

//////////////////////////////////////////////////////////////////////
// CMrm::Init(CObjectList *List, 
//            char *MRMFilePath,
//            char *MRMFileName,
//            const WF_GUID *AppID, 
//            const WF_GUID *ContextID,
//            WF_CONST_DATA *Certificate, 
//            WF_CONST_DATA *Password, 
//            WF_CONST_DATA *SignerName, 
//            WF_CERTGROUP_PTR CertGroup,
//            CAttributeList *SignerAttributes)
//  initializes a CMrm object based on a packed MRM in a memory buffer
//  The CMrm object is initialized with the contents of the buffer, 
//  with the FilePath and FileName used for its external archive 
//  representation, should the object ever call Commit().
//
// Parameters:
//  MRMFilePath(input)     
//      filepath for the MRM (.mrm) file to be used if the archive is
//      to be commited to persistant representation
//  MRMFileName(input)
//      filename for the MRM file (w/o extension) to be used if the 
//      archive is to be commited to persistant representation
//  AppID(input)
//  ContextID(input)
//  Certificate(input)
//  Password(input)
//  SignerName(input)
//  CertGroup(input)
//  SignerAttributes(input)
// Returns:
//  WF_OK              initialization succeeded
//  WF_FAIL            initialization failed
// Side effects:
//  CSSM & ISL will be loaded implicitly
//  All manifests enclosed in MRM will be commited to disk
//  internally calls Init(CObjectList *List, 
//    char *MRMFilePath, char *MRMFileName, const WF_GUID *AppID, 
//    const WF_GUID *ContextID, WF_CONST_DATA *Certificate, 
//    WF_CONST_DATA *Password, WF_CONST_DATA *SignerName, 
//    WF_CERTGROUP_PTR CertGroup, CAttributeList *SignerAttributes,
//    WF_HANDLE hCSP)
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::Init(CObjectList *List, 
                     char *MRMFilePath,
                     char *MRMFileName,
                     const WF_GUID *AppID, 
                     const WF_GUID *ContextID,
                     WF_CONST_DATA *Certificate, 
                     WF_CONST_DATA *Password, 
                     WF_CONST_DATA *SignerName, 
                     WF_CERTGROUP_PTR CertGroup,
                     CAttributeList *SignerAttributes)
{
    return(Init(List, MRMFilePath, MRMFileName, AppID, ContextID,
                Certificate, Password, SignerName, CertGroup,
                SignerAttributes, NULL, NULL));
}
//////////////////////////////////////////////////////////////////////
// CMrm::Init(CObjectList *List, 
//            char *MRMFilePath,
//            char *MRMFileName,
//            const WF_GUID *AppID, 
//            const WF_GUID *ContextID,
//            WF_CONST_DATA *Certificate, 
//            WF_CONST_DATA *Password, 
//            WF_CONST_DATA *SignerName, 
//            WF_CERTGROUP_PTR CertGroup,
//            CAttributeList *SignerAttributes,
//            WF_GUID *CSPGuid)
//  initializes a CMrm object based on a packed MRM in a memory buffer
//  The CMrm object is initialized with the contents of the buffer, 
//  with the FilePath and FileName used for its external archive 
//  representation, should the object ever call Commit().
//
// Parameters:
//  MRMFilePath(input)     
//      filepath for the MRM (.mrm) file to be used if the archive is
//      to be commited to persistant representation
//  MRMFileName(input)
//      filename for the MRM file (w/o extension) to be used if the 
//      archive is to be commited to persistant representation
//  AppID(input)
//  ContextID(input)
//  Certificate(input)
//  Password(input)
//  SignerName(input)
//  CertGroup(input)
//  SignerAttributes(input)
//  CSPGuid(input)
//      guid of the CDSA CSP to be used for all cryptographic 
//      operations WRT manifests
// Returns:
//  WF_OK              initialization succeeded
//  WF_FAIL            initialization failed
// Side effects:
//  CSSM & ISL will be loaded implicitly
//  All manifests enclosed in MRM will be commited to disk, then 
//   opened/locked and constructed into their operational 
//   representation
//  MRMFilePath\MRMFilename will be opened and locked for exclusive 
//   writing
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::Init(CObjectList *List, 
                     char *MRMFilePath,
                     char *MRMFileName,
                     const WF_GUID *AppID, 
                     const WF_GUID *ContextID,
                     WF_CONST_DATA *Certificate, 
                     WF_CONST_DATA *Password, 
                     WF_CONST_DATA *SignerName, 
                     WF_CERTGROUP_PTR CertGroup,
                     CAttributeList *SignerAttributes,
                     WF_GUID *CSPGuid,
                     WF_CSSM_METHODS *CSSMMethods)
{
    // check input
    if( !List || !MRMFilePath || !MRMFileName ||
        !Certificate || !Password || !SignerName )
    {
        // invalid input params
        return WF_FAIL;
    }

    char tmp[WF_MAX_PATH];
    memset( tmp, 0, WF_MAX_PATH );

    // ContentList
    mContentList = new CObjectList(List);
    
    // AppID
    if(AppID)
	{
        mAppID = (WF_GUID*)calloc(sizeof(WF_GUID), 1);
		memcpy(mAppID, AppID, sizeof(WF_GUID));
	}
    else
    {
        // invalid AppID
        Recycle();
        return WF_FAIL;
    }
    
    //ContextID
    if(ContextID)
	{
        mContextID = (WF_GUID*)calloc(sizeof(WF_GUID), 1);
		memcpy(mContextID, ContextID, sizeof(WF_GUID));
	}
    else
    {
        // invalid context id
        Recycle();
        return WF_FAIL;
    }

    // ExternalFilepath
    mExternalFilePath = (char*)calloc(strlen(MRMFilePath)+strlen(MRMFileName)+6, 1);
    sprintf( mExternalFilePath, "%s\\%s.mrm", MRMFilePath, MRMFileName );

    // initialize manifests
    mRequest = new CManifest();
    if(CSPGuid)
    {
        if(mRequest->Init(List, 
                          MRM_REQUEST_FILEPATH,
                          CSPGuid,
                          CSSMMethods) != WF_OK)
        {
            // failed to initialize request manifest
            // MRM operational representation invalid
            Recycle();
            return WF_FAIL;
        }
    }
    else
    {
        if(mRequest->Init(List, 
                          MRM_REQUEST_FILEPATH) != WF_OK)
        {
            // failed to initialize request manifest
            // MRM operational representation invalid
            Recycle();
            return WF_FAIL;
        }
    }

    // zero-ize manifest pointers to be created by reviewers/signers
    mRequestCredential = NULL;
    mSignedManifest = NULL;

    // Version
    mVersion = new CMRMVersion();
    mVersion->mMajor = MRM_VER_MAJOR;
    mVersion->mMinor = MRM_VER_MINOR;

    // create meta-data section with AppID, ContextID, etc.
    CObjectList TmpList;
    WF_CONST_DATA MetaDataAttr, MetaDataValue;
    WF_CONST_DATA ContextIDAttr, ContextIDValue;
    WF_CONST_DATA AppIDAttr, AppIDValue;
    WF_CONST_DATA IntegTagAttr, IntegTagValue;

    IntegTagAttr.Data = (unsigned char*)MRM_REF_ONLY_ATTR;
	IntegTagAttr.Length = strlen(MRM_REF_ONLY_ATTR);
	IntegTagValue.Data = (unsigned char*)MRM_REF_ONLY_VALUE;
	IntegTagValue.Length = strlen((char*)MRM_REF_ONLY_VALUE);
    CAttributeListNode IntegAttr(&IntegTagAttr, &IntegTagValue);

    MetaDataAttr.Data = (unsigned char*)MRM_META_ATTR;
	MetaDataAttr.Length = strlen(MRM_META_ATTR);
	MetaDataValue.Data = (unsigned char*)MRM_META_VALUE;
	MetaDataValue.Length = strlen((char*)MRM_META_VALUE);
    CAttributeListNode MetaAttr(&MetaDataAttr, &MetaDataValue);

	char TmpStr[40];
    memset(TmpStr, 0, 40);
    ContextIDAttr.Data = (unsigned char*)MRM_CTXT_ID_ATTR;
	ContextIDAttr.Length = strlen(MRM_CTXT_ID_ATTR);
    wf_GuidToStr(ContextID, TmpStr);
	ContextIDValue.Data = (unsigned char*)TmpStr;
	ContextIDValue.Length = strlen(TmpStr);
    CAttributeListNode CtxtAttr(&ContextIDAttr, &ContextIDValue);
    
    memset(TmpStr, 0, 40);
    AppIDAttr.Data = (unsigned char*)MRM_APP_ID_ATTR;
	AppIDAttr.Length = strlen(MRM_APP_ID_ATTR);
    wf_GuidToStr(AppID, TmpStr);
    AppIDValue.Data = (unsigned char*)TmpStr;
	AppIDValue.Length = strlen(TmpStr);
    CAttributeListNode AppAttr(&AppIDAttr, &AppIDValue);
    
    CAttributeList AttrList;
    AttrList.AddNode(&IntegAttr);
    AttrList.AddNode(&MetaAttr);
    AttrList.AddNode(&AppAttr);
    AttrList.AddNode(&CtxtAttr);
    
    CObjectRef TmpRef("MetaData", RefFile, "MetaData", "", &AttrList);
    CObjectListNode TmpNode(&TmpRef);
    TmpList.AddNode(&TmpNode);
    mRequest->AddSections(&TmpList);

    // sign request, all Mrm's must have a signed request manifest to
    //  have a valid operational representation
    if(mRequest->SignAllSections(Certificate, 
                                 Password, 
                                 SignerName, 
                                 CertGroup, 
                                 SignerAttributes) != WF_OK)
    {
        // unable to sign request manifest
        // MRM operational representation invalid
        Recycle();
        return WF_FAIL;
    }
    
    return WF_OK;
}

//////////////////////////////////////////////////////////////////////
// CMrm::~CMrm()
//  destructs a CMrm object
//
// Parameters:
//  none
// Returns:
//  none
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
CMrm::~CMrm()
{
    Recycle();
}

//////////////////////////////////////////////////////////////////////
// CMrm::Recycle()
//  frees all memory directly allocated by a mrm object and
//  zero-izes it
//
// Parameters:
//  none
// Returns:
//  Closes/unlocks the archive file associated with the mrm object
//////////////////////////////////////////////////////////////////////
void CMrm::Recycle()
{
    if(mContentList)
        delete mContentList;
    mContentList = NULL;

	if(mVersion)
        delete mVersion;
    mVersion = NULL;

	if(mRequest)
        delete mRequest;
    mRequest = NULL;

	if(mRequestCredential)
        delete mRequestCredential;
	mRequestCredential = NULL;

    if(mSignedManifest)
        delete mSignedManifest;
    mSignedManifest = NULL;

    if(mExternalFilePath)
        free( mExternalFilePath );
	mExternalFilePath = NULL;

    if(mAppID)
    {
        free(mAppID);
        mAppID = NULL;
    }
	
    if(mContextID)
    {
        free(mContextID);
        mContextID = NULL;
    }
}

//////////////////////////////////////////////////////////////////////
// CMrm::SignRequest(WF_CONST_DATA *Certificate, 
//                   WF_CONST_DATA *Password,
//                   WF_CONST_DATA *SignerName,
//                   WF_CERTGROUP_PTR CertGroup,
//                   CAttributeList *SignerAttributes,
//                   MRM_SIGN_FLAG Flag)
//  signs the request credential manifest using the  provided 
//  Certificate, Password, SignerName, CertGroup, and SignerAttributes.  
//  The flag is evaluated to tell how to sign the request.
//
// Parameters:
//  Certificate(input)
//      certificate to be used for the signing operation
//  Password(input)
//      password for use of the private key associated with the 
//      Certificate
//  SignerName(input)
//      name to be used to represent the signer in the manifest
//  CertGroup(input)
//      certificate group associated with the Certificate which may
//      help in determining trust for the Certificate
//  SignerAttributes(input)
//      attributes to be included with the signature
//  Flag(input)
//      flag used to determine the how to sign the request
// Returns:
//  WF_OK              Signing operation succeeded
//  WF_FAIL            Signing operation failed
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::SignRequest(WF_CONST_DATA *Certificate, 
                            WF_CONST_DATA *Password,
                            WF_CONST_DATA *SignerName,
                            WF_CERTGROUP_PTR CertGroup,
                            CAttributeList *SignerAttributes,
                            MRM_SIGN_FLAG Flag )
{
    Flag;
    if( mRequestCredential )
    {
	    if( mRequestCredential->SignAllSections(Certificate, 
                                                Password,
                                                SignerName, 
                                                CertGroup,
                                                SignerAttributes) != WF_OK )
            return WF_FAIL;
    }
    else
    {
        //TO DO: need to create RC based on attr in manifest sections
        mRequestCredential = new CManifest();
        if(mRequestCredential->Init(mContentList, 
                                    MRM_REQUEST_CREDENTIAL_FILEPATH) != WF_OK)
        {
            delete mRequestCredential;
            return WF_FAIL;
        }
        if( mRequestCredential->SignAllSections(Certificate, 
                                                Password,
                                                SignerName, 
                                                CertGroup,
                                                SignerAttributes) != WF_OK )
            return WF_FAIL;
    }

    return WF_OK;
}

//////////////////////////////////////////////////////////////////////
// CMrm::CounterSignRequest(WF_CONST_DATA *Certificate, 
//                          WF_CONST_DATA *Password,
//                          WF_CONST_DATA *CounterSignerName,
//                          WF_CONST_DATA *SignerName,
//                          WF_CERTGROUP_PTR CertGroup,
//                          CAttributeList *SignerAttributes,
//                          MRM_SIGN_FLAG Flag)
//  countersigns a signature in the request credential manifest using 
//  the  provided Certificate, Password, SignerName, CertGroup, and 
//  SignerAttributes. The flag is evaluated to tell how to sign the 
//  request.
//
// Parameters:
//  Certificate(input)
//      certificate to be used for the signing operation
//  Password(input)
//      password for use of the private key associated with the 
//      Certificate
//  CounterSignerName(input)
//      name to be used to represent the signer in the manifest
//  SignerName(input)
//      name of the signer to be countersigned
//  CertGroup(input)
//      certificate group associated with the Certificate which may
//      help in determining trust for the Certificate
//  SignerAttributes(input)
//      attributes to be included with the signature
//  Flag(input)
//      flag used to determine the how to sign the request
// Returns:
//  WF_OK              Signing operation succeeded
//  WF_FAIL            Signing operation failed
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::CounterSignRequest(WF_CONST_DATA *Certificate, 
                                   WF_CONST_DATA *Password,
                                   WF_CONST_DATA *CounterSignerName,
                                   WF_CONST_DATA *SignerName,
                                   WF_CERTGROUP_PTR CertGroup,
                                   CAttributeList *SignerAttributes,
                                   MRM_SIGN_FLAG Flag )
{
    Flag; CertGroup; CounterSignerName; SignerName; Password; Certificate;
//TO DO: FINISH ME
   return WF_OK;
}

//////////////////////////////////////////////////////////////////////
// CMrm::SignManifest(WF_CONST_DATA *Certificate, 
//                    WF_CONST_DATA *Password,
//                    WF_CONST_DATA *SignerName,
//                    WF_CERTGROUP_PTR CertGroup,
//                    CAttributeList *SignerAttributes,
//                    MRM_SIGN_FLAG Flag)
//  signs the signed manifest using the  provided Certificate, 
//  Password, SignerName, CertGroup, and SignerAttributes.  The flag 
//  is evaluated to tell how to sign the request.
//
// Parameters:
//  Certificate(input)
//      certificate to be used for the signing operation
//  Password(input)
//      password for use of the private key associated with the 
//      Certificate
//  SignerName(input)
//      name to be used to represent the signer in the manifest
//  CertGroup(input)
//      certificate group associated with the Certificate which may
//      help in determining trust for the Certificate
//  SignerAttributes(input)
//      attributes to be included with the signature
//  Flag(input)
//      flag used to determine the how to sign the request
// Returns:
//  WF_OK              Signing operation succeeded
//  WF_FAIL            Signing operation failed
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::SignManifest(WF_CONST_DATA *Certificate, 
                             WF_CONST_DATA *Password,
                             WF_CONST_DATA *SignerName,
                             WF_CERTGROUP_PTR CertGroup,
                             CAttributeList *SignerAttributes,
                             MRM_SIGN_FLAG Flag )
{
    Flag; 
    if( mSignedManifest )
    {
    	if( mSignedManifest->SignAllSections(Certificate, 
                                             Password,
                                             SignerName,
                                             CertGroup,
                                             SignerAttributes) != WF_OK )
            return WF_FAIL;
    }
    else
    {
        //TO DO: need to create SM based on attr in manifest sections
        mSignedManifest = new CManifest();
        if(mSignedManifest->Init(mContentList, 
                                 MRM_SIGNED_MANIFEST_FILEPATH) != WF_OK)
        {
            delete mSignedManifest;
            return WF_FAIL;
        }
        if( mSignedManifest->SignAllSections(Certificate, 
                                             Password,
                                             SignerName, 
                                             CertGroup,
                                             SignerAttributes) != WF_OK )
            return WF_FAIL;
    }

    return WF_OK;
}

//////////////////////////////////////////////////////////////////////
// CMrm::CounterSignManifest(WF_CONST_DATA *Certificate, 
//                           WF_CONST_DATA *Password,
//                           WF_CONST_DATA *CounterSignerName,
//                           WF_CONST_DATA *SignerName,
//                           WF_CERTGROUP_PTR CertGroup,
//                           CAttributeList *SignerAttributes,
//                           MRM_SIGN_FLAG Flag)
//  countersigns a signature in the signed manifest using the  
//  provided Certificate, Password, SignerName, CertGroup, and 
//  SignerAttributes. The flag is evaluated to tell how to sign the 
//  manifest.
//
// Parameters:
//  Certificate(input)
//      certificate to be used for the signing operation
//  Password(input)
//      password for use of the private key associated with the 
//      Certificate
//  CounterSignerName(input)
//      name to be used to represent the signer in the manifest
//  SignerName(input)
//      name of the signer to be countersigned
//  CertGroup(input)
//      certificate group associated with the Certificate which may
//      help in determining trust for the Certificate
//  SignerAttributes(input)
//      attributes to be included with the signature
//  Flag(input)
//      flag used to determine the how to sign the request
// Returns:
//  WF_OK              Signing operation succeeded
//  WF_FAIL            Signing operation failed
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::CounterSignManifest(WF_CONST_DATA *Certificate, 
                                    WF_CONST_DATA *Password,
                                    WF_CONST_DATA *CounterSignerName,
                                    WF_CONST_DATA *SignerName,
                                    WF_CERTGROUP_PTR CertGroup,
                                    CAttributeList *SignerAttributes,
                                    MRM_SIGN_FLAG Flag)
{
    Flag; CertGroup; CounterSignerName; SignerName; Password; Certificate;
//TO DO: FINISH ME
	return WF_OK;
}

//////////////////////////////////////////////////////////////////////
// CMrm::VerifyIntegrity()
//  verifies all signers in each manifest object.  This verification will
//  check that the internal structure of each manifest is intact by
//  walking from each signer to all of the nodes signed by that signer
//
// Parameters:
//  none
// Returns:
//  WF_OK              integrity verified successfully
//  WF_FAIL            integrity verified successfully
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::VerifyIntegrity()
{
    // check request
    if( mRequest->VerifyAllSigners() != WF_OK )
        return WF_FAIL;
    // check credential
    if( mRequestCredential )
    {
        if( mRequestCredential->VerifyAllSigners() != WF_OK )
            return WF_FAIL;
    }
    // check signed manifest
    if( mSignedManifest )
    {
        if( mSignedManifest->VerifyAllSigners() != WF_OK )
            return WF_FAIL;
    }

    // if we make it here, everything checks out
    return WF_OK;
}

//////////////////////////////////////////////////////////////////////
// CMrm::VerifyContent()
//  verifies all sections in each manifest object.  This verification will
//  check that the hash values contained in each section match the 
//  actual hash values for the referant object
//
// Parameters:
//  none
// Returns:
//  WF_OK              all content verified successfully
//  WF_FAIL            content verification failed
// Side effects:
//  MRM will do a commit operation prior to verification
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::VerifyContent()
{
    Commit();

    // check request
    if( mRequest->VerifyAllSections() != WF_OK )
        return WF_FAIL;
    // check credential
    if( mRequestCredential )
    {
        if( mRequestCredential->VerifyAllSections() != WF_OK )
            return WF_FAIL;
    }
    // check signed manifest
    if( mSignedManifest )
    {
        if( mSignedManifest->VerifyAllSections() != WF_OK )
            return WF_FAIL;
    }

    // if we make it here, everything checks out
    return WF_OK;
}

//////////////////////////////////////////////////////////////////////
// CMrm::Verify()
//  verifies both integrity and content for all signers and all
//  sections in each contained manifest.  This verification will
//  check that the hash values contained in each section match the 
//  actual hash values for the referant object.  It will also walk the
//  manifests for each signer contained in it.
//
// Parameters:
//  none
// Returns:
//  WF_OK              MRM verification successful
//  WF_FAIL            MRM verification failed
// Side effects:
//  Calls CMrm::VerifyIntegrity() and CMrm::VerifyContent()
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::Verify()
{
    WF_RETURN Rtn;

    // verify MRM integrity
    Rtn = VerifyIntegrity();
    if( Rtn != WF_OK )
    {
        return WF_FAIL;
    }

    // verify MRM content references
    Rtn = VerifyContent();
    if( Rtn != WF_OK )
    {
        return WF_FAIL;
    }

    // assertion: all verifications succeeded
    return WF_OK;
}

//////////////////////////////////////////////////////////////////////
// CMrm::GetSignerGraph(CSignatureGraph* SigGraph)
//  retrieves the graph of all signatures applied to the signed 
//  manifest portion of the MRM
//
// Parameters:
//  SigGraph(output)    
//      pointer to an empty SigGraph object to be populated by this call
// Returns:
//  WF_OK              signature graph retrieved successfully
//  WF_FAIL            signature graph not retrieved successfully
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::GetSignerGraph(CSignatureGraph* SigGraph)
{
    if(mSignedManifest)	
        return( mSignedManifest->GetSignatureGraph(SigGraph));
    else 
        return WF_FAIL;
}

//////////////////////////////////////////////////////////////////////
// CMrm::GetReviewerGraph(CSignatureGraph* SigGraph)
//  retrieves the graph of all signatures applied to the request 
//  credential manifest in the MRM
//
// Parameters:
//  SigGraph(output)    
//      pointer to an empty SigGraph object to be populated by this call
// Returns:
//  WF_OK              signature graph retrieved successfully
//  WF_FAIL            signature graph not retrieved successfully
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::GetReviewerGraph(CSignatureGraph* SigGraph)
{
    if(mRequestCredential)	
        return( mRequestCredential->GetSignatureGraph(SigGraph));
    else 
        return WF_FAIL;
}

//////////////////////////////////////////////////////////////////////
// CMrm::GetVersion()
//  retrieves the version object associated with the MRM
//
// Parameters:
//  none
// Returns:
//  valid ptr           version object for the MRM
//  NULL                no version found
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
CMRMVersion CMrm::GetVersion()
{
	return( *mVersion );
}

//////////////////////////////////////////////////////////////////////
// CMrm::GetAppID()
//  retrieves the AppID associated with the MRM
//
// Parameters:
//  none
// Returns:
//  AppID
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
WF_GUID* CMrm::GetAppID()
{
	return( mAppID );
}

//////////////////////////////////////////////////////////////////////
// CMrm::GetContextID()
//  retrieves the context ID associated with the MRM
//
// Parameters:
//  none
// Returns:
//  ContextID
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
WF_GUID* CMrm::GetContextID()
{
	return( mContextID );
}

//////////////////////////////////////////////////////////////////////
// CMrm::Commit()
//  commits the operational representation of the mrm object to its 
//  persistent archive representation
//
// Parameters:
//  none
// Returns:
//  WF_OK              commit operation completed successfully
//  WF_FAIL            commit operation did not complete successfully
// Side effects:
//  The archive file is unlocked/closed, the relocked/opened after the
//   commit operation completes
//  mArchiveFileHandle will have a new value on exit
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::Commit()
{
    WF_DATA PrincipalID;
    uint8 Tmp[20];
    memset(Tmp, 0, 20);
    PrincipalID.Data = Tmp;

    return(Commit(&PrincipalID));
}
WF_RETURN CMrm::Commit(WF_DATA *PrincipalID)
{
    WF_CONST_DATA PathData = {strlen(mExternalFilePath), (unsigned char*)mExternalFilePath};
    WF_CONST_DATA RequestPath = 
                    {strlen(MRM_REQUEST_FILEPATH), 
                     (unsigned char*)MRM_REQUEST_FILEPATH},
                  ReqCredPath = 
                    {strlen(MRM_REQUEST_CREDENTIAL_FILEPATH), 
                     (unsigned char*)MRM_REQUEST_CREDENTIAL_FILEPATH},
                  ManifestPath = 
                    {strlen(MRM_SIGNED_MANIFEST_FILEPATH),
                     (unsigned char*)MRM_SIGNED_MANIFEST_FILEPATH},
                  TmpPath;
    PK_NEW_ARCHIVE_PTR	pNewArchive = NULL;
	WF_CONST_DATA	NullData = {0, NULL};
    CObjectListNode* node = NULL;
    char tmp[WF_MAX_PATH];
    int node_ct = 0;

    if( !mRequest )
        return WF_FAIL;

    pNewArchive = mrm_pk_CreateNewArchive();
    if (!pNewArchive)
		return WF_FAIL;
	if( mRequest->Commit() != WF_OK )
    {
        mrm_pk_RecycleNewArchive(pNewArchive);
        return WF_FAIL;
    }
    if (mrm_pk_AddFileToNewArchive(pNewArchive, RequestPath, NullData) != PK_OK)
    {
        mrm_pk_RecycleNewArchive(pNewArchive);
        return WF_FAIL;
    }
    if( mRequestCredential )
    {
        if(mRequestCredential->Commit() != WF_OK)
        {
            mrm_pk_RecycleNewArchive(pNewArchive);
            return WF_FAIL;
        }
        if(mrm_pk_AddFileToNewArchive(pNewArchive, ReqCredPath, NullData) != PK_OK)
        {
            mrm_pk_RecycleNewArchive(pNewArchive);
            return WF_FAIL;
        }
    }
    if( mSignedManifest )
    {
        if(mSignedManifest->Commit() != WF_OK)
        {
            mrm_pk_RecycleNewArchive(pNewArchive);
            return WF_FAIL;
        }
        if(mrm_pk_AddFileToNewArchive(pNewArchive, ManifestPath, NullData) != PK_OK)
        {
            mrm_pk_RecycleNewArchive(pNewArchive);
            return WF_FAIL;
        }
    }
	
/* ADD ALL CONTENT OBJECTS TO ARCHIVE */
    node = mContentList->mHead;
	node_ct = mContentList->GetNumberOfNodes();
    for( int i=0; i<node_ct && node; i++ )
	{
#ifdef WIN32
        sprintf((char*)tmp,"%s\\%s", MRM_SCRATCH_DIR, node->mRef->mReferencePath);
#else
        sprintf((char*)tmp,"%s/%s", MRM_SCRATCH_DIR, node->mRef->mReferencePath);
#endif
        TmpPath.Data = (unsigned char*)tmp;
        TmpPath.Length = strlen( (char*)TmpPath.Data );
        if(mrm_pk_AddFileToNewArchive(pNewArchive, TmpPath, NullData) != PK_OK)
        {
            mrm_pk_RecycleNewArchive(pNewArchive);
            return WF_FAIL;
        }
        node = node->mNext;
	}

    if(mrm_pk_WriteNewMRMArchive(pNewArchive, PathData, mVersion, mAppID, mContextID, PrincipalID) != PK_OK)
    {
        mrm_pk_RecycleNewArchive(pNewArchive);
        return WF_FAIL;
    }
	mrm_pk_RecycleNewArchive(pNewArchive);

    // if we got here, everything is committed
    return WF_OK;
}

//////////////////////////////////////////////////////////////////////
// CMrm::ExtractSignedManifest( char* OutputFilePath )
//  extracts the signed manifest from the MRM and archives it at the
//  provided OutputFielPath
//
// Parameters:
//  OutputFilePath(input)
//      path where the extracted signed manifest should be written
//      persistently
// Returns:
//  WF_OK              commit operation completed successfully
//  WF_FAIL            commit operation did not complete successfully
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::ExtractSignedManifest( char* OutputFilePath )
{
    FILE *fh1, *fh2;
    void* buffer;
	struct _stat st;

    fh1 = fopen(MRM_SIGNED_MANIFEST_FILEPATH, "rb");
    if( !fh1 )
        return WF_FAIL;
    fh2 = fopen(OutputFilePath, "wb");
    if( !fh2 )
    {
        fclose(fh1);
        return WF_FAIL;
    }
	/* stat the file and find out how much memory we need */
	if ((_stat(MRM_SIGNED_MANIFEST_FILEPATH,&st)) != 0)
		return WF_FAIL;

    buffer = malloc(st.st_size);
    if(fread(buffer, 1, st.st_size, fh1) != (unsigned int)st.st_size)
    {
        fclose(fh1);
        fclose(fh2);
        free(buffer);
        return WF_FAIL;
    }
    fclose(fh1);
    if(fwrite(buffer, 1, st.st_size, fh2) != (unsigned int)st.st_size)
    {
        fclose(fh2);
        free(buffer);
        return WF_FAIL;
    }
    free(buffer);
    fclose(fh2);

    return WF_OK;
}

//////////////////////////////////////////////////////////////////////
// CMrm::GetPackedSize()
//  gets the size of the MRM in its packed memory form
//
// Parameters:
//  none
// Returns:
//  size of the MRM if packed
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
WF_SIZE CMrm::GetPackedSize()
{
    WF_SIZE Size = 0;
    WF_CONST_DATA PathData = {strlen(mExternalFilePath), (unsigned char*)mExternalFilePath};
    WF_CONST_DATA RequestPath = 
                    {strlen(MRM_REQUEST_FILEPATH), 
                     (unsigned char*)MRM_REQUEST_FILEPATH},
                   ReqCredPath = 
                    {strlen(MRM_REQUEST_CREDENTIAL_FILEPATH), 
                     (unsigned char*)MRM_REQUEST_CREDENTIAL_FILEPATH},
                   ManifestPath = 
                    {strlen(MRM_SIGNED_MANIFEST_FILEPATH),
                     (unsigned char*)MRM_SIGNED_MANIFEST_FILEPATH},
                  TmpPath;
    PK_NEW_ARCHIVE_PTR	pNewArchive = NULL;
	WF_CONST_DATA	NullData = {0, NULL};
    CObjectListNode* node = NULL;
    char tmp[WF_MAX_PATH];
    int node_ct = 0;

    pNewArchive = mrm_pk_CreateNewArchive();
    if (!pNewArchive)
		return 0;
	
    if( mRequest->Commit() != WF_OK )
    {
        mrm_pk_RecycleNewArchive(pNewArchive);
        return 0;
    }
    if (mrm_pk_AddFileToNewArchive(pNewArchive, RequestPath, NullData) != PK_OK)
		return 0;
    if( mRequestCredential )
    {
        if(mRequestCredential->Commit() != WF_OK)
        {
            mrm_pk_RecycleNewArchive(pNewArchive);
            return 0;
        }
        if(mrm_pk_AddFileToNewArchive(pNewArchive, ReqCredPath, NullData) != PK_OK)
        {
            mrm_pk_RecycleNewArchive(pNewArchive);
            return 0;
        }
    }
    if( mSignedManifest )
    {
        if(mSignedManifest->Commit() != WF_OK)
        {
            mrm_pk_RecycleNewArchive(pNewArchive);
            return 0;
        }
        if(mrm_pk_AddFileToNewArchive(pNewArchive, ManifestPath, NullData) != PK_OK)
        {
            mrm_pk_RecycleNewArchive(pNewArchive);
            return 0;
        }
    }
	
/* ADD ALL CONTENT OBJECTS TO ARCHIVE */
    node = mContentList->mHead;
	node_ct = mContentList->GetNumberOfNodes();
    for( int i=0; i<node_ct && node; i++ )
	{
#ifdef WIN32
        sprintf((char*)tmp,"%s\\%s", MRM_SCRATCH_DIR, node->mRef->mReferencePath);
#else
        sprintf((char*)tmp,"%s/%s", MRM_SCRATCH_DIR, node->mRef->mReferencePath);
#endif
        TmpPath.Data = (unsigned char*)tmp;
        TmpPath.Length = strlen( (char*)TmpPath.Data );
        if(mrm_pk_AddFileToNewArchive(pNewArchive, TmpPath, NullData) != PK_OK)
        {
            mrm_pk_RecycleNewArchive(pNewArchive);
            return 0;
        }
        node = node->mNext;
	}

    WF_DATA PrincipalID;
    Size = mrm_pk_GetSizeOfArchive(pNewArchive, PathData, mVersion, 
                                   mAppID, mContextID, &PrincipalID);
    mrm_pk_RecycleNewArchive(pNewArchive);
    
    return Size;
}

//////////////////////////////////////////////////////////////////////
// CMrm::Pack( WF_CONST_DATA_PTR PackedMrm )
//  packs the operational image of the MRM into the provided buffer
//  The buffer must be pre-allocated with its length field reflecting
//  its length.
//
// Parameters:
//  PackedMrm(output)
//      pre-allocated buffer for storage of the packed MRM
// Returns:
//  WF_OK              MRM verification successful
//  WF_FAIL            MRM verification failed
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
WF_RETURN CMrm::Pack(WF_CONST_DATA_PTR PackedMrm, 
                     WF_DATA* PrincipalID)
{
    WF_CONST_DATA PathData = {strlen(mExternalFilePath), (unsigned char*)mExternalFilePath};
    WF_CONST_DATA RequestPath = 
                    {strlen(MRM_REQUEST_FILEPATH), 
                     (unsigned char*)MRM_REQUEST_FILEPATH},
                   ReqCredPath = 
                    {strlen(MRM_REQUEST_CREDENTIAL_FILEPATH), 
                     (unsigned char*)MRM_REQUEST_CREDENTIAL_FILEPATH},
                   ManifestPath = 
                    {strlen(MRM_SIGNED_MANIFEST_FILEPATH),
                     (unsigned char*)MRM_SIGNED_MANIFEST_FILEPATH},
                  TmpPath;
    PK_NEW_ARCHIVE_PTR	pNewArchive = NULL;
	WF_CONST_DATA	NullData = {0, NULL};
    CObjectListNode* node = NULL;
    char tmp[WF_MAX_PATH];
    int node_ct = 0;

    // validate inputs
    if(!PackedMrm)
        return WF_FAIL;
    if(GetPackedSize() > PackedMrm->Length)
        return WF_FAIL;

    pNewArchive = mrm_pk_CreateNewArchive();
    if (!pNewArchive)
    {
        // pk file failure
		return WF_FAIL;
    }
	
    if( mRequest->Commit() != WF_OK )
    {
        // pk file failure
        mrm_pk_RecycleNewArchive(pNewArchive);
        return WF_FAIL;
    }
    if (mrm_pk_AddFileToNewArchive(pNewArchive, RequestPath, NullData) != PK_OK)
    {
        // pk file failure
        mrm_pk_RecycleNewArchive(pNewArchive);
        return WF_FAIL;
    }
    if( mRequestCredential )
    {
        if(mRequestCredential->Commit() != WF_OK)
        {
            // pk file failure
            mrm_pk_RecycleNewArchive(pNewArchive);
            return WF_FAIL;
        }
        if(mrm_pk_AddFileToNewArchive(pNewArchive, ReqCredPath, NullData) != PK_OK)
        {
            // pk file failure
            mrm_pk_RecycleNewArchive(pNewArchive);
            return WF_FAIL;
        }
    }
    if( mSignedManifest )
    {
        if(mSignedManifest->Commit() != WF_OK)
        {
            // pk file failure
            mrm_pk_RecycleNewArchive(pNewArchive);
            return WF_FAIL;
        }
        if(mrm_pk_AddFileToNewArchive(pNewArchive, ManifestPath, NullData) != PK_OK)
        {
            // pk file failure
            mrm_pk_RecycleNewArchive(pNewArchive);
            return WF_FAIL;
        }
    }
	
/* ADD ALL CONTENT OBJECTS TO ARCHIVE */
    node = mContentList->mHead;
	node_ct = mContentList->GetNumberOfNodes();
    for( int i=0; i<node_ct && node; i++ )
	{
#ifdef WIN32
        sprintf((char*)tmp,"%s\\%s", MRM_SCRATCH_DIR, node->mRef->mReferencePath);
#else
        sprintf((char*)tmp,"%s/%s", MRM_SCRATCH_DIR, node->mRef->mReferencePath);
#endif
        TmpPath.Data = (unsigned char*)tmp;
        TmpPath.Length = strlen( (char*)TmpPath.Data );
        if(mrm_pk_AddFileToNewArchive(pNewArchive, TmpPath, NullData) != PK_OK)
        {
            // pk file failure
            mrm_pk_RecycleNewArchive(pNewArchive);
            return WF_FAIL;
        }
        node = node->mNext;
	}
    
    if( mrm_pk_WriteNewMRMToBuffer(pNewArchive, 
                                   mVersion, 
                                   mAppID,
                                   mContextID,
                                   PrincipalID,
                                   PackedMrm) != PK_OK )
    {
        // pk file failure
        mrm_pk_RecycleNewArchive(pNewArchive);
        return WF_FAIL;
    }

    mrm_pk_RecycleNewArchive(pNewArchive);
    
    return WF_OK;
}

CAttributeList* CMrm::GetAttributes()
{
    CObjectList* Sections = mRequest->GetSectionList();
    uint32 NumOfSections = Sections->GetNumberOfNodes();
    CObjectListNode* currRefNode = Sections->mHead;
	CAttributeList* attribList = new CAttributeList();
    for(uint32 i=0; i<NumOfSections; i++)
    {   
	    uint32 NumOfAttr = currRefNode->mRef->mAttributes->GetNumberOfNodes();
	    CAttributeListNode* currAttr = currRefNode->mRef->mAttributes->mHead;

        for(unsigned int j=0; j<NumOfAttr; j++)
        {
			attribList->AddNode ( currAttr );
			currAttr = currAttr->mNext;
		}
        currRefNode = currRefNode->mNext;
    }
    delete Sections;
	return attribList;
}
