/*-----------------------------------------------------------------------
 * File: MDS_UTIL_QUERY.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/*
 * This file contains the code used to Query MDS.
 */

#ifdef VMS
#include <string.h>
#endif
#include <sys/stat.h>
#include "cssmerr.h"
#include "mds_util_intrn.h"
#include "mds.h"
#include "mds_schema.h"
#include "mds_app_util.h"
#include <cssmMemTrack.h>

void CSSMAPI 
mdsutil_FreeRecord (uint32 NumAttributes,
    CSSM_DB_ATTRIBUTE_DATA *OutputAttributeData)
{
    uint32 i;

    for (i = 0; i < NumAttributes; i++) 
    {
        if (OutputAttributeData[i].Value)
        {
            if (OutputAttributeData[i].Value->Data)
            {
                MDSUTIL_free (OutputAttributeData[i].Value->Data, NULL);
            }

            MDSUTIL_free (OutputAttributeData[i].Value, NULL);
        }
    }
}


CSSM_RETURN CSSMAPI mdsutil_GetRecord(CSSM_DL_DB_HANDLE hDLDB,
                                      CSSM_QUERY Query,
                                      CSSM_DB_RECORDTYPE RecordType,
                                      uint32 NumAttributes,
                                      CSSM_HANDLE_PTR ResultsHandle,
                                      CSSM_DB_ATTRIBUTE_DATA *OutputAttributeData)
{
    CSSM_DB_UNIQUE_RECORD_PTR RecordId = NULL;
    CSSM_DB_RECORD_ATTRIBUTE_DATA OutputAttributes;
    CSSM_RETURN retValue = CSSM_OK;

    memset (&OutputAttributes, 0, sizeof (CSSM_DB_RECORD_ATTRIBUTE_DATA));

    /* We want to get NumAttributes back */
    OutputAttributes.DataRecordType = RecordType;
    OutputAttributes.NumberOfAttributes = NumAttributes;
    OutputAttributes.AttributeData = OutputAttributeData;

    if (0 == *ResultsHandle) 
    {
        if( !IsBadCodePtr((CSSM_PROC_ADDR)MDSFuncs.DataGetFirst)) 
        {
            retValue = MDSFuncs.DataGetFirst (hDLDB, /* DLDBHandle */
                                              &Query, 
                                              ResultsHandle, 
                                              &OutputAttributes, 
                                              NULL, 
                                              &RecordId);
        }
    } else {
        if( !IsBadCodePtr((CSSM_PROC_ADDR)MDSFuncs.DataGetNext)) 
        {
            retValue = MDSFuncs.DataGetNext (hDLDB, /* DLDBHandle */
                                             *ResultsHandle, 
                                             &OutputAttributes, 
                                             NULL, 
                                             &RecordId);
        }
    }

    if (retValue != CSSM_OK)
        return retValue;

    if (!IsBadCodePtr((CSSM_PROC_ADDR)MDSFuncs.FreeUniqueRecord)) 
        MDSFuncs.FreeUniqueRecord(hDLDB, RecordId);

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_unconditionalGetRecord(
    CSSM_DL_DB_HANDLE hDLDB,
    CSSM_DB_RECORDTYPE RecordType,
    uint32 NumAttributes,
    CSSM_HANDLE_PTR ResultsHandle,
    CSSM_DB_ATTRIBUTE_DATA *OutputAttributeData)
{
    CSSM_QUERY Query;

    memset (&Query, 0, sizeof (CSSM_QUERY));
    Query.RecordType = RecordType;
    Query.Conjunctive = CSSM_DB_NONE;
    Query.NumSelectionPredicates = 0;
    Query.SelectionPredicate = NULL;
    Query.QueryLimits.TimeLimit = CSSM_QUERY_TIMELIMIT_NONE;
    Query.QueryLimits.SizeLimit = CSSM_QUERY_SIZELIMIT_NONE;
    Query.QueryFlags = 0;

    return(mdsutil_GetRecord(hDLDB,
                             Query,
                             RecordType,
                             NumAttributes,
                             ResultsHandle,
                             OutputAttributeData));
}

CSSM_RETURN CSSMAPI mdsutil_GetRecordByGuid(CSSM_DL_DB_HANDLE hDLDB,
                                            char* GuidStr,
                                            CSSM_DB_RECORDTYPE RecordType,
                                            uint32 NumAttributes,
                                            CSSM_HANDLE_PTR ResultsHandle,
                                            CSSM_DB_ATTRIBUTE_DATA *OutputAttributeData)
{
    CSSM_RETURN retValue = CSSM_OK;
    CSSM_QUERY Query;
    CSSM_SELECTION_PREDICATE Predicate;

    memset (&Query, 0, sizeof (CSSM_QUERY));
    memset (&Predicate, 0, sizeof (CSSM_SELECTION_PREDICATE));
    Query.RecordType = RecordType;
    Query.Conjunctive = CSSM_DB_NONE;
    Query.NumSelectionPredicates = 1;
    Query.SelectionPredicate = &Predicate;
    Query.QueryLimits.TimeLimit = CSSM_QUERY_TIMELIMIT_NONE;
    Query.QueryLimits.SizeLimit = CSSM_QUERY_SIZELIMIT_NONE;
    Query.QueryFlags = 0;

    Predicate.DbOperator = CSSM_DB_EQUAL;
    
    /* NOTE: The first attribute *NEEDS* to be an index */
    Predicate.Attribute.Info = s_MdsAttrInfo_ModuleId;
    
    Predicate.Attribute.Value = (CSSM_DATA *)MDSUTIL_malloc(sizeof (CSSM_DATA), NULL);
    if (Predicate.Attribute.Value == NULL)
        return CSSMERR_CSSM_MEMORY_ERROR;

    Predicate.Attribute.NumberOfValues = 1;
    Predicate.Attribute.Value->Data = (unsigned char *)GuidStr;
    Predicate.Attribute.Value->Length = (uint32)strlen (GuidStr) + 1;

    retValue = mdsutil_GetRecord(hDLDB,
                                 Query,
                                 RecordType,
                                 NumAttributes,
                                 ResultsHandle,
                                 OutputAttributeData);
    
    MDSUTIL_free(Predicate.Attribute.Value, NULL);

    return retValue;
}

CSSM_RETURN CSSMAPI mdsutil_GetPrimaryRecordManifest(char* GuidStr,
                                                     uint32 RecordType,
                                                     uint32 NumAttrib,
                                                     CSSM_DB_ATTRIBUTE_DATA_PTR PrimaryAttributes,
                                                     CSSM_DATA_PTR Credential)
{
    CSSM_HANDLE ResultsHandle = 0;
    CSSM_RETURN retValue;
    CSSM_DATA Manifest;
    
    if ((retValue = mdsutil_GetRecordByGuid(hDLDBCdsa,
                                            GuidStr,
                                            RecordType,
                                            NumAttrib,
                                            &ResultsHandle,
                                            PrimaryAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    MDSFuncs.DataAbortQuery( hDLDBCdsa, ResultsHandle );

    if (Credential)
    {
        Manifest = *PrimaryAttributes[2].Value;     
        if(  (Manifest.Length != 0 ) && (Manifest.Data != NULL) )
        {
            if ((Credential->Data = (uint8 *)MDSUTIL_calloc (1, Manifest.Length, NULL)) == NULL)
            {
                mdsutil_FreeRecord (NumAttrib, PrimaryAttributes);
                return (CSSM_ERRCODE_MEMORY_ERROR);
            }
            memcpy (Credential->Data, Manifest.Data, Manifest.Length);
            Credential->Length = Manifest.Length;
        }
        else
        {
            Credential->Length = 0;
            Credential->Data = NULL;
        }
    }

    mdsutil_FreeRecord (NumAttrib, PrimaryAttributes);

    return retValue;

}

CSSM_RETURN CSSMAPI mdsutil_ExtractCSPPrimaryManifest(char* GuidStr,
                                                      CSSM_DATA_PTR Credential)
{
    cssm_MDS_CDSADIR_CSP_PRIMARY Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA PrimaryAttributes[MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES];
    CSSM_RETURN retValue = CSSM_OK;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_CSP_PRIMARY ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &PrimaryAttributes, 0, MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = PrimaryAttributes;
    if ((retValue = csp_primary_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);

    retValue = mdsutil_GetPrimaryRecordManifest(GuidStr,
                                                MDS_CDSADIR_CSP_PRIMARY_RECORDTYPE,
                                                MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES,
                                                PrimaryAttributes,
                                                Credential);

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_ExtractACPrimaryManifest(char* GuidStr,
                                                     CSSM_DATA_PTR Credential)
{
    cssm_MDS_CDSADIR_AC_PRIMARY Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA PrimaryAttributes[MDS_CDSADIR_AC_PRIMARY_NUM_ATTRIBUTES];
    CSSM_RETURN retValue = CSSM_OK;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_AC_PRIMARY ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &PrimaryAttributes, 0, MDS_CDSADIR_AC_PRIMARY_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = PrimaryAttributes;
    if ((retValue = ac_primary_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);

    retValue = mdsutil_GetPrimaryRecordManifest(GuidStr,
                                                MDS_CDSADIR_AC_PRIMARY_RECORDTYPE,
                                                MDS_CDSADIR_AC_PRIMARY_NUM_ATTRIBUTES,
                                                PrimaryAttributes,
                                                Credential);

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_ExtractDLPrimaryManifest(char* GuidStr,
                                                     CSSM_DATA_PTR Credential)
{
    cssm_MDS_CDSADIR_DL_PRIMARY Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA PrimaryAttributes[MDS_CDSADIR_DL_PRIMARY_NUM_ATTRIBUTES];
    CSSM_RETURN retValue = CSSM_OK;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_DL_PRIMARY ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &PrimaryAttributes, 0, MDS_CDSADIR_DL_PRIMARY_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = PrimaryAttributes;
    if ((retValue = dl_primary_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);

    retValue = mdsutil_GetPrimaryRecordManifest(GuidStr,
                                                MDS_CDSADIR_DL_PRIMARY_RECORDTYPE,
                                                MDS_CDSADIR_DL_PRIMARY_NUM_ATTRIBUTES,
                                                PrimaryAttributes,
                                                Credential);
    
    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_ExtractCLPrimaryManifest(char* GuidStr,
                                                     CSSM_DATA_PTR Credential)
{
    cssm_MDS_CDSADIR_CL_PRIMARY Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA PrimaryAttributes[MDS_CDSADIR_CL_PRIMARY_NUM_ATTRIBUTES];
    CSSM_RETURN retValue = CSSM_OK;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_CL_PRIMARY ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &PrimaryAttributes, 0, MDS_CDSADIR_CL_PRIMARY_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = PrimaryAttributes;
    if ((retValue = cl_primary_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);

    retValue = mdsutil_GetPrimaryRecordManifest(GuidStr,
                                                MDS_CDSADIR_CL_PRIMARY_RECORDTYPE,
                                                MDS_CDSADIR_CL_PRIMARY_NUM_ATTRIBUTES,
                                                PrimaryAttributes,
                                                Credential);

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_ExtractTPPrimaryManifest(char* GuidStr,
                                                     CSSM_DATA_PTR Credential)
{
    cssm_MDS_CDSADIR_TP_PRIMARY Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA PrimaryAttributes[MDS_CDSADIR_TP_PRIMARY_NUM_ATTRIBUTES];
    CSSM_RETURN retValue = CSSM_OK;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_TP_PRIMARY ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &PrimaryAttributes, 0, MDS_CDSADIR_TP_PRIMARY_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = PrimaryAttributes;
    if ((retValue = tp_primary_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);

    retValue = mdsutil_GetPrimaryRecordManifest(GuidStr,
                                                MDS_CDSADIR_TP_PRIMARY_RECORDTYPE,
                                                MDS_CDSADIR_TP_PRIMARY_NUM_ATTRIBUTES,
                                                PrimaryAttributes,
                                                Credential);

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_ExtractEMMPrimaryManifest(char* GuidStr,
                                                      CSSM_DATA_PTR Credential)
{
    cssm_MDS_CDSADIR_EMM_PRIMARY Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA PrimaryAttributes[MDS_CDSADIR_EMM_PRIMARY_NUM_ATTRIBUTES];
    CSSM_RETURN retValue = CSSM_OK;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_EMM_PRIMARY ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &PrimaryAttributes, 0, MDS_CDSADIR_EMM_PRIMARY_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = PrimaryAttributes;
    if ((retValue = emm_primary_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);

    retValue = mdsutil_GetPrimaryRecordManifest(GuidStr,
                                                MDS_CDSADIR_EMM_PRIMARY_RECORDTYPE,
                                                MDS_CDSADIR_EMM_PRIMARY_NUM_ATTRIBUTES,
                                                PrimaryAttributes,
                                                Credential);

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_GetPrimaryManifest(const CSSM_GUID *ModuleGUID,
                                               CSSM_DATA *Credential,
                                               CSSM_SERVICE_TYPE ServiceType)
{
    CSSM_RETURN ret;
    char GuidStr[MDS_UTIL_MAX_PATH];

    if( ( ModuleGUID == NULL ) || ( Credential == NULL ) )
        return CSSMERR_CSSM_INVALID_POINTER;

    /*
     * Convert the Guid to a string.
     */
    mdsutil_GUIDToStr (ModuleGUID, GuidStr);
    
    switch(ServiceType)
    {
        case CSSM_SERVICE_TP:
            ret = mdsutil_ExtractTPPrimaryManifest(GuidStr, Credential);
            break;
        case CSSM_SERVICE_CSP:
            ret = mdsutil_ExtractCSPPrimaryManifest(GuidStr, Credential);
            break;
        case CSSM_SERVICE_CL:
            ret = mdsutil_ExtractCLPrimaryManifest(GuidStr, Credential);
            break;
        case CSSM_SERVICE_DL:
            ret = mdsutil_ExtractDLPrimaryManifest(GuidStr, Credential);
            break;
        case CSSM_SERVICE_AC:
            ret = mdsutil_ExtractACPrimaryManifest(GuidStr, Credential);
            break;
        default:
            ret = mdsutil_ExtractEMMPrimaryManifest(GuidStr, Credential);
            break;
    }

    return (ret);
}

CSSM_RETURN CSSMAPI mdsutil_GetRecordByGuidAndSSID(CSSM_DL_DB_HANDLE hDLDB,
                                                   char* GuidStr,
                                                   uint32 SubserviceID,
                                                   CSSM_DB_RECORDTYPE RecordType,
                                                   uint32 NumAttributes,
                                                   CSSM_HANDLE_PTR ResultsHandle,
                                                   CSSM_DB_ATTRIBUTE_DATA *OutputAttributeData)
{
    CSSM_RETURN retValue = CSSM_OK;
    CSSM_QUERY Query;
    CSSM_SELECTION_PREDICATE Predicate[2];

    memset (&Query, 0, sizeof (CSSM_QUERY));
    memset (&Predicate[0], 0, sizeof (CSSM_SELECTION_PREDICATE));
    memset (&Predicate[1], 0, sizeof (CSSM_SELECTION_PREDICATE));

    Query.RecordType = RecordType;
    Query.Conjunctive = CSSM_DB_AND;
    Query.NumSelectionPredicates = 2;

    Query.SelectionPredicate = Predicate;
    Query.QueryLimits.TimeLimit = CSSM_QUERY_TIMELIMIT_NONE;
    Query.QueryLimits.SizeLimit = CSSM_QUERY_SIZELIMIT_NONE;
    Query.QueryFlags = 0;

    Predicate[0].DbOperator = CSSM_DB_EQUAL;
    Predicate[0].Attribute.Info = s_MdsAttrInfo_ModuleId;
    Predicate[0].Attribute.Value = (CSSM_DATA *) MDSUTIL_malloc(sizeof (CSSM_DATA), NULL);
    if (Predicate[0].Attribute.Value == NULL)
        return CSSM_ERRCODE_MEMORY_ERROR;
    Predicate[0].Attribute.NumberOfValues = 1;
    Predicate[0].Attribute.Value->Data = (unsigned char *)GuidStr;
    Predicate[0].Attribute.Value->Length = (uint32)strlen (GuidStr) + 1;

    Predicate[1].DbOperator = CSSM_DB_EQUAL;
    Predicate[1].Attribute.Info = s_MdsAttrInfo_SSID;
    Predicate[1].Attribute.Value = (CSSM_DATA *) MDSUTIL_malloc(sizeof (CSSM_DATA), NULL);
    if (Predicate[1].Attribute.Value == NULL)
    {
        MDSUTIL_free(Predicate[0].Attribute.Value, NULL);
        return CSSM_ERRCODE_MEMORY_ERROR;
    }
    Predicate[1].Attribute.NumberOfValues = 1;
    Predicate[1].Attribute.Value->Data = (unsigned char *)&SubserviceID;
    Predicate[1].Attribute.Value->Length = (uint32)sizeof(uint32);

    retValue = mdsutil_GetRecord(hDLDB,
                                 Query,
                                 RecordType,
                                 NumAttributes,
                                 ResultsHandle,
                                 OutputAttributeData);

    MDSUTIL_free(Predicate[0].Attribute.Value, NULL);
    MDSUTIL_free(Predicate[1].Attribute.Value, NULL);

    return retValue;
}

CSSM_RETURN CSSMAPI mdsutil_GetRecordByGuidAndSSIDAndUSEE(CSSM_DL_DB_HANDLE hDLDB,
                                                          char* GuidStr,
                                                          uint32 SubserviceID,
                                                          CSSM_USEE_TAG USEERequest,
                                                          CSSM_DB_RECORDTYPE RecordType,
                                                          uint32 NumAttributes,
                                                          CSSM_HANDLE_PTR ResultsHandle,
                                                          CSSM_DB_ATTRIBUTE_DATA *OutputAttributeData)
{
    CSSM_RETURN retValue = CSSM_OK;
    CSSM_QUERY Query;
    CSSM_SELECTION_PREDICATE Predicate[3];

    memset (&Query, 0, sizeof (CSSM_QUERY));
    memset (&Predicate[0], 0, sizeof (CSSM_SELECTION_PREDICATE));
    memset (&Predicate[1], 0, sizeof (CSSM_SELECTION_PREDICATE));
    memset (&Predicate[2], 0, sizeof (CSSM_SELECTION_PREDICATE));

    Query.RecordType = RecordType;
    Query.Conjunctive = CSSM_DB_AND;
    Query.SelectionPredicate = Predicate;
    Query.NumSelectionPredicates = 3;
    Query.QueryLimits.TimeLimit = CSSM_QUERY_TIMELIMIT_NONE;
    Query.QueryLimits.SizeLimit = CSSM_QUERY_SIZELIMIT_NONE;
    Query.QueryFlags = 0;

    Predicate[0].DbOperator = CSSM_DB_EQUAL;
    Predicate[0].Attribute.Info = s_MdsAttrInfo_ModuleId;
    Predicate[0].Attribute.Value = (CSSM_DATA *) MDSUTIL_malloc(sizeof (CSSM_DATA), NULL);
    if (Predicate[0].Attribute.Value == NULL)
        return CSSM_ERRCODE_MEMORY_ERROR;
    Predicate[0].Attribute.NumberOfValues = 1;
    Predicate[0].Attribute.Value->Data = (unsigned char *)GuidStr;
    Predicate[0].Attribute.Value->Length = (uint32)strlen (GuidStr) + 1;
    
    Predicate[1].DbOperator = CSSM_DB_EQUAL;
    Predicate[1].Attribute.Info = s_MdsAttrInfo_UseeTag;
    Predicate[1].Attribute.Value = (CSSM_DATA *) MDSUTIL_malloc(sizeof (CSSM_DATA), NULL);
    if (Predicate[1].Attribute.Value == NULL)
    {
        MDSUTIL_free(Predicate[0].Attribute.Value, NULL);
        return CSSM_ERRCODE_MEMORY_ERROR;
    }
    Predicate[1].Attribute.NumberOfValues = 1;
    Predicate[1].Attribute.Value->Data = (unsigned char *)&USEERequest;
    Predicate[1].Attribute.Value->Length = (uint32)sizeof(uint32);

    Predicate[2].DbOperator = CSSM_DB_EQUAL;
    Predicate[2].Attribute.Info = s_MdsAttrInfo_SSID;
    Predicate[2].Attribute.Value = (CSSM_DATA *) MDSUTIL_malloc(sizeof (CSSM_DATA), NULL);
    if (Predicate[2].Attribute.Value == NULL)
    {
        MDSUTIL_free(Predicate[0].Attribute.Value, NULL);
        MDSUTIL_free(Predicate[1].Attribute.Value, NULL);
        return CSSM_ERRCODE_MEMORY_ERROR;
    }
    Predicate[2].Attribute.NumberOfValues = 1;
    Predicate[2].Attribute.Value->Data = (unsigned char *)&SubserviceID;
    Predicate[2].Attribute.Value->Length = (uint32)sizeof(uint32);

    retValue = mdsutil_GetRecord(hDLDB,
                                 Query,
                                 RecordType,
                                 NumAttributes,
                                 ResultsHandle,
                                 OutputAttributeData);

    MDSUTIL_free(Predicate[0].Attribute.Value, NULL);
    MDSUTIL_free(Predicate[1].Attribute.Value, NULL);
    MDSUTIL_free(Predicate[2].Attribute.Value, NULL);

    return retValue;
}

CSSM_RETURN CSSMAPI mdsutil_GetObjectRecord(CSSM_DL_DB_HANDLE hDLDB,
                                            char* GuidStr,
                                            cssm_MDS_OBJECT_RECORD_PTR Record,
                                            CSSM_DB_ATTRIBUTE_DATA_PTR ObjectAttributes)
{
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_HANDLE ResultsHandle = 0;
    CSSM_RETURN retValue = CSSM_OK;

    memset(&Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));

    Attributes.AttributeData = ObjectAttributes;
    if ((retValue = object_schema_GetAttributes (Record, &Attributes)) != CSSM_OK)
        return (retValue);
    
    if ((retValue = mdsutil_GetRecordByGuid(hDLDB,
                                            GuidStr,
                                            MDS_OBJECT_RECORDTYPE,
                                            MDS_OBJECT_NUM_ATTRIBUTES,
                                            &ResultsHandle,
                                            ObjectAttributes)) != CSSM_OK)
    {
        return (retValue);
    }
    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );

    Record->guid = *ObjectAttributes[0].Value;
    Record->Manifest = *ObjectAttributes[1].Value;
    Record->ModuleName = *ObjectAttributes[2].Value;
    Record->Path = *ObjectAttributes[3].Value;
    Record->ProductVersion = *ObjectAttributes[4].Value;

    return (retValue);
}


CSSM_RETURN CSSMAPI mdsutil_GetCommonRecord(CSSM_DL_DB_HANDLE hDLDB,
                                            char* GuidStr,
                                            cssm_MDS_CDSADIR_COMMON_PTR Record,
                                            CSSM_DB_ATTRIBUTE_DATA_PTR CommonAttributes)
{
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_HANDLE ResultsHandle = 0;
    CSSM_RETURN retValue = CSSM_OK;

    memset(&Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));

    Attributes.AttributeData = CommonAttributes;
    if ((retValue = common_schema_GetAttributes (Record, &Attributes)) != CSSM_OK)
        return (retValue);
    
    if ((retValue = mdsutil_GetRecordByGuid(hDLDB,
                                            GuidStr,
                                            MDS_CDSADIR_COMMON_RECORDTYPE,
                                            MDS_CDSADIR_COMMON_NUM_ATTRIBUTES,
                                            &ResultsHandle,
                                            CommonAttributes)) != CSSM_OK)
    {
        return (retValue);
    }
    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );

    Record->guid = *CommonAttributes[0].Value;
    Record->Manifest = *CommonAttributes[1].Value;
    Record->ModuleName = *CommonAttributes[2].Value;
    Record->Path = *CommonAttributes[3].Value;
    Record->Version = *CommonAttributes[4].Value;
    Record->Desc = *CommonAttributes[5].Value;
    Record->Dynamic = *CommonAttributes[6].Value;
    Record->MultiThreadFlag = *CommonAttributes[7].Value;
    Record->ServiceMask = *CommonAttributes[8].Value;
    
    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_GetCSSMRecord(CSSM_DL_DB_HANDLE hDLDB,
                                          char* GuidStr,
                                          cssm_MDS_CDSADIR_CSSM_PTR Record,
                                          CSSM_DB_ATTRIBUTE_DATA_PTR CssmAttributes)
{
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_HANDLE ResultsHandle = 0;
    CSSM_RETURN retValue = CSSM_OK;

    memset(&Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));

    Attributes.AttributeData = CssmAttributes;
    if ((retValue = cssm_schema_GetAttributes (Record, &Attributes)) != CSSM_OK)
        return (retValue);

    if ((retValue = mdsutil_GetRecordByGuid(hDLDB,
                                            GuidStr,
                                            MDS_CDSADIR_CSSM_RECORDTYPE,
                                            MDS_CDSADIR_CSSM_NUM_ATTRIBUTES,
                                            &ResultsHandle,
                                            CssmAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );

    Record->moduleid = *CssmAttributes[0].Value;
    Record->version = *CssmAttributes[1].Value;
    Record->vendor = *CssmAttributes[2].Value;
    Record->desc = *CssmAttributes[3].Value;
    Record->nativeservices = *CssmAttributes[4].Value;

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_mds_GetEmmRecord(CSSM_DL_DB_HANDLE hDLDB,
                                             char* GuidStr,
                                             cssm_MDS_CDSADIR_EMM_RECORD_PTR Record,
                                             CSSM_DB_ATTRIBUTE_DATA_PTR emmAttributes)
{
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_HANDLE ResultsHandle = 0;
    CSSM_RETURN retValue = CSSM_OK;

    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));

    Attributes.AttributeData = emmAttributes;
    if ((retValue = emm_schema_GetAttributes (Record, &Attributes)) != CSSM_OK)
        return (retValue);

    if ((retValue = mdsutil_GetRecordByGuid(hDLDB,
                                            GuidStr,
                                            MDS_CDSADIR_EMM_RECORDTYPE,
                                            MDS_CDSADIR_EMM_NUM_ATTRIBUTES,
                                            &ResultsHandle,
                                            emmAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );

    Record->moduleid = *emmAttributes[0].Value;
    Record->manifest = *emmAttributes[1].Value;
    Record->modulename = *emmAttributes[2].Value;
    Record->path = *emmAttributes[3].Value;
    Record->CDSAVersion = *emmAttributes[4].Value;
    Record->EMMSpecVersion = *emmAttributes[5].Value;
    Record->desc = *emmAttributes[6].Value;
    Record->policystmt = *emmAttributes[7].Value;
    Record->emmversion = *emmAttributes[8].Value;
    Record->emmvendor = *emmAttributes[9].Value;
    Record->emmtype = *emmAttributes[10].Value;

    return (retValue);
}



void CSSMAPI mdsutil_StringToVersion(unsigned char *VersionStr,
                                     CSSM_VERSION_PTR Version)
{
    uint32 Major=0;
    uint32 Minor=0;
    char* p;

    Major = (uint32)atoi ((const char *)VersionStr);
    p = strrchr ((const char *)VersionStr, '.');
    if(p != NULL)
    {
        p++;
        Minor = (uint32)atoi (p);
    }
    Version->Major = Major;
    Version->Minor = Minor;
}


CSSM_RETURN CSSMAPI mdsutil_mds_get_csp_primary(CSSM_DL_DB_HANDLE hDLDB,
                                                char* GuidStr,
                                                uint32 SubserviceID,
                                                MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    cssm_MDS_CDSADIR_CSP_PRIMARY Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA PrimaryAttributes[MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES];
    CSSM_HANDLE ResultsHandle = 0;
    MDSUTIL_SERVICE_INFO_PTR pServ;
    MDSUTIL_CSPSUBSERVICE_PTR pSubServ;
    CSSM_RETURN retValue = CSSM_OK;
    
/*  CSSM_USEE_TAG refUSEERequest = USEERequest;*/
    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_CSP_PRIMARY ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &PrimaryAttributes, 0, MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = PrimaryAttributes;
    if ((retValue = csp_primary_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);
    
    if ((retValue = mdsutil_GetRecordByGuidAndSSID(hDLDB,
                                                   GuidStr,
                                                   SubserviceID,
                                                   MDS_CDSADIR_CSP_PRIMARY_RECORDTYPE,
                                                   MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES,
                                                   &ResultsHandle,
                                                   PrimaryAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );

    Record.moduleid = *PrimaryAttributes[0].Value;
    Record.ssid = *PrimaryAttributes[1].Value;
    Record.manifest = *PrimaryAttributes[2].Value;
    Record.modulename = *PrimaryAttributes[3].Value;
    Record.addinversion = *PrimaryAttributes[4].Value;
    Record.addinvendor = *PrimaryAttributes[5].Value;
    Record.csptype = *PrimaryAttributes[6].Value;
    Record.cspflags = *PrimaryAttributes[7].Value;
    Record.cspcustomflags = *PrimaryAttributes[8].Value;
    Record.useetags = *PrimaryAttributes[9].Value;
    Record.sampletypes = *PrimaryAttributes[10].Value;
    Record.aclsubjecttypes = *PrimaryAttributes[11].Value;
    Record.authtags = *PrimaryAttributes[12].Value;

    pModuleInfo->NumberOfServices++;
    if ((pModuleInfo->ServiceList = (MDSUTIL_SERVICE_INFO_PTR)MDSUTIL_realloc (pModuleInfo->ServiceList,
                                        (uint32)(sizeof (MDSUTIL_SERVICE_INFO) * pModuleInfo->NumberOfServices), NULL)) == NULL) {
        mdsutil_FreeRecord (MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }

    mdsutil_StringToVersion (Record.addinversion.Data, &pModuleInfo->Version);
    strncpy (pModuleInfo->Vendor, (const char *)Record.addinvendor.Data, Record.addinvendor.Length);

    pServ = &pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1];
    strcpy (pServ->Description, pModuleInfo->Description);
    pServ->Type = CSSM_SERVICE_CSP;
    pServ->Flags = 0;
    pServ->NumberOfSubServices = 1;

    if ((pServ->SubServiceList = (MDSUTIL_SUBSERVICE_LIST_PTR)MDSUTIL_calloc (1, (uint32)sizeof (MDSUTIL_SUBSERVICE_LIST), NULL)) == NULL) {
        MDSUTIL_free (pModuleInfo->ServiceList, NULL);
        mdsutil_FreeRecord (MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }

    if (Record.ssid.Data)
        pServ->SubServiceList->SubServiceId = *(uint32 *)Record.ssid.Data;

    if ((pServ->SubServiceList->SubServiceInfo = (void*)MDSUTIL_calloc (1, (uint32)sizeof (MDSUTIL_CSPSUBSERVICE), NULL)) == NULL) {
        MDSUTIL_free (pServ->SubServiceList, NULL);
        MDSUTIL_free (pModuleInfo->ServiceList, NULL);
        mdsutil_FreeRecord (MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }

    pSubServ = (MDSUTIL_CSPSUBSERVICE_PTR)pServ->SubServiceList->SubServiceInfo;

    if (Record.ssid.Data)
        pSubServ->SubServiceId = *(uint32 *) Record.ssid.Data;
    if (Record.useetags.Data)
        pSubServ->USEETag = *(CSSM_USEE_TAG *) Record.useetags.Data;
    if (pModuleInfo->Description)
        strcpy (pSubServ->Description, pModuleInfo->Description);
    if (Record.cspflags.Data)
        pSubServ->CspFlags = *(CSSM_CSP_FLAGS *) Record.cspflags.Data;
    if (Record.cspcustomflags.Data)
        pSubServ->CspCustomFlags = *(uint32 *) Record.cspcustomflags.Data;
    if (Record.csptype.Data)
        pSubServ->CspType = *(CSSM_CSPTYPE *) Record.csptype.Data;

    mdsutil_FreeRecord (MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_mds_get_additional_csp_capabilities(CSSM_DL_DB_HANDLE hDLDB,
                                                                char* GuidStr, 
                                                                uint32 SubserviceID,
                                                                CSSM_USEE_TAG USEERequest,
                                                                MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    uint32 i;
    cssm_MDS_CDSADIR_CSP_CAPABILITY Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA CapAttributes[MDS_CDSADIR_CSP_CAPABILITY_NUM_ATTRIBUTES];
    CSSM_HANDLE ResultsHandle = 0;
    MDSUTIL_SOFTWARE_CSPSUBSERVICE_INFO_PTR pSubServ;
    CSSM_CONTEXT_PTR pCap;
    CSSM_CONTEXT_ATTRIBUTE_PTR attr = NULL;
    uint32 gpid;
    CSSM_CONTEXT_TYPE contextType=CSSM_ALGCLASS_NONE;
    CSSM_ALGORITHMS algType = CSSM_ALGID_NONE;
    CSSM_ATTRIBUTE_TYPE AttributeType = 0;
    uint32 attrLength = 0;
    uint32 numAttr = 0;
    uint32 attrIndex = 0;
    uint8 *attrPtr = 0;
    MDSUTIL_CSPSUBSERVICE_PTR pCspSubServ = NULL;
    CSSM_RETURN retValue = CSSM_OK;
    
    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_CSP_CAPABILITY ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &CapAttributes, 0, MDS_CDSADIR_CSP_CAPABILITY_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = CapAttributes;
    if ((retValue = csp_capability_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);

    pCspSubServ = (MDSUTIL_CSPSUBSERVICE_PTR)pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1].
                                    SubServiceList->SubServiceInfo;
    pCspSubServ->USEETag = USEERequest;

    pSubServ = &((MDSUTIL_CSPSUBSERVICE_PTR)(pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1].
                                    SubServiceList->SubServiceInfo))->SubServiceInfo.SoftwareCspSubService;
    
    retValue = CSSM_OK;
    for (;;)
    {
        if ((retValue = mdsutil_GetRecordByGuidAndSSIDAndUSEE(hDLDB,
                                                              GuidStr,
                                                              SubserviceID,
                                                              USEERequest,
                                                              MDS_CDSADIR_CSP_CAPABILITY_RECORDTYPE,
                                                              MDS_CDSADIR_CSP_CAPABILITY_NUM_ATTRIBUTES,
                                                              &ResultsHandle,
                                                              CapAttributes)) != CSSM_OK)
        {
            if (CSSMERR_DL_ENDOFDATA == retValue)
            {       
                retValue = CSSM_OK;
            }   
            MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
            return (retValue);
        }

        Record.moduleid = *CapAttributes[0].Value;
        Record.ssid = *CapAttributes[1].Value;
        Record.useetag = *CapAttributes[2].Value;
        Record.contexttype = *CapAttributes[3].Value;
        Record.algtype = *CapAttributes[4].Value;
        Record.groupid = *CapAttributes[5].Value;
        Record.attributetype = *CapAttributes[6].Value;
        Record.attributevalue = *CapAttributes[7].Value;
        Record.description = *CapAttributes[8].Value;
        
        if (Record.contexttype.Data)
            contextType = *(CSSM_CONTEXT_TYPE *)Record.contexttype.Data;
        if (Record.algtype.Data)
            algType = *(CSSM_ALGORITHMS *)Record.algtype.Data;

        gpid = *(uint32 *)Record.groupid.Data;

        if (gpid >= pSubServ->NumberOfCapabilities) 
        {
            uint32 lastid = pSubServ->NumberOfCapabilities;
            pSubServ->NumberOfCapabilities = gpid + 1;
            if ((pSubServ->CapabilityList = (CSSM_CONTEXT_PTR)MDSUTIL_realloc (pSubServ->CapabilityList,
                            (uint32)(sizeof (CSSM_CONTEXT) * pSubServ->NumberOfCapabilities), NULL)) == NULL)
            {
                MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
                if (pSubServ->CapabilityList)
                    MDSUTIL_free (pSubServ->CapabilityList, NULL);
                mdsutil_FreeRecord (MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES, CapAttributes);
                return (CSSM_ERRCODE_MEMORY_ERROR);
            }
                  
            
            for (i = lastid; i < pSubServ->NumberOfCapabilities; i++) {
                pCap = &pSubServ->CapabilityList[i];
                memset (pCap, 0, sizeof (CSSM_CONTEXT));

                pCap->ContextType = contextType;
                pCap->AlgorithmType = algType;
            }

        }
        pCap = &pSubServ->CapabilityList[gpid];
        
        if (Record.attributetype.Data)
            AttributeType = *(uint32 *)Record.attributetype.Data;
        switch (AttributeType) {
        case CSSM_ATTRIBUTE_KEY_LENGTH:
        case CSSM_ATTRIBUTE_PADDING:
        case CSSM_ATTRIBUTE_BLOCK_SIZE:
        case CSSM_ATTRIBUTE_OUTPUT_SIZE:
        case CSSM_ATTRIBUTE_ROUNDS:
        case CSSM_ATTRIBUTE_IV_SIZE:
        case CSSM_ATTRIBUTE_KEY_TYPE:
        case CSSM_ATTRIBUTE_MODE:
        case CSSM_ATTRIBUTE_EFFECTIVE_BITS:
        case CSSM_ATTRIBUTE_KEYUSAGE:
        case CSSM_ATTRIBUTE_KEYATTR:
        case CSSM_ATTRIBUTE_ALG_ID:
        case CSSM_ATTRIBUTE_ITERATION_COUNT:
            numAttr = (uint32)(Record.attributevalue.Length/sizeof (uint32));
            attrLength = (uint32)sizeof (uint32);
            break;
        case CSSM_ATTRIBUTE_KEY_LENGTH_RANGE:
        case CSSM_ATTRIBUTE_ROUNDS_RANGE:
            /* we have to replace the attrib*/
            numAttr = (uint32)(Record.attributevalue.Length/sizeof (CSSM_RANGE));
            attrLength = (uint32)sizeof (CSSM_RANGE);
            break;
        case CSSM_ATTRIBUTE_VERSION:
            numAttr = (uint32)(Record.attributevalue.Length/sizeof (CSSM_VERSION));
            attrLength = (uint32)sizeof (CSSM_VERSION);
            break;
        default:
            break;
        }
        if( AttributeType != CSSM_ATTRIBUTE_KEY_LENGTH_RANGE && AttributeType != CSSM_ATTRIBUTE_ROUNDS_RANGE)
        {
            attrIndex = pCap->NumberOfAttributes;
            if( pCap->NumberOfAttributes == 0 )
            {
                if (Record.description.Data != NULL)
                    pCap->NumberOfAttributes++;     
            }

            pCap->NumberOfAttributes += numAttr;
        }
        else
        {
            attrIndex = pCap->NumberOfAttributes;
            //pCap->NumberOfAttributes += numAttr;
            //if (Record.description.Data != NULL)
            //  pCap->NumberOfAttributes++;     

        }
        if(pCap->ContextAttributes == NULL)
        {
            if ((pCap->ContextAttributes = (CSSM_CONTEXT_ATTRIBUTE_PTR)MDSUTIL_calloc(
                            (uint32)(sizeof (CSSM_CONTEXT_ATTRIBUTE)*pCap->NumberOfAttributes), 1,NULL)) == NULL) {
                if (pCap->ContextAttributes)
                    MDSUTIL_free (pCap->ContextAttributes, NULL);
                MDSUTIL_free (pSubServ->CapabilityList, NULL);
                mdsutil_FreeRecord (MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES, CapAttributes);
                MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
                return (CSSM_ERRCODE_MEMORY_ERROR);
            }

        }
        else
        {
            if ((pCap->ContextAttributes = (CSSM_CONTEXT_ATTRIBUTE_PTR)MDSUTIL_realloc(pCap->ContextAttributes,
                            (uint32)(sizeof (CSSM_CONTEXT_ATTRIBUTE)*pCap->NumberOfAttributes), NULL)) == NULL) {
                if (pCap->ContextAttributes)
                    MDSUTIL_free (pCap->ContextAttributes, NULL);
                MDSUTIL_free (pSubServ->CapabilityList, NULL);
                mdsutil_FreeRecord (MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES, CapAttributes);
                MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
                return (CSSM_ERRCODE_MEMORY_ERROR);
            }
        }
        if( AttributeType != CSSM_ATTRIBUTE_KEY_LENGTH_RANGE && AttributeType != CSSM_ATTRIBUTE_ROUNDS_RANGE)
        {
            attr = &pCap->ContextAttributes[attrIndex];
            attr->AttributeType = AttributeType;
            attrPtr = Record.attributevalue.Data;
        }
        switch (AttributeType) {
        case CSSM_ATTRIBUTE_KEY_LENGTH:
        case CSSM_ATTRIBUTE_PADDING:
        case CSSM_ATTRIBUTE_BLOCK_SIZE:
        case CSSM_ATTRIBUTE_OUTPUT_SIZE:
        case CSSM_ATTRIBUTE_ROUNDS:
        case CSSM_ATTRIBUTE_IV_SIZE:
        case CSSM_ATTRIBUTE_KEY_TYPE:
        case CSSM_ATTRIBUTE_MODE:
        case CSSM_ATTRIBUTE_EFFECTIVE_BITS:
        case CSSM_ATTRIBUTE_KEYUSAGE:
        case CSSM_ATTRIBUTE_KEYATTR:
        case CSSM_ATTRIBUTE_ALG_ID:
        case CSSM_ATTRIBUTE_ITERATION_COUNT:
            for (i = 0; i < numAttr; i++) {
                attr[i].AttributeType = AttributeType;
                attr[i].AttributeLength = attrLength;
                attr[i].Attribute.Uint32 = *(uint32 *)attrPtr;      
                attrPtr += sizeof (uint32);
            }
            break;
        case CSSM_ATTRIBUTE_KEY_LENGTH_RANGE:
        case CSSM_ATTRIBUTE_ROUNDS_RANGE:
            attr = &pCap->ContextAttributes[0];
            attrPtr = Record.attributevalue.Data;

            for (i = 0; i < attrIndex; i++) {
                if(attr[i].AttributeType == AttributeType)
                    break;
            }
            if( i != attrIndex)
            {
                attr[i].AttributeLength = attrLength;
                if(attr[i].Attribute.Range == NULL  )
                {
                    if ((attr[i].Attribute.Range = (CSSM_RANGE_PTR)MDSUTIL_malloc ((uint32)sizeof (CSSM_RANGE), NULL)) == NULL) {
                        if (pCap->ContextAttributes)
                            MDSUTIL_free (pCap->ContextAttributes, NULL);
                        MDSUTIL_free (pSubServ->CapabilityList, NULL);
                        mdsutil_FreeRecord (MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES, CapAttributes);
                        MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
                        return (CSSM_ERRCODE_MEMORY_ERROR);
                    }
                }
                else
                {
                    if( !IsBadWritePtr ( attr[i].Attribute.Range , (uint32)sizeof( CSSM_RANGE)))
                        MDSUTIL_free(  attr[i].Attribute.Range , NULL);
                    if ((attr[i].Attribute.Range = (CSSM_RANGE_PTR)MDSUTIL_malloc ((uint32)sizeof (CSSM_RANGE), NULL)) == NULL) {
                        if (pCap->ContextAttributes)
                            MDSUTIL_free (pCap->ContextAttributes, NULL);
                        MDSUTIL_free (pSubServ->CapabilityList, NULL);
                        mdsutil_FreeRecord (MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES, CapAttributes);
                        MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
                        return (CSSM_ERRCODE_MEMORY_ERROR);
                    }

                }
                memcpy (attr[i].Attribute.Range, attrPtr, sizeof (CSSM_RANGE));                                                                         
                attrPtr += sizeof (CSSM_RANGE);
            }
            break;
        case CSSM_ATTRIBUTE_VERSION:
            for (i = 0; i < numAttr; i++) {
                attr[i].AttributeType = AttributeType;
                attr[i].AttributeLength = attrLength;
                if ((attr[i].Attribute.Version = (CSSM_VERSION_PTR)MDSUTIL_malloc ((uint32)sizeof (CSSM_VERSION), NULL)) == NULL) {
                    if (pCap->ContextAttributes)
                        MDSUTIL_free (pCap->ContextAttributes, NULL);
                    MDSUTIL_free (pSubServ->CapabilityList, NULL);
                    mdsutil_FreeRecord (MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES, CapAttributes);
                    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
                    return (CSSM_ERRCODE_MEMORY_ERROR);
                }
                memcpy (attr[i].Attribute.Version, attrPtr, sizeof (CSSM_VERSION));         
                attrPtr += sizeof (CSSM_VERSION);
            }
            break;
        default:
            break;
        }
        if( (pCap->NumberOfAttributes - numAttr) == 1 )
        {
            if (Record.description.Data != NULL) {
                attr[numAttr].AttributeType = (uint32)CSSM_ATTRIBUTE_DESCRIPTION;
                attr[numAttr].AttributeLength = Record.description.Length;
                if ((attr[numAttr].Attribute.String = (char *)MDSUTIL_malloc (Record.description.Length, NULL)) == NULL) {
                    if (pCap->ContextAttributes)
                        MDSUTIL_free (pCap->ContextAttributes, NULL);
                    MDSUTIL_free (pSubServ->CapabilityList, NULL);
                    mdsutil_FreeRecord (MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES, CapAttributes);
                    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
                    return (CSSM_ERRCODE_MEMORY_ERROR);
                }       

                memcpy (attr[numAttr].Attribute.String, (char *)Record.description.Data,
                                                                    Record.description.Length); 
            }
        }
        mdsutil_FreeRecord (MDS_CDSADIR_CSP_CAPABILITY_NUM_ATTRIBUTES, CapAttributes);
    }
}

CSSM_RETURN CSSMAPI mdsutil_mds_get_software_csp_capabilities(CSSM_DL_DB_HANDLE hDLDB,
                                                              char* GuidStr,
                                                              uint32 SubserviceID,
                                                              CSSM_USEE_TAG USEERequest,
                                                              MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    uint32 i;
    cssm_MDS_CDSADIR_CSP_CAPABILITY Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA CapAttributes[MDS_CDSADIR_CSP_CAPABILITY_NUM_ATTRIBUTES];
    CSSM_HANDLE ResultsHandle = 0;
    MDSUTIL_SOFTWARE_CSPSUBSERVICE_INFO_PTR pSubServ;
    CSSM_CONTEXT_PTR pCap;
    CSSM_CONTEXT_ATTRIBUTE_PTR attr;
    CSSM_CONTEXT_TYPE contextType=CSSM_ALGCLASS_NONE;
    CSSM_ALGORITHMS algType = CSSM_ALGID_NONE;
    CSSM_RETURN retValue = CSSM_OK;
    uint32 AttributeType = 0;
    uint32 attrLength = 0;
    uint32 numAttr = 0;
    uint32 attrIndex = 0;
    uint32 gpid;
    uint8 *attrPtr = 0;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_CSP_CAPABILITY ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &CapAttributes, 0, MDS_CDSADIR_CSP_CAPABILITY_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = CapAttributes;
    if ((retValue = csp_capability_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);

    pSubServ = &((MDSUTIL_CSPSUBSERVICE_PTR)(pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1].
                                    SubServiceList->SubServiceInfo))->SubServiceInfo.SoftwareCspSubService;

    retValue = CSSM_OK;
    for (;;)
    {
        if ((retValue = mdsutil_GetRecordByGuidAndSSIDAndUSEE(hDLDB,
                                                              GuidStr,
                                                              SubserviceID,
                                                              USEERequest,
                                                              MDS_CDSADIR_CSP_CAPABILITY_RECORDTYPE,
                                                              MDS_CDSADIR_CSP_CAPABILITY_NUM_ATTRIBUTES,
                                                              &ResultsHandle,
                                                              CapAttributes)) != CSSM_OK) 
        {
            if (CSSMERR_DL_ENDOFDATA == retValue)
            {
                retValue = CSSM_OK;
            }   
            MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
            return (retValue);      
        }
            
        Record.moduleid = *CapAttributes[0].Value;
        Record.ssid = *CapAttributes[1].Value;
        Record.useetag = *CapAttributes[2].Value;
        Record.contexttype = *CapAttributes[3].Value;
        Record.algtype = *CapAttributes[4].Value;
        Record.groupid = *CapAttributes[5].Value;
        Record.attributetype = *CapAttributes[6].Value;
        Record.attributevalue = *CapAttributes[7].Value;
        Record.description = *CapAttributes[8].Value;
        
        if (Record.contexttype.Data)
            contextType = *(CSSM_CONTEXT_TYPE *)Record.contexttype.Data;
        if (Record.algtype.Data)
            algType = *(CSSM_ALGORITHMS *)Record.algtype.Data;

        gpid = *(uint32 *)Record.groupid.Data;

        if (gpid >= pSubServ->NumberOfCapabilities) {
            uint32 lastid = pSubServ->NumberOfCapabilities;
            pSubServ->NumberOfCapabilities = gpid + 1;
            if ((pSubServ->CapabilityList = (CSSM_CONTEXT_PTR)MDSUTIL_realloc (pSubServ->CapabilityList,
                            (uint32)(sizeof (CSSM_CONTEXT) * pSubServ->NumberOfCapabilities), NULL)) == NULL) {
                if (pSubServ->CapabilityList)
                    MDSUTIL_free (pSubServ->CapabilityList, NULL);
                mdsutil_FreeRecord (MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES, CapAttributes);
                MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
                return (CSSM_ERRCODE_MEMORY_ERROR);
            }
            
            for (i = lastid; i < pSubServ->NumberOfCapabilities; i++) {
                pCap = &pSubServ->CapabilityList[i];
                memset (pCap, 0, sizeof (CSSM_CONTEXT));

                pCap->ContextType = contextType;
                pCap->AlgorithmType = algType;
            }
        }
        pCap = &pSubServ->CapabilityList[gpid];
        
        if (Record.attributetype.Data)
            AttributeType = *(uint32 *)Record.attributetype.Data;
        switch (AttributeType) {
        case CSSM_ATTRIBUTE_KEY_LENGTH:
        case CSSM_ATTRIBUTE_PADDING:
        case CSSM_ATTRIBUTE_BLOCK_SIZE:
        case CSSM_ATTRIBUTE_OUTPUT_SIZE:
        case CSSM_ATTRIBUTE_ROUNDS:
        case CSSM_ATTRIBUTE_IV_SIZE:
        case CSSM_ATTRIBUTE_KEY_TYPE:
        case CSSM_ATTRIBUTE_MODE:
        case CSSM_ATTRIBUTE_EFFECTIVE_BITS:
        case CSSM_ATTRIBUTE_KEYUSAGE:
        case CSSM_ATTRIBUTE_KEYATTR:
        case CSSM_ATTRIBUTE_ALG_ID:
        case CSSM_ATTRIBUTE_ITERATION_COUNT:
            numAttr = (uint32)(Record.attributevalue.Length/sizeof (uint32));
            attrLength = (uint32)sizeof (uint32);
            break;
        case CSSM_ATTRIBUTE_KEY_LENGTH_RANGE:
        case CSSM_ATTRIBUTE_ROUNDS_RANGE:
            numAttr = (uint32)(Record.attributevalue.Length/sizeof (CSSM_RANGE));
            attrLength = (uint32)sizeof (CSSM_RANGE);
            break;
        case CSSM_ATTRIBUTE_VERSION:
            numAttr = (uint32)(Record.attributevalue.Length/sizeof (CSSM_VERSION));
            attrLength = (uint32)sizeof (CSSM_VERSION);
            break;
        default:
            break;
        }
        attrIndex = pCap->NumberOfAttributes;
        if( pCap->NumberOfAttributes == 0 )
        {
            if (Record.description.Data != NULL)
            pCap->NumberOfAttributes++;     
        }


        pCap->NumberOfAttributes += numAttr;
        if ((pCap->ContextAttributes = (CSSM_CONTEXT_ATTRIBUTE_PTR)MDSUTIL_realloc(pCap->ContextAttributes,
                        (uint32)(sizeof (CSSM_CONTEXT_ATTRIBUTE)*pCap->NumberOfAttributes), NULL)) == NULL) {
            if (pCap->ContextAttributes)
                MDSUTIL_free (pCap->ContextAttributes, NULL);
            MDSUTIL_free (pSubServ->CapabilityList, NULL);
            mdsutil_FreeRecord (MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES, CapAttributes);
            MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
            return (CSSM_ERRCODE_MEMORY_ERROR);
        }

        attr = &pCap->ContextAttributes[attrIndex];
        attr->AttributeType = AttributeType;
        attrPtr = Record.attributevalue.Data;

        switch (AttributeType) {
        case CSSM_ATTRIBUTE_KEY_LENGTH:
        case CSSM_ATTRIBUTE_PADDING:
        case CSSM_ATTRIBUTE_BLOCK_SIZE:
        case CSSM_ATTRIBUTE_OUTPUT_SIZE:
        case CSSM_ATTRIBUTE_ROUNDS:
        case CSSM_ATTRIBUTE_IV_SIZE:
        case CSSM_ATTRIBUTE_KEY_TYPE:
        case CSSM_ATTRIBUTE_MODE:
        case CSSM_ATTRIBUTE_EFFECTIVE_BITS:
        case CSSM_ATTRIBUTE_KEYUSAGE:
        case CSSM_ATTRIBUTE_KEYATTR:
        case CSSM_ATTRIBUTE_ALG_ID:
        case CSSM_ATTRIBUTE_ITERATION_COUNT:
            for (i = 0; i < numAttr; i++) {
                attr[i].AttributeType = AttributeType;
                attr[i].AttributeLength = attrLength;
                attr[i].Attribute.Uint32 = *(uint32 *)attrPtr;      
                attrPtr += sizeof (uint32);
            }
            break;
        case CSSM_ATTRIBUTE_KEY_LENGTH_RANGE:
        case CSSM_ATTRIBUTE_ROUNDS_RANGE:
            for (i = 0; i < numAttr; i++) {
                attr[i].AttributeType = AttributeType;
                attr[i].AttributeLength = attrLength;
                if ((attr[i].Attribute.Range = (CSSM_RANGE_PTR)MDSUTIL_malloc ((uint32)sizeof (CSSM_RANGE), NULL)) == NULL) {
                    if (pCap->ContextAttributes)
                        MDSUTIL_free (pCap->ContextAttributes, NULL);
                    MDSUTIL_free (pSubServ->CapabilityList, NULL);
                    mdsutil_FreeRecord (MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES, CapAttributes);
                    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
                    return (CSSM_ERRCODE_MEMORY_ERROR);
                }
                memcpy (attr[i].Attribute.Range, attrPtr, sizeof (CSSM_RANGE));                                                                         
                attrPtr += sizeof (CSSM_RANGE);
            }
            break;
        case CSSM_ATTRIBUTE_VERSION:
            for (i = 0; i < numAttr; i++) {
                attr[i].AttributeType = AttributeType;
                attr[i].AttributeLength = attrLength;
                if ((attr[i].Attribute.Version = (CSSM_VERSION_PTR)MDSUTIL_malloc ((uint32)sizeof (CSSM_VERSION), NULL)) == NULL) {
                    if (pCap->ContextAttributes)
                        MDSUTIL_free (pCap->ContextAttributes, NULL);
                    MDSUTIL_free (pSubServ->CapabilityList, NULL);
                    mdsutil_FreeRecord (MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES, CapAttributes);
                    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
                    return (CSSM_ERRCODE_MEMORY_ERROR);
                }
                memcpy (attr[i].Attribute.Version, attrPtr, sizeof (CSSM_VERSION));         
                attrPtr += sizeof (CSSM_VERSION);
            }
            break;
        default:
            break;
        }
        if( (pCap->NumberOfAttributes - numAttr) == 1 )
        {
            if (Record.description.Data != NULL) {
                attr[numAttr].AttributeType = (uint32)CSSM_ATTRIBUTE_DESCRIPTION;
                attr[numAttr].AttributeLength = Record.description.Length;
                if ((attr[numAttr].Attribute.String = (char *)MDSUTIL_malloc (Record.description.Length, NULL)) == NULL) {
                    if (pCap->ContextAttributes)
                        MDSUTIL_free (pCap->ContextAttributes, NULL);
                    MDSUTIL_free (pSubServ->CapabilityList, NULL);
                    mdsutil_FreeRecord (MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES, CapAttributes);
                    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
                    return (CSSM_ERRCODE_MEMORY_ERROR);
                }       

                memcpy (attr[numAttr].Attribute.String, (char *)Record.description.Data,
                                                                    Record.description.Length); 
            }
        }
        mdsutil_FreeRecord (MDS_CDSADIR_CSP_CAPABILITY_NUM_ATTRIBUTES, CapAttributes);
    }
}

CSSM_RETURN CSSMAPI mdsutil_mds_get_hardware_csp_capabilities(CSSM_DL_DB_HANDLE hDLDB,
                                                              char* GuidStr,
                                                              uint32 SubserviceID,
                                                              CSSM_USEE_TAG USEERequest,
                                                              MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{

    CSSM_RETURN retValue = CSSM_OK;

    if( USEERequest == CSSM_USEE_NONE )
    {
        if ((retValue = mdsutil_mds_get_software_csp_capabilities(hDLDB,
                                                                  GuidStr,
                                                                  SubserviceID,
                                                                  USEERequest,
                                                                  pModuleInfo))
                                                                  != CSSM_OK)
        {
            return (retValue);
        }
    }
    else
    {
        if ((retValue = mdsutil_mds_get_software_csp_capabilities(hDLDB,
                                                                  GuidStr,
                                                                  SubserviceID,
                                                                  CSSM_USEE_NONE,
                                                                  pModuleInfo))
                                                                  != CSSM_OK)
        {
            return (retValue);
        }

        if ((retValue = mdsutil_mds_get_additional_csp_capabilities(hDLDB,
                                                                    GuidStr,
                                                                    SubserviceID,
                                                                    USEERequest,
                                                                    pModuleInfo))
                                                                    != CSSM_OK)
        {
            return (retValue);
        }

    }

    return retValue;
}

CSSM_RETURN mdsutil_mds_get_csp_SC_info(CSSM_DL_DB_HANDLE hDLDB,
                                        char* GuidStr,
                                        uint32 SubserviceID,
                                        MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    cssm_MDS_CDSADIR_CSP_SC_INFO Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA SCAttributes[MDS_CDSADIR_CSP_SC_INFO_NUM_ATTRIBUTES];
    CSSM_HANDLE ResultsHandle = 0;
    MDSUTIL_HARDWARE_CSPSUBSERVICE_INFO_PTR pSubServ;
    CSSM_RETURN retValue = CSSM_OK;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_CSP_SC_INFO ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &SCAttributes, 0, MDS_CDSADIR_CSP_SC_INFO_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = SCAttributes;
    if ((retValue = csp_SC_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);

    if ((retValue = mdsutil_GetRecordByGuidAndSSID(hDLDB,
                                                   GuidStr,
                                                   SubserviceID,
                                                   MDS_CDSADIR_CSP_SC_INFO_RECORDTYPE,
                                                   MDS_CDSADIR_CSP_SC_INFO_NUM_ATTRIBUTES,
                                                   &ResultsHandle,
                                                   SCAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );

    Record.moduleid = *SCAttributes[0].Value;
    Record.ssid = *SCAttributes[1].Value;
    Record.scdescription = *SCAttributes[2].Value;
    Record.scvendor = *SCAttributes[3].Value;
    Record.scversion = *SCAttributes[4].Value;
    Record.scfirmwareversion = *SCAttributes[5].Value;
    Record.scflags = *SCAttributes[6].Value;
    Record.sccustomflags = *SCAttributes[7].Value;
    Record.scserialnumber = *SCAttributes[8].Value;


    pSubServ = &((MDSUTIL_CSPSUBSERVICE_PTR)(pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1].
        SubServiceList->SubServiceInfo))->SubServiceInfo.HardwareCspSubService;

    strncpy (pSubServ->TokenDescription, (const char *)Record.scdescription.Data, 
                                        Record.scdescription.Length);
    strncpy (pSubServ->TokenVendor, (const char *)Record.scvendor.Data,
                                        Record.scvendor.Length);
    strncpy (pSubServ->TokenSerialNumber, (const char *)Record.scserialnumber.Data,
                                        Record.scserialnumber.Length);
    mdsutil_StringToVersion (Record.scversion.Data, &pSubServ->TokenHardwareVersion);
    mdsutil_StringToVersion (Record.scfirmwareversion.Data, &pSubServ->TokenFirmwareVersion);
    if (Record.scflags.Data)
        pSubServ->TokenFlags = *(uint32 *)Record.scflags.Data;
    if (Record.sccustomflags.Data)
        pSubServ->TokenCustomFlags = *(uint32 *)Record.sccustomflags.Data;

    mdsutil_FreeRecord (MDS_CDSADIR_CSP_SC_INFO_NUM_ATTRIBUTES, SCAttributes);

    return (retValue);
}

CSSM_RETURN CSSMAPI 
mdsutil_mds_get_csp_reader_info(cssm_MDS_CDSADIR_CSP_ENCAPSULATED_PRODUCT_PTR Record,
                         MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    MDSUTIL_HARDWARE_CSPSUBSERVICE_INFO_PTR pSubServ;

    pSubServ = &((MDSUTIL_CSPSUBSERVICE_PTR)(pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1].
        SubServiceList->SubServiceInfo))->SubServiceInfo.HardwareCspSubService;

    strncpy (pSubServ->ReaderDescription, (const char *)Record->readerdescription.Data, 
                            Record->readerdescription.Length);
    strncpy (pSubServ->ReaderVendor, (const char *)Record->readervendor.Data,
                            Record->readervendor.Length);
    strncpy (pSubServ->ReaderSerialNumber, (const char *)Record->readerserialnumber.Data,
                                        Record->readerserialnumber.Length);
    mdsutil_StringToVersion (Record->readerversion.Data, &pSubServ->ReaderHardwareVersion);
    mdsutil_StringToVersion (Record->readerfirmwareversion.Data, &pSubServ->ReaderFirmwareVersion);
    if (Record->readerflags.Data)
        pSubServ->ReaderFlags = *(uint32 *)Record->readerflags.Data;
    if (Record->readercustomflags.Data)
        pSubServ->ReaderCustomFlags = *(uint32 *)Record->readercustomflags.Data;

    return (CSSM_OK);
}



CSSM_RETURN CSSMAPI mdsutil_mds_get_csp_encapsulated_product(CSSM_DL_DB_HANDLE hDLDB,
                                                             char* GuidStr,
                                                             CSSM_CSPTYPE cspType,
                                                             uint32 SubserviceID,
                                                             MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    cssm_MDS_CDSADIR_CSP_ENCAPSULATED_PRODUCT Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA EncapAttributes[MDS_CDSADIR_CSP_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES];
    CSSM_HANDLE ResultsHandle = 0;
    MDSUTIL_CSP_WRAPPEDPRODUCTINFO_PTR pWrap;
    CSSM_RETURN retValue = CSSM_OK;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_CSP_ENCAPSULATED_PRODUCT ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &EncapAttributes, 0, MDS_CDSADIR_CSP_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = EncapAttributes;
    if ((retValue = csp_encapsulated_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);
    
    if ((retValue = mdsutil_GetRecordByGuidAndSSID(hDLDB,
                                                   GuidStr,
                                                   SubserviceID,
                                                   MDS_CDSADIR_CSP_ENCAPSULATED_PRODUCT_RECORDTYPE,
                                                   MDS_CDSADIR_CSP_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES,
                                                   &ResultsHandle,
                                                   EncapAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );

    Record.moduleid = *EncapAttributes[0].Value;
    Record.ssid = *EncapAttributes[1].Value;
    Record.productdesc = *EncapAttributes[2].Value;
    Record.productvendor = *EncapAttributes[3].Value;
    Record.productversion = *EncapAttributes[4].Value;
    Record.productflags = *EncapAttributes[5].Value;
    Record.customflags = *EncapAttributes[6].Value;
    Record.standarddesc = *EncapAttributes[7].Value;
    Record.standardversion = *EncapAttributes[8].Value;
    Record.readerdescription = *EncapAttributes[9].Value;
    Record.readervendor = *EncapAttributes[10].Value;
    Record.readerversion = *EncapAttributes[11].Value;
    Record.readerfirmwareversion = *EncapAttributes[12].Value;
    Record.readerflags = *EncapAttributes[13].Value;
    Record.readercustomflags = *EncapAttributes[14].Value;
    Record.readerserialnumber = *EncapAttributes[15].Value;

    pWrap = &(((MDSUTIL_CSPSUBSERVICE_PTR)(pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1].
                                            SubServiceList->SubServiceInfo))->WrappedProduct);
    mdsutil_StringToVersion (Record.standardversion.Data, &pWrap->StandardVersion);
    strncpy (pWrap->StandardDescription, (const char *)Record.standarddesc.Data, Record.standarddesc.Length);
    mdsutil_StringToVersion (Record.productversion.Data, &pWrap->ProductVersion);
    strncpy (pWrap->ProductDescription, (const char *)Record.productdesc.Data, Record.productdesc.Length);
    strncpy (pWrap->ProductVendor, (const char *)Record.productvendor.Data, Record.productvendor.Length);
    if (Record.productflags.Data)
        pWrap->ProductFlags = *(uint32 *)Record.productflags.Data;
    if (Record.customflags.Data)
        pWrap->ProductCustomFlags = *(uint32 *)Record.customflags.Data;

    if (cspType == CSSM_CSP_HARDWARE || cspType == CSSM_CSP_HYBRID) {
        if ((retValue = mdsutil_mds_get_csp_reader_info(&Record, pModuleInfo)) != CSSM_OK) {
            mdsutil_FreeRecord (MDS_CDSADIR_CSP_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES, EncapAttributes);
            return (retValue);
        }
        mdsutil_mds_get_csp_SC_info(hDLDB, GuidStr, SubserviceID, pModuleInfo);
    }
    
    mdsutil_FreeRecord (MDS_CDSADIR_CSP_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES, EncapAttributes);

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_mds_get_tp_primary(CSSM_DL_DB_HANDLE hDLDB,
                                               char* GuidStr,
                                               uint32 SubserviceID,
                                               MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    cssm_MDS_CDSADIR_TP_PRIMARY Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA PrimaryAttributes[MDS_CDSADIR_TP_PRIMARY_NUM_ATTRIBUTES];
    MDSUTIL_SERVICE_INFO_PTR pServ;
    CSSM_RETURN retValue = CSSM_OK;
    uint32 type = 0;
    MDSUTIL_TPSUBSERVICE_PTR pSubServ;
    CSSM_HANDLE ResultsHandle = 0;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_TP_PRIMARY ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &PrimaryAttributes, 0, MDS_CDSADIR_TP_PRIMARY_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = PrimaryAttributes;
    if ((retValue = tp_primary_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);
    
    if ((retValue = mdsutil_GetRecordByGuidAndSSID(hDLDB,
                                                   GuidStr,
                                                   SubserviceID,
                                                   MDS_CDSADIR_TP_PRIMARY_RECORDTYPE,
                                                   MDS_CDSADIR_TP_PRIMARY_NUM_ATTRIBUTES,
                                                   &ResultsHandle,
                                                   PrimaryAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );

    Record.moduleid = *PrimaryAttributes[0].Value;
    Record.ssid = *PrimaryAttributes[1].Value;
    Record.manifest = *PrimaryAttributes[2].Value;
    Record.modulename = *PrimaryAttributes[3].Value;
    Record.addinversion = *PrimaryAttributes[4].Value;
    Record.addinvendor = *PrimaryAttributes[5].Value;
    Record.certtypeformat = *PrimaryAttributes[6].Value;
    Record.sampletypes = *PrimaryAttributes[7].Value;
    Record.aclsubjecttypes = *PrimaryAttributes[8].Value;
    Record.authtags = *PrimaryAttributes[9].Value;

    pModuleInfo->NumberOfServices++;
    if ((pModuleInfo->ServiceList = (MDSUTIL_SERVICE_INFO_PTR)MDSUTIL_realloc (pModuleInfo->ServiceList,
                                        (uint32)(sizeof (MDSUTIL_SERVICE_INFO) * pModuleInfo->NumberOfServices), NULL)) == NULL) {
        mdsutil_FreeRecord (MDS_CDSADIR_TP_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }

    mdsutil_StringToVersion (Record.addinversion.Data, &pModuleInfo->Version);
    strncpy (pModuleInfo->Vendor, (const char *)Record.addinvendor.Data, Record.addinvendor.Length);

    pServ = &pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1];
    
    strcpy (pServ->Description, pModuleInfo->Description);
    pServ->Type = CSSM_SERVICE_TP;
    pServ->Flags = 0;
    pServ->NumberOfSubServices = 1;

    if ((pServ->SubServiceList = (MDSUTIL_SUBSERVICE_LIST_PTR)MDSUTIL_calloc (1, (uint32)sizeof (MDSUTIL_SUBSERVICE_LIST), NULL)) == NULL) {
        MDSUTIL_free (pModuleInfo->ServiceList, NULL);
        mdsutil_FreeRecord (MDS_CDSADIR_TP_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }

    if (Record.ssid.Data)
        pServ->SubServiceList->SubServiceId = *(uint32 *)Record.ssid.Data;
    
    if ((pServ->SubServiceList->SubServiceInfo = (void*)MDSUTIL_calloc (1, (uint32)sizeof (MDSUTIL_TPSUBSERVICE), NULL)) == NULL) {
        MDSUTIL_free (pServ->SubServiceList, NULL);
        MDSUTIL_free (pModuleInfo->ServiceList, NULL);
        mdsutil_FreeRecord (MDS_CDSADIR_TP_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }
    pSubServ = (MDSUTIL_TPSUBSERVICE_PTR)pServ->SubServiceList->SubServiceInfo;

    if (Record.ssid.Data)
        pSubServ->SubServiceId = *(uint32 *)Record.ssid.Data;;
    strcpy (pSubServ->Description, pModuleInfo->Description);
    if (Record.certtypeformat.Data)
        type = *(uint32 *)Record.certtypeformat.Data;
    pSubServ->CertType = (CSSM_CERT_TYPE)(type >> 16);
    pSubServ->CertEncoding = (CSSM_CERT_ENCODING)(type & 0x0000FFFF);

    mdsutil_FreeRecord (MDS_CDSADIR_TP_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
    
    return (CSSM_OK);
}

CSSM_RETURN CSSMAPI mdsutil_mds_get_tp_policy_oids(CSSM_DL_DB_HANDLE hDLDB,
                                                   char* GuidStr,
                                                   uint32 SubserviceID,
                                                   MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    cssm_MDS_CDSADIR_TP_OIDS Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA OidAttributes[MDS_CDSADIR_TP_OIDS_NUM_ATTRIBUTES];
    CSSM_HANDLE ResultsHandle = 0;
    MDSUTIL_TPSUBSERVICE_PTR pSubServ;
    CSSM_RETURN retValue = CSSM_OK;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_TP_OIDS ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &OidAttributes, 0, MDS_CDSADIR_TP_OIDS_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = OidAttributes;
    if ((retValue = tp_policy_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);
    
    pSubServ = pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1].SubServiceList->SubServiceInfo;
    
    retValue = CSSM_OK;
    for (;;)
    {
        if ((retValue = mdsutil_GetRecordByGuidAndSSID(hDLDB,
                                                       GuidStr,
                                                       SubserviceID,
                                                       MDS_CDSADIR_TP_OIDS_RECORDTYPE,
                                                       MDS_CDSADIR_TP_OIDS_NUM_ATTRIBUTES,
                                                       &ResultsHandle,
                                                       OidAttributes)) != CSSM_OK) 
        {
            if (CSSMERR_DL_ENDOFDATA == retValue)
            {           
                retValue = CSSM_OK;
            }   
            MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
            return (retValue);
        }
        
        Record.moduleid = *OidAttributes[0].Value;
        Record.ssid = *OidAttributes[1].Value;
        Record.oid = *OidAttributes[2].Value;   
        Record.value = *OidAttributes[3].Value;
    
        pSubServ->NumberOfPolicyIdentifiers++;
        if ((pSubServ->PolicyIdentifiers = (CSSM_FIELD_PTR)MDSUTIL_realloc (pSubServ->PolicyIdentifiers,
                    (uint32)(sizeof (CSSM_FIELD) * pSubServ->NumberOfPolicyIdentifiers), NULL)) == NULL)
        {
            MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
            mdsutil_FreeRecord (MDS_CDSADIR_TP_OIDS_NUM_ATTRIBUTES, OidAttributes);
            return (CSSM_ERRCODE_MEMORY_ERROR);
        }

        pSubServ->PolicyIdentifiers[pSubServ->NumberOfPolicyIdentifiers-1].FieldOid = Record.oid;
        pSubServ->PolicyIdentifiers[pSubServ->NumberOfPolicyIdentifiers-1].FieldValue = Record.value;

        if (OidAttributes[0].Value->Data)
            MDSUTIL_free (OidAttributes[0].Value->Data, NULL);
        if (OidAttributes[1].Value->Data)
            MDSUTIL_free (OidAttributes[1].Value->Data, NULL);
    }
}

CSSM_RETURN CSSMAPI mdsutil_mds_get_tp_encapsulated_product(CSSM_DL_DB_HANDLE hDLDB,
                                                            char* GuidStr,
                                                            uint32 SubserviceID,
                                                            MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    cssm_MDS_CDSADIR_TP_ENCAPSULATED_PRODUCT Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA EncapAttributes[MDS_CDSADIR_TP_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES];
    CSSM_HANDLE ResultsHandle = 0;
    MDSUTIL_TP_WRAPPEDPRODUCTINFO_PTR pWrap;
    CSSM_RETURN retValue = CSSM_OK;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_TP_ENCAPSULATED_PRODUCT ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &EncapAttributes, 0, MDS_CDSADIR_TP_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = EncapAttributes;
    if ((retValue = tp_encapsulated_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);
    
    if ((retValue = mdsutil_GetRecordByGuidAndSSID(hDLDB,
                                                   GuidStr,
                                                   SubserviceID,
                                                   MDS_CDSADIR_TP_ENCAPSULATED_PRODUCT_RECORDTYPE,
                                                   MDS_CDSADIR_TP_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES,
                                                   &ResultsHandle,
                                                   EncapAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );

    Record.moduleid = *EncapAttributes[0].Value;
    Record.ssid = *EncapAttributes[1].Value;
    Record.productdesc = *EncapAttributes[2].Value;
    Record.productvendor = *EncapAttributes[3].Value;
    Record.productversion = *EncapAttributes[4].Value;
    Record.productflags = *EncapAttributes[5].Value;
    Record.authorityrequesttype = *EncapAttributes[6].Value;
    Record.standarddesc = *EncapAttributes[7].Value;
    Record.standardversion = *EncapAttributes[8].Value;
    Record.protocoldesc = *EncapAttributes[9].Value;
    Record.protocolflags = *EncapAttributes[10].Value;
    Record.certclassname = *EncapAttributes[11].Value;
    Record.rootcertificate = *EncapAttributes[12].Value;
    Record.rootcerttypeformat = *EncapAttributes[13].Value;

        
    pWrap = &(((MDSUTIL_TPSUBSERVICE_PTR)(pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1].
                                            SubServiceList->SubServiceInfo))->WrappedProduct);
    mdsutil_StringToVersion (Record.standardversion.Data, &pWrap->StandardVersion);
    strncpy (pWrap->StandardDescription, (const char *)Record.standarddesc.Data, Record.standarddesc.Length);
    mdsutil_StringToVersion (Record.productversion.Data, &pWrap->ProductVersion);
    strncpy (pWrap->ProductDescription, (const char *)Record.productdesc.Data, Record.productdesc.Length);      
    strncpy (pWrap->ProductVendor, (const char *)Record.productvendor.Data, Record.productvendor.Length);

    mdsutil_FreeRecord (MDS_CDSADIR_TP_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES, EncapAttributes);
    
    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_mds_get_ac_primary(CSSM_DL_DB_HANDLE hDLDB,
                                               char* GuidStr,
                                               uint32 SubserviceID,
                                               MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    cssm_MDS_CDSADIR_AC_PRIMARY Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA PrimaryAttributes[MDS_CDSADIR_AC_PRIMARY_NUM_ATTRIBUTES];
    CSSM_HANDLE ResultsHandle = 0;
    CSSM_RETURN retValue = CSSM_OK;
    MDSUTIL_SERVICE_INFO_PTR pServ;
    MDSUTIL_ACSUBSERVICE_PTR pSubServ;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_AC_PRIMARY ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &PrimaryAttributes, 0, MDS_CDSADIR_AC_PRIMARY_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = PrimaryAttributes;
    if ((retValue = ac_primary_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);
    
    if((retValue = mdsutil_GetRecordByGuidAndSSID(hDLDB,
                                                  GuidStr,
                                                  SubserviceID,
                                                  MDS_CDSADIR_AC_PRIMARY_RECORDTYPE,
                                                  MDS_CDSADIR_AC_PRIMARY_NUM_ATTRIBUTES,
                                                  &ResultsHandle,
                                                  PrimaryAttributes)) != CSSM_OK)
        return (retValue);
    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
    
    Record.moduleid = *PrimaryAttributes[0].Value;
    Record.ssid = *PrimaryAttributes[1].Value;
    Record.manifest = *PrimaryAttributes[2].Value;
    Record.modulename = *PrimaryAttributes[3].Value;
    Record.addinversion = *PrimaryAttributes[4].Value;
    Record.addinvendor = *PrimaryAttributes[5].Value;

    pModuleInfo->NumberOfServices++;
    if ((pModuleInfo->ServiceList = (MDSUTIL_SERVICE_INFO_PTR)MDSUTIL_realloc (pModuleInfo->ServiceList,
                                (uint32)(sizeof (MDSUTIL_SERVICE_INFO) * pModuleInfo->NumberOfServices), NULL)) == NULL) {
        mdsutil_FreeRecord (MDS_CDSADIR_AC_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }

    mdsutil_StringToVersion (Record.addinversion.Data, &pModuleInfo->Version);
    strncpy (pModuleInfo->Vendor, (const char *)Record.addinvendor.Data, Record.addinvendor.Length);

    pServ = &pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1];
    strcpy (pServ->Description, pModuleInfo->Description);
    pServ->Type = CSSM_SERVICE_AC;
    pServ->Flags = 0;
    pServ->NumberOfSubServices = 1;

    if ((pServ->SubServiceList = (MDSUTIL_SUBSERVICE_LIST_PTR)MDSUTIL_calloc (1, (uint32)sizeof (MDSUTIL_SUBSERVICE_LIST), NULL)) == NULL)
    {
        MDSUTIL_free (pModuleInfo->ServiceList, NULL);
        mdsutil_FreeRecord (MDS_CDSADIR_AC_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }

    if (Record.ssid.Data)
        pServ->SubServiceList->SubServiceId = *(uint32 *)Record.ssid.Data;
    
    if ((pServ->SubServiceList->SubServiceInfo = (void*)MDSUTIL_calloc (1, (uint32)sizeof (MDSUTIL_ACSUBSERVICE), NULL)) == NULL)
    {
        MDSUTIL_free (pServ->SubServiceList, NULL);
        MDSUTIL_free (pModuleInfo->ServiceList, NULL);
        mdsutil_FreeRecord(MDS_CDSADIR_AC_PRIMARY_NUM_ATTRIBUTES,
                           PrimaryAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }
    pSubServ = (MDSUTIL_ACSUBSERVICE_PTR)pServ->SubServiceList->SubServiceInfo;
    
    if (Record.ssid.Data)
        pSubServ->SubServiceId = *(uint32 *)Record.ssid.Data;;
    strcpy (pSubServ->Description, pModuleInfo->Description);

    mdsutil_FreeRecord(MDS_CDSADIR_AC_PRIMARY_NUM_ATTRIBUTES,
                       PrimaryAttributes);
    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_mds_get_cl_primary(CSSM_DL_DB_HANDLE hDLDB,
                                               char* GuidStr,
                                               uint32 SubserviceID,
                                               MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    cssm_MDS_CDSADIR_CL_PRIMARY Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA PrimaryAttributes[MDS_CDSADIR_CL_PRIMARY_NUM_ATTRIBUTES];
    CSSM_HANDLE ResultsHandle = 0;
    CSSM_RETURN retValue = CSSM_OK;
    MDSUTIL_SERVICE_INFO_PTR pServ;
    MDSUTIL_CLSUBSERVICE_PTR pSubServ;
    uint32 type = 0;;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_CL_PRIMARY ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &PrimaryAttributes, 0, MDS_CDSADIR_CL_PRIMARY_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = PrimaryAttributes;
    if ((retValue = cl_primary_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);
    
    if ((retValue = mdsutil_GetRecordByGuidAndSSID(hDLDB,
                                                   GuidStr,
                                                   SubserviceID,
                                                   MDS_CDSADIR_CL_PRIMARY_RECORDTYPE,
                                                   MDS_CDSADIR_CL_PRIMARY_NUM_ATTRIBUTES,
                                                   &ResultsHandle,
                                                   PrimaryAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );

    Record.moduleid = *PrimaryAttributes[0].Value;
    Record.ssid = *PrimaryAttributes[1].Value;
    Record.manifest = *PrimaryAttributes[2].Value;
    Record.modulename = *PrimaryAttributes[3].Value;
    Record.addinversion = *PrimaryAttributes[4].Value;
    Record.addinvendor = *PrimaryAttributes[5].Value;
    Record.certtypeformat = *PrimaryAttributes[6].Value;
    Record.crltypeformat = *PrimaryAttributes[7].Value;
    Record.certfieldnames = *PrimaryAttributes[8].Value;
    Record.bundletypeformat = *PrimaryAttributes[9].Value;
    Record.xlationtypeformat = *PrimaryAttributes[10].Value;
    Record.defaulttemplatetype = *PrimaryAttributes[11].Value;
    Record.templatefieldnames = *PrimaryAttributes[12].Value;

    pModuleInfo->NumberOfServices++;
    if ((pModuleInfo->ServiceList = (MDSUTIL_SERVICE_INFO_PTR)MDSUTIL_realloc (pModuleInfo->ServiceList,
                                (uint32)(sizeof (MDSUTIL_SERVICE_INFO) * pModuleInfo->NumberOfServices), NULL)) == NULL) {
        mdsutil_FreeRecord (MDS_CDSADIR_CL_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }

    mdsutil_StringToVersion (Record.addinversion.Data, &pModuleInfo->Version);
    strncpy (pModuleInfo->Vendor, (const char *)Record.addinvendor.Data, Record.addinvendor.Length);

    pServ = &pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1];
    strcpy (pServ->Description, pModuleInfo->Description);
    pServ->Type = CSSM_SERVICE_CL;
    pServ->Flags = 0;
    pServ->NumberOfSubServices = 1;

    if ((pServ->SubServiceList = (MDSUTIL_SUBSERVICE_LIST_PTR)MDSUTIL_calloc (1, (uint32)sizeof (MDSUTIL_SUBSERVICE_LIST), NULL)) == NULL) {
        MDSUTIL_free (pModuleInfo->ServiceList, NULL);
        mdsutil_FreeRecord (MDS_CDSADIR_CL_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }

    if (Record.ssid.Data)
        pServ->SubServiceList->SubServiceId = *(uint32 *)Record.ssid.Data;
    
    if ((pServ->SubServiceList->SubServiceInfo = (void*)MDSUTIL_calloc (1, (uint32)sizeof (MDSUTIL_CLSUBSERVICE), NULL)) == NULL) {
        MDSUTIL_free (pServ->SubServiceList, NULL);
        MDSUTIL_free (pModuleInfo->ServiceList, NULL);
        mdsutil_FreeRecord (MDS_CDSADIR_CL_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }
    pSubServ = (MDSUTIL_CLSUBSERVICE_PTR)pServ->SubServiceList->SubServiceInfo;
    
    if (Record.ssid.Data)
        pSubServ->SubServiceId = *(uint32 *)Record.ssid.Data;;
    strcpy (pSubServ->Description, pModuleInfo->Description);
    if (Record.certtypeformat.Data)
        type = *(uint32 *)Record.certtypeformat.Data;
    pSubServ->CertType = (CSSM_CERT_TYPE) (type >> 16);
    pSubServ->CertEncoding = (CSSM_CERT_ENCODING) (type & 0x0000FFFF);

    if (Record.bundletypeformat.Length) {
        pSubServ->NumberOfBundleInfos = (uint32)(Record.bundletypeformat.Length/sizeof (CSSM_CERT_BUNDLE_HEADER)) + 1;
        if ((pSubServ->BundleInfo = (CSSM_CERT_BUNDLE_HEADER_PTR)MDSUTIL_calloc (pSubServ->NumberOfBundleInfos,
                                            (uint32)sizeof (CSSM_CERT_BUNDLE_HEADER), NULL)) == NULL) {
            MDSUTIL_free (pServ->SubServiceList, NULL);
            MDSUTIL_free (pModuleInfo->ServiceList, NULL);
            mdsutil_FreeRecord (MDS_CDSADIR_CL_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
            return (CSSM_ERRCODE_MEMORY_ERROR);
        }
        memcpy (pSubServ->BundleInfo, Record.bundletypeformat.Data, Record.bundletypeformat.Length);
    }

    if (Record.xlationtypeformat.Length) {
        pSubServ->NumberOfTranslationTypes = (uint32)(Record.xlationtypeformat.Length/sizeof (CSSM_CERT_TYPE));

        if ((pSubServ->CertTranslationTypes = (CSSM_CERT_TYPE_PTR)MDSUTIL_calloc (pSubServ->NumberOfTranslationTypes,
                                                    (uint32)sizeof (CSSM_CERT_TYPE), NULL)) == NULL) {
            if (pSubServ->CertTemplate) {
                MDSUTIL_free (pSubServ->CertTemplate->Data, NULL);
                MDSUTIL_free (pSubServ->CertTemplate, NULL);
            }
            MDSUTIL_free (pSubServ->BundleInfo, NULL);
            MDSUTIL_free (pServ->SubServiceList, NULL);
            MDSUTIL_free (pModuleInfo->ServiceList, NULL);
            mdsutil_FreeRecord (MDS_CDSADIR_CL_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
            return (CSSM_ERRCODE_MEMORY_ERROR);
        }
        memcpy (pSubServ->CertTranslationTypes, Record.xlationtypeformat.Data, Record.xlationtypeformat.Length);
    }

    mdsutil_FreeRecord (MDS_CDSADIR_CL_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_mds_get_cl_encapsulated_product(CSSM_DL_DB_HANDLE hDLDB,
                                                            char* GuidStr,
                                                            uint32 SubserviceID,
                                                            MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    cssm_MDS_CDSADIR_CL_ENCAPSULATED_PRODUCT Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA EncapAttributes[MDS_CDSADIR_CL_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES];
    CSSM_HANDLE ResultsHandle = 0;
    CSSM_RETURN retValue = CSSM_OK;
    MDSUTIL_CL_WRAPPEDPRODUCTINFO_PTR pWrap;
    uint32 type = 0;
    MDSUTIL_CL_ENCODER_PRODUCTINFO_PTR pEncInfo;
    MDSUTIL_CL_CA_PRODUCTINFO_PTR pCAInfo;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_CL_ENCAPSULATED_PRODUCT ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &EncapAttributes, 0, MDS_CDSADIR_CL_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = EncapAttributes;
    if ((retValue = cl_encapsulated_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);
    
    if ((retValue = mdsutil_GetRecordByGuidAndSSID(hDLDB,
                                                   GuidStr,
                                                   SubserviceID,
                                                   MDS_CDSADIR_CL_ENCAPSULATED_PRODUCT_RECORDTYPE,
                                                   MDS_CDSADIR_CL_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES,
                                                   &ResultsHandle,
                                                   EncapAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );

    Record.moduleid = *EncapAttributes[0].Value;
    Record.ssid = *EncapAttributes[1].Value;
    Record.productdesc = *EncapAttributes[2].Value;
    Record.productvendor = *EncapAttributes[3].Value;
    Record.productversion = *EncapAttributes[4].Value;
    Record.productflags = *EncapAttributes[5].Value;
    Record.standarddesc = *EncapAttributes[6].Value;
    Record.standardversion = *EncapAttributes[7].Value;

    pWrap = &(((MDSUTIL_CLSUBSERVICE_PTR)(pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1].
                                            SubServiceList->SubServiceInfo))->WrappedProduct);
    if ((pWrap->EmbeddedEncoderProducts = (MDSUTIL_CL_ENCODER_PRODUCTINFO_PTR)MDSUTIL_calloc (1, 
                                (uint32)sizeof (MDSUTIL_CL_ENCODER_PRODUCTINFO), NULL)) == NULL) {
        mdsutil_FreeRecord (MDS_CDSADIR_CL_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES, EncapAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }

    if ((pWrap->AccessibleCAProducts = (MDSUTIL_CL_CA_PRODUCTINFO_PTR)MDSUTIL_calloc (1, 
                                        (uint32)sizeof (MDSUTIL_CL_CA_PRODUCTINFO), NULL)) == NULL) {
        mdsutil_FreeRecord (MDS_CDSADIR_CL_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES, EncapAttributes);
        MDSUTIL_free (pWrap->EmbeddedEncoderProducts, NULL);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }

    pWrap->NumberOfEncoderProducts = 1;
    pWrap->NumberOfCAProducts = 1;

    pEncInfo = pWrap->EmbeddedEncoderProducts;
    mdsutil_StringToVersion (Record.standardversion.Data, &pEncInfo->StandardVersion);
    strncpy (pEncInfo->StandardDescription, (const char *)Record.standarddesc.Data, Record.standarddesc.Length);
    mdsutil_StringToVersion (Record.productversion.Data, &pEncInfo->ProductVersion);
    strncpy (pEncInfo->ProductDescription, (const char *)Record.productdesc.Data, Record.productdesc.Length);
    strncpy (pEncInfo->ProductVendor, (const char *)Record.productvendor.Data, Record.productvendor.Length);
    
    pEncInfo->CertType = (CSSM_CERT_TYPE) (type >> 16);

    pCAInfo = pWrap->AccessibleCAProducts;
    mdsutil_StringToVersion (Record.standardversion.Data, &pCAInfo->StandardVersion);
    strncpy (pCAInfo->StandardDescription, (const char *)Record.standarddesc.Data, Record.standarddesc.Length);
    mdsutil_StringToVersion (Record.productversion.Data, &pCAInfo->ProductVersion);
    strncpy (pCAInfo->ProductDescription, (const char *)Record.productdesc.Data, Record.productdesc.Length);
    strncpy (pCAInfo->ProductVendor, (const char *)Record.productvendor.Data, Record.productvendor.Length);
    type =0;
    pCAInfo->CertType = (CSSM_CERT_TYPE) (type >> 16);
    pCAInfo->CertEncoding = (CSSM_CERT_ENCODING) (type & 0x0000FFFF);
    pCAInfo->NumberOfCertClasses = 1;
    if ((pCAInfo->CertClasses = (MDSUTIL_CL_CA_CERT_CLASSINFO_PTR)MDSUTIL_calloc (1, 
                                        (uint32)sizeof (MDSUTIL_CL_CA_CERT_CLASSINFO), NULL)) == NULL) {
        MDSUTIL_free (pWrap->EmbeddedEncoderProducts, NULL);
        MDSUTIL_free (pWrap->AccessibleCAProducts, NULL);
        mdsutil_FreeRecord (MDS_CDSADIR_CL_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES, EncapAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }

    mdsutil_FreeRecord (MDS_CDSADIR_CL_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES, EncapAttributes);
        
    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_mds_get_dl_primary(CSSM_DL_DB_HANDLE hDLDB,
                                               char* GuidStr,
                                               uint32 SubserviceID,
                                               MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    cssm_MDS_CDSADIR_DL_PRIMARY Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA PrimaryAttributes[MDS_CDSADIR_DL_PRIMARY_NUM_ATTRIBUTES];
    CSSM_HANDLE ResultsHandle = 0;
    MDSUTIL_SERVICE_INFO_PTR pServ;
    MDSUTIL_DLSUBSERVICE_PTR pSubServ;
    CSSM_RETURN retValue = CSSM_OK;
    uint32 dltype =0;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_DL_PRIMARY ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &PrimaryAttributes, 0, MDS_CDSADIR_DL_PRIMARY_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));
    
    Attributes.AttributeData = PrimaryAttributes;
    if ((retValue = dl_primary_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);
    
    if ((retValue = mdsutil_GetRecordByGuidAndSSID(hDLDB,
                                                   GuidStr,
                                                   SubserviceID,
                                                   MDS_CDSADIR_DL_PRIMARY_RECORDTYPE,
                                                   MDS_CDSADIR_DL_PRIMARY_NUM_ATTRIBUTES,
                                                   &ResultsHandle,
                                                   PrimaryAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );

    Record.moduleid = *PrimaryAttributes[0].Value;
    Record.ssid = *PrimaryAttributes[1].Value;
    Record.manifest = *PrimaryAttributes[2].Value;
    Record.modulename = *PrimaryAttributes[3].Value;
    Record.addinversion = *PrimaryAttributes[4].Value;
    Record.addinvendor = *PrimaryAttributes[5].Value;
    Record.dltype = *PrimaryAttributes[6].Value;
    Record.querylimitsflag = *PrimaryAttributes[7].Value;
    Record.sampletypes = *PrimaryAttributes[8].Value;
    Record.aclsubjecttypes = *PrimaryAttributes[9].Value;
    Record.authtags = *PrimaryAttributes[10].Value;
    Record.conjunctiveops = *PrimaryAttributes[11].Value;
    Record.relationalops = *PrimaryAttributes[12].Value;

    pModuleInfo->NumberOfServices++;
    if ((pModuleInfo->ServiceList = (MDSUTIL_SERVICE_INFO_PTR)MDSUTIL_realloc (pModuleInfo->ServiceList,
                                    (uint32)(sizeof (MDSUTIL_SERVICE_INFO) * pModuleInfo->NumberOfServices), NULL)) == NULL) {
        mdsutil_FreeRecord (MDS_CDSADIR_DL_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }
  
    mdsutil_StringToVersion (Record.addinversion.Data, &pModuleInfo->Version);
    strncpy (pModuleInfo->Vendor, (const char *)Record.addinvendor.Data, Record.addinvendor.Length);
    
    pServ = &pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1];
    strcpy (pServ->Description, pModuleInfo->Description);
    pServ->Type = CSSM_SERVICE_DL;
    pServ->Flags = 0;
    pServ->NumberOfSubServices = 1;

    if ((pServ->SubServiceList = (MDSUTIL_SUBSERVICE_LIST_PTR)MDSUTIL_calloc (1, (uint32)sizeof (MDSUTIL_SUBSERVICE_LIST), NULL)) == NULL) {
        MDSUTIL_free (pModuleInfo->ServiceList, NULL);
        mdsutil_FreeRecord (MDS_CDSADIR_DL_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }

    if (Record.ssid.Data)
        pServ->SubServiceList->SubServiceId = *(uint32 *)Record.ssid.Data;
    
    if ((pServ->SubServiceList->SubServiceInfo = (void*)MDSUTIL_calloc (1, (uint32)sizeof (MDSUTIL_DLSUBSERVICE), NULL)) == NULL) {
        MDSUTIL_free (pServ->SubServiceList, NULL);
        MDSUTIL_free (pModuleInfo->ServiceList, NULL);
        mdsutil_FreeRecord (MDS_CDSADIR_DL_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }
    pSubServ = (MDSUTIL_DLSUBSERVICE_PTR)pServ->SubServiceList->SubServiceInfo;
    
    if (Record.ssid.Data)
        pSubServ->SubServiceId = *(uint32 *)Record.ssid.Data;;
    strcpy (pSubServ->Description, pModuleInfo->Description);
    if (Record.dltype.Data)
    {
        dltype = *(uint32 *)Record.dltype.Data;
        switch(dltype)
        {
        case 0:
            pSubServ->Type = CSSM_DL_UNKNOWN;
            break;
        case 1:
            pSubServ->Type = CSSM_DL_CUSTOM;
            break;
        case 2:
            pSubServ->Type = CSSM_DL_LDAP;
            break;
        case 3:
            pSubServ->Type = CSSM_DL_ODBC;
            break;
        case 4:
            pSubServ->Type = CSSM_DL_PKCS11;
            break;
        case 5:
            pSubServ->Type = CSSM_DL_FFS;
            break;
        case 6:
            pSubServ->Type = CSSM_DL_MEMORY;
            break;
        case 7:
            pSubServ->Type = CSSM_DL_REMOTEDIR;
            break;
        default:
            break;
        }
    }

    if (Record.relationalops.Length) {
        pSubServ->NumberOfRelOperatorTypes = (uint32)(Record.relationalops.Length/sizeof (CSSM_DB_OPERATOR)); 

        if ((pSubServ->RelOperatorTypes = (CSSM_DB_OPERATOR_PTR)MDSUTIL_calloc (pSubServ->NumberOfRelOperatorTypes,
                                                (uint32)sizeof (CSSM_DB_OPERATOR), NULL)) == NULL) {
            MDSUTIL_free (pServ->SubServiceList, NULL);
            MDSUTIL_free (pModuleInfo->ServiceList, NULL);
            mdsutil_FreeRecord (MDS_CDSADIR_DL_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
            return (CSSM_ERRCODE_MEMORY_ERROR);
        }
        memcpy (pSubServ->RelOperatorTypes, Record.relationalops.Data, Record.relationalops.Length);
    }

    if (Record.conjunctiveops.Length) {
        pSubServ->NumberOfConjOperatorTypes = (uint32)(Record.conjunctiveops.Length/sizeof (CSSM_DB_CONJUNCTIVE)); 
        
        if ((pSubServ->ConjOperatorTypes = (CSSM_DB_CONJUNCTIVE_PTR)MDSUTIL_calloc (pSubServ->NumberOfConjOperatorTypes,
                                                (uint32)sizeof (CSSM_DB_CONJUNCTIVE), NULL)) == NULL) {
            MDSUTIL_free (pSubServ->RelOperatorTypes, NULL);
            MDSUTIL_free (pServ->SubServiceList, NULL);
            MDSUTIL_free (pModuleInfo->ServiceList, NULL);
            mdsutil_FreeRecord (MDS_CDSADIR_DL_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
            return (CSSM_ERRCODE_MEMORY_ERROR);
        }
        memcpy (pSubServ->ConjOperatorTypes, Record.conjunctiveops.Data, Record.conjunctiveops.Length);
    }

    if (Record.querylimitsflag.Data)
        pSubServ->QueryLimitsSupported =  *(CSSM_BOOL *)Record.querylimitsflag.Data; 
    
    mdsutil_FreeRecord (MDS_CDSADIR_DL_PRIMARY_NUM_ATTRIBUTES, PrimaryAttributes);
        
    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_mds_get_dl_encapsulated_product(CSSM_DL_DB_HANDLE hDLDB,
                                                            char* GuidStr,
                                                            uint32 SubserviceID,
                                                            MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    cssm_MDS_CDSADIR_DL_ENCAPSULATED_PRODUCT Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA EncapAttributes[MDS_CDSADIR_DL_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES];
    CSSM_HANDLE ResultsHandle = 0;
    MDSUTIL_DL_WRAPPEDPRODUCTINFO_PTR pWrap;
    CSSM_RETURN retValue = CSSM_OK;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_DL_ENCAPSULATED_PRODUCT ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &EncapAttributes, 0, MDS_CDSADIR_DL_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = EncapAttributes;
    if ((retValue = dl_encapsulated_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);
    
    if ((retValue = mdsutil_GetRecordByGuidAndSSID(hDLDB,
                                                   GuidStr,
                                                   SubserviceID,
                                                   MDS_CDSADIR_DL_ENCAPSULATED_PRODUCT_RECORDTYPE,
                                                   MDS_CDSADIR_DL_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES,
                                                   &ResultsHandle,
                                                   EncapAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );

    Record.moduleid = *EncapAttributes[0].Value;
    Record.ssid = *EncapAttributes[1].Value;
    Record.productdesc = *EncapAttributes[2].Value;
    Record.productvendor = *EncapAttributes[3].Value;
    Record.productversion = *EncapAttributes[4].Value;
    Record.productflags = *EncapAttributes[5].Value;
    Record.standarddesc = *EncapAttributes[6].Value;
    Record.standardversion = *EncapAttributes[7].Value;
    Record.protocol = *EncapAttributes[8].Value;
    Record.retrievalmode = *EncapAttributes[9].Value;

    pWrap = &(((MDSUTIL_DLSUBSERVICE_PTR)(pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1].
                                            SubServiceList->SubServiceInfo))->WrappedProduct);
    mdsutil_StringToVersion (Record.standardversion.Data, &pWrap->StandardVersion);
    strncpy (pWrap->StandardDescription, (const char *)Record.standarddesc.Data, Record.standarddesc.Length);
    strncpy (pWrap->StandardDescription, (const char *)Record.standarddesc.Data, Record.standarddesc.Length);
    strncpy (pWrap->ProductDescription, (const char *)Record.productdesc.Data, Record.productdesc.Length);

    if (Record.protocol.Data)
        pWrap->NetworkProtocol = *(CSSM_NET_PROTOCOL *)Record.protocol.Data;

    mdsutil_FreeRecord (MDS_CDSADIR_DL_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES, EncapAttributes);

    return (retValue);
}

/*---------------------------------------------------------------
 *
 *Name: mdsutil_StrToGUID
 *
 *Description:
 *   Given a string, covert it to a guid.
 *
 *Parameters: 
 *   ID - pointer to the guid string
 *   GUID - pointer to a CSSM_GUID to receive the guid
 *
 *Returns:
 *  NOne
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI mdsutil_StrToGUID (char *ID, CSSM_GUID_PTR Guid)
{
    uint32 i, Data1;
    uint32 Data2, Data3;
    uint32 Data4[8];

    if ((ID == NULL) || (Guid == NULL))
        return (CSSMERR_CSSM_INVALID_POINTER);

    sscanf (ID, GUID_FORMAT_STRING, 
                     &Data1, &Data2, &Data3, &Data4[0],
                     &Data4[1], &Data4[2], &Data4[3],
                     &Data4[4], &Data4[5], &Data4[6],
                     &Data4[7]);
        
    Guid->Data1 = Data1;
    Guid->Data2 = (uint16) Data2;
    Guid->Data3 = (uint16) Data3;
    for (i=0; i<8; i++)
      Guid->Data4[i] = (uint8) Data4[i];

    return CSSM_OK;
}

CSSM_RETURN CSSMAPI mdsutil_get_cssm_info(CSSM_DL_DB_HANDLE hDLDB,
                                          char* GuidStr,
                                          MDSUTIL_MODULE_INFO_PTR *pModuleInfo)
{
    cssm_MDS_CDSADIR_CSSM Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA CssmAttributes[MDS_CDSADIR_CSSM_NUM_ATTRIBUTES];
    CSSM_HANDLE ResultsHandle = 0;
    CSSM_RETURN retValue = CSSM_OK;
    
    *pModuleInfo = NULL;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_CSSM ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &CssmAttributes, 0, MDS_CDSADIR_CSSM_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = CssmAttributes;
    if ((retValue = cssm_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);
    
    if ((retValue = mdsutil_GetRecordByGuid(hDLDB,
                                            GuidStr,
                                            MDS_CDSADIR_CSSM_RECORDTYPE,
                                            MDS_CDSADIR_CSSM_NUM_ATTRIBUTES,
                                            &ResultsHandle,
                                            CssmAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );

    Record.moduleid = *CssmAttributes[0].Value;
    Record.version = *CssmAttributes[1].Value;
    Record.vendor = *CssmAttributes[2].Value;
    Record.desc = *CssmAttributes[3].Value;
    Record.nativeservices = *CssmAttributes[4].Value;


    if ((*pModuleInfo = (MDSUTIL_MODULE_INFO_PTR)MDSUTIL_calloc (1, (uint32)sizeof (MDSUTIL_MODULE_INFO), NULL)) == NULL) {
        mdsutil_FreeRecord (MDS_CDSADIR_CSSM_NUM_ATTRIBUTES, CssmAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }
    
    mdsutil_StrToGUID ((char *)Record.moduleid.Data, &(*pModuleInfo)->ModuleGuid);
    mdsutil_StringToVersion (Record.version.Data, &(*pModuleInfo)->Version);
    strncpy ((*pModuleInfo)->Description, (const char *)Record.desc.Data, Record.desc.Length);
    strncpy ((*pModuleInfo)->Vendor, (const char *)Record.vendor.Data, Record.vendor.Length);

    mdsutil_FreeRecord (MDS_CDSADIR_CSSM_NUM_ATTRIBUTES, CssmAttributes);
    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_get_common_info(CSSM_DL_DB_HANDLE hDLDB,
                                            char* GuidStr,
                                            uint32 *Dynamic,
                                            MDSUTIL_MODULE_INFO_PTR *pModuleInfo)
{
    cssm_MDS_CDSADIR_COMMON Record;
    CSSM_DB_ATTRIBUTE_DATA CommonAttributes[MDS_CDSADIR_COMMON_NUM_ATTRIBUTES];
    CSSM_RETURN retValue = CSSM_OK;

    *pModuleInfo = NULL;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_COMMON ));
    memset( &CommonAttributes, 0, MDS_CDSADIR_COMMON_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    if ((retValue = mdsutil_GetCommonRecord(hDLDB,
                                            GuidStr,
                                            &Record,
                                            CommonAttributes)) != CSSM_OK)
    {
        return (retValue);
    }
    
    if ((*pModuleInfo = (MDSUTIL_MODULE_INFO_PTR)MDSUTIL_calloc (1, (uint32)sizeof (MDSUTIL_MODULE_INFO), NULL)) == NULL) {
        mdsutil_FreeRecord (MDS_CDSADIR_COMMON_NUM_ATTRIBUTES, CommonAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }
    
    mdsutil_StrToGUID ((char *)Record.guid.Data, &(*pModuleInfo)->ModuleGuid);  
    mdsutil_StringToVersion (Record.Version.Data, &(*pModuleInfo)->CompatibleCSSMVersion);
    strncpy ((*pModuleInfo)->Description, (const char *)Record.Desc.Data, Record.Desc.Length);

    if (Record.ServiceMask.Data)
        (*pModuleInfo)->ServiceMask = *(CSSM_SERVICE_MASK *)Record.ServiceMask.Data;
    
    if (Record.Dynamic.Data && Dynamic)
        *Dynamic = *(uint32 *)Record.Dynamic.Data;
    
    mdsutil_FreeRecord (MDS_CDSADIR_COMMON_NUM_ATTRIBUTES, CommonAttributes);
    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_get_csp_info(CSSM_DL_DB_HANDLE hDLDB,
                                         char* GuidStr,
                                         uint32 SubserviceID,
                                         CSSM_USEE_TAG USEERequest,
                                         MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    MDSUTIL_SERVICE_INFO_PTR pServ;
    CSSM_CSPTYPE cspType;
    CSSM_RETURN retValue = CSSM_OK;

    if ((retValue = mdsutil_mds_get_csp_primary(hDLDB, 
                                                GuidStr,
                                                SubserviceID,
                                                pModuleInfo)) != CSSM_OK)
    {
        return (retValue);
    }
    
    pServ = &pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1];
    cspType = ((MDSUTIL_CSPSUBSERVICE_PTR)(pServ->SubServiceList->SubServiceInfo))->CspType;

    switch (cspType) 
    {
        case CSSM_CSP_SOFTWARE:
            if( USEERequest == CSSM_USEE_NONE )
            {
                if ((retValue = mdsutil_mds_get_software_csp_capabilities(hDLDB,
                                                                          GuidStr,
                                                                          SubserviceID,
                                                                          USEERequest,
                                                                          pModuleInfo))
                                                                          != CSSM_OK)
                {
                    return (retValue);
                }
            }
            else
            {
                if ((retValue = mdsutil_mds_get_software_csp_capabilities(hDLDB,
                                                                          GuidStr,
                                                                          SubserviceID,
                                                                          CSSM_USEE_NONE,
                                                                          pModuleInfo))
                                                                          != CSSM_OK)
                {
                    return (retValue);
                }

                if ((retValue = mdsutil_mds_get_additional_csp_capabilities(hDLDB,
                                                                            GuidStr,
                                                                            SubserviceID,
                                                                            USEERequest,
                                                                            pModuleInfo))
                                                                            != CSSM_OK)
                {
                    return (retValue);
                }
            }
            break;
        case CSSM_CSP_HARDWARE:
        case CSSM_CSP_HYBRID:
            if ((retValue = mdsutil_mds_get_hardware_csp_capabilities(hDLDB,
                                                                      GuidStr,
                                                                      SubserviceID,
                                                                      USEERequest,
                                                                      pModuleInfo))
                                                                      != CSSM_OK)
            {
                return (retValue);
            }
            break;
        default:
            return (CSSMERR_CSSM_INTERNAL_ERROR);
    }

    if ((retValue = mdsutil_mds_get_csp_encapsulated_product(hDLDB,
                                                             GuidStr,
                                                             cspType,
                                                             SubserviceID,
                                                             pModuleInfo))
                                                             != CSSM_OK)
    {
        return (retValue);
    }

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_get_tp_info(CSSM_DL_DB_HANDLE hDLDB,
                                        char* GuidStr,
                                        uint32 SubserviceID,
                                        MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    MDSUTIL_SERVICE_INFO_PTR pServ;
    CSSM_RETURN retValue = CSSM_OK;

    if ((retValue = mdsutil_mds_get_tp_primary(hDLDB,
                                               GuidStr,
                                               SubserviceID,
                                               pModuleInfo)) != CSSM_OK)
    {
        return (retValue);
    }

    if ((retValue = mdsutil_mds_get_tp_policy_oids(hDLDB,
                                                   GuidStr,
                                                   SubserviceID,
                                                   pModuleInfo)) != CSSM_OK)
    {
        pServ = &pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1];
        MDSUTIL_free (pServ->SubServiceList->SubServiceInfo, NULL);
        MDSUTIL_free (pServ->SubServiceList, NULL);
        MDSUTIL_free (pModuleInfo->ServiceList, NULL);
        return (retValue);
    }

    if ((retValue = mdsutil_mds_get_tp_encapsulated_product(hDLDB,
                                                            GuidStr,
                                                            SubserviceID,
                                                            pModuleInfo))
                                                            != CSSM_OK)
    {
        pServ = &pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1];
        MDSUTIL_free (pServ->SubServiceList->SubServiceInfo, NULL);
        MDSUTIL_free (pServ->SubServiceList, NULL);
        MDSUTIL_free (pModuleInfo->ServiceList, NULL);
        return (retValue);
    }
      
    return (retValue);
}


CSSM_RETURN CSSMAPI mdsutil_get_cl_info(CSSM_DL_DB_HANDLE hDLDB,
                                        char* GuidStr,
                                        uint32 SubserviceID,
                                        MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    MDSUTIL_SERVICE_INFO_PTR pServ;
    CSSM_RETURN retValue = CSSM_OK;

    if ((retValue = mdsutil_mds_get_cl_primary(hDLDB,
                                               GuidStr,
                                               SubserviceID,
                                               pModuleInfo)) != CSSM_OK)
    {
        return (retValue);
    }

    if ((retValue = mdsutil_mds_get_cl_encapsulated_product(hDLDB,
                                                            GuidStr,
                                                            SubserviceID,
                                                            pModuleInfo))
                                                            != CSSM_OK) 
    {
        pServ = &pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1];
        MDSUTIL_free (pServ->SubServiceList->SubServiceInfo, NULL);
        MDSUTIL_free (pServ->SubServiceList, NULL);
        MDSUTIL_free (pModuleInfo->ServiceList, NULL);
        return (retValue);
    }

    return (retValue);
}



CSSM_RETURN CSSMAPI mdsutil_get_ac_info(CSSM_DL_DB_HANDLE hDLDB,
                                        char* GuidStr,
                                        uint32 SubserviceID,
                                        MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    CSSM_RETURN retValue = CSSM_OK;

    retValue = mdsutil_mds_get_ac_primary(hDLDB,
                                          GuidStr,
                                          SubserviceID,
                                          pModuleInfo);

    return (retValue);
}



CSSM_RETURN CSSMAPI mdsutil_get_dl_info(CSSM_DL_DB_HANDLE hDLDB,
                                        char* GuidStr,
                                        uint32 SubserviceID,
                                        MDSUTIL_MODULE_INFO_PTR pModuleInfo)
{
    MDSUTIL_SERVICE_INFO_PTR pServ;
    CSSM_RETURN retValue = CSSM_OK;

    if ((retValue = mdsutil_mds_get_dl_primary(hDLDB,
                                               GuidStr,
                                               SubserviceID,
                                               pModuleInfo)) != CSSM_OK)
    {
        return (retValue);
    }

    if ((retValue = mdsutil_mds_get_dl_encapsulated_product(hDLDB,
                                                            GuidStr,
                                                            SubserviceID,
                                                            pModuleInfo))
                                                            != CSSM_OK)
    {
        pServ = &pModuleInfo->ServiceList[pModuleInfo->NumberOfServices-1];
        MDSUTIL_free (pServ->SubServiceList->SubServiceInfo, NULL);
        MDSUTIL_free (pServ->SubServiceList, NULL);
        MDSUTIL_free (pModuleInfo->ServiceList, NULL);
        return (retValue);
    }

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_mds_get_moduleInfo(const CSSM_GUID *ModuleGUID,
                                               CSSM_SERVICE_MASK UsageMask,
                                               uint32 SubserviceID,
                                               CSSM_USEE_TAG USEERequest,
                                               MDSUTIL_MODULE_INFO_PTR *pModuleInfo)
{
    CSSM_RETURN retValue = CSSM_OK;
    CSSM_SERVICE_MASK mask;
    uint32 Dynamic;
    char GuidStr[MDS_UTIL_MAX_PATH];

    /*
     * Convert the ModuleGuid to a string.
     */
    mdsutil_GUIDToStr (ModuleGUID, GuidStr);

    /*
     * In case of CSSM, simply return the information from CSSM record.
     */
    if (UsageMask == CSSM_SERVICE_CSSM) 
    {
        retValue = mdsutil_get_cssm_info (hDLDBCdsa, GuidStr, pModuleInfo);
        return (retValue);
    }

    if ((retValue = mdsutil_get_common_info(hDLDBCdsa,
                                            GuidStr,
                                            &Dynamic,
                                            pModuleInfo)) != CSSM_OK)
    {
        return (retValue);
    }

    mask = UsageMask;
    while (mask)
    {
        if (mask & CSSM_SERVICE_CSP)
        {
            if((retValue = mdsutil_get_csp_info(hDLDBCdsa,
                                                GuidStr,
                                                SubserviceID,
                                                USEERequest,
                                                *pModuleInfo)) != CSSM_OK)
            {
                mdsutil_FreeModule(*pModuleInfo);
                *pModuleInfo = NULL;
                return (retValue);
            }

            /*always need to set the following information because CSSM_FreeModule( ) requires it */
            if ((*pModuleInfo) &&
                (*pModuleInfo)->ServiceList &&
                (*pModuleInfo)->ServiceList->SubServiceList)
            {
                (*pModuleInfo)->ServiceList->SubServiceList->ModuleType = Dynamic;
            }
            mask &= ~CSSM_SERVICE_CSP;

            if (retValue != CSSM_OK)
            {
                break;
            }
        } 
        else if (mask & CSSM_SERVICE_TP)
        {
            retValue = mdsutil_get_tp_info(hDLDBCdsa,
                                           GuidStr,
                                           SubserviceID,
                                           *pModuleInfo);

            /*always need to set the following information because CSSM_FreeModule( ) requires it */
            if ((*pModuleInfo) &&
                (*pModuleInfo)->ServiceList &&
                (*pModuleInfo)->ServiceList->SubServiceList)
            {
                (*pModuleInfo)->ServiceList->SubServiceList->ModuleType = Dynamic;
            }
            mask &= ~CSSM_SERVICE_TP;
            if ( retValue != CSSM_OK )
            {
                break;
            }
        }
        else if (mask & CSSM_SERVICE_CL) 
        {
            retValue = mdsutil_get_cl_info(hDLDBCdsa,
                                           GuidStr,
                                           SubserviceID,
                                           *pModuleInfo);

            /*always need to set the following information because CSSM_FreeModule( ) requires it */
            if ((*pModuleInfo) &&
                (*pModuleInfo)->ServiceList &&
                (*pModuleInfo)->ServiceList->SubServiceList)
            {
                (*pModuleInfo)->ServiceList->SubServiceList->ModuleType = Dynamic;
            }
            mask &= ~CSSM_SERVICE_CL;

            if ( retValue != CSSM_OK )
            {
                break;
            }
        }
        else if (mask & CSSM_SERVICE_DL) 
        {
            retValue = mdsutil_get_dl_info(hDLDBCdsa,
                                           GuidStr,
                                           SubserviceID,
                                           *pModuleInfo);

            /*always need to set the following information because CSSM_FreeModule( ) requires it */
            if ((*pModuleInfo)&&
                (*pModuleInfo)->ServiceList &&
                (*pModuleInfo)->ServiceList->SubServiceList)
            {
                (*pModuleInfo)->ServiceList->SubServiceList->ModuleType = Dynamic;
            }
            mask &= ~CSSM_SERVICE_DL;
            if ( retValue != CSSM_OK )
            {
                break;
            }
        }
        else if (mask & CSSM_SERVICE_AC) 
        {
            if((retValue = mdsutil_get_ac_info(hDLDBCdsa,
                                               GuidStr,
                                               SubserviceID,
                                               *pModuleInfo)) != CSSM_OK)
            {
                *pModuleInfo = NULL;
                return (retValue);
            }

            /*always need to set the following information because CSSM_FreeModule( ) requires it */
            if ((*pModuleInfo) &&
                (*pModuleInfo)->ServiceList &&
                (*pModuleInfo)->ServiceList->SubServiceList)
            {
                (*pModuleInfo)->ServiceList->SubServiceList->ModuleType = Dynamic;
            }
            mask &= ~CSSM_SERVICE_AC;

            if ( retValue != CSSM_OK )
            {
                break;
            }
        }
    } /*while*/

    /*cleanup if necessary*/
    if ( retValue != CSSM_OK )
    {
        mdsutil_FreeModule( *pModuleInfo );
        *pModuleInfo = NULL;
    }

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_GetObjectCredentialInfo(char* GuidStr,
                                                    CSSM_DATA_PTR Credential,
                                                    CSSM_DATA_PTR ModuleName,
                                                    CSSM_DATA_PTR ModulePath)
{
    cssm_MDS_OBJECT_RECORD Record;
    CSSM_DB_ATTRIBUTE_DATA ObjectAttributes[MDS_OBJECT_NUM_ATTRIBUTES];
    CSSM_RETURN retValue = CSSM_OK;

    memset( &Record, 0, sizeof( cssm_MDS_OBJECT_RECORD ));
    memset( &ObjectAttributes, 0, MDS_OBJECT_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    if ((retValue = mdsutil_GetObjectRecord(hDLDBObject,
                                            GuidStr,
                                            &Record,
                                            ObjectAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    if (Credential)
    {
        if ((Credential->Data = (uint8 *)MDSUTIL_calloc (1, Record.Manifest.Length, NULL)) == NULL)
        {
            mdsutil_FreeRecord (MDS_OBJECT_NUM_ATTRIBUTES, ObjectAttributes);
            return (CSSM_ERRCODE_MEMORY_ERROR);
        }

        memcpy (Credential->Data, Record.Manifest.Data, Record.Manifest.Length);
        Credential->Length = Record.Manifest.Length;
    }

    if (ModuleName)
    {
        if ((ModuleName->Data = (uint8 *)MDSUTIL_calloc (1, Record.ModuleName.Length, NULL)) == NULL)
        {
            mdsutil_FreeRecord (MDS_OBJECT_NUM_ATTRIBUTES, ObjectAttributes);
            return (CSSM_ERRCODE_MEMORY_ERROR);
        }

        memcpy (ModuleName->Data, Record.ModuleName.Data, Record.ModuleName.Length);
        ModuleName->Length = Record.ModuleName.Length;
    }

    if (ModulePath)
    {
        if ((ModulePath->Data = (uint8 *)MDSUTIL_calloc (1, Record.Path.Length, NULL)) == NULL)
        {
            mdsutil_FreeRecord (MDS_OBJECT_NUM_ATTRIBUTES, ObjectAttributes);
            return (CSSM_ERRCODE_MEMORY_ERROR);
        }

        memcpy (ModulePath->Data, Record.Path.Data, Record.Path.Length);
        ModulePath->Length = Record.Path.Length;
    }
    
    mdsutil_FreeRecord (MDS_OBJECT_NUM_ATTRIBUTES, ObjectAttributes);

    return (retValue);
}


CSSM_RETURN CSSMAPI 
mdsutil_updateListEntry (MDSUTIL_LIST_PTR pList, cssm_MDS_CDSADIR_COMMON Record, uint32 SubServiceId)

{
    MDSUTIL_LIST_ITEM_PTR pItem;

    pList->NumberItems++;

    if ((pList->Items = (MDSUTIL_LIST_ITEM_PTR)MDSUTIL_realloc (pList->Items, 
                        (uint32)(sizeof (MDSUTIL_LIST_ITEM) * pList->NumberItems), NULL)) == NULL)
        return (CSSM_ERRCODE_MEMORY_ERROR);


    pItem = &pList->Items[pList->NumberItems-1];

    memset (pItem, 0, sizeof (MDSUTIL_LIST_ITEM));
                        
    if ((pItem->Name = (char*)MDSUTIL_calloc (1, Record.Desc.Length, NULL)) == NULL)
        return (CSSM_ERRCODE_MEMORY_ERROR);

    strncpy (pItem->Name, (const char *)Record.Desc.Data, Record.Desc.Length);

    mdsutil_StrToGUID ((char *)Record.guid.Data, &pItem->SubserviceUid.Guid);
    mdsutil_StringToVersion (Record.Version.Data, &pItem->SubserviceUid.Version);
    pItem->SubserviceUid.SubserviceId = SubServiceId;
    if (Record.ServiceMask.Data)
        pItem->SubserviceUid.SubserviceType = *(CSSM_SERVICE_TYPE *)Record.ServiceMask.Data;

    return (CSSM_OK);
}

CSSM_RETURN CSSMAPI mdsutil_GetSubServiceInfo(CSSM_DL_DB_HANDLE hDLDB,
                                              char* GuidStr,
                                              CSSM_SERVICE_TYPE mask,
                                              uint32 **SubServiceIds,
                                              uint32 *numSub)

{
    CSSM_HANDLE ResultsHandle = 0;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA CommonAttributes[MDS_CDSADIR_CL_PRIMARY_NUM_ATTRIBUTES];
    CSSM_RETURN retValue = CSSM_OK;
    uint32 recordType = 0;
    uint32 numAttributes = 0;

    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &CommonAttributes, 0, MDS_CDSADIR_CL_PRIMARY_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = CommonAttributes;

    *SubServiceIds = NULL;
    *numSub = 0;
    while (mask) {
        if (mask & CSSM_SERVICE_CSP) {
            cssm_MDS_CDSADIR_CSP_PRIMARY Record;
            memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_CSP_PRIMARY ));

            recordType = MDS_CDSADIR_CSP_PRIMARY_RECORDTYPE;
            numAttributes = MDS_CDSADIR_CSP_PRIMARY_NUM_ATTRIBUTES;
            
            csp_primary_schema_GetAttributes (&Record, &Attributes);
                
            mask &= ~CSSM_SERVICE_CSP;

        } else if (mask & CSSM_SERVICE_TP) {
            cssm_MDS_CDSADIR_TP_PRIMARY Record;
            memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_TP_PRIMARY ));

            recordType = MDS_CDSADIR_TP_PRIMARY_RECORDTYPE;
            numAttributes = MDS_CDSADIR_TP_PRIMARY_NUM_ATTRIBUTES;

            tp_primary_schema_GetAttributes (&Record, &Attributes);
                
            mask &= ~CSSM_SERVICE_TP;

        } else if (mask & CSSM_SERVICE_CL) {
            cssm_MDS_CDSADIR_CL_PRIMARY Record;
            memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_CL_PRIMARY ));

            recordType = MDS_CDSADIR_CL_PRIMARY_RECORDTYPE;
            numAttributes = MDS_CDSADIR_CL_PRIMARY_NUM_ATTRIBUTES;

            cl_primary_schema_GetAttributes (&Record, &Attributes);
        
            mask &= ~CSSM_SERVICE_CL;

        } else if (mask & CSSM_SERVICE_DL) {
            cssm_MDS_CDSADIR_DL_PRIMARY Record;
            memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_DL_PRIMARY ));

            recordType = MDS_CDSADIR_DL_PRIMARY_RECORDTYPE;
            numAttributes = MDS_CDSADIR_DL_PRIMARY_NUM_ATTRIBUTES;
                
            dl_primary_schema_GetAttributes (&Record, &Attributes);
            mask &= ~CSSM_SERVICE_DL;
        } else if (mask & CSSM_SERVICE_AC) {
            cssm_MDS_CDSADIR_AC_PRIMARY Record;
            memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_AC_PRIMARY ));

            recordType = MDS_CDSADIR_AC_PRIMARY_RECORDTYPE;
            numAttributes = MDS_CDSADIR_AC_PRIMARY_NUM_ATTRIBUTES;
                
            ac_primary_schema_GetAttributes (&Record, &Attributes);
            mask &= ~CSSM_SERVICE_AC;
        }

        ResultsHandle = 0;
        retValue = CSSM_OK;

        for (;;)    
        {
            if ((retValue = mdsutil_GetRecordByGuid(hDLDB,
                                                    GuidStr,
                                                    recordType,
                                                    numAttributes,
                                                    &ResultsHandle,
                                                    CommonAttributes))
                                                    != CSSM_OK)
            {
                if (CSSMERR_DL_ENDOFDATA == retValue)
                {   
                    retValue = CSSM_OK;
                    break;              
                }       
                MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
                return (retValue);
            }
                
            (*numSub)++;
            if ((*SubServiceIds = (uint32*)MDSUTIL_realloc (*SubServiceIds,
                                            (uint32)(sizeof(uint32) * (*numSub)), NULL)) == NULL) {
                if (*SubServiceIds)
                    MDSUTIL_free (*SubServiceIds, NULL);
                *SubServiceIds = NULL;
                mdsutil_FreeRecord (numAttributes, CommonAttributes);
                MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
                return (CSSM_ERRCODE_MEMORY_ERROR);
            }
            if (CommonAttributes[1].Value->Data)
                (*SubServiceIds)[*numSub-1] = *(uint32 *)CommonAttributes[1].Value->Data;
            else
                (*SubServiceIds)[*numSub-1] = 0;

            mdsutil_FreeRecord (numAttributes, CommonAttributes);
        }

        MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
    }
    return retValue;
}

CSSM_RETURN CSSMAPI 
mdsutil_ListCSSMModules (CSSM_DL_DB_HANDLE hDLDB,
                    MDSUTIL_LIST_PTR pList,
                    CSSM_SERVICE_MASK UsageMask,
                    CSSM_BOOL MatchAll)
{
    CSSM_HANDLE ResultsHandle = 0;
    cssm_MDS_CDSADIR_CSSM Record;
    CSSM_VERSION Version;
    CSSM_BOOL MatchRes;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA CSSMAttributes[MDS_CDSADIR_CSSM_NUM_ATTRIBUTES];
    CSSM_RETURN retValue = CSSM_OK;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_CSSM ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &CSSMAttributes, 0, MDS_CDSADIR_CSSM_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = CSSMAttributes;
    if ((retValue = cssm_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);

    retValue = CSSM_OK;
    for (;;)
    {
        if ((retValue = mdsutil_unconditionalGetRecord(
                    hDLDB,
                    MDS_CDSADIR_CSSM_RECORDTYPE,
                    MDS_CDSADIR_CSSM_NUM_ATTRIBUTES,
                    &ResultsHandle,
                    CSSMAttributes)) != CSSM_OK) 
        {
            if (CSSMERR_DL_ENDOFDATA == retValue)
            {   
                retValue = CSSM_OK;
            }       
            MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
            return (retValue);
        }
    
        Record.moduleid = *CSSMAttributes[0].Value;
        Record.version = *CSSMAttributes[1].Value;
        Record.vendor = *CSSMAttributes[2].Value;
        Record.desc = *CSSMAttributes[3].Value;
        Record.nativeservices = *CSSMAttributes[4].Value;


        mdsutil_StringToVersion (Record.version.Data, &Version);

        //if (Version.Major != CSSM_MAJOR || Version.Minor != CSSM_MINOR)
        //  continue;
        
        if (MatchAll)
            MatchRes = (CSSM_BOOL)(CSSM_SERVICE_CSSM == UsageMask);
        else
            MatchRes = (CSSM_BOOL)(CSSM_SERVICE_CSSM & UsageMask);

        if (MatchRes) {

            MDSUTIL_LIST_ITEM_PTR pItem;

            pList->NumberItems++;

            if ((pList->Items = (MDSUTIL_LIST_ITEM_PTR)MDSUTIL_realloc (pList->Items, 
                            (uint32)(sizeof (MDSUTIL_LIST_ITEM) * pList->NumberItems), NULL)) == NULL) {
                mdsutil_FreeRecord (MDS_CDSADIR_CSSM_NUM_ATTRIBUTES, CSSMAttributes);
                MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
                return (CSSM_ERRCODE_MEMORY_ERROR);
            }

            pItem = &pList->Items[pList->NumberItems-1];

            memset (pItem, 0, sizeof (MDSUTIL_LIST_ITEM));          
                        
            if ((pItem->Name = (char*)MDSUTIL_calloc (1, Record.desc.Length, NULL)) == NULL)
            {
                MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
                return (CSSM_ERRCODE_MEMORY_ERROR);
            }

            strncpy (pItem->Name, (const char *)Record.desc.Data, Record.desc.Length);
                                    
            mdsutil_StrToGUID ((char *)Record.moduleid.Data, &pItem->SubserviceUid.Guid);
            mdsutil_StringToVersion (Record.version.Data, &pItem->SubserviceUid.Version);
            pItem->SubserviceUid.SubserviceId = 0;
            pItem->SubserviceUid.SubserviceType = CSSM_SERVICE_CSSM;

            mdsutil_FreeRecord (MDS_CDSADIR_CSSM_NUM_ATTRIBUTES, CSSMAttributes);
            MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
            return (CSSM_OK);
        }
        mdsutil_FreeRecord (MDS_CDSADIR_CSSM_NUM_ATTRIBUTES, CSSMAttributes);
    }
}

CSSM_RETURN CSSMAPI mdsutil_ListModules (CSSM_SERVICE_MASK UsageMask,
                                         CSSM_BOOL MatchAll,
                                         MDSUTIL_LIST_PTR *pList)
{
    uint32 i;
    uint32 numSub;
    uint32 *SubServiceIds = NULL;
    CSSM_VERSION Version;
    CSSM_HANDLE ResultsHandle = 0;
    cssm_MDS_CDSADIR_COMMON Record;
    CSSM_SERVICE_TYPE Mask = 0;
    CSSM_BOOL MatchRes;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA CommonAttributes[MDS_CDSADIR_COMMON_NUM_ATTRIBUTES];
    CSSM_RETURN retValue = CSSM_OK;
        
#ifdef VMS
    if( pList == NULL ) return (CSSM_ERRCODE_INVALID_OUTPUT_POINTER);
#endif
    if ((*pList = (MDSUTIL_LIST_PTR)MDSUTIL_calloc (1, (uint32)sizeof (CSSM_LIST), NULL)) == NULL)
        return (CSSM_ERRCODE_MEMORY_ERROR);

    if (UsageMask & CSSM_SERVICE_CSSM) 

    {
        if ((retValue = mdsutil_ListCSSMModules (hDLDBCdsa, *pList, UsageMask, MatchAll)) != CSSM_OK) 
        {
            if ((*pList)->Items)
                MDSUTIL_free ((*pList)->Items, NULL);
            MDSUTIL_free (*pList, NULL);
            *pList = NULL;
            return (retValue);
        }
    }
    
    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_COMMON ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &CommonAttributes, 0, MDS_CDSADIR_COMMON_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = CommonAttributes;
    if ((retValue = common_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);

    retValue = CSSM_OK;
    for (;;)
    {
        if ((retValue = mdsutil_unconditionalGetRecord(hDLDBCdsa,
                                                       MDS_CDSADIR_COMMON_RECORDTYPE,
                                                       MDS_CDSADIR_COMMON_NUM_ATTRIBUTES,
                                                       &ResultsHandle,
                                                       CommonAttributes)) != CSSM_OK) 
        {
            if (CSSMERR_DL_ENDOFDATA == retValue)
            {   
                MDSFuncs.DataAbortQuery( hDLDBCdsa, ResultsHandle );
                return (CSSM_OK);
            }       
            if ((*pList)->Items)
                MDSUTIL_free ((*pList)->Items, NULL);
            MDSUTIL_free (*pList, NULL);    
            *pList = NULL;
            MDSFuncs.DataAbortQuery( hDLDBCdsa, ResultsHandle );            
            return (retValue);      
        }

        Record.guid = *CommonAttributes[0].Value;
        Record.Manifest = *CommonAttributes[1].Value;
        Record.ModuleName = *CommonAttributes[2].Value;
        Record.Path = *CommonAttributes[3].Value;
        Record.Version = *CommonAttributes[4].Value;
        Record.Desc = *CommonAttributes[5].Value;
        Record.Dynamic = *CommonAttributes[6].Value;
        Record.MultiThreadFlag = *CommonAttributes[7].Value;
        Record.ServiceMask = *CommonAttributes[8].Value;

        mdsutil_StringToVersion (Record.Version.Data, &Version);

        //if (Version.Major != CSSM_MAJOR || Version.Minor != CSSM_MINOR)
        //  continue;
        
        if (Record.ServiceMask.Data)
            Mask = *(CSSM_SERVICE_TYPE *)Record.ServiceMask.Data;

        if (MatchAll)
            MatchRes = (CSSM_BOOL)(Mask == UsageMask);
        else
            MatchRes = (CSSM_BOOL)(Mask & UsageMask);

        if (MatchRes) 
        {
            if ((retValue = mdsutil_GetSubServiceInfo (hDLDBCdsa, 
                                                       (char *)Record.guid.Data,
                                                       UsageMask & Mask, 
                                                       &SubServiceIds, 
                                                       &numSub)) != CSSM_OK) 
            {
                MDSUTIL_free ((*pList)->Items, NULL);
                MDSUTIL_free (*pList, NULL);
                *pList = NULL;
                mdsutil_FreeRecord (MDS_CDSADIR_COMMON_NUM_ATTRIBUTES, CommonAttributes);           
                MDSFuncs.DataAbortQuery( hDLDBCdsa, ResultsHandle );
                return (retValue);
            }

            for (i = 0; i < numSub; i++) 
            {
                if ((retValue = mdsutil_updateListEntry (*pList, Record, SubServiceIds[i])) != CSSM_OK) 
                {
                    MDSUTIL_free ((*pList)->Items, NULL);
                    MDSUTIL_free (*pList, NULL);
                    mdsutil_FreeRecord (MDS_CDSADIR_COMMON_NUM_ATTRIBUTES, CommonAttributes);               
                    MDSFuncs.DataAbortQuery( hDLDBCdsa, ResultsHandle );
                    *pList = NULL;
                    return (retValue);
                }
            }
        }

        if ( SubServiceIds != NULL )
        {
            MDSUTIL_free( SubServiceIds, NULL );
            SubServiceIds = NULL;
        }
        mdsutil_FreeRecord (MDS_CDSADIR_COMMON_NUM_ATTRIBUTES, CommonAttributes);
    }
}


CSSM_RETURN CSSMAPI mdsutil_GetCredLocationFromGUID(const CSSM_GUID *ModuleGUID,
                                                    CSSM_DATA *pModulePath, /* full path to module*/
                                                    CSSM_DATA *pModuleCredentialPath,
                                                    CSSM_API_MEMORY_FUNCS_PTR MemoryFuncs)/* Full path to ModuleCredential*/ 
{
    CSSM_DATA tmpCredential;
    CSSM_DATA tmpModuleName;
    CSSM_DATA tmpPath;
    CSSM_DATA ModulePath;
    CSSM_RETURN retValue = CSSM_OK;

#ifdef VMS
    if( pModulePath == NULL ) return (CSSM_ERRCODE_INVALID_OUTPUT_POINTER);
    if( pModuleCredentialPath == NULL ) return (CSSM_ERRCODE_INVALID_OUTPUT_POINTER);
#endif

    if ((retValue = mdsutil_GetModuleCredentialInfo(ModuleGUID,
                                                    &tmpCredential,
                                                    &tmpModuleName,
                                                    &tmpPath)) != CSSM_OK)
    {
        return (retValue);
    }

    if ((retValue = mdsutil_GetModulePath (tmpModuleName, tmpPath, &ModulePath)) != CSSM_OK)
        return (retValue);

    MDSUTIL_free (tmpModuleName.Data, NULL);
    MDSUTIL_free (tmpPath.Data, NULL);

    pModulePath->Length = ModulePath.Length;
    if (pModulePath->Length) {
        if ((pModulePath->Data = CSSM_MALLOC_WRAPPER(
                                        MemoryFuncs->malloc_func,
                                        ModulePath.Length,
                                        MemoryFuncs->AllocRef)) == NULL)
        {
            MDSUTIL_free (tmpCredential.Data, NULL);
            MDSUTIL_free (ModulePath.Data, NULL);
            return (CSSM_ERRCODE_MEMORY_ERROR);
        }

        memcpy (pModulePath->Data, ModulePath.Data, pModulePath->Length);
        MDSUTIL_free (ModulePath.Data, NULL);
    } else {
        pModulePath->Data = NULL;
    }

    if (NULL != pModuleCredentialPath) {
        pModuleCredentialPath->Length = tmpCredential.Length;
        if (pModuleCredentialPath->Length) {
            if ((pModuleCredentialPath->Data = CSSM_MALLOC_WRAPPER(
                                                    MemoryFuncs->malloc_func,
                                                    pModuleCredentialPath->Length,
                                                    MemoryFuncs->AllocRef)) == NULL)
            {
                MDSUTIL_free (tmpCredential.Data, NULL);
                CSSM_FREE_WRAPPER(MemoryFuncs->free_func,
                                  pModulePath->Data,
                                  MemoryFuncs->AllocRef);
                memset (pModulePath, 0, sizeof (CSSM_DATA));
                return (CSSM_ERRCODE_MEMORY_ERROR);
            }

            memcpy (pModuleCredentialPath->Data, tmpCredential.Data, pModuleCredentialPath->Length);
            MDSUTIL_free (tmpCredential.Data, NULL);
        } else {
            pModuleCredentialPath->Data = NULL;
        }
    }

    return (retValue);  
}

CSSM_RETURN CSSMAPI mdsutil_ListInstalledEmms(CSSM_GUID_PTR *ModuleManagerGuids,
                                              uint32* NumberOfModuleManagers)
{
    cssm_MDS_CDSADIR_EMM_RECORD Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA EmmAttributes[MDS_CDSADIR_EMM_NUM_ATTRIBUTES];
    CSSM_HANDLE ResultsHandle = 0;
    CSSM_GUID_PTR GuidPtr = NULL;
    CSSM_GUID_PTR tempptr = NULL;
    uint32 emmcount = 0;
    CSSM_RETURN retValue = CSSM_OK;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_EMM_RECORD ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &EmmAttributes, 0, MDS_CDSADIR_EMM_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = EmmAttributes;
    if ((retValue = emm_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);

    if ((GuidPtr = (CSSM_GUID_PTR)MDSUTIL_calloc (1, (uint32)sizeof (CSSM_GUID), NULL)) == NULL)
        return (CSSM_ERRCODE_MEMORY_ERROR);
    
    retValue = CSSM_OK;
    for (;;)
    {
        if ((retValue = mdsutil_unconditionalGetRecord(hDLDBCdsa,
                                                  MDS_CDSADIR_EMM_RECORDTYPE,
                                                  MDS_CDSADIR_EMM_NUM_ATTRIBUTES,
                                                  &ResultsHandle,
                                                  EmmAttributes)) != CSSM_OK)
        {
            if (CSSMERR_DL_ENDOFDATA == retValue)
            {
                *ModuleManagerGuids = GuidPtr;
                *NumberOfModuleManagers = emmcount;
                MDSFuncs.DataAbortQuery( hDLDBCdsa, ResultsHandle );
                return (CSSM_OK);
            }
            if( GuidPtr != NULL)
                MDSUTIL_free( GuidPtr, NULL);           
            MDSFuncs.DataAbortQuery( hDLDBCdsa, ResultsHandle );
            return (retValue);          
        }
        emmcount++;
        GuidPtr = (CSSM_GUID_PTR)MDSUTIL_realloc (GuidPtr, (uint32)(sizeof (CSSM_GUID)*emmcount), NULL);
        tempptr = GuidPtr + (sizeof(CSSM_GUID)*(emmcount-1));
        if(tempptr == NULL)
        {
            if( GuidPtr != NULL)
                MDSUTIL_free( GuidPtr, NULL);
            mdsutil_FreeRecord (MDS_CDSADIR_EMM_NUM_ATTRIBUTES, EmmAttributes);     
            MDSFuncs.DataAbortQuery( hDLDBCdsa, ResultsHandle );
            return (CSSM_ERRCODE_MEMORY_ERROR);
        }
        mdsutil_StrToGUID ((char *)EmmAttributes[0].Value->Data, tempptr);  
        mdsutil_FreeRecord (MDS_CDSADIR_EMM_NUM_ATTRIBUTES, EmmAttributes);
    }
}

CSSM_RETURN CSSMAPI mdsutil_get_emm_info(CSSM_DL_DB_HANDLE hDLDB,
                                         char* GuidStr,
                                         MDSUTIL_MODULE_MANAGER_INFO_PTR *pModuleManagerInfo)
{
    cssm_MDS_CDSADIR_EMM_RECORD Record;
    CSSM_DB_ATTRIBUTE_DATA EmmAttributes[MDS_CDSADIR_EMM_NUM_ATTRIBUTES];
    CSSM_RETURN retValue = CSSM_OK;

    *pModuleManagerInfo = NULL;
    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_EMM_RECORD ));
    memset( &EmmAttributes, 0, MDS_CDSADIR_EMM_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    if ((retValue = mdsutil_mds_GetEmmRecord(hDLDB,
                                             GuidStr,
                                             &Record,
                                             EmmAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    if ((*pModuleManagerInfo = (MDSUTIL_MODULE_MANAGER_INFO_PTR)MDSUTIL_calloc (1, (uint32)sizeof (MDSUTIL_MODULE_MANAGER_INFO), NULL)) == NULL) {
        mdsutil_FreeRecord (MDS_CDSADIR_EMM_NUM_ATTRIBUTES, EmmAttributes);
        return (CSSM_ERRCODE_MEMORY_ERROR);
    }
    
    mdsutil_StringToVersion (Record.CDSAVersion.Data, &(*pModuleManagerInfo)->CompatibleCSSMVersion);
    mdsutil_StringToVersion (Record.EMMSpecVersion.Data, &(*pModuleManagerInfo)->EMMSpecVersion);

    strncpy ((*pModuleManagerInfo)->Description, (const char *)Record.desc.Data, Record.desc.Length);
    strncpy ((*pModuleManagerInfo)->Vendor, (const char *)Record.emmvendor.Data, Record.emmvendor.Length);
    mdsutil_StringToVersion (Record.emmversion.Data, &(*pModuleManagerInfo)->EmmVersion);
    if (Record.emmtype.Data)
        memcpy(&(*pModuleManagerInfo)->ServiceType , Record.emmtype.Data, sizeof(uint32));

    mdsutil_FreeRecord (MDS_CDSADIR_EMM_NUM_ATTRIBUTES, EmmAttributes);
    
    return (retValue);
}


CSSM_RETURN CSSMAPI mdsutil_mds_get_modulemanagerInfo(const CSSM_GUID *ModuleGUID,
                                                      CSSM_SERVICE_MASK ServiceType,
                                                      MDSUTIL_MODULE_MANAGER_INFO_PTR *pInfo)
{
    CSSM_RETURN retValue = CSSM_OK;
    char GuidStr[MDS_UTIL_MAX_PATH];

    *pInfo = NULL;

    /*
     * Convert the Guid to a string.
     */
    mdsutil_GUIDToStr (ModuleGUID, GuidStr);
    
    if ((retValue = mdsutil_get_emm_info(hDLDBCdsa,
                                         GuidStr,
                                         pInfo)) != CSSM_OK)
    {
        return (retValue);
    }

    if( ServiceType != (*pInfo)->ServiceType)
    {
        MDSUTIL_free(*pInfo, NULL);
        return (CSSMERR_CSSM_INVALID_SERVICE_MASK);
    }
    
    return (retValue);
}

/*---------------------------------------------------------------
 *
 *Name: mdsutil_GetModuleLocationfromGuid
 *
 *Description:
 *   Get the data associated with the local label
 *
 *Parameters: 
 *   AddInType - Identifier for add-in type
 *   GUID - pointer to the guid for the add-in
 *   path - pointer to buffer to store the path to the add-in
 *
 *Returns:
 *   CSSM_FAIL - unable to get the path
 *   CSSM_OK - location to the add-in is retreive
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI mdsutil_GetModuleLocationfromGuid(const CSSM_GUID_PTR ModuleGUID,
                                                      char* path)
{
    CSSM_DATA tmpPath;
    CSSM_DATA tmpModuleName;
    CSSM_DATA ModulePath;
    CSSM_RETURN retValue = CSSM_OK;

    if ((retValue = mdsutil_GetModuleCredentialInfo(ModuleGUID,
                                                    NULL,
                                                    &tmpModuleName,
                                                    &tmpPath)) != CSSM_OK)
    {
        return (retValue);
    }

    if ((retValue = mdsutil_GetModulePath (tmpModuleName, tmpPath, &ModulePath)) != CSSM_OK)    
        return (retValue);

    if (ModulePath.Length)
        memcpy (path, ModulePath.Data, ModulePath.Length);
    else
        path = NULL;

    MDSUTIL_free (tmpModuleName.Data, NULL);
    MDSUTIL_free (tmpPath.Data, NULL);
    MDSUTIL_free (ModulePath.Data, NULL);

    return (retValue);
}

CSSM_RETURN CSSMAPI MDSUTIL_GetCredLocationFromGUID(const CSSM_GUID *ModuleGUID,
                                                    CSSM_DATA *pModulePath, /* full path to module */
                                                    CSSM_DATA *pModuleCredentialPath, /* Full path to ModuleCredential */
                                                    CSSM_API_MEMORY_FUNCS_PTR MemoryFuncs) 
{
    /* check for bad GUID */
    if (ModuleGUID == NULL ||
                IsBadReadPtr(ModuleGUID, (uint32)sizeof(CSSM_GUID)) == CSSM_TRUE)
        return (CSSMERR_CSSM_INVALID_GUID);

    /* check for bad Modulepath */
    if (pModulePath == NULL ||
                IsBadReadPtr(pModulePath, (uint32)sizeof(CSSM_DATA)) == CSSM_TRUE)
        return (CSSMERR_CSSM_MDS_ERROR);

    /* check for bad ModuleCredentialPath */
    if (pModuleCredentialPath == NULL ||
                IsBadReadPtr(pModuleCredentialPath, (uint32)sizeof(CSSM_DATA)) == CSSM_TRUE)
        return (CSSM_ERRCODE_MDS_ERROR);

    return (mdsutil_GetCredLocationFromGUID( ModuleGUID,pModulePath,
                                        pModuleCredentialPath,MemoryFuncs));
}

CSSM_RETURN CSSMAPI mdsutil_DeleteRecordByGuid(CSSM_DL_DB_HANDLE hDLDB,
                                               const CSSM_GUID *guid,
                                               CSSM_DB_RECORDTYPE RecordType)
{
    CSSM_QUERY Query;
    CSSM_SELECTION_PREDICATE Predicate;
    CSSM_DB_UNIQUE_RECORD_PTR RecordId = NULL;
    CSSM_RETURN retValue = CSSM_OK;
    CSSM_HANDLE ResultsHandle = 0;
    char GuidStr[MDS_UTIL_MAX_PATH];

    memset (&Query, 0, sizeof (CSSM_QUERY));
    memset (&Predicate, 0, sizeof (CSSM_SELECTION_PREDICATE));
    Query.RecordType = RecordType;
    Query.Conjunctive = CSSM_DB_NONE;
    Query.NumSelectionPredicates = 1;
    Query.SelectionPredicate = &Predicate;
    Query.QueryLimits.TimeLimit = CSSM_QUERY_TIMELIMIT_NONE;
    Query.QueryLimits.SizeLimit = CSSM_QUERY_SIZELIMIT_NONE;
    Query.QueryFlags = 0;

    Predicate.DbOperator = CSSM_DB_EQUAL;

    /*
     * Convert the Guid to a string.
     */
    mdsutil_GUIDToStr (guid, GuidStr);
    
    Predicate.Attribute.Info = s_MdsAttrInfo_ModuleId;
    Predicate.Attribute.Value = (CSSM_DATA *) MDSUTIL_malloc(sizeof (CSSM_DATA), 0);
    if (Predicate.Attribute.Value == NULL)
        return (CSSM_ERRCODE_MEMORY_ERROR);
    Predicate.Attribute.NumberOfValues = 1;
    Predicate.Attribute.Value->Data = (unsigned char *)GuidStr;
    Predicate.Attribute.Value->Length = (uint32)strlen (GuidStr) + 1;

    if( !IsBadCodePtr((CSSM_PROC_ADDR)MDSFuncs.DataGetFirst)) 
        retValue = MDSFuncs.DataGetFirst(hDLDB, /* DLDBHandle */
                                             &Query, 
                                             &ResultsHandle, 
                                             NULL, 
                                             NULL, 
                                             &RecordId);

    while (CSSM_OK == retValue)
    {
        if( !IsBadCodePtr((CSSM_PROC_ADDR)MDSFuncs.DataDelete)) 
            retValue = MDSFuncs.DataDelete(hDLDB, RecordId);

        if( !IsBadCodePtr((CSSM_PROC_ADDR)MDSFuncs.FreeUniqueRecord)) 
            MDSFuncs.FreeUniqueRecord(hDLDB, RecordId);
        
        if( !IsBadCodePtr((CSSM_PROC_ADDR)MDSFuncs.DataGetNext)) 
            retValue = MDSFuncs.DataGetNext(hDLDB, /* DLDBHandle */
                                            ResultsHandle, 
                                            NULL, 
                                            NULL, 
                                            &RecordId);
    }

    if (CSSMERR_DL_ENDOFDATA == retValue)
    {   
        retValue = CSSM_OK;
    }

    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
    MDSUTIL_free(Predicate.Attribute.Value, NULL);
    return (retValue);
}
CSSM_RETURN CSSMAPI mdsutil_ModuleManagerUninstall(const CSSM_GUID *Guid)
{
    CSSM_RETURN retValue = CSSM_OK;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA EmmAttributes[MDS_CDSADIR_EMM_NUM_ATTRIBUTES];
    cssm_MDS_CDSADIR_EMM_RECORD Record;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_EMM_RECORD ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &EmmAttributes, 0, MDS_CDSADIR_EMM_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = EmmAttributes;
    if ((retValue = emm_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);
    
    return (mdsutil_DeleteRecordByGuid(hDLDBCdsa,
                                       Guid,
                                       MDS_CDSADIR_EMM_RECORDTYPE));
}


CSSM_RETURN CSSMAPI mdsutil_delete_csp_record(CSSM_DL_DB_HANDLE hDLDB,
                                              const CSSM_GUID *ModuleGUID)
{
    CSSM_RETURN retValue = CSSM_OK;

    if ((retValue = mdsutil_DeleteRecordByGuid(hDLDB,
                                               ModuleGUID,
                                               MDS_CDSADIR_CSP_PRIMARY_RECORDTYPE))
                                               != CSSM_OK)
    {
        return (retValue);
    }

    if ((retValue = mdsutil_DeleteRecordByGuid(hDLDB,
                                               ModuleGUID,
                                               MDS_CDSADIR_CSP_CAPABILITY_RECORDTYPE))
                                               != CSSM_OK)
    {
        return (retValue);
    }

    if ((retValue = mdsutil_DeleteRecordByGuid(hDLDB,
                                               ModuleGUID,
                                               MDS_CDSADIR_CSP_ENCAPSULATED_PRODUCT_RECORDTYPE))
                                               != CSSM_OK)
    {
        return (retValue);
    }

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_delete_dl_record(CSSM_DL_DB_HANDLE hDLDB,
                                             const CSSM_GUID *ModuleGUID)
{
    CSSM_RETURN retValue = CSSM_OK;

    if ((retValue = mdsutil_DeleteRecordByGuid(hDLDB,
                                               ModuleGUID,
                                               MDS_CDSADIR_DL_PRIMARY_RECORDTYPE)) != CSSM_OK)
    {
        return (retValue);
    }

    if ((retValue = mdsutil_DeleteRecordByGuid(hDLDB,
                                               ModuleGUID,
                                               MDS_CDSADIR_DL_ENCAPSULATED_PRODUCT_RECORDTYPE))
                                               != CSSM_OK)
    {
        return (retValue);
    }

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_delete_cl_record(CSSM_DL_DB_HANDLE hDLDB,
                                             const CSSM_GUID *ModuleGUID)
{
    CSSM_RETURN retValue = CSSM_OK;

    if ((retValue = mdsutil_DeleteRecordByGuid(hDLDB,
                                               ModuleGUID,
                                               MDS_CDSADIR_CL_PRIMARY_RECORDTYPE))
                                               != CSSM_OK)
    {
        return (retValue);
    }

    if ((retValue = mdsutil_DeleteRecordByGuid(hDLDB,
                                               ModuleGUID,
                                               MDS_CDSADIR_CL_ENCAPSULATED_PRODUCT_RECORDTYPE))
                                               != CSSM_OK)
    {
        return (retValue);
    }

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_delete_tp_record(CSSM_DL_DB_HANDLE hDLDB,
                                             const CSSM_GUID *ModuleGUID)
{
    CSSM_RETURN retValue = CSSM_OK;

    if ((retValue = mdsutil_DeleteRecordByGuid(hDLDB,
                                               ModuleGUID,
                                               MDS_CDSADIR_TP_PRIMARY_RECORDTYPE))
                                               != CSSM_OK)
    {
        return (retValue);
    }

    if ((retValue = mdsutil_DeleteRecordByGuid(hDLDB,
                                               ModuleGUID,
                                               MDS_CDSADIR_TP_OIDS_RECORDTYPE))
                                               != CSSM_OK)
    {
        return (retValue);
    }

    if ((retValue = mdsutil_DeleteRecordByGuid(hDLDB,
                                               ModuleGUID,
                                               MDS_CDSADIR_TP_ENCAPSULATED_PRODUCT_RECORDTYPE))
                                               != CSSM_OK)
    {
        return (retValue);
    }

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_ModuleUninstall(const CSSM_GUID *ModuleGUID)
{
    cssm_MDS_CDSADIR_COMMON Record;
    CSSM_DB_ATTRIBUTE_DATA CommonAttributes[MDS_CDSADIR_COMMON_NUM_ATTRIBUTES];
    CSSM_SERVICE_MASK mask;
    CSSM_RETURN retValue = CSSM_OK;
    char GuidStr[MDS_UTIL_MAX_PATH];

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_COMMON ));
    memset( &CommonAttributes, 0, MDS_CDSADIR_COMMON_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    /*
     * Convert the Guid to a string.
     */
    mdsutil_GUIDToStr(ModuleGUID, GuidStr);

    if ((retValue = mdsutil_GetCommonRecord(hDLDBCdsa,
                                            GuidStr,
                                            &Record,
                                            CommonAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    mask = *(CSSM_SERVICE_MASK *)Record.ServiceMask.Data;

    mdsutil_FreeRecord (MDS_CDSADIR_COMMON_NUM_ATTRIBUTES, CommonAttributes);

    while (mask) {
        if (mask & CSSM_SERVICE_CSP) {
            if ((retValue = mdsutil_delete_csp_record (hDLDBCdsa, ModuleGUID)) != CSSM_OK)
                return (retValue);

            mask &= ~CSSM_SERVICE_CSP;

        } else if (mask & CSSM_SERVICE_TP) {
            if ((retValue = mdsutil_delete_tp_record (hDLDBCdsa, ModuleGUID)) != CSSM_OK)
                return (retValue);

            mask &= ~CSSM_SERVICE_TP;

        } else if (mask & CSSM_SERVICE_CL) {
                if ((retValue = mdsutil_delete_cl_record (hDLDBCdsa, ModuleGUID)) != CSSM_OK)
                return (retValue);

            mask &= ~CSSM_SERVICE_CL;

        } else if (mask & CSSM_SERVICE_DL) {
            if ((retValue = mdsutil_delete_dl_record (hDLDBCdsa, ModuleGUID)) != CSSM_OK)
                return (retValue);
        
            mask &= ~CSSM_SERVICE_DL;
        }
        else
            break;
    }
    
    return mdsutil_DeleteRecordByGuid(hDLDBCdsa,
                                          ModuleGUID,
                                          MDS_CDSADIR_COMMON_RECORDTYPE);
}

/*---------------------------------------------------------------
 *
 *Name: mdsutil_ReadUsage
 *
 *Description:
 *     Function to read GUID usage mask
 *
 *Parameters:
 *     GUID (input) - the GUID that identifies the addin
 *
 *Returns:
 *     Usage mask
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI mdsutil_RegReadUsage(const CSSM_GUID *ModuleGUID,
                                         CSSM_SERVICE_MASK *ServiceMask)
{
    CSSM_RETURN retValue = CSSM_OK;
    cssm_MDS_CDSADIR_COMMON Record;
    CSSM_DB_ATTRIBUTE_DATA CommonAttributes[MDS_CDSADIR_COMMON_NUM_ATTRIBUTES];
    char GuidStr[MDS_UTIL_MAX_PATH];

    *ServiceMask = 0;
    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_COMMON ));
    memset( &CommonAttributes, 0, MDS_CDSADIR_COMMON_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));
    
    /*
     * Convert the Guid to a string.
     */
    mdsutil_GUIDToStr (ModuleGUID, GuidStr);

    if ((retValue = mdsutil_GetCommonRecord(hDLDBCdsa,
                                            GuidStr,
                                            &Record,
                                            CommonAttributes)) == CSSM_OK)
    {
        if(Record.ServiceMask.Length != 0 )
        {
            memcpy(ServiceMask, Record.ServiceMask.Data, Record.ServiceMask.Length);
        }
        mdsutil_FreeRecord (MDS_CDSADIR_COMMON_NUM_ATTRIBUTES, CommonAttributes);
    }
    return ( retValue );
}

CSSM_RETURN CSSMAPI mdsutil_RegGetThreadSafe(const CSSM_GUID *ModuleGUID,
                                             CSSM_BOOL *ThreadSafe)
{
    CSSM_RETURN retValue = CSSM_OK;
    cssm_MDS_CDSADIR_COMMON Record;
    CSSM_DB_ATTRIBUTE_DATA CommonAttributes[MDS_CDSADIR_COMMON_NUM_ATTRIBUTES];
    char GuidStr[MDS_UTIL_MAX_PATH];
	uint32 threadData;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_COMMON ));
    memset( &CommonAttributes, 0, MDS_CDSADIR_COMMON_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));
    
    /*
     * Convert the Guid to a string.
     */
    mdsutil_GUIDToStr(ModuleGUID, GuidStr);

    if ((retValue = mdsutil_GetCommonRecord(hDLDBCdsa,
                                            GuidStr,
                                            &Record,
                                            CommonAttributes)) == CSSM_OK)
    {
		memcpy(&threadData, Record.MultiThreadFlag.Data, sizeof(uint32));
        if (threadData != 0)
        {
            *ThreadSafe = CSSM_FALSE;
        }
        else
        {
            *ThreadSafe = CSSM_TRUE;
        }

        mdsutil_FreeRecord(MDS_CDSADIR_COMMON_NUM_ATTRIBUTES,
                           CommonAttributes);
    }
    
    return ( retValue );
}

CSSM_RETURN CSSMAPI mdsutil_GetEmmRecordbyType(CSSM_DL_DB_HANDLE hDLDB,
                                               const CSSM_SERVICE_MASK EMMType,
                                               CSSM_GUID_PTR pManagerGUID,
                                               CSSM_DATA_PTR EMMCredential,
                                               CSSM_DATA_PTR EMMModuleName,
                                               CSSM_DATA_PTR EMMPath)
{
    cssm_MDS_CDSADIR_EMM_RECORD Record;
    CSSM_DB_RECORD_ATTRIBUTE_DATA Attributes;
    CSSM_DB_ATTRIBUTE_DATA EmmAttributes[MDS_CDSADIR_EMM_NUM_ATTRIBUTES];
    CSSM_HANDLE ResultsHandle = 0;
    CSSM_SERVICE_MASK emmtype = 0;
    CSSM_RETURN retValue = CSSM_OK;
    CSSM_BOOL bFound = CSSM_FALSE;

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_EMM_RECORD ));
    memset( &Attributes, 0, sizeof(CSSM_DB_RECORD_ATTRIBUTE_DATA));
    memset( &EmmAttributes, 0, MDS_CDSADIR_EMM_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    Attributes.AttributeData = EmmAttributes;
    if ((retValue = emm_schema_GetAttributes (&Record, &Attributes)) != CSSM_OK)
        return (retValue);

    for (;;)
    {
        if ((retValue = mdsutil_unconditionalGetRecord(hDLDB,
                                                  MDS_CDSADIR_EMM_RECORDTYPE,
                                                  MDS_CDSADIR_EMM_NUM_ATTRIBUTES,
                                                  &ResultsHandle,
                                                  EmmAttributes)) != CSSM_OK)
        {           
            MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
            return (retValue);
        }
        if(EmmAttributes[10].Value->Length != 0) {
            memcpy( &emmtype, EmmAttributes[10].Value->Data, EmmAttributes[10].Value->Length);
            if( emmtype == EMMType)
            {
                bFound = CSSM_TRUE;
                break;
            }
            else
                mdsutil_FreeRecord (MDS_CDSADIR_EMM_NUM_ATTRIBUTES, EmmAttributes);
        }
    }

    MDSFuncs.DataAbortQuery( hDLDB, ResultsHandle );
    if (NULL != pManagerGUID)
        mdsutil_StrToGUID ((char *)EmmAttributes[0].Value->Data, pManagerGUID);

    if (NULL != EMMPath) {
        EMMPath->Length = EmmAttributes[3].Value->Length;
        if (EMMPath->Length) {
            if ((EMMPath->Data = (uint8*)MDSUTIL_malloc (EMMPath->Length, NULL)) == NULL)
                return (CSSM_ERRCODE_MEMORY_ERROR);
            memcpy(EMMPath->Data, EmmAttributes[3].Value->Data, EmmAttributes[3].Value->Length);
        } else {
            EMMPath->Data = NULL;
        }
    }
    if (NULL != EMMModuleName) {
        EMMModuleName->Length = EmmAttributes[2].Value->Length;
        if (EMMModuleName->Length) {
            if ((EMMModuleName->Data = (uint8*)MDSUTIL_malloc (EMMModuleName->Length, NULL)) == NULL)
                return (CSSM_ERRCODE_MEMORY_ERROR);
            memcpy(EMMModuleName->Data, EmmAttributes[2].Value->Data, EmmAttributes[2].Value->Length);
        } else {
            EMMModuleName->Data = NULL;
        }
    }

    if (NULL != EMMCredential) {
        EMMCredential->Length = EmmAttributes[1].Value->Length;
        if (EMMCredential->Length) {
            if ((EMMCredential->Data = (uint8*)MDSUTIL_malloc (EMMCredential->Length, NULL)) == NULL)
                return (CSSM_ERRCODE_MEMORY_ERROR);
            memcpy(EMMCredential->Data, EmmAttributes[1].Value->Data, EmmAttributes[1].Value->Length);
        } else {
            EMMCredential->Data = NULL;
        }
    }

    mdsutil_FreeRecord (MDS_CDSADIR_EMM_NUM_ATTRIBUTES, EmmAttributes);
    return (CSSM_OK);
}

/*---------------------------------------------------------------
 *
 *Name: mdsutil_ReadEMMInfo
 *
 *Description:
 *     Function to read emm info based on type 
 *
 *Parameters:
 *     GUID (input) - the GUID that identifies the emm
 *
 *Returns:
 *     Usage mask
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI mdsutil_RegGetEMMInfo(const CSSM_SERVICE_MASK EMMType,
                                          CSSM_GUID_PTR pManagerGUID,
                                          CSSM_DATA_PTR EMMCredential,
                                          CSSM_DATA_PTR EMMModuleName,
                                          CSSM_DATA_PTR EMMPath)
{
    CSSM_RETURN retValue = CSSM_OK;

    retValue = mdsutil_GetEmmRecordbyType(hDLDBCdsa,
                                          EMMType,
                                          pManagerGUID,
                                          EMMCredential,
                                          EMMModuleName,
                                          EMMPath);

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_GetCommonCredentialInfo(const CSSM_GUID *guid,
                                                    CSSM_DATA_PTR Credential,
                                                    CSSM_DATA_PTR ModuleName,
                                                    CSSM_DATA_PTR ModulePath)
{
    cssm_MDS_CDSADIR_COMMON Record;
    CSSM_DB_ATTRIBUTE_DATA CommonAttributes[MDS_CDSADIR_COMMON_NUM_ATTRIBUTES];
    CSSM_RETURN retValue = CSSM_OK;
    char GuidStr[MDS_UTIL_MAX_PATH];

    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_COMMON ));
    memset( &CommonAttributes, 0, MDS_CDSADIR_COMMON_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));
    
    /*
     * Convert the Guid to a string.
     */
    mdsutil_GUIDToStr(guid, GuidStr);

    if ((retValue = mdsutil_GetCommonRecord(hDLDBCdsa,
                                            GuidStr,
                                            &Record,
                                            CommonAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    if (Credential)
    {
        if ((Credential->Data = (uint8 *)MDSUTIL_calloc (1, Record.Manifest.Length, NULL)) == NULL)
        {
            mdsutil_FreeRecord (MDS_CDSADIR_COMMON_NUM_ATTRIBUTES, CommonAttributes);
            return (CSSM_ERRCODE_MEMORY_ERROR);
        }

        memcpy (Credential->Data, Record.Manifest.Data, Record.Manifest.Length);
        Credential->Length = Record.Manifest.Length;
    }

    if (ModuleName)
    {
        if ((ModuleName->Data = (uint8 *)MDSUTIL_calloc (1, Record.ModuleName.Length, NULL)) == NULL)
        {
            mdsutil_FreeRecord (MDS_CDSADIR_COMMON_NUM_ATTRIBUTES, CommonAttributes);
            return (CSSM_ERRCODE_MEMORY_ERROR);
        }

        memcpy (ModuleName->Data, Record.ModuleName.Data, Record.ModuleName.Length);
        ModuleName->Length = Record.ModuleName.Length;
    }

    if (ModulePath)
    {
        if ((ModulePath->Data = (uint8 *)MDSUTIL_calloc (1, Record.Path.Length, NULL)) == NULL)
        {
            mdsutil_FreeRecord (MDS_CDSADIR_COMMON_NUM_ATTRIBUTES, CommonAttributes);
            return (CSSM_ERRCODE_MEMORY_ERROR);
        }

        memcpy (ModulePath->Data, Record.Path.Data, Record.Path.Length);
        ModulePath->Length = Record.Path.Length;
    }
    
    mdsutil_FreeRecord (MDS_CDSADIR_COMMON_NUM_ATTRIBUTES, CommonAttributes);

    return (retValue);
}

CSSM_RETURN CSSMAPI mdsutil_GetEmmCredentialInfo(char* GuidStr,
                                                 CSSM_DATA_PTR Credential,
                                                 CSSM_DATA_PTR ModuleName,
                                                 CSSM_DATA_PTR ModulePath)
{
    cssm_MDS_CDSADIR_EMM_RECORD Record;
    CSSM_DB_ATTRIBUTE_DATA EmmAttributes[MDS_CDSADIR_EMM_NUM_ATTRIBUTES];
    CSSM_RETURN retValue = CSSM_OK;
        
    memset( &Record, 0, sizeof( cssm_MDS_CDSADIR_EMM_RECORD ));
    memset( &EmmAttributes, 0, MDS_CDSADIR_EMM_NUM_ATTRIBUTES * sizeof( CSSM_DB_ATTRIBUTE_DATA));

    if ((retValue = mdsutil_mds_GetEmmRecord(hDLDBCdsa,
                                             GuidStr,
                                             &Record,
                                             EmmAttributes)) != CSSM_OK)
    {
        return (retValue);
    }

    if (Credential)
    {
        if ((Credential->Data = (uint8 *)MDSUTIL_calloc (1, Record.manifest.Length, NULL)) == NULL)
        {
            mdsutil_FreeRecord (MDS_CDSADIR_EMM_NUM_ATTRIBUTES, EmmAttributes);
            return (CSSM_ERRCODE_MEMORY_ERROR);
        }

        memcpy (Credential->Data, Record.manifest.Data, Record.manifest.Length);
        Credential->Length = Record.manifest.Length;
    }

    if (ModuleName)
    {
        if ((ModuleName->Data = (uint8 *)MDSUTIL_calloc (1, Record.modulename.Length, NULL)) == NULL)
        {
            mdsutil_FreeRecord (MDS_CDSADIR_EMM_NUM_ATTRIBUTES, EmmAttributes);
            return (CSSM_ERRCODE_MEMORY_ERROR);
        }

        memcpy (ModuleName->Data, Record.modulename.Data, Record.modulename.Length);
        ModuleName->Length = Record.modulename.Length;
    }

    if (ModulePath)
    {
        if ((ModulePath->Data = (uint8 *)MDSUTIL_calloc (1, Record.path.Length, NULL)) == NULL)
        {
            mdsutil_FreeRecord (MDS_CDSADIR_EMM_NUM_ATTRIBUTES, EmmAttributes);
            return (CSSM_ERRCODE_MEMORY_ERROR);
        }

        memcpy (ModulePath->Data, Record.path.Data, Record.path.Length);
        ModulePath->Length = Record.path.Length;
    }
    
    mdsutil_FreeRecord (MDS_CDSADIR_EMM_NUM_ATTRIBUTES, EmmAttributes);

    return (retValue);
}


CSSM_RETURN CSSMAPI mdsutil_GetModuleCredentialInfo(const CSSM_GUID *guid,
                                                    CSSM_DATA_PTR Credential,
                                                    CSSM_DATA_PTR ModuleName,
                                                    CSSM_DATA_PTR ModulePath)
{
    char GuidStr[MDS_UTIL_MAX_PATH];

    /*
     * Convert the Guid to a string.
     */
    mdsutil_GUIDToStr(guid, GuidStr);

    if (mdsutil_GetObjectCredentialInfo(GuidStr,
                                        Credential,
                                        ModuleName,
                                        ModulePath) == CSSM_OK)
    {
        return (CSSM_OK);
    }

    if (mdsutil_GetCommonCredentialInfo(guid,
                                        Credential,
                                        ModuleName,
                                        ModulePath) == CSSM_OK)
    {
        return (CSSM_OK);
    }

    return (mdsutil_GetEmmCredentialInfo(GuidStr,
                                         Credential,
                                         ModuleName,
                                         ModulePath));
}


CSSM_RETURN CSSMAPI 
mdsutil_GetModulePath (CSSM_DATA ModuleName, CSSM_DATA ModuleSearchPath, CSSM_DATA_PTR ModulePath)
{
    uint8 *SearchPtr;
    uint8 *ptr;
    uint32 len;
    CSSM_DATA tmpPath;
    struct _stat st;

    len = ModuleSearchPath.Length + ModuleName.Length + 2; 
    if ((ptr = (uint8*)MDSUTIL_malloc (len, NULL)) == NULL)
        return (CSSM_ERRCODE_MEMORY_ERROR);

    SearchPtr = ModuleSearchPath.Data;
    tmpPath.Data = ModuleSearchPath.Data;
    tmpPath.Length = ModuleSearchPath.Length;

    while (SearchPtr) {
        SearchPtr = memchr (tmpPath.Data, PATH_SEPARATOR, tmpPath.Length);
        if (SearchPtr != NULL) {
            tmpPath.Length = (uint32)(SearchPtr - tmpPath.Data);
            SearchPtr++;            /* Skip the PATH_SEPARATOR */
        } else {    
            /*
             * get rid of the null character at the end of search path, if any.
             */
            if (tmpPath.Data[tmpPath.Length-1] == '\0')
                tmpPath.Length--;
        }
        
        ModulePath->Data = ptr;
        memcpy(ptr, tmpPath.Data, tmpPath.Length);
        ptr = ptr + tmpPath.Length;
#ifdef VMS
        if( *(ptr-1) != DIRECTORY_SEPARATOR )
#endif
            *ptr++ = DIRECTORY_SEPARATOR;
        memcpy(ptr, ModuleName.Data, ModuleName.Length);   /* copy relative path */
        ptr = ptr + ModuleName.Length;                     /* move to end of string */
        *ptr = '\0';                                       /* null terminate */
        ModulePath->Length = (uint32)(ptr - ModulePath->Data);
        /* 
         * stat the file and find out if we have found it.
         * If yes, then return else continue searching the
         * search path.
         */
        if ((_stat((const char *)ModulePath->Data, &st)) == 0)
            return (CSSM_OK);

        tmpPath.Data = SearchPtr;
        if(SearchPtr != NULL)
            tmpPath.Length = (uint32)(ModuleSearchPath.Length - (uint32)(SearchPtr - ModuleSearchPath.Data));
    }

    return (CSSM_ERRCODE_MDS_ERROR);
}

/*---------------------------------------------------------------
 *
 *Name: MDSUTIL_ListModules
 *
 *Description:
 *   List the registered addin modules
 *
 *Parameters: 
 *   UsageMask - mask to describe the usage of addin
 *   MatchAll - flag to indicate if addin has to match all 
 *              the conditions express in UsageMask
 *
 *Returns:
 *   NULL - unable to locate modules that match mask or an error
 *   not NULL - list of modules that match mask
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI MDSUTIL_ListModules (CSSM_SERVICE_MASK UsageMask,
                                         CSSM_BOOL MatchAll, 
                                         MDSUTIL_LIST_PTR *pList)
{
    CSSM_RETURN retValue = CSSM_OK;
    CSSM_SERVICE_MASK validMask = CSSM_SERVICE_CSSM | 
                                  CSSM_SERVICE_CSP  | 
                                  CSSM_SERVICE_DL   | 
                                  CSSM_SERVICE_CL   |
                                  CSSM_SERVICE_TP   |
#ifdef KEY_RECOVERY
                                  CSSM_SERVICE_KR   |
#endif
                                  CSSM_SERVICE_AC;

    validMask = ~validMask;

    /* check for invalid UsageMask */
    if (UsageMask & validMask)
        return (CSSM_ERRCODE_MDS_ERROR);

    if (pList == NULL)
        return (CSSMERR_CSSM_INVALID_POINTER);

    *pList = NULL;

    /* make sure that app has done a CSSM_Init */
    if ((retValue = mdsutil_CheckInit ()) != CSSM_OK)
        return(retValue);

    return (mdsutil_ListModules (UsageMask, MatchAll, pList));
}

/*---------------------------------------------------------------
 *
 *Name: MDSUTIL_GetModuleInfo
 *
 *Description:
 *   Get from the registry module info of the addin
 *
 *Parameters: 
 *   ModuleGUID - GUID of the addin module
 *   UsageMask - the type of service
 *   SubserviceID - internal ID of service
 *
 *Returns:
 *   NULL - error in retreiving info from the registry
 *   not NULL - pointer to module info
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI MDSUTIL_GetModuleInfo(const CSSM_GUID *ModuleGUID,
                                          CSSM_SERVICE_MASK UsageMask,
                                          uint32 SubserviceID,
                                          CSSM_USEE_TAG USEERequest,
                                          MDSUTIL_MODULE_INFO_PTR *pModuleInfo)
{
    CSSM_RETURN retValue = CSSM_OK;
    CSSM_SERVICE_MASK validMask = CSSM_SERVICE_CSSM | 
                                  CSSM_SERVICE_CSP  | 
                                  CSSM_SERVICE_DL   | 
                                  CSSM_SERVICE_CL   |
                                  CSSM_SERVICE_TP   |
#ifdef KEY_RECOVERY
                                  CSSM_SERVICE_KR   |
#endif
                                  CSSM_SERVICE_AC;

    validMask = ~validMask;

    if (pModuleInfo == NULL ||
                IsBadReadPtr(pModuleInfo, (uint32)sizeof(MDSUTIL_MODULE_INFO)) == CSSM_TRUE)
        return (CSSMERR_CSSM_INVALID_POINTER);

    *pModuleInfo = NULL;

    /* check for bad GUID */
    if (ModuleGUID == NULL ||
                IsBadReadPtr(ModuleGUID, (uint32)sizeof(CSSM_GUID)) == CSSM_TRUE)
        return (CSSMERR_CSSM_INVALID_GUID);

    /* check for invalid UsageMask */
    if ((UsageMask == 0) || (UsageMask & validMask))
        return (CSSM_ERRCODE_MDS_ERROR);

    /* make sure that app as done a CSSM_Init */
    if ((retValue = mdsutil_CheckInit ()) != CSSM_OK)
        return(retValue);

    return(mdsutil_mds_get_moduleInfo(ModuleGUID,
                                      UsageMask,
                                      SubserviceID,
                                      USEERequest,
                                      pModuleInfo));
}

/*---------------------------------------------------------------
 *
 *Name: mdsutil_FreeTPSubService
 *
 *Description:
 *   Free memory associated with the TPINFO structure
 *
 *Parameters: 
 *   Info (input) - pointer to TP Info
 *
 *Returns:
 *   CSSM_FAIL - unable to free
 *   CSSM_OK - free memory
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI mdsutil_FreeTPSubService (MDSUTIL_TPSUBSERVICE_PTR Sub)
{
    uint32 cnt;

    if (IsBadReadPtr (Sub, sizeof (MDSUTIL_TPSUBSERVICE)))
        return (CSSMERR_CSSM_INVALID_POINTER);

    /* Free PolicyIdentifiers memory */
    for (cnt = 0; cnt < Sub->NumberOfPolicyIdentifiers; cnt++) {
        CSSM_OID_PTR ptr;
    
        if (IsBadReadPtr (&Sub->PolicyIdentifiers[cnt], sizeof (CSSM_FIELD)))
            return (CSSMERR_CSSM_INVALID_POINTER);

        ptr = &Sub->PolicyIdentifiers[cnt].FieldOid;
        if (IsBadReadPtr (ptr->Data, ptr->Length))
            return (CSSMERR_CSSM_INVALID_POINTER);

        MDSUTIL_free (ptr->Data, NULL);

        ptr = &Sub->PolicyIdentifiers[cnt].FieldValue;
        if (IsBadReadPtr (ptr->Data, ptr->Length))
            return (CSSMERR_CSSM_INVALID_POINTER);

        MDSUTIL_free (ptr->Data, NULL);
    }
    
    /* Free Array of OID memory */
    if (Sub->NumberOfPolicyIdentifiers > 0)
        if (!IsBadReadPtr (Sub->PolicyIdentifiers, sizeof (CSSM_FIELD) * Sub->NumberOfPolicyIdentifiers)) {
            MDSUTIL_free (Sub->PolicyIdentifiers, NULL);
        }
 
    return CSSM_OK;
}

/*---------------------------------------------------------------
 *
 *Name: mdsutil_FreeSWInfo
 *
 *Description:
 *   Free memory associated with the SOFTWARE_SUBSERVICEINFO structure
 *
 *Parameters: 
 *   Info (input) - pointer to SOFTWARE Info
 *
 *Returns:
 *   CSSM_FAIL - unable to free
 *   CSSM_OK - free memory
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI mdsutil_FreeSWInfo (MDSUTIL_SOFTWARE_CSPSUBSERVICE_INFO_PTR Info)
{
    uint32 i;

    for (i = 0; i < Info->NumberOfCapabilities; i++)
    {
        CSSM_CONTEXT_PTR pCon = &Info->CapabilityList[i];
        mdsutil_FreeAttributes(pCon->NumberOfAttributes,
                               pCon->ContextAttributes,
                               _MDSUTIL_free,
                               NULL,
                               1);
    }

    /* Free the pointer to capabilities list */
    MDSUTIL_free(Info->CapabilityList, NULL);

    return CSSM_OK;
}

/*---------------------------------------------------------------
 *
 *Name: mdsutil_FreeAttribute
 *
 *Description:
 *   Free memory that was allocated for attributes
 *
 *Parameters: 
 *   NumberOfAttributes (input) - count of attributes to free
 *   Attributes (input) - pointer to the attributes info
 *   MDSUTIL_free (input) - free function to deallocate memory
 *   Free (input) - flag to indicate if the top level pointer
 *                  needs to be freed
 *
 *Returns:
 *   none
 *
 *-------------------------------------------------------------*/
void CSSMAPI mdsutil_FreeAttributes (uint32 NumberOfAttributes, 
                                  CSSM_CONTEXT_ATTRIBUTE_PTR Attributes,
                                  CSSM_FREE MDSUTIL_free, void *mem_ref, 
                                  uint32 Free)
{
    sint16 i;

    for (i=0; i < (sint16)NumberOfAttributes; i++) 
    {
        /* make sure there is something that needs freeing */
        if ((Attributes[i].Attribute.Data != NULL) && ((Attributes[i].AttributeType & CSSM_ATTRIBUTE_TYPE_MASK) 
                != CSSM_ATTRIBUTE_DATA_UINT32))
        {
            /* for every attribute type, the info is different, free stuff INSIDE the structure */
            /* DO NOT FREE WHAT THE POINTER ACTUALLY POINTS TO */
            switch (Attributes[i].AttributeType & CSSM_ATTRIBUTE_TYPE_MASK)
            {
                case CSSM_ATTRIBUTE_DATA_CRYPTO_DATA: 
                    {
                        /* These types use a CSSM_CRYPTO_DATA structure */
                        CSSM_CRYPTO_DATA_PTR q = Attributes[i].Attribute.CryptoData;
                        if ((q->Param.Data) && (!IsBadReadPtr (q->Param.Data, q->Param.Length)))
                            MDSUTIL_free (q->Param.Data, mem_ref);
                    }
                    break;

                case CSSM_ATTRIBUTE_DATA_KEY : 
                    {
                        /* These types use a CSSM_DATA structure */
                        CSSM_KEY_PTR q = Attributes[i].Attribute.Key;
                        if (q->KeyData.Data) {
                            MDSUTIL_free (q->KeyData.Data, mem_ref);
                        }
                    }
                    break;

                case CSSM_ATTRIBUTE_DATA_CSSM_DATA:
                    if (Attributes[i].Attribute.Data->Data != NULL)
                        MDSUTIL_free(Attributes[i].Attribute.Data->Data, NULL);
                    break;

                default:
                    break;
            } /* end switch */

            /* now free the structure that pointer actually points to */
            MDSUTIL_free(Attributes[i].Attribute.Data, mem_ref);
        } /* endif */
    } /* endfor */

    /* check for need to free the top most pointer */
    if (Free && Attributes)
        MDSUTIL_free (Attributes, mem_ref);
    return;
}

/*---------------------------------------------------------------
 *
 *Name:  mdsutil_FreeHardwareInfo
 *
 *Description:
 *   Free memory associated with the hradware structure
 *
 *Parameters: 
 *   Info (input) - pointer to hradware Info
 *
 *Returns:
 *   CSSM_FAIL - unable to free
 *   CSSM_OK - free memory
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI mdsutil_FreeHardwareInfo (MDSUTIL_HARDWARE_CSPSUBSERVICE_INFO_PTR Info)
{
    CSSM_RETURN retValue = CSSM_OK;

    if (!IsBadReadPtr (Info->UserCACertificate.Data, Info->UserCACertificate.Length)) {
        MDSUTIL_free (Info->UserCACertificate.Data, NULL);
    }

    /* free the common info structure */
    if ((retValue = mdsutil_FreeSWInfo ((MDSUTIL_SOFTWARE_CSPSUBSERVICE_INFO_PTR)Info)) != CSSM_OK)
        return (retValue);

    return (retValue);
}

/*---------------------------------------------------------------
 *
 *Name: mdsutil_FreeCSPWrapped
 *
 *Description:
 *   Free memory associated with the CSP_WRAPPEDPRODUCT_INFO structure
 *
 *Parameters: 
 *   Info (input) - pointer to CSP_WRAPPEDPRODUCT_INFO Info
 *
 *Returns:
 *   CSSM_FAIL - unable to free
 *   CSSM_OK - free memory
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI mdsutil_FreeCSPWrapped (MDSUTIL_CSP_WRAPPEDPRODUCTINFO_PTR Info)
{
/*  MDSUTIL_CSP_WRAPPEDPRODUCTINFO_PTR refInfo = Info;/* to remove level 4 warning*/

    return CSSM_OK;
}

/*---------------------------------------------------------------
 *
 *Name: mdsutil_FreeCSPSubService
 *
 *Description:
 *   Free memory associated with the CSPINFO structure
 *
 *Parameters: 
 *   Info (input) - pointer to CSP Info
 *
 *Returns:
 *   CSSM_FAIL - unable to free
 *   CSSM_OK - free memory
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI mdsutil_FreeCSPSubService (MDSUTIL_CSPSUBSERVICE_PTR Sub)
{
    CSSM_RETURN retValue = CSSM_OK;

    if (IsBadReadPtr (Sub, sizeof (MDSUTIL_CSPSUBSERVICE)))
        return (CSSMERR_CSSM_INVALID_POINTER);
        
    if (Sub->CspType == CSSM_CSP_SOFTWARE) {
        if ((retValue = mdsutil_FreeSWInfo (&Sub->SubServiceInfo.SoftwareCspSubService)) != CSSM_OK)
            return (retValue);
    } else if (Sub->CspType == CSSM_CSP_HARDWARE) {
        if ((retValue = mdsutil_FreeHardwareInfo (&Sub->SubServiceInfo.HardwareCspSubService)) != CSSM_OK)
            return (retValue);
    } else if (Sub->CspType == CSSM_CSP_HYBRID) {
        if ((retValue = mdsutil_FreeHardwareInfo (&Sub->SubServiceInfo.HybridCspSubService)) != CSSM_OK)
            return (retValue);
    }

    return (retValue);
}

/*---------------------------------------------------------------
 *
 *Name: mdsutil_FreeDLSubService
 *
 *Description:
 *   Free memory associated with the DLINFO structure
 *
 *Parameters: 
 *   Info (input) - pointer to DL Info
 *
 *Returns:
 *   CSSM_FAIL - unable to free
 *   CSSM_OK - free memory
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI mdsutil_FreeDLSubService (MDSUTIL_DLSUBSERVICE_PTR Sub)
{

    /* check to make sure that array is OK */
    if (IsBadReadPtr (Sub, sizeof (MDSUTIL_DLSUBSERVICE)))
        return (CSSMERR_CSSM_INVALID_POINTER);

    /* free every sub service */
    switch (Sub->Type) {
        case CSSM_DL_PKCS11 :
        case CSSM_DL_UNKNOWN:
        case CSSM_DL_CUSTOM:
        case CSSM_DL_LDAP:
        case CSSM_DL_ODBC:
        case CSSM_DL_FFS:
        case CSSM_DL_MEMORY:
        case CSSM_DL_REMOTEDIR:
            break;
        default:
            break;
    }

    if (Sub->RelOperatorTypes)
        MDSUTIL_free  (Sub->RelOperatorTypes, NULL);
    if (Sub->ConjOperatorTypes)
        MDSUTIL_free  (Sub->ConjOperatorTypes, NULL);

    if (Sub->NumberOfDataStores > 0)
    {
        uint32 j;

        if (IsBadReadPtr (Sub->DataStoreInfo, sizeof(CSSM_DBINFO) * Sub->NumberOfDataStores))
                return (CSSMERR_CSSM_INVALID_POINTER);

        if (IsBadReadPtr (Sub->DataStoreNames, sizeof(CSSM_NAME_LIST)))
            return (CSSMERR_CSSM_INVALID_POINTER);

        if (!IsBadReadPtr (Sub->DataStoreNames->String, sizeof(char *))) {
            for (j = 0; j < Sub->DataStoreNames->NumStrings; j++)
            {
                if (!IsBadReadPtr (Sub->DataStoreNames->String[j], sizeof(char *)))
                    MDSUTIL_free (Sub->DataStoreNames->String[j], NULL);
            }
            MDSUTIL_free (Sub->DataStoreNames->String, NULL);
        }

        for (j=0; j < Sub->NumberOfDataStores; j++) {

            mdsutil_FreeDbInfo (&Sub->DataStoreInfo[j]);
        }

        if (Sub->DataStoreNames)
           MDSUTIL_free  (Sub->DataStoreNames, NULL);

        if (Sub->DataStoreInfo)
            MDSUTIL_free  (Sub->DataStoreInfo, NULL);
    }

    return CSSM_OK;
}

/*---------------------------------------------------------------
 *
 *Name: mdsutil_FreeDBAttributeInfo
 *
 *Description:
 *   Free memory associated with the DB_ATTRIBUTE structure
 *
 *Parameters: 
 *   Info (input) - pointer to DB_ATTRIBUTE Info
 *
 *Returns:
 *   CSSM_FAIL - unable to free
 *   CSSM_OK - free memory
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI mdsutil_FreeDbAttributeInfo(CSSM_DB_ATTRIBUTE_INFO_PTR Info)
{

    if (Info->AttributeNameFormat == CSSM_DB_ATTRIBUTE_NAME_AS_STRING)
    {
        if (Info->Label.AttributeName != NULL)
            MDSUTIL_free(Info->Label.AttributeName, NULL);
    }
    else
    {
        if (Info->Label.AttributeOID.Data != NULL)
            MDSUTIL_free(Info->Label.AttributeOID.Data, NULL);
    }

    return (CSSM_OK);
}

/*---------------------------------------------------------------
 *
 *Name: mdsutil_FreeDBInfo
 *
 *Description:
 *   Free memory associated with the DBINFO structure
 *
 *Parameters: 
 *   Info (input) - pointer to DB Info
 *
 *Returns:
 *   CSSM_FAIL - unable to free
 *   CSSM_OK - free memory
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI mdsutil_FreeDbInfo (CSSM_DBINFO_PTR Info)
{
    uint32 i;

    if (IsBadReadPtr (Info->RecordAttributeNames, sizeof (CSSM_DB_RECORD_ATTRIBUTE_INFO)
            * Info->NumberOfRecordTypes))
        return(CSSMERR_CSSM_INVALID_POINTER);

    if (IsBadReadPtr (Info->RecordIndexes, sizeof (CSSM_DB_RECORD_INDEX_INFO)
            * Info->NumberOfRecordTypes))
        return(CSSMERR_CSSM_INVALID_POINTER);

    /* free all record types */
    for (i = 0; i < Info->NumberOfRecordTypes; i++)
    {
        uint32 j;
        if (Info->RecordAttributeNames[i].AttributeInfo)
        {
            if (IsBadReadPtr (Info->RecordAttributeNames[i].AttributeInfo, sizeof 
                        (CSSM_DB_ATTRIBUTE_INFO) * Info->RecordAttributeNames[i].NumberOfAttributes))
                return (CSSMERR_CSSM_INVALID_POINTER);

            for (j = 0; j < Info->RecordAttributeNames[i].NumberOfAttributes; j++)
            {
                mdsutil_FreeDbAttributeInfo(&Info->RecordAttributeNames[i].AttributeInfo[j]);
            }
            MDSUTIL_free (Info->RecordAttributeNames[i].AttributeInfo, NULL);
        }
        
        if (Info->RecordIndexes[i].IndexInfo)
        {
            if (IsBadReadPtr (Info->RecordIndexes[i].IndexInfo, sizeof 
                        (CSSM_DB_INDEX_INFO) * Info->RecordIndexes[i].NumberOfIndexes))
                return (CSSMERR_CSSM_INVALID_POINTER);

            for (j = 0; j < Info->RecordIndexes[i].NumberOfIndexes; j++)
            {
                mdsutil_FreeDbAttributeInfo(&Info->RecordIndexes[i].IndexInfo[j].Info);
            }

            MDSUTIL_free (Info->RecordIndexes[i].IndexInfo, NULL);
            
        }
    }

    if (Info->DefaultParsingModules)
        MDSUTIL_free (Info->DefaultParsingModules, NULL);
    if (Info->RecordAttributeNames)
        MDSUTIL_free (Info->RecordAttributeNames, NULL);
    if (Info->RecordIndexes)
        MDSUTIL_free (Info->RecordIndexes, NULL);
    if (Info->AccessPath)
        MDSUTIL_free (Info->AccessPath, NULL);
    if (Info->Reserved && ((CSSM_DATA_PTR)Info->Reserved)->Data)
        MDSUTIL_free (((CSSM_DATA_PTR)Info->Reserved)->Data, NULL);

    return(CSSM_OK);
}

/*---------------------------------------------------------------
 *
 *Name:mdsutil_FreeCLWrapped
 *
 *Description:
 *   Free memory associated with the CL_WRAPPEDPRODUCT structure
 *
 *Parameters: 
 *   Info (input) - pointer to CL_WRAPPEDPRODUCT Info
 *
 *Returns:
 *   CSSM_FAIL - unable to free
 *   CSSM_OK - free memory
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI mdsutil_FreeCLWrapped(MDSUTIL_CL_WRAPPEDPRODUCTINFO_PTR Wrapped)
{

    if (Wrapped->NumberOfEncoderProducts > 0)
    {
        if (IsBadReadPtr (Wrapped->EmbeddedEncoderProducts, sizeof (MDSUTIL_CL_ENCODER_PRODUCTINFO) 
                * Wrapped->NumberOfEncoderProducts))
            return (CSSMERR_CSSM_INVALID_POINTER);
        
        MDSUTIL_free (Wrapped->EmbeddedEncoderProducts,NULL);
    }

    if (Wrapped->NumberOfCAProducts > 0)
    {
        uint32 i;

        MDSUTIL_CL_CA_PRODUCTINFO_PTR CA = Wrapped->AccessibleCAProducts;
        if (IsBadReadPtr(CA, sizeof(MDSUTIL_CL_CA_PRODUCTINFO) * Wrapped->NumberOfCAProducts))
            return (CSSMERR_CSSM_INVALID_POINTER);
        
        for (i = 0; i < Wrapped->NumberOfCAProducts; i++)
        {
            uint32 j;
            for (j = 0; j < CA[i].NumberOfCertClasses; j++)
            {
                if (CA[i].CertClasses[j].CACert.Length > 0)
                {
                    if (IsBadReadPtr(CA[i].CertClasses[j].CACert.Data, 
                                CA[i].CertClasses[j].CACert.Length))
                        return (CSSMERR_CSSM_INVALID_POINTER);
                    MDSUTIL_free (CA[i].CertClasses[j].CACert.Data, NULL);
                }
            }
            MDSUTIL_free (CA[i].CertClasses, NULL);
        }

        MDSUTIL_free (CA,NULL);
    }

    return (CSSM_OK);
}

/*---------------------------------------------------------------
 *
 *Name:mdsutil_FreeCLSubService
 *
 *Description:
 *   Free memory associated with the CLINFO structure
 *
 *Parameters: 
 *   Info (input) - pointer to CL Info
 *
 *Returns:
 *   CSSM_FAIL - unable to free
 *   CSSM_OK - free memory
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI mdsutil_FreeCLSubService (MDSUTIL_CLSUBSERVICE_PTR Sub)
{
    uint32 cnt;

    /* check to make sure that array is OK */
    if (IsBadReadPtr (Sub, sizeof (MDSUTIL_CLSUBSERVICE)))
        return (CSSMERR_CSSM_INVALID_POINTER);

    /* free every sub service */
    /* Free certTemplate memory */
    for (cnt = 0; cnt < Sub->NumberOfTemplateFields; cnt++) {
        CSSM_OID_PTR ptr = &Sub->CertTemplate[cnt];
    
        if (IsBadReadPtr (ptr, sizeof (CSSM_OID)))
            return (CSSMERR_CSSM_INVALID_POINTER);
        
        if (IsBadReadPtr (ptr->Data, ptr->Length))
            return (CSSMERR_CSSM_INVALID_POINTER);

        MDSUTIL_free (ptr->Data, NULL);
     }
        
    /* Free Array of OID memory */
    if (Sub->NumberOfTemplateFields > 0)
        if (!IsBadReadPtr (Sub->CertTemplate, sizeof (CSSM_OID) * Sub->NumberOfTemplateFields)) {
            MDSUTIL_free (Sub->CertTemplate, NULL);
        }

    /* Free certTranslationTypes memory */
    if (Sub->NumberOfTranslationTypes > 0)
        if (!IsBadReadPtr (Sub->CertTranslationTypes, sizeof (CSSM_CERT_TYPE) * Sub->NumberOfTranslationTypes)) {
            MDSUTIL_free (Sub->CertTranslationTypes, NULL);
        }

    /* free wrapped product */
    if (mdsutil_FreeCLWrapped(&Sub->WrappedProduct) != CSSM_OK)
        return CSSM_ERRCODE_MDS_ERROR;

    /* free bundle info */
    if ( Sub->BundleInfo )
    {
        MDSUTIL_free(Sub->BundleInfo, NULL);
        Sub->BundleInfo = NULL;
    }

    return CSSM_OK;
}


/*---------------------------------------------------------------
 *
 *Name: mdsutil_FreeModule
 *
 *Description:
 *   Internal function that frees module description structure.
 *
 *Parameters: 
 *   ModuleDescription (input) -  Description of addin module
 *
 *Returns:
 *   CSSM_FAIL - unable to free structure
 *   CSSM_OK - structure is freed
 *
 *-------------------------------------------------------------*/
CSSM_RETURN mdsutil_FreeModule (const MDSUTIL_MODULE_INFO_PTR ModuleDescription)
{
    uint32 i=0;
    CSSM_RETURN retValue = CSSM_OK;

    if (ModuleDescription == NULL || IsBadReadPtr(ModuleDescription, (uint32)sizeof(MDSUTIL_MODULE_INFO_PTR)))
        return (CSSMERR_CSSM_INVALID_POINTER);


    for (i = 0; i < ModuleDescription->NumberOfAppAuthenRootKeys; i++) 
    {
        if(ModuleDescription->AppAuthenRootKeys[i].Data != NULL)
            MDSUTIL_free (ModuleDescription->AppAuthenRootKeys[i].Data, NULL);
    }

    if(ModuleDescription->AppAuthenRootKeys != NULL)
        MDSUTIL_free (ModuleDescription->AppAuthenRootKeys, NULL);

    if (ModuleDescription->ServiceMask != CSSM_SERVICE_CSSM) 
    {
        MDSUTIL_SERVICE_INFO_PTR Service;

        if (ModuleDescription->ServiceList != NULL) 
        {
            Service = ModuleDescription->ServiceList;
            for (i=0; i<ModuleDescription->NumberOfServices; i++) 
            {
                switch (Service[i].Type)
                {
                    uint32 j;

                    case CSSM_SERVICE_TP:
                        for (j=0; j<Service[i].NumberOfSubServices; j++)
                        {
                            if( Service[i].SubServiceList[j].ModuleType == MODULE_STATIC_TYPE)
                            {
                                retValue = mdsutil_FreeTPSubService (
                                        (MDSUTIL_TPSUBSERVICE_PTR)Service[i].SubServiceList[j].SubServiceInfo);
                            }
                        }

                        if(Service[i].SubServiceList != NULL)
                        {
                            if(Service[i].SubServiceList->SubServiceInfo != NULL)
                            {
                                MDSUTIL_free (Service[i].SubServiceList->SubServiceInfo, NULL);
                                Service[i].SubServiceList->SubServiceInfo = NULL;
                            }
                            MDSUTIL_free (Service[i].SubServiceList,NULL);
                            Service[i].SubServiceList = NULL;
                        }
                    break;
        
                    case CSSM_SERVICE_CSP:
                        for (j=0; j<Service[i].NumberOfSubServices; j++)
                        {
                            if( Service[i].SubServiceList[j].ModuleType == MODULE_STATIC_TYPE)
                            {
                                retValue = mdsutil_FreeCSPSubService (
                                    (MDSUTIL_CSPSUBSERVICE_PTR)Service[i].SubServiceList[j].SubServiceInfo);
                            }
                        }

                        if(Service[i].SubServiceList != NULL)
                        {
                            if(Service[i].SubServiceList->SubServiceInfo != NULL)
                            {
                                MDSUTIL_free (Service[i].SubServiceList->SubServiceInfo, NULL);
                                Service[i].SubServiceList->SubServiceInfo = NULL;
                            }
                            MDSUTIL_free (Service[i].SubServiceList,NULL);
                            Service[i].SubServiceList = NULL;
                        }
                    break;
        
                    case CSSM_SERVICE_CL:
                        for (j=0; j<Service[i].NumberOfSubServices; j++)
                        {
                            if( Service[i].SubServiceList[j].ModuleType == MODULE_STATIC_TYPE)
                            {

                                retValue = mdsutil_FreeCLSubService (
                                        (MDSUTIL_CLSUBSERVICE_PTR)Service[i].SubServiceList[j].SubServiceInfo);
                            }
                        }

                        if(Service[i].SubServiceList != NULL)
                        {
                            if(Service[i].SubServiceList->SubServiceInfo != NULL)
                            {
                                MDSUTIL_free (Service[i].SubServiceList->SubServiceInfo, NULL);
                                Service[i].SubServiceList->SubServiceInfo = NULL;
                            }
                            MDSUTIL_free (Service[i].SubServiceList,NULL);
                            Service[i].SubServiceList = NULL;
                        }
                    break;
        
                    case CSSM_SERVICE_DL:
                        for (j=0; j<Service[i].NumberOfSubServices; j++)
                        {
                            if( Service[i].SubServiceList[j].ModuleType == MODULE_STATIC_TYPE)
                            {
                                retValue = mdsutil_FreeDLSubService (
                                        (MDSUTIL_DLSUBSERVICE_PTR)Service[i].SubServiceList[j].SubServiceInfo);
                            }
                        }

                        if(Service[i].SubServiceList != NULL)
                        {
                            if(Service[i].SubServiceList->SubServiceInfo != NULL)
                            {
                                MDSUTIL_free (Service[i].SubServiceList->SubServiceInfo, NULL);
                                Service[i].SubServiceList->SubServiceInfo = NULL;
                            }
                            MDSUTIL_free (Service[i].SubServiceList,NULL);
                            Service[i].SubServiceList = NULL;
                        }
                    break;
        
                    case CSSM_SERVICE_AC:
                        if(Service[i].SubServiceList != NULL)
                        {
                            if(Service[i].SubServiceList->SubServiceInfo != NULL)
                            {
                                MDSUTIL_free (Service[i].SubServiceList->SubServiceInfo, NULL);
                                Service[i].SubServiceList->SubServiceInfo = NULL;
                            }
                            MDSUTIL_free (Service[i].SubServiceList,NULL);
                            Service[i].SubServiceList = NULL;
                        }
                    break;
        
                    default:
                    break;
                }

                if (retValue != CSSM_OK)
                    return(retValue);
            }
            
            if(ModuleDescription->ServiceList != NULL)
                MDSUTIL_free (ModuleDescription->ServiceList, NULL);
        }
    }
    
    if(ModuleDescription != NULL)
        MDSUTIL_free (ModuleDescription, NULL);

    return(retValue);
}


/*---------------------------------------------------------------
 *
 *Name: MDSUTIL_FreeModuleInfo
 *
 *Description:
 *   Free module info of the addin
 *
 *Parameters: 
 *   ModuleInfo - data to be freed
 *
 *Returns:
 *   CSSM_FAIL - error in freeing the info
 *   CSSM_OK -  info is freed
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI MDSUTIL_FreeModuleInfo (MDSUTIL_MODULE_INFO_PTR ModuleInfo)
{
    return(mdsutil_FreeModule (ModuleInfo));
}

/*---------------------------------------------------------------
 *
 *Name: MDSUTIL_GetModuleManagerInfo
 *
 *Description:
 *  This function returns descriptive information about the elective module manager
 *  identified by the GUID or the service mask.
 *
 *Parameters: 
 *  GUID - GUID identifying the module manager.
 *  ServiceType - Unique servicemask identifying the module manager.
 *
 *Returns:
 *   NULL - unable to get the module manager information.
 *   pointer to the module manager info.
 *
 *-------------------------------------------------------------*/
CSSM_RETURN CSSMAPI MDSUTIL_GetModuleManagerInfo(const CSSM_GUID *ModuleGUID,
                                                 CSSM_SERVICE_MASK ServiceType,
                                                 MDSUTIL_MODULE_MANAGER_INFO_PTR *moduleManagerInfo)
{
    CSSM_RETURN retValue = CSSM_OK;

    if (moduleManagerInfo == NULL)
        return (CSSMERR_CSSM_INVALID_POINTER);

    *moduleManagerInfo = NULL;

    /* make sure that app as done a CSSM_Init */
    if ((retValue = mdsutil_CheckInit ()) != CSSM_OK)
        return(retValue);
    
    /* if the Guid is given then find the manager using guid*/
    if (!IsBadReadPtr(ModuleGUID, (uint32)sizeof(CSSM_GUID)))
         retValue = mdsutil_mds_get_modulemanagerInfo (ModuleGUID, ServiceType, moduleManagerInfo);
    else
        return(CSSMERR_CSSM_INVALID_GUID);

    return (retValue);
  
}

CSSM_RETURN CSSMAPI MDSUTIL_ListModuleManagers(CSSM_GUID_PTR *ModuleManagerGuids,
                                               uint32 *NumberOfModuleManagers)
{
    CSSM_RETURN retValue = CSSM_OK;

    /* make sure that app as done a CSSM_Init */
    if ((retValue = mdsutil_CheckInit ()) != CSSM_OK)
        return(retValue);

    /*
     * Check for valid paramerers 
     */
    if (ModuleManagerGuids == NULL || IsBadReadPtr(ModuleManagerGuids, (uint32)sizeof(CSSM_GUID_PTR)) )
        return (CSSMERR_CSSM_INVALID_POINTER);

    if (NumberOfModuleManagers == NULL || IsBadReadPtr(NumberOfModuleManagers, (uint32)sizeof(uint32)) )
        return (CSSMERR_CSSM_INVALID_POINTER);

    retValue = mdsutil_ListInstalledEmms(ModuleManagerGuids,NumberOfModuleManagers);
    return (retValue);
}
